#ifndef dgNorm2d_hh
#define dgNorm2d_hh

#include "toolbox.hh"
#include "operator.hh"
#include "integration.hh"
#include "vectorial.hh"

// **************************************************************** DGNorm2d **

/** Provides methods to calculate norms
    @author Christoph Winkelmann, 2003
*/
class DGNorm2d
{
 public:
  /** Calculates the norm defined by \c matrix of \c coeffs
      @param coeffs The coefficent vector
      @param matrix The matrix which defines the norm
  */
  static Real matrix(const concepts::Vector<Real>& coeffs,
                     concepts::Operator<Real>& matrix) {
    concepts::Vector<Real> temp(coeffs);
    matrix(coeffs, temp);
    return std::sqrt(coeffs*temp);    
  }
  /** Calculates the L2 norm of a mesh function defined by \c coeffs
      minus a general function given by an exact formula \c exactFrm
      evaluated at time \c time.
      @param coeffs The coefficent vector
      @param exactFrm The formula for the exact solution
      @param nQuadPts Number of Gauss quadrature points per dimension
      @param time Time to evaluate \c exactFrmX and \c exactFrmY at
  */
  static Real l2(const concepts::Vector<Real>& coeffs,
                 concepts::Formula& exactFrm, uint nQuadPts,
                 Real time = 0) {
    Real norm2 = 0;
    try {
      const linDG2D::SpaceP1& spc =
        dynamic_cast<const linDG2D::SpaceP1&>(coeffs.space());
      std::auto_ptr<linDG2D::SpaceP1::Scanner> elemSc(spc.scan());

      concepts::Quadrature<4> quad(nQuadPts);
      const Real* absc = quad.abscissas();
      const Real* weig = quad.weights();
      concepts::Array<Real> xi(nQuadPts);
      concepts::Array<Real> w(nQuadPts);
      for (uint iQuad=0; iQuad<nQuadPts; ++iQuad) {
        xi[iQuad] = 0.5*(1.+absc[iQuad]);
        w[iQuad] = 0.5*weig[iQuad];
      }

      while (*elemSc) {
        Real integral = 0;
        Real temp;
        const linDG2D::TriangleP1& elm = 
          dynamic_cast<const linDG2D::TriangleP1&>((*elemSc)++);
        for (uint iQuad1=0; iQuad1<nQuadPts; ++iQuad1) {
          for (uint iQuad2=0; iQuad2<nQuadPts; ++iQuad2) {
            temp = elm.solution(coeffs, xi[iQuad1], xi[iQuad2]*xi[iQuad1]) - 
              exactFrm(elm.cell().chi(xi[iQuad1], xi[iQuad2]), time);
            integral += temp * temp * w[iQuad1] * w[iQuad2] * xi[iQuad1];
          }
        }
        norm2 += integral * 2 * elm.cell().area();
      }
    } catch (std::bad_cast& e) {
      std::cerr << e.what() << std::endl;
      conceptsAssert(false, concepts::Assertion());
    }
    return std::sqrt(norm2);
  }
  /** Calculates the energy norm of a vectorial mesh function defined by \c
      coeffs minus a general function given by exact formulae for x
      (\c exactFrmX) and y (\c exactFrmY) component, both evaluated at time
      \c time. The definition of the energy norm is according to T. Wihler.
      @param coeffs The coefficent vector
      @param exactFrmX The formula for the x component of the exact solution
      @param exactFrmY The formula for the y component of the exact solution
      @param exactFrmXderX d/dx of \c exactFrmX
      @param exactFrmXderY d/dy of \c exactFrmX
      @param exactFrmYderX d/dx of \c exactFrmY
      @param exactFrmYderY d/dy of \c exactFrmY
      @param nQuadPts Number of Gauss quadrature points per dimension
      @param time Time to evaluate \c exactFrm at
  */
  static Real energyTW(const concepts::Vector<Real>& coeffs,
                       concepts::Formula& exactFrmX,
                       concepts::Formula& exactFrmY,
                       concepts::Formula& exactFrmXderX,
                       concepts::Formula& exactFrmXderY,
                       concepts::Formula& exactFrmYderX,
                       concepts::Formula& exactFrmYderY,
                       uint nQuadPts, Real time = 0) {
    Real norm2 = 0;
    try {
      const vectorial::Space<Real>& spc =
        dynamic_cast<const vectorial::Space<Real>&>(coeffs.space());
      std::auto_ptr<vectorial::Space<Real>::Scanner> elemSc(spc.scan());

      concepts::Quadrature<4> quad(nQuadPts);
      const Real* absc = quad.abscissas();
      const Real* weig = quad.weights();
      concepts::Array<Real> xi(nQuadPts);
      concepts::Array<Real> w(nQuadPts);
      for (uint iQuad=0; iQuad<nQuadPts; ++iQuad) {
        xi[iQuad] = 0.5*(1.+absc[iQuad]);
        w[iQuad] = 0.5*weig[iQuad];
      }

      while (*elemSc) {
        Real integral = 0;
        vectorial::Element<Real>& vElm = (*elemSc)++;
        const linDG2D::TriangleP1& elm =
          dynamic_cast<const linDG2D::TriangleP1&>(*vElm.get(0));
        
        concepts::Mapping<Real, 2> numSymmGrad = elmSymmGrad(vElm, coeffs);
        
        for (uint iQuad1=0; iQuad1<nQuadPts; ++iQuad1) {
          for (uint iQuad2=0; iQuad2<nQuadPts; ++iQuad2) {
            Real e11 = 
              exactFrmXderX(elm.cell().chi(xi[iQuad1], xi[iQuad2]), time);
            Real e12 = 0.5 * (
              exactFrmXderY(elm.cell().chi(xi[iQuad1], xi[iQuad2]), time) +
              exactFrmYderX(elm.cell().chi(xi[iQuad1], xi[iQuad2]), time));
            Real e22 =
              exactFrmYderY(elm.cell().chi(xi[iQuad1], xi[iQuad2]), time);
            concepts::Mapping<Real, 2> temp(-e11, -e12, -e12, -e22);
            temp += numSymmGrad;
            integral += mappingNorm<2>(temp) * 
              w[iQuad1] * w[iQuad2] * xi[iQuad1];
          }
        }
        norm2 += integral * 2 * elm.cell().area();
      }
      // here the integrals over edges have to be added
    } catch (std::bad_cast& e) {
      std::cerr << e.what() << std::endl;
      conceptsAssert(false, concepts::Assertion());
    }
    return std::sqrt(norm2);
  }
  /** Calculates the L2 norm of a vectorial mesh function defined by \c coeffs
      minus a general function given by exact formulae for x (\c exactFrmX)
      and y (\c exactFrmY) component, both evaluated at time \c time.
      @param coeffs The coefficent vector
      @param exactFrmX The formula for the x component of the exact solution
      @param exactFrmY The formula for the y component of the exact solution
      @param nQuadPts Number of Gauss quadrature points per dimension
      @param time Time to evaluate \c exactFrm at
  */
  static Real l2(const concepts::Vector<Real>& coeffs,
                 concepts::Formula& exactFrmX, concepts::Formula& exactFrmY,
                 uint nQuadPts, Real time = 0) {
    Real norm2 = 0;
    try {
      const vectorial::Space<Real>& spc =
        dynamic_cast<const vectorial::Space<Real>&>(coeffs.space());
      std::auto_ptr<vectorial::Space<Real>::Scanner> elemSc(spc.scan());

      concepts::Quadrature<4> quad(nQuadPts);
      const Real* absc = quad.abscissas();
      const Real* weig = quad.weights();
      concepts::Array<Real> xi(nQuadPts);
      concepts::Array<Real> w(nQuadPts);
      for (uint iQuad=0; iQuad<nQuadPts; ++iQuad) {
        xi[iQuad] = 0.5*(1.+absc[iQuad]);
        w[iQuad] = 0.5*weig[iQuad];
      }

      while (*elemSc) {
        Real integral = 0;
        concepts::Real2d temp;
        vectorial::Element<Real>& vElm = (*elemSc)++;
        const linDG2D::TriangleP1& elm =
          dynamic_cast<const linDG2D::TriangleP1&>(*vElm.get(0));
        for (uint iQuad1=0; iQuad1<nQuadPts; ++iQuad1) {
          for (uint iQuad2=0; iQuad2<nQuadPts; ++iQuad2) {
            temp[0] = exactFrmX(elm.cell().chi(xi[iQuad1], xi[iQuad2]), time);
            temp[1] = exactFrmY(elm.cell().chi(xi[iQuad1], xi[iQuad2]), time);
            temp -= elmSol(vElm, coeffs, xi[iQuad1], xi[iQuad2]*xi[iQuad1]);
            integral += temp * temp * w[iQuad1] * w[iQuad2] * xi[iQuad1];
          }
        }
        norm2 += integral * 2 * elm.cell().area();
      }
    } catch (std::bad_cast& e) {
      std::cerr << e.what() << std::endl;
      conceptsAssert(false, concepts::Assertion());
    }
    return std::sqrt(norm2);
  }
  // Calculate symmetric gradient of solution on vectorial element
  static concepts::Mapping<Real, 2> elmSymmGrad
  (vectorial::Element<Real>& vElm, const concepts::Vector<Real>& coeffs) {
    concepts::Mapping<Real, 2> sol;
    try {
      const linDG2D::TriangleP1& elm =
        dynamic_cast<const linDG2D::TriangleP1&>(*vElm.get(0));
      const uint nShapeFct = elm.T().m();
      concepts::Array<Real> localCoeffs(nShapeFct);
      concepts::Array<concepts::Real2d> gradArray(2);
      for (uint iDim=0; iDim<2; ++iDim) {
        vectorial::Element<Real>::vecGraphics()->choose(iDim);
        vElm.T().extract(coeffs, localCoeffs);
        for (int iShapeFct=0; iShapeFct<3; ++iShapeFct) {
          gradArray[iDim] +=
            elm.shapeFctGradient(iShapeFct) * localCoeffs[iShapeFct];
        }
      }
      concepts::Mapping<Real, 2> grad(gradArray[0], gradArray[1]);
      grad += grad.transpose();
      grad *= 0.5;
      sol = grad;
    } catch (std::bad_cast& e) {
      std::cerr << e.what() << std::endl;
      conceptsAssert(false, concepts::Assertion());
    }
    return sol;
  }
  private:
  /// Constructor. Is private in order to prevent creating an instance.
  DGNorm2d() {};
  /// Calculate solution on vectorial element
  static concepts::Real2d elmSol(vectorial::Element<Real>& vElm,
                                 const concepts::Vector<Real>& coeffs,
                                 Real xi1, Real xi2) {
    concepts::Real2d sol(0, 0);
    try {
      const linDG2D::TriangleP1& elm =
        dynamic_cast<const linDG2D::TriangleP1&>(*vElm.get(0));
      const uint nShapeFct = elm.T().m();
      concepts::Array<Real> localCoeffs(nShapeFct);
      for (uint iDim=0; iDim<2; ++iDim) {
        vectorial::Element<Real>::vecGraphics()->choose(iDim);
        vElm.T().extract(coeffs, localCoeffs);
        for (int iShapeFct=0; iShapeFct<3; ++iShapeFct) {
          sol[iDim] += localCoeffs[iShapeFct] *
            elm.shapeFct(iShapeFct, xi1, xi2);
        }
      } 
    } catch (std::bad_cast& e) {
      std::cerr << e.what() << std::endl;
      conceptsAssert(false, concepts::Assertion());
    }
    return sol;
  }
  /// Calculate norm of a mapping
  template <uint dim>
  static Real mappingNorm(concepts::Mapping<Real, dim>& map) {
    Real norm2 = 0;
    for (uint i=0; i<dim; ++i) {
      for (uint j=0; j<dim; ++j) {
        norm2 += map(i,j)*map(i,j);
      }
    }
    return std::sqrt(norm2);
  }
};

#endif // dgNorm2d_hh
