#include <iostream>
#include <stdlib.h>

#include "geometry/meshImport2D.hh"
#include "geometry/uniformlyRefinedMesh2.hh"
#include "geometry/boundaryConditions.hh"
#include "operator.hh"
#include "timestepping.hh"
#include "vectorial.hh"
#include "graphics/dataDX.hh"
#include "eigensolver.hh"
#include "basics/fortranException.hh"
#include "toolbox/resourceMonitor.hh"
#include "toolbox/formulaLiCo.hh"
#include "linDG2D.hh"

#if defined(UMFPACK_41)
#include "operator/umfpack.hh"
#elif defined(HAS_SuperLU)
#include "operator/superLU.hh"
#endif

#include "linearForm.hh"
#include "bilinearForm.hh"
#include "elasticityTensor.hh"
#include "../app-fvdgfem/formula.hh"
#include "../app-fvdgfem/pslogger.hh"
#include "dgNorm2d.hh"
#include "equivalentStress.hh"
#include "periodicSquare.hh"

#define DEBUG_TEST 0

using concepts::Real;
using std::auto_ptr;

// ************************************************************* writeOutput **
void writeOutput(const concepts::InOutParameters& input,
                 const concepts::InOutParameters& output) {
  std::string outFileName = input.getString("outFileName");
  std::ofstream outFile(outFileName.c_str());
  outFile << output;
}

// ************************************************************ solverFabric **
concepts::SolverFabric<Real>* solverFabric
(const concepts::InOutParameters& input, const uint spcDim) {
  bool directSolver = input.getBool("directSolver");
  if (directSolver) {
#if defined(UMFPACK_41)
    return new concepts::UmfpackFabric();
#elif defined(HAS_SuperLU)
    return new concepts::SuperLUFabric();
#else
#error either Umfpack or SuperLU needed
#endif
  } else {
    Real tol = input.getDouble("tolLinSys");
    Real tau = input.getDouble("tau");
#if defined(HAS_PETSC)
    std::string ksptype;
    std::string pctype;
    if(tau == -1) {
      ksptype = input.getString("symmSolver");
      pctype = input.getString("symmPc");
    } else {
      ksptype = input.getString("asymmSolver");
      pctype = input.getString("asymmPc");
    }
    return new concepts::PETScFabric(tol, ksptype, pctype);
#else
    uint iter = input.getInt("maxIterLinSys");
    if (iter < 10) {
      iter = iter * spcDim + 100;
    }
    if (tau == -1) {
      return new concepts::CGFabric<Real>(tol, iter, true);
    } else {
      uint restart = input.getInt("restartLinSys");
      return new concepts::GMResFabric<Real>(tol, iter, restart, true);
    }
#endif
  }
}

// ******************************************************** solverIterations **
uint solverIterations(const concepts::Operator<Real>& solver) {
  const concepts::CG<Real>* cg = 
    dynamic_cast<const concepts::CG<Real>*>(&solver);
  if (cg) {
    return cg->iterations();
  }
  const concepts::GMRes<Real>* gmres = 
    dynamic_cast<const concepts::GMRes<Real>*>(&solver);
  if (gmres) {
    return gmres->iterations();
  }
  const concepts::PETSc* petsc = 
    dynamic_cast<const concepts::PETSc*>(&solver);
  if (petsc) {
    return petsc->iterations();
  }
  return 1;
}

// ****************************************************************** TestBf **
class TestBf : public concepts::BilinearForm<Real> {
 public:
  TestBf(uint iV, uint iU) : iV_(iV), iU_(iU) {}
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmU,
                          concepts::ElementMatrix<Real>& em) {
    const uint mV = elmV.T().m();
    const uint mU = elmU.T().m();
    em.resize(mV, mU);
    em.zeros(); 
    
    if (&elmV == &elmU) {
      conceptsAssert(iV_ < mV, concepts::Assertion());
      conceptsAssert(iU_ < mU, concepts::Assertion());
      em(iV_,iU_) = 1;
    }
  }
 private:
  const uint iV_;
  const uint iU_;
};

// ****************************************************************** LExact **
// exact solution in l shaped domain
class LExact : public concepts::Formula {
 public:
  LExact(Real mu, Real lambda, uint vDim) 
    : mu_(mu), lambda_(lambda), vDim_(vDim) {
      if (alpha_ == 0.) {
        alpha_ = 0.5;
        Real alphaOld = 0;
        Real eps = std::numeric_limits<Real>::epsilon();
        while (std::fabs(alphaOld-alpha_) > 2*eps) {
          alphaOld = alpha_;
          alpha_ = (1. - std::asin(alpha_)/pi_) * 2./3.;
        }
      }
      c1_ = -std::cos((alpha_+1)*omega_) / std::cos((alpha_-1)*omega_);
      c2_ = 2. * (lambda_ + 2.*mu_) / (lambda_ + mu_);
    }
  concepts::Formula* clone() const { return new LExact(mu_, lambda_, vDim_); }
  Real operator() (const Real x, const Real t = 0.) const { return 0.; }
  Real operator() (const concepts::Real2d& x, const Real t = 0.) const {
    Real r = x.l2();
    Real theta = std::atan2(x[1], x[0]);
    Real uR = std::pow(r,alpha_) *
      (-(alpha_+1.)*std::cos((alpha_+1.)*theta)
       +(c2_-(alpha_+1.))*c1_*std::cos((alpha_-1)*theta)) / (2.*mu_);
    Real uTheta = std::pow(r,alpha_) *
      ((alpha_+1.)*std::sin((alpha_+1.)*theta)
       +(c2_+alpha_-1.)*c1_*std::sin((alpha_-1)*theta)) / (2.*mu_);
    switch (vDim_) {
    case 0:
      return uR*std::cos(theta) - uTheta*std::sin(theta);
      break;
    case 1:
      return uR*std::sin(theta) + uTheta*std::cos(theta);
      break;
    }
    return 0;
  }
  Real operator() (const concepts::Real3d& x, const Real t = 0.) const {
    return 0.;
  }
 private:
  const Real mu_;
  const Real lambda_;
  uint vDim_;

  Real c1_;
  Real c2_;

  static Real pi_;
  static Real omega_;
  static Real alpha_;
};

Real LExact::pi_ = std::acos(-1.);
Real LExact::omega_ = pi_ * 3./4.;
Real LExact::alpha_ = 0.;

// ************************************************************ TimeExactSol **
// exact solution for time dependent elasticity problem
class TimeExactSol : public concepts::Formula {
 public:
  TimeExactSol(const concepts::Formula& initial, Real omega) 
    : initial_(initial), omega_(omega) { }
  concepts::Formula* clone() const {
    return new TimeExactSol(initial_, omega_);
  }
  Real operator() (const Real x, const Real t = 0.) const { return 0.; }
  Real operator() (const concepts::Real2d& x, const Real t = 0.) const {
    return std::cos(omega_*t)*initial_(x);
  }
  Real operator() (const concepts::Real3d& x, const Real t = 0.) const {
    return 0.;
  }
 private:
  const concepts::Formula& initial_;
  Real omega_;
};

// ************************************************************ TimeExactRhs **
// exact right hand side for time dependent elasticity problem
class TimeExactRhs : public concepts::Formula {
 public:
  TimeExactRhs(const concepts::Formula& initial,
            const concepts::Formula& flux, Real omega) 
    : initial_(initial), flux_(flux), omega_(omega) { }
  concepts::Formula* clone() const {
    return new TimeExactRhs(initial_, flux_, omega_);
  }
  Real operator() (const Real x, const Real t = 0.) const { return 0.; }
  Real operator() (const concepts::Real2d& x, const Real t = 0.) const {
    return std::cos(omega_*t)*(-omega_*omega_*initial_(x)+flux_(x));
  }
  Real operator() (const concepts::Real3d& x, const Real t = 0.) const {
    return 0.;
  }
 private:
  const concepts::Formula& initial_;
  const concepts::Formula& flux_;
  Real omega_;
};

// ************************************************************ testMappings **
void testMappings(const concepts::InOutParameters& input,
                  concepts::InOutParameters& output) {
  // data
  uint level = input.getInt("level");
  uint iV = input.getInt("iV");
  uint iU = input.getInt("iU");
  uint jV = input.getInt("jV");
  uint jU = input.getInt("jU");
    
  // mesh
  concepts::Import2dMesh coarseMesh(input.getString("meshCoord"),
                                    input.getString("meshElem"),
                                    input.getString("meshBnd"),
                                    input.getInt("idxStart"));
  concepts::UniformlyRefinedMesh2 mesh(coarseMesh, level);

  // space
  linDG2D::SpaceP1 spc1(mesh);
  const uint vdim = 2;
  vectorial::Space<Real> spc(vdim);
  spc.put(spc1);
  spc.put(spc1);

  // mass matrix
  TestBf testBf(iV, iU);
  vectorial::BilinearForm<Real> testBfv(vdim);
  testBfv.put(testBf, jV, jU);
  concepts::DenseMatrix<Real> mass(spc, spc);
  concepts::Matrix<Real>::assembly(mass, spc, testBfv);

  std::cout << "MC = " << mass << std::endl;
}

// ************************************************************ testMatrices **
void testMatrices(const concepts::InOutParameters& input,
                  concepts::InOutParameters& output) {
  // data
  Real tau = input.getDouble("tau");
  Real gammaMu = input.getDouble("gammaMu");
  Real gammaLambda = input.getDouble("gammaLambda");
  concepts::ParsedFormula boundaryFrmX(input.getString("bndFrmX").c_str());
  concepts::ParsedFormula boundaryFrmY(input.getString("bndFrmY").c_str());
  uint level = input.getInt("minLevel");
  Real mu = input.getDouble("mu");
  Real lambda = input.getDouble("minLambda");

  // mesh
  concepts::Import2dMesh coarseMesh(input.getString("meshCoord"),
                                    input.getString("meshElem"),
                                    input.getString("meshBnd"),
                                    input.getInt("idxStart"));
  concepts::UniformlyRefinedMesh2 mesh(coarseMesh, level);
  linDG2D::MeshInfo mi(mesh);

  // boundary
  concepts::Array<auto_ptr<concepts::BoundaryConditions> > bc(2);
  bc[0].reset(new concepts::BoundaryConditions());
  bc[0]->add(concepts::Attribute(1),
             concepts::Boundary(concepts::Boundary::DIRICHLET,
                                boundaryFrmX));
  bc[1].reset(new concepts::BoundaryConditions());
  bc[1]->add(concepts::Attribute(1),
             concepts::Boundary(concepts::Boundary::DIRICHLET,
                                boundaryFrmY));

  // space
  linDG2D::SpaceP1 spc1(mesh);
  linDG2D::ElementPairList epl11(spc1, spc1, mi);

  const uint vdim = 2;
  vectorial::Space<Real> spcU(vdim);
  spcU.put(spc1);
  spcU.put(spc1);

  vectorial::ElementPairListIdentic<Real> eplU(epl11, spcU);

  // mass matrix
  linDG2D::IdentityBf idBf1;
  vectorial::BilinearForm<Real> idBf(vdim);
  for (uint i=0; i<vdim; ++i) {
    idBf.put(idBf1, i, i);
  }
  concepts::DenseMatrix<Real> mass(spcU, spcU);
  concepts::Matrix<Real>::assembly(mass, spcU, idBf);

  // stiffness matrices
  ElasticityTensorIsotropic<vdim> et(mu, lambda);

  concepts::Array<auto_ptr<DGElasticVolBf2d> > elaVolBf1(vdim*vdim);
  vectorial::BilinearForm<Real> elaVolBf(vdim);
  for (uint k=0; k<vdim*vdim; ++k) {
    uint iV = (uint)k/vdim;
    uint iU = k%vdim;
    elaVolBf1[k].reset(new DGElasticVolBf2d(et(iV,iU)));
    elaVolBf.put(*elaVolBf1[k], iV, iU);
  }

  concepts::DenseMatrix<Real> stiff_vol(spcU, spcU);
  concepts::Matrix<Real>::assembly(stiff_vol, spcU, elaVolBf);

  concepts::Array<auto_ptr<DGElasticFluxBf2d> > elaFluxBf1(vdim*vdim);
  vectorial::BilinearForm<Real> elaFluxBf(vdim);
  for (uint k=0; k<vdim*vdim; ++k) {
    uint iV = (uint)k/vdim;
    uint iU = k%vdim;
    elaFluxBf1[k].reset(new DGElasticFluxBf2d(et(iV,iU), bc[0].get()));
    elaFluxBf.put(*elaFluxBf1[k], iV, iU);
  }
  concepts::DenseMatrix<Real> stiff_flux_one(spcU, spcU);
  concepts::Matrix<Real>::assembly(stiff_flux_one, elaFluxBf, eplU);
  concepts::DenseMatrix<Real> stiff_flux_oneT(stiff_flux_one, true);
  concepts::LiCo<Real> temp(stiff_flux_one, stiff_flux_oneT, -1, tau);
  concepts::DenseMatrix<Real> stiff_flux(temp);

  DGElasticStabBf2d elaStabBf1(gammaMu*mu, bc[0].get());
  vectorial::BilinearForm<Real> elaStabBf(vdim);
  for (uint i=0; i<vdim; ++i) {
    elaStabBf.put(elaStabBf1, i, i);
  }
  concepts::DenseMatrix<Real> stiff_stab(spcU, spcU);
  concepts::Matrix<Real>::assembly(stiff_stab, elaStabBf, eplU);

  concepts::Array<auto_ptr<DGElasticNormalStabBf2d> >
    elaNormalStabBf1(vdim*vdim);
  vectorial::BilinearForm<Real> elaNormalStabBf(vdim);
  for (uint k=0; k<vdim*vdim; ++k) {
    uint iV = (uint)k/vdim;
    uint iU = k%vdim;
    elaNormalStabBf1[k].reset
      (new DGElasticNormalStabBf2d(lambda*gammaLambda, iV, iU, bc[0].get()));
    elaNormalStabBf.put(*elaNormalStabBf1[k], iV, iU);
  }
  concepts::DenseMatrix<Real> stiff_normalStab(spcU, spcU);
  concepts::Matrix<Real>::assembly(stiff_normalStab, elaNormalStabBf, eplU);

  // output

  std::cout << "AvC = " << stiff_vol << std::endl;
  std::cout << "AfC1 = " << stiff_flux_one << std::endl;
  std::cout << "AfC = " << stiff_flux << std::endl;
  std::cout << "AsC = " << stiff_stab << std::endl;
  std::cout << "AnsC = " << stiff_normalStab << std::endl;
  std::cout << "MC = " << mass << std::endl;
}

// ***************************************************************** testRun **
void testRun(const concepts::InOutParameters& input,
             concepts::InOutParameters& output) {
  // logging
  PsLogger logger(input.getString("logFileName"),
                  input.getString("loggerData"));
  logger.log("S beginning");
  concepts::ResourceMonitor rsm;

  // data
  bool runEgvl = input.getBool("runEgvl");
  bool runStatic = input.getBool("runStatic");
  bool runDynamic = input.getBool("runDynamic");
  if (!runEgvl && !runStatic & !runDynamic) {
    return;
  }

  Real tau = input.getDouble("tau");
  Real gammaMu = input.getDouble("gammaMu");
  Real gammaLambda = input.getDouble("gammaLambda");

  uint maxLevel = input.getInt("maxLevel");
  uint minLevel = input.getInt("minLevel");
  Real mu = input.getDouble("mu");
  Real maxLambda = input.getDouble("maxLambda");
  Real minLambda = input.getDouble("minLambda");
  uint nQuadPts = input.getInt("nQuadPts");

  bool dxOutput = input.getBool("dxOutput");
  uint verbosity = input.getInt("verbosity");

  const uint vdim = 2;

  bool periodicMesh = input.getBool("periodicMesh");

  // output
  output.addArrayDouble("nDoF");
  output.addArrayDouble("lambda");
  output.addArrayDouble("utime_level_init");
  output.addArrayDouble("utime_lambda_init");

  uint iArray = 0;
  uint iArrayDyn = 0;

  if (runStatic) {
    output.addArrayDouble("L2_error");
    output.addArrayDouble("max_equiv_stress_principal");
    output.addArrayDouble("max_equiv_stress_shear");
    output.addArrayDouble("max_equiv_stress_energy");
    output.addArrayDouble("iterLinSys");
    output.addArrayDouble("utime_static_init");
    output.addArrayDouble("utime_static_calc");
    output.addArrayDouble("utime_static_output");
  }

  if (runEgvl) {
    uint nEgvl = input.getInt("nEgvl");

    std::stringstream arrayName;
    for (uint iEgvl=0; iEgvl<nEgvl; ++iEgvl) {
      arrayName << "eigenvalueR_" << iEgvl;
      output.addArrayDouble(arrayName.str().c_str());
      arrayName.str("");
    }
    if (tau != -1) {
      for (uint iEgvl=0; iEgvl<nEgvl; ++iEgvl) {
        arrayName << "eigenvalueI_" << iEgvl;
        output.addArrayDouble(arrayName.str().c_str());
        arrayName.str("");
      }
    }
    output.addArrayInt("iterEgvl");
    output.addArrayInt("converged");
    output.addArrayDouble("utime_egvl_initCalc");
    output.addArrayDouble("utime_egvl_output");
  }

  if (runDynamic) {
    // array output
    output.addArrayInt("arrayOffsetDyn");
    output.addArrayDouble("time");
    output.addArrayDouble("M_normDyn");
    output.addArrayDouble("A_normDyn");
    output.addArrayDouble("max_normDyn");
    output.addArrayDouble("L2_errorDyn");
    // output.addArrayInt("iterLinSysDyn");
    output.addArrayDouble("utime_dyn_init");
    output.addArrayDouble("utime_dyn_calcOutput");
    output.addArrayInt("nsteps");
    output.addArrayDouble("dt_omega");
    output.addArrayDouble("dt_input");
    output.addArrayDouble("dt");
    bool implicit = input.getBool("implicit");
    if (!implicit) {
      output.addArrayDouble("dt_cfl");
      output.addArrayDouble("eigenvalue_max");
    }
  }

  // mesh
  auto_ptr<concepts::Mesh2> coarseMesh;
  if (!periodicMesh) {
    coarseMesh.reset(new concepts::Import2dMesh(input.getString("meshCoord"),
                                                input.getString("meshElem"),
                                                input.getString("meshBnd"),
                                                input.getInt("idxStart")));
  }

  logger.log("S_1 global_initialization");
  output.addDouble("utime_init_global", rsm.utime());
  writeOutput(input, output);
  std::stringstream logMsg;

  uint iLevel=0;
  for (uint level=minLevel; level<=maxLevel; ++level, ++iLevel) {

    if (verbosity > 0) {
      std::cout << "level = " << level << std::endl;
    }

    // refined mesh
    auto_ptr<concepts::Mesh2> mesh;
    if (periodicMesh) {
      uint n = 1 << level;
      Real jitter = input.getDouble("meshJitter");
      mesh.reset(new PeriodicSquare(n, jitter));
    } else {
      mesh.reset(new concepts::UniformlyRefinedMesh2(*coarseMesh, level));
    }
    linDG2D::MeshInfo mi(*mesh);

    // spaces
    linDG2D::SpaceP1 spc1(*mesh);
    linDG2D::ElementPairList epl11(spc1, spc1, mi);

    vectorial::Space<Real> spcU(vdim);
    for (uint i=0; i<vdim; ++i) {
      spcU.put(spc1);
    }
    vectorial::ElementPairListIdentic<Real> eplU(epl11, spcU);
    
    graphics::DataDX<Real> dxX
      (graphics::drawDataDX(spcU, input.getString("dxFileNameX"), 2));
    graphics::DataDX<Real> dxY
      (graphics::drawDataDX(spcU, input.getString("dxFileNameY"), 2));
    graphics::DataDX<Real> dxStress
      (graphics::drawDataDX(spc1, input.getString("dxFileNameStress"), 2));

    logMsg << "S_1_" << iLevel << " level_initialization";
    logger.log(logMsg.str());
    logMsg.str("");
    output.addArrayDouble("utime_level_init", iLevel, rsm.utime());
    writeOutput(input, output);

    uint iLambda=0;
    for (Real lambda=minLambda; lambda<=maxLambda; lambda*=10, ++iLambda) {

      if (verbosity > 0) {
        std::cout << "  lambda = " << lambda << std::endl;
      }

      // exact solution
      concepts::ParsedFormula exactFrmX(input.getString("exactFrmX").c_str());
      concepts::ParsedFormula exactFrmY(input.getString("exactFrmY").c_str());

      // LExact exactFrmX(mu, lambda, 0);
      // LExact exactFrmY(mu, lambda, 1);

      concepts::Formula& boundaryFrmX = exactFrmX;
      concepts::Formula& boundaryFrmY = exactFrmY;

      concepts::ParsedFormula
        neumannFrmX(input.getString("neumannFrmX").c_str());
      concepts::ParsedFormula
        neumannFrmY(input.getString("neumannFrmY").c_str());

      // source term
      concepts::ParsedFormula
        sourceFrmMuX(input.getString("sourceFrmMuX").c_str());
      concepts::ParsedFormula
        sourceFrmMuY(input.getString("sourceFrmMuY").c_str());
      concepts::ParsedFormula
        sourceFrmLambdaX(input.getString("sourceFrmLambdaX").c_str());
      concepts::ParsedFormula
        sourceFrmLambdaY(input.getString("sourceFrmLambdaY").c_str());

      concepts::Array<auto_ptr<concepts::Formula> > sourceFrm(2);
      sourceFrm[0].reset(new concepts::FormulaLiCo(sourceFrmMuX,
                                                   sourceFrmLambdaX, mu,
                                                   lambda));
      sourceFrm[1].reset(new concepts::FormulaLiCo(sourceFrmMuY,
                                                   sourceFrmLambdaY, mu,
                                                   lambda));
      concepts::Array<concepts::Formula*> sourceFrmP(2);
      sourceFrmP[0] = sourceFrm[0].get();
      sourceFrmP[1] = sourceFrm[1].get();

      // boundary
      concepts::Array<auto_ptr<concepts::BoundaryConditions> > bc(vdim);
      bc[0].reset(new concepts::BoundaryConditions());
      bc[0]->add(concepts::Attribute(1),
                 concepts::Boundary(concepts::Boundary::DIRICHLET,
                                    boundaryFrmX));
      bc[0]->add(concepts::Attribute(2),
                 concepts::Boundary(concepts::Boundary::NEUMANN,
                                    neumannFrmX));
      bc[1].reset(new concepts::BoundaryConditions());
      bc[1]->add(concepts::Attribute(1),
                 concepts::Boundary(concepts::Boundary::DIRICHLET,
                                    boundaryFrmY));
      bc[1]->add(concepts::Attribute(2),
                 concepts::Boundary(concepts::Boundary::NEUMANN,
                                    neumannFrmY));
      concepts::Array<concepts::BoundaryConditions*> bcp(2);
      bcp[0] = bc[0].get();
      bcp[1] = bc[1].get();

      // stiffness matrix
      ElasticityTensorIsotropic<vdim> et(mu, lambda);

      // stiffness matrix: flux contribution
      concepts::Array<auto_ptr<DGElasticFluxBf2d> >
        elaFluxBf1(vdim*vdim);
      vectorial::BilinearForm<Real> elaFluxBf(vdim);
      for (uint k=0; k<vdim*vdim; ++k) {
        uint iV = (uint)k/vdim;
        uint iU = k%vdim;
        elaFluxBf1[k].reset(new DGElasticFluxBf2d(et(iV,iU), bc[0].get()));
        elaFluxBf.put(*elaFluxBf1[k], iV, iU);
      }
      concepts::SparseMatrix<Real> stiff_flux(spcU, spcU);
      concepts::Matrix<Real>::assembly(stiff_flux, elaFluxBf, eplU);
      concepts::SparseMatrix<Real> stiff_fluxT(stiff_flux, true);
      concepts::LiCo<Real> temp(stiff_flux, stiff_fluxT, -1, tau);
      concepts::SparseMatrix<Real> stiff(temp);
      if (tau == -1) {
        stiff.symmetrize();
      }

      // stiffness matrix: volume contribution
      concepts::Array<auto_ptr<DGElasticVolBf2d> > elaVolBf1(vdim*vdim);
      vectorial::BilinearForm<Real> elaVolBf(vdim);
      for (uint k=0; k<vdim*vdim; ++k) {
        uint iV = (uint)k/vdim;
        uint iU = k%vdim;
        elaVolBf1[k].reset(new DGElasticVolBf2d(et(iV,iU)));
        elaVolBf.put(*elaVolBf1[k], iV, iU);
      }
      concepts::Matrix<Real>::assembly(stiff, spcU, elaVolBf);

      // stiffness matrix: stabilization
      DGElasticStabBf2d elaStabBfX(gammaMu*mu, bc[0].get());
      DGElasticStabBf2d elaStabBfY(gammaMu*mu, bc[1].get());
      vectorial::BilinearForm<Real> elaStabBf(vdim);
      elaStabBf.put(elaStabBfX, 0, 0);
      elaStabBf.put(elaStabBfY, 1, 1);
      concepts::Matrix<Real>::assembly(stiff, elaStabBf, eplU);

      // stiffness matrix: normal stabilization
      concepts::Array<auto_ptr<DGElasticNormalStabBf2d> >
        elaNormalStabBf1(vdim*vdim);
      vectorial::BilinearForm<Real> elaNormalStabBf(vdim);
      for (uint k=0; k<vdim*vdim; ++k) {
        uint iV = (uint)k/vdim;
        uint iU = k%vdim;
        elaNormalStabBf1[k].reset
          (new DGElasticNormalStabBf2d(lambda * gammaLambda, iV, iU,
                                       bc[0].get()));
        elaNormalStabBf.put(*elaNormalStabBf1[k], iV, iU);
      }
      concepts::Matrix<Real>::assembly(stiff, elaNormalStabBf, eplU);

      // equivalent stress operators
      MaximumPrincipalStressHypothesis<2> esh_principal;
      MaximumShearStressHypothesis<2> esh_shear;
      MaximumShearStrainEnergyHypothesis<2> esh_energy;
      EquivalentStress2d eqs_principal(esh_principal, et, spcU);
      EquivalentStress2d eqs_shear(esh_shear, et, spcU);
      EquivalentStress2d eqs_energy(esh_energy, et, spcU);

      // array output
      output.addArrayDouble("nDoF", iArray, spcU.dim());
      output.addArrayDouble("lambda", iArray, lambda);

      logMsg << "S_1_" << iLevel << "_" << iLambda << " lambda_initialization";
      logger.log(logMsg.str());
      logMsg.str("");
      output.addArrayDouble("utime_lambda_init", iArray, rsm.utime());
      writeOutput(input, output);

      if (runStatic) {
        // right hand side
        DGElasticTimeLf2d elaLfX(sourceFrmP, et, 0, bcp, mi, mu*gammaMu,
                                 lambda*gammaLambda, tau, nQuadPts);
        DGElasticTimeLf2d elaLfY(sourceFrmP, et, 1, bcp, mi, mu*gammaMu,
                                 lambda*gammaLambda, tau, nQuadPts);
        vectorial::TimeLinearForm elaLf(vdim);
        elaLf.put(elaLfX, 0);
        elaLf.put(elaLfY, 1);

        concepts::Vector<Real> rhs(spcU, elaLf);

        // solution vector
        concepts::Vector<Real> sol(rhs);

        // solver fabric
        auto_ptr<concepts::SolverFabric<Real> > fabric;
        fabric.reset(solverFabric(input, spcU.dim()));
        if (verbosity > 2) {
          std::cout << *fabric << std::endl;
        }
        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_1 static_initialization";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_static_init", iArray, rsm.utime());
        writeOutput(input, output);

        // calculations
        auto_ptr<concepts::Operator<Real> > solver;
        solver.reset((*fabric)(stiff));
        (*solver)(rhs, sol);
        output.addArrayDouble("iterLinSys", iArray, solverIterations(*solver));

        concepts::Vector<Real> stress_principal(spc1);
        concepts::Vector<Real> stress_shear(spc1);
        concepts::Vector<Real> stress_energy(spc1);
        eqs_principal(sol, stress_principal);
        eqs_shear(sol, stress_shear);
        eqs_energy(sol, stress_energy);

        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_2 static_calculation";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_static_calc", iArray, rsm.utime());
        writeOutput(input, output);

        // array output
        output.addArrayDouble("L2_error", iArray,
                              DGNorm2d::l2(sol, exactFrmX, exactFrmY,
                                           nQuadPts));
        output.addArrayDouble("max_equiv_stress_principal",
                              iArray, stress_principal.max());
        output.addArrayDouble("max_equiv_stress_shear",
                              iArray, stress_shear.max());
        output.addArrayDouble("max_equiv_stress_energy",
                              iArray, stress_energy.max());

        if (dxOutput) {
          // opendx output
          std::stringstream frameName;
          frameName << "u_exact, lambda = " << lambda;
          dxX(exactFrmX, frameName.str());
          
          frameName.str("");
          vectorial::Element<Real>::vecGraphics()->choose(0);
          frameName << "u_num, lambda = " << lambda;
          dxX(sol, frameName.str());

          frameName.str("");
          frameName << "u_exact, lambda = " << lambda;
          dxY(exactFrmY, frameName.str());

          frameName.str("");
          vectorial::Element<Real>::vecGraphics()->choose(1);
          frameName << "u_num, lambda = " << lambda;
          dxY(sol, frameName.str());

          frameName.str("");
          frameName << "equivalent stress_principal, lambda = " << lambda;
          dxStress(stress_principal, frameName.str());

          frameName.str("");
          frameName << "equivalent stress_shear, lambda = " << lambda;
          dxStress(stress_shear, frameName.str());

          frameName.str("");
          frameName << "equivalent stress_energy, lambda = " << lambda;
          dxStress(stress_energy, frameName.str());
        }

        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_3 static_output";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_static_output", iArray, rsm.utime());
        writeOutput(input, output);

      } // if (runStatic)

      // eigenvalues
      if (runEgvl) {
        // mass matrix
        linDG2D::IdentityBf idBf1;
        vectorial::BilinearForm<Real> idBf(vdim);
        for (uint i=0; i<vdim; ++i) {
          idBf.put(idBf1, i, i);
        }
        concepts::DiagonalMatrix<Real> mass(spcU);
        concepts::Matrix<Real>::assembly(mass, spcU, idBf);
         
        // solve for eigenvalues
        uint nEgvl = input.getInt("nEgvl");
        nEgvl = nEgvl > spcU.dim()-1 ? spcU.dim()-1 : nEgvl;

        Real tolEgvl = input.getDouble("tolEgvl");
        uint maxIterEgvl = input.getInt("maxIterEgvl");
        std::string whichEgvlStr = input.getString("whichEgvl");
        concepts::DiagonalSolver<Real> massSolver(mass);

        auto_ptr<eigensolver::EigenSolver> es;
        if (tau == -1) {
          eigensolver::ArPack::which whichEgvl;
          if (whichEgvlStr == "LA") {
            whichEgvl = eigensolver::ArPack::LA;
          } else if (whichEgvlStr == "SA") {
            whichEgvl = eigensolver::ArPack::SA;
          } else if (whichEgvlStr == "LM") {
            whichEgvl = eigensolver::ArPack::LM;
          } else if (whichEgvlStr == "SM") {
            whichEgvl = eigensolver::ArPack::SM;
          } else if (whichEgvlStr == "BE") {
            whichEgvl = eigensolver::ArPack::BE;
          } else {
            std::cerr << "eigensolver::ArPack::" << whichEgvlStr
                      << " does not exist" << std::endl;
            return;
          }
          es.reset(new eigensolver::ArPack(massSolver, stiff, mass, nEgvl,
                                           tolEgvl, maxIterEgvl, whichEgvl));
        } else {
          bool schur = input.getBool("schur");
          eigensolver::ArPackNonSymm::which whichEgvl;
          if (whichEgvlStr == "LM") {
            whichEgvl = eigensolver::ArPackNonSymm::LM;
          } else if (whichEgvlStr == "SM") {
            whichEgvl = eigensolver::ArPackNonSymm::SM;
          } else if (whichEgvlStr == "LR") {
            whichEgvl = eigensolver::ArPackNonSymm::LR;
          } else if (whichEgvlStr == "SR") {
            whichEgvl = eigensolver::ArPackNonSymm::SR;
          } else if (whichEgvlStr == "LI") {
            whichEgvl = eigensolver::ArPackNonSymm::LI;
          } else if (whichEgvlStr == "SI") {
            whichEgvl = eigensolver::ArPackNonSymm::SI;
          } else {
            std::cerr << "eigensolver::ArPackNonSymm::" << whichEgvlStr
                      << " does not exist" << std::endl;
            return;
          }
          es.reset(new eigensolver::ArPackNonSymm
                   (massSolver, stiff, mass, nEgvl, tolEgvl, maxIterEgvl,
                    whichEgvl, eigensolver::ArPackNonSymm::REGINV, 0., 0,
                    schur));
        }

        try {
          const concepts::Array<Real>& ev = es->getEV();
          const concepts::Array<concepts::Vector<Real>*>& ef = es->getEF();
          
          logMsg << "S_1_" << iLevel << "_" << iLambda
                 << "_4 egvl_initializationAndCalculation";
          logger.log(logMsg.str());
          logMsg.str("");
          output.addArrayDouble("utime_egvl_initCalc", iArray, rsm.utime());
          writeOutput(input, output);

          // array and opendx output
          output.addArrayInt("converged", iArray, es->converged());
          output.addArrayInt("iterEgvl", iArray, es->iterations());

          // eigenvalues to arrays in output file
          for (uint iEgvl=0; iEgvl<nEgvl; ++iEgvl) {
            Real egvl = ev[iEgvl];
            if (iEgvl >= es->converged()) {
              egvl = 0;
            }
            std::stringstream arrayName;
            arrayName << "eigenvalueR_" << iEgvl;
            output.addArrayDouble(arrayName.str().c_str(), iArray, egvl);
          }
          if (tau != -1) {
            for (uint iEgvl=0; iEgvl<nEgvl; ++iEgvl) {
              Real egvl = ev[nEgvl+iEgvl];
              if (iEgvl >= es->converged()) {
                egvl = 0;
              }
              std::stringstream arrayName;
              arrayName << "eigenvalueI_" << iEgvl;
              output.addArrayDouble(arrayName.str().c_str(), iArray, egvl);
            }
          }
          if (dxOutput) {
            for (uint iEgvl=0; iEgvl<nEgvl; ++iEgvl) {

              // eigenfunctions to dx file
              std::stringstream frameNameBase;
              frameNameBase << "lambda = " << lambda
                            << ", eigenfunction #" << iEgvl;
              if (iEgvl >= es->converged()) {
                frameNameBase << ", not converged";
              } else {
                if (tau == -1) {
                  frameNameBase << " to eigenvalue " << ev[iEgvl];
                } else {
                  frameNameBase << " to eigenvalue " << ev[iEgvl]
                                << "+" << ev[iEgvl+nEgvl] << "i";
                }
              }
              std::stringstream frameName;
              frameName << frameNameBase.str();
              if (tau != -1) {
                frameName << ", real part";
              }
              vectorial::Element<Real>::vecGraphics()->choose(0);
              dxX(*(ef[iEgvl]), frameName.str());
              vectorial::Element<Real>::vecGraphics()->choose(1);
              dxY(*(ef[iEgvl]), frameName.str());
            }
          }
          logMsg << "S_1_" << iLevel << "_" << iLambda
                 << "_5 egvl_output";
          logger.log(logMsg.str());
          logMsg.str("");
          output.addArrayDouble("utime_egvl_output", iArray, rsm.utime());
          writeOutput(input, output);
        } catch (concepts::FortranException& e) {
          std::cout << e << std::endl;
        }
      } // if (runEgvl)
      // timestepping
      if (runDynamic) {
        // data
        output.addArrayInt("arrayOffsetDyn", iArray, iArrayDyn);
        Real dt = input.getDouble("dt");
        Real tFinal = input.getDouble("tFinal");
        uint dxStride = input.getInt("dxStride");
        uint arrayStride = input.getInt("arrayStride");
        bool implicit = input.getBool("implicit");

        // mass matrix
        linDG2D::IdentityBf idBf1;
        vectorial::BilinearForm<Real> idBf(vdim);
        for (uint i=0; i<vdim; ++i) {
          idBf.put(idBf1, i, i);
        }
        auto_ptr<concepts::Matrix<Real> > mass;
        if (implicit) {
          mass.reset(new concepts::SparseMatrix<Real>(spcU, spcU));
        } else {
          mass.reset(new concepts::DiagonalMatrix<Real>(spcU));
        }
        concepts::Matrix<Real>::assembly(*mass, spcU, idBf);

        // solver fabric
        auto_ptr<concepts::SolverFabric<Real> > fabric;
        if (implicit) {
          fabric.reset(solverFabric(input, spcU.dim()));
        } else {
          fabric.reset(new concepts::DiagonalSolverFabric<Real>);
        }
        if (verbosity > 2) {
          std::cout << *fabric << std::endl;
        }

        // right hand side
        Real omega = input.getDouble("omega");
        concepts::Array<auto_ptr<concepts::Formula> > sourceFrmT(2);
        sourceFrmT[0].reset(new TimeExactRhs(exactFrmX, *(sourceFrm[0]),
                                             omega));
        sourceFrmT[1].reset(new TimeExactRhs(exactFrmY, *(sourceFrm[1]),
                                             omega));
        concepts::Array<concepts::Formula*> sourceFrmTP(2);
        sourceFrmTP[0] = sourceFrmT[0].get();
        sourceFrmTP[1] = sourceFrmT[1].get();

        DGElasticTimeLf2d elaLfX(sourceFrmTP, et, 0, bcp, mi, mu*gammaMu,
                                 lambda*gammaLambda, tau, nQuadPts);
        DGElasticTimeLf2d elaLfY(sourceFrmTP, et, 1, bcp, mi, mu*gammaMu,
                                 lambda*gammaLambda, tau, nQuadPts);
        vectorial::TimeLinearForm elaLf(vdim);
        elaLf.put(elaLfX, 0);
        elaLf.put(elaLfY, 1);
        timestepping::TimeVector rhs(spcU, elaLf);

        // solution vector
        concepts::Vector<Real> sol(rhs);

        // exact solution
        TimeExactSol exactFrmXT(exactFrmX, omega);
        TimeExactSol exactFrmYT(exactFrmY, omega);

        // initial conditions
        auto_ptr<concepts::Operator<Real> > massSolver((*fabric)(*mass));

        linDG2D::ScalarProductLf u0XLf(exactFrmX, nQuadPts);
        linDG2D::ScalarProductLf u0YLf(exactFrmY, nQuadPts);
        vectorial::LinearForm<Real> u0Lf(vdim);
        u0Lf.put(u0XLf, 0);
        u0Lf.put(u0YLf, 1);
        concepts::Vector<Real> u0(spcU, u0Lf);
        (*massSolver)(u0, u0);

        concepts::Vector<Real> v0(spcU);

        Real dt_omega = 0.314159265359 / omega; // sampling...
        output.addArrayDouble("dt_input", iArray, dt);
        output.addArrayDouble("dt_omega", iArray, dt_omega);

        // timestepping
        auto_ptr<timestepping::TimeStepStrategy> tss;
        uint nsteps;
        if (implicit) {
          Real alpha = input.getDouble("alpha");
          dt = dt > dt_omega ? dt_omega : dt;
          // make sure the overall error is O(h)...
          dt /= 1 << (level-minLevel);
          nsteps = static_cast<uint>(std::ceil(tFinal / dt));
          dt = tFinal / nsteps;
          tss.reset(new timestepping::Alpha(*fabric, *mass, stiff, rhs, u0, v0,
                                            dt, alpha));
        } else {
          // estimate timestep via largest eigenvalue
          concepts::DiagonalSolver<Real> massSolver(*mass);
          Real egvl_max;
          uint nEgvl = 10;
          if (tau == -1) {
            eigensolver::ArPack es(massSolver, stiff, *mass, nEgvl, 0.,
                                   1000, eigensolver::ArPack::LM);
            const concepts::Array<Real>& ev = es.getEV();
            egvl_max = std::fabs(ev[nEgvl-1]);
          } else {
            eigensolver::ArPackNonSymm es(massSolver, stiff, *mass, nEgvl,
                                          0., 1000,
                                          eigensolver::ArPackNonSymm::LM);
            const concepts::Array<Real>& ev = es.getEV();
            egvl_max = std::sqrt(ev[0]*ev[0]+ev[nEgvl]*ev[nEgvl]);
          }
          output.addArrayDouble("eigenvalue_max", iArray, egvl_max);
          Real omega_max = std::sqrt(egvl_max);
          Real cfl = input.getDouble("cfl");
          Real dt_cfl = cfl / omega_max;
          output.addArrayDouble("dt_cfl", iArray, dt_cfl);
          dt = dt > dt_cfl ? dt_cfl : dt;
          dt = dt > dt_omega ? dt_omega : dt;
          nsteps = static_cast<uint>(std::ceil(tFinal / dt));
          dt = tFinal / nsteps;
          tss.reset(new timestepping::Nystroem(*fabric, *mass, stiff, rhs, u0,
                                               v0, dt));
        }
        output.addArrayInt("nsteps", iArray, nsteps);
        output.addArrayDouble("dt", iArray, dt);
        if (arrayStride == 0) {
          // sampling
          arrayStride = static_cast<uint>(dt_omega/dt);
        }
        if (dxStride == 0) {
          // sampling
          dxStride = static_cast<uint>(dt_omega/dt);
        }
        timestepping::TimeStepping timeSolver(*tss);

        if (verbosity > 1) {
          std::cout << "|0%    |      2|0%     |      4|0%     |"
                    << "      6|0%     |      8|0%     |   100%|" << std::endl;
        }
        uint nDots = 0;

        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_6 timestepping_initialization";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_dyn_init", iArray, rsm.utime());
        writeOutput(input, output);

        Real mNormUpperBound = 2*DGNorm2d::matrix(u0, *mass);

        for (uint i=0; i<=nsteps; ++i) {
          try {
            timeSolver(sol, i);
            if (i % arrayStride == 0 || i==nsteps) {
              output.addArrayDouble("time", iArrayDyn, i*dt);
              output.addArrayDouble("max_normDyn", iArrayDyn, sol.max());
              Real mNorm = DGNorm2d::matrix(sol, *mass);
              output.addArrayDouble("M_normDyn", iArrayDyn, mNorm);
              output.addArrayDouble("A_normDyn", iArrayDyn,
                                    DGNorm2d::matrix(sol, stiff));
              output.addArrayDouble("L2_errorDyn", iArrayDyn,
                                    DGNorm2d::l2(sol, exactFrmXT, exactFrmYT,
                                                 nQuadPts, i*dt));
              // output.addArrayInt("iterLinSysDyn", iArrayDyn,
              //                    solverIterations(tss->solver()));
              writeOutput(input, output);
              if (mNorm > mNormUpperBound) {
                // stop due to divergence in time
                i = nsteps+1;
              }
              ++iArrayDyn;
            }
            if (dxOutput && (i % dxStride == 0)) {
              std::stringstream frameName;
              frameName << "lambda = " << lambda << ", t = " 
                        << std::setprecision(3) << i*dt;
              vectorial::Element<Real>::vecGraphics()->choose(0);
              dxX(sol, frameName.str());
              vectorial::Element<Real>::vecGraphics()->choose(1);
              dxY(sol, frameName.str());
            }
          } catch (concepts::NoConvergence& e) {
            // stop due to nonconvergence in space
            i = nsteps+1;
          }
          if (verbosity > 1) {
            uint s = 80*i/nsteps - nDots;
            if (s > 0) {
              for (uint j=0; j<s; ++j) {
                std::cout << "." << std::flush;
              }
              nDots += s;
            }
          }
        } // timestepping loop
        if (verbosity > 1) {
          std::cout << std::endl;
        }

        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_7 timestepping_calculationAndOutput";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_dyn_calcOutput", iArray, rsm.utime());
        writeOutput(input, output);

      } // if (runDynamic)
      
      ++iArray;
      
    } // for lambda
  } // for level
}

// ******************************************************************** Main **
int main(int argc, char** argv) {
  if (argc != 2) {
    std::cout << "usage: " << argv[0] << " <input file>" << std::endl;
    return 0;
  }
  concepts::InputParser parser;
  parser.parse(argv[1]);
  concepts::InOutParameters& output = parser.outputParameters();
  const concepts::InOutParameters& input = parser.inputParameters();

  // just to make sure (before calculation) this doesn't cause an error...
  std::string outFileName = input.getString("outFileName");

  std::stringstream inputstream;
  inputstream << input;

  if (inputstream.str().find("syntax error") != std::string::npos) {
    std::cout << input;
    std::cout << "syntax error in input file" << std::endl;
  } else {
    testRun(input, output);
    // testMatrices(input, output);
    // testMappings(input, output);

    writeOutput(input, output);
  }

  return 0;
}
