// periodic uint square mesh

#ifndef periodicSquare_hh
#define periodicSquare_hh

#include "basics/typedefs.hh"
#include "geometry/mesh.hh"
#include "geometry/cell2D.hh"

using concepts::Real;

/** Periodic square. A topologically periodic mesh of the unit
    square (0,1)^2 into 2*n*n triangles.
    @author Christoph Winkelmann, 2004
 */
class PeriodicSquare : public concepts::Mesh2 {
public:
  /** Constructor. The mesh has 2*\c n*\c n elements. The inner points
      are moved in all directions by a random value between -\c jitter/\c n and
      \c jitter/\c n.
      @param n mesh size
      @param jitter jitter parameter
  */
  PeriodicSquare(uint n, Real jitter = 0);
  /// Destructor.
  virtual ~PeriodicSquare();
  /// Returns the number of cells.
  virtual unsigned int ncell() const { return 2*n_*n_; }
  /// Returns a scanner over the mesh cells.
  virtual concepts::Scan2* scan() { return new S(cell_, n_); }
  /// Returns information about the mesh.
  virtual std::ostream& info(std::ostream& os) const;
  /// Returns the mesh size n.
  uint n() { return n_; };
private:
  /// Scanner over mesh cells
  class S : public concepts::Scan<concepts::Cell2> {
    /// position of scanner
    unsigned int idx_;
    /// size of mesh
    unsigned int n_;
    /// pointer to array of cells
    concepts::Triangle2d** cell_;
  public:
    /// Constructor. Takes a cell array and its size
    S(concepts::Triangle2d* cell[], uint n) : idx_(0), n_(n), cell_(cell) {}
    /// Copy constructor.
    S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    /// Returns true if at the end of the scan
    bool eos() const { return idx_ == 2*n_*n_; }
    /// Increment operator.
    concepts::Cell2& operator++(int) { return *cell_[idx_++]; }
    /// Returns a clone of this scanner.
    concepts::Scan2* clone() const { return new S(*this); }
  };
  /// Calculates a geometric vertex index
  uint ii(uint a, uint b) { 
    return a + (1+n_)*b;
  };
  /// Calculates a topologic vertex index.
  uint j(uint a, uint b) { return (a%n_) + n_*(b%n_); };
  /// Calculates a topologic edge index
  uint k(uint a, uint b, uint c) { return c + 3*j(a, b); };
  /// Calculates a topologic triangle index
  uint l(uint a, uint b, uint d) { return d + 2*j(a, b); };
  /// Returns a random number in the suitable range for jittering
  Real dRand() {
    return jitter_*2*((static_cast<Real>(std::rand()) /
      static_cast<Real>(RAND_MAX))-0.5);
  }
  /// Calculates a coordinate from a one dimensional index
  double x(uint a) { return static_cast<double>(a)/static_cast<double>(n_); }
  /// mesh size
  uint n_;
  /// jitter parameter
  Real jitter_;
  /// coordinate array
  concepts::Real2d** coord_;
  /// vertex array
  concepts::Vertex** vtx_;
  /// edge array
  concepts::Edge** edg_;
  /// triangle array
  concepts::Triangle** tri_;
  /// cell array
  concepts::Triangle2d** cell_;
};

#endif // periodicSquare_hh
