#include <iostream>
#include <stdlib.h>

#include "toolbox.hh"
#include "geometry/meshImport2D.hh"
#include "geometry/uniformlyRefinedMesh2.hh"
#include "geometry/boundaryConditions.hh"
#include "operator.hh"
#include "timestepping.hh"
#include "graphics/meshDX.hh"
#include "graphics/dataDX.hh"
#include "eigensolver.hh"
#include "basics/exceptions.hh"
#include "linDG2D.hh"

#if defined(UMFPACK_41)
#include "operator/umfpack.hh"
#include "operator/umfpackFabric.hh"
#elif defined(HAS_SuperLU)
#include "operator/superLU.hh"
#include "operator/superLUfabric.hh"
#endif

#include "linearForm.hh"
#include "bilinearForm.hh"
#include "../app-fvdgfem/formula.hh"
#include "dgNorm2d.hh"

#define DEBUG_TEST 0

using concepts::Real;

class ExactFrm : public concepts::Formula {
 public:
  virtual concepts::Formula* clone () const {
    return new ExactFrm;
  }
  virtual Real operator() (const Real p, const Real t=0.0) const {
    return 0;
  }
  virtual Real operator() (const concepts::Real2d &p, const Real t=0.0) const {
    Real pi = std::acos(-1.);
    return std::sin(pi*p[0]) * std::sin(pi*p[1]) * 
      std::cos(pi*std::sqrt(2.)*t);
  }
  virtual Real operator() (const concepts::Real3d &p, const Real t=0.0) const {
    return 0;
  }
};
  
// ******************************************************** timesteppingTest **
void timesteppingTest(const concepts::InOutParameters& input,
                      concepts::InOutParameters& output) {
  try {
    uint nQuadPts = input.getInt("nQuadPts");

    concepts::Import2dMesh mesh(input.getString("meshCoord"),
                                input.getString("meshElem"),
                                input.getString("meshBnd"),
                                input.getInt("idxStart"));
    linDG2D::SpaceP1 spc(mesh);
    linDG2D::MeshInfo mi(mesh);

    linDG2D::IdentityBf idBf;
    concepts::SparseMatrix<Real> mass(spc, spc);
    concepts::Matrix<Real>::assembly(mass, spc, idBf);

    concepts::Vector<Real> sol(spc);

    TimeSquared rhsFrm;
    linDG2D::LaplaceTimeLf driver_tlf(rhsFrm, 0, mi, 0, 0, nQuadPts);
    timestepping::TimeVector driver(spc, driver_tlf);

    concepts::ParsedFormula u0Frm("(-1)");
    concepts::ParsedFormula v0Frm("(0.5)");
    linDG2D::ScalarProductLf u0Lf(u0Frm, nQuadPts);
    linDG2D::ScalarProductLf v0Lf(v0Frm, nQuadPts);
    concepts::Vector<Real> u0(spc, u0Lf);
    concepts::Vector<Real> v0(spc, v0Lf);

    std::auto_ptr<concepts::SolverFabric<Real> > fabric;
#if defined(UMFPACK_41)
    fabric.reset(new concepts::UmfpackFabric());
#elif defined(HAS_SuperLU)
    fabric.reset(new concepts::SuperLUFabric());
#else
#error either Umfpack or SuperLU needed
#endif

    std::auto_ptr<concepts::Operator<Real> > massSolver((*fabric)(mass));
    (*massSolver)(u0, u0);
    (*massSolver)(v0, v0);
    
    const Real exact = -1. + std::cos(1.) + 0.5*std::sin(1.);
    std::cout << "h\t\terror\t\torder" << std::endl;
    uint steps = 1;
    Real oldError = 0;
    Real newError = 0;
    Real order = 0;
    while (true) {
      Real dt = 1./steps;
      timestepping::Newmark timeStepStrategy(*fabric, mass, mass, driver,
                                             u0, v0, dt, 0.25, 0.5);
      timestepping::TimeStepping timeSolver(timeStepStrategy);
      timeSolver(sol, steps);
      newError = std::fabs(-sol(0) - exact);
      order = std::log(newError/oldError)/std::log(0.5);
      std::cout << dt << "\t" << newError << "\t" << order << std::endl;
      oldError = newError;
      steps *= 2;
    }
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}

// ************************************************************ testMatrices **
void testMatrices(const concepts::InOutParameters& input,
                  concepts::InOutParameters& output)
{
  try {
    // data
    Real tau = input.getDouble("tau");
    Real omega = input.getDouble("omega");
    concepts::ParsedFormula boundaryFrm(input.getString("bndFrm").c_str());
    uint level = input.getInt("minLevel");

    // mesh
    concepts::Import2dMesh coarseMesh(input.getString("meshCoord"),
                                      input.getString("meshElem"),
                                      input.getString("meshBnd"),
                                      input.getInt("idxStart"));
    concepts::UniformlyRefinedMesh2 mesh(coarseMesh, level);
    linDG2D::MeshInfo mi(mesh);

    // boundary
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
            concepts::Boundary(concepts::Boundary::DIRICHLET, boundaryFrm));

    // space
    linDG2D::SpaceP1 spc(mesh);
    linDG2D::ElementPairList epl(spc, spc, mi);

    // bilinear forms
    linDG2D::IdentityBf idBf;
    linDG2D::LaplaceVolBf lapVolBf;
    linDG2D::LaplaceFluxBf lapFluxBf(bc.get());
    linDG2D::LaplaceStabBf lapStabBf(omega, bc.get());

    // mass matrix
    concepts::DenseMatrix<Real> mass(spc, spc);
    concepts::Matrix<Real>::assembly(mass, spc, idBf);

    // stiffness matrices
    concepts::DenseMatrix<Real>*
      stiffFlux = new concepts::DenseMatrix<Real>(spc, spc);
    concepts::Matrix<Real>::assembly(*stiffFlux, lapFluxBf, epl);
    concepts::DenseMatrix<Real>*
      stiffFluxT = new concepts::DenseMatrix<Real>(*stiffFlux, true);
    concepts::LiCo<Real>* 
      temp = new concepts::LiCo<Real>(*stiffFlux, *stiffFluxT,
                                      -1, tau);
    concepts::DenseMatrix<Real> stiff_flux(*temp);
    delete stiffFlux;
    delete stiffFluxT;
    delete temp;

    concepts::DenseMatrix<Real> stiff_vol(spc, spc);
    concepts::Matrix<Real>::assembly(stiff_vol, spc, lapVolBf);
    concepts::DenseMatrix<Real> stiff_pen(spc, spc);
    concepts::Matrix<Real>::assembly(stiff_pen, lapStabBf, epl);

    // output

    std::cout << "AvC = " << stiff_vol << std::endl;
    std::cout << "AfC = " << stiff_flux << std::endl;
    std::cout << "AsC = " << stiff_pen << std::endl;
    std::cout << "MC = " << mass << std::endl;

  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}

// ***************************************************************** waveEqn **
void waveEqn(const concepts::InOutParameters& input,
             concepts::InOutParameters& output)
{
  try {
    // data
    Real tau = input.getDouble("tau");
    Real omega = input.getDouble("omega");
    concepts::ParsedFormula sourceFrm(input.getString("sourceFrm").c_str());
    concepts::ParsedFormula bndFrm(input.getString("bndFrm").c_str());
    concepts::ParsedFormula ic0Frm(input.getString("ic0Frm").c_str());
    concepts::ParsedFormula ic1Frm(input.getString("ic1Frm").c_str());
    Real dt = input.getDouble("dt");
    uint nsteps = input.getInt("nsteps");
    uint dxStride = input.getInt("dxStride");
    uint arrayStride = input.getInt("arrayStride");
    uint minLevel = input.getInt("minLevel");
    uint maxLevel = input.getInt("maxLevel");
    uint nQuadPts = input.getInt("nQuadPts");
    bool implicit = input.getBool("implicit");
    
    // exact solution
    ExactFrm exactFrm;

    // mesh
    concepts::Import2dMesh coarseMesh(input.getString("meshCoord"),
                                      input.getString("meshElem"),
                                      input.getString("meshBnd"),
                                      input.getInt("idxStart"));

    output.addArrayDouble("L2_error");
    output.addArrayDouble("nDoF");

    uint arrayIdx = 0;

    for(uint level=minLevel; level<=maxLevel; ++level) {

      std::cout << "level = " << level << std::endl;

      concepts::UniformlyRefinedMesh2 mesh(coarseMesh, level);
      linDG2D::MeshInfo mi(mesh);
      
      // boundary
      std::auto_ptr<concepts::BoundaryConditions>
        bc(new concepts::BoundaryConditions());
      bc->add(concepts::Attribute(1),
              concepts::Boundary(concepts::Boundary::DIRICHLET, bndFrm));
      
      // space
      linDG2D::SpaceP1 spc(mesh);
      linDG2D::ElementPairList epl(spc, spc, mi);

      // mass matrix
      std::auto_ptr<concepts::Matrix<Real> > mass;
      if(implicit) {
        mass.reset(new concepts::SparseMatrix<Real>(spc, spc));
      } else {
        mass.reset(new concepts::DiagonalMatrix<Real>(spc));
      }
      linDG2D::IdentityBf idBf;
      concepts::Matrix<Real>::assembly(*mass, spc, idBf);
      
      // stiffness matrix
      concepts::SparseMatrix<Real>*
        stiffFlux = new concepts::SparseMatrix<Real>(spc,spc);
      linDG2D::LaplaceFluxBf lapFluxBf(bc.get());
      concepts::Matrix<Real>::assembly(*stiffFlux, lapFluxBf, epl);
      concepts::SparseMatrix<Real>*
        stiffFluxT = new concepts::SparseMatrix<Real>(*stiffFlux,
                                                          true);
      concepts::LiCo<Real>* 
        temp = new concepts::LiCo<Real>(*stiffFlux, *stiffFluxT,
                                        -1, tau);
      concepts::SparseMatrix<Real> stiff(*temp);
      delete stiffFlux;
      delete stiffFluxT;
      delete temp;
      
      linDG2D::LaplaceVolBf lapVolBf;
      concepts::Matrix<Real>::assembly(stiff, spc, lapVolBf);
      linDG2D::LaplaceStabBf lapStabBf(omega, bc.get());
      concepts::Matrix<Real>::assembly(stiff, lapStabBf, epl);

      // solution vector and right hand side
      concepts::Vector<Real> sol(spc);
      linDG2D::LaplaceTimeLf rhsLf(sourceFrm, bc.get(), mi, omega, tau, nQuadPts);
      timestepping::TimeVector rhs(spc, rhsLf);
      
      // initial conditions
      linDG2D::ScalarProductLf u0Lf(ic0Frm, nQuadPts);
      linDG2D::ScalarProductLf v0Lf(ic1Frm, nQuadPts);
      concepts::Vector<Real> u0(spc, u0Lf);
      concepts::Vector<Real> v0(spc, v0Lf);

      std::auto_ptr<concepts::SolverFabric<Real> > fabric;
      if(implicit) {
#if defined(UMFPACK_41)
        fabric.reset(new concepts::UmfpackFabric());
#elif defined(HAS_SuperLU)
        fabric.reset(new concepts::SuperLUFabric());
#else
#error either Umfpack or SuperLU needed
#endif
      } else {
        fabric.reset(new concepts::DiagonalSolverFabric<Real>);
      }
      
      std::auto_ptr<concepts::Operator<Real> > massSolver((*fabric)(*mass));
      (*massSolver)(u0, u0);
      (*massSolver)(v0, v0);
      
      // graphical output
      // graphics::MeshDX<Real>(spc, "wave.mesh", 2);
      graphics::DataDX<Real>
        dx(graphics::drawDataDX(spc, input.getString("dxFileName"), 2));
      
      // array output
      output.addArrayDouble("time");
      output.addArrayDouble("M_norm");
      output.addArrayDouble("A_norm");
      output.addArrayDouble("max_norm");
      output.addArrayDouble("L2_error");
      
      // timestepping
      std::auto_ptr<timestepping::TimeStepStrategy> tss;
      if(implicit) {
        Real alpha = input.getDouble("alpha");
        tss.reset(new timestepping::Alpha(*fabric, *mass, stiff, rhs, u0, v0,
                                          dt, alpha));
      } else {
        // estimate timestep via largest eigenvalue
        concepts::DiagonalSolver<Real> massSolver(*mass);
        std::auto_ptr<eigensolver::EigenSolver> es;
        if(tau == -1) {
          es.reset(new eigensolver::ArPack(massSolver, stiff, *mass, 10, 0.,
                                           1000, eigensolver::ArPack::LM));
        } else {
          es.reset(new eigensolver::ArPackNonSymm
                   (massSolver, stiff, *mass, 10, 0., 1000,
                    eigensolver::ArPackNonSymm::LM));
        }
        const concepts::Array<Real>& ev = es->getEV();
        Real omega_max = std::sqrt(ev[0]);
        Real cfl = input.getDouble("cfl");
        Real dt_cfl = cfl / omega_max;
        output.addDouble("dt_input", dt);
        output.addDouble("dt_cfl", dt_cfl);
        dt = dt > dt_cfl ? dt_cfl : dt;
        output.addDouble("dt", dt);
        tss.reset(new timestepping::Nystroem(*fabric, *mass, stiff, rhs, u0,
                                             v0, dt));
      }
      timestepping::TimeStepping timeSolver(*tss);

      std::cout << "|0%    |      2|0%     |      4|0%     |"
                << "      6|0%     |      8|0%     |   100%|" << std::endl;

      uint nDots = 0;
      for(uint i=0; i<=nsteps; ++i) {
        timeSolver(sol, i);
        if (i % arrayStride == 0) {
          output.addArrayDouble("time", arrayIdx, i*dt);
          output.addArrayDouble("max_norm", arrayIdx, sol.max());
          output.addArrayDouble("M_norm", arrayIdx,
                                DGNorm2d::matrix(sol, *mass));
          output.addArrayDouble("A_norm", arrayIdx,
                                DGNorm2d::matrix(sol, stiff));
          output.addArrayDouble("L2_error", arrayIdx,
                                DGNorm2d::l2(sol, exactFrm, nQuadPts, i*dt));
          ++arrayIdx;
        }
        if (i % dxStride == 0) {
          std::stringstream frameName;
          frameName << "t = " << std::setprecision(3) << i*dt;
          dx(sol, frameName.str());
        }
        uint s = 80*i/nsteps - nDots;
        if (s > 0) {
          for (uint j=0; j<s; ++j) {
            std::cout << "." << std::flush;
          }
          nDots += s;
        }
      }
      std::cout << std::endl;
      output.addArrayDouble("nDoF", level, spc.dim());
    }
  } catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}

// ******************************************************************** diff **
void diff(const concepts::InOutParameters& input,
          concepts::InOutParameters& output)
{
  try {
    Real pi = std::acos(-1.);
    // data
    Real tau = input.getDouble("tau");
    Real omega = input.getDouble("omega");
    concepts::ParsedFormula sourceFrm(input.getString("sourceFrm").c_str());
    concepts::ParsedFormula bndFrm(input.getString("bndFrm").c_str());
    uint minLevel = input.getInt("minLevel");
    uint maxLevel = input.getInt("maxLevel");
    uint nQuadPts = input.getInt("nQuadPts");
    uint nEgvl = input.getInt("nEgvl");
    
    // mesh
    concepts::Import2dMesh coarseMesh(input.getString("meshCoord"),
                                      input.getString("meshElem"),
                                      input.getString("meshBnd"),
                                      input.getInt("idxStart"));

    output.addArrayDouble("nDoF");
    output.addArrayDouble("L2_error");
    output.addArrayDouble("order");
    Real oldError = 0;
    Real newError = 0;

    std::stringstream arrayName;
    for(uint iEgvl=0; iEgvl<nEgvl; ++iEgvl) {
      arrayName << "eigenvalueR_" << iEgvl;
      output.addArrayDouble(arrayName.str().c_str());
      arrayName.str("");
      if(tau != -1) {
        arrayName << "eigenvalueI_" << iEgvl;
        output.addArrayDouble(arrayName.str().c_str());
        arrayName.str("");
      }
    }
    output.addArrayInt("iterations");
    output.addArrayInt("converged");

    for(uint level=minLevel; level<=maxLevel; ++level) {
      // mesh
      concepts::UniformlyRefinedMesh2 mesh(coarseMesh, level);
      linDG2D::MeshInfo mi(mesh);

      // boundary
      std::auto_ptr<concepts::BoundaryConditions>
        bc(new concepts::BoundaryConditions());
      bc->add(concepts::Attribute(1),
              concepts::Boundary(concepts::Boundary::DIRICHLET, bndFrm));

      // space
      linDG2D::SpaceP1 spc(mesh);
      linDG2D::ElementPairList epl(spc, spc, mi);

      // mass matrix
      concepts::DiagonalMatrix<Real> mass(spc);
      linDG2D::IdentityBf idBf;
      concepts::Matrix<Real>::assembly(mass, spc, idBf);
      
      // stiffness matrix
      concepts::SparseMatrix<Real>*
        stiffFlux = new concepts::SparseMatrix<Real>(spc,spc);
      linDG2D::LaplaceFluxBf lapFluxBf(bc.get());
      concepts::Matrix<Real>::assembly(*stiffFlux, lapFluxBf, epl);
      concepts::SparseMatrix<Real>*
        stiffFluxT = new concepts::SparseMatrix<Real>(*stiffFlux,
                                                          true);
      concepts::LiCo<Real>* 
        temp = new concepts::LiCo<Real>(*stiffFlux, *stiffFluxT,
                                        -1, tau);
      concepts::SparseMatrix<Real> stiff(*temp);
      delete stiffFlux;
      delete stiffFluxT;
      delete temp;
      
      linDG2D::LaplaceVolBf lapVolBf;
      concepts::Matrix<Real>::assembly(stiff, spc, lapVolBf);
      linDG2D::LaplaceStabBf lapStabBf(omega, bc.get());
      concepts::Matrix<Real>::assembly(stiff, lapStabBf, epl);
      if (tau == -1) {
        stiff.symmetrize();
      }

      // solution vector and right hand side
      concepts::Vector<Real> sol(spc);
      linDG2D::LaplaceTimeLf rhsLf(sourceFrm, bc.get(), mi, omega, tau, nQuadPts);
      concepts::Vector<Real> rhs(spc, rhsLf);

      graphics::DataDX<Real>
        dx(graphics::drawDataDX(spc, input.getString("dxFileName"), 2));

      std::auto_ptr<concepts::Operator<Real> > solver;
#if defined(UMFPACK_41)
      solver.reset(new concepts::Umfpack(stiff));
#elif defined(HAS_SuperLU)
      solver.reset(new concepts::SuperLU(stiff));
#else
#error either Umfpack or SuperLU needed
#endif

      (*solver)(rhs, sol);
      dx(rhs, "right hand side");
      dx(sol, "solution");
      output.addArrayDouble("nDoF", level, spc.dim());
      newError = DGNorm2d::l2(sol, bndFrm, nQuadPts);
      output.addArrayDouble("L2_error", level, newError);
      output.addArrayDouble("order", level,
                            std::log(newError/oldError)/std::log(0.5));
      oldError = newError;

      // eigenvalues
      concepts::DiagonalSolver<Real> massSolver(mass);
      std::auto_ptr<eigensolver::EigenSolver> es;
      if(tau == -1) {
        es.reset(new eigensolver::ArPack(massSolver, stiff, mass, nEgvl, 0.,
                                         1000, eigensolver::ArPack::SA));
      } else {
        es.reset(new eigensolver::ArPackNonSymm
                 (massSolver, stiff, mass, nEgvl,0., 1000,
                  eigensolver::ArPackNonSymm::SR));
      }
      const concepts::Array<Real>& ev = es->getEV();
      const concepts::Array<concepts::Vector<Real>*>& efR = es->getEF();
      output.addArrayInt("converged", level, es->converged());
      output.addArrayInt("iterations", level, es->iterations());
      
      for(uint iEgvl=0; iEgvl<nEgvl; ++iEgvl) {
        arrayName << "eigenvalueR_" << iEgvl;
        output.addArrayDouble(arrayName.str().c_str(), level,
                              ev[iEgvl]/(pi*pi));
        arrayName.str("");
        arrayName << "eigenfunctionR_" << iEgvl;
        dx(*efR[iEgvl], arrayName.str());
        arrayName.str("");
        if(tau != -1) {
          arrayName << "eigenvalueI_" << iEgvl;
          output.addArrayDouble(arrayName.str().c_str(), level,
                                ev[iEgvl+nEgvl]/(pi*pi));
          arrayName.str("");
        }
      }
    }
  } catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}

// ******************************************************************** Main **
int main(int argc, char** argv) {
  if (argc != 2) {
    std::cout << "usage: " << argv[0] << " <input file>" << std::endl;
    return 0;
  }
  concepts::InputParser parser;
  parser.parse(argv[1]);
  concepts::InOutParameters& output = parser.outputParameters();
  const concepts::InOutParameters& input = parser.inputParameters();

  // timesteppingTest(input, output);
  // testMatrices(input, output);
  waveEqn(input, output);
  // diff(input, output);

  std::ofstream logfile(input.getString("outFileName").c_str());
  logfile << output;

  return 0;
}
