#include "advection.hh"

// *************************************************************** Advection **
template<typename NumFlux>
std::ostream& Advection<NumFlux>::info (std::ostream &os) const {
  os << "Space: " << spc_;
  return os;
}

template<typename NumFlux>
void Advection<NumFlux>::operator() (const concepts::Function<Real>& fncY,
				     concepts::Function<Real>& fncX) {
  DEBUGL(DEBUG_ADVECTION, "begin of Advection::operator()");
  try{
    // Downcast functions to vectors
    const concepts::Vector<Real>& vecY =
      dynamic_cast<const concepts::Vector<Real>&> (fncY);
    concepts::Vector<Real>& vecX  =
      dynamic_cast<concepts::Vector<Real>&> (fncX);
    // Get time, if vecX is a time vector
    const timestepping::TimeVector* const tVec = 
      dynamic_cast<const timestepping::TimeVector* const> (&vecY);
    if(tVec) {
      time_ = tVec->time();
    }
    // Set up quadrature vectors
    std::vector<Real> weights;
    std::vector<concepts::Real3d> xi1;
    std::vector<concepts::Real3d> xi2;
    std::vector<concepts::Real3d> x;

    // Reset solution vector to zero
    vecX *= 0;
    DEBUGL(DEBUG_ADVECTION, "vecY " << vecY << " vecX " << vecX);
    // Scanner for inner and boundary faces
    std::auto_ptr<typename concepts::ElementPairList<Real>::Scanner> 
      innerSc(innerList_.scan());
    conceptsAssert(innerSc.get(), concepts::Assertion());
    std::auto_ptr<typename concepts::ElementPairList<Real>::Scanner> 
      boundarySc(boundaryList_.scan());
    conceptsAssert(boundarySc.get(), concepts::Assertion());
    // Loop over inner faces
    while (*innerSc) {
      try {
	const concepts::ElementPair<Real>& epBase = (*innerSc)++;
	const linDG3D::FvdgElementPair& ep =
	  dynamic_cast<const linDG3D::FvdgElementPair&>(epBase);
	// Elements adjacent to face
	const linDG3D::FvdgElement& elm1 = ep.elm1();
	const linDG3D::FvdgElement& elm2 = ep.elm2();
	ep.quadrature(weights, xi1, xi2);
	// No. Dof. of elements
	const uint nDof = elm1.T().m();
	conceptsAssert(nDof == elm2.T().m(), concepts::Assertion());
	// Loop over edgePairs
	const concepts::Real3d faceNormal = ep.normal();
	const uint NOfQuadPts = weights.size();
	for (uint i = 0; i < NOfQuadPts; ++i) {
	  // Calculate flux
	  const Real flux = ep.area()*weights[i]*
	    numFlux_(elm1.solution(vecY, xi1[i]),
		     elm2.solution(vecY, xi2[i]),
		     faceNormal);
	  DEBUGL(DEBUG_FLUXES, "flux   = " << flux);
	  // "Distribute values"
	  for (uint j = 0; j < nDof; ++j) {
	    vecX(elm1.T().index(j)) +=
	      elm1.shapeFct(j, xi1[i])*flux;
	    vecX(elm2.T().index(j)) -=
	      elm2.shapeFct(j, xi2[i])*flux;
	  }
	}
      } catch (std::bad_cast& e) {
	std::cerr << e.what() << std::endl;
	conceptsAssert(false, concepts::Assertion());
      } // end try/catch
    } // end while(*innerSc)
    DEBUGL(DEBUG_ADVECTION, "After inner scan\nvecX: " << vecX);

    // Loop over boundary faces
    while (*boundarySc) {
      try {
	const concepts::ElementPair<Real>& epBase = (*boundarySc)++;
	const linDG3D::FvdgElementPair& ep =
	  dynamic_cast<const linDG3D::FvdgElementPair&>(epBase);
	// Element adjacent to face
	const linDG3D::FvdgElement& elm = ep.elm1();
	ep.quadrature(weights, xi1, xi2, x);
	  DEBUGL(DEBUG_ADVECTION, "weights = [" << weights[0] << ", " << weights[1] << ", " 
		 << weights[2] << ", " << weights[3] << ", " << weights[4] << ", " << weights[5] 
		 << ", " << weights[6] << "]" );
	  DEBUGL(DEBUG_ADVECTION, "x = [" << x[0] << ", " << x[1] << ", " 
		 << x[2] << ", " << x[3] << ", " << x[4] << ", " << x[5] 
		 << ", " << x[6] << "]" );
	// No. Dof. of elements
	const uint nDof = elm.T().m();
	// Loop over edgePairs
	const concepts::Real3d faceNormal = ep.normal();
	const uint NOfQuadPts = weights.size();
	std::vector<Real> u(NOfQuadPts, 0.);
	for (uint i = 0; i < NOfQuadPts; ++i) {
	  u[i] = elm.solution(vecY, xi1[i]);
	}
	bool isOutflow = false;
	if (pcb_) {
	  const Real uMean = 
	    std::inner_product(u.begin(), u.end(), weights.begin(), 0.);
	  DEBUGL(DEBUG_ADVECTION, "u = [" << u[0] << ", " << u[1] << ", " 
		 << u[2] << ", " << u[3] << ", " << u[4] << ", " << u[5] 
		 << ", " << u[6] << "]" );
	  DEBUGL(DEBUG_ADVECTION, "uMean = " << uMean);
	  // Distinguish between inflow and outflow boundary
	  isOutflow = (numFlux_.fluxFct_.deriv(uMean, faceNormal) > 0);
	}
	DEBUGL(DEBUG_ADVECTION, (isOutflow ? "outflow" : "inflow"));
	for (uint i = 0; i < NOfQuadPts; ++i) {
	  Real uRight;
	  try {
	    uRight = (isOutflow ? 
		      u[i] :
		      uBoundary_(elm, ep.faceIdx1(),x[i]));
	  } catch(concepts::MissingParameter& e) {
	    std::cerr << "Warning: " << e.what() <<
	      "\nDoing 0th order extrapolation instead.\n";
	    uRight = u[i];
	  }
	  const Real flux = 
	    ep.area()*numFlux_(u[i], uRight, faceNormal)*weights[i];
	  DEBUGL(DEBUG_FLUXES, "area   = " << ep.area());
	  DEBUGL(DEBUG_FLUXES, "uLeft  = " << u[i] );
	  DEBUGL(DEBUG_FLUXES, "uRight = " << uRight);
	  DEBUGL(DEBUG_FLUXES, "normal = " << faceNormal);
	  DEBUGL(DEBUG_FLUXES, "flux   = " << flux);
	  for(uint j = 0; j < nDof; ++j) {
	    vecX(elm.T().index(j)) +=
	      elm.shapeFct(j, xi1[i])*flux;
	  } // end for j
	} // end for i
      } catch (std::bad_cast& e) {
	std::cerr << e.what() << std::endl;
	conceptsAssert(false, concepts::Assertion());
      } // end try/catch
    } // end while
    DEBUGL(DEBUG_ADVECTION, "After boundary scan\nvecX: " << vecX);

    // Loop over all elements, calculating volume integral
    std::auto_ptr<typename concepts::Scan<linDG3D::FvdgElement> > 
      elemSc(spc_.scan());
    conceptsAssert(elemSc.get(), concepts::Assertion());
    std::vector<Real> solEdgeCenter(6);
    std::vector<Real> solVertex(4);
    concepts::Real3d v1, v2, v3;
    while (*elemSc) {
      const linDG3D::FvdgElement& elm = (*elemSc)++;
      // Nothing to do for one Ansatz function
      const uint nDof = elm.T().m();
      if (nDof == 1) continue;
      // Build up solution vectors for edge centers and vertices
      for (uint j = 0; j < 6; ++j) {
	solEdgeCenter[j] = elm.solutionEdgeCenter(vecY, j);
      }
      for (uint j = 0; j < 4; ++j) {
	solVertex[j] = elm.solutionVertex(vecY, j);
      }
      // Calculate volume of element
      v1 = elm.cell().vertex(1) - elm.cell().vertex(0);
      v2 = elm.cell().vertex(2) - elm.cell().vertex(0);
      v3 = elm.cell().vertex(3) - elm.cell().vertex(0);
      const concepts::Mapping<Real, 3> B(v1, v2, v3);
      const Real vol = B.determinant()/6.;
      // Loop over all local degress of freedom
      // Assumption: the Gradient is constant for all shapefunctions
      for (uint i = 1; i < nDof; ++i) {
	const concepts::Real3d gradient = elm.shapeFctGradient(i);
	Real fluxIntegral = 0.;
	// Quadrature on element:
	// 1. Loop over edge centers
	for (uint j = 0; j < 6; ++j) {
	  fluxIntegral += 
	    numFlux_.fluxFct_.flux(solEdgeCenter[j], gradient);
	}
	fluxIntegral *= 4.;
	// 2. Loop over vertices
	for (uint j = 0; j < 4; ++j) {
	  fluxIntegral -= 
	    numFlux_.fluxFct_.flux(solVertex[j], gradient);
	}
	vecX(elm.T().index(i)) -= fluxIntegral*vol/20.;
      } // end for i
    } // end while(*elemSc)
    DEBUGL(DEBUG_ADVECTION, "After volume contribution\nvecX: " << vecX);
  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
  DEBUGL(DEBUG_ADVECTION, "end of Advection::operator()");
}

template <typename NumFlux>
Real Advection<NumFlux>::uBoundary_(const linDG3D::FvdgElement& elm,
				    const uint faceIdx,
				    const concepts::Real3d &x) const {
  concepts::Attribute attrib = elm.support().face(faceIdx)->attrib();
  DEBUGL(DEBUG_ADVECTION, "type  = " << (*bc_)(attrib).type() );
  if ((*bc_)(attrib).type() == concepts::Boundary::DIRICHLET) {
    return (*bc_)(attrib)(x, time_);
  } 
  // Don't know what to do, throw an exception
  throw conceptsException(concepts::MissingParameter("No Dirichlet boundary"));
  return 0;
}

// Numerical Fluxes

//************************************************************ LaxFriedrichs **
template <typename T>
Real LaxFriedrichs<T>::operator() (Real u1, Real u2,
				   const concepts::Real3d& normal) const {
  const Real lambda = std::max(std::fabs(fluxFct_.deriv(u1, normal)), 
			  std::fabs(fluxFct_.deriv(u2, normal)));
  DEBUGL(DEBUG_FLUXES, "u1:" << u1 << ", abs:" << std::fabs(-u1));
  return 0.5*(fluxFct_.flux(u1, normal) + fluxFct_.flux(u2, normal) 
    - lambda*(u2-u1));
}

//******************************************************************* Upwind **
template <typename T>
Real Upwind<T>::operator()
  (Real u1, Real u2, const concepts::Real3d& normal) const {
  const Real uMean = 0.5*(u1 + u2);
  if (fluxFct_.deriv(uMean, normal) > 0) {
    return fluxFct_.flux(u1, normal);
  } else {
    return fluxFct_.flux(u2, normal);
  }
}

//****************************************************************** Godunov **
template <typename T>
Real Godunov<T>::operator() 
  (Real u1, Real u2, const concepts::Real3d& normal) const {
  const Real uMin = fluxFct_.uMin(normal);
  const Real fluxU1 = fluxFct_.flux(u1, normal);
  const Real fluxU2 = fluxFct_.flux(u2, normal);
  Real shockVel = (u2 != u1 ? (fluxU2 - fluxU1)/(u2 - u1) :
		   fluxFct_.deriv(u1, normal));
  // Check for convexity!
  if(!fluxFct_.isConvex(normal)) {
    shockVel *= -1.;
    const Real temp = u1;
    u1 = u2; u2 = temp;
  }
  if (u1 <= uMin && u2 >= uMin)
    return fluxFct_.flux(uMin, normal);
  DEBUGL(DEBUG_FLUXES,
	 "u1:" << u1 << ", u2:" << u2 <<
	 "\ns:" << shockVel << ", normal:" << normal);
  if(u1 > uMin && shockVel >= 0) return fluxFct_.flux(u1, normal);
  if(u2 < uMin && shockVel <= 0) return fluxFct_.flux(u2, normal);
  conceptsAssert(false, concepts::Assertion());
  return 0;
}


//******************************************************************** Osher **

template <typename T>
Real Osher<T>::operator()
  (Real u1, Real u2, const concepts::Real3d& normal) const {
  const Real uMin = fluxFct_.uMin(normal);
  // Check for convexity!
  if(!fluxFct_.isConvex(normal)) {
    const Real temp = u1;
    u1 = u2; u2 = temp;
  }
  
  if (u1 > uMin) {
    if (u2 > uMin) return fluxFct_.flux(u1, normal);
    else return fluxFct_.flux(u1, normal) + fluxFct_.flux(u2, normal) - 
	   fluxFct_.flux(uMin, normal);
  } else {
    if (u2 < uMin) return fluxFct_.flux(u2, normal);
    else return fluxFct_.flux(uMin, normal);
  }
  conceptsAssert(false, concepts::Assertion());
  return 0;
}

//*********************************************************** FvdgPi0Limiter **
FvdgPi0Limiter::FvdgPi0Limiter(const linDG3D::FvdgSpace& spc) : diag_(spc) {
  std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc.scan());
  // Loop over all elements
  while (*elemSc) {
    const linDG3D::FvdgElement& elm = (*elemSc)++;
    const uint nDof = elm.T().m();
    // Set weight of constant shapefunction to 1
    uint idx = elm.T().index(0);
    diag_(idx, idx) = 1.;
    // Set the rest to zero
    for(uint i = 1; i < nDof; ++i) {
      idx = elm.T().index(i);
      diag_(idx, idx) = 0.;
    }
  } // end while
}

//****************************************************** FvdgAdaptiveLimiter **
FvdgAdaptiveLimiter::FvdgAdaptiveLimiter(const linDG3D::FvdgSpace& spc,
					 const Real alpha,
					 const Real lambda = 0.,
					 concepts::BoundaryConditions* bc = 0):
  spc_(spc), bc_(bc), alpha_(alpha), lambda_(lambda), elmIndicator_() {
  std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc.scan());
  conceptsAssert(elemSc.get(), concepts::Assertion());
  while (*elemSc) {
    const linDG3D::FvdgElement& elm = (*elemSc)++;
    elmIndicator_.insert(std::pair<const linDG3D::FvdgElement*, 
			 Real>(&elm, 0.));
  } // end while
}

void FvdgAdaptiveLimiter::reset_() {
  for (std::map<const linDG3D::FvdgElement*, Real>::iterator mi =
	 elmIndicator_.begin();
       mi != elmIndicator_.end(); ++mi) {
    mi->second = 0.;
  }
}

void FvdgAdaptiveLimiter::innerContrib_(const concepts::Vector<Real>& vecY, 
					concepts::Vector<Real>& vecX) {
  try{
    const concepts::ElementPairList<Real>& innerList_ = 
      spc_.innerElmPairList();
    std::auto_ptr<concepts::ElementPairList<Real>::Scanner> 
      innerSc(innerList_.scan());
    conceptsAssert(innerSc.get(), concepts::Assertion());
    // Loop over inner faces
    while (*innerSc) {
      // Calculate square of face integral
      const concepts::ElementPair<Real>& epBase = (*innerSc)++;
      const linDG3D::FvdgElementPair& ep =
	dynamic_cast<const linDG3D::FvdgElementPair&>(epBase);
      // Elements adjacent to face
      const linDG3D::FvdgElement& elm1 = ep.elm1();
      const linDG3D::FvdgElement& elm2 = ep.elm2();
      // Calculate volume of elements
      //      const Real vol1 = elmVolume_(elm1);
      //      const Real vol2 = elmVolume_(elm2);
      // Loop over edgePairs
      const uint nPairs = ep.numPairs();
      Real faceIntegral = 0.;
      for (uint i = 0; i < nPairs; ++i) {
	const Real tmp = elm1.solutionEdgeCenter(vecY, ep.edgeIndex(0, i)) - 
	  elm2.solutionEdgeCenter(vecY, ep.edgeIndex(1, i));
	faceIntegral += tmp*tmp; 
      }
      // Scale integral
      faceIntegral /= 3.;
      //  faceIntegral *= ep.area()/(3.*ep.diam()*ep.diam());
      // Distribute values and scale appropriately
      elmIndicator_[&elm1] += faceIntegral;
      elmIndicator_[&elm2] += faceIntegral;
      //  elmIndicator_[&elm1] += faceIntegral/std::pow(vol1, (alpha_-2.)/3.);
      //  elmIndicator_[&elm2] += faceIntegral/std::pow(vol2, (alpha_-2.)/3.);
    } // end while
  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
}

void FvdgAdaptiveLimiter::bndryContrib_(const concepts::Vector<Real>& vecY, 
					concepts::Vector<Real>& vecX) {
  try {
    const concepts::ElementPairList<Real>& boundaryList_ =
      spc_.boundaryElmPairList();
    std::auto_ptr<concepts::ElementPairList<Real>::Scanner> 
      boundarySc(boundaryList_.scan());
    conceptsAssert(boundarySc.get(), concepts::Assertion());
    while (*boundarySc) {
      // Calculate square of face integral
      const concepts::ElementPair<Real>& epBase = (*boundarySc)++;
      const linDG3D::FvdgElementPair& ep =
	dynamic_cast<const linDG3D::FvdgElementPair&>(epBase);
      // Face element
      const linDG3D::FvdgElement& elm = ep.elm1();
      // Calculate volume of element
      //      const Real vol = elmVolume_(elm);
      // Loop over edgePairs
      const uint nPairs = ep.numPairs();
      Real faceIntegral = 0.;
      for (uint i = 0; i < nPairs; ++i) {
	concepts::Attribute attrib = 
	  elm.support().face(ep.faceIdx1())->attrib();
	const Real uBoundary = 
	  ((*bc_)(attrib).type() == concepts::Boundary::DIRICHLET ?
	   (*bc_)(attrib)(elm.edgeCenter(ep.edgeIndex(0, i))) : 0.);
	const Real tmp = elm.solutionEdgeCenter(vecY, ep.edgeIndex(0, i))
	  - uBoundary;
	faceIntegral += tmp*tmp;
      }
      // Scale integral
      faceIntegral /= 3.;
      //      faceIntegral *= ep.area()/(3.*ep.diam()*ep.diam());
      // Accumulate value
      elmIndicator_[&elm] += faceIntegral;
      //elmIndicator_[&elm] += faceIntegral/std::pow(vol, (alpha_-2.)/3.);
    } // end while 
  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
}

Real FvdgAdaptiveLimiter::elmVolume_(const linDG3D::FvdgElement& elm) const {
  concepts::Real3d v1 = elm.cell().vertex(1) - elm.cell().vertex(0);
  concepts::Real3d v2 = elm.cell().vertex(2) - elm.cell().vertex(0);
  concepts::Real3d v3 = elm.cell().vertex(3) - elm.cell().vertex(0);
  const concepts::Mapping<Real, 3> B(v1, v2, v3);
  return B.determinant()/6.;
}

void FvdgAdaptiveLimiter::limit_(concepts::Vector<Real>& vecX,
				  const linDG3D::FvdgElement& elm) const {
  const uint nDof = elm.T().m();
  uint idx = elm.T().index(0);
  // Set elements belonging to non-constant shapefunction to zero
  for (uint i = 1; i < nDof; ++i) {
    idx = elm.T().index(i);
    vecX(idx) = 0.;
  } // end for
}
void FvdgAdaptiveLimiter::operator() (const concepts::Function<Real> &fncY,
				      concepts::Function<Real> &fncX) {
  try {
    // Downcast functions to vectors
    const concepts::Vector<Real>& vecY =
      dynamic_cast<const concepts::Vector<Real>&> (fncY);
    concepts::Vector<Real>& vecX  =
      dynamic_cast<concepts::Vector<Real>&> (fncX);
    // Assign values from vecY to vecX and alter the rest later
    vecX = vecY;
    DEBUGL(DEBUG_REF_ASSIGNMENT, "Before: " << vecX);
    // Reset element indicator to zero
    reset_();
    // Calculate indicator functions for all elements, using loop over faces
    // 1. Contributions from inner faces
    innerContrib_(vecY, vecX);
    // 2. Contributions from boundary faces
    bndryContrib_(vecY, vecX);
    
    // Decide if element needs limiting and modify vector accordingly
    for (std::map<const linDG3D::FvdgElement*, Real>::iterator mi = 
	   elmIndicator_.begin();
	 mi != elmIndicator_.end(); ++mi ) {
      DEBUGL(DEBUG_ADAPT_LIM_INDICATOR,
	     std::endl << "element: " << mi->first->cell().connector().key()
	     << " indicator: " << mi->second);
      const linDG3D::FvdgElement& elm = *(mi->first);
      if (((mi->second)/std::pow(elmVolume_(elm), (alpha_-2.)/3.)) > 1) {
	DEBUGL(DEBUG_ADAPT_LIM_ELM, 
	       "limited element: " << elm.cell().connector().key());
	limit_(vecX, elm);
      }
    }
    DEBUGL(DEBUG_REF_ASSIGNMENT, "After: " << vecX); 
  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
}

void FvdgAdaptiveLimiter::getIndicator(concepts::Vector<Real> &indicator) {
  try {
    std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc_.scan());
    conceptsAssert(elemSc.get(), concepts::Assertion());

    // Reset indicator
    indicator = 0.;

    while (*elemSc) {
      const linDG3D::FvdgElement& elm = (*elemSc)++;

      std::map<const linDG3D::FvdgElement*, Real>::iterator it = 
	elmIndicator_.find(&elm);
      indicator(elm.T().index(0)) = it->second;
    } // end while (*elemSc)

  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
}


// Specializations
template class LaxFriedrichs<LinearAdvectionFlux>;
template class LaxFriedrichs<BurgerFlux>;
template class Upwind<LinearAdvectionFlux>;
template class Upwind<BurgerFlux>;
template class Godunov<BurgerFlux>;
template class Osher<BurgerFlux>;

template class Advection<LaxFriedrichs<LinearAdvectionFlux> >;
template class Advection<LaxFriedrichs<BurgerFlux> >;
template class Advection<Upwind<LinearAdvectionFlux> >;
template class Advection<Upwind<BurgerFlux> >;
template class Advection<Godunov<BurgerFlux> >;
template class Advection<Osher<BurgerFlux> >;
