// Numerical tests for fvdg project

#include <iostream>
#include <cmath>
#include <cstdlib>
#include <sstream>
#include <fstream>
#include <unistd.h>

#include "basics.hh"
#include "toolbox.hh"
#include "graphics.hh"
#include "timestepping.hh"
#include "linDG3D.hh"
#include "geometry/meshImport3D.hh"
#include "function/vector.hh"
#include "operator.hh"
#include "formula.hh"
#include "periodicube.hh"
#include "advection.hh"
#include "utility.hh"
#include "pslogger.hh"

#define DEBUG_TEST 0

using concepts::Real;

//************************************************************ advectionTest **
void advectionTest(uint nsteps, Real tfinal, bool p0,
		   bool periodicMesh, uint n, Real jitter, Real cfl,
		   uint fluxNbr, uint limNbr, Real alpha,
		   bool tvd_rk, std::string fileNameBase)
{
  try {
    // set up timing
    concepts::ResourceMonitor timer;
    timer.utime();
    // set up memory logging
    std::stringstream fileName;
    fileName << fileNameBase << "_pslog.txt" << std::ends;
    PsLogger logger(fileName.str(), "rss pcpu etime cputime");
    logger.log("beginning:");
    // set up input parser (for output)
    concepts::InputParser input;
    concepts::InOutParameters& outputParameters = input.outputParameters();
    outputParameters.addString("basic: fileNameBase", fileNameBase.c_str());
    // set up fvdg stuff
    concepts::Mesh3* mshP = 0;
    if (periodicMesh) {
      mshP = new PeriodiCube(n, jitter);
      // static_cast<PeriodiCube*>(mshP)->saveMesh();
    } else {
      mshP = new concepts::Import3dMesh("mesh.coord", "mesh.elms",
					"mesh.boundary", 1, false);
    }
    logger.log("+ mesh =");
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, "(0)"));
    logger.log("+ boundary conditions = ");
    linDG3D::FvdgSpace* spcP = 0;
    if (p0) {
      spcP = new linDG3D::FvdgSpaceP0(*mshP);
    } else {
      spcP = new linDG3D::FvdgSpaceP1(*mshP);
    }
    logger.log("+ space =");
    // save mesh for dx
    fileName.str("");
    fileName << fileNameBase << "_mesh.dx" << std::ends;
    graphics::MeshDX<Real>(*spcP, fileName.str(), 3);
    // set up mass matrix
    linDG3D::Identity id;
    concepts::DiagonalMatrix<Real> mass(*spcP);
    concepts::Matrix<Real>::assembly(mass, *spcP, id);
    logger.log("+ mass matrix =");
    // set up advection
    Upwind<LinearAdvectionFlux> numfluxLU;
    LaxFriedrichs<LinearAdvectionFlux> numfluxLL;
    Upwind<BurgerFlux> numfluxBU;
    LaxFriedrichs<BurgerFlux> numfluxBL;
    Osher<BurgerFlux> numfluxBO;
    Godunov<BurgerFlux> numfluxBG;
    concepts::Operator<Real>* advectionP = 0;
    switch (fluxNbr) {
    case 0: 
      advectionP = new Advection<Upwind<LinearAdvectionFlux> >(numfluxLU,*spcP,
							       true, bc.get());
      break;
    case 1: 
      advectionP = 
	new Advection<LaxFriedrichs<LinearAdvectionFlux> >(numfluxLL, *spcP,
							   true, bc.get());
      break;
    case 2:
      advectionP = new Advection<Upwind<BurgerFlux> >(numfluxBU, *spcP,
						      true, bc.get());
      break;
    case 3: 
      advectionP = new Advection<LaxFriedrichs<BurgerFlux> >(numfluxBL, *spcP,
							     true, bc.get());
      break;
    case 4:
      advectionP = new Advection<Osher<BurgerFlux> >(numfluxBO, *spcP,
						     true, bc.get());
      break;
    case 5:
      advectionP = new Advection<Godunov<BurgerFlux> >(numfluxBG, *spcP,
						       true, bc.get());
      break;
    }
    concepts::Operator<Real>* limiterP = 0;
    if (p0) limNbr = 0;
    switch (limNbr) {
    case 1: limiterP = new FvdgPi0Limiter(*spcP); break;
    case 2: limiterP = new FvdgAdaptiveLimiter(*spcP, alpha, 0, bc.get());
      break;
    case 3: limiterP = new FvdgPi0Limiter(*spcP); break;
    }
    concepts::Compose<Real>* limAdvectionP = 0;
    concepts::Compose<Real>* limAdvectionP2 = 0;
    concepts::Operator<Real>* advP;
    if (limiterP) {
      limAdvectionP = new concepts::Compose<Real>(*advectionP, *limiterP);
      if (limNbr == 3) {
	limAdvectionP2 = 
	  new concepts::Compose<Real>(*limiterP, *limAdvectionP);
	advP = limAdvectionP2;
      } else {
	advP = limAdvectionP;
      }
    } else {
      advP = advectionP;
    }
    // do diffusion only
    concepts::Operator<Real>* D0P = 0;
    D0P = advP;
    logger.log("+ advection =");
    // set up solution vector, right hand side and indicator
    concepts::Vector<Real> sol(*spcP);
    logger.log("+ solution vector =");
    linDG3D::ConstRhs driver_tlf(0.);
    timestepping::TimeVector driver(*spcP, driver_tlf);
    logger.log("+ right hand side =");
    concepts::Vector<Real> indicator(*spcP);
    logger.log("+ indicator =");
    // Set up Burger error estimator
    BoxSin boxSin;
    NodalBurgerError burgerEstimator(boxSin, *spcP);
    concepts::Array<Real> berror;
    logger.log("+ burger error estimator =");
    // Set up initial condition
    linDG3D::FvdgTimeRhs initial_tlf(boxSin);
    timestepping::TimeVector initial(*spcP, initial_tlf);
    concepts::DiagonalSolver<Real> diag(mass);
    diag(initial, initial);
    logger.log("+ initial vector =");
    // set up timestepping
    TimestepEstimator<LinearAdvectionFlux> tseL(*spcP, 1.0);
    TimestepEstimator<BurgerFlux> tseB(*spcP, 1.0);
    Real dt = cfl*(fluxNbr < 2 ? tseL(initial) : tseB(initial));
    if (nsteps == 0) {
      nsteps = static_cast<uint>(std::ceil(tfinal/dt));
      dt = tfinal/nsteps;
    } else {
      tfinal = static_cast<Real>(nsteps)*dt;
    }
    outputParameters.addInt("basic: number of elements", spcP->nelm());
    outputParameters.addInt("basic: number of timesteps", nsteps);
    outputParameters.addDouble("basic: timestep", dt);
    outputParameters.addDouble("basic: final time", tfinal);
    outputParameters.addDouble("basic: timestep for writing", tfinal);
    timestepping::TimeStepStrategy* tssP = 0;
    concepts::DiagonalSolverFabric<Real> fabric;
    if (limiterP) {
      if (tvd_rk) {
	tssP = new timestepping::LimitingTvdRK2(fabric, mass, *D0P, *limiterP,
						driver, initial, dt);
      } else {
	tssP = new timestepping::LimitingEuler(fabric, mass, *D0P, *limiterP,
					       driver, initial, dt);
      }
    } else {
      if (tvd_rk) {
	tssP = new timestepping::TvdRK2(fabric, mass, *D0P,
					driver, initial, dt);
      } else {
	tssP = new timestepping::Euler(fabric, mass, *D0P,
				       driver, initial, dt);
      }
    }
    DEBUGL(DEBUG_TEST, *tssP);
    timestepping::TimeStepping timeSolver(*tssP);
    logger.log("+ timestepping =");
    // set up output arrays
    outputParameters.addArrayDouble("snapshot: M-norm");
    outputParameters.addArrayDouble("snapshot: max-norm");
    outputParameters.addArrayDouble("snapshot: time");
    outputParameters.addArrayDouble("snapshot: burger-mx-nrm");
    // do timestepping
    // create graphics object for numeric solution
    fileName.str("");
    fileName << fileNameBase << "_data.dx" << std::ends;
    graphics::DataDX<Real> dx(graphics::drawDataDX(*spcP, fileName.str(), 3));
    // create graphics object for exact solution (linear advection)
    fileName.str("");
    fileName << fileNameBase << "_exact.dx" << std::ends;
    graphics::DataDX<Real> dxEx(graphics::drawDataDX(*spcP,fileName.str(), 3));
    // create graphics object for limiter indicator
    fileName.str("");
    fileName << fileNameBase << "_ind.dx" << std::ends;
    graphics::DataDX<Real> 
      dxInd(graphics::drawDataDX(*spcP, fileName.str(), 3));
    // do some timing
    float uTime_setup = timer.utime();
    float uTime_timestepping = 0;
    float uTime_snapshots = 0;
    uint snapshot = 0;
    // timestepping loop
    //for(uint n = 0;  n <= nsteps; n+=stride, ++snapshot) {
    timeSolver(sol, nsteps);
    DEBUGL(DEBUG_TEST, "limNbr = " << limNbr);
    if (limNbr == 2) {
      FvdgAdaptiveLimiter* adLim = 
	dynamic_cast<FvdgAdaptiveLimiter*>(limiterP);
      conceptsAssert(adLim, concepts::Assertion());
      DEBUGL(DEBUG_TEST, "Before: " << indicator);
      adLim->getIndicator(indicator);
      DEBUGL(DEBUG_TEST, "After: " << indicator);
    }
    uTime_timestepping += timer.utime();
    // do test outputs
    //       // cfl test output
    //       std::cout << "cfl(" << n*dt << ") = " 
    //                 << dt / (fluxNbr < 2 ? tseL(sol) : tseB(sol)) << std::endl;
    // ||u||L2 test output
    std::cout << "||u||L2(" << tfinal << ") = " << Norm::matrix(sol, mass)
	      << std::endl;
    // ||u||Max test output
    std::cout << "||u||Max(" << tfinal << ") = " << Norm::max(sol) 
	      << std::endl;
    // write dx file
    std::stringstream frameName;
    frameName << "t = " << std::setprecision(3)
	      << tfinal;
    dx(sol, frameName.str());
    if (limNbr == 2) {
      DEBUGL(DEBUG_TEST, frameName.str());
      dxInd(indicator, frameName.str());
    }
    // calculate error vector for linear advection
    initial_tlf.time(tfinal);
    timestepping::TimeVector tempTv(*spcP, initial_tlf);
    diag(tempTv, tempTv); // projected exact solution
    concepts::Vector<Real>& temp = 
      dynamic_cast<concepts::Vector<Real>&> (tempTv);
    dxEx(temp, frameName.str());
    temp -= sol; // approximation error
    // Linear Advection Output
    outputParameters.addArrayDouble("snapshot: M-norm", snapshot,
				    Norm::matrix(temp, mass));
    outputParameters.addArrayDouble("snapshot: max-norm", snapshot,
				      Norm::max(temp));
    //    std::cout << "linadv error(" << n*dt << ") = " << Norm::max(temp) 
    //		<< std::endl;
    outputParameters.addArrayDouble("snapshot: time", snapshot,
				    tfinal);
    // error norm for burger flux
    burgerEstimator(sol, berror, tfinal);
    Real berr = 0.;
    for (uint i = 0; i < berror.size(); ++i) {
      if (std::fabs(berror[i]) > berr) berr = std::fabs(berror[i]);
    }
    //   std::cout << "burger error(" << n*dt << ") = " << berr << std::endl;
    outputParameters.addArrayDouble("snapshot: burger-mx-nrm", snapshot,
				    berr);
    uTime_snapshots += timer.utime();
    //} // end for timestepping loop
    logger.log("end");
    //doing time calculations
    outputParameters.addDouble("timer: setup", uTime_setup);
    outputParameters.addDouble("timer: setup per element",
			       uTime_setup / spcP->nelm());
    outputParameters.addDouble("timer: timestepping",
			       uTime_timestepping);
    outputParameters.addDouble("timer: timestepping per timestep",
			       uTime_timestepping / nsteps);
    outputParameters.addDouble("timer: timestepping per timestep and element",
			       uTime_timestepping / nsteps / spcP->nelm());
    outputParameters.addDouble("timer: snapshots",
			       uTime_snapshots);
    outputParameters.addDouble("timer: snapshots per snapshot",
			       uTime_snapshots / snapshot);
    outputParameters.addDouble("timer: snapshots per snapshot and element",
			       uTime_snapshots / snapshot / spcP->nelm());
    // writing log file
    fileName.str("");
    fileName << fileNameBase << "_log.txt" << std::ends;
    std::ofstream logfile(fileName.str().c_str());
    logfile << input;
    // delete objects on heap
    delete tssP;
    tssP = 0;
    delete limAdvectionP;
    limAdvectionP = 0;
    delete limiterP;
    limiterP = 0;
    delete advectionP;
    advectionP = 0;
    D0P = 0;
    delete spcP;
    spcP = 0;
    delete mshP;
    mshP = 0;
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}

//****************************************************************** cflTest **
void cflTest(uint nsteps, Real tfinal, bool p0, bool periodicMesh, 
	     uint n, Real jitter, Real cfl, uint fluxNbr, uint limNbr,
	     Real alpha, bool tvd_rk,
	     std::string fileNameBase) {
  try {
    // constant for break
    const Real c = 100;
    // Mesh
    concepts::Mesh3 *mshP = 0;
    if (periodicMesh) {
      mshP = new PeriodiCube(n, jitter);
      // static_cast<PeriodiCube*>(mshP)->saveMesh();
    } else {
      mshP = new concepts::Import3dMesh("mesh.coord", "mesh.elms",
					"mesh.boundary", 1, false);
    }
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, "(1)"));
    // Space
    linDG3D::FvdgSpace* spcP = 0;
    if (p0) {
      spcP = new linDG3D::FvdgSpaceP0(*mshP);
    } else {
      spcP = new linDG3D::FvdgSpaceP1(*mshP);
    }
    // set up mass matrix
    linDG3D::Identity id;
    concepts::DiagonalMatrix<Real> mass(*spcP);
    concepts::Matrix<Real>::assembly(mass, *spcP, id);

    // set up advection
    Upwind<LinearAdvectionFlux> numfluxLU;
    LaxFriedrichs<LinearAdvectionFlux> numfluxLL;
    Upwind<BurgerFlux> numfluxBU;
    LaxFriedrichs<BurgerFlux> numfluxBL;
    Osher<BurgerFlux> numfluxBO;
    Godunov<BurgerFlux> numfluxBG;
    concepts::Operator<Real>* advectionP = 0;
    switch (fluxNbr) {
    case 0: 
      advectionP = new Advection<Upwind<LinearAdvectionFlux> >(numfluxLU,*spcP,
							       true, bc.get());
      break;
    case 1: 
      advectionP = 
	new Advection<LaxFriedrichs<LinearAdvectionFlux> >(numfluxLL, *spcP,
							   true, bc.get());
      break;
    case 2:
      advectionP = new Advection<Upwind<BurgerFlux> >(numfluxBU, *spcP,
						      true, bc.get());
      break;
    case 3: 
      advectionP = new Advection<LaxFriedrichs<BurgerFlux> >(numfluxBL,*spcP,
							     true, bc.get());
      break;
    case 4:
      advectionP = new Advection<Osher<BurgerFlux> >(numfluxBO, *spcP,
						     true, bc.get());
      break;
    case 5:
      advectionP = new Advection<Godunov<BurgerFlux> >(numfluxBG, *spcP,
						       true, bc.get());
      break;
    }
    // set up limiter
    concepts::Operator<Real>* limiterP = 0;
    if (p0) limNbr = 0;
    switch (limNbr) {
    case 1: limiterP = new FvdgPi0Limiter(*spcP); break;
    case 2: limiterP = new FvdgAdaptiveLimiter(*spcP, alpha, 0, bc.get());
      break;
    }
    concepts::Compose<Real>* limAdvectionP = 0;
    concepts::Compose<Real>* limAdvectionP2 = 0;
    concepts::Operator<Real>* D0P;
    if (limiterP) {
      limAdvectionP = new concepts::Compose<Real>(*advectionP, *limiterP);
      if (limNbr == 3) {
	limAdvectionP2 = 
	  new concepts::Compose<Real>(*limAdvectionP, *limiterP);
	D0P = limAdvectionP2;
      } else {
	D0P = limAdvectionP;
      }
    } else {
      D0P = advectionP;
    }
    // set up solution vector and right hand side
    concepts::Vector<Real> sol(*spcP);
    linDG3D::ConstRhs driver_tlf(0.0);
    timestepping::TimeVector driver(*spcP, driver_tlf);

    // set up initial condition
    BoxSin initFctSin;
    linDG3D::FvdgTimeRhs rhs(initFctSin);
    timestepping::TimeVector initial(*spcP, rhs);
    concepts::DiagonalSolver<Real> diag(mass);
    diag(initial, initial);

    // set up timestepping
    TimestepEstimator<LinearAdvectionFlux> tseL(*spcP, 1.0);
    TimestepEstimator<BurgerFlux> tseB(*spcP, 1.0);
    Real dt = cfl*(fluxNbr < 2 ? tseL(initial) : tseB(initial));
    std::cout << "dt = " << dt << std::endl;
    uint stride;
    if (nsteps == 0) {
      stride = static_cast<uint>(std::ceil(tfinal/(20.0*dt)));
      nsteps = static_cast<uint>(std::ceil(tfinal/dt));
    } else {
      stride = ((nsteps-1)/20)+1;
    }
    nsteps = stride*(((nsteps-1)/stride)+1);
    tfinal = static_cast<Real>(nsteps)*dt;
    timestepping::TimeStepStrategy* tssP = 0;
    concepts::DiagonalSolverFabric<Real> fabric;
    if (limiterP) {
      if (tvd_rk) {
	tssP = new timestepping::LimitingTvdRK2(fabric, mass, *D0P, *limiterP,
						driver, initial, dt);
      } else {
	tssP = new timestepping::LimitingEuler(fabric, mass, *D0P, *limiterP,
					       driver, initial, dt);
      }
    } else {
      if (tvd_rk) {
	tssP = new timestepping::TvdRK2(fabric, mass, *D0P,
					driver, initial, dt);
      } else {
	tssP = new timestepping::Euler(fabric, mass, *D0P,
				       driver, initial, dt);
      }
    }
    timestepping::TimeStepping timeSolver(*tssP);
    // do timestepping
    const Real uMax0 = 0.25;
    Real uMax = 0;
    for(uint n = 0;  n <= nsteps; n+=stride) {
      timeSolver(sol, n);
      uMax = Norm::max(sol);
      if(uMax > c*uMax0) {
	exit(3);
      }
    }
    // delete objects on heap
    delete tssP;
    tssP = 0;
    delete limAdvectionP;
    limAdvectionP = 0;
    delete limiterP;
    limiterP = 0;
    delete advectionP;
    advectionP = 0;
    D0P = 0;
    delete spcP;
    spcP = 0;
    delete mshP;
    mshP = 0;
    exit(4);
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}

//************************************************************* ellipticTest **
void ellipticTest(std::string fileNameBase, bool symm) {
  try {
    // formulae on (0,1)^3
    concepts::ParsedFormula boundaryFrm("(x*(1-x)+y*(1-y)+z*(1-z))");
    concepts::ParsedFormula exactFrm("(x*(1-x)+y*(1-y)+z*(1-z))");
    concepts::ParsedFormula gradX("(1-2*x)");
    concepts::ParsedFormula gradY("(1-2*y)");
    concepts::ParsedFormula gradZ("(1-2*z)");
    concepts::ParsedFormula sourceFrm("(6)");

    // formulae on unit sphere
//     concepts::ParsedFormula boundaryFrm("(0)");
//     concepts::ParsedFormula exactFrm("(x*x+y*y+z*z-1)");
//     concepts::ParsedFormula gradX("(2*x)");
//     concepts::ParsedFormula gradY("(2*y)");
//     concepts::ParsedFormula gradZ("(2*z)");
//     concepts::ParsedFormula sourceFrm("(-6)");

    //    std::cout << "loading mesh ... " << std::flush;
    concepts::Mesh3* mshP =
      new concepts::Import3dMesh("mesh.coord", "mesh.elms",
				 "mesh.boundary", 1, false);
    //    std::cout << "done" << std::endl;
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, boundaryFrm));
    //    std::cout << "setting up space ... " << std::flush;
    linDG3D::FvdgSpaceP1 spc(*mshP, bc.get());
    //    std::cout << "done" << std::endl;

    // save mesh for dx
    std::stringstream fileName;
    fileName << fileNameBase << "_mesh.dx" << std::ends;
    graphics::MeshDX<Real>(spc, fileName.str(), 3);

    linDG3D::FvdgTimeRhs exactLF(exactFrm);
    concepts::Vector<Real> exact(spc, exactLF);

    linDG3D::Identity id;
    concepts::DiagonalMatrix<Real> mass(spc);
    concepts::Matrix<Real>::assembly(mass, spc, id);

    concepts::DiagonalSolver<Real> diag(mass);
    diag(exact, exact);

    // std::cout << exact << std::endl;

    linDG3D::LaplaceVolBF lapBfVol;
    linDG3D::LaplaceInnerBF lapBfInn(symm);
    linDG3D::LaplaceBoundaryBF lapBfBnd(symm);
    linDG3D::PenaltyBF penBf;

    concepts::DenseMatrix<Real> diff(spc, spc);

    //    std::cout << "assembling volume laplace matrix ... " << std::flush;
    concepts::Matrix<Real>::assembly(diff, spc, lapBfVol);
    //    std::cout << "done" << std::endl 
    //	      << "assembling inner laplace matrix ... " << std::flush;
    linDG3D::Assembly::inner(diff, lapBfInn, spc.innerElmPairList(), true);
    //    std::cout << "done" << std::endl 
    //	      << "assembling boundary laplace matrix ... " << std::flush;
    linDG3D::Assembly::boundary(diff, lapBfBnd, spc.boundaryElmPairList(),
				true);
    //    std::cout << "done" << std::endl 
    //	      << "assembling inner penalty matrix ... " << std::flush;
    linDG3D::Assembly::inner(diff, penBf, spc.innerElmPairList(), true);
    //    std::cout << "done" << std::endl 
    //	      << "assembling boundary penalty matrix ... " << std::flush;
    linDG3D::Assembly::boundary(diff, penBf, spc.boundaryElmPairList(), true);
    // std::cout << diff << std::endl;
    //    std::cout << "done" << std::endl;
    linDG3D::FvdgTimeRhsDiff rhsLF(sourceFrm, 1, bc.get(), symm);
    //    std::cout << "assembling load vector ... " << std::flush;
    concepts::Vector<Real> rhs(spc, rhsLF);
    // std::cout << rhs << std::endl;
    //    std::cout << "done" << std::endl 
    //	      << "creating solver ... " << std::flush;
    concepts::GMRes<Real> solver(diff, 1e-4, spc.dim(), 0, 1);
    //    std::cout << "done" << std::endl 
    //	      << "creating solution vector ... " << std::flush;
    linDG3D::ConstRhs initLF(1);
    concepts::Vector<Real> sol(spc, initLF);
    //    std::cout << "done" << std::endl 
    //	      << "solving the system ... " << std::flush;
    try {
      solver(rhs, sol);
    } catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    //    std::cout << "done" << std::endl;

    fileName.str("");
    fileName << fileNameBase << "_data.dx" << std::ends;
    graphics::DataDX<Real> dx(graphics::drawDataDX(spc, fileName.str(), 3));
    std::stringstream frameName;
    frameName << "solution";
    dx(sol, frameName.str());
    
    fileName.str("");
    fileName << fileNameBase << "_exact.dx" << std::ends;
    graphics::DataDX<Real> dxEx(graphics::drawDataDX(spc, fileName.str(), 3));
    dxEx(exact, frameName.str());

    concepts::Vector<Real> residual(spc);
    diff(sol, residual);
    residual -= rhs;
    std::cout << "||A*u_FV-l||L2   = " << Norm::matrix(residual, mass)
	      << std::endl;
//     diff(exact, residual);
//     residual -= rhs;
//     std::cout << "||A*u_proj-l||L2 = " << Norm::matrix(residual, mass)
// 	      << std::endl;

    std::cout << spc.dim() << "\t"
	      << Norm::broL2Grad(sol, gradX, gradY, gradZ) << std::endl;
//     std::cout << "|u_proj-u|H1 = "
// 	      << Norm::broL2Grad(exact, gradX, gradY, gradZ) << std::endl;
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}

//************************************************************ parabolicTest **
void parabolicTest(uint nsteps, Real tfinal, bool periodicMesh, uint n,
		   Real jitter, Real cfl, bool tvd_rk, bool symm,
		   std::string fileNameBase, std::string meshNameBase)
{
  try {
    // formulae on (0,1)^3
    HeatSolution exactFrm;
    concepts::Formula& boundaryFrm = exactFrm;
    concepts::Formula& initFrm = exactFrm;
    concepts::ParsedFormula sourceFrm("(0)");
    // set up timing
    concepts::ResourceMonitor timer;
    timer.utime();
    // set up memory logging
    std::stringstream fileName;
    fileName << fileNameBase << "_pslog.txt" << std::ends;
    PsLogger logger(fileName.str(), "rss pcpu etime cputime");
    logger.log("beginning:");
    // set up input parser (for output)
    concepts::InputParser input;
    concepts::InOutParameters& outputParameters = input.outputParameters();
    outputParameters.addString("basic: fileNameBase", fileNameBase.c_str());
    // set up fvdg stuff
    concepts::Mesh3* mshP = 0;
    if (periodicMesh) {
      mshP = new PeriodiCube(n, jitter);
      // static_cast<PeriodiCube*>(mshP)->saveMesh();
    } else {
      std::stringstream coordName;
      std::stringstream elmsName;
      std::stringstream boundaryName;
      coordName << meshNameBase << ".coord" << std::ends;
      elmsName << meshNameBase << ".elms" << std::ends;
      boundaryName << meshNameBase << ".boundary" << std::ends;
      mshP = new concepts::Import3dMesh(coordName.str(), elmsName.str(),
					boundaryName.str(), 1, false);
    }
    logger.log("+ mesh =");
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, boundaryFrm));
    logger.log("+ boundary conditions = ");
    linDG3D::FvdgSpace* spcP = new linDG3D::FvdgSpaceP1(*mshP);
    logger.log("+ space =");
    // save mesh for dx
    fileName.str("");
    fileName << fileNameBase << "_mesh.dx" << std::ends;
    graphics::MeshDX<Real>(*spcP, fileName.str(), 3);
    // set up mass matrix
    linDG3D::Identity id;
    concepts::DiagonalMatrix<Real> mass(*spcP);
    concepts::Matrix<Real>::assembly(mass, *spcP, id);
    logger.log("+ mass matrix =");
    // set up diffusion
    linDG3D::LaplaceVolBF lapBfVol;
    linDG3D::LaplaceInnerBF lapBfInn(symm);
    linDG3D::LaplaceBoundaryBF lapBfBnd(symm);
    linDG3D::PenaltyBF penBf;
    concepts::SparseMatrix<Real> diff(*spcP, *spcP);
    concepts::Matrix<Real>::assembly(diff, *spcP, lapBfVol);
    linDG3D::Assembly::inner(diff, lapBfInn, spcP->innerElmPairList(), true);
    linDG3D::Assembly::boundary(diff, lapBfBnd, spcP->boundaryElmPairList(),
				true);
    linDG3D::Assembly::inner(diff, penBf, spcP->innerElmPairList(), true);
    linDG3D::Assembly::boundary(diff, penBf, spcP->boundaryElmPairList(),
				true);
    logger.log("+ diffusion =");
    // set up solution vector and right hand side
    concepts::Vector<Real> sol(*spcP);
    logger.log("+ solution vector =");
    linDG3D::FvdgTimeRhsDiff driver_tlf(sourceFrm, 1, bc.get(), symm);
    timestepping::TimeVector driver(*spcP, driver_tlf);
    logger.log("+ right hand side =");
    // Set up initial condition
    linDG3D::FvdgTimeRhs rhs(initFrm);
    timestepping::TimeVector initial(*spcP, rhs);
    concepts::DiagonalSolver<Real> diag(mass);
    diag(initial, initial);
    logger.log("+ initial vector =");
    // set up timestepping
    TimestepEstimatorParabolic tse(0.01);
    Real dt = cfl*tse(*spcP);
    // std::cout << "dt = " << dt << std::endl;
    uint stride;
    if (nsteps == 0) {
      stride = static_cast<uint>(std::ceil(tfinal/(20.0*dt)));
      nsteps = static_cast<uint>(std::ceil(tfinal/dt));
    } else {
      stride = ((nsteps-1)/20)+1;
    }
    nsteps = stride*(((nsteps-1)/stride)+1);
    tfinal = static_cast<Real>(nsteps)*dt;
    outputParameters.addInt("basic: number of elements", spcP->nelm());
    outputParameters.addInt("basic: number of timesteps", nsteps);
    outputParameters.addDouble("basic: timestep", dt);
    outputParameters.addDouble("basic: final time", tfinal);
    outputParameters.addDouble("basic: timestep for writing", stride*dt);
    timestepping::TimeStepStrategy* tssP = 0;
    concepts::DiagonalSolverFabric<Real> fabric;
    if (tvd_rk) {
      tssP = new timestepping::TvdRK2(fabric, mass, diff,
				      driver, initial, dt);
    } else {
      tssP = new timestepping::Euler(fabric, mass, diff,
				     driver, initial, dt);
    }
    timestepping::TimeStepping timeSolver(*tssP);
    logger.log("+ timestepping =");
    // set up output arrays
    outputParameters.addArrayDouble("snapshot: L1-norm");
    outputParameters.addArrayDouble("snapshot: max-norm");
    outputParameters.addArrayDouble("snapshot: time");
    // do timestepping
    // create graphics object for numeric solution
    fileName.str("");
    fileName << fileNameBase << "_data.dx" << std::ends;
    graphics::DataDX<Real> dx(graphics::drawDataDX(*spcP, fileName.str(), 3));
    // do some timing
    float uTime_setup = timer.utime();
    float uTime_timestepping = 0;
    float uTime_snapshots = 0;
    uint snapshot = 0;
    // timestepping loop
    for(uint n = 0;  n <= nsteps; n+=stride, ++snapshot) {
      timeSolver(sol, n);
      uTime_timestepping += timer.utime();
      // do test outputs
      //       // cfl test output
      //       std::cout << "cfl(" << n*dt << ") = " 
      //           << dt / (fluxNbr < 2 ? tseL(sol) : tseB(sol)) << std::endl;
      // ||u||L2 test output
//       std::cout << "||u||L2(" << n*dt << ") = " << Norm::matrix(sol, mass)
// 		<< std::endl;
      // ||u||Max test output
//       std::cout << "||u||Max(" << n*dt << ") = " << Norm::max(sol) 
// 		<< std::endl;
      // write dx file
      std::stringstream frameName;
      frameName << "t = " << std::setprecision(3)
		<< n*dt;
      dx(sol, frameName.str());
      // output
      Real l1norm = Norm::broL1(sol, exactFrm, n*dt);
//     std::cout << "||u-u_ex||L1(" << n*dt << ")  = " << l1norm << std::endl;
      outputParameters.addArrayDouble("snapshot: L1-norm", snapshot,
				      l1norm);
      Real maxNorm = Norm::broMax(sol, exactFrm, n*dt);
//     std::cout << "||u-u_ex||max(" << n*dt << ") = " << maxNorm << std::endl;
      outputParameters.addArrayDouble("snapshot: max-norm", snapshot,
				      maxNorm);
      outputParameters.addArrayDouble("snapshot: time", snapshot,
				      n*dt);
      uTime_snapshots += timer.utime();
    } // end for timestepping loop
    logger.log("end");
    //doing time calculations
    outputParameters.addDouble("timer: setup", uTime_setup);
    outputParameters.addDouble("timer: setup per element",
			       uTime_setup / spcP->nelm());
    outputParameters.addDouble("timer: timestepping",
			       uTime_timestepping);
    outputParameters.addDouble("timer: timestepping per timestep",
			       uTime_timestepping / nsteps);
    outputParameters.addDouble("timer: timestepping per timestep and element",
			       uTime_timestepping / nsteps / spcP->nelm());
    outputParameters.addDouble("timer: snapshots",
			       uTime_snapshots);
    outputParameters.addDouble("timer: snapshots per snapshot",
			       uTime_snapshots / snapshot);
    outputParameters.addDouble("timer: snapshots per snapshot and element",
			       uTime_snapshots / snapshot / spcP->nelm());
    // writing log file
    fileName.str("");
    fileName << fileNameBase << "_log.txt" << std::ends;
    std::ofstream logfile(fileName.str().c_str());
    logfile << input;
    // delete objects on heap
    delete tssP;
    tssP = 0;
    delete spcP;
    spcP = 0;
    delete mshP;
    mshP = 0;
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}

// ************************************************** advectionDiffusionTest **
void advectionDiffusionTest(uint nsteps, Real tfinal, Real cfl,
			    uint fluxNbr, uint limNbr, Real alpha,
			    Real epsilon, Real ny, bool tvd_rk,
			    std::string fileNameBase, std::string meshNameBase)
{
  try {
    if (epsilon <= 0) exit(-1);
    // set up timing
    concepts::ResourceMonitor timer;
    timer.utime();
    // set up memory logging
    std::stringstream fileName;
    fileName << fileNameBase << "_pslog.txt" << std::ends;
    PsLogger logger(fileName.str(), "rss pcpu etime cputime");
    logger.log("beginning:");
    // set up input parser (for output)
    concepts::InputParser input;
    concepts::InOutParameters& outputParameters = input.outputParameters();
    outputParameters.addString("basic: fileNameBase", fileNameBase.c_str());
    // set up fvdg stuff
    concepts::Mesh3* mshP = 0;
    std::stringstream coordName;
    std::stringstream elmsName;
    std::stringstream boundaryName;
    coordName << meshNameBase << ".coord" << std::ends;
    elmsName << meshNameBase << ".elms" << std::ends;
    boundaryName << meshNameBase << ".boundary" << std::ends;
    mshP = new concepts::Import3dMesh(coordName.str(), elmsName.str(),
				      boundaryName.str(), 1, false);
    logger.log("+ mesh =");
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, "(0)"));
    logger.log("+ boundary conditions = ");
    linDG3D::FvdgSpace* spcP = 0;
    spcP = new linDG3D::FvdgSpaceP1(*mshP);
    logger.log("+ space =");
    // save mesh for dx
    fileName.str("");
    fileName << fileNameBase << "_mesh.dx" << std::ends;
    graphics::MeshDX<Real>(*spcP, fileName.str(), 3);
    // set up mass matrix
    linDG3D::Identity id;
    concepts::DiagonalMatrix<Real> mass(*spcP);
    concepts::Matrix<Real>::assembly(mass, *spcP, id);
    logger.log("+ mass matrix =");
    // set up advection
    Upwind<BurgerFlux> numfluxBU;
    LaxFriedrichs<BurgerFlux> numfluxBL;
    Osher<BurgerFlux> numfluxBO;
    Godunov<BurgerFlux> numfluxBG;
    concepts::Operator<Real>* advectionP = 0;
    switch (fluxNbr) {
    case 2:
      advectionP = new Advection<Upwind<BurgerFlux> >(numfluxBU, *spcP,
						      false, bc.get());
      break;
    case 3: 
      advectionP = new Advection<LaxFriedrichs<BurgerFlux> >(numfluxBL, *spcP,
							     false, bc.get());
      break;
    case 4:
      advectionP = new Advection<Osher<BurgerFlux> >(numfluxBO, *spcP,
						     false, bc.get());
      break;
    case 5:
      advectionP = new Advection<Godunov<BurgerFlux> >(numfluxBG, *spcP,
						       false, bc.get());
      break;
    default:
      exit(-1);
    }
    concepts::Operator<Real>* limiterP = 0;
    switch (limNbr) {
    case 1: limiterP = new FvdgPi0Limiter(*spcP); break;
    case 2: limiterP = new FvdgAdaptiveLimiter(*spcP, alpha, 0, bc.get());
      break;
    case 3: limiterP = new FvdgPi0Limiter(*spcP); break;
    }
    concepts::Compose<Real>* limAdvectionP = 0;
    concepts::Compose<Real>* limAdvectionP2 = 0;
    concepts::Operator<Real>* advP;
    if (limiterP) {
      limAdvectionP = new concepts::Compose<Real>(*advectionP, *limiterP);
      if (limNbr == 3) {
	limAdvectionP2 = 
	  new concepts::Compose<Real>(*limiterP, *limAdvectionP);
	advP = limAdvectionP2;
      } else {
	advP = limAdvectionP;
      }
    } else {
      advP = advectionP;
    }

    logger.log("+ advection =");
    // set up diffusion
    concepts::SparseMatrix<Real>* diffP = 0;
    concepts::Operator<Real>* D0P;
    diffP = new concepts::SparseMatrix<Real>(*spcP, *spcP);
    linDG3D::LaplaceVolBF lapBfVol;
    linDG3D::LaplaceInnerBF lapBfInn(false);
    linDG3D::LaplaceBoundaryBF lapBfBnd(false);
    linDG3D::PenaltyBF penBf;
    concepts::Matrix<Real>::assembly(*diffP, *spcP, lapBfVol);
    linDG3D::Assembly::inner(*diffP, lapBfInn, spcP->innerElmPairList(),
			     true);
    linDG3D::Assembly::boundary(*diffP, lapBfBnd, spcP->boundaryElmPairList(), true);
    linDG3D::Assembly::inner(*diffP, penBf, spcP->innerElmPairList(), true);
    linDG3D::Assembly::boundary(*diffP, penBf, spcP->boundaryElmPairList(),
				true);
    logger.log("+ diffusion =");
    D0P = new concepts::LiCo<Real>(*advP, *diffP, 1.0, epsilon);
    // set up solution vector, right hand side and indicator
    concepts::Vector<Real> sol(*spcP);
    logger.log("+ solution vector =");
    ADLoad loadFrm(epsilon, ny);
    linDG3D::FvdgTimeRhsDiff driver_tlf(loadFrm, epsilon, bc.get(), false);
    timestepping::TimeVector driver(*spcP, driver_tlf);
    logger.log("+ right hand side =");
    concepts::Vector<Real> indicator(*spcP);
    logger.log("+ indicator =");
    ADSolution exactFrm(ny);
    // Set up initial condition
    linDG3D::FvdgTimeRhs rhs(exactFrm);
    timestepping::TimeVector initial(*spcP, rhs);
    concepts::DiagonalSolver<Real> diag(mass);
    diag(initial, initial);
    logger.log("+ initial vector =");
    // set up timestepping
    TimestepEstimator<LinearAdvectionFlux> tseL(*spcP, 1.0);
    TimestepEstimator<BurgerFlux> tseB(*spcP, 1.0);
    TimestepEstimatorParabolic tseP(0.027);
    Real dtA = (fluxNbr < 2 ? tseL(initial) : tseB(initial));
    Real dtP = tseP(*spcP)/epsilon;
//     Real dt = dtA < dtP ? dtA : dtP;
//     dt *= cfl;
    Real dt = (1.001/15)*std::pow(90./spcP->dim(),1./3);
    uint stride;

    // cfl-primat
//     if (nsteps == 0) {
//       stride = static_cast<uint>(std::ceil(tfinal/(15.0*dt)));
//       nsteps = static_cast<uint>(std::ceil(tfinal/dt));
//     } else {
//       stride = ((nsteps-1)/15)+1;
//     }
//     nsteps = stride*(((nsteps-1)/stride)+1);
//     tfinal = static_cast<Real>(nsteps)*dt;

    // tfinal-primat
    if (nsteps == 0) {
      if(tfinal != 0) {
	stride = static_cast<uint>(std::ceil(tfinal/(15.0*dt)));
	nsteps = 15*stride;
	dt = tfinal / nsteps;
      }
    } else {
      stride = ((nsteps-1)/15)+1;
      nsteps = stride*(((nsteps-1)/stride)+1);
    }
    outputParameters.addInt("basic: number of elements", spcP->nelm());
    outputParameters.addInt("basic: number of timesteps", nsteps);
    outputParameters.addDouble("basic: timestep", dt);
    outputParameters.addDouble("basic: final time", tfinal);
    outputParameters.addDouble("basic: timestep for writing", stride*dt);
    outputParameters.addDouble("basic: timestep cfl advection", dtA);
    outputParameters.addDouble("basic: timestep cfl diffusion", dtP);
    timestepping::TimeStepStrategy* tssP = 0;
    concepts::DiagonalSolverFabric<Real> fabric;
    if (limNbr == 2) {
      if (tvd_rk) {
	tssP = new timestepping::LimitingTvdRK2(fabric, mass, *D0P, *limiterP,
						driver, initial, dt);
      } else {
	tssP = new timestepping::LimitingEuler(fabric, mass, *D0P, *limiterP,
					       driver, initial, dt);
      }
    } else {
      if (tvd_rk) {
	tssP = new timestepping::TvdRK2(fabric, mass, *D0P,
					driver, initial, dt);
      } else {
	tssP = new timestepping::Euler(fabric, mass, *D0P,
				       driver, initial, dt);
      }
    }
    DEBUGL(DEBUG_TEST, *tssP);
    timestepping::TimeStepping timeSolver(*tssP);
    logger.log("+ timestepping =");
    // set up output arrays
    outputParameters.addArrayDouble("snapshot: M-norm");
    outputParameters.addArrayDouble("snapshot: max-norm");
    outputParameters.addArrayDouble("snapshot: time");
    // do timestepping
    // create graphics object for numeric solution
    fileName.str("");
    fileName << fileNameBase << "_data.dx" << std::ends;
    graphics::DataDX<Real> dx(graphics::drawDataDX(*spcP, fileName.str(), 3));
    // create graphics object for exact solution (linear advection)
    fileName.str("");
    fileName << fileNameBase << "_exact.dx" << std::ends;
    graphics::DataDX<Real> dxEx(graphics::drawDataDX(*spcP,fileName.str(), 3));
    // create graphics object for limiter indicator
    fileName.str("");
    fileName << fileNameBase << "_ind.dx" << std::ends;
    graphics::DataDX<Real> 
      dxInd(graphics::drawDataDX(*spcP, fileName.str(), 3));
    // do some timing
    float uTime_setup = timer.utime();
    float uTime_timestepping = 0;
    float uTime_snapshots = 0;
    uint snapshot = 0;
    // timestepping loop
    for(uint n = 0;  n <= nsteps; n+=stride, ++snapshot) {
      timeSolver(sol, n);
      DEBUGL(DEBUG_TEST, "limNbr = " << limNbr);
      if (limNbr == 2) {
	FvdgAdaptiveLimiter* adLim = 
	  dynamic_cast<FvdgAdaptiveLimiter*>(limiterP);
	conceptsAssert(adLim, concepts::Assertion());
	DEBUGL(DEBUG_TEST, "Before: " << indicator);
	adLim->getIndicator(indicator);
	DEBUGL(DEBUG_TEST, "After: " << indicator);
      }
      uTime_timestepping += timer.utime();
      // do test outputs
//       // cfl test output
//       std::cout << "cfl(" << n*dt << ") = " 
//                 << dt / (fluxNbr < 2 ? tseL(sol) : tseB(sol)) << std::endl;
      // ||u||L2 test output
//       std::cout << "||u||L2(" << n*dt << ") = " << Norm::matrix(sol, mass)
// 		<< std::endl;
      // ||u||Max test output
//       std::cout << "||u||Max(" << n*dt << ") = " << Norm::max(sol) 
// 		<< std::endl;
      // write dx file
      std::stringstream frameName;
      frameName << "t = " << std::setprecision(3)
		<< n*dt;
      dx(sol, frameName.str());
      if (limNbr == 2) {
	DEBUGL(DEBUG_TEST, frameName.str());
	dxInd(indicator, frameName.str());
      }
      // calculate error vector
      rhs.time(n*dt);
      timestepping::TimeVector tempTv(*spcP, rhs);
      diag(tempTv, tempTv); // projected exact solution
      concepts::Vector<Real>& temp = 
	dynamic_cast<concepts::Vector<Real>&> (tempTv);
      dxEx(temp, frameName.str());
      temp -= sol; // approximation error
      // error output
      outputParameters.addArrayDouble("snapshot: M-norm", snapshot,
				      Norm::matrix(temp, mass));
      outputParameters.addArrayDouble("snapshot: max-norm", snapshot,
				      Norm::max(temp));
      //    std::cout << "linadv error(" << n*dt << ") = " << Norm::max(temp) 
      //		<< std::endl;
      outputParameters.addArrayDouble("snapshot: time", snapshot,
				      n*dt);
      uTime_snapshots += timer.utime();
    } // end for timestepping loop
    logger.log("end");
    //doing time calculations
    outputParameters.addDouble("timer: setup", uTime_setup);
    outputParameters.addDouble("timer: setup per element",
			       uTime_setup / spcP->nelm());
    outputParameters.addDouble("timer: timestepping",
			       uTime_timestepping);
    outputParameters.addDouble("timer: timestepping per timestep",
			       uTime_timestepping / nsteps);
    outputParameters.addDouble("timer: timestepping per timestep and element",
			       uTime_timestepping / nsteps / spcP->nelm());
    outputParameters.addDouble("timer: snapshots",
			       uTime_snapshots);
    outputParameters.addDouble("timer: snapshots per snapshot",
			       uTime_snapshots / snapshot);
    outputParameters.addDouble("timer: snapshots per snapshot and element",
			       uTime_snapshots / snapshot / spcP->nelm());
    // writing log file
    fileName.str("");
    fileName << fileNameBase << "_log.txt" << std::ends;
    std::ofstream logfile(fileName.str().c_str());
    logfile << input;
    // delete objects on heap
    delete tssP;
    tssP = 0;
    delete limAdvectionP;
    limAdvectionP = 0;
    delete limiterP;
    limiterP = 0;
    delete advectionP;
    advectionP = 0;
    D0P = 0;
    delete spcP;
    spcP = 0;
    delete mshP;
    mshP = 0;
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}


// ******************************************************************** Main **
int main(int argc, char** argv) {
  std::srand(0);
  bool p0 = false;
  bool periodic = true;
  std::string meshNameBase = "mesh";
  uint n = 10;
  uint nsteps = 20;
  Real tfinal = 0;
  Real jitter = 0.0;
  Real cfl = 1.0;
  uint fluxNbr = 0;
  uint limNbr = 2;
  Real alpha = 5;
  Real epsilon = 0;
  Real ny = 0.1;
  bool tvd_rk = false;
  bool symm = false;
  std::string fileNameBase = "output";
  bool timeFlag = false;
  int opt;
  while ((opt = getopt(argc, argv, "-s:t:p:iI:n:j:c:l:L:a:e:N:rSf:")) != EOF)
    switch(opt) {
    case 's': 
      if (timeFlag) {
	std::cerr << "Error: -s and -t may not be specified at the same time"
		  << std::endl;
	std::exit(1);
      }
      nsteps = std::atoi(optarg);
      timeFlag = true;
      break;
    case 't':
      if (timeFlag) {
	std::cerr << "Error: -s and -t may not be specified at the same time"
		  << std::endl;
	std::exit(1);
      }
      tfinal = std::atof(optarg);
      nsteps = 0;
      timeFlag = true;
      break;
    case 'p': p0 = (std::atoi(optarg)!=1); break;
    case 'i': periodic = false; break;
    case 'I': periodic = false; meshNameBase = optarg; break;
    case 'n': n = std::atoi(optarg); break;
    case 'j': jitter = std::atof(optarg); break;
    case 'c': cfl = std::atof(optarg); break;
    case 'l': switch(256*optarg[0]+optarg[1]) {
    case 19532: fluxNbr = 1; break; // LL
    case 16981: fluxNbr = 2; break; // BU
    case 16972: fluxNbr = 3; break; // BL
    case 16975: fluxNbr = 4; break; // BO
    case 16967: fluxNbr = 5; break; // BG
    }; break;
    case 'L': limNbr = std::atoi(optarg); break;
    case 'a': alpha = std::atof(optarg); break;
    case 'e': epsilon = std::atof(optarg); break;
    case 'N': ny = std::atof(optarg); break;
    case 'r': tvd_rk = true; break;
    case 'S': symm = true; break;
    case 'f': fileNameBase = optarg; break;
    default:
      std::cout << "Usage: " << argv[0] << std::endl
		<< "  [-s STEPS] [-t T_FINAL] [-p P_DEGREE] [-i] [-n N]" 
		<< " [-j JITTER] [-c CFL]" << std::endl
		<< "  [-l FLUX] [-L LIM] [-a ALPHA] [-e EPSILON] [-N NY] [-r]"
		<< " [-f FILENAME]" << std::endl 
		<< "where" << std::endl
		<< "  -s STEPS    does STEPS timesteps" << std::endl
		<< "  -t T_FINAL  simulates until t>=T_FINAL" << std::endl
		<< "  -p P_DEGREE sets polynomial degree to P_DEGREE" 
		<< std::endl
		<< "  -i          uses imported mesh instead of periodic cube"
		<< std::endl
		<< "  -I MESH     as -i, but loads <MESH>.* instead of mesh.*"
		<< std::endl
		<< "  -n N        sets size of periodic cube mesh to 6*N^3"
		<< std::endl
		<< "  -j JITTER   moves meshpoints of periodic cube"
		<< std::endl
		<< "              by +/- JITTER element diameters"
		<< std::endl
		<< "  -c CFL      initial cfl-number to determine timestep"
		<< std::endl
		<< "  -l FLUX     chooses flux FLUX (LU, LL, BU, BL, BO, BG)"
		<< std::endl
		<< "  -L LIM      uses limiter LIM: 0=none, 1=Pi0, 2=adaptive, 3=double Pi0"
		<< std::endl
		<< "  -a ALPHA    sets parameter alpha for adaptive limiter"
		<< std::endl
		<< "  -e EPSILON  sets coefficient of diffusion to EPSILON"
		<< std::endl
		<< "  -N NY       sets parameter of exact solution to NY"
		<< std::endl
		<< "  -r          uses TVD-RungeKutta2 instead of Euler"
		<< std::endl
		<< "  -S          uses symmetric diffusion instead of antis."
		<< std::endl
		<< "  -f FILENAME saves data under <FILENAME>_*.*"
		<< std::endl;
      std::exit(1); break;
    }
  //cflTest(nsteps, tfinal, p0, periodic, n, jitter, cfl, fluxNbr,
  //		limNbr, alpha, tvd_rk, fileNameBase);
  // ellipticTest(fileNameBase, symm);
  // parabolicTest(nsteps, tfinal, periodic, n, jitter, cfl, tvd_rk, symm,
  // 		fileNameBase, meshNameBase);
  //advectionDiffusionTest(nsteps, tfinal, cfl,
  //			 fluxNbr, limNbr, alpha, epsilon, ny, tvd_rk,
  //			 fileNameBase, meshNameBase);
  advectionTest(nsteps, tfinal, p0, periodic, n, jitter, cfl,
  		fluxNbr, limNbr, alpha, tvd_rk, fileNameBase);
  //advectionDiffusionTest(nsteps, tfinal, cfl,
  //			 fluxNbr, limNbr, alpha, epsilon, ny, tvd_rk,
  //			 fileNameBase, meshNameBase);

  return 0;
}
