#include <fstream>

#include "periodicube.hh"

PeriodiCube::PeriodiCube(uint n, Real jitter) : n_(n), jitter_(jitter) {
  if ( (jitter_ < -0.4) || (jitter > 0.4) ) {
    jitter_ = 0.4;
  }
  jitter_ /= static_cast<Real>(n_);
  // arrays
  coord_ = new concepts::Real3d*[(n_+1)*(n_+1)*(n_+1)];
  vtx_ = new concepts::Vertex*[n_*n_*n_];
  edg_ = new concepts::Edge*[7*n_*n_*n_];
  tri_ = new concepts::Triangle*[12*n_*n_*n_];
  tet_ = new concepts::Tetrahedron*[6*n_*n_*n_];
  cell_ = new concepts::Tetrahedron3d*[6*n_*n_*n_];
  // coordinates
  for (uint a=0; a<=n_; ++a) {
    for (uint b=0; b<=n_; ++b) {
      for (uint c=0; c<=n_; ++c) {
	  coord_[ii(a,b,c)-1] = new concepts::Real3d(x(a), x(b), x(c));
	if( (jitter_!=0) && (a%n_!=0) && (b%n_!=0) && (c%n_!=0) ) {
	  *coord_[ii(a,b,c)-1] += concepts::Real3d(dRand(), dRand(), dRand());
	}
      }
    }
  }
  // vertices
  for (uint i = 0; i < n_*n_*n_; ++i)
    vtx_[i] = new concepts::Vertex();
  // edges
  for(uint a=0; a<n_; ++a) {
    for(uint b=0; b<n_; ++b) {
      for(uint c=0; c<n_; ++c) {
	edg_[k(a,b,c,0)] = 
	  new concepts::Edge(*vtx_[j(a,b,c)], *vtx_[j(a+1,b,c)]);
	edg_[k(a,b,c,1)] = 
	  new concepts::Edge(*vtx_[j(a,b,c)], *vtx_[j(a,b+1,c)]);
	edg_[k(a,b,c,2)] = 
	  new concepts::Edge(*vtx_[j(a,b,c)], *vtx_[j(a,b,c+1)]);
	edg_[k(a,b,c,3)] = 
	  new concepts::Edge(*vtx_[j(a,b,c)], *vtx_[j(a+1,b+1,c)]);
	edg_[k(a,b,c,4)] = 
	  new concepts::Edge(*vtx_[j(a,b,c+1)], *vtx_[j(a,b+1,c)]);
	edg_[k(a,b,c,5)] = 
	  new concepts::Edge(*vtx_[j(a,b,c+1)], *vtx_[j(a+1,b,c)]);
	edg_[k(a,b,c,6)] = 
	  new concepts::Edge(*vtx_[j(a,b,c+1)], *vtx_[j(a+1,b+1,c)]);
      }
    }
  }

  // Faces
  for(uint a=0; a<n_; ++a) {
    for(uint b=0; b<n_; ++b) {
      for(uint c=0; c<n_; ++c) {
	tri_[l(a,b,c,0)] = new concepts::Triangle(*edg_[k(a,b,c,0)],
						  *edg_[k(a,b,c,3)],
						  *edg_[k(a+1,b,c,1)]);
	tri_[l(a,b,c,1)] = new concepts::Triangle(*edg_[k(a,b,c,1)],
						  *edg_[k(a,b,c,3)],
						  *edg_[k(a,b+1,c,0)]);
	tri_[l(a,b,c,2)] = new concepts::Triangle(*edg_[k(a,b,c,1)],
						  *edg_[k(a,b,c,2)],
						  *edg_[k(a,b,c,4)]);
	tri_[l(a,b,c,3)] = new concepts::Triangle(*edg_[k(a,b,c,4)],
						  *edg_[k(a,b+1,c,2)],
						  *edg_[k(a,b,c+1,1)]);
	tri_[l(a,b,c,4)] = new concepts::Triangle(*edg_[k(a,b,c,0)],
						  *edg_[k(a,b,c,2)],
						  *edg_[k(a,b,c,5)]);
	tri_[l(a,b,c,5)] = new concepts::Triangle(*edg_[k(a,b,c+1,0)],
						  *edg_[k(a+1,b,c,2)],
						  *edg_[k(a,b,c,5)]);
	tri_[l(a,b,c,6)] = new concepts::Triangle(*edg_[k(a,b,c+1,0)],
						  *edg_[k(a+1,b,c,4)],
						  *edg_[k(a,b,c,6)]);
        tri_[l(a,b,c,7)] = new concepts::Triangle(*edg_[k(a,b+1,c,0)],
						  *edg_[k(a,b,c,4)],
						  *edg_[k(a,b,c,6)]);
	tri_[l(a,b,c,8)] = new concepts::Triangle(*edg_[k(a,b,c+1,1)],
						  *edg_[k(a,b+1,c,5)],
						  *edg_[k(a,b,c,6)]);
	tri_[l(a,b,c,9)] = new concepts::Triangle(*edg_[k(a+1,b,c,1)],
						  *edg_[k(a,b,c,5)],
						  *edg_[k(a,b,c,6)]);
	tri_[l(a,b,c,10)] = new concepts::Triangle(*edg_[k(a,b,c,2)],
						   *edg_[k(a,b,c,3)],
						   *edg_[k(a,b,c,6)]);
	tri_[l(a,b,c,11)] = new concepts::Triangle(*edg_[k(a+1,b+1,c,2)],
						   *edg_[k(a,b,c+1,3)],
						   *edg_[k(a,b,c,6)]);
      }
    }
  }

  // Tets
  for(uint a=0; a<n_; ++a) {
    for(uint b=0; b<n_; ++b) {
      for(uint c=0; c<n_; ++c) {
	tet_[m(a,b,c,0)] = new concepts::Tetrahedron(*tri_[l(a,b,c,0)],
						     *tri_[l(a,b,c,4)],
						     *tri_[l(a,b,c,9)],
						     *tri_[l(a,b,c,10)]);
	tet_[m(a,b,c,1)] = new concepts::Tetrahedron(*tri_[l(a,b,c,9)],
						     *tri_[l(a,b,c,5)],
						     *tri_[l(a+1,b,c,2)],
						     *tri_[l(a,b,c,6)]);
	tet_[m(a,b,c,2)] = new concepts::Tetrahedron(*tri_[l(a,b,c,6)],
						     *tri_[l(a,b,c+1,0)],
						     *tri_[l(a+1,b,c,3)],
						     *tri_[l(a,b,c,11)]);
	tet_[m(a,b,c,3)] = new concepts::Tetrahedron(*tri_[l(a,b,c,1)],
						     *tri_[l(a,b,c,2)],
						     *tri_[l(a,b,c,10)],
						     *tri_[l(a,b,c,7)]);
	tet_[m(a,b,c,4)] = new concepts::Tetrahedron(*tri_[l(a,b,c,7)],
						     *tri_[l(a,b+1,c,4)],
						     *tri_[l(a,b,c,3)],
						     *tri_[l(a,b,c,8)]);
	tet_[m(a,b,c,5)] = new concepts::Tetrahedron(*tri_[l(a,b,c,8)],
						     *tri_[l(a,b+1,c,5)],
						     *tri_[l(a,b,c+1,1)],
						     *tri_[l(a,b,c,11)]);
      }
    }
  }

  // Cells
  for(uint a=0; a<n_; ++a) {
    for(uint b=0; b<n_; ++b) {
      for(uint c=0; c<n_; ++c) {
        cell_[m(a,b,c,0)] = new concepts::Tetrahedron3d(*tet_[m(a,b,c,0)],
	  concepts::MapTetrahedron3d(*(coord_[ii(a, b, c)-1]),
				     *(coord_[ii(a+1, b, c)-1]),
				     *(coord_[ii(a+1, b+1, c)-1]),
				     *(coord_[ii(a, b, c+1)-1])));
        cell_[m(a,b,c,1)] = new concepts::Tetrahedron3d(*tet_[m(a,b,c,1)],
	  concepts::MapTetrahedron3d(*(coord_[ii(a, b, c+1)-1]),
				     *(coord_[ii(a+1, b, c)-1]),
				     *(coord_[ii(a+1, b+1, c)-1]),
				     *(coord_[ii(a+1, b, c+1)-1])));
        cell_[m(a,b,c,2)] = new concepts::Tetrahedron3d(*tet_[m(a,b,c,2)],
	  concepts::MapTetrahedron3d(*(coord_[ii(a, b, c+1)-1]),
				     *(coord_[ii(a+1, b, c+1)-1]),
				     *(coord_[ii(a+1, b+1, c)-1]),
				     *(coord_[ii(a+1, b+1, c+1)-1])));
        cell_[m(a,b,c,3)] = new concepts::Tetrahedron3d(*tet_[m(a,b,c,3)],
	  concepts::MapTetrahedron3d(*(coord_[ii(a, b+1, c)-1]),
				     *(coord_[ii(a, b, c)-1]),
				     *(coord_[ii(a+1, b+1, c)-1]),
				     *(coord_[ii(a, b, c+1)-1])));
        cell_[m(a,b,c,4)] = new concepts::Tetrahedron3d(*tet_[m(a,b,c,4)],
	  concepts::MapTetrahedron3d(*(coord_[ii(a+1, b+1, c)-1]),
				     *(coord_[ii(a, b+1, c)-1]),
				     *(coord_[ii(a, b, c+1)-1]),
				     *(coord_[ii(a, b+1, c+1)-1])));
        cell_[m(a,b,c,5)] = new concepts::Tetrahedron3d(*tet_[m(a,b,c,5)],
	  concepts::MapTetrahedron3d(*(coord_[ii(a+1, b+1, c)-1]),
				     *(coord_[ii(a, b+1, c+1)-1]),
				     *(coord_[ii(a, b, c+1)-1]),
				     *(coord_[ii(a+1, b+1, c+1)-1])));
      }
    }
  }
}

PeriodiCube::~PeriodiCube() {
  for (uint i = 0; i < 6*n_*n_*n_; ++i)
    delete cell_[i];
  delete [] cell_;
  for (uint i = 0; i < 6*n_*n_*n_; ++i)
    delete tet_[i];
  delete [] tet_;
  for (uint i = 0; i < 12*n_*n_*n_; ++i)
    delete tri_[i];
  delete [] tri_;
  for (uint i = 0; i < 7*n_*n_*n_; ++i)
    delete edg_[i];
  delete [] edg_;
  for (uint i = 0; i < n_*n_*n_; ++i)
    delete vtx_[i];
  delete [] vtx_;
  for (uint i = 0; i < (n_+1)*(n_+1)*(n_+1); ++i)
    delete coord_[i];
  delete [] coord_;
}

std::ostream& PeriodiCube::info(std::ostream& os) const {
  os << "PeriodiCube(n = " << n_ << ")";
  return os;
}

void PeriodiCube::saveMesh() {
  std::ofstream file( "periodicube.mesh" );
  conceptsAssert(file != 0, concepts::Assertion());
  // write vertices
  file << (n_+1)*(n_+1)*(n_+1) << std::endl;
  uint i = 1;
  concepts::Real3d* vtx = 0;
  for(uint c=0; c<=n_; ++c) {
    for(uint b=0; b<=n_; ++b) {
      for(uint a=0; a<=n_; ++a, ++i) {
	DEBUGL(DEBUG_SAVE_MESH, ii(a,b,c) << " " << i);
	vtx = coord_[ii(a,b,c)-1];
	file << "  " << (*vtx)[0] << " " << (*vtx)[1] << " " << (*vtx)[2]
	     << std::endl;
      }
    }
  }
  // write tetrahedrons
  file << 6*n_*n_*n_ << std::endl;
  i = 0;
  for(uint c=0; c<n_; ++c) {
    for(uint b=0; b<n_; ++b) {
      for(uint a=0; a<n_; ++a, ++i) {
        // tetrahedron 0
	file << "   1  ";
	file << "        " << ii(a+1, b, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b, c) +
	       cell_[i]->connector().vertex(1)->key());
	file << "        " << ii(a, b, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b, c) +
	       cell_[i]->connector().vertex(0)->key());
	file << "        " << ii(a+1, b+1, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b+1, c) +
	       cell_[i]->connector().vertex(2)->key());
	file << "        " << ii(a, b, c+1) << std::endl;
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b, c+1) +
	       cell_[i]->connector().vertex(3)->key());
        ++i;
        // tetrahedron 1
	file << "   1  ";
	file << "        " << ii(a+1, b, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b, c) +
	       cell_[i]->connector().vertex(1)->key());
	file << "        " << ii(a, b, c+1);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b, c+1) +
	       cell_[i]->connector().vertex(0)->key());
	file << "        " << ii(a+1, b+1, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b+1, c) +
	       cell_[i]->connector().vertex(2)->key());
	file << "        " << ii(a+1, b, c+1) << std::endl;
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b, c+1) +
	       cell_[i]->connector().vertex(3)->key());
        ++i;
        // tetrahedron 2
	file << "   1  ";
	file << "        " << ii(a+1, b, c+1);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b, c+1) +
	       cell_[i]->connector().vertex(1)->key());
	file << "        " << ii(a, b, c+1);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b, c+1) +
	       cell_[i]->connector().vertex(0)->key());
	file << "        " << ii(a+1, b+1, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b+1, c) +
	       cell_[i]->connector().vertex(2)->key());
	file << "        " << ii(a+1, b+1, c+1) << std::endl;
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b+1, c+1) +
	       cell_[i]->connector().vertex(3)->key());
        ++i;
        // tetrahedron 3
	file << "   1  ";
	file << "        " << ii(a, b, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b, c) +
	       cell_[i]->connector().vertex(1)->key());
	file << "        " << ii(a, b+1, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b+1, c) +
	       cell_[i]->connector().vertex(0)->key());
	file << "        " << ii(a+1, b+1, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b+1, c) +
	       cell_[i]->connector().vertex(2)->key());
	file << "        " << ii(a, b, c+1) << std::endl;
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b, c+1) +
	       cell_[i]->connector().vertex(3)->key());
        ++i;
        // tetrahedron 4
	file << "   1  ";
	file << "        " << ii(a, b+1, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b+1, c) +
	       cell_[i]->connector().vertex(1)->key());
	file << "        " << ii(a+1, b+1, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b+1, c) +
	       cell_[i]->connector().vertex(0)->key());
	file << "        " << ii(a, b, c+1);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b, c+1) +
	       cell_[i]->connector().vertex(2)->key());
	file << "        " << ii(a, b+1, c+1) << std::endl;
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b+1, c+1) +
	       cell_[i]->connector().vertex(3)->key());
        ++i;
        // tetrahedron 5
	file << "   1  ";
	file << "        " << ii(a, b+1, c+1);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b+1, c+1) +
	       cell_[i]->connector().vertex(1)->key());
	file << "        " << ii(a+1, b+1, c);
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b+1, c) +
	       cell_[i]->connector().vertex(0)->key());
	file << "        " << ii(a, b, c+1);
	DEBUGL(DEBUG_SAVE_MESH, -j(a, b, c+1) +
	       cell_[i]->connector().vertex(2)->key());
	file << "        " << ii(a+1, b+1, c+1) << std::endl;
	DEBUGL(DEBUG_SAVE_MESH, -j(a+1, b+1, c+1) +
	       cell_[i]->connector().vertex(3)->key());
      }
    }
  }
  // write boundary
  file << 12*n_*n_ << std::endl;
  for(uint q=0; q<n_; ++q) {
    for(uint p=0; p<n_; ++p) {
      file << "   1    ";
      file << "        " << ii(0,p,q);
      file << "        " << ii(0,p,q+1);
      file << "        " << ii(0,p+1,q) << std::endl;

      file << "   1    ";
      file << "        " << ii(0,p+1,q);
      file << "        " << ii(0,p,q+1);
      file << "        " << ii(0,p+1,q+1) << std::endl;

      file << "   2    ";
      file << "        " << ii(n_,p,q);
      file << "        " << ii(n_,p+1,q);
      file << "        " << ii(n_,p,q+1) << std::endl;

      file << "   2    ";
      file << "        " << ii(n_,p+1,q);
      file << "        " << ii(n_,p+1,q+1);
      file << "        " << ii(n_,p,q+1) << std::endl;

      file << "   3    ";
      file << "        " << ii(p,0,q);
      file << "        " << ii(p+1,0,q);
      file << "        " << ii(p,0,q+1) << std::endl;

      file << "   3    ";
      file << "        " << ii(p,0,q+1);
      file << "        " << ii(p+1,0,q);
      file << "        " << ii(p+1,0,q+1) << std::endl;

      file << "   4    ";
      file << "        " << ii(p,n_,q);
      file << "        " << ii(p,n_,q+1);
      file << "        " << ii(p+1,n_,q) << std::endl;

      file << "   4    ";
      file << "        " << ii(p,n_,q+1);
      file << "        " << ii(p+1,n_,q+1);
      file << "        " << ii(p+1,n_,q) << std::endl;

      file << "   5    ";
      file << "        " << ii(p,q,0);
      file << "        " << ii(p+1,q+1,0);
      file << "        " << ii(p+1,q,0) << std::endl;

      file << "   5    ";
      file << "        " << ii(p,q,0);
      file << "        " << ii(p,q+1,0);
      file << "        " << ii(p+1,q+1,0) << std::endl;

      file << "   6    ";
      file << "        " << ii(p,q,n_);
      file << "        " << ii(p+1,q,n_);
      file << "        " << ii(p+1,q+1,n_) << std::endl;

      file << "   6    ";
      file << "        " << ii(p,q,n_);
      file << "        " << ii(p+1,q+1,n_);
      file << "        " << ii(p,q+1,n_) << std::endl;
    }
  }
}
