#include "utility.hh"

// ******************************************************************** Norm **
Real Norm::max(const concepts::Vector<Real>& coeffs) {
  Real maxNorm = 0;
  try {
    const linDG3D::FvdgSpace& spc =
      dynamic_cast<const linDG3D::FvdgSpace&>(coeffs.space());
    std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc.scan());
    Real uLoc = 0;
    while (*elemSc) {
      const linDG3D::FvdgElement& elm = (*elemSc)++;
      for (uint i=0; i<4; ++i) {
	uLoc = std::fabs(elm.solutionVertex(coeffs, i));
	maxNorm = uLoc > maxNorm ? uLoc : maxNorm;
      }
    }
  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
  return maxNorm;
}

Real Norm::matrix(const concepts::Vector<Real>& coeffs,
		  concepts::Operator<Real>& matrix) {
  concepts::Vector<Real> temp(coeffs);
  matrix(coeffs, temp);
  return std::sqrt(coeffs*temp);
}

Real Norm::broMax(const concepts::Vector<Real>& coeffs,
		  concepts::Formula& frm, Real time = 0) {
  Real norm = 0;
  try {
    Real temp;
    const linDG3D::FvdgSpace& spc = 
      dynamic_cast<const linDG3D::FvdgSpace&>(coeffs.space());
    std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc.scan());

    while (*elemSc) {
      const linDG3D::FvdgElement& elm = (*elemSc)++;
      // 1. Loop over egde centers
      for (uint j = 0; j < 6; ++j) {
	temp = std::fabs(elm.solutionEdgeCenter(coeffs, j) - 
			 frm(elm.edgeCenter(j), time));
	norm = norm > temp ? norm : temp;
      }
      // 2. Loop over vertices
      for (uint j = 0; j < 4; ++j) {
	temp = std::fabs(elm.solutionVertex(coeffs, j) -
			 frm(elm.vertex(j), time));
	norm = norm > temp ? norm : temp;
      }
    }
    DEBUGL(DEBUG_NORM_BROL1, "norm = " << norm);
  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
  return norm;
}


Real Norm::broL1(const concepts::Vector<Real>& coeffs,
		 concepts::Formula& frm, Real time = 0) {
  Real norm = 0;
  try {
    concepts::Real3d v1, v2, v3;
    const linDG3D::FvdgSpace& spc = 
      dynamic_cast<const linDG3D::FvdgSpace&>(coeffs.space());
    std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc.scan());

    while (*elemSc) {
      const linDG3D::FvdgElement& elm = (*elemSc)++;
      Real integral = 0;
      Real temp;
      // 1. Loop over egde centers
      for (uint j = 0; j < 6; ++j) {
	temp = elm.solutionEdgeCenter(coeffs, j)- frm(elm.edgeCenter(j), time);
	integral += std::fabs(temp);
      }
      integral *= 4.;
      // 2. Loop over vertices
      for (uint j = 0; j < 4; ++j) {
	temp = elm.solutionVertex(coeffs, j) - frm(elm.vertex(j), time);
	integral -= std::fabs(temp);
      }
      // Calculation of the volume
      v1 = elm.cell().vertex(1) - elm.cell().vertex(0);
      v2 = elm.cell().vertex(2) - elm.cell().vertex(0);
      v3 = elm.cell().vertex(3) - elm.cell().vertex(0);
      const concepts::Mapping<Real, 3> B(v1, v2, v3);
      const Real vol = B.determinant()/6.;

      norm += integral*vol/20;
    }
    DEBUGL(DEBUG_NORM_BROL1, "norm = " << norm);
  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
  return norm;
}

Real Norm::broL2Grad(const concepts::Vector<Real>& coeffs,
		     concepts::Formula& gradX,
		     concepts::Formula& gradY,
		     concepts::Formula& gradZ) {
  Real norm = 0;
  try {
    const linDG3D::FvdgSpace& spc = 
      dynamic_cast<const linDG3D::FvdgSpace&>(coeffs.space());
    std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc.scan());

    concepts::Real3d fvGradient, v1, v2, v3;
    concepts::Array<Real> localCoeffs;

    while (*elemSc) {
      const linDG3D::FvdgElement& elm = (*elemSc)++;
      // Build FV-Gradient
      fvGradient *=0;
      elm.T().extract(coeffs, localCoeffs);
      for(uint j = 0; j<elm.T().m(); ++j) {
	fvGradient += elm.shapeFctGradient(j) * localCoeffs[j];
      }
      
      Real integral = 0;
      concepts::Real3d diff;
      // 1. Loop over egde centers
      for (uint j = 0; j < 6; ++j) {
	diff[0] = gradX(elm.edgeCenter(j));
	diff[1] = gradY(elm.edgeCenter(j));
	diff[2] = gradZ(elm.edgeCenter(j));
	diff -= fvGradient;
	integral += diff.l2_2();
      }
      integral *= 4.;
      // 2. Loop over vertices
      for (uint j = 0; j < 4; ++j) {
	diff[0] = gradX(elm.vertex(j));
	diff[1] = gradY(elm.vertex(j));
	diff[2] = gradZ(elm.vertex(j));
	diff -= fvGradient;
	integral -= diff.l2_2();
      }
      // Calculation of the volume
      v1 = elm.cell().vertex(1) - elm.cell().vertex(0);
      v2 = elm.cell().vertex(2) - elm.cell().vertex(0);
      v3 = elm.cell().vertex(3) - elm.cell().vertex(0);
      const concepts::Mapping<Real, 3> B(v1, v2, v3);
      const Real vol = B.determinant()/6.;

      norm += integral*vol/20;
    }
    norm = std::sqrt(norm);
  } catch (std::bad_cast& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
  return norm;
}


// ******************************************************* TimestepEstimator **
template <typename T>
Real TimestepEstimator<T>::operator()(const concepts::Vector<Real>& sol)const {
  std::auto_ptr<typename concepts::Scan<linDG3D::FvdgElement> >
    sc(spc_.scan());
  Real u = 0;
  Real diam = 0;
  Real temp = 0;
  Real invTimestep = 0;
  while (*sc) {
    const linDG3D::FvdgElement& elm = (*sc)++;
    concepts::Real3d v1 = elm.cell().vertex(1) - elm.cell().vertex(0);
    concepts::Real3d v2 = elm.cell().vertex(2) - elm.cell().vertex(0);
    concepts::Real3d v3 = elm.cell().vertex(3) - elm.cell().vertex(0);
    const concepts::Mapping<Real, 3> B(v1, v2, v3);
    const Real vol = B.determinant()/6.;
    diam = std::pow(vol, 1./3.);
    for (uint j = 0; j < 4; ++j) {
      u = elm.solutionVertex(sol, j);
      temp = std::fabs((fluxFct_.deriv(u)).l2())/ diam;
      if (temp > invTimestep) invTimestep = temp;
    }
  }
  return c_/invTimestep;
}

template <typename T>
Real TimestepEstimator<T>::testOld(const concepts::Vector<Real>& sol)const {
  std::auto_ptr<concepts::Scan<concepts::ElementPair<Real> > >
    sc(spc_.innerElmPairList().scan());
  Real u = 0;
  Real temp = 0;
  Real invTimestep = 0;
  while (*sc) {
    concepts::ElementPair<Real>* ep = &(*sc)++;
    const linDG3D::FvdgElementPair* fep = 
      dynamic_cast<const linDG3D::FvdgElementPair*>(ep);
    conceptsAssert(fep, concepts::Assertion());
    for(uint i=0; i<fep->numPairs(); ++i) {
      u = fep->elm1().solutionEdgeCenter(sol, fep->edgeIndex(0,i));
      temp = std::fabs(fluxFct_.deriv(u, fep->normal())) / fep->diam();
      invTimestep = temp < invTimestep ? invTimestep : temp;
      u = fep->elm2().solutionEdgeCenter(sol, fep->edgeIndex(1,i));
      temp = std::fabs(fluxFct_.deriv(u, fep->normal())) / fep->diam();
      invTimestep = temp < invTimestep ? invTimestep : temp;
    }
  }
  sc.reset(spc_.boundaryElmPairList().scan());
  while (*sc) {
    concepts::ElementPair<Real>* ep = &(*sc)++;
    const linDG3D::FvdgElementPair* fep =
      dynamic_cast<const linDG3D::FvdgElementPair*>(ep);
    conceptsAssert(fep, concepts::Assertion());
    for(uint i=0; i<fep->numPairs(); ++i) {
      u = fep->elm1().solutionEdgeCenter(sol, fep->edgeIndex(0,i));
      temp = std::fabs(fluxFct_.deriv(u, fep->normal())) / fep->diam();
      invTimestep = temp < invTimestep ? invTimestep : temp;
    }
  }
  return c_ / invTimestep;
}

// ********************************************** TimestepEstimatorParabolic **
Real TimestepEstimatorParabolic::operator()(const linDG3D::FvdgSpace& spc) 
  const {
  Real h2 = 1e200;
  concepts::Real3d vertices[4];
  concepts::Real3d edge;
  Real e2;
  std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc.scan());
  while (*elemSc) {
    const linDG3D::FvdgElement& elm = (*elemSc)++;
    for (uint i=0; i<4; ++i) {
      vertices[i] = elm.vertex(i);
    }
    for (uint k=0; k<3; ++k) {
      for(uint l=k+1; l<4; ++l) {
	edge = vertices[k] - vertices[l];
	e2 = edge.l2_2();
	h2 = h2 < e2 ? h2 : e2;
      }
    }
  }
  return c_ * h2;
}

// ************************************************************* BurgerError **

void NodalBurgerError::operator() (const concepts::Vector<Real>& sol,
				   concepts::Array<Real>& error,
				   Real time) {
  // Get scanner
  std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > elemSc(spc_.scan());
  conceptsAssert(elemSc.get(), concepts::Assertion());
  // Vector for nodal solution and coordinates
  std::vector<std::pair<Real, concepts::Real3d> > nodal;
  // Loop over all elements
  while (*elemSc) {
    const linDG3D::FvdgElement& elm = (*elemSc)++;
    // Calculate nodal values
    for (uint i = 0; i < 4; ++i) {
      nodal.push_back( std::make_pair(elm.solutionVertex(sol, i),
				      elm.cell().vertex(i)) );
    }
  }
  // Resize array
  error.resize(nodal.size());
  for (uint i = 0; i < nodal.size(); ++i) {
    concepts::Real3d x0 = nodal[i].second - normal_*(nodal[i].first*time);
    error[i] = nodal[i].first - initCondition_(x0, 0);
  }
}

// Specialization
template class TimestepEstimator<BurgerFlux>;
template class TimestepEstimator<LinearAdvectionFlux>;
