#ifndef utility_hh
#define utility_hh

#include "linDG3D/space.hh"
#include "linDG3D/elementPair.hh"
#include "function/vector.hh"
#include "toolbox/scannerConnectors.hh"
#include "operator/compositions.hh"
#include "advection.hh"

#define DEBUG_NORM_BROL1 0

using concepts::Real;

// ******************************************************************** Norm **

/** Provides methods to calculate norms
    @author Christoph Winkelmann, 2003
*/
class Norm
{
public:
  /** Calculates the maximum norm of \c coeffs, which has to be in an
      FvdgSpace.
      @param coeffs The coefficent vector
  */
  static Real max(const concepts::Vector<Real>& coeffs);
  /** Calculates the norm defined by \c matrix of \c coeffs
      @param coeffs The coefficent vector
      @param matrix The matrix which defines the norm
  */
  static Real matrix(const concepts::Vector<Real>& coeffs,
		     concepts::Operator<Real>& matrix);
  /** Calculates the broken max norm of the difference between
      an FV solution given by \c coeffs and the exact solution given by
      \c frm by quadrature.
      @param coeffs The coefficent vector of the FV solution
      @param frm The exact solution
      @param time The time to evaluate \c frm at
  */
  static Real broMax(const concepts::Vector<Real>& coeffs,
		     concepts::Formula& frm, Real time);
  /** Calculates the broken L1 norm of the difference between
      an FV solution given by \c coeffs and the exact solution given by
      \c frm by quadrature. The quadrature formula used is exact for functions
      up to total degree 2.
      @param coeffs The coefficent vector of the FV solution
      @param frm The exact solution
      @param time The time to evaluate \c frm at
  */
  static Real broL1(const concepts::Vector<Real>& coeffs,
		    concepts::Formula& frm, Real time);
  /** Calculates the broken L2 norm of the gradient of the difference between
      an FV solution given by \c coeffs and the exact solution given by
      \c gradX, \c gradY and \c gradZ by quadrature. The quadrature formula
      used is exact for functions up to total degree 2.
      @param coeffs The coefficent vector of the FV solution
      @param gradX The x component of the exact gradient
      @param gradY The y component of the exact gradient
      @param gradZ The z component of the exact gradient
  */
  static Real broL2Grad(const concepts::Vector<Real>& coeffs,
			concepts::Formula& gradX,
			concepts::Formula& gradY,
			concepts::Formula& gradZ);
private:
  /// Constructor. Is private in order to prevent creating an instance.
  Norm() {};
};


// ******************************************************* TimestepEstimator **
/** Timestep Estimator. Gives an estimate for the maximal timestep
    @author Christoph Winkelmann, 2002
*/
template <typename T>
class TimestepEstimator {
public:
  /** Constructor.
      @param spc space to estimate the timestep for
      @param c constant in the CFL condition
  */
  TimestepEstimator(const linDG3D::FvdgSpace& spc, const Real c) : 
    spc_(spc), fluxFct_(), c_(c) {};
  /** Application operator.
      @param sol Solution vector to estimate the timestep for
  */
  Real operator()(const concepts::Vector<Real>& sol) const;
  Real testOld(const concepts::Vector<Real>& sol) const;
private:
  const linDG3D::FvdgSpace& spc_;
  const T fluxFct_;
  const Real c_;
};


// ********************************************** TimestepEstimatorParabolic **
/** Timestep Estimator for parabolic problems. Gives an estimate for the
    maximal timestep
    @author Christoph Winkelmann, 2003
*/
class TimestepEstimatorParabolic {
public:
  /** Constructor.
      @param c constant in the CFL condition
  */
  TimestepEstimatorParabolic(const Real c) : c_(c) {};
  /** Application operator.
      @param spc Space to estimate the timestep for
  */
  Real operator()(const linDG3D::FvdgSpace& spc) const;
private:
  const Real c_;
};


// ************************************************************* BurgerError **
class NodalBurgerError {
private:
  concepts::Formula& initCondition_;
  /// The scanner of the space belonging to the solution vector
  linDG3D::FvdgSpace& spc_;
  /// The normal vector of the Burger flux
  concepts::Real3d normal_;
public:
  /// Constructor
  NodalBurgerError(concepts::Formula& initCondition,
		   linDG3D::FvdgSpace& spc) :
    initCondition_(initCondition),
    spc_(spc),
    normal_(concepts::Real3d(1., 1., 1.)) {}

  /** Application operator. Takes a solution vector belonging to the space \c 
      spc and calculates a pointwise error on the mesh's nodes in the 
      following way: The solution in each node is followed backwards in time
      using characteristics to obtain a virtual initial condition. The error is
      calculated comparing this virtual initial condition to the one that was
      originally specified in \c initCondition.
      @param sol Solution vector
      @param error Error vector (nodal)
      @param time Physical time in simulation
  */
  void operator() (const concepts::Vector<Real>& sol,
		   concepts::Array<Real>& error,
		   Real time);
};

#endif // utility_hh
