/* System include files. The first is for input and output operations and
   the second one declares <span class="code">auto_ptr</span>.
 */
#include <sstream>
#include <iostream>
#include <memory>
#include <unistd.h> // for command line parsing
/* Concepts include files.
 */
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
#include "linearFEM.hh"
#include "graphics.hh"
#include "timestepping.hh"

/* With this using directive, we do not need to prepend
   <span class="code">concepts::</span> before <span class="code">Real</span>
   everytime.
*/
using concepts::Real;

/* <h3>Mesh</h3>

   The mesh is the interval from 1 to 5 with two cells.  One
   is <span class="math">(1,2)</span> and the other is <span
   class="math">(2,5)</span>. They are coupled in the class Line
   which makes it possible to loop over the cells.
*/
// ******************************************************************** Mesh **
#include "../applications/meshes.cc"

/* <h3>Linear Form</h3>

   The linear form is used to compute the element contributions to the
   load vector. The computations are done in the application operator
   of this class.
 */
// ************************************************************* Linear Form **

class RHS : public timestepping::TimeLinearForm {
public:
  RHS(Real c = 1.0) : c_(c) {}
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();
    const linearFEM::Quad* eQ = dynamic_cast<const linearFEM::Quad*>(&elm);
    if (eQ) { this->operator()(*eQ, em); return; }
    const linearFEM::Triangle* eT =
      dynamic_cast<const linearFEM::Triangle*>(&elm);
    if (eT) { this->operator()(*eT, em); return; }
    std::stringstream error;
    error << "element " << elm << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }
  void operator()(const linearFEM::Quad& elm,
		  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    /* Get the vertices of the element. They are needed to compute the
       Jacobian of the element map.
     */
    concepts::Real2d y0(elm.cell().vertex(0));
    concepts::Real2d y1(elm.cell().vertex(1));
    concepts::Real2d y2(elm.cell().vertex(2));
    concepts::Real2d y3(elm.cell().vertex(3));

    concepts::Real2d p(y1 - y0);
    concepts::Real2d q(y3 - y0);
    concepts::Real2d r(y1 - y0 + y3 - y2);

    concepts::Real2d s, t;
    /* Get the quadrature points.
     */
    const concepts::Quadrature<0> quadrature(3);
    const Real* abscissas = quadrature.abscissas();
    const Real* weights = quadrature.weights();
    const uint qn = quadrature.n();
    /* Get the values of the shape functions in the quadrature points.
     */
    shpfct_.resize(n*qn*qn);
    elm.evaluate((concepts::Real*)shpfct_, abscissas, qn);
    /* Numerical integration: loop over all quadrature points and compute
       the Jacobian. Then, loop over all reference element shape functions
       and add the contributions up.
     */
    for (uint qi = 0; qi < qn; ++qi) {
      for (uint qj = 0; qj < qn; ++qj) {
	s.lincomb(p, r, 1.0, -(1.0+abscissas[qj])/2.0);
	t.lincomb(q, r, 1.0, -(1.0+abscissas[qi])/2.0);
	const Real dx = fabs(s.cross(t)) * weights[qi] * weights[qj] / 4.0
	  * c_;
	for (uint i = 0; i < n; ++i)
	  em(i, 0) += dx * shpfct_[i*qn*qn+qi*qn+qj];
      }
    }
  }
  void operator()(const linearFEM::Triangle& elm,
		  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    concepts::Real2d x0(elm.cell().vertex(0));
    concepts::Real2d x1(elm.cell().vertex(1));
    concepts::Real2d x2(elm.cell().vertex(2));
    const concepts::MapReal3d volMat(1.0, 1.0, 1.0,
				     x0[0], x1[0], x2[0],
				     x0[1], x1[1], x2[1]);
    const Real volTriangle_6 = volMat.determinant() / 6.0 * c_;
    for (uint i = 0; i < n; ++i)
      em(i,0) = volTriangle_6;
  }
private:
  concepts::Array<concepts::Real> shpfct_;
  Real c_;
};

// *********************************************************** Bilinear Form **

class Laplace : public concepts::BilinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    if (em.size()) {
      const linearFEM::Quad* eQX = dynamic_cast<const linearFEM::Quad*>(&elmX);
      if (eQX) {
	const linearFEM::Quad* eQY =
	  dynamic_cast<const linearFEM::Quad*>(&elmY);
	if (eQY) { this->operator()(*eQX, *eQY, em); return; }
      }
      const linearFEM::Triangle* eTX =
	dynamic_cast<const linearFEM::Triangle*>(&elmX);
      if (eTX) {
	const linearFEM::Triangle* eTY =
	  dynamic_cast<const linearFEM::Triangle*>(&elmY);
	if (eTY) {
	  this->operator()(*eTX, *eTY, em); return; }
      }
      std::stringstream error;
      error << "element " << elmX << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }
  }
  void operator()(const linearFEM::Quad& elmX,
		  const linearFEM::Quad& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint n = elmX.T().m();
    if (elmX.support().key() == elmY.support().key()) {
      concepts::Real2d y0(elmX.cell().vertex(0));
      concepts::Real2d y1(elmX.cell().vertex(1));
      concepts::Real2d y2(elmX.cell().vertex(2));
      concepts::Real2d y3(elmX.cell().vertex(3));

      concepts::Real2d p(y1 - y0);
      concepts::Real2d q(y3 - y0);
      concepts::Real2d r(y1 - y0 + y3 - y2);

      concepts::Real2d s, t;

      const concepts::Quadrature<0> quadrature(3);
      const Real* abscissas = quadrature.abscissas();
      const Real* weights = quadrature.weights();
      const uint qn = quadrature.n();

      shpfct_.resize(n*qn*qn);
      elmX.evaluateD((concepts::Real2d*)shpfct_, abscissas, qn);

      for (uint qi = 0; qi < qn; ++qi) {
	for (uint qj = 0; qj < qn; ++qj) {
	  s.lincomb(p, r, 1.0, -(1.0+abscissas[qj])/2.0);
	  t.lincomb(q, r, 1.0, -(1.0+abscissas[qi])/2.0);
	  const Real dx = fabs(4.0 / s.cross(t))
	    * weights[qi] * weights[qj] / 4.0;
	  const Real ss = dx * s.l2_2();
	  const Real st = -dx * (s*t);
	  const Real tt = dx * t.l2_2();
	  for (uint i = 0; i < n; ++i) {
	    concepts::Real2d shpfcti = shpfct_[i*qn*qn+qi*qn+qj];
	    for (uint j = i; j < n; ++j) {
	      concepts::Real2d shpfctj = shpfct_[j*qn*qn+qi*qn+qj];
	      em(i, j) += shpfcti[0] * shpfctj[0] * tt
		+ (shpfcti[0] * shpfctj[1] + shpfcti[1] * shpfctj[0]) * st
		+ shpfcti[1] * shpfctj[1] * ss;
	    }
	  }
	}
      }
      for(uint i = 0; i < n; ++i) 
	for (uint j = i + 1; j < n; ++j) 
	  em(j, i) = em(i, j);
    }
  }
  void operator()(const linearFEM::Triangle& elmX,
		  const linearFEM::Triangle& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint n = elmX.T().m();
    if (elmX.support().key() == elmY.support().key()) {
      concepts::Real2d x0(elmX.cell().vertex(0));
      concepts::Real2d x1(elmX.cell().vertex(1));
      concepts::Real2d x2(elmX.cell().vertex(2));
      /* First compute the volume of the triangle.
       */
      const concepts::MapReal3d volMat(1.0, 1.0, 1.0,
				       x0[0], x1[0], x2[0],
				       x0[1], x1[1], x2[1]);
      const Real volTriangle_4 = 1.0 / (2.0 * volMat.determinant());
      /* The stiffness matrix is easily computable from the coordinates of
	 the corners of the triangle.
      */
      concepts::Real2d p1_p2(x1 - x2);
      concepts::Real2d p2_p0(x2 - x0);
      concepts::Real2d p0_p1(x0 - x1);

      em(0,0) = p1_p2.l2_2() * volTriangle_4;
      em(0,1) = p1_p2*p2_p0 * volTriangle_4;
      em(0,2) = p1_p2*p0_p1 * volTriangle_4;
      em(1,1) = p2_p0.l2_2() * volTriangle_4;
      em(1,2) = p2_p0*p0_p1 * volTriangle_4;
      em(2,2) = p0_p1.l2_2() * volTriangle_4;

      for (uint i = 0; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
    }
  }
private:
  concepts::Array<concepts::Real2d> shpfct_;
};
 
class Identity : public concepts::BilinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    if (em.size()) {
      const linearFEM::Quad* eQX = dynamic_cast<const linearFEM::Quad*>(&elmX);
      if (eQX) {
	const linearFEM::Quad* eQY =
	  dynamic_cast<const linearFEM::Quad*>(&elmY);
	if (eQY) { this->operator()(*eQX, *eQY, em); return; }
      }
      const linearFEM::Triangle* eTX =
	dynamic_cast<const linearFEM::Triangle*>(&elmX);
      if (eTX) {
	const linearFEM::Triangle* eTY =
	  dynamic_cast<const linearFEM::Triangle*>(&elmY);
	if (eTY) {
	  this->operator()(*eTX, *eTY, em); return; }
      }
      std::stringstream error;
      error << "element " << elmX << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }
  }
  void operator()(const linearFEM::Quad& elmX,
		  const linearFEM::Quad& elmY,
		  concepts::ElementMatrix<Real>& em) {
    conceptsAssert(false, concepts::Assertion());
  }
  void operator()(const linearFEM::Triangle& elmX,
		  const linearFEM::Triangle& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint n = elmX.T().m();
    if (elmX.support().key() == elmY.support().key()) {
      concepts::Real2d x0(elmX.cell().vertex(0));
      concepts::Real2d x1(elmX.cell().vertex(1));
      concepts::Real2d x2(elmX.cell().vertex(2));
      /* First compute the volume of the triangle.
       */
      const concepts::MapReal3d volMat(1.0, 1.0, 1.0,
				       x0[0], x1[0], x2[0],
				       x0[1], x1[1], x2[1]);
      const Real volTriangle_12 = volMat.determinant()/24.0;

      em(2,2) = em(1,1) = em(0,0) = volTriangle_12*2.0;
      em(0,1) = em(0,2) = em(1,2) = volTriangle_12;

      for (uint i = 0; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
    }
  }
private:
  concepts::Array<concepts::Real2d> shpfct_;
};

/* <h3>Main program</h3>

   In the main program, all the classes above are used and built together
   in the right way to solve to PDE.
*/
// ************************************************************ Main Program **

int main(int argc, char** argv) {
  /* The whole program is in a big <span
     class="code">try...catch</span> brace to catch exceptions thrown
     by the code.
  */
  try {
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, "(0)"));
    bc->add(concepts::Attribute(2),
	    concepts::Boundary(concepts::Boundary::NEUMANN, "(0)"));
    std::cout << "Boundary conditions: " << *bc << std::endl;

    /* <div class="solution">
       Default values for the variable which could be set by a command line
       argument
    */
    uint level = 1;
    uint gauss_p = 1;
    uint steps = 1;
    uint strategy = 0;
    Real end_time = 1.0;
    /* The <span class="code">inputParser</span> takes care of all input
       parameters and the results which are gathered. At the end, they can
       be printed.
    */
    concepts::InputParser inputParser;
    concepts::InOutParameters& input = inputParser.inputParameters();
    /* Register the parameters.
     */
    input.addInt("level", level);
    input.addInt("gauss_p", gauss_p);
    input.addInt("steps", steps);
    input.addInt("strategy", strategy);
    input.addDouble("end_time", end_time);
    /* Parse the command line options.
     */
    int opt;
    while ((opt = getopt(argc, argv, "-l:p:n:t:s:")) != EOF)
      switch(opt) {
      case 'l': input.addInt("level", std::atoi(optarg)); break;
      case 'p': input.addInt("gauss_p", std::atoi(optarg)); break;
      case 'n': input.addInt("steps", std::atoi(optarg)); break;
      case 's': input.addInt("strategy", std::atoi(optarg)); break;
      case 't': input.addDouble("end_time", std::atof(optarg)); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-l LEVEL] [-p GAUSS_P] [-n STEPS] [-t TIME] [-s STRATEGY]" << std::endl
		  << "where" << std::endl
		  << "  LEVEL: maximal level of refinement" << std::endl
		  << "  GAUSS_P: number of points in Gauss integration" << std::endl
		  << "  STEPS: number of steps for simulation" << std::endl
		  << "  TIME: end time of simulation" << std::endl
		  << "  STRATEGY: time step strategy" << std::endl
		  << "    0 = Newmark" << std::endl
		  << "    1 = Theta" << std::endl
		  << "    2 = RungeKutta2" << std::endl
		  << "    3 = RungeKutta4" << std::endl;
	std::exit(1); break;
      }
    level = input.getInt("level");
    gauss_p = input.getInt("gauss_p");
    steps = input.getInt("steps");
    strategy = input.getInt("strategy");
    end_time = input.getDouble("end_time");

    // Set up the mesh.
    // LshapedTriangle msh;
    concepts::Import2dMesh msh("../../../app-walser/coord.dat",
			       "../../../app-walser/elms.dat",
			       "../../../app-walser/boundary.dat");
    std::cout << "Mesh: " << msh << std::endl;

      std::cout << "level = " << level << std::endl;
      /* And set up the space according to the current level.
	 </div>
      */
      linearFEM::Linear2d spc(msh, level, bc.get());
      //         Linear1d spc(msh, level);
      spc.rebuild();
      std::cout << "Space: " << spc << std::endl;
      graphics::drawMeshEPS(spc, "mesh.eps");
      graphics::MeshDX<Real>(spc, "mesh.dx", 2);

      /* Compute the time dependet external driver
       */
//       GaussFnOscillator driver_tlf(0.0005, 0.5, 0.05, 2*3.1415*5, gauss_p);
      RHS driver_tlf(0.0);
      timestepping::TimeVector driver(spc, driver_tlf);

      /* Compute the initial conditions */
      RHS init_tlf;
      concepts::Vector<Real> initial(spc, init_tlf);
      initial = 0;
      initial[spc.dim()/2] = 1;
      initial[spc.dim()/2-1] = 1;
      initial[spc.dim()/2+1] = 1;
      std::cout << "inital = " << initial << std::endl;

      concepts::Vector<Real> zero(spc);
      std::cout << "zero = " << zero << std::endl;

      concepts::Vector<Real> peak(zero);
      peak(peak.n()/2)=1;

      /* Compute the stiffness matrix and the mass matrix.
      */
      Laplace stiff_bf;
      concepts::SparseMatrix<Real> stiff(spc, stiff_bf);
      //concepts::DenseMatrix<Real> stiff(spc, stiff_bf);
      //std::cout << "stiff = " << stiff << std::endl; 

      Identity mass_bf;
      concepts::SparseMatrix<Real> mass(spc, mass_bf);
      //concepts::DenseMatrix<Real> mass(spc, mass_bf);
      std::cout << "mass = " << mass << std::endl; 

//       concepts::LiCo<Real> L(mass, stiff);
//       concepts::CG<Real> pdesol(L, 1e-10, 200);
//       pdesol(driver, zero);
//       std::cout << "zero = " << zero << std::endl;
//       graphics::DataGnuplot(spc, "driver.gnuplot", driver);
//       graphics::DataGnuplot(spc, "zero.gnuplot", zero);


//       for (uint n = 0; n <= 5; ++n) {
// 	v = driver(n*end_time/5.0);
// 	concepts::Vector<Real> u(spc);
// 	pdesol(v, u);
// 	std::stringstream drivername;
// 	drivername << "gnuplot/l2proj-driver1d-" << n << ".gnuplot" << std::ends;
// 	graphics::DataGnuplot(spc, drivername.str(), u);
//       }

//       concepts::SuperLUFabric fabric;
      concepts::CGFabric<Real> fabric;

      std::auto_ptr<timestepping::TimeStepStrategy> timeStepStrategy;
      switch (strategy) {
      /* Generate a Newmark1 time stepping strategy  */
      case 0:  timeStepStrategy.reset(new timestepping::Newmark(fabric, mass, stiff, driver, initial, zero, end_time/steps));
	break;
      /* Generate a Theta-Scheme time stepping strategy  */
      case 1:  timeStepStrategy.reset(new timestepping::Theta(fabric, mass, stiff, driver, initial, end_time/steps));
	break;
      /* Generate a RungeKutta2-Scheme time stepping strategy */
      case 2: timeStepStrategy.reset(new timestepping::RungeKutta2(fabric, mass, stiff, driver, initial, end_time/steps));
	break;
     /* Generate a RungeKutta4-Scheme time stepping strategy */
      case 3: timeStepStrategy.reset(new timestepping::RungeKutta4(fabric, mass, stiff, driver, initial, end_time/steps));
	break;
      }

      /* Solve the PDE with timestepping 
      */
      std::cout << "TimeStepStrategy = " << *timeStepStrategy << std::endl;
      timestepping::TimeStepping solver(*timeStepStrategy);
      std::cout << "TimeStepping = " << solver << std::endl;

      std::cout << "Number of steps = " << steps << ", end time = "
		<< end_time 
		<< " dt = " << end_time/steps << std::endl; // in solver output
      concepts::Vector<Real> sol(spc);
      graphics::DataDX<Real> dx(graphics::drawDataDX(spc, "data2D.dx"));
      for(uint n = 0; n <= steps; ++n) {
	solver(sol, n);
	// write data to file in Gnuplot format, look at them with ...
	std::stringstream filename;
	//	filename << "gnuplot/sol1d-l" << level << "-n"
	//	 << n << "-N" << steps << ".gnuplot" << std::ends;
// 	filename << "gnuplot/sol1d-" << n  << ".gnuplot" << std::ends;
// 	graphics::DataGnuplot(spc, filename.str(), sol);

	filename << "t = " << std::setprecision(3)
		 << n*end_time/steps << std::ends;
	dx(sol, filename.str());
      }
//       std::cout << "Trigger exception:" << std::endl;
//       solver(sol, 1);
//       std::cout << "Shoul not be reached." << std::endl;

  }

  /* Here, the exceptions are caught and some output is generated. Normally,
     this should not happen.
  */
//   catch(concepts::NoConvergence& e) {
//     std::cout << "solver did not converge!" << std::endl << e << std::endl;
//   }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
//   catch(...) {
//     // catch all
//   }

  /* Sane return value.
   */
  return 0;
}

      /*
gnuplot> plot 'solution1d-3.gnuplot'
gnuplot> set data style linespoints
gnuplot> replot
gnuplot> replot
gnuplot> plot 'solution1d-3.gnuplot', 'solution1d-3.gnuplot'
gnuplot> plot 'solution1d-3.gnuplot' title '1', 'solution1d-3.gnuplot'
gnuplot> load 'gnuplot.script'
       */
