/* System include files. The first is for input and output operations and
   the second one declares <span class="code">auto_ptr</span>.

   segfaults with
   timesteps -l 3 -s 0 -n 128 -t 1 -m 4
*/
#include <sstream>
#include <iostream>
#include <memory>
#include <unistd.h> // for command line parsing

/* Concepts include files.
 */
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
#include "linearFEM.hh"
#include "graphics.hh"

/* Timestepping include files.
 */
#include "timestepping.hh"

const double Pi = 3.1415926535875;

// ********************************************************* 1D Test Example **

/* With this using directive, we do not need to prepend
   <span class="code">concepts::</span> before <span class="code">Real</span>
   everytime.
*/
using concepts::Real;

/* <h3>Mesh</h3>

The mesh is the interval from 1 to Pi with two cells.  One
is <span class="math">(0,Pi/2)</span> and the other is <span
class="math">(Pi/2,Pi)</span>. They are coupled in the class Line
which makes it possible to loop over the cells.
*/
// ******************************************************************** Mesh **


class Line : public concepts::Mesh1 {
public:
  Line();
  virtual ~Line();
  unsigned int ncell() const { return 2; }
  concepts::Scan1* scan() { return new S(cell_); }
  virtual std::ostream& info(std::ostream& os) const;
private:
  concepts::Vertex *vtx_[3];
  concepts::Edge *edg_[2];
  concepts::Cell1 *cell_[2];

  /* A scanner is used to loop over all cells in a mesh.
     <span class="code">S</span> is the scanner for this mesh. It is used
     in the constructor of the space to create the elements on the cells
     of the mesh. With the ++ operator it is possible to loop over the cells
     of the mesh. The constructor of the space shows an example of the usage
     of a scanner and its ++ operator.
  */
  class S : public concepts::Scan<concepts::Cell1> {
    unsigned int idx_;
    concepts::Cell1 *(&cell_)[2];
  public:
    S(concepts::Cell1 *(&cell)[2]) : idx_(0), cell_(cell) {}
    S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    bool eos() const { return idx_ == 2; }
    concepts::Cell1& operator++(int) { return *cell_[idx_++]; }
    concepts::Scan1* clone() const { return new S(*this); }
  };
};

/* The constructor of the mesh Line sets up everything which is needed:
   <ol>
*/
Line::Line() {
  /* <li> the vertices
   */
  vtx_[0] = new concepts::Vertex(1);
  vtx_[1] = new concepts::Vertex(0);
  vtx_[2] = new concepts::Vertex(2);
  /* <li> the edges (they share the middle vertex)
   */
  edg_[0] = new concepts::Edge(*vtx_[0], *vtx_[1]);
  edg_[1] = new concepts::Edge(*vtx_[1], *vtx_[2]);
  /* <li> and the cells which live on the edges. Here, the element map is
     given. 
     <div class="solution">
     The domain of interest is (-1,1)
  */
  cell_[0] = new concepts::Edge1d(*edg_[0], concepts::MapEdge1d(0.0, Pi/2));
  cell_[1] = new concepts::Edge1d(*edg_[1], concepts::MapEdge1d(Pi/2, Pi));
}
/* Therefore, the constructor arguments of the element map had to be
   fixed.</div>
   </ol>

   The destructor of the mesh removes all components which where created
   in the constructor.
*/
Line::~Line() {
  for (int i = 0; i < 2; ++i)
    delete cell_[i];
  for (int i = 0; i < 2; ++i)
    delete edg_[i];
  for (int i = 0; i < 3; ++i)
    delete vtx_[i];
}

/* The info member prints the two cells of the mesh. It is used by the
   output operator of the base class <span
   class="code">OutputOperator</span>.
*/
std::ostream& Line::info(std::ostream& os) const {
  os << "Line(" << *cell_[0] << ", " << *cell_[1] << ')';
  return os;
}

/* <h3>Linear Form</h3>

The linear form is used to compute the element contributions to the
load vector. The computations are done in the application operator
of this class.
*/
// ************************************************************* Linear Form **

class Gauss_lf : public concepts::LinearForm<Real> {
  /* <div class="solution">
     Linear form that can be used as initial conditions for a PDE. The
     function is a Gauss Function.
     (<span class="code">gauss_p</span>).
  */
private:
  const Real ampl_, mu_, sigma_;
  const concepts::Quadrature<4> quad_;
public:
  Gauss_lf(const Real ampl = 1, const Real mu = 0, const Real sigma = 1, 
	  const uint gauss_p = 1) : 
    ampl_(ampl), mu_(mu), sigma_(sigma), quad_(gauss_p) {}
  /* </div>
   */
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();

    const linearFEM::Line* e = dynamic_cast<const linearFEM::Line*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    const Real h = e->cell().size();
    const Real* w = quad_.weights();   // quadrature weights
    const Real* q = quad_.abscissas(); // quadrature points

    for (uint i = 0; i < n; ++i) { // loop over shape functions
      Real& m = em(i,0);
      for (uint k = 0; k < quad_.n(); ++k) { // loop over quadrature points
	m += w[k] * h / 2.0
	  * e->shapefct(i, q[k])
	  * exp(-pow(e->cell().chi( (1.0+q[k])/2.0 ) - mu_, 2.0) / (2 * sigma_ * sigma_))
	  * 0.398942 / sigma_ * ampl_;
      } // for k
    } // for i
    /* </div>
     */
  }
};

class Sin_lf : public concepts::LinearForm<Real> {
  /* <div class="solution">
     Linear form that can be used as initial conditions for a PDE. The
     function is A * sin( k * (x - mu) ).
     (<span class="code">gauss_p</span>).
  */
private:
  const Real ampl_, k_, mu_;
  const concepts::Quadrature<4> quad_;
public:
  Sin_lf(const Real ampl = 1, const Real k = 1, const Real mu = 0,
	 const uint gauss_p = 1) : 
    ampl_(ampl), k_(k), mu_(mu), quad_(gauss_p) {}
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();

    const linearFEM::Line* e = dynamic_cast<const linearFEM::Line*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    const Real h = e->cell().size();
    const Real* w = quad_.weights();   // quadrature weights
    const Real* q = quad_.abscissas(); // quadrature points

    for (uint i = 0; i < n; ++i) { // loop over shape functions
      Real& m = em(i,0);
      for (uint k = 0; k < quad_.n(); ++k) { // loop over quadrature points
	m += w[k] * h / 2.0
	  * e->shapefct(i, q[k])
	  * sin((e->cell().chi( (1.0+q[k])/2.0 ) - mu_)*k_) * ampl_;
      } // for k
    } // for i
  }
};

class Gauss_tlf : public timestepping::TimeLinearForm {
  /* <div class="solution">
     Timedependent linear form that can be used as external driver for
     the rhs of a  PDE. It is a Gaussfunction with an amplitude that
     is oscillating with cos(omega t)
     (<span class="code">gauss_p</span>).
  */
private:
  const Real ampl_, mu_, sigma_, omega_;
  const concepts::Quadrature<4> quad_;
public:
  Gauss_tlf(const Real ampl = 1, const Real mu = 0, const Real sigma = 1, 
		    const Real omega = 1, const uint gauss_p = 1) : 
    ampl_(ampl), mu_(mu), sigma_(sigma), omega_(omega), quad_(gauss_p) {}
  /* </div>
   */
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();

    const linearFEM::Line* e = dynamic_cast<const linearFEM::Line*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    const Real h = e->cell().size();
    const Real* w = quad_.weights();   // quadrature weights
    const Real* q = quad_.abscissas(); // quadrature points

    for (uint i = 0; i < n; ++i) { // loop over shape functions
      Real& m = em(i,0);
      for (uint k = 0; time_ < 2*2*Pi/omega_ && k < quad_.n(); ++k) { 
	// loop over quadrature points
	m += w[k] * h / 2.0
	  * e->shapefct(i, q[k])
	  * exp(-pow(e->cell().chi( (1.0+q[k])/2.0 ) - mu_, 2.0) / 
		(2 * sigma_ * sigma_))
	  * 0.398942 / sigma_ * cos(omega_*time_) * ampl_;
      } // for k
    } // for i
    /* </div>
     */
  }
};

class Sin_tlf : public timestepping::TimeLinearForm {
  /* Timedependent linear form that can be used as driver for a
     PDE. The function is A * sin( k * (x - mu)) * cos( w * t).
   */
private:
  const Real ampl_, k_, mu_, omega_;
  const concepts::Quadrature<4> quad_;
public:
  Sin_tlf(const Real ampl = 1, const Real k = 1, const Real mu = 0,
	  const Real omega = 1, const uint gauss_p = 1) : 
    ampl_(ampl), k_(k), mu_(mu), omega_(omega), quad_(gauss_p) {}
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();

    const linearFEM::Line* e = dynamic_cast<const linearFEM::Line*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    const Real h = e->cell().size();
    const Real* w = quad_.weights();   // quadrature weights
    const Real* q = quad_.abscissas(); // quadrature points

    for (uint i = 0; i < n; ++i) { // loop over shape functions
      Real& m = em(i,0);
      for (uint k = 0; k < quad_.n(); ++k) { // loop over quadrature points
	m += w[k] * h / 2.0
	  * e->shapefct(i, q[k])
	  * sin((e->cell().chi( (1.0+q[k])/2.0 ) - mu_)*k_) 
	  * cos(omega_*time_) * ampl_;
      } // for k
    } // for i
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

/* The application operator of the class <span class="code">Laplace</span>
   computes the element matrix belonging to the Laplacian in the PDE.
*/
class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    /* Since the intersection of the supports of the two different
       elements has at most measure zero, we do only integrate if the
       elements are equal.

       Here, we don't integrate, since the element stiffness matrices can
       be computed beforehand.
    */
    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size();

      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());

      const Real s = 2.0/h;
      em(0,0) = em(1,1) = 1.0/2.0*s;
      em(0,1) = em(1,0) = -1.0/2.0*s;
    }
  }
};

/* The application operator of the class <span class="code">Identity</span>
   computes the element matrix belonging to the absolute term in the PDE.
*/
class Identity : public concepts::BilinearForm<Real> {
  /* <div class="solution">
     As for the linear form, we need numerical integration.
  */
private:
  const concepts::Quadrature<4> quad_;
public:
  Identity(const uint gauss_p = 1) : quad_(gauss_p) {}
  /* </div>
   */
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size();

      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());
      const Real s = h/3.0/2.0;
      em(0,0) = em(1,1) = 2.0*s;
      em(0,1) = em(1,0) = s;
      /* </div>
       */
    }
  }
};

/* The exact solutions and the exact energies for the wave and the
   heat equation of example 1).
*/

/* Wave equation
   initial = sin(2x), d_dt_initial = sin(3x), driver = sin(x)
*/
Real sol_wave_eq(const Real x, const Real t) {
  return sin(x)*(1.0-cos(t))+sin(2.0*x)*cos(2.0*t)+
      1/3.0*sin(3.0*x)*sin(3.0*t);
}
Real energy_wave_eq(const Real t) {
  return 3*Pi-12*Pi*pow(cos(t),2)-cos(t)*Pi+20*Pi*pow(cos(t),4)-8*Pi*pow(cos(t),6);
}

/* Heat equation
   initial = sin(2x), driver = sin(x)
*/
Real sol_heat_eq(const Real x, const Real t) {
  return sin(x)*(1-exp(-t))+sin(2.0*x)*exp(-4.0*t);
}
Real energy_heat_eq(const Real t) {
  return Pi/2.0*exp(-8.0*t)*(4.0+exp(6.0*t)-2.0*exp(7.0*t)+exp(8.0*t));
}

/* <h3>Main program</h3>

In the main program, all the classes above are used and built together
in the right way to solve different PDE. Three examples can be
executed by commenting and uncommenting the corresponding initial
conditions of the partial differential equation.

1) Test example

This is the standard configuration of the file. The advantage is that
the energy error of all strategy schemes is calculated and so the
implementation can be tested in detail. The disadvantage is that the
solutions are not intuitively understandable.

2) Wavepacket on rope

This example simulates the propagation of a wave packet on an initially
horizontal rope, that is free an the left side and fixed on the right
side. The wave packet is produced by an external up and down movement
of the right hand side. Please notice the different reflections of the
wave packet on the different boundary conditions. In this example the
energy error can not be calculated.

3) Diffusing ink drop

This example simulates the diffusion of an ink drop in water. The
initial condition is a sharp peak that is spreading out proportional
to the square root of time. In this example the energy error can not
be calculated.

Use the command line arguments to change the diffent parameters of a
simulation. Every simulation outputs 5 graphics of the solution vector
and 5 graphics of the external driver at equally spaced timesteps. You
can look at them with the GNUPLOT scripts in the gnuplot directory.

gnuplot> load 'sol.gnuplot.script'
gnuplot> load 'driver.gnuplot.script'

If you use example 1) and calculate several levels of refinement, you
can also look at the evolution of the energy errors.  

gnuplot> load 'err-s0.gnuplot.script' 
gnuplot> load 'err-s1.gnuplot.script' 
gnuplot> load 'err-s23.gnuplot.script'
*/

// ************************************************************ Main Program **

/* </div>
 */
int main(int argc, char** argv) {
  /* The whole program is in a big <span
     class="code">try...catch</span> brace to catch exceptions thrown
     by the code.
  */
  try {   
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, 
			       concepts::ParsedFormula("(0)")));
    bc->add(concepts::Attribute(2),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, 
			       concepts::ParsedFormula("(0)")));
    // Example 2): replace DIRICHLET with NEUMANN
    std::cout << "Boundary conditions: " << *bc << std::endl;

    /* <div class="solution">
       Default values for the variable which could be set by a command line
       argument
    */
    uint level = 1;
    uint gauss_p = 1;
    uint steps = 1;
    uint steps_level = 0;
    uint strategy = 0;
    Real end_time = 1.0;
    /* The <span class="code">inputParser</span> takes care of all input
       parameters and the results which are gathered. At the end, they can
       be printed.
    */
    concepts::InputParser inputParser;
    concepts::InOutParameters& input = inputParser.inputParameters();
    /* Register the parameters.
     */
    input.addInt("gauss_p", gauss_p);
    input.addInt("level", level);
    input.addInt("steps", steps);
    input.addInt("steps_level", steps_level);
    input.addInt("strategy", strategy);
    input.addDouble("end_time", end_time);
    /* Parse the command line options.
     */
    int opt;
    while ((opt = getopt(argc, argv, "-l:p:n:m:t:s:")) != EOF)
      switch(opt) {
      case 'p': input.addInt("gauss_p", std::atoi(optarg)); break;
      case 'l': input.addInt("level", std::atoi(optarg)); break;
      case 'n': input.addInt("steps", std::atoi(optarg)); break;
      case 'm': input.addInt("steps_level", std::atoi(optarg)); break;
      case 's': input.addInt("strategy", std::atoi(optarg)); break;
      case 't': input.addDouble("end_time", std::atof(optarg)); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-l LEVEL] [-p GAUSS_P] [-n STEPS] [-m STEPS_L]"
		  << " [-t TIME] [-s STRATEGY]" << std::endl
		  << "where" << std::endl
		  << "  GAUSS_P: number of points in Gauss integration" 
		  << std::endl
		  << "  LEVEL: number of space refinements" << std::endl
		  << "  STEPS: minimal number of time steps" << std::endl
		  << "  STEPS_L: number of time refinements" << std::endl
		  << "  TIME: end time of simulation" << std::endl
		  << "  STRATEGY: time step strategy" << std::endl
		  << "    0 = Newmark" << std::endl
		  << "    1 = Theta" << std::endl
		  << "    2 = RungeKutta2" << std::endl
		  << "    3 = RungeKutta4" << std::endl;
	std::exit(1); break;
      }
    gauss_p = input.getInt("gauss_p");
    level = input.getInt("level");
    steps = input.getInt("steps");
    steps_level = input.getInt("steps_level");
    strategy = input.getInt("strategy");
    end_time = input.getDouble("end_time");

    /* </div> Prepare the results part to gather the results and compute the
       exact FE energy and the element size.
    */
    concepts::InOutParameters& output = inputParser.outputParameters();
    Real exEnergy, feEnergy;
    output.addArrayDouble("h");
    output.addArrayDouble("time");
    output.addArrayDouble("errEnergy");
    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::DOUBLE, "h", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "time", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "errEnergy", output);

    for (uint m=0; m<=steps_level; ++m) {
      /* </div>
	 Set up the mesh.
      */
      Line msh;
      std::cout << "Mesh: " << msh << std::endl;

      /* And set up the space according to the current level.
	 </div>
      */
      std::cout << "level = " << level << std::endl;
      linearFEM::Linear1d spc(msh, level, bc.get());
      std::cout << "Space: " << spc << std::endl;

      /* Compute the time dependet external driver
	 Example 1)
      */
      Sin_tlf driver_tlf(1.0, 1.0 , 0.0, 0.0, gauss_p);
      /* Example 2)
       */
      // Gauss_tlf driver_tlf(1, Pi, 0.1, 2*3.1415*2, gauss_p);
      /* Example 3)
       */
      //Gauss_tlf driver_tlf(0, Pi, 0.1, 2*3.1415*2, gauss_p);

      timestepping::TimeVector driver(spc, driver_tlf);

      /* Compute the initial conditions
	 Example 1)
      */
      Sin_lf initial_lf(1.0, 2.0, 0.0, gauss_p);
      Sin_lf d_dt_initial_lf(1.0, 3.0, 0, gauss_p);
      /* Example 2)
       */
      //Gauss_lf initial_lf(0, Pi/2 , 0.1, gauss_p);
      //Gauss_lf d_dt_initial_lf(0, Pi/2 , 0.1, gauss_p);
      /* Example 3)
       */
      //Gauss_lf initial_lf(1, Pi/2 , 0.1, gauss_p);

      concepts::Vector<Real> temp(spc, initial_lf);
      concepts::Vector<Real> initial(spc);
      Identity identity_bf(gauss_p);
      concepts::SparseMatrix<Real> identity(spc, identity_bf);
      concepts::SuperLU inv(identity);
      inv(temp, initial);
      std::cout << "inital = " << initial << std::endl;

      // Initial velocity field for the wave equation
      concepts::Vector<Real> d_dt_initial(spc);
      temp = concepts::Vector<Real>(spc, d_dt_initial_lf);
      inv(temp, d_dt_initial);
      std::cout << "d/dt initial = " << d_dt_initial << std::endl;

      /* Write external driver data to file in Gnuplot format
       */
      for (uint n = 0; n <= 5; ++n) {
	driver.time(n*end_time/5.0);
	inv(driver, temp);
	std::stringstream drivername;
	drivername << "gnuplot/driver1d-" << n << ".gnuplot" << std::ends;
	graphics::drawDataGnuplot(spc, drivername.str(), temp);
      }

      /* Compute the stiffness matrix and the mass matrix.
       */
      Laplace stiff_bf;
      concepts::SparseMatrix<Real> stiff(spc, stiff_bf);

      Identity mass_bf(gauss_p);
      concepts::SparseMatrix<Real> mass(spc, mass_bf);

      /* Generate the choosen time stepping strategy and the exact results 
       */
      concepts::SuperLUFabric fabric;
      std::auto_ptr<timestepping::TimeStepStrategy> timeStepStrategy;
      double (*ex_sol) (double, double);
      double (*ex_energy) (double);

      switch (strategy) {
      case 0:  
	timeStepStrategy.reset(new timestepping::Newmark(fabric, mass, stiff, driver, initial, d_dt_initial, end_time/steps));
	ex_sol = &sol_wave_eq;
	ex_energy = &energy_wave_eq;
	break;
      case 1:  
	timeStepStrategy.reset(new timestepping::Theta(fabric, mass, stiff, driver, initial, end_time/steps,1));
	ex_sol = &sol_heat_eq;
	ex_energy = &energy_heat_eq;
	break;
      case 2: 
	timeStepStrategy.reset(new timestepping::RungeKutta2(fabric, mass, stiff, driver, initial, end_time/steps));
	ex_sol = &sol_heat_eq;
	ex_energy = &energy_heat_eq;
	break;
      case 3: 
	timeStepStrategy.reset(new timestepping::RungeKutta4(fabric, mass, stiff, driver, initial, end_time/steps));
	ex_sol = &sol_heat_eq;
	ex_energy = &energy_heat_eq;
	break;
      }

      /* Solve the PDE with timestepping 
       */
      timestepping::TimeStepping solver(*timeStepStrategy);
      std::cout << "TimeStepping = " << solver << std::endl;

      std::cout << "Number of steps = " << steps << ", end time = "
		<< end_time << std::endl;
      concepts::Vector<Real> sol(spc);
      double time;
      const Real h = 1.0/(1<<level);
      for(uint n = 0; n <= 5; ++n) {
	solver(sol, n*steps/5);
	time = solver.time();
 
	// Write solution data to file in Gnuplot format
	std::stringstream filename;
	filename << "gnuplot/sol1d-" << n << ".gnuplot" << std::ends;
	graphics::drawDataGnuplot(spc, filename.str(), sol);
      }
      /* Calcute FE energy error.
       */
      stiff(sol,temp);
      feEnergy = sol * temp;
      exEnergy = ex_energy(time);
      output.addArrayDouble("h", steps, h);
      output.addArrayDouble("time", steps, time);
      output.addArrayDouble("errEnergy", steps, 
			    fabs(exEnergy-feEnergy)/exEnergy);
      steps *= 2;
      std::cout << std::endl;
    } // end steps loop 

    /* </div> Write the energy errors to file in Gnuplot format
     */
    table.print<concepts::ResultsTable::GNUPLOT>(std::cout);
    std::stringstream filename;
    filename << "gnuplot/err1d-l" << level << "-s" << strategy 
	     <<".gnuplot" << std::ends;
    std::ofstream* ofs = new std::ofstream(filename.str().c_str());
    *ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(*ofs);
    delete ofs;
  } //end try
  /* Here, the exceptions are caught and some output is generated. Normally,
     this should not happen.
  */
  catch(concepts::NoConvergence& e) {
    std::cout << "solver did not converge!" << std::endl << e << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
  // catch(...) {
  //   // catch all
  // }
  
  /* Sane return value.
   */
  return 0;
}
