/* System include files. The first is for input and output operations and
   the second one declares <span class="code">auto_ptr</span>.
 */
#include <iostream>
#include <memory>
#include <unistd.h> // for command line parsing
/* Concepts include files.
 */
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
/* With this using directive, we do not need to prepend
   <span class="code">concepts::</span> before <span class="code">Real</span>
   everytime.
*/
using concepts::Real;

/* <h3>Mesh</h3>

   The mesh is the interval from 1 to 5 with two cells.  One
   is <span class="math">(1,2)</span> and the other is <span
   class="math">(2,5)</span>. They are coupled in the class Line
   which makes it possible to loop over the cells.
*/
// ******************************************************************** Mesh **

class Line : public concepts::Mesh1 {
public:
  Line();
  virtual ~Line();
  unsigned int ncell() const { return 2; }
  concepts::Scan1* scan() { return new S(cell_); }
  virtual std::ostream& info(std::ostream& os) const;
private:
  concepts::Vertex *vtx_[3];
  concepts::Edge *edg_[2];
  concepts::Cell1 *cell_[2];
  /* A scanner is used to loop over all cells in a mesh.
     <span class="code">S</span> is the scanner for this mesh. It is used
     in the constructor of the space to create the elements on the cells
     of the mesh. With the ++ operator it is possible to loop over the cells
     of the mesh. The constructor of the space shows an example of the usage
     of a scanner and its ++ operator.
  */
  class S : public concepts::Scan<concepts::Cell1> {
    unsigned int idx_;
    concepts::Cell1 *(&cell_)[2];
  public:
    S(concepts::Cell1 *(&cell)[2]) : idx_(0), cell_(cell) {}
    S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    bool eos() const { return idx_ == 2; }
    concepts::Cell1& operator++(int) { return *cell_[idx_++]; }
    concepts::Scan1* clone() const { return new S(*this); }
  };
};

/* The constructor of the mesh Line sets up everything which is needed:
   <ol>
*/
Line::Line() {
  /* <li> the vertices
   */
  vtx_[0] = new concepts::Vertex(2);
  vtx_[1] = new concepts::Vertex(0);
  vtx_[2] = new concepts::Vertex(2);
  /* <li> the edges (they share the middle vertex)
   */
  edg_[0] = new concepts::Edge(*vtx_[0], *vtx_[1]);
  edg_[1] = new concepts::Edge(*vtx_[1], *vtx_[2]);
  /* <li> and the cells which live on the edges. Here, the element map is
     given. 
     <div class="solution">
     The domain of interest is (-1,1)
  */
  cell_[0] = new concepts::Edge1d(*edg_[0], concepts::MapEdge1d(-1.0, 0.0));
  cell_[1] = new concepts::Edge1d(*edg_[1], concepts::MapEdge1d(0.0, 1.0));
}
/* Therefore, the constructor arguments of the element map had to be
   fixed.</div>
   </ol>

   The destructor of the mesh removes all components which where created
   in the constructor.
*/
Line::~Line() {
  for (int i = 0; i < 2; ++i)
    delete cell_[i];
  for (int i = 0; i < 2; ++i)
    delete edg_[i];
  for (int i = 0; i < 3; ++i)
    delete vtx_[i];
}

/* The info member prints the two cells of the mesh. It is used by the
   output operator of the base class <span
   class="code">OutputOperator</span>.
*/
std::ostream& Line::info(std::ostream& os) const {
  os << "Line(" << *cell_[0] << ", " << *cell_[1] << ')';
  return os;
}

/* <h3>Element</h3>

   There is not much happening in the element yet. At a later time, one
   wants to be able to evaluate the shape functions at any point in the
   reference element using this class. Right now, only the T matrices
   are generated from the information coming from the space (ie. the
   indices).
*/
// ***************************************************************** Element **

class HatFunctions1d : public concepts::Element<Real> {
public:
  HatFunctions1d(const concepts::Edge1d& cell, uint idx[]);
  const concepts::TMatrixBase<Real>& T() const { return *T_; }
  const concepts::Edge& support() const { return cell_.connector(); }
  concepts::Real3d vertex(const uint i) const { return cell_.vertex(i); }
  const concepts::Edge1d& cell() const { return cell_; }
  Real shapefct(const uint i, const Real xi) const;
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  const concepts::Edge1d& cell_;
  /* <div class="solution">Here, the T matrices are stored.</div>
   */
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
};

/* <div class="solution">Here, the T matrices are created from the data in
   <span class="code">idx</span>.</div>
*/
HatFunctions1d::HatFunctions1d(const concepts::Edge1d& cell, uint idx[]) :
  cell_(cell), T_(new concepts::TIndex<Real>(2,2,idx)) {}

/* <div class="solution">
   The class <span class="code">HatFunctions1d</span> got a new
   member function <span class="code">shapefct</span> which computes the
   value of a shape function in a given point in the reference element (-1,1).
*/
Real HatFunctions1d::shapefct(const uint i, const Real xi) const {
  switch(i) {
  case 0: return (1.0+xi)/2.0;
  case 1: return (1.0-xi)/2.0;
  default: conceptsAssert(1 == 0, concepts::Assertion()); return 0.0;
  }
}
/* </div>
*/

std::ostream& HatFunctions1d::info(std::ostream& os) const {
  return os << "HatFunctions1d(" << cell_ << ", T = " << *T_ << ')';
}

/* <h3>Scanner for the Elements</h3>

   This declaration indicates the relationship between a scanner over generic
   elements and a scanner over elements of type
   <span class="code">HatFunctions1d</span>. The declaration has to be in
   the namespace <span class="code">concepts</span>.
*/
// ***************************************************************** Scanner **

namespace concepts {
  template<>
  class Scan<HatFunctions1d> : public Scan<Element<Real> > {
  public:
    HatFunctions1d& operator++(int) = 0;
  };
} // namespace concepts

/* <h3>Space</h3>

   A space ties the elements together and has methods to loop over the
   elements with a scanner (as in the mesh). This is used in the assembling
   procedure.

   In the constructor of the space, the degrees of freedom are counted
   and the continuity is enforced. This is done using so called T matrices
   which are used for the assembling. The T matrices are stored in the
   elements.
*/
// ******************************************************************* Space **

class Linear1d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<HatFunctions1d> Scan;
  Linear1d(concepts::Mesh1& line, uint level);
  virtual ~Linear1d();
  uint dim() const { return dim_; }
  uint nelm() const { return nelm_; }
  Scan* scan() const {
    return new concepts::PListScan<HatFunctions1d>(*elm_);
  }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  concepts::Joiner<HatFunctions1d*, 1>* elm_;
  void constructor_(concepts::Edge1d* cell, uint level,
		    concepts::DynArray<uint>& map);
};

/* The constructor calls <span class="code">constructor_</span> on
   every suitable element.
*/
Linear1d::Linear1d(concepts::Mesh1& line, uint level) :
  dim_(0), nelm_(0), elm_(0) {

  concepts::DynArray<uint> map(2, 3, (uint)-1);
  std::auto_ptr<concepts::Scan1> sc(line.scan());
  while (*sc) {
    concepts::Cell1& cell = (*sc)++;
    concepts::Edge1d* cellE = dynamic_cast<concepts::Edge1d*>(&cell);
    if (cellE)
      constructor_(cellE, level, map);
    else
      throw conceptsException(concepts::MissingFeature("cell not supported"));
  }
}

Linear1d::~Linear1d() {
  delete elm_;
}

std::ostream& Linear1d::info(std::ostream& os) const {
  return os << "Linear1d(dim = " << dim_ << ", nelm = " << nelm_ << ')';
}

/* <span class="code">constructor_</span> is called recursively
   on the children of the current cell if the
   correct level of refinement is not reached yet. Otherwise, the degrees
   of freedom are assigned and a new element is added to the list of
   elements.
*/
void Linear1d::constructor_(concepts::Edge1d* cell, uint level,
			    concepts::DynArray<uint>& map) {
  if (level) {
    --level;
    constructor_(cell->child(0), level, map);
    constructor_(cell->child(1), level, map);
  } else {
    /* <div class="solution">Here, the data for the T matrix is computed and
       stored in <span class="code">idx</span> which is then given as an
       argument to the constructor of the Element
       <span class="code">HatFunctions1d</span>.</div>
    */
    uint idx[2];
    for (uint j = 0; j < 2; ++j) {
      uint& i = map[cell->connector().vertex(j)->key()];
      if (i == (uint) -1)
	i = dim_++;
      idx[j] = i;
    }
    HatFunctions1d* elm = new HatFunctions1d(*cell, idx);

    nelm_++;
    elm_ = new concepts::Joiner<HatFunctions1d*, 1>(elm, elm_);
  }
}

/* <h3>Linear Form</h3>

   The linear form is used to compute the element contributions to the
   load vector. The computations are done in the application operator
   of this class.
 */
// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
  /* <div class="solution">
     As we need numerical integration, a new member is added to the class which
     evaluates the linear form: quad_. It contains the quadrature points and
     weights for given number of quadrature points
     (<span class="code">gauss_p</span>).
  */
private:
  const concepts::Quadrature<4> quad_;
public:
  RHS(const uint gauss_p = 1) : quad_(gauss_p) {}
  /* </div>
  */
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    /* Resize the element load vector to the correct size and set all entries
       to zero. This way, the computed values can be added into
       <span class="code">em</span>.
    */
    em.resize(n, 1);
    em.zeros();

    const HatFunctions1d* e = dynamic_cast<const HatFunctions1d*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    const Real h = e->cell().size();
    /* <div class="solution">
       <span class="math">f = x<sup>2</sup></span>,
       hat functions living on <span class="math">(-1,1)</span>
       and mapped onto the element. The element map is called by
       <span class="code">chi</span> and returns the point in the physical
       space. The reference element for the element map is
       <span class="math">(0,1)</span> (therefore the transformation).

       <span class="math">h / 2.0</span> is the Jacobian resulting from the
       variable transformation in the integration.
    */
    const Real* w = quad_.weights();   // quadrature weights
    const Real* q = quad_.abscissas(); // quadrature points

    for (uint i = 0; i < n; ++i) { // loop over shape functions
      Real& m = em(i,0);
      for (uint k = 0; k < quad_.n(); ++k) { // loop over quadrature points
	m += w[k] * h / 2.0
	  * e->shapefct(i, q[k])
	  * pow(e->cell().chi( (1.0+q[k])/2.0 ), 2.0);
      } // for k
    } // for i
    /* </div>
    */
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

/* The application operator of the class <span class="code">Laplace</span>
   computes the element matrix belonging to the Laplacian in the PDE.
*/
class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const HatFunctions1d* eX = dynamic_cast<const HatFunctions1d*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const HatFunctions1d* eY = dynamic_cast<const HatFunctions1d*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    /* Since the intersection of the supports of the two different
       elements has at most measure zero, we do only integrate if the
       elements are equal.

       Here, we don't integrate, since the element stiffness matrices can
       be computed beforehand.
    */
    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size();

      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());

      const Real s = 2.0/h;
      em(0,0) = em(1,1) = 1.0/2.0*s;
      em(0,1) = em(1,0) = -1.0/2.0*s;
    }
  }
};

/* The application operator of the class <span class="code">Identity</span>
   computes the element matrix belonging to the absolute term in the PDE.
*/
class Identity : public concepts::BilinearForm<Real> {
  /* <div class="solution">
     As for the linear form, we need numerical integration.
  */
private:
  const concepts::Quadrature<4> quad_;
public:
  Identity(const uint gauss_p = 1) : quad_(gauss_p) {}
  /* </div>
  */
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const HatFunctions1d* eX = dynamic_cast<const HatFunctions1d*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const HatFunctions1d* eY = dynamic_cast<const HatFunctions1d*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size();

      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());

      /* <div class="solution">
	 First get the quadrature rule. The outer loops are for the shape
	 functions (Ansatz and test functions). The inner loop is for the
	 numerical quadrature.

	 The symmetry of the mass matrix is not taken into account.
      */
//       const Real* w = quad_.weights();   // quadrature weights
//       const Real* q = quad_.abscissas(); // quadrature points

//       for (uint i = 0; i < n; ++i) { // loop over shape functions
// 	for (uint j = 0; j < n; ++j) { // loop over shape functions
// 	  Real& m = em(i,j);
// 	  for (uint k = 0; k < quad_.n(); ++k) { // loop over quadrature points
// 	    m += w[k] * h / 2.0
// 	      * eX->shapefct(i, q[k]) * eX->shapefct(j, q[k])
// 	      * 1.0;
// 	  } // for k
// 	} // for j
//       } // for i
      // The plots are done with the following code:
      const Real s = h/3.0/2.0;
      em(0,0) = em(1,1) = 2.0*s;
      em(0,1) = em(1,0) = s;
      /* </div>
       */
    }
  }
};

/* <h3>Main program</h3>

   In the main program, all the classes above are used and built together
   in the right way to solve to PDE.
*/
// ************************************************************ Main Program **
/* <div class="solution">
   The exact solution.
*/
Real exact(const Real x) {
  return -3.0*cosh(x)/cosh(1.0)+x*x+2.0;
}
/* </div>
 */
int main(int argc, char** argv) {
  /* The whole program is in a big <span
     class="code">try...catch</span> brace to catch exceptions thrown
     by the code.
  */
  try {
    /* <div class="solution">
       Default values for the variable which could be set by a command line
       argument
    */
    uint level = 1;
    uint gauss_p = 1;
    /* The <span class="code">inputParser</span> takes care of all input
       parameters and the results which are gathered. At the end, they can
       be printed.
    */
    concepts::InputParser inputParser;
    concepts::InOutParameters& input = inputParser.inputParameters();
    /* Register the parameters.
     */
    input.addInt("level", level);
    input.addInt("gauss_p", gauss_p);
    /* Parse the command line options.
     */
    int opt;
    while ((opt = getopt(argc, argv, "-l:p:")) != EOF)
      switch(opt) {
      case 'l': input.addInt("level", std::atoi(optarg)); break;
      case 'p': input.addInt("gauss_p", std::atoi(optarg)); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-l LEVEL] [-p GAUSS_P]" << std::endl
		  << "where" << std::endl
		  << "  LEVEL: maximal level of refinement" << std::endl
		  << "  GAUSS_P: number of points in Gauss integration"
		  << std::endl;
	std::exit(1); break;
      }
    level = input.getInt("level");
    gauss_p = input.getInt("gauss_p");
    /* Prepare the results part to gather the results and compute the
       exact FE energy and the element size.
     */
    concepts::InOutParameters& output = inputParser.outputParameters();
    const Real exEnergie = 206.0/15.0
      + 36.0*exp(2.0)/(exp(2.0)+1.0)/(exp(2.0)+1.0)*sinh(2.0)
      - 72.0*exp(1.0)/(exp(2.0)+1.0)*sinh(1.0);
    output.addDouble("exEnergie", exEnergie);
    output.addArrayDouble("h");
    output.addArrayDouble("energieFehler");
    output.addArrayDouble("maxNodalFehler");
    /* </div>

       Set up the mesh.
    */
    Line msh;
    std::cout << "Mesh: " << msh << std::endl;
    /* <div class="solution">
       Start the loop over the different levels.
    */
    for (uint l = 1; l < level; ++l) {
      std::cout << "l = " << l << std::endl;
      /* And set up the space according to the current level.
	 </div>
      */
      Linear1d spc(msh, l);
      std::cout << "Space: " << spc << std::endl;
      /* Compute the load vector.
       */
      RHS rhs_lf(gauss_p);
      concepts::Vector<Real> rhs(spc, rhs_lf);
      /* Compute the stiffness matrix and the mass matrix.
      */
      Laplace stiff_bf;
      concepts::DenseMatrix<Real> stiff(spc, stiff_bf);

      Identity mass_bf(gauss_p);
      concepts::DenseMatrix<Real> mass(spc, mass_bf);
      /* Enforce homogeneous Dirichlet boundary conditions.
       */
      const uint n = stiff.spaceX().dim();
      for(uint i = 0; i < n; ++i)
	stiff.concepts::ElementMatrix<Real>::operator()(0,i) =
	  stiff.concepts::ElementMatrix<Real>::operator()(i,0) = 0.0;
      stiff.concepts::ElementMatrix<Real>::operator()(0,0) = 1.0;
      rhs[0] = 0.0;

      for(uint i = 0; i < n; ++i)
	stiff.concepts::ElementMatrix<Real>::operator()(n-1,i) =
	  stiff.concepts::ElementMatrix<Real>::operator()(i,n-1) = 0.0;
      stiff.concepts::ElementMatrix<Real>::operator()(n-1,n-1) = 1.0;
      rhs[n-1] = 0.0;
      /* <div class="solution">
	 This is also necessary for the mass matrix.
      */
      for(uint i = 0; i < n; ++i)
	mass.concepts::ElementMatrix<Real>::operator()(0,i) =
	  mass.concepts::ElementMatrix<Real>::operator()(i,0) = 0.0;
      mass.concepts::ElementMatrix<Real>::operator()(0,0) = 1.0;
      for(uint i = 0; i < n; ++i)
	mass.concepts::ElementMatrix<Real>::operator()(n-1,i) =
	  mass.concepts::ElementMatrix<Real>::operator()(i,n-1) = 0.0;
      mass.concepts::ElementMatrix<Real>::operator()(n-1,n-1) = 1.0;
      /* The linear combination is set up properly.
	 </div>
      */
      concepts::LiCo<Real> L(stiff, mass, 1.0, 1.0);
      /* Solve the linear system iteratively -- catch divergence errors
	 right here.
      */
      concepts::CG<Real> solver(L, 1e-6, 2000);
      concepts::Vector<Real> sol(spc);
      try {
	solver(rhs, sol);
      }
      catch(concepts::NoConvergence& e) {
	std::cout << "solver did not converge!" << std::endl << e << std::endl;
      }
      std::cout << "solver = " << solver << std::endl;
      std::cout << "solution = " << sol << std::endl;
      /* <div class="solution">
	 Gather the results of the computation: maximal nodal error.
      */
      const Real h = 1.0/(1<<l);
      output.addArrayDouble("h", l, h);
      Real x = -1.0;
      Real maxNodalFehler = 0.0;
      Real nodalFehler = 0.0;
      for (int p = 0; p < (2<<l)+1; ++p, x += h) {
	nodalFehler = fabs(sol[p] - exact(x));
	if (nodalFehler > maxNodalFehler)
	  maxNodalFehler = nodalFehler;
      }
      output.addArrayDouble("maxNodalFehler", l, maxNodalFehler);
      /* FE energy error.
       */
      const Real feEnergie(sol*rhs);
      output.addArrayDouble("energieFehler", l,
			    fabs(exEnergie-feEnergie)/exEnergie);
      /* End of the loop over the different levels.
       */
    } // for l
    /* Print the results.
     */
    std::cout << inputParser << std::endl;
    /* </div>
     */
  }
  /* Here, the exceptions are caught and some output is generated. Normally,
     this should not happen.
  */
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  /* Sane return value.
   */
  return 0;
}
