/* System include files. The first is for input and output operations
   and the second one declares <span class="code">auto_ptr</span>.
   <span class="code">auto_ptr</span> is used to make the handling of
   dynamic objects (which are created with <span
   class="code">new</span>) easier: <span class="code">delete</span>
   is no longer necessary as the variable is destroyed automatically
   at the end of the block. (More information can be found in the <a
   href="http://wwwinfo.cern.ch/asd/lhc++/RW/stdlibcr/aut_3512.htm">Standard
   C++ Library Class Reference of Rogue Wave Software</a> or in <a
   href="http://cseng.aw.com/book/related/0,3833,020163371X+80,00.html">Scott
   Meyers' More Effective C++</a>.)
 */
#include <iostream>
#include <memory>
/* Concepts include files.
 */
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
/* With this using directive, we do not need to prepend <span
   class="code">concepts::</span> before <span
   class="code">Real</span> everytime. (More information on namespaces
   and how to used them can be found in the <a
   href="http://www.cplusplus.com/doc/tutorial/tut5-2.html">Complete
   C++ language tutorial of cplusplus.com</a> or in an article of <a
   href="http://www.mvps.org/windev/cpp/nspaces.html">Tomas
   Restrepo</a>).
*/
using concepts::Real;

/* <h3>Mesh</h3>

   The mesh is taken from the <a href="meshes.html#line">meshes tutorial</a>:
   <span class="code">Line</span>. Here, the relevant code is included:
*/
#include "meshes.cc"

/* <h3>Element</h3>

   There is not much happening in the element yet. At a later time,
   one wants to be able to evaluate the shape functions at any point
   in the reference element using this class. Right now, only the <a
   href="../#classDocumentation#classconcepts_1_1TMatrixBase.html">T
   matrices</a> are generated from the information coming from the
   space (ie. the indices in <span class="code">idx</span>) in the
   constructor.

   As described <a href="../geometry.html">here</a>, an element
   contains a cell. A cell holds geometrical (ie. the coordinates) and
   topological (ie. neighbours, connectivity etc.) information. The
   member function <span class="code">support()</span> returns the
   topological part of the element and <span
   class="code">vertex(i)</span> returns the coordinates of the ith
   vertex.
*/
// ***************************************************************** Element **

class HatFunctions1d : public concepts::Element<Real> {
public:
  HatFunctions1d(const concepts::Edge1d& cell, uint idx[]);
  const concepts::TMatrixBase<Real>& T() const { return *T_; }
  const concepts::Edge& support() const { return cell_.connector(); }
  concepts::Real3d vertex(const uint i) const { return cell_.vertex(i); }
  const concepts::Edge1d& cell() const { return cell_; }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  const concepts::Edge1d& cell_;
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
};

HatFunctions1d::HatFunctions1d(const concepts::Edge1d& cell, uint idx[]) :
  cell_(cell), T_(new concepts::TIndex<Real>(2,2,idx)) {}

std::ostream& HatFunctions1d::info(std::ostream& os) const {
  return os << "HatFunctions1d(" << cell_ << ", T = " << *T_ << ')';
}

/* <h3>Scanner for the Elements</h3>

   This declaration indicates the relationship between a scanner over generic
   elements and a scanner over elements of type
   <span class="code">HatFunctions1d</span>. The declaration has to be in
   the namespace <span class="code">concepts</span>.
*/
// ***************************************************************** Scanner **

namespace concepts {
  template<>
  class Scan<HatFunctions1d> : public Scan<Element<Real> > {
  public:
    HatFunctions1d& operator++(int) = 0;
  };
} // namespace concepts

/* <h3>Space</h3>

   A space ties the elements together and has methods to loop over the
   elements with a scanner (as a mesh has methods to loop over the
   cells of the mesh). This is used in the assembling procedure.

   In the constructor of the space, the degrees of freedom are counted
   and the continuity is enforced. This is done using so called T matrices
   which are used for the assembling. The T matrices are stored in the
   elements.
*/
// ******************************************************************* Space **

class Linear1d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<HatFunctions1d> Scan;
  Linear1d(concepts::Mesh1& line, uint level);
  virtual ~Linear1d();
  uint dim() const { return dim_; }
  uint nelm() const { return nelm_; }
  Scan* scan() const {
    return new concepts::PListScan<HatFunctions1d>(*elm_);
  }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  concepts::Joiner<HatFunctions1d*, 1>* elm_;
  void constructor_(concepts::Edge1d* cell, uint level,
		    concepts::DynArray<uint>& map);
};

/* The constructor calls <span class="code">constructor_</span> on
   every suitable element. <span class="code">map</span> is a dynamic
   array containing the global number of the degree of freedom, if the
   degree of freedom was already visited. The default value of an
   entry in <span class="code">map</span> is <span
   class="code">(uint)-1</span>, which is the largest value an <span
   class="code">uint</span> (unsigned integer) can take. The dynamic
   array <a
   href="../#classDocumentation#classconcepts_1_1DynArray.html"
   class="code">DynArray</a> adjusts its size on its own in blocks of
   size 8 = 2<sup>3</sup>.
*/
Linear1d::Linear1d(concepts::Mesh1& line, uint level) :
  dim_(0), nelm_(0), elm_(0) {

  concepts::DynArray<uint> map(2, 3, (uint)-1);
  std::auto_ptr<concepts::Scan1> sc(line.scan());
  while (*sc) {
    concepts::Cell1& cell = (*sc)++;
    concepts::Edge1d* cellE = dynamic_cast<concepts::Edge1d*>(&cell);
    if (cellE)
      constructor_(cellE, level, map);
    else
      throw conceptsException(concepts::MissingFeature("cell not supported"));
  }
}

Linear1d::~Linear1d() {
  delete elm_;
}

std::ostream& Linear1d::info(std::ostream& os) const {
  return os << "Linear1d(dim = " << dim_ << ", nelm = " << nelm_ << ')';
}

/* <span class="code">constructor_</span> is called recursively
   on the children of the current cell if the
   correct level of refinement is not reached yet. Otherwise, the degrees
   of freedom are assigned and a new element is added to the list of
   elements.
*/
void Linear1d::constructor_(concepts::Edge1d* cell, uint level,
			    concepts::DynArray<uint>& map) {
  if (level) {
    --level;
    constructor_(cell->child(0), level, map);
    constructor_(cell->child(1), level, map);
  } else {
    /* <span class="code">idx</span> is going to hold the global
       numbers of the degrees of freedom of the two vertices. The size
       of <span class="code">idx</span> is 2 as there are two local
       degrees of freedom in each element (namely the two linear shape
       functions which are going to be combined to hat functions).
    */
    uint idx[2];
    for (uint j = 0; j < 2; ++j) { // loop over the two vertices
      /* <span class="code">i</span> is a reference to the entry in
	 <span class="code">map</span> which corresponds to the jth
	 vertex of this cell. Every <a
	 href="../#classDocumentation#classconcepts_1_1Vertex.html"
	 class="code">Vertex</a> as a topological entity holds a
	 unique number (its <a
	 href="../#classDocumentation#classconcepts_1_1Key.html"
	 class="code">Key</a>). This key is used as the index of
	 the array <span class="code">map</span>.

	 If <span class="code">i</span> is equal to <span
	 class="code">(uint)-1</span>, there is no degree of freedom
	 associated to this vertex yet. Therefore, the dimension of
	 the space has to be increased and a new degree of freedom is
	 allocated in the same step.

	 Finally, the global number of the degree of freedom is stored
	 in <span class="code">idx</span> to associated it to the
	 local degree of freedom.
      */
      uint& i = map[cell->connector().vertex(j)->key()];
      if (i == (uint) -1)
	i = dim_++;
      idx[j] = i;
    } // for i
    /* All necessary information (<span class="code">idx</span>) is
       collected and a new element is created. The number of elements
       of the space (<span class="code">nelm_</span>) is increased and
       the element is added to the list of elements (<span
       class="code">elm_</span>).
    */
    HatFunctions1d* elm = new HatFunctions1d(*cell, idx);
    nelm_++;
    elm_ = new concepts::Joiner<HatFunctions1d*, 1>(elm, elm_);
  }
}

/* <h3>Linear Form</h3>

   The linear form is used to compute the element contributions to the
   load vector. The computations are done in the application operator
   of this class.
 */
// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    /* Resize the element load vector to the correct size and set all entries
       to zero. This way, the computed values can be added into
       <span class="code">em</span>.
    */
    em.resize(n, 1);
    em.zeros();

    /* <span class="math">f = 1</span>, hat functions living on <span
       class="math">(-1,1)</span> and mapped onto the element.
    */
    const HatFunctions1d* e = dynamic_cast<const HatFunctions1d*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    const Real h = e->cell().size();
    for (uint i = 0; i < n; ++i) {
      em(i,0) = h/2.0;
    }
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

/* The application operator of the class <span class="code">Laplace</span>
   computes the element matrix belonging to the Laplacian in the PDE.
*/
class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const HatFunctions1d* eX = dynamic_cast<const HatFunctions1d*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const HatFunctions1d* eY = dynamic_cast<const HatFunctions1d*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    /* Since the intersection of the supports of the two different
       elements has at most measure zero, we do only integrate if the
       elements are equal.

       Here, we don't integrate, since the element stiffness matrices can
       be computed beforehand.
    */
    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size();

      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());

      const Real s = 2.0/h;
      em(0,0) = em(1,1) = 1.0/2.0*s;
      em(0,1) = em(1,0) = -1.0/2.0*s;
    }
  }
};

/* The application operator of the class <span class="code">Identity</span>
   computes the element matrix belonging to the absolute term in the PDE.
   This code is not really used. Therefore, it does not compute anything.
*/
class Identity : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    // Is not used, ie. no code here.
  }
};

/* <h3>Main program</h3>

   In the main program, all the classes above are used and built together
   in the right way to solve the PDE.
*/
// ************************************************************ Main Program **

int main(int argc, char** argv) {
  /* The whole program is in a big <span
     class="code">try...catch</span> brace to catch exceptions thrown
     by the code.
  */
  try {
    /* Set up the mesh and the space.
     */
    Line msh;
    std::cout << "Mesh: " << msh << std::endl;

    Linear1d spc(msh, 1);
    std::cout << "Space: " << spc << std::endl;
    /* Compute the load vector.
     */
    RHS rhs_lf;
    concepts::Vector<Real> rhs(spc, rhs_lf);
    std::cout << "rhs = " << rhs << std::endl;
    /* Compute the stiffness matrix and the mass matrix. The mass matrix
       is not really used afterwards.
    */
    Laplace stiff_bf;
    concepts::DenseMatrix<Real> stiff(spc, stiff_bf);

    Identity mass_bf;
    concepts::DenseMatrix<Real> mass(spc, mass_bf);
    /* Enforce homogeneous Dirichlet boundary conditions. The first
       and last row and column of the stiffness matrix are erased and
       a 1 is placed on the diagonal element. Together with setting
       the first and the last entry of <span class="code">rhs</span>
       to 0, this enforces the solution <span class="code">sol</span>
       to be 0 in the first and the last entry (which correspond to
       the rightmost and leftmost points in the mesh).
    */
    const uint n = stiff.spaceX().dim();
    for(uint i = 0; i < n; ++i)
      stiff.concepts::ElementMatrix<Real>::operator()(0,i) =
	stiff.concepts::ElementMatrix<Real>::operator()(i,0) = 0.0;
    stiff.concepts::ElementMatrix<Real>::operator()(0,0) = 1.0;
    rhs[0] = 0.0;

    for(uint i = 0; i < n; ++i)
      stiff.concepts::ElementMatrix<Real>::operator()(n-1,i) =
	stiff.concepts::ElementMatrix<Real>::operator()(i,n-1) = 0.0;
    stiff.concepts::ElementMatrix<Real>::operator()(n-1,n-1) = 1.0;
    rhs[n-1] = 0.0;
    /* <a href="../#classDocumentation#classconcepts_1_1LiCo.html"
       class="code">LiCo</a> combines <span class="code">stiff</span>
       und <span class="code">mass</span> in a linear combinatoin with
       the given coefficients.
    */
    concepts::LiCo<Real> L(stiff, mass, 1.0, 0.0);
    std::cout << "L = stiff + mass = " << L << std::endl;
    /* Solve the linear system iteratively -- catch divergence errors
       right here.
    */
    concepts::CG<Real> solver(L, 1e-6, 20);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
    std::cout << "solution = " << sol << std::endl;
  }
  /* Here, the exceptions are caught and some output is generated. Normally,
     this should not happen.
  */
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
  /* Sane return value.
   */
  return 0;
}
