/* The include files are the same as in the previous tutorial.
 */
#include <iostream>
#include <memory>
#include <sstream>
#include <unistd.h> // for command line parsing
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
using concepts::Real;

/* <h3>Mesh</h3>

   The domain is <span class="math">(-1,1)<sup>2</sup></span>. It is meshed
   by two triangles.
*/
// ******************************************************************** Mesh **

class SquareTriangle : public concepts::Mesh2 {
public:
  SquareTriangle();
  virtual ~SquareTriangle();
  unsigned int ncell() const { return 2; }
  concepts::Scan2* scan() { return new S(cell_); }
  virtual std::ostream& info(std::ostream& os) const;
private:
  concepts::Vertex *vtx_[4];
  concepts::Edge *edg_[5];
  concepts::Triangle *tri_[2];
  concepts::Cell2 *cell_[2];

  class S : public concepts::Scan<concepts::Cell2> {
    unsigned int idx_;
    concepts::Cell2 *(&cell_)[2];
  public:
    S(concepts::Cell2 *(&cell)[2]) : idx_(0), cell_(cell) {}
    S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    bool eos() const { return idx_ == 2; }
    concepts::Cell2& operator++(int) { return *cell_[idx_++]; }
    concepts::Scan2* clone() const { return new S(*this); }
  };
};

SquareTriangle::SquareTriangle() {
  /* <div class="solution">
     The attributes are fixed in a way that one vertex has an attribute
     and the edges on the boundaries have attributes.
  */
  vtx_[0] = new concepts::Vertex(5);
  for (uint i = 1; i < 4; ++i)
    vtx_[i] = new concepts::Vertex();

  edg_[0] = new concepts::Edge(*vtx_[0], *vtx_[1], 1);
  edg_[1] = new concepts::Edge(*vtx_[2], *vtx_[1], 2);
  edg_[2] = new concepts::Edge(*vtx_[0], *vtx_[2]);
  edg_[3] = new concepts::Edge(*vtx_[2], *vtx_[3], 3);
  edg_[4] = new concepts::Edge(*vtx_[0], *vtx_[3], 4);
  /* </div>
   */
  tri_[0] = new concepts::Triangle(*edg_[0], *edg_[1], *edg_[2]);
  tri_[1] = new concepts::Triangle(*edg_[2], *edg_[3], *edg_[4]);

  cell_[0] = new concepts::Triangle2d
    (*tri_[0], concepts::VertexTriangle2d(concepts::Real2d(-1.0, -1.0),
					  concepts::Real2d(1.0, -1.0),
					  concepts::Real2d(1.0, 1.0)));
  cell_[1] = new concepts::Triangle2d
    (*tri_[1], concepts::VertexTriangle2d(concepts::Real2d(-1.0, -1.0),
					  concepts::Real2d(1.0, 1.0),
					  concepts::Real2d(-1.0, 1.0)));
}

SquareTriangle::~SquareTriangle() {
  for (int i = 0; i < 2; ++i)
    delete cell_[i];
  for (int i = 0; i < 2; ++i)
    delete tri_[i];
  for (int i = 0; i < 5; ++i)
    delete edg_[i];
  for (int i = 0; i < 4; ++i)
    delete vtx_[i];
}

std::ostream& SquareTriangle::info(std::ostream& os) const {
  os << "SquareTriangle(" << *cell_[0] << ", " << *cell_[1] << ')';
  return os;
}

/* <h3>Elements</h3>
*/
// **************************************************************** Elements **

class FEMElement2d : public concepts::Element<Real> {
public:
  FEMElement2d(const uint i, uint idx[]) :
    T_(new concepts::TIndex<Real>(i,i,idx)) {}
  FEMElement2d(concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1);
  const concepts::TMatrixBase<Real>& T() const { return *T_; }
  virtual concepts::Real2d shapefctD(const uint i) const = 0;
  virtual Real shapefct(const uint i,
			const Real xi1, const Real xi2) const = 0;
protected:
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
};

FEMElement2d::FEMElement2d(concepts::TColumn<Real>* T0,
			   concepts::TColumn<Real>* T1) :
  T_(new concepts::TMatrix<Real>(T1)) {
  if (T0) {
    concepts::TMatrix<Real>* T =
      dynamic_cast<concepts::TMatrix<Real>*>(T_.get());
    conceptsAssert(T, concepts::Assertion());
    T->append(T0);
  }
}

/* <h4>Triangle</h4>
 */
class FEMTriangle : public FEMElement2d {
public:
  FEMTriangle(const concepts::Triangle2d& cell, uint idx[]);
  FEMTriangle(const concepts::Triangle2d& cell, concepts::TColumn<Real>* T0,
	      concepts::TColumn<Real>* T1 = 0);
  const concepts::Triangle& support() const { return cell_.connector(); }
  concepts::Real3d vertex(const uint i) const { return cell_.vertex(i); }
  const concepts::Triangle2d& cell() const { return cell_; }
  virtual concepts::Real2d shapefctD(const uint i) const;
  virtual Real shapefct(const uint i,
				    const Real xi1, const Real xi2) const;
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  const concepts::Triangle2d& cell_;
};

FEMTriangle::FEMTriangle(const concepts::Triangle2d& cell, uint idx[]) :
  FEMElement2d(3, idx), cell_(cell) {}

FEMTriangle::FEMTriangle(const concepts::Triangle2d& cell,
			 concepts::TColumn<Real>* T0,
			 concepts::TColumn<Real>* T1) :
  FEMElement2d(T0, T1), cell_(cell) {}

concepts::Real2d FEMTriangle::shapefctD(const uint i) const {
  switch(i) {
  case 0: return concepts::Real2d(-1, -1);
  case 1: return concepts::Real2d(1, 0);
  case 2: return concepts::Real2d(0, 1);
  default:
    throw conceptsException(concepts::MissingFeature("wrong shapefct. index"));
  }
}

Real FEMTriangle::shapefct(const uint i,
			   const Real xi1, const Real xi2) const {
  conceptsAssert(xi1 >= 0.0, concepts::Assertion());
  conceptsAssert(xi2 >= 0.0, concepts::Assertion());
  conceptsAssert(xi1+xi2 <= 1.0, concepts::Assertion());
  switch(i) {
  case 0: return 1.0-xi1-xi2;
  case 1: return xi1;
  case 2: return xi2;
  default:
    throw conceptsException(concepts::MissingFeature("wrong shapefct. index"));
  }
}

std::ostream& FEMTriangle::info(std::ostream& os) const {
  return os << "FEMTriangle(" << cell_ << ", T = " << *T_ << ')';
}

/* <h3>Scanner for the Elements</h3>
*/
// ***************************************************************** Scanner **

namespace concepts {
  template<>
  class Scan<FEMTriangle> : public Scan<Element<Real> > {
  public:
    FEMTriangle& operator++(int) = 0;
  };
} // namespace concepts

/* <h3>Space</h3>

   The space in 2D looks almost the same as the space in 1D with the slight
   difference that the Dirichlet boundary conditions are enforced while
   creating the space.

   This is done by not assembling the degrees of freedom on a Dirichlet
   boundary. Hence, the system matrix gets smaller: the degrees of freedom
   which are 0 anyway are left out.
*/
// ******************************************************************* Space **

class Linear2d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<concepts::Element<Real> > Scan;
  Linear2d(concepts::Mesh2& msh, uint level,
	   concepts::BoundaryConditions* bc = 0);
  virtual ~Linear2d();
  uint dim() const { return dim_; }
  uint nelm() const { return nelm_; }
  Scan* scan() const {
    return new concepts::PListScan<concepts::Element<Real> >(*elm_);
  }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  concepts::BoundaryConditions* bc_;
  concepts::Joiner<concepts::Element<Real>*, 1>* elm_;
  void enforceBC_(concepts::Cell2* cell, uint level,
		  concepts::DynArray<uint>& bcMap);
  void constructor_(concepts::Cell2* cell, uint level,
		    concepts::DynArray<uint>& dofMap,
		    const concepts::DynArray<uint>& bcMap);
};

/* The constructor first calls <span class="code">enforceBC_</span> on
   each element which notes where the Dirichlet boundary conditions
   have to be enforced and then calls <span
   class="code">constructor_</span> (same as in the 1D code).
*/
Linear2d::Linear2d(concepts::Mesh2& msh, uint level,
		   concepts::BoundaryConditions* bc) :
  dim_(0), nelm_(0), bc_(bc), elm_(0) {

  /* <span class="code">bcMap</span> stores for every visited vertex
     if it has a valid degree of freedom (1) or not (0). <span
     class="code">bcMap</span> is filled in <span
     class="code">enforceBC_</span> depending on the level and
     boundary conditions.
  */
  concepts::DynArray<uint> bcMap(2, 3, 1);
  std::auto_ptr<concepts::Scan2> sc(msh.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    enforceBC_(&cell, level, bcMap);
  }
  /* <span class="code">dofMap</span> stores for every visited vertex the
     assigned degree of freedom or the value <span class="code">(uint)-1</span>
     if no degree of freedom is assigned yet.
  */
  concepts::DynArray<uint> dofMap(2, 3, (uint)-1);
  sc.reset(msh.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    concepts::Triangle2d* cellT = dynamic_cast<concepts::Triangle2d*>(&cell);
    if (cellT) {
      constructor_(cellT, level, dofMap, bcMap);
      continue;
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }
}

Linear2d::~Linear2d() {
  delete elm_;
}

std::ostream& Linear2d::info(std::ostream& os) const {
  return os << "Linear2d(dim = " << dim_ << ", nelm = " << nelm_ << ')';
}

/* <span class="code">enforceBC_</span> calls itself recursively if the
   desired level of refinement is not reached yet. The code in this routine
   is independent of the exact element type.
*/
void Linear2d::enforceBC_(concepts::Cell2* cell, uint level,
			  concepts::DynArray<uint>& bcMap) {
  if (level) {
    --level;
    for (uint i = 0; i < 4; ++i)
      enforceBC_(cell->child(i), level, bcMap);
  } else {
    /* If the desired level of refinement is reached, every edge of the
       element is checked for Dirichlet boundary conditions.

       If an edge has Dirichlet boundary conditions, both endpoints are
       marked: the corresponding degrees of freedom will not be assembled.
       If no Dirichlet boundary conditions are specified on the edge,
       the entries in <span class="code">bcMap</span>
       are touched to make sure they exist.
    */
    concepts::Connector1* edg;
    for (uint j = 0; (edg = cell->connector().edge(j)) != 0; ++j) {
      if ((*bc_)(edg->attrib()).type() == concepts::Boundary::DIRICHLET) {
	bcMap[edg->vertex(0)->key()] = 0;
	bcMap[edg->vertex(1)->key()] = 0;
      } else {
	bcMap[edg->vertex(0)->key()];
	bcMap[edg->vertex(1)->key()];
      }
    }
    /* <div class="solution">
       In addition to the check above, we check all vertices for Dirichlet
       boundary conditions (with this method, a single vertex may be fixed).
    */
    concepts::Connector0* vtx;
    for (uint j = 0; (vtx = cell->connector().vertex(j)) != 0; ++j) {
      if ((*bc_)(vtx->attrib()).type() == concepts::Boundary::DIRICHLET)
	bcMap[vtx->key()] = 0;
      else
	bcMap[vtx->key()];
    }
    /* </div>
     */
  }
}

/* As in the 1D case, <span class="code">constructor_</span> calls itself
   recursively if the desired level of refinement is not reached yet.
*/
void Linear2d::constructor_(concepts::Cell2* cell, uint level,
			    concepts::DynArray<uint>& dofMap,
			    const concepts::DynArray<uint>& bcMap) {
  if (level) {
    --level;
    for (uint i = 0; i < 4; ++i)
      constructor_(cell->child(i), level, dofMap, bcMap);
  } else {
    /* If the desired level of refinement is reached, the degrees of
       freedom are counted. A triangle has three local shape functions
       (a quadrilateral would have four).
    */
    concepts::Triangle2d* cellT = dynamic_cast<concepts::Triangle2d*>(cell);
    uint n = 0;
    if (cellT)
      n = 3;
    conceptsAssert(n, concepts::Assertion());
    concepts::TColumn<Real>* T = 0;
    concepts::Connector0* vtx;
    /* In the loop over all vertices of the element, a new column is
       added to the list of columns of the T matrix for every degree
       of freedom. This is done only if the vertex does not have
       Dirichlet boundary conditions (the information is stored in
       <span class="code">bcMap</span>).

       <span class="code">i</span> is the index of the global basis
       function and <span class="code">j</span> is the index of the local
       shape function.
    */
    for (uint j = 0; (vtx = cell->connector().vertex(j)) != 0; ++j) {
      if (bcMap[vtx->key()]) {
	uint& i = dofMap[vtx->key()];
	if (i == (uint)-1)
	  i = dim_++;
	T = new concepts::TColumn<Real>(n, i, T);
	(*T)[j] = 1.0;
      }
    }
    /* After all vertices are checked, the new element is created with the
       assembling information in the T columns.
    */
    concepts::Element<Real>* elm;
    if (cellT) {
      elm = new FEMTriangle(*cellT, T);
    }
    nelm_++;
    elm_ = new concepts::Joiner<concepts::Element<Real>*, 1>(elm, elm_);
  }
}

/* <h3>Linear Form</h3>
 */
// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
private:
  const concepts::Quadrature<4> quad_;
  concepts::BoundaryConditions* bc_;
  Real f(const Real x1, const Real x2) const {
    return 4.0*(1.0+x1)*(1.0+x2)*(2.0-x1-x2);
  }
public:
  RHS(const uint gauss_p = 1, concepts::BoundaryConditions* bc = 0) :
    quad_(gauss_p), bc_(bc) {}
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();

    const FEMTriangle* e = dynamic_cast<const FEMTriangle*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    concepts::Real2d x0(e->vertex(0));
    concepts::Real2d x1(e->vertex(1));
    concepts::Real2d x2(e->vertex(2));

    const concepts::MapReal2d B_T(x1-x0, x2-x0);
    const concepts::Real2d b(x0);
    const Real detB_T = B_T.determinant();

    const Real* q = quad_.abscissas();
    const Real* w = quad_.weights();

    for(uint k = 0; k < quad_.n(); k++) { // loop over xi_1 quad. points
      for(uint l = 0; l < quad_.n(); l++) { // loop over xi_2 quad.
	const concepts::Real2d xi((1.0+q[k])*(1.0-q[l])/4.0,
				  (1.0+q[l])/2.0);
	const concepts::Real2d x(B_T * xi + b);

	const Real F = f(x[0], x[1]);
	const Real dx = w[k] * w[l] * detB_T * (1.0-q[l])/8.0;
	for (uint i = 0; i < n; ++i) {
	  em(i,0) += F*dx*e->shapefct(i, xi[0], xi[1]);
	} // for i
      } // for l
    } // for k

    if (bc_) {
      // the element map
      concepts::Real2d x0(e->vertex(0));
      concepts::Real2d x1(e->vertex(1));
      concepts::Real2d x2(e->vertex(2));
      const concepts::MapReal2d B_T(x1-x0, x2-x0);
      const concepts::Real2d b(x0);

      concepts::Edge* edg = 0;
      for (uint i = 0; (edg = e->support().edge(i)) != 0; ++i) {
	concepts::Attribute attrib = edg->attrib();
	if ( (*bc_)(attrib).type() == concepts::Boundary::NEUMANN ) {
	  const concepts::Real2d x0(e->vertex(i));
	  const concepts::Real2d x1(e->vertex( (i+1)%3 ));
	  const concepts::Real2d edge(x1-x0);
	  const Real length = edge.l2();

	  const Real* q = quad_.abscissas();
	  const Real* w = quad_.weights();
	  for (uint k = 0; k < quad_.n(); ++k) { // integration

	    Real qk = 0.0, ql = 0.0;
	    switch(i) {
	    case 0: qk = q[k]; ql = -1.0; break;
	    case 1: qk = 1.0; ql = q[k]; break;
	    case 2: qk = -1.0; ql = q[k]; break;
	    default: conceptsAssert(1 == 0, concepts::Assertion());
	    }

	    const concepts::Real2d xi((1.0+qk)*(1.0-ql)/4.0,
				      (1.0+ql)/2.0);
	    const concepts::Real2d x(B_T * xi + b);

	    Real f = (*bc_)(attrib)(x) * length / 2.0 * w[k];

	    for (uint j = 0; j < n; ++j) // shape functions
	      em(j,0) += e->shapefct(j, xi[0], xi[1]) * f;
	  } // for k
	} // Neumann bc
      } // for i
    } // if bc
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

class Laplace : public concepts::BilinearForm<Real> {
private:
  const concepts::Quadrature<4> quad_;
  concepts::MapReal2d a(const Real x1, const Real x2) const {
    return concepts::MapReal2d(2.0+x1, 0.0, 0.0, 2.0+x2);
  }
public:
  Laplace(const uint gauss_p = 1) : quad_(gauss_p) {}
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const FEMTriangle* eX = dynamic_cast<const FEMTriangle*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const FEMTriangle* eY = dynamic_cast<const FEMTriangle*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      concepts::Real2d x0(eX->vertex(0));
      concepts::Real2d x1(eX->vertex(1));
      concepts::Real2d x2(eX->vertex(2));

      const concepts::MapReal2d B_T(x1-x0, x2-x0);
      const concepts::Real2d b(x0);
      const Real detB_T = B_T.determinant();
      const concepts::MapReal2d B_TinvT(B_T.inverse().transpose());

      const Real* q = quad_.abscissas();
      const Real* w = quad_.weights();

      Real inta[4] = { 0.0, 0.0, 0.0, 0.0 };

      for(uint k = 0; k < quad_.n(); k++) { // loop over xi_1 quad. points
	for(uint l = 0; l < quad_.n(); l++) { // loop over xi_2 quad.

	  const concepts::Real2d xi((1.0+q[k])*(1.0-q[l])/4.0,
				    (1.0+q[l])/2.0);
	  const concepts::Real2d x(B_T * xi + b);

	  const concepts::MapReal2d A(a(x[0], x[1]));
	  const Real dDdet = (1.0-q[l])/4.0*0.5;
	  const Real dx = fabs(w[k] * w[l] * detB_T * dDdet);
	  inta[0] += dx * A.column(0)[0];
	  inta[1] += dx * A.column(1)[0];
	  inta[2] += dx * A.column(0)[1];
	  inta[3] += dx * A.column(1)[1];
	} // for l
      } // for k

      const concepts::MapReal2d A(inta[0], inta[1], inta[2], inta[3]);

      for (uint i = 0; i < n; ++i) { // loop over u shape fcts.
	const concepts::Real2d u(B_TinvT * eX->shapefctD(i));
	for (uint j = 0; j < n; ++j) { // loop over v shape fcts.
	  const concepts::Real2d v(B_TinvT * eX->shapefctD(j));
	  em(i,j) += (A * u) * v;
	} // for j
      } // for i
    }
  }
};

/* <div class="solution">
   <h3>Solver: cg</h3>

   This cg variant includes a projection of the approximated solution
   in every step onto the orthogonal complement of the kernel of the
   matrix.  Here, the kernel is spanned by at most one vector (e)
   which can be given as a parameter to the constructor of the solver.

   Most of the code of the following class is copied from the cg algorithm
   in the library.
*/
// ****************************************************************** Solver **

class projCG : public concepts::Operator<Real> {
public:
  projCG(concepts::Operator<Real>& A, Real maxeps, int maxit = 0,
	 const concepts::Vector<Real>* e = 0) :
    W_(0), A_(A), maxeps_(maxeps*maxeps), maxit_(maxit), eps_(1.0),
    it_(0), e_(e) {}
  const concepts::Space<Real>& spaceX() const { return A_.spaceY(); }
  const concepts::Space<Real>& spaceY() const { return A_.spaceX(); }
  uint iterations() const { return it_; }
  Real epsilon() const { return eps_; }
  void operator()(const concepts::Function<Real>& fncY,
		  concepts::Function<Real>& fncX) {
    const concepts::Vector<Real>* fncYA;
    if ((fncYA = dynamic_cast<const concepts::Vector<Real>*>(&fncY))) {
      concepts::Vector<Real>* fncXA;
      if ((fncXA = dynamic_cast<concepts::Vector<Real>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException
      (concepts::MissingFeature("function not supported"));
  }
  void operator()(const concepts::Vector<Real>& fncY,
		  concepts::Vector<Real>& fncX) {
    if (W_ == 0) {
      fncX = fncY;
      /* The first approximation of the solution has to be projected.
       */
      if (e_) projectOntoComplementOfe_(fncX);
      concepts::Vector<Real> r(fncY.space()); A_(fncX, r); r -= fncY;
      /* The first residual has to be projected as the updates to the
	 approximation of the solution are taken from a vector space which
	 is spanned (among others) by this first residual.
      */
      if (e_) projectOntoComplementOfe_(r);
      concepts::Vector<Real> p(fncX.space()); p = r;
      concepts::Vector<Real> Ap(fncY.space());

      eps_ = r.l2_2();
      it_  = 0;

      for(it_ = 0; !(eps_ < maxeps_ || (maxit_ > 0 && it_ > maxit_)); ++it_) {
	A_(p, Ap);

	Real lopt = eps_ / (Ap * p);
	fncX.add(p, -lopt);
	r.add(Ap, -lopt);

	Real eps = r.l2_2();
	p *= eps / eps_; p += r;

	eps_ = eps;
      }

      if (!(eps_ < maxeps_)) {
	std::stringstream errMsg;
	errMsg << "it: " << it_ << ", eps: " << eps_
	       << ", wanted: " << maxeps_ << std::ends;
	throw conceptsException(concepts::NoConvergence(errMsg.str()));
      }
    } else
      throw conceptsException
	(concepts::MissingFeature("operator W not supported"));
  }
private:
  /* The projection method computes the projection of x onto the
     kernel of the matrix and subtracts it from x.
  */
  void projectOntoComplementOfe_(concepts::Vector<Real>& x) const {
    conceptsAssert(e_, concepts::Assertion());
    const Real c = (*e_) * x / e_->l2_2();
    x.add(*e_, -c);
  }
  concepts::Operator<Real>*  W_;
  concepts::Operator<Real>&  A_;
  Real maxeps_;
  uint maxit_;
  Real eps_;
  uint it_;
  const concepts::Vector<Real>* e_;
};
/* </div>
   <h3>Main program</h3>
*/
// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    uint level = 1;
    uint gauss_p = 3;

    concepts::InputParser inputParser;
    concepts::InOutParameters& input = inputParser.inputParameters();

    input.addInt("level", level);
    input.addInt("gauss_p", gauss_p);
    /* Parse the command line options.
     */
    int opt;
    while ((opt = getopt(argc, argv, "-l:p:")) != EOF)
      switch(opt) {
      case 'l': input.addInt("level", std::atoi(optarg)); break;
      case 'p': input.addInt("gauss_p", std::atoi(optarg)); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-l LEVEL] [-p GAUSS_P]" << std::endl
		  << "where" << std::endl
		  << "  LEVEL: maximal level of refinement" << std::endl
		  << "  GAUSS_P: number of points in Gauss integration"
		  << std::endl;
	std::exit(1); break;
      }
    level = input.getInt("level");
    gauss_p = input.getInt("gauss_p");

    concepts::InOutParameters& output = inputParser.outputParameters();
    const Real exEnergie = 512.0/45.0;
    output.addDouble("exEnergie", exEnergie);
    output.addArrayDouble("h");
    output.addArrayDouble("energieFehler");
    output.addArrayDouble("energie");

    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    /* <div class="solution">
       The boundary conditions. The boundary condition with the number 5 is
       only used for the case where the solution is fixed in one point.
    */
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::NEUMANN,
			       "(2*(x*x-1))"));
    bc->add(concepts::Attribute(2),
	    concepts::Boundary(concepts::Boundary::NEUMANN,
			       "(6*(y*y-1))"));
    bc->add(concepts::Attribute(3),
	    concepts::Boundary(concepts::Boundary::NEUMANN,
			       "(6*(x*x-1))"));
    bc->add(concepts::Attribute(4),
	    concepts::Boundary(concepts::Boundary::NEUMANN,
			       "(2*(y*y-1))"));
//     bc->add(concepts::Attribute(5),
// 	    concepts::Boundary(concepts::Boundary::DIRICHLET, "(0)"));
    std::cout << "Boundary conditions: " << *bc << std::endl;
    /* </div>
     */
    SquareTriangle msh;
    std::cout << "Mesh: " << msh << std::endl;

    for (uint l = 0; l < level; ++l) {
      std::cout << "l = " << l << std::endl;

      Linear2d spc(msh, l, bc.get());
      std::cout << "Space: " << spc << std::endl;

      Laplace stiff_bf(gauss_p);
      concepts::PETScMat stiff(spc, stiff_bf);

      RHS rhs_lf(gauss_p, bc.get());
      concepts::Vector<Real> rhs(spc, rhs_lf);

      concepts::Vector<Real> e(spc);
      e = 1.0;
//       projCG solver(stiff, 1e-6, 2000, &e);
      concepts::CG<Real> solver(stiff, 1e-6, 2000);
      concepts::Vector<Real> sol(spc);
      try {
	solver(rhs, sol);
      }
      catch(concepts::NoConvergence& e) {
	std::cout << "solver did not converge!" << std::endl << e << std::endl;
      }
      std::cout << "solver = " << solver << std::endl;
      /* Gather the results of the computation.
       */
      const Real h = 1.0/(1<<l);
      output.addArrayDouble("h", l, h);
      const Real feEnergie(sol*rhs);
      output.addArrayDouble("energie", l, feEnergie);
      output.addArrayDouble("energieFehler", l,
			    fabs(exEnergie-feEnergie)/exEnergie);
    } // for l
    /* Print the results.
     */
    std::cout << inputParser << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
