/* The include files are the same as in the previous tutorial.
 */
#include <iostream>
#include <memory>
#include <unistd.h> // for command line parsing
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
using concepts::Real;

/* <h3>Mesh</h3>

   <div class="solution">
   The domain is <span class="math">(-1,1)<sup>2</sup></span>. It is meshed
   by two triangles.
*/
// ******************************************************************** Mesh **

class SquareTriangle : public concepts::Mesh2 {
public:
  SquareTriangle();
  virtual ~SquareTriangle();
  unsigned int ncell() const { return 2; }
  concepts::Scan2* scan() { return new S(cell_); }
  virtual std::ostream& info(std::ostream& os) const;
private:
  concepts::Vertex *vtx_[4];
  concepts::Edge *edg_[5];
  concepts::Triangle *tri_[2];
  concepts::Cell2 *cell_[2];

  class S : public concepts::Scan<concepts::Cell2> {
    unsigned int idx_;
    concepts::Cell2 *(&cell_)[2];
  public:
    S(concepts::Cell2 *(&cell)[2]) : idx_(0), cell_(cell) {}
    S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    bool eos() const { return idx_ == 2; }
    concepts::Cell2& operator++(int) { return *cell_[idx_++]; }
    concepts::Scan2* clone() const { return new S(*this); }
  };
};

SquareTriangle::SquareTriangle() {
  for (uint i = 0; i < 4; ++i)
    vtx_[i] = new concepts::Vertex();

  edg_[0] = new concepts::Edge(*vtx_[0], *vtx_[1], 1);
  edg_[1] = new concepts::Edge(*vtx_[2], *vtx_[1], 1);
  edg_[2] = new concepts::Edge(*vtx_[0], *vtx_[2]);
  edg_[3] = new concepts::Edge(*vtx_[2], *vtx_[3], 1);
  edg_[4] = new concepts::Edge(*vtx_[0], *vtx_[3], 1);

  tri_[0] = new concepts::Triangle(*edg_[0], *edg_[1], *edg_[2]);
  tri_[1] = new concepts::Triangle(*edg_[2], *edg_[3], *edg_[4]);

  cell_[0] = new concepts::Triangle2d
    (*tri_[0], concepts::VertexTriangle2d(concepts::Real2d(-1.0, -1.0),
					  concepts::Real2d(1.0, -1.0),
					  concepts::Real2d(1.0, 1.0)));
  cell_[1] = new concepts::Triangle2d
    (*tri_[1], concepts::VertexTriangle2d(concepts::Real2d(-1.0, -1.0),
					  concepts::Real2d(1.0, 1.0),
					  concepts::Real2d(-1.0, 1.0)));
}

SquareTriangle::~SquareTriangle() {
  for (int i = 0; i < 2; ++i)
    delete cell_[i];
  for (int i = 0; i < 2; ++i)
    delete tri_[i];
  for (int i = 0; i < 5; ++i)
    delete edg_[i];
  for (int i = 0; i < 4; ++i)
    delete vtx_[i];
}

std::ostream& SquareTriangle::info(std::ostream& os) const {
  os << "SquareTriangle(" << *cell_[0] << ", " << *cell_[1] << ')';
  return os;
}

/* </div>

   <h3>Elements</h3>

   In 2D, two different elements are possible: triangles and quadrilaterals.
   Since some parts of the two elements are equal, they are derived from
   a common parent. As only triangles are used in this tutorial, the
   quadrilaterals are not shown.

   Compared to the 1D case, the elements look almost the same. The most
   significant difference is the new constructor which makes the creation
   of the T matrices somewhat more flexible.
*/
// **************************************************************** Elements **

class FEMElement2d : public concepts::Element<Real> {
public:
  FEMElement2d(const uint i, uint idx[]) :
    T_(new concepts::TIndex<Real>(i,i,idx)) {}
  FEMElement2d(concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1);
  const concepts::TMatrixBase<Real>& T() const { return *T_; }
  virtual concepts::Real2d shapefctD(const uint i) const = 0;
  virtual Real shapefct(const uint i,
			const Real xi1, const Real xi2) const = 0;
protected:
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
};

/* The new constructor takes two linear lists of T matrix columns and
   creates a T matrix out of them.
*/
FEMElement2d::FEMElement2d(concepts::TColumn<Real>* T0,
			   concepts::TColumn<Real>* T1) :
  T_(new concepts::TMatrix<Real>(T1)) {
  if (T0) {
    concepts::TMatrix<Real>* T =
      dynamic_cast<concepts::TMatrix<Real>*>(T_.get());
    conceptsAssert(T, concepts::Assertion());
    T->append(T0);
  }
}

/* <h4>Triangle</h4>
 */
class FEMTriangle : public FEMElement2d {
public:
  FEMTriangle(const concepts::Triangle2d& cell, uint idx[]);
  FEMTriangle(const concepts::Triangle2d& cell, concepts::TColumn<Real>* T0,
	      concepts::TColumn<Real>* T1 = 0);
  const concepts::Triangle& support() const { return cell_.connector(); }
  concepts::Real3d vertex(const uint i) const { return cell_.vertex(i); }
  const concepts::Triangle2d& cell() const { return cell_; }
  virtual concepts::Real2d shapefctD(const uint i) const;
  virtual Real shapefct(const uint i,
				    const Real xi1, const Real xi2) const;
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  const concepts::Triangle2d& cell_;
};

FEMTriangle::FEMTriangle(const concepts::Triangle2d& cell, uint idx[]) :
  FEMElement2d(3, idx), cell_(cell) {}

/* In the constructor, the T matrices are forwarded to the constructor of
   the general element.
*/
FEMTriangle::FEMTriangle(const concepts::Triangle2d& cell,
			 concepts::TColumn<Real>* T0,
			 concepts::TColumn<Real>* T1) :
  FEMElement2d(T0, T1), cell_(cell) {}

/* <div class="solution">
   Gradient of the shape functions on the reference element.
*/
concepts::Real2d FEMTriangle::shapefctD(const uint i) const {
  switch(i) {
  case 0: return concepts::Real2d(-1, -1);
  case 1: return concepts::Real2d(1, 0);
  case 2: return concepts::Real2d(0, 1);
  default:
    throw conceptsException(concepts::MissingFeature("wrong shapefct. index"));
  }
}
/* Shape functions on the reference element.
 */
Real FEMTriangle::shapefct(const uint i,
			   const Real xi1, const Real xi2) const {
  switch(i) {
  case 0: return 1.0-xi1-xi2;
  case 1: return xi1;
  case 2: return xi2;
  default:
    throw conceptsException(concepts::MissingFeature("wrong shapefct. index"));
  }
}
/* </div>
 */

std::ostream& FEMTriangle::info(std::ostream& os) const {
  return os << "FEMTriangle(" << cell_ << ", T = " << *T_ << ')';
}

/* <h3>Scanner for the Elements</h3>
*/
// ***************************************************************** Scanner **

namespace concepts {
  template<>
  class Scan<FEMTriangle> : public Scan<Element<Real> > {
  public:
    FEMTriangle& operator++(int) = 0;
  };
} // namespace concepts

/* <h3>Space</h3>

   The space in 2D looks almost the same as the space in 1D with the slight
   difference that the Dirichlet boundary conditions are enforced while
   creating the space.

   This is done by not assembling the degrees of freedom on a Dirichlet
   boundary. Hence, the system matrix gets smaller: the degrees of freedom
   which are 0 anyway are left out.
*/
// ******************************************************************* Space **

class Linear2d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<concepts::Element<Real> > Scan;
  Linear2d(concepts::Mesh2& msh, uint level,
	   concepts::BoundaryConditions* bc = 0);
  virtual ~Linear2d();
  uint dim() const { return dim_; }
  uint nelm() const { return nelm_; }
  Scan* scan() const {
    return new concepts::PListScan<concepts::Element<Real> >(*elm_);
  }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  concepts::BoundaryConditions* bc_;
  concepts::Joiner<concepts::Element<Real>*, 1>* elm_;
  void enforceBC_(concepts::Cell2* cell, uint level,
		  concepts::DynArray<uint>& bcMap);
  void constructor_(concepts::Cell2* cell, uint level,
		    concepts::DynArray<uint>& dofMap,
		    const concepts::DynArray<uint>& bcMap);
};

/* The constructor first calls <span class="code">enforceBC_</span> on
   each element which notes where the Dirichlet boundary conditions
   have to be enforced and then calls <span
   class="code">constructor_</span> (same as in the 1D code).
*/
Linear2d::Linear2d(concepts::Mesh2& msh, uint level,
		   concepts::BoundaryConditions* bc) :
  dim_(0), nelm_(0), bc_(bc), elm_(0) {

  /* <span class="code">bcMap</span> stores for every visited vertex
     if it has a valid degree of freedom (1) or not (0). <span
     class="code">bcMap</span> is filled in <span
     class="code">enforceBC_</span> depending on the level and
     boundary conditions.
  */
  concepts::DynArray<uint> bcMap(2, 3, 1);
  std::auto_ptr<concepts::Scan2> sc(msh.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    enforceBC_(&cell, level, bcMap);
  }
  /* <span class="code">dofMap</span> stores for every visited vertex the
     assigned degree of freedom or the value <span class="code">(uint)-1</span>
     if no degree of freedom is assigned yet.
  */
  concepts::DynArray<uint> dofMap(2, 3, (uint)-1);
  sc.reset(msh.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    concepts::Triangle2d* cellT = dynamic_cast<concepts::Triangle2d*>(&cell);
    if (cellT) {
      constructor_(cellT, level, dofMap, bcMap);
      continue;
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }
}

Linear2d::~Linear2d() {
  delete elm_;
}

std::ostream& Linear2d::info(std::ostream& os) const {
  return os << "Linear2d(dim = " << dim_ << ", nelm = " << nelm_ << ')';
}

/* <span class="code">enforceBC_</span> calls itself recursively if the
   desired level of refinement is not reached yet. The code in this routine
   is independent of the exact element type.
*/
void Linear2d::enforceBC_(concepts::Cell2* cell, uint level,
			  concepts::DynArray<uint>& bcMap) {
  if (level) {
    --level;
    for (uint i = 0; i < 4; ++i)
      enforceBC_(cell->child(i), level, bcMap);
  } else {
    /* If the desired level of refinement is reached, every edge of the
       element is checked for Dirichlet boundary conditions.

       If an edge has Dirichlet boundary conditions, both endpoints are
       marked: the corresponding degrees of freedom will not be assembled.
       If no Dirichlet boundary conditions are specified on the edge,
       the entries in <span class="code">bcMap</span>
       are touched to make sure they exist.
    */
    concepts::Connector1* edg;
    for (uint j = 0; (edg = cell->connector().edge(j)) != 0; ++j) {
      if ((*bc_)(edg->attrib()).type() == concepts::Boundary::DIRICHLET) {
	bcMap[edg->vertex(0)->key()] = 0;
	bcMap[edg->vertex(1)->key()] = 0;
      } else {
	bcMap[edg->vertex(0)->key()];
	bcMap[edg->vertex(1)->key()];
      }
    }
  }
}

/* As in the 1D case, <span class="code">constructor_</span> calls itself
   recursively if the desired level of refinement is not reached yet.
*/
void Linear2d::constructor_(concepts::Cell2* cell, uint level,
			    concepts::DynArray<uint>& dofMap,
			    const concepts::DynArray<uint>& bcMap) {
  if (level) {
    --level;
    for (uint i = 0; i < 4; ++i)
      constructor_(cell->child(i), level, dofMap, bcMap);
  } else {
    /* If the desired level of refinement is reached, the degrees of
       freedom are counted. A triangle has three local shape functions
       (a quadrilateral would have four).
    */
    concepts::Triangle2d* cellT = dynamic_cast<concepts::Triangle2d*>(cell);
    uint n = 0;
    if (cellT)
      n = 3;
    conceptsAssert(n, concepts::Assertion());
    concepts::TColumn<Real>* T = 0;
    concepts::Connector0* vtx;
    /* In the loop over all vertices of the element, a new column is
       added to the list of columns of the T matrix for every degree
       of freedom. This is done only if the vertex does not have
       Dirichlet boundary conditions (the information is stored in
       <span class="code">bcMap</span>).

       <span class="code">i</span> is the index of the global basis
       function and <span class="code">j</span> is the index of the local
       shape function.
    */
    for (uint j = 0; (vtx = cell->connector().vertex(j)) != 0; ++j) {
      if (bcMap[vtx->key()]) {
	uint& i = dofMap[vtx->key()];
	if (i == (uint)-1)
	  i = dim_++;
	T = new concepts::TColumn<Real>(n, i, T);
	(*T)[j] = 1.0;
      }
    }
    /* After all vertices are checked, the new element is created with the
       assembling information in the T columns.
    */
    concepts::Element<Real>* elm;
    if (cellT) {
      elm = new FEMTriangle(*cellT, T);
    }
    nelm_++;
    elm_ = new concepts::Joiner<concepts::Element<Real>*, 1>(elm, elm_);
  }
}

/* <h3>Linear Form</h3>
 */
// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
  /* <div class="solution">
     The quadrature rule and the right hand side function f.
   */
private:
  const concepts::Quadrature<4> quad_;
  Real f(const Real x1, const Real x2) const {
    return 4.0*(1.0+x1)*(1.0+x2)*(2.0-x1-x2);
  }
public:
  /* The constructor which initializes the quadrature rule.
   */
  RHS(const uint gauss_p = 1) : quad_(gauss_p) {}
  /* </div>
   */
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();

    const FEMTriangle* e = dynamic_cast<const FEMTriangle*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    concepts::Real2d x0(e->vertex(0));
    concepts::Real2d x1(e->vertex(1));
    concepts::Real2d x2(e->vertex(2));
    /* <div class="solution">
       Create the element map <span class="math">x = B<sub>T</sub>
       &xi; + b</span>.
    */
    const concepts::MapReal2d B_T(x1-x0, x2-x0);
    const concepts::Real2d b(x0);
    const Real detB_T = B_T.determinant();
    /* Get the weights and points from the quadrature rule.
     */
    const Real* q = quad_.abscissas();
    const Real* w = quad_.weights();
    /* Loop over all quadrature points.
     */
    for(uint k = 0; k < quad_.n(); k++) { // loop over xi_1 quad. points
      for(uint l = 0; l < quad_.n(); l++) { // loop over xi_2 quad.
	/* Get the point in the reference element and in the physical element.
	 */
	const concepts::Real2d xi((1.0+q[k])*(1.0-q[l])/4.0,
				  (1.0+q[l])/2.0);
	const concepts::Real2d x(B_T * xi + b);
	/* Evaluate the right hand side f.
	 */
	const Real F = f(x[0], x[1]);
	const Real dx = w[k] * w[l] * detB_T * (1.0-q[l])/8.0;
	/* Loop over all shape functions and add their contribution into
	   the local load vector.
	*/
	for (uint i = 0; i < n; ++i) {
	  em(i,0) += F*dx*e->shapefct(i, xi[0], xi[1]);
	} // for i
      } // for l
    } // for k
    /* </div>
     */
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

class Laplace : public concepts::BilinearForm<Real> {
  /* <div class="solution">
     The quadrature rule and the diffusion matrix a, depending on x.
  */
private:
  const concepts::Quadrature<4> quad_;
  concepts::MapReal2d a(const Real x1, const Real x2) const {
    return concepts::MapReal2d(2.0+x1, 0.0, 0.0, 2.0+x2);
  }
public:
  /* The constructor which initializes the quadrature rule.
   */
  Laplace(const uint gauss_p = 1) : quad_(gauss_p) {}
  /* </div>
   */
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const FEMTriangle* eX = dynamic_cast<const FEMTriangle*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const FEMTriangle* eY = dynamic_cast<const FEMTriangle*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      concepts::Real2d x0(eX->vertex(0));
      concepts::Real2d x1(eX->vertex(1));
      concepts::Real2d x2(eX->vertex(2));

      /* <div class="solution">
	 Create the element map <span class="math">x = B<sub>T</sub>
	 &xi; + b</span>.
       */
      const concepts::MapReal2d B_T(x1-x0, x2-x0);
      const concepts::Real2d b(x0);
      const Real detB_T = B_T.determinant();
      /* The inverse of the element map is used to transform the gradients
	 of the shape functions onto the reference element.
      */
      const concepts::MapReal2d B_TinvT(B_T.inverse().transpose());
      /* Get the quadrature points and weights.
       */
      const Real* q = quad_.abscissas();
      const Real* w = quad_.weights();
      /* In this array the integral of (a) * (the Jacobian of the Duffy
	 transformation) (the quad - triangle transformation) is stored.
      */
      Real inta[4] = { 0.0, 0.0, 0.0, 0.0 };
      /* Loop over all quadrature points.
       */
      for(uint k = 0; k < quad_.n(); k++) { // loop over xi_1 quad. points
	for(uint l = 0; l < quad_.n(); l++) { // loop over xi_2 quad.
	  /* Get the point in the reference element and in
	     the physical element.
	  */
	  const concepts::Real2d xi((1.0+q[k])*(1.0-q[l])/4.0,
				    (1.0+q[l])/2.0);
	  const concepts::Real2d x(B_T * xi + b);
	  /* Evaluate the diffusion matrix and the Jacobian of the
	     transformation D which maps the reference element onto
	     <span class="math">(-1,1)<sup>2</sup></span>.
	   */
	  const concepts::MapReal2d A(a(x[0], x[1]));
	  const Real dDdet = (1.0-q[l])/4.0*0.5;
	  const Real dx = fabs(w[k] * w[l] * detB_T * dDdet);
	  inta[0] += dx * A.column(0)[0];
	  inta[1] += dx * A.column(1)[0];
	  inta[2] += dx * A.column(0)[1];
	  inta[3] += dx * A.column(1)[1];
	} // for l
      } // for k
      /* Store the previously computed integral of a in a suitable form
	 to make it easy to use.
      */
      const concepts::MapReal2d A(inta[0], inta[1], inta[2], inta[3]);
      /* Loop over the shape functions and add their contributions into
	 the local stiffness matrix.
      */
      for (uint i = 0; i < n; ++i) { // loop over u shape fcts.
	const concepts::Real2d u(B_TinvT * eX->shapefctD(i));
	for (uint j = 0; j < n; ++j) { // loop over v shape fcts.
	  const concepts::Real2d v(B_TinvT * eX->shapefctD(j));
	  em(i,j) += (A * u) * v;
	} // for j
      } // for i
      /* </div>
       */
    }
  }
};

/* <h3>Main program</h3>
*/
// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    /* <div class="solution">
    */
    uint level = 1;
    uint gauss_p = 1;

    concepts::InputParser inputParser;
    concepts::InOutParameters& input = inputParser.inputParameters();

    input.addInt("level", level);
    input.addInt("gauss_p", gauss_p);
    /* Parse the command line options.
     */
    int opt;
    while ((opt = getopt(argc, argv, "-l:p:")) != EOF)
      switch(opt) {
      case 'l': input.addInt("level", std::atoi(optarg)); break;
      case 'p': input.addInt("gauss_p", std::atoi(optarg)); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-l LEVEL] [-p GAUSS_P]" << std::endl
		  << "where" << std::endl
		  << "  LEVEL: maximal level of refinement" << std::endl
		  << "  GAUSS_P: number of points in Gauss integration"
		  << std::endl;
	std::exit(1); break;
      }
    level = input.getInt("level");
    gauss_p = input.getInt("gauss_p");

    concepts::InOutParameters& output = inputParser.outputParameters();
    const Real exEnergie = 512.0/45.0;
    output.addDouble("exEnergie", exEnergie);
    output.addArrayDouble("h");
    output.addArrayDouble("energieFehler");
    output.addArrayDouble("energie");
    /* </div>
     */
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET, "(0)"));
    std::cout << "Boundary conditions: " << *bc << std::endl;

    SquareTriangle msh;
    std::cout << "Mesh: " << msh << std::endl;
    /* <div class="solution">
       Start the loop over the different levels.
    */       
    for (uint l = 1; l < level; ++l) {
      std::cout << "l = " << l << std::endl;
      /* </div>
       */
      Linear2d spc(msh, l, bc.get());
      std::cout << "Space: " << spc << std::endl;

      Laplace stiff_bf(gauss_p);
      concepts::PETScMat stiff(spc, stiff_bf);

      RHS rhs_lf(gauss_p);
      concepts::Vector<Real> rhs(spc, rhs_lf);

      concepts::PETSc solver(stiff, 1e-6, "bcgs");
      concepts::Vector<Real> sol(spc);
      try {
	solver(rhs, sol);
      }
      catch(concepts::NoConvergence& e) {
	std::cout << "solver did not converge!" << std::endl << e << std::endl;
      }
      std::cout << "solver = " << solver << std::endl;
      /* <div class="solution">
	 Gather the results of the computation.
      */
      const Real h = 1.0/(1<<l);
      output.addArrayDouble("h", l, h);
      const Real feEnergie(sol*rhs);
      output.addArrayDouble("energie", l, feEnergie);
      output.addArrayDouble("energieFehler", l,
			    fabs(exEnergie-feEnergie)/exEnergie);
      /* End of the loop over the different levels.
       */
    } // for l
    /* Print the results.
     */
    std::cout << inputParser << std::endl;
    /* </div>
     */
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
