/* The include files are the same as in the previous tutorial.
 */
#include <iostream>
#include <memory>
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"

using concepts::Real;

/* <h3>Mesh</h3>

   The mesh is taken from the <a href="meshes.html#LshapedTriangle">meshes
   tutorial</a>:
   <span class="code">LshapedTriangle</span>. Here, the relevant code is
   included:
*/
#include "meshes.cc"

/* <h3>Elements</h3>

   In 2D, two different elements are possible: triangles and quadrilaterals.
   Since some parts of the two elements are equal, they are derived from
   a common parent. As only triangles are used in this tutorial, the
   quadrilaterals are not shown.

   Compared to the 1D case, the elements look almost the same. The
   most significant difference is the new constructor which makes the
   creation of the T matrices somewhat more flexible. There exist two
   constructors: one for creating the T matrix out of a <span
   class="code">idx</span> array as seen in the <a
   href="fem1D.html">1D tutorial</a> and one for creating the T
   matrices out of linked lists of <a
   href="../#classDocumentation#classconcepts_1_1TColumn.html">T
   matrix columns</a> <span class="code">T0</span> and <span
   class="code">T1</span>.
*/
// **************************************************************** Elements **

class FEMElement2d : public concepts::Element<Real> {
public:
  FEMElement2d(const uint i, uint idx[]) :
    T_(new concepts::TIndex<Real>(i,i,idx)) {}
  FEMElement2d(concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1);
  const concepts::TMatrixBase<Real>& T() const { return *T_; }
protected:
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
};

/* The new constructor takes two linear lists of <a
   href="../#classDocumentation#classconcepts_1_1TColumn.html">T
   matrix columns</a> and creates a <a
   href="../#classDocumentation#classconcepts_1_1TMatrixBase.html">T
   matrix</a> out of them.
*/
FEMElement2d::FEMElement2d(concepts::TColumn<Real>* T0,
			   concepts::TColumn<Real>* T1) :
  T_(new concepts::TMatrix<Real>(T1)) {
  if (T0) {
    /* The following <span class="code">dynamic_cast</span> business
       is necessary since only the <a
       href="../#classDocumentation#classconcepts_1_1TMatrix.html"
       class="code">TMatrix</a> has the needed member function <span
       class="code">append()</span> and <a
       href="../#classDocumentation#classconcepts_1_1TIndex.html"
       class="code">TIndex</a> does not.
    */
    concepts::TMatrix<Real>* T =
      dynamic_cast<concepts::TMatrix<Real>*>(T_.get());
    conceptsAssert(T, concepts::Assertion());
    T->append(T0);
  }
}

/* <h4>Quadrilateral</h4>

   We do not use quadrilaterals in this code.

   <h4>Triangle</h4>
*/
class FEMTriangle : public FEMElement2d {
public:
  FEMTriangle(const concepts::Triangle2d& cell, uint idx[]);
  FEMTriangle(const concepts::Triangle2d& cell, concepts::TColumn<Real>* T0,
	      concepts::TColumn<Real>* T1 = 0);
  const concepts::Triangle& support() const { return cell_.connector(); }
  concepts::Real3d vertex(const uint i) const { return cell_.vertex(i); }
  const concepts::Triangle2d& cell() const { return cell_; }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  const concepts::Triangle2d& cell_;
};

FEMTriangle::FEMTriangle(const concepts::Triangle2d& cell, uint idx[]) :
  FEMElement2d(3, idx), cell_(cell) {}

/* In the constructor, the T matrix columns are forwarded to the
   constructor of the general element.
*/
FEMTriangle::FEMTriangle(const concepts::Triangle2d& cell,
			 concepts::TColumn<Real>* T0,
			 concepts::TColumn<Real>* T1) :
  FEMElement2d(T0, T1), cell_(cell) {}

std::ostream& FEMTriangle::info(std::ostream& os) const {
  return os << "FEMTriangle(" << cell_ << ", T = " << *T_ << ')';
}

/* <h3>Scanner for the Elements</h3>
*/
// ***************************************************************** Scanner **

namespace concepts {
  template<>
  class Scan<FEMTriangle> : public Scan<Element<Real> > {
  public:
    FEMTriangle& operator++(int) = 0;
  };
} // namespace concepts

/* <h3>Space</h3>

   The space in 2D looks almost the same as the space in 1D with the slight
   difference that the Dirichlet boundary conditions are enforced while
   creating the space.

   This is done by not assembling the degrees of freedom on a Dirichlet
   boundary. Hence, the system matrix gets smaller: the degrees of freedom
   which are 0 anyway are left out.
*/
// ******************************************************************* Space **

class Linear2d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<concepts::Element<Real> > Scan;
  Linear2d(concepts::Mesh2& msh, uint level,
	   concepts::BoundaryConditions* bc = 0);
  virtual ~Linear2d();
  uint dim() const { return dim_; }
  uint nelm() const { return nelm_; }
  Scan* scan() const {
    return new concepts::PListScan<concepts::Element<Real> >(*elm_);
  }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  concepts::BoundaryConditions* bc_;
  concepts::Joiner<concepts::Element<Real>*, 1>* elm_;
  void enforceBC_(concepts::Cell2* cell, uint level,
		  concepts::DynArray<uint>& bcMap);
  void constructor_(concepts::Cell2* cell, uint level,
		    concepts::DynArray<uint>& dofMap,
		    const concepts::DynArray<uint>& bcMap);
};

/* The constructor first calls <span class="code">enforceBC_</span> on
   each element which notes where the Dirichlet boundary conditions
   have to be enforced and then calls <span
   class="code">constructor_</span> (same idea as in the <a
   href="fem1D.html">1D code</a>).
*/
Linear2d::Linear2d(concepts::Mesh2& msh, uint level,
		   concepts::BoundaryConditions* bc) :
  dim_(0), nelm_(0), bc_(bc), elm_(0) {
  /* <span class="code">bcMap</span> stores for every visited vertex
     if it has a valid degree of freedom (1) or not (0). <span
     class="code">bcMap</span> is filled in <span
     class="code">enforceBC_</span> depending on the level and
     boundary conditions. The default value for newly created entries
     in the <a
     href="../#classDocumentation#classconcepts_1_1DynArray.html">dynamic
     array</a> is 1.
  */
  concepts::DynArray<uint> bcMap(2, 3, 1);
  std::auto_ptr<concepts::Scan2> sc(msh.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    enforceBC_(&cell, level, bcMap);
  }
  /* <span class="code">dofMap</span> stores for every visited vertex the
     assigned degree of freedom or the value <span class="code">(uint)-1</span>
     if no degree of freedom is assigned yet.
  */
  concepts::DynArray<uint> dofMap(2, 3, (uint)-1);
  sc.reset(msh.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    concepts::Triangle2d* cellT = dynamic_cast<concepts::Triangle2d*>(&cell);
    if (cellT) {
      constructor_(cellT, level, dofMap, bcMap);
      continue;
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }
}

Linear2d::~Linear2d() {
  delete elm_;
}

std::ostream& Linear2d::info(std::ostream& os) const {
  return os << "Linear2d(dim = " << dim_ << ", nelm = " << nelm_ << ')';
}

/* <span class="code">enforceBC_</span> calls itself recursively if the
   desired level of refinement is not reached yet. The code in this routine
   is independent of the exact element type.
*/
void Linear2d::enforceBC_(concepts::Cell2* cell, uint level,
			  concepts::DynArray<uint>& bcMap) {
  if (level) {
    --level;
    for (uint i = 0; i < 4; ++i)
      enforceBC_(cell->child(i), level, bcMap);
  } else {
    /* If the desired level of refinement is reached, every edge of the
       element is checked for Dirichlet boundary conditions.

       If an edge has Dirichlet boundary conditions, both endpoints are
       marked: the corresponding degrees of freedom will not be assembled.
       If no Dirichlet boundary conditions are specified on the edge,
       the entries in <span class="code">bcMap</span>
       are touched to make sure they exist.
    */
    concepts::Connector1* edg;
    for (uint j = 0; (edg = cell->connector().edge(j)) != 0; ++j) {
      if ((*bc_)(edg->attrib()).type() == concepts::Boundary::DIRICHLET) {
	bcMap[edg->vertex(0)->key()] = 0;
	bcMap[edg->vertex(1)->key()] = 0;
      } else {
	bcMap[edg->vertex(0)->key()];
	bcMap[edg->vertex(1)->key()];
      }
    }
  }
}

/* As in the 1D case, <span class="code">constructor_</span> calls itself
   recursively if the desired level of refinement is not reached yet.
*/
void Linear2d::constructor_(concepts::Cell2* cell, uint level,
			    concepts::DynArray<uint>& dofMap,
			    const concepts::DynArray<uint>& bcMap) {
  if (level) {
    --level;
    for (uint i = 0; i < 4; ++i)
      constructor_(cell->child(i), level, dofMap, bcMap);
  } else {
    /* If the desired level of refinement is reached, the degrees of
       freedom are counted. A triangle has three local shape functions
       (a quadrilateral would have four).
    */
    concepts::Triangle2d* cellT = dynamic_cast<concepts::Triangle2d*>(cell);
    uint n = 0;
    if (cellT)
      n = 3;
    conceptsAssert(n, concepts::Assertion());
    concepts::TColumn<Real>* T = 0;
    concepts::Connector0* vtx;
    /* In the loop over all vertices of the element, a new column is
       added to the list of columns of the T matrix for every degree
       of freedom. This is done only if the vertex does not have
       Dirichlet boundary conditions (the information is stored in
       <span class="code">bcMap</span>).

       <span class="code">i</span> is the index of the global basis
       function and <span class="code">j</span> is the index of the local
       shape function.
    */
    for (uint j = 0; (vtx = cell->connector().vertex(j)) != 0; ++j) {
      if (bcMap[vtx->key()]) {
	uint& i = dofMap[vtx->key()];
	if (i == (uint)-1)
	  i = dim_++;
	T = new concepts::TColumn<Real>(n, i, T);
	(*T)[j] = 1.0;
      }
    }
    /* After all vertices are checked, the new element is created with the
       assembling information in the T columns.
    */
    concepts::Element<Real>* elm;
    if (cellT) {
      elm = new FEMTriangle(*cellT, T);
    }
    nelm_++;
    elm_ = new concepts::Joiner<concepts::Element<Real>*, 1>(elm, elm_);
  }
}

/* <h3>Linear Form</h3>

   <span class="math">f = 1</span>.
 */
// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();
    const FEMTriangle* e = dynamic_cast<const FEMTriangle*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    concepts::Real2d x0(e->vertex(0));
    concepts::Real2d x1(e->vertex(1));
    concepts::Real2d x2(e->vertex(2));
    const concepts::MapReal3d volMat(1.0, 1.0, 1.0,
				     x0[0], x1[0], x2[0],
				     x0[1], x1[1], x2[1]);
    const Real volTriangle_6 = volMat.determinant() / 6.0;
    for (uint i = 0; i < n; ++i)
      em(i,0) = volTriangle_6;
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const FEMTriangle* eX = dynamic_cast<const FEMTriangle*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const FEMTriangle* eY = dynamic_cast<const FEMTriangle*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      concepts::Real2d x0(eX->vertex(0));
      concepts::Real2d x1(eX->vertex(1));
      concepts::Real2d x2(eX->vertex(2));
      /* First compute the volume of the triangle.
       */
      const concepts::MapReal3d volMat(1.0, 1.0, 1.0,
				       x0[0], x1[0], x2[0],
				       x0[1], x1[1], x2[1]);
      const Real volTriangle_4 = 1.0 / (2.0 * volMat.determinant());
      /* The stiffness matrix is easily computable from the coordinates of
	 the corners of the triangle.
      */
      concepts::Real2d p1_p2(x1 - x2);
      concepts::Real2d p2_p0(x2 - x0);
      concepts::Real2d p0_p1(x0 - x1);

      em(0,0) = p1_p2.l2_2() * volTriangle_4;
      em(0,1) = p1_p2*p2_p0 * volTriangle_4;
      em(0,2) = p1_p2*p0_p1 * volTriangle_4;
      em(1,1) = p2_p0.l2_2() * volTriangle_4;
      em(1,2) = p2_p0*p0_p1 * volTriangle_4;
      em(2,2) = p0_p1.l2_2() * volTriangle_4;

      for (uint i = 0; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
    }
  }
};

/* <h3>Main program</h3>
*/
// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    /* The edges on the boundary of the domain have unique attributes
       (ie. numbers) given in the constructor of the mesh (see near
       the <a href="#CommProg">top of the code</a>). Here, the
       corresponding boundary conditions are created and assigned to
       the same attributes.
    */
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(3),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(4),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(6),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    std::cout << "Boundary conditions: " << *bc << std::endl;

    LshapedTriangle msh;
    std::cout << "Mesh: " << msh << std::endl;
    /* The mesh was set up as in the 1D case. The space on the other
       hand, needs to know about the boundary conditions since they
       might influence the way the degrees of freedom are counted.
       The rest is the same as in the 1D case again.
    */
    Linear2d spc(msh, 1, bc.get());
    std::cout << "Space: " << spc << std::endl;

    Laplace stiff_bf;
    concepts::DenseMatrix<Real> stiff(spc, stiff_bf);
    std::cout << "stiff = " << stiff << std::endl;

    RHS rhs_lf;
    concepts::Vector<Real> rhs(spc, rhs_lf);
    std::cout << "rhs = " << rhs << std::endl;

    concepts::CG<Real> solver(stiff, 1e-6, 200);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
    std::cout << "solution = " << sol << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
