/* The include files are the same as in the previous tutorial.
 */
#include <iostream>
#include <memory>
#include <vector>
#include <sstream>
#include <cstdlib>
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
using concepts::Real;

/* <h3>Element</h3>
*/
// **************************************************************** Elements **

class LinearTet3d : public concepts::Element<Real> {
public:
  LinearTet3d(const concepts::Tetrahedron3d& cell, concepts::TColumn<Real>* T0,
	      concepts::TColumn<Real>* T1 = 0);
  const concepts::TMatrixBase<Real>& T() const { return *T_; }
  const concepts::Tetrahedron& support() const { return cell_.connector(); }
  concepts::Real3d vertex(const uint i) const { return cell_.vertex(i); }
  const concepts::Tetrahedron3d& cell() const { return cell_; }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
  const concepts::Tetrahedron3d& cell_;
};

LinearTet3d::LinearTet3d(const concepts::Tetrahedron3d& cell,
			 concepts::TColumn<Real>* T0,
			 concepts::TColumn<Real>* T1) :
  T_(new concepts::TMatrix<Real>(T1)), cell_(cell) {
  if (T0) {
    concepts::TMatrix<Real>* T =
      dynamic_cast<concepts::TMatrix<Real>*>(T_.get());
    conceptsAssert(T, concepts::Assertion());
    T->append(T0);
  }
}

std::ostream& LinearTet3d::info(std::ostream& os) const {
  return os << "FEMTriangle(" << cell_ << ", T = " << *T_ << ')';
}

/* <h3>Scanner for the Elements</h3>
*/
// ***************************************************************** Scanner **

namespace concepts {
  template<>
  class Scan<LinearTet3d> : public Scan<Element<Real> > {
  public:
    LinearTet3d& operator++(int) = 0;
  };
} // namespace concepts

/* <h3>Space</h3>
*/
// ******************************************************************* Space **

class Linear3d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<concepts::Element<Real> > Scan;
  Linear3d(concepts::Mesh3& msh, concepts::BoundaryConditions* bc = 0);
  uint dim() const { return dim_; }
  uint nelm() const { return nelm_; }
  Scan* scan() const {
    return new concepts::PListScan<concepts::Element<Real> >(*elm_);
  }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  concepts::BoundaryConditions* bc_;
  concepts::Joiner<concepts::Element<Real>*, 1>* elm_;
};

Linear3d::Linear3d(concepts::Mesh3& msh, concepts::BoundaryConditions* bc) :
  dim_(0), nelm_(0), bc_(bc), elm_(0) {
  /* <span class="code">bcMap</span> stores for every visited vertex
     if it has a valid degree of freedom (1) or not (0). <span
     class="code">bcMap</span> is filled in <span
     class="code">enforceBC_</span> depending on the level and
     boundary conditions.
  */
  concepts::DynArray<uint> bcMap(2, 3, 1);
  std::auto_ptr<concepts::Scan3> sc(0);
  if (bc_) {
    sc.reset(msh.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      concepts::Connector2* face;
      for (uint j = 0; (face = cell.connector().face(j)) != 0; ++j) {
	concepts::Connector0* vtx;
	for (uint k = 0; (vtx = face->vertex(k)) != 0; ++k) {
	  if ((*bc_)(face->attrib()).type() == concepts::Boundary::DIRICHLET)
	    bcMap[vtx->key()] = 0;
	  else
	    bcMap[vtx->key()];
	}
      }
    }
  }

  /* <span class="code">dofMap</span> stores for every visited vertex the
     assigned degree of freedom or the value <span class="code">(uint)-1</span>
     if no degree of freedom is assigned yet.
  */
  concepts::DynArray<uint> dofMap(2, 3, (uint)-1);
  sc.reset(msh.scan());
  while (*sc) {
    concepts::Cell3& cell = (*sc)++;
    const concepts::Tetrahedron3d* cellT =
      dynamic_cast<const concepts::Tetrahedron3d*>(&cell);
    conceptsAssert(cellT, concepts::Assertion());
    concepts::TColumn<Real>* T = 0;
    const uint n = 4;
    concepts::Connector0* vtx;
    for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j) {
      if (bcMap[vtx->key()]) {
	uint& i = dofMap[vtx->key()];
	if (i == (uint)-1)
	  i = dim_++;
	T = new concepts::TColumn<Real>(n, i, T);
	(*T)[j] = 1.0;
      }
    }
    nelm_++;
    elm_ = new concepts::Joiner<concepts::Element<Real>*, 1>
      (new LinearTet3d(*cellT, T), elm_);
  }
}

std::ostream& Linear3d::info(std::ostream& os) const {
  return os << "Linear3d(dim = " << dim_ << ", nelm = " << nelm_ << ')';
}

/* <h3>Linear Form</h3>
 */
// ************************************************************* Linear Form **

// right hand side
Real f(const concepts::Real3d x) {
  return 2.0*x[0]*x[1]*(1.0-x[0])*(1.0-x[1])
    + 2.0*x[0]*x[2]*(1.0-x[0])*(1.0-x[2])
    + 2.0*x[1]*x[2]*(1.0-x[1])*(1.0-x[2]);
}

class RHS : public concepts::LinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();

    const LinearTet3d* e = dynamic_cast<const LinearTet3d*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    /* <div class="solution">
       Get the vertices of the physical element.
    */
    const concepts::Real3d x0(e->vertex(0));
    const concepts::Real3d x1(e->vertex(1));
    const concepts::Real3d x2(e->vertex(2));
    const concepts::Real3d x3(e->vertex(3));
    /* Compute the center of the tetrahedron. In this point, <span
       class="math">f</span> is evaluated to integrate the right hand
       side with a one point formula.
     */
    concepts::Real3d center(x0 + x1 + x2 + x3);
    center *= 0.25;
    /* The Jacobian of element map. It is used to determine the weight
       of the one point integration rule. The weight is <span
       class="math">g/6</span>.
     */
    const Real g = x0[2]*x1[1]*x2[0] - x0[1]*x1[2]*x2[0] - x0[2]*x1[0]*x2[1] +
      x0[0]*x1[2]*x2[1] + x0[1]*x1[0]*x2[2] - x0[0]*x1[1]*x2[2] -
      x0[2]*x1[1]*x3[0] + x0[1]*x1[2]*x3[0] + x0[2]*x2[1]*x3[0] -
      x1[2]*x2[1]*x3[0] - x0[1]*x2[2]*x3[0] + x1[1]*x2[2]*x3[0] +
      x0[2]*x1[0]*x3[1] - x0[0]*x1[2]*x3[1] - x0[2]*x2[0]*x3[1] +
      x1[2]*x2[0]*x3[1] + x0[0]*x2[2]*x3[1] - x1[0]*x2[2]*x3[1] -
      x0[1]*x1[0]*x3[2] + x0[0]*x1[1]*x3[2] + x0[1]*x2[0]*x3[2] -
      x1[1]*x2[0]*x3[2] - x0[0]*x2[1]*x3[2] + x1[0]*x2[1]*x3[2];
    /* Integrate and set all entries of the local load vector to the
       same value. This happends since all shape functions have the
       same value (1/4) in the center of the reference element.
     */
    const Real E = f(center) / 24.0 * g;
    for (uint i = 0; i < n; ++i)
      em(i,0) = E;
    /* </div>
     */
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const LinearTet3d* eX = dynamic_cast<const LinearTet3d*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const LinearTet3d* eY = dynamic_cast<const LinearTet3d*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      /* <div class="solution">
	 Get the vertices of the physical element.
      */
      concepts::Real3d x0(eX->vertex(0));
      concepts::Real3d x1(eX->vertex(1));
      concepts::Real3d x2(eX->vertex(2));
      concepts::Real3d x3(eX->vertex(3));
      /* Set <span class="math">em = H * H<sup>T</sup></span> where
	 <span class="math">H = G * g</span>. As in the linear form,
	 g is the Jacobian of the element map.
       */
      em(0, 0) = pow(x1[1]*x2[0] - x1[0]*x2[1] - x1[1]*x3[0] + x2[1]*x3[0] + x1[0]*x3[1] - x2[0]*x3[1],2) + pow(-(x1[2]*x2[0]) + x1[0]*x2[2] + x1[2]*x3[0] - x2[2]*x3[0] - x1[0]*x3[2] + x2[0]*x3[2],2) + pow(x1[2]*x2[1] - x1[1]*x2[2] - x1[2]*x3[1] + x2[2]*x3[1] + x1[1]*x3[2] - x2[1]*x3[2],2);
      em(0, 1) = (x1[1]*x2[0] - x1[0]*x2[1] - x1[1]*x3[0] + x2[1]*x3[0] + x1[0]*x3[1] - x2[0]*x3[1])*(-(x0[1]*x2[0]) + x0[0]*x2[1] + x0[1]*x3[0] - x2[1]*x3[0] - x0[0]*x3[1] + x2[0]*x3[1]) + (x0[2]*x2[0] - x0[0]*x2[2] - x0[2]*x3[0] + x2[2]*x3[0] + x0[0]*x3[2] - x2[0]*x3[2])*(-(x1[2]*x2[0]) + x1[0]*x2[2] + x1[2]*x3[0] - x2[2]*x3[0] - x1[0]*x3[2] + x2[0]*x3[2]) + (x1[2]*x2[1] - x1[1]*x2[2] - x1[2]*x3[1] + x2[2]*x3[1] + x1[1]*x3[2] - x2[1]*x3[2])*(-(x0[2]*x2[1]) + x0[1]*x2[2] + x0[2]*x3[1] - x2[2]*x3[1] - x0[1]*x3[2] + x2[1]*x3[2]);
      em(0, 2) = (x0[1]*x1[0] - x0[0]*x1[1] - x0[1]*x3[0] + x1[1]*x3[0] + x0[0]*x3[1] - x1[0]*x3[1])*(x1[1]*x2[0] - x1[0]*x2[1] - x1[1]*x3[0] + x2[1]*x3[0] + x1[0]*x3[1] - x2[0]*x3[1]) + (-(x0[2]*x1[0]) + x0[0]*x1[2] + x0[2]*x3[0] - x1[2]*x3[0] - x0[0]*x3[2] + x1[0]*x3[2])*(-(x1[2]*x2[0]) + x1[0]*x2[2] + x1[2]*x3[0] - x2[2]*x3[0] - x1[0]*x3[2] + x2[0]*x3[2]) + (x0[2]*x1[1] - x0[1]*x1[2] - x0[2]*x3[1] + x1[2]*x3[1] + x0[1]*x3[2] - x1[1]*x3[2])*(x1[2]*x2[1] - x1[1]*x2[2] - x1[2]*x3[1] + x2[2]*x3[1] + x1[1]*x3[2] - x2[1]*x3[2]);
      em(0, 3) = (-(x0[1]*x1[0]) + x0[0]*x1[1] + x0[1]*x2[0] - x1[1]*x2[0] - x0[0]*x2[1] + x1[0]*x2[1])*(x1[1]*x2[0] - x1[0]*x2[1] - x1[1]*x3[0] + x2[1]*x3[0] + x1[0]*x3[1] - x2[0]*x3[1]) + (x0[2]*x1[0] - x0[0]*x1[2] - x0[2]*x2[0] + x1[2]*x2[0] + x0[0]*x2[2] - x1[0]*x2[2])*(-(x1[2]*x2[0]) + x1[0]*x2[2] + x1[2]*x3[0] - x2[2]*x3[0] - x1[0]*x3[2] + x2[0]*x3[2]) + (-(x0[2]*x1[1]) + x0[1]*x1[2] + x0[2]*x2[1] - x1[2]*x2[1] - x0[1]*x2[2] + x1[1]*x2[2])*(x1[2]*x2[1] - x1[1]*x2[2] - x1[2]*x3[1] + x2[2]*x3[1] + x1[1]*x3[2] - x2[1]*x3[2]);
      em(1, 1) = pow(-(x0[1]*x2[0]) + x0[0]*x2[1] + x0[1]*x3[0] - x2[1]*x3[0] - x0[0]*x3[1] + x2[0]*x3[1],2) + pow(x0[2]*x2[0] - x0[0]*x2[2] - x0[2]*x3[0] + x2[2]*x3[0] + x0[0]*x3[2] - x2[0]*x3[2],2) + pow(-(x0[2]*x2[1]) + x0[1]*x2[2] + x0[2]*x3[1] - x2[2]*x3[1] - x0[1]*x3[2] + x2[1]*x3[2],2);
      em(1, 2) = (x0[1]*x1[0] - x0[0]*x1[1] - x0[1]*x3[0] + x1[1]*x3[0] + x0[0]*x3[1] - x1[0]*x3[1])*(-(x0[1]*x2[0]) + x0[0]*x2[1] + x0[1]*x3[0] - x2[1]*x3[0] - x0[0]*x3[1] + x2[0]*x3[1]) + (-(x0[2]*x1[0]) + x0[0]*x1[2] + x0[2]*x3[0] - x1[2]*x3[0] - x0[0]*x3[2] + x1[0]*x3[2])*(x0[2]*x2[0] - x0[0]*x2[2] - x0[2]*x3[0] + x2[2]*x3[0] + x0[0]*x3[2] - x2[0]*x3[2]) + (x0[2]*x1[1] - x0[1]*x1[2] - x0[2]*x3[1] + x1[2]*x3[1] + x0[1]*x3[2] - x1[1]*x3[2])*(-(x0[2]*x2[1]) + x0[1]*x2[2] + x0[2]*x3[1] - x2[2]*x3[1] - x0[1]*x3[2] + x2[1]*x3[2]);
      em(1, 3) = (-(x0[1]*x1[0]) + x0[0]*x1[1] + x0[1]*x2[0] - x1[1]*x2[0] - x0[0]*x2[1] + x1[0]*x2[1])*(-(x0[1]*x2[0]) + x0[0]*x2[1] + x0[1]*x3[0] - x2[1]*x3[0] - x0[0]*x3[1] + x2[0]*x3[1]) + (x0[2]*x1[0] - x0[0]*x1[2] - x0[2]*x2[0] + x1[2]*x2[0] + x0[0]*x2[2] - x1[0]*x2[2])*(x0[2]*x2[0] - x0[0]*x2[2] - x0[2]*x3[0] + x2[2]*x3[0] + x0[0]*x3[2] - x2[0]*x3[2]) + (-(x0[2]*x1[1]) + x0[1]*x1[2] + x0[2]*x2[1] - x1[2]*x2[1] - x0[1]*x2[2] + x1[1]*x2[2])*(-(x0[2]*x2[1]) + x0[1]*x2[2] + x0[2]*x3[1] - x2[2]*x3[1] - x0[1]*x3[2] + x2[1]*x3[2]);
      em(2, 2) = pow(x0[1]*x1[0] - x0[0]*x1[1] - x0[1]*x3[0] + x1[1]*x3[0] + x0[0]*x3[1] - x1[0]*x3[1],2) + pow(-(x0[2]*x1[0]) + x0[0]*x1[2] + x0[2]*x3[0] - x1[2]*x3[0] - x0[0]*x3[2] + x1[0]*x3[2],2) + pow(x0[2]*x1[1] - x0[1]*x1[2] - x0[2]*x3[1] + x1[2]*x3[1] + x0[1]*x3[2] - x1[1]*x3[2],2);
      em(2, 3) = (-(x0[1]*x1[0]) + x0[0]*x1[1] + x0[1]*x2[0] - x1[1]*x2[0] - x0[0]*x2[1] + x1[0]*x2[1])*(x0[1]*x1[0] - x0[0]*x1[1] - x0[1]*x3[0] + x1[1]*x3[0] + x0[0]*x3[1] - x1[0]*x3[1]) + (x0[2]*x1[0] - x0[0]*x1[2] - x0[2]*x2[0] + x1[2]*x2[0] + x0[0]*x2[2] - x1[0]*x2[2])*(-(x0[2]*x1[0]) + x0[0]*x1[2] + x0[2]*x3[0] - x1[2]*x3[0] - x0[0]*x3[2] + x1[0]*x3[2]) + (-(x0[2]*x1[1]) + x0[1]*x1[2] + x0[2]*x2[1] - x1[2]*x2[1] - x0[1]*x2[2] + x1[1]*x2[2])*(x0[2]*x1[1] - x0[1]*x1[2] - x0[2]*x3[1] + x1[2]*x3[1] + x0[1]*x3[2] - x1[1]*x3[2]);
      em(3, 3) = pow(-(x0[1]*x1[0]) + x0[0]*x1[1] + x0[1]*x2[0] - x1[1]*x2[0] - x0[0]*x2[1] + x1[0]*x2[1],2) + pow(x0[2]*x1[0] - x0[0]*x1[2] - x0[2]*x2[0] + x1[2]*x2[0] + x0[0]*x2[2] - x1[0]*x2[2],2) + pow(-(x0[2]*x1[1]) + x0[1]*x1[2] + x0[2]*x2[1] - x1[2]*x2[1] - x0[1]*x2[2] + x1[1]*x2[2],2);
      /* Jacobian of element map.
       */
      const Real g = x0[2]*x1[1]*x2[0] - x0[1]*x1[2]*x2[0] - x0[2]*x1[0]*x2[1] +
	x0[0]*x1[2]*x2[1] + x0[1]*x1[0]*x2[2] - x0[0]*x1[1]*x2[2] -
	x0[2]*x1[1]*x3[0] + x0[1]*x1[2]*x3[0] + x0[2]*x2[1]*x3[0] -
	x1[2]*x2[1]*x3[0] - x0[1]*x2[2]*x3[0] + x1[1]*x2[2]*x3[0] +
	x0[2]*x1[0]*x3[1] - x0[0]*x1[2]*x3[1] - x0[2]*x2[0]*x3[1] +
	x1[2]*x2[0]*x3[1] + x0[0]*x2[2]*x3[1] - x1[0]*x2[2]*x3[1] -
	x0[1]*x1[0]*x3[2] + x0[0]*x1[1]*x3[2] + x0[1]*x2[0]*x3[2] -
	x1[1]*x2[0]*x3[2] - x0[0]*x2[1]*x3[2] + x1[0]*x2[1]*x3[2];
      const Real c = 1.0/g/6.0;
      /* Exploit symmetry and multiply every entry of the stiffness matrix
	 with c.
      */
      for (uint i = 0; i < n; ++i)
	for (uint j = i; j < n; ++j)
	  em(j,i) = em(i,j) *= c;
      /* </div>
       */
    }
  }
};

/* <h3>Main program</h3>
*/
// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    /* <div class="solution">
       Set up the structures to capture the results. The exact energy is
       1/900.
     */
    concepts::InputParser inputParser;
    concepts::InOutParameters& output = inputParser.outputParameters();
    const Real exEnergie = 1.0/900.0;
    output.addDouble("exEnergie", exEnergie);
    output.addArrayDouble("energieFehler");
    /* Start the loop over the four meshes given in the files <span
       class="filename">serie4-1.mesh</span> and so on.
     */
    for (uint i = 1; i <= 6; ++i) {
      /* The filename of the mesh file together with the command to
	 create the input files for the class <span
	 class="class">Import3DTetMesh</span> is created in <span
	 class="code">meshName</span>. This command will be executed
	 later.
       */
      std::stringstream meshName;
      meshName << "meshgen.pl serie4-" << i << ".mesh" << std::ends;
      std::cout << "Converting mesh: " << meshName.str() << std::endl;
      /* Here, the command is executed.
       */
      std::system(meshName.str().c_str());
      std::cout << "Mesh conversion done, importing." << std::endl;
      /* The input files were generated, import them.
	 </div>
       */
      concepts::Import3DTetMesh msh("Coordinates.dat", "Elements3.dat",
				    "Dirichlet.dat", true);
      std::cout << "Mesh = " << msh << std::endl;

      std::auto_ptr<concepts::BoundaryConditions>
	bc(new concepts::BoundaryConditions());
      bc->add(concepts::Attribute(1),
	      concepts::Boundary(concepts::Boundary::DIRICHLET, "(0)"));
      std::cout << "Boundary conditions: " << *bc << std::endl;

      Linear3d spc(msh, bc.get());
      std::cout << "Space: " << spc << std::endl;

      Laplace stiff_bf;
      concepts::PETScMat stiff(spc, stiff_bf);

      RHS rhs_lf;
      concepts::Vector<Real> rhs(spc, rhs_lf);
//       std::cout << "rhs = " << rhs << std::endl;

      concepts::PETSc solver(stiff, 1e-12, "bcgs");
      concepts::Vector<Real> sol(spc);
      try {
	solver(rhs, sol);
      }
      catch(concepts::NoConvergence& e) {
	std::cout << "solver did not converge!" << std::endl << e << std::endl;
      }
      std::cout << "solver = " << solver << std::endl;
//       std::cout << "solution = " << sol << std::endl;
      /* <div class="solution">
	 Compute the FE energy and the energy error and store the results
	 in the prepared array.
      */
      std::cout << "FE energy = " << sol*rhs << std::endl;
      output.addArrayDouble("energieFehler", i ,
			    fabs(exEnergie-sol*rhs)/exEnergie);
      std::cout << "=======================" << std::endl;
      /* Here, the loop over the four meshes ends.
       */
    } // for i
    /* Print the results which were gathered in the previous loop.
     */
    std::cout << inputParser << std::endl;
    /* </div>
     */
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
