/* The include files are the same as in the previous tutorial.
 */
#include <iostream>
#include <memory>
#include <vector>
#include <fstream>
#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
using concepts::Real;

/* <h3>Element</h3>
*/
// **************************************************************** Elements **

class LinearTet3d : public concepts::Element<Real> {
public:
  LinearTet3d(const concepts::Tetrahedron3d& cell, concepts::TColumn<Real>* T0,
	      concepts::TColumn<Real>* T1 = 0);
  const concepts::TMatrixBase<Real>& T() const { return *T_; }
  const concepts::Tetrahedron& support() const { return cell_.connector(); }
  concepts::Real3d vertex(const uint i) const { return cell_.vertex(i); }
  const concepts::Tetrahedron3d& cell() const { return cell_; }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
  const concepts::Tetrahedron3d& cell_;
};

LinearTet3d::LinearTet3d(const concepts::Tetrahedron3d& cell,
			 concepts::TColumn<Real>* T0,
			 concepts::TColumn<Real>* T1) :
  T_(new concepts::TMatrix<Real>(T1)), cell_(cell) {
  if (T0) {
    concepts::TMatrix<Real>* T =
      dynamic_cast<concepts::TMatrix<Real>*>(T_.get());
    conceptsAssert(T, concepts::Assertion());
    T->append(T0);
  }
}

std::ostream& LinearTet3d::info(std::ostream& os) const {
  return os << "FEMTriangle(" << cell_ << ", T = " << *T_ << ')';
}

/* <h3>Scanner for the Elements</h3>
*/
// ***************************************************************** Scanner **

namespace concepts {
  template<>
  class Scan<LinearTet3d> : public Scan<Element<Real> > {
  public:
    LinearTet3d& operator++(int) = 0;
  };
} // namespace concepts

/* <h3>Space</h3>
*/
// ******************************************************************* Space **

class Linear3d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<concepts::Element<Real> > Scan;
  Linear3d(concepts::Mesh3& msh, concepts::BoundaryConditions* bc = 0);
  uint dim() const { return dim_; }
  uint nelm() const { return nelm_; }
  Scan* scan() const {
    return new concepts::PListScan<concepts::Element<Real> >(*elm_);
  }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  concepts::BoundaryConditions* bc_;
  concepts::Joiner<concepts::Element<Real>*, 1>* elm_;
};

Linear3d::Linear3d(concepts::Mesh3& msh, concepts::BoundaryConditions* bc) :
  dim_(0), nelm_(0), bc_(bc), elm_(0) {
  /* <span class="code">bcMap</span> stores for every visited vertex
     if it has a valid degree of freedom (1) or not (0). <span
     class="code">bcMap</span> is filled in <span
     class="code">enforceBC_</span> depending on the level and
     boundary conditions.
  */
  concepts::DynArray<uint> bcMap(2, 3, 1);
  std::auto_ptr<concepts::Scan3> sc(0);
  if (bc_) {
    sc.reset(msh.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      concepts::Connector2* face;
      for (uint j = 0; (face = cell.connector().face(j)) != 0; ++j) {
	concepts::Connector0* vtx;
	for (uint k = 0; (vtx = face->vertex(k)) != 0; ++k) {
	  if ((*bc_)(face->attrib()).type() == concepts::Boundary::DIRICHLET)
	    bcMap[vtx->key()] = 0;
	  else
	    bcMap[vtx->key()];
	}
      }
    }
  }

  /* <span class="code">dofMap</span> stores for every visited vertex the
     assigned degree of freedom or the value <span class="code">(uint)-1</span>
     if no degree of freedom is assigned yet.
  */
  concepts::DynArray<uint> dofMap(2, 3, (uint)-1);
  sc.reset(msh.scan());
  while (*sc) {
    concepts::Cell3& cell = (*sc)++;
    const concepts::Tetrahedron3d* cellT =
      dynamic_cast<const concepts::Tetrahedron3d*>(&cell);
    conceptsAssert(cellT, concepts::Assertion());
    concepts::TColumn<Real>* T = 0;
    const uint n = 4;
    concepts::Connector0* vtx;
    for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j) {
      if (bcMap[vtx->key()]) {
	uint& i = dofMap[vtx->key()];
	if (i == (uint)-1)
	  i = dim_++;
	T = new concepts::TColumn<Real>(n, i, T);
	(*T)[j] = 1.0;
      }
    }
    nelm_++;
    elm_ = new concepts::Joiner<concepts::Element<Real>*, 1>
      (new LinearTet3d(*cellT, T), elm_);
  }
}

std::ostream& Linear3d::info(std::ostream& os) const {
  return os << "Linear3d(dim = " << dim_ << ", nelm = " << nelm_ << ')';
}

/* <h3>Linear Form</h3>
 */
// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();
  }
};

/* <h3>Main program</h3>
*/
// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    concepts::Import3dMesh msh("Coordinates.dat", "Elements3.dat",
				  "Dirichlet.dat", 1, true);
    std::cout << "Mesh = " << msh << std::endl;

    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    std::cout << "Boundary conditions: " << *bc << std::endl;

    Linear3d spc(msh, bc.get());
    std::cout << "Space: " << spc << std::endl;

    Laplace stiff_bf;
    concepts::DenseMatrix<Real> stiff(spc, stiff_bf);
    std::cout << "stiff = " << stiff << std::endl;

    RHS rhs_lf;
    concepts::Vector<Real> rhs(spc, rhs_lf);
    std::cout << "rhs = " << rhs << std::endl;

    concepts::CG<Real> solver(stiff, 1e-6, 200);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
    std::cout << "solution = " << sol << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
