/* The include files are the same as in the other tutorials.
 */
#include <cmath>
#include <iostream>
#include <sstream>
#include <memory>
#include "basics.hh"
#include "function.hh"
#include "geometry.hh"
#include "integration.hh"
#include "operator.hh"
#include "space.hh"
#include "toolbox.hh"

#include "basics/debug.hh"

using concepts::Real;

/* <h3>Mesh</h3>

   The mesh is taken from the <a href="meshes.html#LshapedQuad">meshes
   tutorial</a>:
   <span class="code">LshapedQuad</span>. Here, the relevant code is
   included:
*/
#include "meshes.cc"

/* <h3>Elements</h3>
*/
// **************************************************************** Elements **

class FEMElement2d : public concepts::Element<Real> {
public:
  FEMElement2d(const uint i, uint idx[]) :
    T_(new concepts::TIndex<Real>(i,i,idx)) {}
  FEMElement2d(concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1);
  const concepts::TMatrixBase<Real>& T() const { return *T_; }
protected:
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
};

FEMElement2d::FEMElement2d(concepts::TColumn<Real>* T0,
			   concepts::TColumn<Real>* T1) :
  T_(new concepts::TMatrix<Real>(T1)) {
  if (T0) {
    concepts::TMatrix<Real>* T =
      dynamic_cast<concepts::TMatrix<Real>*>(T_.get());
    conceptsAssert(T, concepts::Assertion());
    T->append(T0);
  }
}

/* <h4>Quadrilateral</h4>

   Since the inverse of a bilinear element map is no longer bilinear,
   we need to be able to evaluate the reference element shape functions
   at an arbitraty point. This is done by <span class="code">evaluate</span>
   and <span class="code">evaluateD</span> (for the derivatives).
 */
class FEMQuad : public FEMElement2d {
public:
  FEMQuad(const concepts::Quad2d& cell, uint idx[]);
  FEMQuad(const concepts::Quad2d& cell, concepts::TColumn<Real>* T0,
	  concepts::TColumn<Real>* T1 = 0);
  const concepts::Quad& support() const { return cell_.connector(); }
  concepts::Real3d vertex(const uint i) const { return cell_.vertex(i); }
  const concepts::Quad2d& cell() const { return cell_; }
  void evaluate(concepts::Real* res, const Real* x, const uint n) const;
  void evaluateD(concepts::Real2d* res, const Real* x, const uint n) const;
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  const concepts::Quad2d& cell_;
};

FEMQuad::FEMQuad(const concepts::Quad2d& cell, uint idx[]) :
  FEMElement2d(4, idx), cell_(cell) {}

FEMQuad::FEMQuad(const concepts::Quad2d& cell, concepts::TColumn<Real>* T0,
		 concepts::TColumn<Real>* T1) :
  FEMElement2d(T0, T1), cell_(cell) {}

/* <span class="code">evaluate</span> evaluates the shape functions in
   the given one dimensional points. The points have to be between -1
   and 1 and are tensorised to get two dimensional points.
*/
void FEMQuad::evaluate(concepts::Real* res, const Real* x,
			const uint n) const {
  for (uint i = 0; i < n; ++i)
    for (uint j = 0; j < n; ++j)
      *res++ = 0.25*(1.0-x[i])*(1.0-x[j]);
  for (uint i = 0; i < n; ++i)
    for (uint j = 0; j < n; ++j)
      *res++ = 0.25*(x[i]+1.0)*(1.0-x[j]);
  for (uint i = 0; i < n; ++i)
    for (uint j = 0; j < n; ++j)
      *res++ = 0.25*(x[i]+1.0)*(x[j]+1.0);
  for (uint i = 0; i < n; ++i)
    for (uint j = 0; j < n; ++j)
      *res++ = 0.25*(1.0-x[i])*(x[j]+1.0);
}

/* <span class="code">evaluateD</span> evaluates the derivatives of the
   shape functions in the given n one dimensional points. The points have
   to be between -1 and 1 and are tensorised to get two dimensional points.
*/
void FEMQuad::evaluateD(concepts::Real2d* res, const Real* x,
			const uint n) const {
  for (uint i = 0; i < n; ++i)
    for (uint j = 0; j < n; ++j)
      *res++ = concepts::Real2d( -0.25+0.25*x[j] , -0.25+0.25*x[i] );
  for (uint i = 0; i < n; ++i)
    for (uint j = 0; j < n; ++j)
      *res++ = concepts::Real2d( 0.25-0.25*x[j] , -0.25*x[i]-0.25 );
  for (uint i = 0; i < n; ++i)
    for (uint j = 0; j < n; ++j)
      *res++ = concepts::Real2d( 0.25*x[j]+0.25 , 0.25*x[i]+0.25 );
  for (uint i = 0; i < n; ++i)
    for (uint j = 0; j < n; ++j)
      *res++ = concepts::Real2d( -0.25*x[j]-0.25 , 0.25-0.25*x[i] );
}

std::ostream& FEMQuad::info(std::ostream& os) const {
  return os << "FEMQuad(" << cell_ << ", T = " << *T_ << ')';
}

/* <h3>Scanner for the Elements</h3>
*/
// ***************************************************************** Scanner **

namespace concepts {
  template<>
  class Scan<FEMQuad> : public Scan<Element<Real> > {
  public:
    FEMQuad& operator++(int) = 0;
  };
} // namespace concepts

/* <h3>Space</h3>
*/
// ******************************************************************* Space **

class SpaceNotBuilt : public concepts::ExceptionBase {};

class Linear2d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<concepts::Element<Real> > Scan;
  typedef void (*SMap)(const concepts::TColumn<Real>&,
		       concepts::TColumn<Real>&);
  Linear2d(concepts::Mesh2& msh, uint level,
	   concepts::BoundaryConditions* bc = 0);
  virtual ~Linear2d();
  uint dim() const { return dim_; }
  uint nelm() const { return nelm_; }
  Scan* scan() const;
  void adjust(const FEMQuad& elm, const short level);
  void rebuild();
  /* The S matrices are used for the irregular refinement of a
     mesh. An S matrix is independent of the geometry but it depends on
     the topological relation of the children with the parent: the
     bottom left child has a different S matrix than the top right etc.
  */
  const SMap S(uint i) const;
  static void S0(const concepts::TColumn<Real>& src,
		 concepts::TColumn<Real>& dst); // bottom left
  static void S1(const concepts::TColumn<Real>& src,
		 concepts::TColumn<Real>& dst); // bottom right
  static void S2(const concepts::TColumn<Real>& src,
		 concepts::TColumn<Real>& dst); // top right
  static void S3(const concepts::TColumn<Real>& src,
		 concepts::TColumn<Real>& dst); // top left
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  /* <span class="code">rebuild_</span> is set to <span
     class="code">true</span> if the degrees of freedom and the elements
     have to be rebuilt. This rebuilding is done by the member <span
     class="code">rebuild</span>.
  */
  bool rebuild_;
  concepts::Mesh2& msh_;
  concepts::BoundaryConditions* bc_;
  concepts::Joiner<concepts::Element<Real>*, 1>* elm_;
  /* The following five dynamic arrays are all indexed with a
     topological key. They grow in size if necessary.

     <span class="code">adj_</span> holds adjustment information for
     the cells, ie. which cell has to be refined in the next step.
  */
  concepts::DynArray<short> adj_;
  /* The following three arrays store control information of
     topological entities with the given dimension, ie. <span
     class="code">ctrl0_</span> stores information on vertices, <span
     class="code">ctrl1_</span> on edges and <span
     class="code">ctrl2_</span> on whole quadrilaterals. The entries
     in the arrays have to be taken as bit fields. The different bits
     have to following meaning:
     <ol>
     <li><b>Member of the space:</b> this entity is present in the space (1).
     <li><b>Passive:</b> this entity is passive (1, ie. its degrees of freedom
         are counted altough it might be member of the space) or active (0)
     <li><b>Valid index:</b> the index of this entity is valid, ie. it was
         it has a valid degree of freedom assigned.
     </ol>
  */
  concepts::DynArray<short> ctrl0_;
  concepts::DynArray<short> ctrl1_;
  concepts::DynArray<short> ctrl2_;
  /* <span class="code">dof_</span> stores the index of the degree of freedom
     which was assigned to a vertex.
  */
  concepts::DynArray<uint> dof_;
  static SMap S_[4];
  void rebuild0_(concepts::Connector2& cntr, const int l, int& L);
  void rebuild1_(concepts::Quad2d& cell, concepts::TColumn<Real>* T0 = 0);
  void deactivate_(concepts::Connector0& vtx);
  void deactivate_(concepts::Connector1& edg);
};

/* The constructor of the space initializes all private variables.  In
   addition, the parameter <span class="code">level</span> is assigned
   to the entry in <span class="code">adj_</span> of all cells in the
   mesh.  The control information of the cells in the mesh is set to
   <em>active member of the space with invalid index</em>.
*/
Linear2d::Linear2d(concepts::Mesh2& msh, uint level,
		   concepts::BoundaryConditions* bc) :
  dim_(0), nelm_(0), rebuild_(true), msh_(msh), bc_(bc), elm_(0),
  adj_(2, 3, 0), ctrl0_(2, 3, 0), ctrl1_(2, 3, 0), ctrl2_(2, 3, 0),
  dof_(2, 3, (uint)-1) {

  std::auto_ptr<concepts::Scan2> sc(msh_.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    concepts::Key key(cell.connector().key());
    adj_[key] = level;
    ctrl2_[key] = 1;
  }
}

Linear2d::~Linear2d() {
  delete elm_;
}

/* Since <span class="code">scan</span> is <span class="code">const</span>,
   it cannot call <span class="code">rebuild</span> directly, if this
   was necessary. Instead, <span class="code">info</span> throws an exception
   indicating that the information currently stored in the space has
   to be rebuilt.
*/
Linear2d::Scan* Linear2d::scan() const {
  if (rebuild_) throw conceptsException(SpaceNotBuilt());
  return new concepts::PListScan<concepts::Element<Real> >(*elm_);
}

/* <span class="code">adjust</span> marks an element for refinement and
   the space for rebuilding.
*/
void Linear2d::adjust(const FEMQuad& elm, const short level) {
  adj_[elm.support().key()] = level;
  rebuild_ = true;
}

/* <span class="code">rebuild</span> is called to recount the degrees
   of freedom and rebuild the elements. This happens after a refinement
   of the space.

   The first step is to delete all old information.
 */
void Linear2d::rebuild() {
  delete elm_;
  elm_ = 0;
  dim_ = 0;
  nelm_ = 0;
  ctrl0_.reset();
  ctrl1_.reset();
  /* For every cell in the coarsest mesh, <span
     class="code">rebuild0_</span> is called. It fills the control
     arrays up to very small changes for hanging nodes (these changes
     occur in <span class="code">rebuild1_</span>).
  */
  std::auto_ptr<concepts::Scan2> sc(msh_.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    int L = -1;
    rebuild0_(cell.connector(), 0, L);
  }
  /* Next, <span class="code">rebuild1_</span> is called for every
     cell in the coarsest mesh. It assigns the degrees of freedom and
     creates the elements.
  */
  sc.reset(msh_.scan());
  while (*sc) {
    concepts::Cell2& cell = (*sc)++;
    concepts::Quad2d* cellQ = dynamic_cast<concepts::Quad2d*>(&cell);
    if (cellQ) {
      rebuild1_(*cellQ);
      continue;
    }
    std::stringstream error;
    error << "cell " << cell << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }
  /* At the end, the <span class="code">rebuild_</span> flag is turned
     off and the adjustment information are deleted.
  */
  rebuild_ = false;
  adj_.reset();
}

const Linear2d::SMap Linear2d::S(uint i) const {
  conceptsAssert(i < 4, concepts::Assertion());
  return S_[i];
}

Linear2d::SMap Linear2d::S_[4] = { Linear2d::S0, Linear2d::S1,
				   Linear2d::S2, Linear2d::S3 };

void Linear2d::S0(const concepts::TColumn<Real>& src,
		  concepts::TColumn<Real>& dst) {
  dst[0] += 1.0 * src[0];
  dst[1] += 0.5 * src[0] + 0.5 * src[1];
  dst[2] += 0.25 * src[0] + 0.25 * src[1] + 0.25 * src[2] + 0.25 * src[3];
  dst[3] += 0.5 * src[0] + 0.5 * src[3];
}

void Linear2d::S1(const concepts::TColumn<Real>& src,
		  concepts::TColumn<Real>& dst) {
  dst[0] += 0.5 * src[0] + 0.5 * src[1];
  dst[1] += 1.0 * src[1];
  dst[2] += 0.5 * src[1] + 0.5 * src[2];
  dst[3] += 0.25 * src[0] + 0.25 * src[1] + 0.25 * src[2] + 0.25 * src[3];
}

void Linear2d::S2(const concepts::TColumn<Real>& src,
		  concepts::TColumn<Real>& dst) {
  dst[0] += 0.25 * src[0] + 0.25 * src[1] + 0.25 * src[2] + 0.25 * src[3];
  dst[1] += 0.5 * src[1] + 0.5 * src[2];
  dst[2] += 1.0 * src[2];
  dst[3] += 0.5 * src[2] + 0.5 * src[3];
}

void Linear2d::S3(const concepts::TColumn<Real>& src,
		  concepts::TColumn<Real>& dst) {
  dst[0] += 0.5 * src[0] + 0.5 * src[3];
  dst[1] += 0.25 * src[0] + 0.25 * src[1] + 0.25 * src[2] + 0.25 * src[3];
  dst[2] += 0.5 * src[2] + 0.5 * src[3];
  dst[3] += 1.0 * src[3];
}

/* Since <span class="code">info</span> is <span class="code">const</span>,
   it cannot call <span class="code">rebuild</span> directly, if this
   was necessary. Instead, <span class="code">info</span> throws an exception
   indicating that the information currently stored in the space has
   to be rebuilt.
*/
std::ostream& Linear2d::info(std::ostream& os) const {
  if (rebuild_) throw conceptsException(SpaceNotBuilt());
  return os << "Linear2d(dim = " << dim_ << ", nelm = " << nelm_ << ')';
}

/* <span class="code">rebuild0_</span> is called recursively. It
   determines the maximal level in the children of the current cell
   and sets the control information accordingly.
*/
void Linear2d::rebuild0_(concepts::Connector2& cntr, const int l, int& L) {
  short& ctrl(ctrl2_[cntr.key()]);
  if ((L == -1) && (ctrl & 1)) {
    short& adj(adj_[cntr.key()]);
    L = l + adj;
    if (L < 0)
      L = 0;
  }
  if (L == -1 || l < L) {
    int Lmax = -1;
    concepts::Connector2* chld = 0;
    for(uint i = 0; (chld = cntr.child(i)) != 0; ++i) {
      int LL = L;
      rebuild0_(*chld, l + 1, LL);
      if (LL > Lmax)
	Lmax = LL; 
    }
    L = Lmax;
  }
  if (l < L) {
    /* If the cell is not on the maximal level, it is not member of
       the space.
    */
    ctrl = 0;
  } else {
    /* On the maximal level, the cell is member of the space and its
       edges and vertices too.
    */
    ctrl = 1;
    concepts::Connector1* cntr1 = 0;
    for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i)
      ctrl1_[cntr1->key()] = 1;
    concepts::Connector0* cntr0 = 0;
    for (uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i)
      ctrl0_[cntr0->key()] = 1;
  }
  /* At the end, the Dirichlet boundary conditions are enforced by
     setting some topological entities to <em>passive</em>.
  */
  if (bc_) {
    concepts::Connector1* cntr1 = 0;
    for (uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
      if ((*bc_)(cntr1->attrib()).type() == concepts::Boundary::DIRICHLET) {
	ctrl1_[cntr1->key()] |= 2;
	concepts::Connector0* cntr0 = 0;
	for (uint j = 0; (cntr0 = cntr1->vertex(j)) != 0; ++j)
	  ctrl0_[cntr0->key()] |= 2;
      } // if DIRICHLET bc
    } // for i
  } // if bc_
}

/* <span class="code">rebuild1_</span> is called recursively. It
   assigns degrees of freedom to vertices, creates the T matrices,
   applies S matrices and creates elements.
*/
void Linear2d::rebuild1_(concepts::Quad2d& cell, concepts::TColumn<Real>* T0) {
  concepts::Quad& cntr = cell.connector();
  short& ctrl2(ctrl2_[cntr.key()]);
  const uint n = 4;
  concepts::TColumn<Real>* T1 = 0;
  /* Loop over all vertices of the current cell. If the vertex is
     member of the space and active, it gets a degree of freedom
     assigned.
  */
  concepts::Connector0* cntr0 = 0;
  for(uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i) {
    const concepts::Key& key = cntr0->key();
    short& ctrl0(ctrl0_[key]);
    if (ctrl0 & 3) {
      if ((ctrl0 & 3) == 1) {
	if ((ctrl0 & 4) == 0) {
	  dof_[key] = dim_++;
	  ctrl0 |= 4;
	}
	T1 = new concepts::TColumn<Real>(n, dof_[key], T1);
	(*T1)[i] = 1.0;
      }
      /* If the vertex is passive, the children are deactivated.
       */
      if ((ctrl2 & 1) == 0) deactivate_(*cntr0);
    }
  }
  concepts::Connector1* cntr1 = 0;
  for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
    short& ctrl1(ctrl1_[cntr1->key()]);
    if (ctrl1 & 3) {
      /* If the edge is passive, the children of the edge (and the
	 vertex in the middle) are deactivated.
      */
      if ((ctrl2 & 1) == 0) deactivate_(*cntr1);
    }
  }
  /* If the current cell is member of the space, a new element is created.
   */
  if (ctrl2 & 1) {
    FEMQuad* elm = new FEMQuad(cell, T0, T1);
    nelm_++;
    elm_ = new concepts::Joiner<concepts::Element<Real>*, 1>(elm, elm_);
  } else {
    /* If the current cell not member of the space, the S matrices are
       applied to the already built columns of the T matrix and <span
       class="code">rebuild0_</span> is called on the children of the
       current cell. If no T columns were created, the S matrices are not
       applied, ie. the S matrices are only used to resolve hanging nodes.
    */
    concepts::Cell2* child = 0;
    for (uint i = 0; (child = cell.child(i)) != 0; ++i) {
      concepts::TColumn<Real>* T2 = 0;
      for (concepts::TColumn<Real>* TT = T0; TT != NULL; TT = TT->link()) {
	T2 = new concepts::TColumn<Real>(n, TT->index(), T2);
	(S(i))(*TT, *T2);
      }
      for (concepts::TColumn<Real>* TT = T1; TT != NULL; TT = TT->link()) {
	T2 = new concepts::TColumn<Real>(n, TT->index(), T2);
	(S(i))(*TT, *T2);
      }
      rebuild1_(*cell.child(i), T2);
    }
  }
}

void Linear2d::deactivate_(concepts::Connector0& vtx) {
  ctrl0_[vtx.child(0)->key()] |= 2;
}

void Linear2d::deactivate_(concepts::Connector1& edg) {
  ctrl1_[edg.child(0)->key()] |= 2;
  ctrl1_[edg.child(1)->key()] |= 2;
  ctrl0_[edg.child(1)->vertex(0)->key()] |= 2;
}

/* <h3>Linear Form</h3>

   As the inverse of a bilinear map is no longer easily representable, we
   need numerical integration. <span class="math">f = 1</span>.
*/
// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();
    const FEMQuad* e = dynamic_cast<const FEMQuad*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    /* Get the vertices of the element. They are needed to compute the
       Jacobian of the element map.
     */
    concepts::Real2d y0(e->vertex(0));
    concepts::Real2d y1(e->vertex(1));
    concepts::Real2d y2(e->vertex(2));
    concepts::Real2d y3(e->vertex(3));

    concepts::Real2d p(y1 - y0);
    concepts::Real2d q(y3 - y0);
    concepts::Real2d r(y1 - y0 + y3 - y2);

    concepts::Real2d s, t;
    /* Get the quadrature points.
     */
    const concepts::Quadrature<0> quadrature(3);
    const Real* abscissas = quadrature.abscissas();
    const Real* weights = quadrature.weights();
    const uint qn = quadrature.n();
    /* Get the values of the shape functions in the quadrature points.
     */
    shpfct_.resize(n*qn*qn);
    e->evaluate((concepts::Real*)shpfct_, abscissas, qn);
    /* Numerical integration: loop over all quadrature points and compute
       the Jacobian. Then, loop over all reference element shape functions
       and add the contributions up.
     */
    for (uint qi = 0; qi < qn; ++qi) {
      for (uint qj = 0; qj < qn; ++qj) {
	s.lincomb(p, r, 1.0, -(1.0+abscissas[qj])/2.0);
	t.lincomb(q, r, 1.0, -(1.0+abscissas[qi])/2.0);
	const Real dx = fabs(s.cross(t)) * weights[qi] * weights[qj] / 4.0;
	for (uint i = 0; i < n; ++i)
	  em(i, 0) += dx * shpfct_[i*qn*qn+qi*qn+qj];
      }
    }
  }
private:
  concepts::Array<concepts::Real> shpfct_;
};

/* <h3>Bilinear Forms</h3>

   As the inverse of a bilinear map is no longer easily representable, we
   need numerical integration.

   The element mapping of a bilinear element looks like <div
   style="text-align: center"><img src="form-82.gif" width="186"
   height="40"></div> where <span class="math">N<sub>i</sub></span>
   are the four bilinear shape functions corresponding to the four
   edges of the quadrilateral. The derivatives of the element mapping
   are <div style="text-align: center"><img src="form-84.gif"
   width="394" height="43"></div> and <div style="text-align:
   center"><img src="form-85.gif" width="398" height="43"></div> The
   Jacobian of the element mapping is <img src="form-86.gif"
   width="67" height="18">.

   Using <span class="math">s</span>, <span class="math">t</span>, the
   Jacobian and <div style="text-align: center"><img src="form-89.gif"
   width="315" height="46"></div> we get <div style="text-align:
   center"><img src="form-90.gif" width="235" height="35"></div> and
   therefore <div style="text-align: center"><img src="form-91.gif"
   width="394" height="35"></div> all derivatives to <span
   class="math">x</span> are eliminated.
*/
// *********************************************************** Bilinear Form **

class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const FEMQuad* eX = dynamic_cast<const FEMQuad*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const FEMQuad* eY = dynamic_cast<const FEMQuad*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      concepts::Real2d y0(eX->vertex(0));
      concepts::Real2d y1(eX->vertex(1));
      concepts::Real2d y2(eX->vertex(2));
      concepts::Real2d y3(eX->vertex(3));

      concepts::Real2d p(y1 - y0);
      concepts::Real2d q(y3 - y0);
      concepts::Real2d r(y1 - y0 + y3 - y2);

      concepts::Real2d s, t;

      const concepts::Quadrature<0> quadrature(3);
      const Real* abscissas = quadrature.abscissas();
      const Real* weights = quadrature.weights();
      const uint qn = quadrature.n();

      shpfct_.resize(n*qn*qn);
      eX->evaluateD((concepts::Real2d*)shpfct_, abscissas, qn);

      for (uint qi = 0; qi < qn; ++qi) {
	for (uint qj = 0; qj < qn; ++qj) {
	  s.lincomb(p, r, 1.0, -(1.0+abscissas[qj])/2.0);
	  t.lincomb(q, r, 1.0, -(1.0+abscissas[qi])/2.0);
	  const Real dx = fabs(4.0 / s.cross(t))
	    * weights[qi] * weights[qj] / 4.0;
	  const Real ss = dx * s.l2_2();
	  const Real st = -dx * (s*t);
	  const Real tt = dx * t.l2_2();
	  for (uint i = 0; i < n; ++i) {
	    concepts::Real2d shpfcti = shpfct_[i*qn*qn+qi*qn+qj];
	    for (uint j = i; j < n; ++j) {
	      concepts::Real2d shpfctj = shpfct_[j*qn*qn+qi*qn+qj];
	      em(i, j) += shpfcti[0] * shpfctj[0] * tt
		+ (shpfcti[0] * shpfctj[1] + shpfcti[1] * shpfctj[0]) * st
		+ shpfcti[1] * shpfctj[1] * ss;
	    }
	  }
	}
      }
      for(uint i = 0; i < n; ++i) 
	for (uint j = i + 1; j < n; ++j) 
	  em(j, i) = em(i, j);
    }
  }
private:
  concepts::Array<concepts::Real2d> shpfct_;
};

/* <h3>Main program</h3>
*/
// ************************************************************ Main Program **

/* <span class="code">contains</span> returns true if the element contains
   the given point.
*/
bool contains(const FEMQuad* elm, const concepts::Real2d& x) {
  for (uint i = 0; i < 4; ++i) {
    if ((elm->vertex(i)[0] == x[0]) && (elm->vertex(i)[1] == x[1]))
      return true;
  }
  return false;
}

int main(int argc, char** argv) {

  try {
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(3),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(4),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(5),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(6),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(8),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    std::cout << "Boundary conditions: " << *bc << std::endl;

    LshapedQuad msh;
    std::cout << "Mesh: " << msh << std::endl;

    Linear2d spc(msh, 1, bc.get());
    spc.rebuild();
    std::cout << "Space: " << spc << std::endl;
    /* The space is refined. An element is refined, if it contains the
       origin. This results in a mesh which is geometrically refined
       towards the origin.
    */
    for (uint i = 0; i < 3; ++i) {
      std::auto_ptr<Linear2d::Scan> sc(spc.scan());
      while (*sc) {
	concepts::Element<Real>& elm = (*sc)++;
	FEMQuad* elmQ = dynamic_cast<FEMQuad*>(&elm);
	conceptsAssert(elmQ, concepts::Assertion());
	if (contains(elmQ, concepts::Real2d(0.0,0.0)))
	  spc.adjust(*elmQ, 1);
      }
      spc.rebuild();
      std::cout << "Space: " << spc << std::endl;
    }

    Laplace stiff_bf;
    concepts::DenseMatrix<Real> stiff(spc, stiff_bf);

    RHS rhs_lf;
    concepts::Vector<Real> rhs(spc, rhs_lf);
    std::cout << "rhs = " << rhs << std::endl;

    concepts::CG<Real> solver(stiff, 1e-6, 200);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
    std::cout << "solution = " << sol << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
