#include <memory>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <sstream>
#include <unistd.h>

#include "basics.hh"
#include "function.hh"
#include "geometry.hh"
#include "graphics.hh"
#include "operator.hh"
#include "space.hh"
#include "toolbox.hh"

#include "hp2D.hh"

using concepts::Real;

// ************************************************************* fem routine **

void fem(const concepts::InOutParameters& input,
	 concepts::BoundaryConditions& bc,
	 concepts::InOutParameters& output) {
  const uint l = input.getInt("level");
  const uint p = input.getInt("polynomial");
  const std::string meshNameBase = input.getString("meshNameBase");
  const Real exenergy = input.getDouble("exenergy");
  const uint solverType = input.getInt("solver");

  std::stringstream fileCoord;
  fileCoord << meshNameBase << "Coord.dat" << std::ends;
  output.addString("coordFile", fileCoord.str().c_str());
  std::stringstream fileElements;
  fileElements << meshNameBase << "Elements.dat" << std::ends;
  output.addString("elementsFile", fileElements.str().c_str());
  std::stringstream fileBoundary;
  fileBoundary << meshNameBase << "Boundary.dat" << std::ends;
  output.addString("boundaryFile", fileBoundary.str().c_str());
  concepts::Import2dMesh msh(fileCoord.str(), fileElements.str(),
			     fileBoundary.str());
  std::cout << "msh = " << msh << std::endl;

  hp2D::Space space(msh, 0, p, &bc);
  for (uint i = 0; i < l; ++i) {
    std::cout << "i = " << i << std::endl;
    concepts::ResourceMonitor timer;
    space.rebuild();
    if (space.dim() > 0) {
      output.addArrayDouble("space_time", i, timer.utime());
      output.addArrayInt("dof", i, space.dim());
      std::cout << "  space = " << space << std::endl;
      graphics::drawMeshEPS(space, "mesh.eps");
      graphics::drawMeshDX(space, "mesh.dx");

      // ** right hand side **
      hp2D::Riesz lform(concepts::ParsedFormula
			(input.getString("f").c_str()), &bc);
      concepts::Vector<Real> rhs(space, lform);

      // ** matrices **
      hp2D::Laplace la;
      hp2D::Identity id;
      concepts::BilinearFormLiCo<Real> bform(la, id, input.getDouble("a"),
					     input.getDouble("c"));

      timer.utime();
      concepts::SparseMatrix<Real> stiff(space, bform);
      output.addArrayDouble("stiff_time", i, timer.utime());
      stiff.compress();
      std::cout << "  matrix: " << stiff << std::endl;

      // ** set up solver **
      std::auto_ptr<concepts::Operator<Real> > solver(0);
      std::auto_ptr<concepts::DiagonalMatrix<Real> > diag(0);
      std::auto_ptr<concepts::DiagonalSolver<Real> > precond(0);
#ifdef HAS_PETSC
	std::auto_ptr<concepts::PETScMat> stiffP(0);
#endif
      switch(solverType) {
      case 0:
	diag.reset(new concepts::DiagonalMatrix<Real>(stiff));
	precond.reset(new concepts::DiagonalSolver<Real>(*diag));
	solver.reset(new concepts::CG<Real>(stiff, *precond, 1e-6,
					    2*stiff.spaceX().dim()+100));
	break;
#ifdef HAS_PETSC
      case 1:
	stiffP.reset(new concepts::PETScMat(stiff));
	solver.reset(new concepts::PETSc(*stiffP, 1e-12, "bcgs", "ilu"));
	break;
#endif
#ifdef HAS_SuperLU
      case 2: solver.reset(new concepts::SuperLU(stiff)); break;
#endif
#ifdef HAS_UMFPACK
      case 3: solver.reset(new concepts::Umfpack(stiff, true)); break;
#endif
#ifdef HAS_PARDISO
      case 4: solver.reset(new concepts::Pardiso(stiff, true)); break;
#endif
      default:
	throw conceptsException
	  (concepts::MissingFeature("solver not present"));
      }

      // ** solve **
      concepts::Vector<Real> sol(space);
      timer.utime();
      (*solver)(rhs, sol);
      output.addArrayDouble("solve_time", i, timer.utime());

      std::cout << "  solver = " << *solver << std::endl;
      std::cout << "  solve time " << output.getArrayDouble("solve_time", i)
		<< std::endl;
      switch(solverType) {
      case 0: {
	const concepts::CG<Real>* s =
	  dynamic_cast<const concepts::CG<Real>*>(solver.get());
	conceptsAssert(s, concepts::Assertion());
	output.addArrayInt("iterations", i, s->iterations());
	break;
      }
#ifdef HAS_PETSC
      case 1: {
	const concepts::PETSc* s =
	  dynamic_cast<const concepts::PETSc*>(solver.get());
	conceptsAssert(s, concepts::Assertion());
	output.addArrayInt("iterations", i, s->iterations());
	break;
      }
#endif
      }

      const Real FEenergy = sol*rhs;
      std::cout << "  FE energy = " << std::setprecision(20)
		<< FEenergy << std::setprecision(6);
      output.addArrayDouble("feenergy", i, FEenergy);
      const Real relError = fabs(FEenergy - exenergy)/exenergy;
      std::cout << ", rel. energy error = " << std::setprecision(20)
		<< relError << std::setprecision(6) << std::endl;
      output.addArrayDouble("energyError", i, relError);

      if (input.getBool("graphics")) {
	std::stringstream name;
	name << "hpFEM2d-" << i << ".gnuplot" << std::ends;
	graphics::drawDataGnuplot(space, name.str(), sol);
      }
    } // if dim > 0

    // ** refinement **
    if (i < l-1) {
      int pMax[2] = { 1, 1 };
      hp2D::APrioriRefinement refineSpace(space, input.getInt("vtxRef"),
					  input.getInt("edgRef"), pMax);
      concepts::GlobalPostprocess<Real> post(space);
      post(refineSpace);
    }
  } // for level i
}

// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    concepts::InputParser inputParser(true);
    concepts::InOutParameters& input = inputParser.inputParameters();
    inputParser.parse("../../../applications/hpFEM2d.concepts");

    input.addInt("level", 3);
    input.addInt("polynomial", 1);
    input.addBool("graphics", false);

    input.addDouble("a", 1);
    input.addDouble("c", 0);
    input.addString("f", "(0)");
    input.addInt("vtxRef", 10);
    input.addInt("edgRef", 0);

    input.addInt("solver", 1);

    input.addString("meshNameBase", "../../../applications/lshape2D");

    input.addString("title" , "singularity in (0,0) in L shaped domain");

    std::stringstream outfile;
    outfile << argv[0] << ".out" << std::ends;
    input.addString("parameterout", outfile.str().c_str());

    std::stringstream outfileG;
    outfileG << argv[0] << ".gnuplot" << std::ends;
    input.addString("gnuplot", outfileG.str().c_str());

    concepts::InOutParameters& output = inputParser.outputParameters();
    output.addString("version", "$Id: hpFEM2d.cc,v 1.2 2004/06/03 05:56:31 pfrauenf Exp $");
    output.addArrayDouble("energyError");
    output.addArrayDouble("feenergy");
    output.addArrayInt("dof");
    output.addArrayInt("iterations");
    output.addArrayDouble("space_time");
    output.addArrayDouble("stiff_time");
    output.addArrayDouble("solve_time");
    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::INT, "dof", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "energyError", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "feenergy", output);
    table.addMap(concepts::ResultsTable::INT, "iterations", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "space_time", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "stiff_time", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "solve_time", output);

    // ********************************************************** input data **
    std::ofstream devnull("/dev/null");
    std::string inputfile;
    int opt;
    while ((opt = getopt(argc, argv, "-f:l:p:g")) != EOF)
      switch(opt) {
      case 'l': input.addInt("level", std::atoi(optarg)); break;
      case 'p': input.addInt("polynomial", std::atoi(optarg)); break;
      case 'f': inputfile = std::string(optarg);
	inputParser.parse(inputfile);
	break;
      case 'g': input.addBool("graphics", true); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-f FILE] [-l LEVEL] [-p DEGREE] [-d]"
		  << std::endl
		  << "where" << std::endl
		  << "  FILE: name of the input file" << std::endl
		  << "  LEVEL: level of refinement" << std::endl
		  << "  DEGREE: polynomial degree" << std::endl
		  << "  -g: graphics of solution" << std::endl
		  << "Options given after the input file override the values "
		  << "read from the"
		  << std::endl << "input file." << std::endl;
	std::exit(1);
	break;
      }

    // ************************************************* boundary conditions **
    concepts::BoundaryConditions bc;
    for (int i = 1; i <= input.getInt("numberbc"); ++i) {
      bc.add(concepts::Attribute(i),
	     concepts::Boundary((concepts::Boundary::boundaryTypes)
				input.getArrayInt("bctype", i),
				concepts::ParsedFormula
				(input.getArrayString("bcform", i).c_str())));
    }

    // ***************************************************** show parameters **
    std::cout << '[' << argv[0] << "]" << std::endl;
    std::cout << "--" << std::endl;
    std::cout << "Parameters:" << std::endl
	      << "  input file = " << inputfile << std::endl
	      << input;
    std::cout << "--" << std::endl;
    std::cout << bc << std::endl;
    std::cout << "--" << std::endl;

    // ******************************************************** computations **
    fem(input, bc, output);

    // ************************************** output of input data and other **
    std::ofstream* ofs = new std::ofstream(input.getString("gnuplot").c_str());
    *ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(*ofs);
    delete ofs;

    std::cout << "--" << std::endl;
    std::cout << output << std::endl;
    std::cout << table << std::endl;
    std::cout << "--" << std::endl
	      << "Writing gathered data to disk: " 
	      << input.getString("parameterout") << std::endl;

    ofs = new std::ofstream(input.getString("parameterout").c_str());
    *ofs << "/* program:\t" << argv[0] << std::endl
	 << " * command:\t";
    for (int i = 0; i < argc; ++i)
      *ofs << argv[i] << " ";
    *ofs << std::endl
	 << " * input file:\t" << inputfile << std::endl;
    *ofs << " */" << std::endl << std::setprecision(20) << inputParser;
    delete ofs;
  }

  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
