/* square made out of one or two hexahedron
 */

#ifndef square_hh
#define square_hh

#include "geometry.hh"

using concepts::Real;

// ****************************************************************** Square **

/** Mesh for \f$[0,1]^2\f$ with one quadrilateral. All boundaries have
    attribute 0.
    @author Philipp Frauenfelder, 2001
 */
class Square : public concepts::Mesh2 {
  static const uint ncell_ = 1;
public:
  Square(const Real sizex = 1.0, const Real sizey = 1.0);
  Square(const Real x0, const Real y0, const Real x1, const Real y1, 
	 const Real x2, const Real y2, const Real x3, const Real y3);
  virtual ~Square();

  inline uint ncell() const { return ncell_; }
  inline concepts::Scan2* scan() { return new S(cell_); }

  virtual std::ostream& info(std::ostream& os) const;
private:
  class S : public concepts::Scan<concepts::Cell2> {
    uint idx_;
    concepts::Quad2d *(&cell_)[ncell_];
  public:
    inline S(concepts::Quad2d *(&cell)[ncell_]) : idx_(0), cell_(cell) {}
    inline S(const S &scan) : idx_(scan.idx_), cell_(scan.cell_) {}

    inline bool eos() const { return idx_ == ncell_; }
    inline concepts::Cell2& operator++(int) { return *cell_[idx_++]; }

    inline concepts::Scan2* clone() const { return new S(*this); }
  };

  concepts::Vertex* vtx_[4];
  concepts::Edge* edg_[4];
  concepts::Quad* quad_[ncell_];

  concepts::Quad2d* cell_[ncell_];
};

// ***************************************************************** Square2 **

/** Mesh for \f$[0,1]^2\f$ with two quadrilaterals. All boundaries have
    attribute 0.

    One quad is \f$[0,1]\times[0,0.5]\f$ and the other one is
    \f$[0,1]\times[0.5,1]\f$.
    
    @author Philipp Frauenfelder, 2001
 */
class Square2 : public concepts::Mesh2 {
  static const uint ncell_ = 2;
public:
  Square2();
  virtual ~Square2();

  inline uint ncell() const { return ncell_; }
  inline concepts::Scan2* scan() { return new S(cell_); }

  virtual std::ostream& info(std::ostream& os) const;
private:
  class S : public concepts::Scan<concepts::Cell2> {
    uint idx_;
    concepts::Quad2d *(&cell_)[ncell_];
  public:
    inline S(concepts::Quad2d *(&cell)[ncell_]) : idx_(0), cell_(cell) {}
    inline S(const S &scan) : idx_(scan.idx_), cell_(scan.cell_) {}

    inline bool eos() const { return idx_ == ncell_; }
    inline concepts::Cell2& operator++(int) { return *cell_[idx_++]; }

    inline concepts::Scan2* clone() const { return new S(*this); }
  };

  concepts::Vertex* vtx_[6];
  concepts::Edge* edg_[7];
  concepts::Quad* quad_[ncell_];

  concepts::Quad2d* cell_[ncell_];
};

#endif // square_hh
