/* Debugging Header File
 */

#ifndef debug_hh
#define debug_hh

#ifdef DEBUG

#include <iostream>
#include <iomanip>
#include <cstring>
#include <sys/types.h>

#if defined (__GNUC__) || defined(__SUNPRO_CC)
// #  ifdef __GNUC__
// #    define DEBUG_FUNCTION __PRETTY_FUNCTION__
// #  else
#    define DEBUG_FUNCTION __func__
// #  endif
#else
#  define __PRETTY_FUNCTION__ "(unknown)"
#  define DEBUG_FUNCTION __PRETTY_FUNCTION__
#endif

namespace concepts {
  char* shorten(const char* FILE);
}

#define DEBUGL(doit, msg)\
DPL(doit, "[34m[" << concepts::shorten(__FILE__) << ':' << std::setw(4) \
    << __LINE__ << "][m " << DEBUG_FUNCTION << " --", msg)

#define DPL(doit, msg, var)\
if(doit!=0) std::cout << msg << " " << var << std::endl;

#define DP(doit, msg, var)\
if(doit!=0) std::cout << msg << " " << var;

#else
 #define DEBUGL(doit, msg)
 #define DP(doit, msg, var)
 #define DPL(doit, msg, var)
 #define NDEBUG // discard assertions
#endif

#include <cassert>
#include <iomanip>

#endif // debug_hh

/** @def DEBUGL
    Debug Line. Prints a debugging message with information about file,
    line and function where it is located. At the end, the line is wrapped.
    The message looks like
    <pre>[file.cc, line 62] void function(double) -- this is the message</pre>

    @param doit The line is only printed if true
    @param msg Text in a stream (can include variables etc.) to be printed
    @ingroup debugging
*/

/** @def DPL
    Debug Print Line. Prints a debugging message with a variable if
    the first parameter is true. At the end, the line is wrapped.

    @param doit The line is only printed if true.
    @param msg Some text
    @param var A variable
*/

/** @def DP
    Debug Print. Prints a debugging message with a variable if
    the first parameter is true. At the end, the line is not wrapped.

    @param doit The line is only printed if true.
    @param msg Some text
    @param var A variable
*/

/** @fn char * concepts::shorten(const char *FILE)
    Shortens the string to the given size. The last part of the string is
    printed, the first part might be skipped. This is used by the macros
    DP, DPL and DEBUGL.
*/

/** @defgroup debugging Debugging Concepts
    Various methods and tools for debugging parts of the library or
    your application.
    <h2>Debugging Concepts with a Debugger</h2>

    This is quite simple: make sure you use the \c -g flag of the
    compiler for compiling the library and your application and then
    use a debugger on the resulting binary. Examples for debuggers
    are:
    <ul>
    <li><a href="http://www.gnu.org/software/gdb/">GDB: The GNU
    Debugger</a>, no graphical frontend
    <li>xxgdb, a graphical frontend to GDB
    <li><a href="http://www.gnu.org/software/ddd/">DDD: The Data
    Display Debugger</a>, a Motif/Lesstif frontend to GDB
    <li><a href="http://members.nextra.at/johsixt/kdbg.html">KDbg</a>, a
    QT/KDE frontend to GDB
    <li><a href="http://www.concerto.demon.co.uk/UPS/">UPS</a>, Ups is a
    source level C,C++ and Fortran debugger that runs under X11. It
    runs native on some versions of UNIX (e.g. BSD/OS, FreeBSD,
    Linux, SunOS & Solaris, etc.), on others it can be built as a
    front end to GDB.
    </ul>

    I prefer DDD, which should be part of every Linux distribution and
    is installed on the ETH workstations.
    <h2>Debugging Concepts with Output Statements and Assertions</h2>
    <h3>Output Statements</h3>

    There exists a header file to simplify the creation of debugging
    output statements: debug.hh. It contains one compiler macros which
    is important: <span class="class">DEBUGL</span>.  This macro is
    only evaluated if the \c DEBUG symbol is defined.

    In order to have fine grained debugging possibilities, you should use
    different symbols in a local debugging header file. For instance:
    \code #include "basics/debug.h"

#ifndef dgDebug_h
#define dgDebug_h

#define dgBDiffApplOperator_D 1
#define dgBDiffIntegrateEdge_D 0

#endif \endcode
    In our example, all
    statements with \c dgBDiffApplOperator_D
    are executed. Such a statement could look like this:
    \code DEBUGL(dgBDiffApplOperator_D, "elmX: " << quadX.key()
       << ", elmY:", quadY.key()); \endcode
    The output of the statment above looks like
    \code [../dg/dgBForm.cc, line 36] dg_BDiffusion::operator() -- elmX: 1, elmY: 0 \endcode

    If you have more complex statements which you would like to hide for
    production use, you can do something like the following
    \code #if dgBDiffApplOperator_D
cout << "[" << __FILE__ << ", line " << __LINE__ 
     << "] dg_BDiffusion::operator() --" << endl;
cout << "blah blub" << endl;
#endif \endcode

    Like before, these statements are only executed (and even only
    compiled) if \c DEBUG is defined and \c dgBDiffApplOperator_D is
    true (ie. non-zero).
    <h3>Assertions</h3>

    The class concepts::Assertion can be used to assert certain
    conditions.
    \code #include "basics/exceptions.hh"

// ...

const int* m = dynamic_cast<const int*>(n);
conceptsAssert(m != 0, Assertion());
cout << *m << endl;

// ...

conceptsAssert3(a == b, Assertion(), "a = " << a << ", b = " << b); \endcode

    If the assertion fails, an exception of type concepts::Assertion
    is thrown. If the macro \c conceptsAssert3 is used instead of \c
    conceptsAssert, the additional information given as the third
    argument is printed in case of an error. It can be caught by
    \code try {
  // here, an assertion might fail
}
catch (Assertion& e) {
  std::cout << e << std::endl;
} \endcode
    Instead of catching \c Assertion it is also possible to catch \c
    ExceptionBase (the base class of all exceptions in the library).
    <h2>Stacktrace</h2>

    In case an assertion fails, a stacktrace is printed by calling <a
    href="http://www.gnu.org/software/gdb/">gdb</a> in
    concepts::Stacktrace::print. This can be controlled by
    concepts::Stacktrace::doit:
    \code concepts::Stacktrace::doit() = false; \endcode
    inhibits printing stack traces.

    @author Philipp Frauenfelder, 2004
*/
