/* exceptions
 *
 * Assert
 * an exeption should be thrown with information about: file, line, function,
 * violated condition, exception name. This exception can be handled in the
 * main program.
 */

#ifndef exceptions_hh
#define exceptions_hh

#include <exception>
#include <iostream>
#include <string>
#include <cstdlib>
#include <sstream>

#include "outputOperator.hh"

namespace concepts {

  class Stacktrace;

  // ********************************************************* ExceptionBase **

  /** Base class for exceptions. This class features informations about 
      the exceptions thrown, ie. filename and line, name of the function and
      of the exception. The ouput operator (<<) can display this
      information. 

      The typical usage of this class is to throw an exception if an
      error has occured. This can be done by 
      <pre>throw conceptsException(ExceptionBase());</pre>

      These exceptions are always present (in a production code as well as in
      development code). If you want to have assertions using exceptions
      (which are only in the development code), have a look at
      Assertion and conceptsAssert.

      There are two different error cases: <ul>
      <li>The error is a real programming error and should not happen
          during normal and correct operation. Such an error is caught
	  with an assertion. Such an assertion can be thrown (using an
	  exception) by 
	  <pre>conceptsAssert(1 == 0, Assertion());</pre>
	  These assertions are not working with the optimised version
	  of the library
      <li>An error which can also happen during normal operation of
          the library which should also be detected in optimised
	  mode. Such an error is caught with a normal throw. To use to
	  provided classes comfortably, use the macro conceptsException like
	  this:
	  <pre>throw conceptsException(ExceptionBase());</pre>
      </ul>

      @see conceptsException
      @see conceptsAssert
      @see Assertion

      @test test::ExceptionTest
      @author Philipp Frauenfelder, 2000
  */
  class ExceptionBase : public std::exception,
			public OutputOperator {
  public:
    /// Default constructor
    ExceptionBase() throw();

    /** Constructor.

	@param file Filename where the exception was thrown from
	@param line Line where the exception was thrown from
	@param function Name of the function that threw the exception
	@param excName The name of the exception
    */
    ExceptionBase(const std::string& file, const unsigned int line, 
	 const std::string& function, const std::string& excName) throw();

    virtual ~ExceptionBase() throw();

    /** Sets the fields of this exception.

	@param file Filename where the exception was thrown from
	@param line Line where the exception was thrown from
	@param function Name of the function that threw the exception
	@param excName The name of the exception
    */
    void setFields(const std::string& file, const unsigned int line, 
		   const std::string& function, 
		   const std::string& excName) throw();

  protected:
    virtual std::ostream& info(std::ostream& os) const throw();

    /// Filename where the exception was thrown from
    std::string file_;

    /// Line where the exception was thrown from
    unsigned int line_;

    /// Name of the function that threw the exception
    std::string function_;

    /// The name of the exception with calling sequence
    std::string excName_;
  };

  // ***************************************************** DimensionMismatch **

  /** Exception class to express dimensions not matching.

      @author Philipp Frauenfelder, 2001
  */
  class DimensionMismatch : public ExceptionBase {};

  // ****************************************************** IndexNotExisting **

  /** Exception class to express that an index in a dynamic array does
      not exist.

      @author Philipp Frauenfelder, 2001
  */
  class IndexNotExisting : public ExceptionBase {};

  // ******************************************************** MissingFeature **

  /** Exception class to express a missing feature. Besides the information
      already present in ExceptionBase, this class features more information
      about the feature which is missing. The output operator (<<) can display
      this information.

      @see conceptsException
      @see ExceptionBase

      @test test::ExceptionTest
      @author Philipp Frauenfelder, 2000
  */
  class MissingFeature : public ExceptionBase {
  public:
    /** Constructor.

	@param feature The missing feature
     */
    MissingFeature(const std::string& feature) throw();

    /** Constructor.

	@param file Filename where the exception was thrown from
	@param line Line where the exception was thrown from
	@param function Name of the function that threw the exception
	@param excName The name of the exception
	@param feature The missing feature
     */
    MissingFeature(const std::string& file, const unsigned int line, 
		   const std::string& function, const std::string& excName, 
		   const std::string& feature = std::string("")) throw();

    virtual ~MissingFeature() throw();
  protected:
    virtual std::ostream& info(std::ostream& os) const throw();

    /// Missing feature
    std::string feature_;
  };

  // ************************************************************* Assertion **

  /** Exception class for assertions. Besides the information already
      present in ExceptionBase, this class features more information about
      the assertion which failed. The ouput operator (<<) can display
      this information.

      The typical usage of this class is to throw an exception if an
      assertion has failed. This can be done by
      <pre>conceptsAssert(1 == 0, Assertion());</pre>

      These exceptions are not always present, since the macro
      conceptsAssert is switched off in production code. If you want
      to have exceptions which are always present, have a look at
      ExceptionBase and conceptsException.

      @see conceptsAssert
      @see conceptsException
      @see ExceptionBase

      @test test::ExceptionTest
      @author Philipp Frauenfelder, 2000
      @ingroup debugging
  */
  class Assertion : public MissingFeature {
  public:
    /// Default constructor
    Assertion() throw();

    /** Constructor.

	@param file Filename where the exception was thrown from
	@param line Line where the exception was thrown from
	@param function Name of the function that threw the exception
	@param excName The name of the exception
	@param cond Condition which was violated
    */
    Assertion(const std::string& file, const unsigned int line, 
	      const std::string& function, const std::string& excName, 
	      const std::string& cond) throw();

    virtual ~Assertion() throw();

    /** Sets the fields of this exception.

	@param file Filename where the exception was thrown from
	@param line Line where the exception was thrown from
	@param function Name of the function that threw the exception
	@param excName The name of the exception
	@param cond Condition which was violated
    */
    void setFieldsAssert(const std::string& file, const unsigned int line, 
			 const std::string& function, 
			 const std::string& excName, 
			 const std::string& cond = std::string("")) throw();
  protected:
    virtual std::ostream& info(std::ostream& os) const throw();
  };

  // ************************************************** exception_set_fields **

#ifndef __GNUC__
#  define __PRETTY_FUNCTION__ "(unknown)"
#endif

  /** Sets fields on exception and throws it. This routine does the main 
      work for the conceptsException macro. This routine should not be called 
      directly, it is used by the conceptsException macro.

      @return An exception which can then be thrown

      @param e Exception to be thrown
      @param file Filename where the exception was thrown from
      @param line Line where the exception was thrown from
      @param function Name of the function that threw the exception
      @param excName The name of the exception

      @see conceptsException
      @see ExceptionBase
      @author Philipp Frauenfelder, 2000
  */
  template<class exc>
  exc exception_set_fields(exc e, const std::string& file, 
			   const unsigned int line,
			   const std::string& function, 
			   const std::string& excName) {
    e.setFields(file, line, function, excName);
#ifdef DEBUG
    Stacktrace::print();
#endif
#ifdef __SUNPRO_CC
    std::cout << e << std::endl;
    std::exit(0);
#endif
    return e;
  }

  // ***************************************************** conceptsException **

  /** Prepares an exception for throwing. 
      This is the main routine in the exception
      mechanism for production mode error checking. See the
      ExceptionBase class for more information.

      @param exc Exception to be thrown

      @see ExceptionBase
      @test test::ExceptionTest
      @author Philipp Frauenfelder, 2000
  */
#define conceptsException(exc) \
  exception_set_fields(exc, std::string(__FILE__), __LINE__, \
		       std::string(__PRETTY_FUNCTION__), std::string(#exc))

  // ************************************************ exception_throw_assert **

  /** Sets the fields of an assertion and throws it. 
      This routine does the main work for the exception generation
      mechanism used in the conceptsAssert macro. This routine should not be
      called directly, it is used by the conceptsAssert macro.
 
      @see conceptsAssert
      @see Assertion
      @author Philipp Frauenfelder, 2000
  */
  template<class exc>
  void exception_throw_assert(const std::string& file, int line, 
			      const std::string& function,
			      const std::string& exc_name, 
			      const std::string& cond, exc e) {
    e.setFieldsAssert(file, line, function, exc_name, cond);
    Stacktrace::print();
#ifdef __SUNPRO_CC
    std::cout << e << std::endl;
    std::exit(0);
#else
    throw e;
#endif
  }

  // ************************************************************ Stacktrace **

  /** Dumps a stack trace using gdb. This class is used if an assertion
      fails using conceptsAssert.
      @see conceptsAssert
      @author Philipp Frauenfelder, 2001
      @ingroup debugging
  */
  class Stacktrace {
  public:
    /// Prints the stacktrace
    static void print();
    /// If doit is set to false, no stacktrace is printed
    static bool& doit() { return doit_; }
  private:
    static bool doit_;
  };

  // ******************************************************** conceptsAssert **

#ifndef DEBUG
  /** Assert that a certain condition is fulfilled. Otherwise, issue an
      exception (if its not catched, the program is aborted). This is
      the main routine in the exception mechanism for debug mode error
      checking. See the Assertion class for more information. 

      @param cond Condition, if unfulfill an exception is thrown
      @param exc Exception to throw, if the condition is unfulfilled

      @see Assertion
      @test test::ExceptionTest
      @author Philipp Frauenfelder, 2000
      @ingroup debugging
  */
#define conceptsAssert(cond, exc) { }
#else
#define conceptsAssert(cond, exc) { \
    if (!(cond)) { \
      std::stringstream errorMsg; \
      errorMsg << "by violating condition <" << #cond << ">"; \
      exception_throw_assert(std::string(__FILE__), __LINE__, \
			     std::string(__PRETTY_FUNCTION__), \
                             std::string(#exc), errorMsg.str(), exc); \
    } \
  }
#endif // DEBUG

  // ******************************************************* conceptsAssert3 **

#ifndef DEBUG
  /** Assert that a certain condition is fulfilled. Otherwise, issue an
      exception (if its not catched, the program is aborted). This is an
      extention to the \c conceptsAssert macro. The additional feature is
      the third parameter \c msg:
      <pre>conceptsAssert3(1 == 0, Assertion(), "add " << "a message"
                << " here");</pre>
      gives
      <pre>[file.cc, line 34] int main(int, char**) -- error Assertion() by violating condition <1 == 0>, add a message here</pre>
      as error message.

      @param cond Condition, if unfulfill an exception is thrown
      @param exc Exception to throw, if the condition is unfulfilled
      @param msg Additional message in stream notation

      @see Assertion
      @test test::ExceptionTest
      @author Philipp Frauenfelder, 2003
      @ingroup debugging
  */
#define conceptsAssert3(cond, exc, msg) { }
#else
#define conceptsAssert3(cond, exc, msg) { \
    if (!(cond)) { \
      std::stringstream errorMsg; \
      errorMsg << "by violating condition <" << #cond << ">, " << msg; \
      exception_throw_assert(std::string(__FILE__), __LINE__, \
			     std::string(__PRETTY_FUNCTION__), \
                             std::string(#exc), errorMsg.str(), exc); \
    } \
  }
#endif // DEBUG

} // namespace

#endif // exceptions_hh
