// Test from Chuck Allison in C/C++ Users Journal, September 2000:
// "The Simplest Automated Unit Test Framework That Could Possibly Work"

#ifndef TEST_H
#define TEST_H

#include <iostream>
#include <string>
#include <iosfwd>

/** @defgroup testing Unit Tests
    Unit tests to automatically test intended behaviour.

    It is good practice to cover all code by unit tests. This should
    certainly be aimed at for all new code. Examples for such unit
    tests can be found in the <span class="filename">testsuite</span>
    directories and in the classes in the <span
    class="code">test</span> namespace.

    test::TestCase is the base class for all tests. A new test (for a
    class or a part of a class or a small set of classes) should be derived
    from test::TestCase. Several test cases can be grouped together by
    using test::TestSuite in the following way:
    \code int main(int argc, char** argv) {
  try {
    test::TestSuite t("Constrained Matrix Eigenvalue Problems");
    test::GolubExample t1;           t.addTest(&t1);
    test::MaxwellTransmissionEVP t2; t.addTest(&t2);
    test::GolubExampleSum t3;        t.addTest(&t3);
    t.run();
    return t.report();
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
} \endcode

    @see Chuck Allison, <a href="http://www.cuj.com/documents/s=8035/cuj0009allison1/">The Simplest Automated Unit Test Framework That Could Possibly Work</a>, C/C++ Users Journal, September 2000.
    @author Philipp Frauenfelder, 2004
*/

/** Unit tests. This namespace contains two basic classes for unit tests:
    TestCase and TestSuite. They are intended to support testing. For a
    test case, a specialization of TestCase has to be written. TestSuite is
    meant to group a few test cases, detailed description in
    \ref testing "unit tests".
    @see Chuck Allison, <a href="http://www.cuj.com/documents/s=8035/cuj0009allison1/">The Simplest Automated Unit Test Framework That Could Possibly Work</a>, C/C++ Users Journal, September 2000.
    @author Philipp Frauenfelder, 2003
    @ingroup testing
*/
namespace test {

  using std::string;
  using std::ostream;

  // The following have underscores because they are macros
  // (and it's impolite to usurp other users' functions!).
  // For consistency, _succeed() also has an underscore.

  /// Tests if \c cond is true (test passed) or not (test failed)
#define _test(cond) do_test(cond, #cond, __FILE__, __LINE__)
  /// Tests if \c num and \c orig are equal up to a rel. difference of 1e-10
#define _numtest(num, orig) do_numtest(num, orig, #num, #orig, __FILE__, __LINE__)
  /// Tests if \c num and \c orig are equal up to a rel. difference of \c tol
#define _numtesttol(num, orig, tol) do_numtest(num, orig, #num, #orig, __FILE__, __LINE__, tol)
  /// Explicitly fails a test
#define _fail(str) do_fail(str, __FILE__, __LINE__)

  /** Base class for tests. Writing a test case is done by deriving from this
      class.
      @author Chuck Allison, 2000.
              Doxygen comments by Philipp Frauenfelder, 2003.
      @see Chuck Allison, <a href="http://www.cuj.com/documents/s=8035/cuj0009allison1/">The Simplest Automated Unit Test Framework That Could Possibly Work</a>, C/C++ Users Journal, September 2000.
      @test test::ToTest
      @ingroup testing
  */
  class TestCase {
  public:
    /** Constructor.
	@param osptr Pointer to the output stream. Defaults to stdout.
    */
    TestCase(ostream* osptr = 0);
    virtual ~TestCase() {}
    /// Runs the tests. Must be overwritten by the specialization.
    virtual void run() = 0;

    /// Returns number of passed tests
    long getNumPassed() const { return m_nPass; }
    /// Returns number of failed tests
    long getNumFailed() const { return m_nFail; }
    /// Returns output stream
    const ostream* getStream() const { return m_osptr; }
    /// Sets the output stream
    void setStream(ostream* osptr) { m_osptr = osptr; }

    /// Explicitly succeds a test
    void _succeed() { ++m_nPass; }
    /** Prints a report on the number of passed and failed tests to
	the output stream.
	@return Number of failed tests.
    */
    long report() const;
    /// Resets the counters for the failed and passed tests
    virtual void reset() { m_nPass = m_nFail = 0; }
  protected:
    /// Internal function to do a test
    void do_test(bool cond, const string& lbl,
                 const char* fname, long lineno);
    /// Internal function to do a numerical test
    void do_numtest(double num, double orig, const string& lbl,
		    const string& lbl2, const char* fname, long lineno,
		    const double tol = 1e-10);
    /** Internal function to report a failed test (besides increasing the
	failed counter)
    */
    void do_fail(const string& lbl,
                 const char* fname, long lineno);
  private:
    ostream* m_osptr;
    long m_nPass;
    long m_nFail;

    /// Disallowed
    TestCase(const TestCase&);
    /// Disallowed
    TestCase& operator=(const TestCase&);
  };

  inline TestCase::TestCase(ostream* osptr) :
    m_osptr(osptr), m_nPass(0), m_nFail(0) {
    if (m_osptr == 0)
      m_osptr = &std::cout;
  }

} // namespace test

#endif // TEST_H
