// TestSuite from Chuck Allison in C/C++ Users Journal, September 2000:
// "The Simplest Automated Unit Test Framework That Could Possibly Work"

#include "testsuite.hh"
#include "exceptions.hh"
#include <iostream>
#include <stdexcept>
#include <cassert>

using namespace std;

namespace test {

  // ******************************************************** TestSuiteError **

  TestSuiteError::TestSuiteError(const std::string& error) throw() :
    concepts::MissingFeature(error) {}

  std::ostream& TestSuiteError::info(std::ostream& os) const throw() {
    return os << '[' << file_ << ", line " << line_ << "] " << function_ 
	      << " -- test suite error: '" << feature_ << "'";
  }

  // ************************************************************* TestSuite **

  void TestSuite::addTest(TestCase* t) throw(TestSuiteError) {
    // Make sure test has a stream:
    if (t == 0)
      throw TestSuiteError("null test");
    else if (m_osptr != 0 && t->getStream() == 0)
      t->setStream(m_osptr);

    m_tests.push_back(t);
    t->reset();
  }

  void TestSuite::addSuite(const TestSuite& s) throw(TestSuiteError) {
    for (size_t i = 0; i < s.m_tests.size(); ++i)
      addTest(s.m_tests[i]);
  }

  void TestSuite::free() {
    // This is not a destructor because tests
    // don't have to be on the heap.
    for (size_t i = 0; i < m_tests.size(); ++i)
      {
        delete m_tests[i];
        m_tests[i] = 0;
      }
  }

  void TestSuite::run() {
    reset();
    for (size_t i = 0; i < m_tests.size(); ++i) {
      conceptsAssert(m_tests[i], concepts::Assertion());
      m_tests[i]->run();
    }
  }


  long TestSuite::report() const {
    if (m_osptr) {
      long totFail = 0;
      *m_osptr << "TestSuite \"" << m_name << "\"\n=======";
      size_t i;
      for (i = 0; i < m_name.size(); ++i)
	*m_osptr << '=';
      *m_osptr << "=\n";

      for (i = 0; i < m_tests.size(); ++i) {
	  conceptsAssert(m_tests[i], concepts::Assertion());
	  totFail += m_tests[i]->report();
        }

      *m_osptr << "=======";
      for (i = 0; i < m_name.size(); ++i)
	*m_osptr << '=';
      *m_osptr << "=\n";
      *m_osptr << "Total:\tPassed: " << getNumPassed()
	       << "\tFailed: " << totFail << endl;
      return totFail;
    } else
      return getNumFailed();
  }

  long TestSuite::getNumPassed() const {
    long totPass = 0;
    for (size_t i = 0; i < m_tests.size(); ++i) {
      conceptsAssert(m_tests[i], concepts::Assertion());
      totPass += m_tests[i]->getNumPassed();
    }
    return totPass;
  }

  long TestSuite::getNumFailed() const {
    long totFail = 0;
    for (size_t i = 0; i < m_tests.size(); ++i) {
      conceptsAssert(m_tests[i], concepts::Assertion());
      totFail += m_tests[i]->getNumFailed();
    }
    return totFail;
  }

  void TestSuite::reset() {
    for (size_t i = 0; i < m_tests.size(); ++i) {
      conceptsAssert(m_tests[i], concepts::Assertion());
      m_tests[i]->reset();
    }
  }

} // namespace test
