// TestSuite from Chuck Allison in C/C++ Users Journal, September 2000:
// "The Simplest Automated Unit Test Framework That Could Possibly Work"

#ifndef testSUITE_H
#define testSUITE_H

#include "testcase.hh"
#include "exceptions.hh"
#include <vector>
using std::string;
using std::ostream;
using std::vector;

namespace test {

  // ******************************************************** TestSuiteError **

  /** Exception for TestSuite
      @author Philipp Frauenfelder, 2003
  */
  class TestSuiteError : public concepts::MissingFeature {
  public:
    TestSuiteError(const std::string& error) throw();
  protected:
    virtual std::ostream& info(std::ostream& os) const throw();
  };

  // ************************************************************* TestSuite **

  /** Suite of tests. Running a group of test cases is best be done by
      adding each test case to a test suite. This suite does then the
      run and report of the whole group.
      @author Chuck Allison, 2000.
              Doxygen comments by Philipp Frauenfelder, 2003.
      @see Chuck Allison, <a href="http://www.cuj.com/documents/s=8035/cuj0009allison1/">The Simplest Automated Unit Test Framework That Could Possibly Work</a>, C/C++ Users Journal, September 2000.
      @ingroup testing
  */
  class TestSuite {
  public:
    /** Constructor.
	@param name Name of the test suite
	@param osptr Pointer to the output stream. Defaults to stdout.
    */
    TestSuite(const string& name, ostream* osptr = 0);

    /// Returns name of test suite
    string getName() const { return m_name; }
    /// Returns number of passed tests
    long getNumPassed() const;
    /// Returns number of failed tests
    long getNumFailed() const;
    /// Returns output stream
    const ostream* getStream() const { return m_osptr; }
    /// Sets the output stream
    void setStream(ostream* osptr) { m_osptr = osptr; }
    
    /// Adds a test case to the suite of tests
    void addTest(TestCase* t) throw (TestSuiteError);
    /// Adds a test suite to this test suite
    void addSuite(const TestSuite&) throw(TestSuiteError);
    /// Runs all test cases in the suite
    void run();
    /** Prints a report on the number of passed and failed tests in
	the whole suite to the output stream.
	@return Number of failed tests.
    */
    long report() const;
    /// Deletes the tests
    void free();
  private:
    string m_name;
    ostream* m_osptr;
    vector<TestCase*> m_tests;
    void reset();

    /// Disallowed
    TestSuite(const TestSuite&);
    /// Disallowed
    TestSuite& operator=(const TestSuite&);
  };

  inline TestSuite::TestSuite(const string& name, ostream* osptr) :
    m_name(name), m_osptr(osptr) {
    if (m_osptr == 0)
      m_osptr = &std::cout;
  }

} // namespace test

#endif // testSUITE_H
