// analytical constraint

#include <memory>
#include "analytical.hh"
#include "basics/typedefs.hh"
#include "element.hh"

// debugging
#include "basics/debug.hh"

#define ConstraintsListAssembly_D 0
#define ConstraintsListAssemblyC_D 0
#define ConstraintsListAppl_D 0
#define ConstraintsListApplLinear_D 0

namespace constraints {

  using concepts::Real;

  // ******************************************************* ConstraintsList **

  template<typename F>
  ConstraintsList<F>::~ConstraintsList() {
    concepts::Joiner<AnalyticalConstraint<F>*, 1>::destructor(constraints_);
  }

  template<typename F>
  void  ConstraintsList<F>::add(const AnalyticalConstraint<F>& c) {
    add(c.clone());
  }

  template<typename F>
  void ConstraintsList<F>::add(AnalyticalConstraint<F>* c) {
    constraints_ = new concepts::Joiner<AnalyticalConstraint<F>*, 1>
      (c, constraints_);
  }

  template<typename F>
  void ConstraintsList<F>::assembly(const concepts::Element<F>& e,
				    concepts::TColumn<F>*& t, uint& n) {
    DEBUGL(ConstraintsListAssembly_D, e);

    // count localy active constraints -> length of TColumn
    uint length = 0;
    std::auto_ptr<concepts::PListScan<AnalyticalConstraint<F> > >
      sc(new concepts::PListScan<AnalyticalConstraint<F> >(*constraints_));
    while (*sc) {
      AnalyticalConstraint<F>& c = (*sc)++;
      DEBUGL(ConstraintsListAssemblyC_D, "c = " << c);
      c.lengthT(e, length);
    }
    DEBUGL(ConstraintsListAssembly_D, "length = " << length);

    uint j = 0;
    if (length > 0) {
      DEBUGL(ConstraintsListAssembly_D, "j = " << j);
      sc.reset
	(new concepts::PListScan<AnalyticalConstraint<F> >(*constraints_));
      while (*sc) {
	AnalyticalConstraint<F>& c = (*sc)++;
	c.assembly(e, t, n, j, length);
      }
    }
  }

  template<typename F>
  void ConstraintsList<F>::operator()(const concepts::Element<F>& elmX,
				      const concepts::Element<F>& elmY,
				      concepts::ElementMatrix<F>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();
    bool transpose = false;

    const concepts::Element<F>* compElm = &elmX;
    const Element<F>* constrElm = dynamic_cast<const Element<F>*>(&elmY);
    if (constrElm == 0) {
      constrElm = dynamic_cast<const Element<F>*>(&elmX);
      conceptsAssert3(constrElm, concepts::Assertion(), "elmX = " << elmX);
      compElm = &elmY;
      transpose = true;
      em.transpose();
    }
    conceptsAssert(constrElm, concepts::Assertion());

    if (compElm == &(constrElm->compElement())) {
      DEBUGL(ConstraintsListAppl_D, elmX << ", " << elmY << " -> " << em);
      concepts::PListScan<AnalyticalConstraint<F> > sc(*constraints_);
      uint j = 0;
      while (sc) {
	AnalyticalConstraint<F>& c = sc++;
	DEBUGL(ConstraintsListAppl_D, c);
	c.localMatrix(em, j, *constrElm);
      }
      DEBUGL(ConstraintsListAppl_D, em);
    }
    if (transpose)
      em.transpose();
  }

  template<typename F>
  void ConstraintsList<F>::operator()(const concepts::Element<F>& elm,
				      concepts::ElementMatrix<F>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();

    const Element<F>* constrElm = dynamic_cast<const Element<F>*>(&elm);
    conceptsAssert(constrElm, concepts::Assertion());

    DEBUGL(ConstraintsListApplLinear_D, elm << " -> " << em);
    concepts::PListScan<AnalyticalConstraint<F> > sc(*constraints_);
    uint j = 0;
    while (sc) {
      AnalyticalConstraint<F>& c = sc++;
      c.localVector(em, j, *constrElm);
    }
    DEBUGL(ConstraintsListApplLinear_D, em);
  }

  template<typename F>
  std::ostream& ConstraintsList<F>::info(std::ostream& os) const {
    os << "ConstraintsList(";
    concepts::PListScan<AnalyticalConstraint<F> > sc(*constraints_);
    while (sc) {
      AnalyticalConstraint<F>& c = sc++;
      os << c;
      if (sc) os << ", ";
    }
    return os << ')';
  }

  template class ConstraintsList<Real>;

} // namespace constraints
