/* analytical constraint:
   - a condition on a part of the boundary (ie. homogeneous Dirichlet)
   - a condition on the whole domain (ie. mean value zero)
*/

#ifndef analyticalConstr_hh
#define analyticalConstr_hh

#include "basics/outputOperator.hh"
#include "toolbox/scannerConnectors.hh"
#include "operator/bilinearForm.hh"
#include "function/linearForm.hh"

namespace concepts {
  template<typename F>
  class TColumn;
}

namespace constraints {

  template<typename F>
  class Element;

  // ************************************************** AnalyticalConstraint **

  /** Models an analytical constraint. This class can generate the information
      for assembling the constraints (\c assembly) for the constraints matrix
      (\c localMatrix) and for the constraints vector (\c localVector). They
      are all used by \c ConstraintsList.

      Examples:
      <ul>
      <li>a condition on a part of the boundary (ie. homogeneous Dirichlet)
      <li>a condition on the whole domain (ie. mean value zero)
      </ul>

      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class AnalyticalConstraint : public concepts::OutputOperator {
  public:
    /** Computes the length of the TColumn. This method decides if this
	constraint is active on element \c e and how many local equations
	it generates. This is added to \c length.
    */
    virtual void lengthT(const concepts::Element<F>& e, uint& length) = 0;
    /** Returns the TColumns \c t for this constraint in element \c e.
	\c n contains the next free global index in the space of
	constraints.

	When treating dof by dof constraints, \c n should be increased
	after the creation of the TColumn (which gets as global index the old
	value of \c n).

	When treating global constraints such as a zero mean value, \c
	n should be increased only once and all local constraints have
	to give this index to \c idx in the constructor of the new
	TColumn.

	A similar idea stands behind \c j: it counts the local constraints
	which have already been treated. It therefore has to be increased
	after treating a constraint. \c j can be taken as index for the
	created TColumn to fill in the correct entry.
	@param e Element of the computational space
	@param t List of TColumns
	@param n Next free global index in the space of constraints. \c n is
	         increased if a global constraint is hit for the first time.
	@param j Next free local index for the constraints in this element.
	         \c j is increased accordingly when treating constraints in
		 this element.
	@param length Length of TColumn
	@see \c concepts::TColumn
    */
    virtual void assembly(const concepts::Element<F>& e,
			  concepts::TColumn<F>*& t, uint& n, uint& j,
			  uint length) = 0;
    /** Creates clone of itself and returns it. This clone function
	implements a virtual constructor (cf. Stroustrup p. 424).
    */    
    virtual AnalyticalConstraint<F>* clone() const = 0;

    /** Fills the entries in the local constraints matrix \c em.
	@param em Local constraints matrix. Has to be filled. The rows
	          correspond to local dofs, the columns correspond to
		  the equations this constraints creates in this element.
	@param j Local number of the constraints. Everytime, when a active
	         constraint on the element is treated, \c j has to be
		 increased by the number of equations this constraint
		 created.
	@param elm Element of the constraints space
    */
    virtual void localMatrix(concepts::ElementMatrix<F>& em, uint& j,
			     const Element<F>& elm) const = 0;

    /** Fills the entries in the local constraints vector \c em.
	The parameters have the same meaning as in \c localMatrix.
    */
    virtual void localVector(concepts::ElementMatrix<F>& em, uint& j,
			     const Element<F>& elm) const = 0;
  };

  // ******************************************************* ConstraintsList **

  /** List of \c AnalyticalConstraint.
      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class ConstraintsList : public concepts::BilinearForm<F>,
			  public concepts::LinearForm<F> {
  public:
    /// Default constructor
    ConstraintsList() : n_(0), constraints_(0) {}
    virtual ~ConstraintsList();

    /** Adds an \c AnalyticalConstraint to the list. \c c is copied before
	adding it to the list.
    */
    void add(const AnalyticalConstraint<F>& c);

    /** Adds an \c AnalyticalConstraint to the list. The list takes care
	of the removal of \c c.
    */
    void add(AnalyticalConstraint<F>* c);

    /** Returns the TColumns \c t and the number of TColumns \c n for
	this constraint in element \c e. Calls \c assembly on every constraint
	of \c constraints_.
    */
    void assembly(const concepts::Element<F>& e,
		  concepts::TColumn<F>*& t, uint& n);

    virtual void operator()(const concepts::Element<F>& elmX,
			    const concepts::Element<F>& elmY,
			    concepts::ElementMatrix<F>& em);

    virtual void operator()(const concepts::Element<F>& elm,
			    concepts::ElementMatrix<F>& em); 
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Number of global constraints on the degree of freedom level
    uint n_;

    /// List of constraints
    concepts::Joiner<AnalyticalConstraint<F>*, 1>* constraints_;
  };

} // namespace constraints

#endif // analyticalConstr_hh
