// ARPACK eigensolver using dnaupd

#include <cstring>
#include "ARPACKdnaupd.hh"
#include "basics/fortranException.hh"
#include "operator/compositions.hh"
#include "toolbox/array.hh"

// debugging
#include "basics/debug.hh"

#define ARPACKcompute_D 0
#define ARPACKdestr_D 0

#if defined(__hpux) || defined(_IBMR2)
#define FORTRAN(X) X
#else
#define FORTRAN(X) X ## _
#endif

#ifdef __cplusplus
extern "C" {
#endif
  extern void FORTRAN(dnaupd)(int& IDO, const char& BMAT, const int& N,
                              const char WHICH[2], const int& NEV,
                              const double& TOL, const double* RESID,
                              const int& NCV, const double* V, const int& LDV,
                              int IPARAM[11], int IPNTR[14],
                              const double* WORKD, const double* WORKL,
                              const int& LWORKL, int& INFO);
  extern void FORTRAN(dneupd)(const int& RVEC, const char& HOWMNY,
                              const int* SELECT, const double* DR,
                              const double* DI, const double* Z, 
                              const int& LDZ, const double& SIGMAR,
                              const double& SIGMAI, const double* WORKEV,
                              const char& BMAT, const int& N,
                              const char WHICH[2], const int& NEV,
                              const double& TOL, const double* RESID,
                              const int& NCV, const double* V, const int& LDV,
                              int IPARAM[11], int IPNTR[14],
                              const double* WORKD, const double* WORKL,
                              const int& LWORKL, int& INFO);
#ifdef __cplusplus
}
#endif

namespace eigensolver {

  // ********************************************************* ArPackNonSymm **

  ArPackNonSymm::ArPackNonSymm(concepts::Operator<Real>& OP,
                               concepts::Operator<Real>& A,
                               concepts::Operator<Real>& B,
                               const int kmax, const Real tol,
                               const int maxiter, enum which target,
                               enum modus mode, const Real sigma,
                               const concepts::Vector<Real>* start,
                               const bool schur) :
    OP_(OP), A_(A), B_(B), kmax_(kmax), maxiter_(maxiter), tol_(tol),
    numop_(0), numopb_(0), numreo_(0), k_conv_(0), iter_(0),
    eigenvalues_(0), eigenvectors_(0), ev_(0),
    target_(target), mode_(mode), sigmaR_(sigma), sigmaI_(0),
    start_(start), schur_(schur), computed_(false) {
      conceptsAssert(OP_.spaceX().dim() == B_.spaceX().dim(),
                     concepts::Assertion());
      conceptsAssert(OP_.spaceX().dim() == A_.spaceX().dim(),
                     concepts::Assertion());
      conceptsAssert(OP_.spaceX().dim() == OP_.spaceY().dim(),
                     concepts::Assertion());
      conceptsAssert(B_.spaceX().dim() == B_.spaceY().dim(),
                     concepts::Assertion());
      conceptsAssert(A_.spaceX().dim() == A_.spaceY().dim(),
                     concepts::Assertion());
      conceptsAssert(kmax_ > 0, concepts::Assertion());
      conceptsAssert((uint)kmax_ < OP_.spaceX().dim(),
                     concepts::Assertion());
      conceptsAssert(kmax_ > 0, concepts::Assertion());
    }

  ArPackNonSymm::~ArPackNonSymm() {
    DEBUGL(ARPACKdestr_D, "start.");
    for (uint i = 0; i < ev_.size(); ++i) {
      delete ev_[i];
    }
    DEBUGL(ARPACKdestr_D, "done.");
  }

  const concepts::Array<Real>& ArPackNonSymm::getEV() {
    if (!computed_) {
      compute_();
    }
    return eigenvalues_;
  }

  const concepts::Array<concepts::Vector<Real>*>& ArPackNonSymm::getEF() {
    if (!computed_) {
      compute_(); // eigenvectors_ is filled
    }
    return ev_;
  }

  std::ostream& ArPackNonSymm::info(std::ostream& os) const {
    os << "ArPackNonSymm(";
    switch (mode_) {
    case NORMAL: os << "normal mode"; break;
    case REGINV: os << "regular inverse mode"; break;
    case SHIFTINV: os << "shift-invert mode"; break;
    case SHIFTINVIMAG: os << "shift-invert imaginary mode"; break;
    }
    os << '(' << mode_ << "), " << (computed_ ? "" : "not yet ")
       << "computed ";
    switch (target_) {
    case LM: os << "largest (in magnitude) eigenvalues"; break;
    case SM: os << "smallest (in magnitude) eigenvalues"; break;
    case LR: os << "largest (in real part) eigenvalues"; break;
    case SR: os << "smallest (in real part) eigenvalues"; break;
    case LI: os << "largest (in imaginary part) eigenvalues"; break;
    case SI: os << "smallest (in imaginary part) eigenvalues"; break;
    }
    return os << ", sigma = " << sigmaR_ << ", n = " << OP_.spaceX().dim()
              << ", tol = " << tol_ << ", conv. eigenpairs = " << k_conv_
              << "(max = " << kmax_ << "), Arnoldi iter = "
              << iter_ << "(max = " << maxiter_ << "), " << numop_
              << " OP*x ops., " << numopb_ << " B*x ops., "
              << numreo_ << " steps of re-orth.)";
  }

  void ArPackNonSymm::compute_() {
    int ido = 0;
    const char bmat = 'G';
    const int n = OP_.spaceX().dim();
    char whichEV[2];
    switch (target_) {
    case LM: strncpy(whichEV, "LM", 2); break;
    case SM: strncpy(whichEV, "SM", 2); break;
    case LR: strncpy(whichEV, "LR", 2); break;
    case SR: strncpy(whichEV, "SR", 2); break;
    case LI: strncpy(whichEV, "LI", 2); break;
    case SI: strncpy(whichEV, "SI", 2); break;
    }
    const int ncv = 2*kmax_+1 <= n ? 2*kmax_+1 : n;
    // suggestion in source code: >= 2*kmax+1
    const int lworkl = ncv*(6+3*ncv);
    concepts::Array<Real> resid(n, 0), workd(3*n, 0),
      workl(lworkl, 0), workev(3*ncv, 0);
    DEBUGL(ARPACKcompute_D, "before resizing");
    eigenvectors_.resize(ncv*n);
    DEBUGL(ARPACKcompute_D, "after resizing");
    const int ldv = n;
    concepts::Array<int> iparam(11, 0), ipntr(14, 0);
    iparam[0] = 1; // ishift = 1: use exact shifts
    iparam[1] = 0; // no longer referenced
    iparam[2] = maxiter_; // max number of Arnoldi iterations allowed
    iparam[3] = 1; // blocksize, only 1 is supported
    iparam[4] = 0; // number of converged eigenvalues
    iparam[5] = 0; // no longer referenced
    iparam[6] = mode_; // mode
    iparam[7] = 0; // not needed for ishift = 1
    int info = 0;

    if (start_) {
      info = 1;
      std::memcpy(resid, *start_, n*sizeof(Real));
    }

    DEBUGL(ARPACKcompute_D, "bmat = " << bmat << ", n = " << n << ", which = "
           << whichEV[0] << whichEV[1] << ", nev = " << kmax_ << ", tol = "
           << tol_ << ", ncv = " << ncv << ", iparam = " << iparam);
    while (ido != 99) {
      FORTRAN(dnaupd)(ido, bmat, n, whichEV, kmax_, tol_, (Real*)resid, ncv,
                      (Real*)eigenvectors_, ldv, (int*)iparam, (int*)ipntr,
                      (Real*)workd, (Real*)workl, lworkl, info);
      DEBUGL(ARPACKcompute_D, "ido = " << ido << ", info = " << info
             << ", ipntr = " << ipntr << ", iparam = " << iparam);
      switch (ido) {
      case -1:  {
        concepts::Vector<Real> y(B_.spaceX(), (Real*)workd+ipntr[1]-1),
          x(B_.spaceY(), (Real*)workd+ipntr[0]-1);
        DEBUGL(ARPACKcompute_D, "y = OP*x -- " << ipntr[0] << ", "
               << ipntr[1] << ", x = " << x);
        if (mode_ == REGINV) {
          A_(x, y);
          x = y;
        }
        if (mode_ == SHIFTINV) {
          B_(x, y);
          x = y;
        }
        OP_(x, y);
        DEBUGL(ARPACKcompute_D, "x = " << x << ", y = " << y);
      } break;
      case 1: {
        concepts::Vector<Real> y(B_.spaceX(), (Real*)workd+ipntr[1]-1),
          x(B_.spaceY(), (Real*)workd+ipntr[(mode_ == SHIFTINV) ? 2 : 0]-1);
        DEBUGL(ARPACKcompute_D, "y = OP*x -- " << ipntr[0] << ", "
               << ipntr[1] << ", x = " << x);
        if (mode_ == REGINV) {
          A_(x, y);
          x = y;
        }
        OP_(x, y);
        DEBUGL(ARPACKcompute_D, "x = " << x << ", y = " << y);
      } break;
      case 2: {
        DEBUGL(ARPACKcompute_D, "y = B*x -- " << ipntr[0] << ", " << ipntr[1]);
        concepts::Vector<Real> y(B_.spaceX(), (Real*)workd+ipntr[1]-1),
          x(B_.spaceY(), (Real*)workd+ipntr[0]-1);
        B_(x, y);
        DEBUGL(ARPACKcompute_D, "x = " << x << ", y = " << y);
      } break;
      case 99: break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
    }
    DEBUGL(ARPACKcompute_D, "resid = " << resid << ", iparam = " << iparam
           << ", ipntr = " << ipntr << ", info = " << info << ", workl = "
           << workl);
    iter_ = iparam[2];
    k_conv_ = iparam[4];
    numop_ = iparam[8];
    numopb_ = iparam[9];
    numreo_ = iparam[10];
    switch (info) {
    case 0: break; // everything ok
    case 1: break; // not all eigenpairs have converged
    default: throw conceptsException(concepts::FortranException
                                     (info, "dnaupd"));
    }

    concepts::Array<int> select(ncv, true);
    eigenvalues_.resize(2*kmax_);
    DEBUGL(ARPACKcompute_D, "before dneupd");
    FORTRAN(dneupd)(1, schur_ ? 'P' : 'A', select, (Real*)eigenvalues_,
                    ((Real*)eigenvalues_)+kmax_,
                    (Real*)eigenvectors_, ldv, sigmaR_, sigmaI_, workev,
                    bmat, n, whichEV, kmax_, tol_, (Real*)resid, ncv,
                    (Real*)eigenvectors_, ldv, (int*)iparam, (int*)ipntr,
                    (Real*)workd, (Real*)workl, lworkl, info);
    DEBUGL(ARPACKcompute_D, "info = " << info);
    if (info != 0) throw conceptsException(concepts::FortranException
                                           (info, "dneupd"));
    DEBUGL(ARPACKcompute_D, "eigenvalues = " << eigenvalues_
           << ", eigenvectors = " << eigenvectors_);
    ev_.resize(kmax_);
    for (int i = 0; i < kmax_; ++i) {
      ev_[i] = new concepts::Vector<Real>(OP_.spaceX(), eigenvectors_+i*n);
    }
    computed_ = true;
  }

} // namespace eigensolver
