// ARPACK eigensolver using dnaupd

#ifndef ARPACKdnaupd_hh
#define ARPACKdnaupd_hh

#include <cfloat>
#include "eigensolver/eigens.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class Operator;
} // namespace concepts

namespace eigensolver {

  // ********************************************************* ArPackNonSymm **

  /** Eigenvalue solver using ARPACK, the routine \c dnaupd.

  \c dnaupd uses implicitly restarted Arnoldi iteration to solve
  the generalized eigenvalue problem \f$ A x = \lambda B x \f$
  with A general non-symmetric and B symmetric and positive definite.
  For A symmetric, use \c ArPack instead.

  @see <a href="http://www.caam.rice.edu/software/ARPACK/">ARPACK Homepage</a>
  @author Christoph Winkelmann, 2004
  */
  class ArPackNonSymm : public EigenSolver {
  public:
    /** Specify which of the Ritz values \f$ \nu \f$ of OP
        (described in \c modus) to compute.
    */
    enum which { LM, SM, LR, SR, LI, SI };
    /** @enum LM
        Compute the largest (in magnitude) eigenvalues. */
    /** @enum SM
        Compute the smallest (in magnitude) eigenvalues. */
    /** @enum LR
        Compute the largest (in real part) eigenvalues. */
    /** @enum SR
        Compute the smallest (in real part) eigenvalues. */
    /** @enum LI
        Compute the largest (in imaginary part) eigenvalues. */
    /** @enum SI
        Compute the smallest (in imaginary part) eigenvalues. */

    /** Specify mode of ARPACK which should be used to compute the Ritz
        values \f$ \nu \f$ of OP.
    */
    enum modus { NORMAL = 1, REGINV = 2, SHIFTINV = 3, SHIFTINVIMAG = 4 };
    /** @enum NORMAL
        Normal mode. Solves \f$ A x = \lambda x \f$ where A is
        general non-symmetric. OP = A. B = I. Not implemented yet. */
    /** @enum REGINV
        Regular inverse mode. Solves \f$ A x = \lambda M x \f$ where A is
        general non-symmetric and M is symmetric positive definite.
        OP = inv(M). B = M. \f$ \lambda = \nu \f$. */
    /** @enum SHIFTINV
        Shift-invert mode. Solves \f$ A x = \lambda M x \f$ where A is
        general non-symmetric and M is symmetric positive semi-definite.
        OP = inv(A - sigma*M). B = M. \f$ \lambda = \sigma + 1/\nu \f$.
        For the eigenvalues \f$ \lambda \f$ just to the right of sigma,
        \c which should be LA, for those just to the left use SA.
        The routine returns eigenvalues \f$ \lambda \f$ and not Ritz
        values \f$ \nu \f$. */
    /** @enum SHIFTINVIMAG
        Not implemented yet. */

    /** Constructor.
        @param OP Operator OP as described in \c modus
        @param A Stiffness matrix
        @param B Operator B as descirbed in \c modus
        @param kmax Number of eigenpairs to be computed
        @param tol Convergence tolerance for the eigenpairs. The default
        value 0.0 is replaced by \c DLAMCH('EPS') from LAPACK.
        @param maxiter Maximum number of Arnoldi iterations allowed
        @param target What sort of eigenvalues to compute
        @param mode Mode in which ARPACK should be used
        @param sigma Shift for the shift-invert modes
        @param start Initial vector for iteration
        @param schur Calculate Schur vector basis instead of eigenvectors
    */
    ArPackNonSymm(concepts::Operator<Real>& OP, concepts::Operator<Real>& A,
                  concepts::Operator<Real>& B, const int kmax = 1,
                  const Real tol = 0.0, const int maxiter = 300,
                  enum which target = SM, enum modus mode = REGINV,
                  const Real sigma = 0.0,
                  const concepts::Vector<Real>* start = 0,
                  const bool schur = false);
    virtual ~ArPackNonSymm();
    /** Returns an array with the real parts of the eigenvalues in the first
        \c kmax entries and the imaginary parts of the eigenvalues in the
        second \c kmax entries
    */
    virtual const concepts::Array<Real>& getEV();
    /** Returns an array with the eigenfunctions. If eigenvalue k is real,
        entry k contains the (real) eigenfunction. If eigenvalues k and k+1
        are complex conjugate, entry k contains the real part of and entry k+1
        contains the imaginary part of eigenfunction k. Eigenfunction k+1 is
        the complex conjugate of eigenfunction k.
    */
    virtual const concepts::Array<concepts::Vector<Real>*>& getEF();
    /// Returns the number of iterations
    virtual uint iterations() const { return (uint)iter_; }
    /// Returns the number of converged eigenpairs
    virtual uint converged() const { return (uint)k_conv_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Operator OP as described in \c modus
    concepts::Operator<Real>& OP_;
    /// Stiffness matrix
    concepts::Operator<Real>& A_;
    /// Operator B as descirbed in \c modus
    concepts::Operator<Real>& B_;
    /// Number of eigenpairs to be computed
    int kmax_;
    /// Maximum number of Arnoldi iterations allowed
    int maxiter_;
    /// Convergence tolerance for the eigenpairs
    Real tol_;
    int numop_, numopb_, numreo_, k_conv_, iter_;
    /// Storage space for eigenvalues
    concepts::Array<Real> eigenvalues_;
    /// Storage space for eigenvectors
    concepts::Array<Real> eigenvectors_;
    /// References into storage for eigenvectors
    concepts::Array<concepts::Vector<Real>*> ev_;
    /// Sort of eigenvalues to compute
    enum which target_;
    /// Mode in which ARPACK should be used
    enum modus mode_;
    /// Shift for the shift-invert modes, real part
    const Real sigmaR_;
    /// Shift for the shift-invert modes, imaginary part
    const Real sigmaI_;
    /// Starting vector, if not given ARPACK invents one
    const concepts::Vector<Real>* start_;
    /// Calculate vectors instead of eigenvectors
    const bool schur_;
    /// Eigenpairs computed?
    bool computed_;
    /// Compute eigenpairs
    void compute_();
  };

} // namespace eigensolver

#endif // ARPACKdsaupd_hh
