// ARPACK eigensolver using dsaupd

#include <cstring>
#include "ARPACKdsaupd.hh"
#include "basics/fortranException.hh"
#include "operator/compositions.hh"
#include "toolbox/array.hh"

// debugging
#include "basics/debug.hh"

#define ARPACKcompute_D 0
#define ARPACKdestr_D 0

#if defined(__hpux) || defined(_IBMR2)
  #define FORTRAN(X) X
#else
  #define FORTRAN(X) X ## _
#endif

#ifdef __cplusplus
extern "C" {
#endif
  extern void FORTRAN(dsaupd)(int& IDO, const char& BMAT, const int& N,
			      const char WHICH[2], const int& NEV,
			      const double& TOL, const double* RESID,
			      const int& NCV, const double* V, const int& LDV,
			      int IPARAM[11], int IPNTR[11],
			      const double* WORKD, const double* WORKL,
			      const int& LWORKL, int& INFO);
  extern void FORTRAN(dseupd)(const int& RVEC, const char& HOWMNY,
			      const int* SELECT, const double* D,
			      const double* Z, const int& LDZ,
			      const double& SIGMA,
			      const char& BMAT, const int& N,
			      const char WHICH[2], const int& NEV,
			      const double& TOL, const double* RESID,
			      const int& NCV, const double* V, const int& LDV,
			      int IPARAM[11], int IPNTR[11],
			      const double* WORKD, const double* WORKL,
			      const int& LWORKL, int& INFO);
#ifdef __cplusplus
}
#endif

namespace eigensolver {

  // **************************************************************** ArPack **

  ArPack::ArPack(concepts::Operator<Real>& OP, concepts::Operator<Real>& A,
		 concepts::Operator<Real>& B,
		 const int kmax, const Real tol, const int maxiter,
		 enum which target, enum modus mode, const Real sigma,
		 const concepts::Vector<Real>* start) :
    OP_(OP), A_(A), B_(B), kmax_(kmax), maxiter_(maxiter), tol_(tol),
    numop_(0), numopb_(0), numreo_(0), k_conv_(0), iter_(0), eigenvalues_(0),
    eigenvectors_(0), ev_(0), target_(target), mode_(mode), sigma_(sigma),
    start_(start), computed_(false) {
    conceptsAssert(OP_.spaceX().dim() == B_.spaceX().dim(),
		   concepts::Assertion());
    conceptsAssert(OP_.spaceX().dim() == A_.spaceX().dim(),
		   concepts::Assertion());
    conceptsAssert(OP_.spaceX().dim() == OP_.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert(B_.spaceX().dim() == B_.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert(A_.spaceX().dim() == A_.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert3(kmax_ > 0, concepts::Assertion(), "kmax = " << kmax);
    conceptsAssert3((uint)kmax_ < OP_.spaceX().dim(),
		    concepts::Assertion(), "kmax = " << kmax << ", dim = "
		    << OP_.spaceX().dim());
  }

  ArPack::~ArPack() {
    DEBUGL(ARPACKdestr_D, "start.");
    for (uint i = 0; i < ev_.size(); ++i) delete ev_[i];
    DEBUGL(ARPACKdestr_D, "done.");
  }

  const concepts::Array<Real>& ArPack::getEV() {
    if (!computed_) {
      compute_();
    }
    return eigenvalues_;
  }

  const concepts::Array<concepts::Vector<Real>*>& ArPack::getEF() {
    if (!computed_) {
      compute_(); // eigenvectors_ is filled
    }
    return ev_;
  }

  std::ostream& ArPack::info(std::ostream& os) const {
    os << "ArPack(";
    switch (mode_) {
    case NORMAL: os << "normal mode"; break;
    case REGINV: os << "regular inverse mode"; break;
    case SHIFTINV: os << "shift-invert mode"; break;
    case BUCKLING: os << "Buckling mode"; break;
    case CAYLEY: os << "Cayley mode"; break;
    }
    os << '(' << mode_ << "), " << (computed_ ? "" : "not yet ")
       << "computed ";
    switch (target_) {
    case LA: os << "largest (algebraic) eigenvalues"; break;
    case SA: os << "smallest (algebraic) eigenvalues"; break;
    case LM: os << "largest (in magnitude) eigenvalues"; break;
    case SM: os << "smallest (in magnitude) eigenvalues"; break;
    case BE: os << "eigenvalues from both ends"; break;
    }
    return os << ", sigma = " << sigma_ << ", n = " << OP_.spaceX().dim()
	      << ", tol = " << tol_ << ", conv. eigenpairs = " << k_conv_
	      << "(max = " << kmax_ << "), Arnoldi iter = "
	      << iter_ << "(max = " << maxiter_ << "), " << numop_
	      << " OP*x ops., " << numopb_ << " B*x ops., "
	      << numreo_ << " steps of re-orth.)";
  }

  void ArPack::compute_() {
    int ido = 0;
    const char bmat = 'G';
    const int n = OP_.spaceX().dim();
    char whichEV[2];
    switch (target_) {
    case LA: strncpy(whichEV, "LA", 2); break;
    case SA: strncpy(whichEV, "SA", 2); break;
    case LM: strncpy(whichEV, "LM", 2); break;
    case SM: strncpy(whichEV, "SM", 2); break;
    case BE: strncpy(whichEV, "BE", 2); break;
    }
    const int ncv = 2*kmax_ <= n ? 2*kmax_ : n; // suggestion by manual: >= 2*kmax
    const int lworkl = ncv*(8+ncv);
    concepts::Array<Real> resid(n, 0), workd(3*n, 0),
      workl(lworkl, 0);
    eigenvectors_.resize(ncv*n);
    const int ldv = n;
    concepts::Array<int> iparam(11, 0), ipntr(11, 0);
    iparam[0] = 1; // ishift = 1: use exact shifts
    iparam[1] = 0; // no longer referenced
    iparam[2] = maxiter_; // max number of Arnoldi iterations allowed
    iparam[3] = 1; // blocksize, only 1 is supported
    iparam[4] = 0; // number of converged eigenvalues
    iparam[5] = 0; // no longer referenced
    iparam[6] = mode_; // mode
    iparam[7] = 0; // not needed for ishift = 1
    int info = 0;

    if (start_) {
      info = 1;
      std::memcpy(resid, *start_, n*sizeof(Real));
    }

    DEBUGL(ARPACKcompute_D, "bmat = " << bmat << ", n = " << n << ", which = "
	   << whichEV[0] << whichEV[1] << ", nev = " << kmax_ << ", tol = "
	   << tol_ << ", ncv = " << ncv << ", iparam = " << iparam);
    while (ido != 99) {
      FORTRAN(dsaupd)(ido, bmat, n, whichEV, kmax_, tol_, (Real*)resid, ncv,
		      (Real*)eigenvectors_, ldv, (int*)iparam, (int*)ipntr,
		      (Real*)workd, (Real*)workl, lworkl, info);
      DEBUGL(ARPACKcompute_D, "ido = " << ido << ", info = " << info
	     << ", ipntr = " << ipntr << ", iparam = " << iparam);
      switch (ido) {
      case -1:  {
	concepts::Vector<Real> y(B_.spaceX(), (Real*)workd+ipntr[1]-1),
	  x(B_.spaceY(), (Real*)workd+ipntr[0]-1);
	DEBUGL(ARPACKcompute_D, "y = OP*x -- " << ipntr[0] << ", "
	       << ipntr[1] << ", x = " << x);
	if (mode_ == REGINV) {
	  A_(x, y);
	  x = y;
	}
	if (mode_ == SHIFTINV) {
	  B_(x, y);
	  x = y;
	}
 	OP_(x, y);
	DEBUGL(ARPACKcompute_D, "x = " << x << ", y = " << y);
      } break;
      case 1: {
	concepts::Vector<Real> y(B_.spaceX(), (Real*)workd+ipntr[1]-1),
	  x(B_.spaceY(), (Real*)workd+ipntr[(mode_ == SHIFTINV) ? 2 : 0]-1);
	DEBUGL(ARPACKcompute_D, "y = OP*x -- " << ipntr[0] << ", "
	       << ipntr[1] << ", x = " << x);
	if (mode_ == REGINV) {
	  A_(x, y);
	  x = y;
	}
 	OP_(x, y);
	DEBUGL(ARPACKcompute_D, "x = " << x << ", y = " << y);
      } break;
      case 2: {
	DEBUGL(ARPACKcompute_D, "y = B*x -- " << ipntr[0] << ", " << ipntr[1]);
	concepts::Vector<Real> y(B_.spaceX(), (Real*)workd+ipntr[1]-1),
	  x(B_.spaceY(), (Real*)workd+ipntr[0]-1);
	B_(x, y);
	DEBUGL(ARPACKcompute_D, "x = " << x << ", y = " << y);
      } break;
      case 99: break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
    }
    DEBUGL(ARPACKcompute_D, "resid = " << resid << ", iparam = " << iparam
	   << ", ipntr = " << ipntr << ", info = " << info << ", workl = "
	   << workl);
    iter_ = iparam[2];
    k_conv_ = iparam[4];
    numop_ = iparam[8];
    numopb_ = iparam[9];
    numreo_ = iparam[10];
    switch (info) {
    case 0: break; // everything ok
    case 1: break; // not all eigenpairs have converged
    default: throw conceptsException(concepts::FortranException
				     (info, "dsaupd"));
    }

    concepts::Array<int> select(ncv, true);
    eigenvalues_.resize(kmax_);
    FORTRAN(dseupd)(1, 'A', select, (Real*)eigenvalues_, (Real*)eigenvectors_,
		    ldv, sigma_,
		    bmat, n, whichEV, kmax_, tol_, (Real*)resid, ncv,
		    (Real*)eigenvectors_, ldv, (int*)iparam, (int*)ipntr,
		    (Real*)workd, (Real*)workl, lworkl, info);
    DEBUGL(ARPACKcompute_D, "info = " << info);
    if (info != 0) throw conceptsException(concepts::FortranException
					   (info, "dseupd"));
    DEBUGL(ARPACKcompute_D, "eigenvalues = " << eigenvalues_
	   << ", eigenvectors = " << eigenvectors_);
    ev_.resize(kmax_);
    for (int i = 0; i < kmax_; ++i)
      ev_[i] = new concepts::Vector<Real>(OP_.spaceX(), eigenvectors_+i*n);
    computed_ = true;
  }

} // namespace eigensolver
