// fabric for eigensolver ARPACK

#ifndef arpackFabric_hh
#define arpackFabric_hh

#include <memory>
#include "ARPACKdsaupd.hh"
#include "solverFabric.hh"

namespace eigensolver {

  // ********************************************************** ArPackFabric **

  /** Fabric class for ARPACK eigenvalue solver for a constrained problem.
      @author Philipp Frauenfelder, 2002
      @test test::GolubExample
      @test test::GolubExampleSum
      @test test::MaxwellTransmissionEVP
  */
  class ArPackFabric : public SolverFabric {
  public:
    /** Constructor.
	@param kmax Number of eigenpairs to be computed
	@param tol Convergence tolerance for the eigenpairs. The default
	           value 0.0 is replaced by \c DLAMCH('EPS') from LAPACK.
	@param maxiter Maximum number of Arnoldi iterations allowed
	@param target What sort of eigenvalues to compute
	@param mode Mode in which ARPACK should be used
	@param sigma Shift for the shift-invert, Buckling or Cayley mode
	@param start Starting vector, if not given ARPACK invents one
    */
    ArPackFabric(int kmax = 1, Real tol = 0.0, const int maxiter = 300,
		 enum ArPack::which target = ArPack::SM,
		 enum ArPack::modus mode = ArPack::REGINV,
		 const Real sigma = 0.0,
		 const concepts::Vector<Real>* start = 0) :
      solver_(0), OP_(0), B_(0), tmp1_(0), tmp2_(0), kmax_(kmax),
      maxiter_(maxiter), tol_(tol), target_(target), mode_(mode),
      sigma_(sigma), start_(start) {}
    virtual ~ArPackFabric();
    virtual ArPack& operator()(concepts::Operator<Real>& stiff,
			       concepts::Operator<Real>& mass);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    std::auto_ptr<ArPack> solver_;
    std::auto_ptr<concepts::Operator<Real> > OP_, B_, tmp1_, tmp2_;
    /// Number of eigenpairs to be computed
    int kmax_;
    /// Maximum number of Arnoldi iterations allowed
    int maxiter_;
    /// Convergence tolerance for the eigenpairs
    Real tol_;
    /// What sort of eigenvalues to compute
    enum ArPack::which target_;
    /// Mode in which ARPACK should be used
    enum ArPack::modus mode_;
    /// Shift for the shift-invert, Buckling or Cayley mode
    const Real sigma_;
    /// Starting vector, if not given ARPACK invents one
    const concepts::Vector<Real>* start_;
    ArPackFabric(const ArPackFabric&);

    void normalMode_(concepts::Operator<Real>& stiff,
		    concepts::Operator<Real>& mass);
    void reginvMode_(concepts::Operator<Real>& stiff,
		    concepts::Operator<Real>& mass);
    void shiftinvMode_(concepts::Operator<Real>& stiff,
		      concepts::Operator<Real>& mass);
    void bucklingMode_(concepts::Operator<Real>& stiff,
		      concepts::Operator<Real>& mass);
    void cayleyMode_(concepts::Operator<Real>& stiff,
		    concepts::Operator<Real>& mass);
  };

} // namespace eigensolver

#endif // arpackFabric_hh
