// jdbsym eigensolver

#ifndef jdbsym_hh
#define jdbsym_hh

#include "eigens.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class Operator;
  template<typename F>
  class SolverFabric;
} // namespace concepts

namespace eigensolver {

  // **************************************************************** JdbSym **

  /** Eigenvalue solver using JDBSYM.

      From the documentation of JDBSYM:
      JDBSYM is an implementation of the Jacobi-Davidson method
      optimized for symmetric eigenvalue problems. It solves
      eigenproblems of the form \f$ A x = \lambda x \f$ and
      \f$ A x = \lambda B x \f$ with or without preconditioning, where
      A is symmetric and B is symmetric positive definite. The
      implementation supports blocking.

      @see <a href="http://www.inf.ethz.ch/personal/geus/software.html">Software
      page of Roman Geus</a>
      @test test::GolubExample
      @test test::GolubExampleSum
      @test test::MaxwellTransmissionEVP
      @author Norbert Fernandes, 2002
  */
  class JdbSym : public EigenSolver {
  public:
    /** Constructor. Initializes the JdbSym class. As parameter it
        takes the few most important variables that are needed for the
        Jacobi-Davidson algorithm. The rest of the parameters are
        given as default and work in most cases. The class then solves
        the problem \f$ A x = \lambda B x \f$ where A, B are the
        stiffness and the mass matrices respectively.

	The computation is only perfomed if \c getEV or \c getEF is
	called.

	@param stiff Stiffness matrix A
        @param mass Mass matrix B
	@param fabric Solver fabric for a linear solver (preconditioner for
	              the shifted stiffness matrix)
        @param tol Convergence tolerance for the eigenpairs. For a pair
	           \f$ (\lambda, u) \f$ convergence is defined by
		   \f$ \| A u - \lambda B u \| < tol \f$
	@param maxit Maximal number of iterations
        @param tau Target value of Jacobi-Davidson algorithm. The code will
	           find the \c kmax eigenvalues closest to \c tau.
        @param jdtype Type of solver required. An older solver and a newer
	              solver are possible to use.
        @param kmax Number of eigenpairs to be computed
	@param start Starting vectors. Used to build the initial search
	             subspace
    */
    JdbSym(concepts::Operator<Real>& stiff, concepts::Operator<Real>& mass,
	   Real tol, uint maxit = 150, Real tau = 0.0, uint jdtype = 1,
	   uint kmax = 1, concepts::SolverFabric<Real>* fabric = 0,
	   const concepts::Array<concepts::Vector<Real>*>* start = 0);
    virtual ~JdbSym();
    virtual const concepts::Array<Real>& getEV();
    virtual const concepts::Array<concepts::Vector<Real>*>& getEF();
    virtual uint iterations() const { return (uint)it_; }
    virtual uint converged() const { return (uint)k_conv_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /** The real compute routine. Fills \c lambda_ and \c ev_
	@see computed_
     */
    void compute_();
    /// If false, the code will have to compute, otherwise not
    bool computed_;
    /// Stiffness matrix (ie. discretazation of spatial problem)
    concepts::Operator<Real>& stiff_;
    /** Mass matrix (ie. discretization of the right hand side of
	\f$ A u = \lambda u \f$
    */
    concepts::Operator<Real>& mass_;
    /// Tolerance
    Real tol_;
    /// Maximal number of iterations
    uint maxit_;
    /// Target value for the eigenvalues
    Real tau_;
    /// Type of eigen value solver
    const uint jdtype_;
    /// Number of eigenvalues
    uint kmax_;
    /// Actual number of iterations
    int it_;
    /// Number of converged eigenpais
    int k_conv_;

    /// Storage space for eigenvalues
    concepts::Array<Real> lambda_;
    /// Storage space for eigenvectors
    concepts::Array<Real> eigenvectors_;
    /// References into storage for eigenvectors
    concepts::Array<concepts::Vector<Real>*> ev_;

    /// Statistics, counts number of evaluations of matrix vector product
    uint stiffCounter_, massCounter_;

    concepts::SolverFabric<Real>* fabric_;
    const concepts::Array<concepts::Vector<Real>*>* start_;

    /// Lock for the eigen value solver
    static int jdbsymLock_;
  };

}

#endif // jdbsym_hh
