// testing constraint EVP

#include "constrained.hh"
#include "operator/denseMatrix.hh"
#ifdef HAS_ARPACK
#  include "eigensolver/arpackFabric.hh"
#endif
#ifdef HAS_JDBSym
#  include "eigensolver/jdbsymFabric.hh"
#endif
#include "eigensolver/constrained.hh"
#include "eigensolver/inexactFabric.hh"
#include "basics/testsuite.hh"

using namespace concepts;

namespace test {

  // ******************************************************** ConstrainedEVP **

  void ConstrainedEVP::run() {
#ifdef HAS_ARPACK
    testARPACKregularInv();
    testARPACKshiftInv();
#endif
#ifdef HAS_JDBSym
    testJDBSym();
#endif
    testInexactInv();
  }

#ifdef HAS_ARPACK
  void ConstrainedEVP::testARPACKregularInv() {
    eigensolver::ArPackFabric fabric(kmax_-1);
    eigensolver::Constrained solver(*a, b, c, fabric);
    testResults(solver.getEV());
    _test(solver.converged() == kmax_-1);
  }

  void ConstrainedEVP::testARPACKshiftInv() {
    eigensolver::ArPackFabric fabric(kmax_-1, 0, 300, eigensolver::ArPack::LA,
				     eigensolver::ArPack::SHIFTINV,
				     exact_[0]*0.9);
    eigensolver::Constrained solver(*a, b, c, fabric);
    testResults(solver.getEV());
    _test(solver.converged() == kmax_-1);
  }
#endif

#ifdef HAS_JDBSym
  void ConstrainedEVP::testJDBSym() {
    eigensolver::JdbSymFabric fabric(1e-10, 150, 0, 1, kmax_);
    eigensolver::Constrained solver(*a, b, c, fabric);
    testResults(solver.getEV());
    _test(solver.converged() == kmax_);
  }
#endif

  void ConstrainedEVP::testInexactInv() {
    Vector<Real> guess(gspc);
    guess = 1;
    eigensolver::InexactInvFabric fabric(guess, 0.5, 100, 1e-10, 100);
    eigensolver::Constrained solver(*a, b, c, fabric);
    testResults(solver.getEV());
    _test(solver.converged() == 1);
  }

  void ConstrainedEVP::testResults(const concepts::Array<Real>& res) {
//     std::cout << res << std::endl;
    for (uint i = 0; i < res.size(); ++i) {
//       if (std::fabs(res[i] - exact_[i]) > 1e-3)
// 	std::cout << i << ": " << res[i] << " != " << exact_[i] << std::endl;
      _numtesttol(res[i], exact_[i], 1e-3);
    }
  }

  // ********************************************************** GolubExample **

  const concepts::Real GolubExample::exact[4] =
    { 0.170039264847579, 1.23788202328080, 4.9176011926100, 9.2744775192616 };

  GolubExample::GolubExample() : ConstrainedEVP(6, 4, 4, 4) {
    exact_ = exact;
    a.reset(new concepts::SparseMatrix<Real>(spc, spc));
    concepts::SparseMatrix<Real>& A =
      *(dynamic_cast<concepts::SparseMatrix<Real>*>(a.get()));
    A(0,0) = 1;  A(0,1) = -1;
    A(1,0) = -1; A(1,1) = 2; A(1,2) = -1;
    A(2,1) = -1; A(2,2) = 2; A(2,3) = -1;
    A(3,2) = -1; A(3,3) = 2; A(3,4) = -1;
    A(4,3) = -1; A(4,4) = 2; A(4,5) = -1;
    A(5,4) = -1; A(5,5) = 2;

    b(0,0) = 6;
    b(0,1) = b(1,0) = b(1,1) = 5;
    b(0,2) = b(2,0) = b(2,2) = b(1,2) = b(2,1) = 4;
    b(3,3) = b(3,2) = b(3,1) = b(3,0) = b(2,3) = b(1,3) = b(0,3) = 3;
    b(4,4) = b(4,3) = b(4,2) = b(4,1) = b(4,0) = b(3,4) = b(2,4) = b(1,4)
      = b(0,4) = 2;
    b(5,5) = b(5,4) = b(5,3) = b(5,2) = b(5,1) = b(5,0) = b(4,5) = b(3,5)
      = b(2,5) = b(1,5) = b(0,5) = 1;

    for (uint i = 0; i < spc.dim(); i += 2) {
      c(i,0) = c(i,1) = 1; c(i,2) = 8; c(i,3) = 5;
      c(i+1,0) = c(i+1,3) = 1; c(i+1,1) = -1; c(i+1,2) = 2;
    }
  }

  // ******************************************************* GolubExampleSum **

  const concepts::Real GolubExampleSum::exact[4] =
    { 0.170039264847579, 1.23788202328080, 4.9176011926100, 9.2744775192616 };

  GolubExampleSum::GolubExampleSum() : ConstrainedEVP(6, 4, 4, 4),
				       a1(spc, spc), a2(spc, spc) {
    exact_ = exact;
    a.reset(new concepts::LiCo<Real>(a1, a2));
    a1(0,0) = 1;  a1(0,1) = -1;
    a1(1,0) = -1; a1(1,1) = 2; a1(1,2) = -1;
    a1(2,1) = -1; a1(2,2) = 2; a1(2,3) = -1;
    a2(3,2) = -1; a2(3,3) = 2; a2(3,4) = -1;
    a2(4,3) = -1; a2(4,4) = 2; a2(4,5) = -1;
    a2(5,4) = -1; a2(5,5) = 2;

    b(0,0) = 6;
    b(0,1) = b(1,0) = b(1,1) = 5;
    b(0,2) = b(2,0) = b(2,2) = b(1,2) = b(2,1) = 4;
    b(3,3) = b(3,2) = b(3,1) = b(3,0) = b(2,3) = b(1,3) = b(0,3) = 3;
    b(4,4) = b(4,3) = b(4,2) = b(4,1) = b(4,0) = b(3,4) = b(2,4) = b(1,4)
      = b(0,4) = 2;
    b(5,5) = b(5,4) = b(5,3) = b(5,2) = b(5,1) = b(5,0) = b(4,5) = b(3,5)
      = b(2,5) = b(1,5) = b(0,5) = 1;

    for (uint i = 0; i < spc.dim(); i += 2) {
      c(i,0) = c(i,1) = 1; c(i,2) = 8; c(i,3) = 5;
      c(i+1,0) = c(i+1,3) = 1; c(i+1,1) = -1; c(i+1,2) = 2;
    }
  }

  // ************************************************ MaxwellTransmissionEVP **

  const concepts::Real MaxwellTransmissionEVP::exact[20] =
    { 3.90138946818128, 5.67614958516675, 6.29319923553879, 6.62453446998068,
      14.4661382447631, 15.841176719629, 17.8055451358582, 20.4388159938776,
      35.7572606890455, 37.024713997502, 45.3839416431495, 49.0315666164476,
      53.7467291774104, 60.7248346280723, 74.7898665707598, 89.9819462970942,
      165.403361036472, 168.607671096364, 174.979664471991, 177.339435309688 };

  MaxwellTransmissionEVP::MaxwellTransmissionEVP() :
    ConstrainedEVP(48, 24, 28, 18) {
    exact_ = exact;
    a.reset(new concepts::SparseMatrix<Real>(spc, spc));
    concepts::SparseMatrix<Real>& A =
      *(dynamic_cast<concepts::SparseMatrix<Real>*>(a.get()));
    matlab_matrix_entry(A, 1, 5, 0.266667);
    matlab_matrix_entry(A, 1, 25, 0.25);
    matlab_matrix_entry(A, 1, 29, -0.116667);
    matlab_matrix_entry(A, 1, 1, 1.2);
    matlab_matrix_entry(A, 1, 6, 0.0833333);
    matlab_matrix_entry(A, 1, 26, -0.75);
    matlab_matrix_entry(A, 1, 30, -0.00555556);
    matlab_matrix_entry(A, 1, 2, -0.7);
    matlab_matrix_entry(A, 1, 27, 0.316667);
    matlab_matrix_entry(A, 1, 3, 0.416667);
    matlab_matrix_entry(A, 1, 4, -0.266667);
    matlab_matrix_entry(A, 1, 28, -0.15);
    matlab_matrix_entry(A, 2, 5, -0.266667);
    matlab_matrix_entry(A, 2, 25, -0.75);
    matlab_matrix_entry(A, 2, 29, -0.05);
    matlab_matrix_entry(A, 2, 1, -0.7);
    matlab_matrix_entry(A, 2, 6, -0.0833333);
    matlab_matrix_entry(A, 2, 26, 0.25);
    matlab_matrix_entry(A, 2, 30, 0.00555556);
    matlab_matrix_entry(A, 2, 2, 1.2);
    matlab_matrix_entry(A, 2, 27, -0.316667);
    matlab_matrix_entry(A, 2, 3, -0.25);
    matlab_matrix_entry(A, 2, 4, 0.266667);
    matlab_matrix_entry(A, 2, 28, 0.15);
    matlab_matrix_entry(A, 3, 5, 0.0833333);
    matlab_matrix_entry(A, 3, 25, -0.116667);
    matlab_matrix_entry(A, 3, 29, -0.0611111);
    matlab_matrix_entry(A, 3, 1, 0.416667);
    matlab_matrix_entry(A, 3, 30, -0.0166667);
    matlab_matrix_entry(A, 3, 6, 0.1);
    matlab_matrix_entry(A, 3, 26, -0.05);
    matlab_matrix_entry(A, 3, 2, -0.25);
    matlab_matrix_entry(A, 3, 27, 0.05);
    matlab_matrix_entry(A, 3, 3, 0.366667);
    matlab_matrix_entry(A, 3, 28, -0.05);
    matlab_matrix_entry(A, 3, 4, -0.0833333);
    matlab_matrix_entry(A, 4, 29, -0.05);
    matlab_matrix_entry(A, 4, 1, -0.266667);
    matlab_matrix_entry(A, 4, 25, -0.15);
    matlab_matrix_entry(A, 4, 5, -0.068254);
    matlab_matrix_entry(A, 4, 30, -0.0166667);
    matlab_matrix_entry(A, 4, 2, 0.266667);
    matlab_matrix_entry(A, 4, 26, 0.15);
    matlab_matrix_entry(A, 4, 6, -0.00555556);
    matlab_matrix_entry(A, 4, 3, -0.0833333);
    matlab_matrix_entry(A, 4, 27, -0.0722222);
    matlab_matrix_entry(A, 4, 28, 0.0722222);
    matlab_matrix_entry(A, 4, 4, 0.234921);
    matlab_matrix_entry(A, 5, 29, 0.05);
    matlab_matrix_entry(A, 5, 1, 0.266667);
    matlab_matrix_entry(A, 5, 25, 0.316667);
    matlab_matrix_entry(A, 5, 5, 0.234921);
    matlab_matrix_entry(A, 5, 30, 0.0166667);
    matlab_matrix_entry(A, 5, 2, -0.266667);
    matlab_matrix_entry(A, 5, 26, -0.316667);
    matlab_matrix_entry(A, 5, 6, 0.0611111);
    matlab_matrix_entry(A, 5, 3, 0.0833333);
    matlab_matrix_entry(A, 5, 27, 0.0722222);
    matlab_matrix_entry(A, 5, 28, -0.0722222);
    matlab_matrix_entry(A, 5, 4, -0.068254);
    matlab_matrix_entry(A, 6, 29, -0.0166667);
    matlab_matrix_entry(A, 6, 1, 0.0833333);
    matlab_matrix_entry(A, 6, 25, -0.00555556);
    matlab_matrix_entry(A, 6, 5, 0.0611111);
    matlab_matrix_entry(A, 6, 2, -0.0833333);
    matlab_matrix_entry(A, 6, 26, 0.00555556);
    matlab_matrix_entry(A, 6, 6, 0.0568254);
    matlab_matrix_entry(A, 6, 27, 0.0166667);
    matlab_matrix_entry(A, 6, 3, 0.1);
    matlab_matrix_entry(A, 6, 28, -0.0166667);
    matlab_matrix_entry(A, 6, 4, -0.00555556);
    matlab_matrix_entry(A, 7, 9, 0.266667);
    matlab_matrix_entry(A, 7, 33, 0.116667);
    matlab_matrix_entry(A, 7, 10, -0.266667);
    matlab_matrix_entry(A, 7, 34, 0.15);
    matlab_matrix_entry(A, 7, 31, -0.25);
    matlab_matrix_entry(A, 7, 35, -0.316667);
    matlab_matrix_entry(A, 7, 7, 1.2);
    matlab_matrix_entry(A, 7, 11, 0.416667);
    matlab_matrix_entry(A, 7, 12, 0.0833333);
    matlab_matrix_entry(A, 7, 32, 0.75);
    matlab_matrix_entry(A, 7, 36, 0.00555556);
    matlab_matrix_entry(A, 7, 8, -0.7);
    matlab_matrix_entry(A, 8, 9, -0.266667);
    matlab_matrix_entry(A, 8, 33, 0.05);
    matlab_matrix_entry(A, 8, 10, 0.266667);
    matlab_matrix_entry(A, 8, 34, -0.15);
    matlab_matrix_entry(A, 8, 31, 0.75);
    matlab_matrix_entry(A, 8, 35, 0.316667);
    matlab_matrix_entry(A, 8, 7, -0.7);
    matlab_matrix_entry(A, 8, 11, -0.25);
    matlab_matrix_entry(A, 8, 12, -0.0833333);
    matlab_matrix_entry(A, 8, 32, -0.25);
    matlab_matrix_entry(A, 8, 36, -0.00555556);
    matlab_matrix_entry(A, 8, 8, 1.2);
    matlab_matrix_entry(A, 9, 33, -0.05);
    matlab_matrix_entry(A, 9, 9, 0.234921);
    matlab_matrix_entry(A, 9, 34, 0.0722222);
    matlab_matrix_entry(A, 9, 10, -0.068254);
    matlab_matrix_entry(A, 9, 7, 0.266667);
    matlab_matrix_entry(A, 9, 11, 0.0833333);
    matlab_matrix_entry(A, 9, 31, -0.316667);
    matlab_matrix_entry(A, 9, 35, -0.0722222);
    matlab_matrix_entry(A, 9, 36, -0.0166667);
    matlab_matrix_entry(A, 9, 8, -0.266667);
    matlab_matrix_entry(A, 9, 32, 0.316667);
    matlab_matrix_entry(A, 9, 12, 0.0611111);
    matlab_matrix_entry(A, 10, 33, 0.05);
    matlab_matrix_entry(A, 10, 9, -0.068254);
    matlab_matrix_entry(A, 10, 34, -0.0722222);
    matlab_matrix_entry(A, 10, 10, 0.234921);
    matlab_matrix_entry(A, 10, 7, -0.266667);
    matlab_matrix_entry(A, 10, 11, -0.0833333);
    matlab_matrix_entry(A, 10, 31, 0.15);
    matlab_matrix_entry(A, 10, 35, 0.0722222);
    matlab_matrix_entry(A, 10, 36, 0.0166667);
    matlab_matrix_entry(A, 10, 8, 0.266667);
    matlab_matrix_entry(A, 10, 32, -0.15);
    matlab_matrix_entry(A, 10, 12, -0.00555556);
    matlab_matrix_entry(A, 11, 9, 0.0833333);
    matlab_matrix_entry(A, 11, 33, 0.0611111);
    matlab_matrix_entry(A, 11, 34, 0.05);
    matlab_matrix_entry(A, 11, 10, -0.0833333);
    matlab_matrix_entry(A, 11, 35, -0.05);
    matlab_matrix_entry(A, 11, 31, 0.116667);
    matlab_matrix_entry(A, 11, 7, 0.416667);
    matlab_matrix_entry(A, 11, 11, 0.366667);
    matlab_matrix_entry(A, 11, 36, 0.0166667);
    matlab_matrix_entry(A, 11, 12, 0.1);
    matlab_matrix_entry(A, 11, 32, 0.05);
    matlab_matrix_entry(A, 11, 8, -0.25);
    matlab_matrix_entry(A, 12, 33, 0.0166667);
    matlab_matrix_entry(A, 12, 9, 0.0611111);
    matlab_matrix_entry(A, 12, 34, 0.0166667);
    matlab_matrix_entry(A, 12, 10, -0.00555556);
    matlab_matrix_entry(A, 12, 35, -0.0166667);
    matlab_matrix_entry(A, 12, 7, 0.0833333);
    matlab_matrix_entry(A, 12, 11, 0.1);
    matlab_matrix_entry(A, 12, 31, 0.00555556);
    matlab_matrix_entry(A, 12, 8, -0.0833333);
    matlab_matrix_entry(A, 12, 32, -0.00555556);
    matlab_matrix_entry(A, 12, 12, 0.0568254);
    matlab_matrix_entry(A, 13, 17, 0.266667);
    matlab_matrix_entry(A, 13, 37, 0.25);
    matlab_matrix_entry(A, 13, 41, -0.116667);
    matlab_matrix_entry(A, 13, 13, 1.2);
    matlab_matrix_entry(A, 13, 18, 0.0833333);
    matlab_matrix_entry(A, 13, 38, -0.75);
    matlab_matrix_entry(A, 13, 42, -0.00555556);
    matlab_matrix_entry(A, 13, 14, -0.7);
    matlab_matrix_entry(A, 13, 39, 0.316667);
    matlab_matrix_entry(A, 13, 15, 0.416667);
    matlab_matrix_entry(A, 13, 16, -0.266667);
    matlab_matrix_entry(A, 13, 40, -0.15);
    matlab_matrix_entry(A, 14, 17, -0.266667);
    matlab_matrix_entry(A, 14, 37, -0.75);
    matlab_matrix_entry(A, 14, 41, -0.05);
    matlab_matrix_entry(A, 14, 13, -0.7);
    matlab_matrix_entry(A, 14, 18, -0.0833333);
    matlab_matrix_entry(A, 14, 38, 0.25);
    matlab_matrix_entry(A, 14, 42, 0.00555556);
    matlab_matrix_entry(A, 14, 14, 1.2);
    matlab_matrix_entry(A, 14, 39, -0.316667);
    matlab_matrix_entry(A, 14, 15, -0.25);
    matlab_matrix_entry(A, 14, 16, 0.266667);
    matlab_matrix_entry(A, 14, 40, 0.15);
    matlab_matrix_entry(A, 15, 17, 0.0833333);
    matlab_matrix_entry(A, 15, 37, -0.116667);
    matlab_matrix_entry(A, 15, 41, -0.0611111);
    matlab_matrix_entry(A, 15, 13, 0.416667);
    matlab_matrix_entry(A, 15, 42, -0.0166667);
    matlab_matrix_entry(A, 15, 18, 0.1);
    matlab_matrix_entry(A, 15, 38, -0.05);
    matlab_matrix_entry(A, 15, 14, -0.25);
    matlab_matrix_entry(A, 15, 39, 0.05);
    matlab_matrix_entry(A, 15, 15, 0.366667);
    matlab_matrix_entry(A, 15, 40, -0.05);
    matlab_matrix_entry(A, 15, 16, -0.0833333);
    matlab_matrix_entry(A, 16, 41, -0.05);
    matlab_matrix_entry(A, 16, 13, -0.266667);
    matlab_matrix_entry(A, 16, 37, -0.15);
    matlab_matrix_entry(A, 16, 17, -0.068254);
    matlab_matrix_entry(A, 16, 42, -0.0166667);
    matlab_matrix_entry(A, 16, 14, 0.266667);
    matlab_matrix_entry(A, 16, 38, 0.15);
    matlab_matrix_entry(A, 16, 18, -0.00555556);
    matlab_matrix_entry(A, 16, 15, -0.0833333);
    matlab_matrix_entry(A, 16, 39, -0.0722222);
    matlab_matrix_entry(A, 16, 40, 0.0722222);
    matlab_matrix_entry(A, 16, 16, 0.234921);
    matlab_matrix_entry(A, 17, 41, 0.05);
    matlab_matrix_entry(A, 17, 13, 0.266667);
    matlab_matrix_entry(A, 17, 37, 0.316667);
    matlab_matrix_entry(A, 17, 17, 0.234921);
    matlab_matrix_entry(A, 17, 42, 0.0166667);
    matlab_matrix_entry(A, 17, 14, -0.266667);
    matlab_matrix_entry(A, 17, 38, -0.316667);
    matlab_matrix_entry(A, 17, 18, 0.0611111);
    matlab_matrix_entry(A, 17, 15, 0.0833333);
    matlab_matrix_entry(A, 17, 39, 0.0722222);
    matlab_matrix_entry(A, 17, 40, -0.0722222);
    matlab_matrix_entry(A, 17, 16, -0.068254);
    matlab_matrix_entry(A, 18, 41, -0.0166667);
    matlab_matrix_entry(A, 18, 13, 0.0833333);
    matlab_matrix_entry(A, 18, 37, -0.00555556);
    matlab_matrix_entry(A, 18, 17, 0.0611111);
    matlab_matrix_entry(A, 18, 14, -0.0833333);
    matlab_matrix_entry(A, 18, 38, 0.00555556);
    matlab_matrix_entry(A, 18, 18, 0.0568254);
    matlab_matrix_entry(A, 18, 39, 0.0166667);
    matlab_matrix_entry(A, 18, 15, 0.1);
    matlab_matrix_entry(A, 18, 40, -0.0166667);
    matlab_matrix_entry(A, 18, 16, -0.00555556);
    matlab_matrix_entry(A, 19, 21, 0.266667);
    matlab_matrix_entry(A, 19, 45, 0.116667);
    matlab_matrix_entry(A, 19, 22, -0.266667);
    matlab_matrix_entry(A, 19, 46, 0.15);
    matlab_matrix_entry(A, 19, 43, -0.25);
    matlab_matrix_entry(A, 19, 47, -0.316667);
    matlab_matrix_entry(A, 19, 19, 1.2);
    matlab_matrix_entry(A, 19, 23, 0.416667);
    matlab_matrix_entry(A, 19, 24, 0.0833333);
    matlab_matrix_entry(A, 19, 44, 0.75);
    matlab_matrix_entry(A, 19, 48, 0.00555556);
    matlab_matrix_entry(A, 19, 20, -0.7);
    matlab_matrix_entry(A, 20, 21, -0.266667);
    matlab_matrix_entry(A, 20, 45, 0.05);
    matlab_matrix_entry(A, 20, 22, 0.266667);
    matlab_matrix_entry(A, 20, 46, -0.15);
    matlab_matrix_entry(A, 20, 43, 0.75);
    matlab_matrix_entry(A, 20, 47, 0.316667);
    matlab_matrix_entry(A, 20, 19, -0.7);
    matlab_matrix_entry(A, 20, 23, -0.25);
    matlab_matrix_entry(A, 20, 24, -0.0833333);
    matlab_matrix_entry(A, 20, 44, -0.25);
    matlab_matrix_entry(A, 20, 48, -0.00555556);
    matlab_matrix_entry(A, 20, 20, 1.2);
    matlab_matrix_entry(A, 21, 45, -0.05);
    matlab_matrix_entry(A, 21, 21, 0.234921);
    matlab_matrix_entry(A, 21, 46, 0.0722222);
    matlab_matrix_entry(A, 21, 22, -0.068254);
    matlab_matrix_entry(A, 21, 19, 0.266667);
    matlab_matrix_entry(A, 21, 23, 0.0833333);
    matlab_matrix_entry(A, 21, 43, -0.316667);
    matlab_matrix_entry(A, 21, 47, -0.0722222);
    matlab_matrix_entry(A, 21, 48, -0.0166667);
    matlab_matrix_entry(A, 21, 20, -0.266667);
    matlab_matrix_entry(A, 21, 44, 0.316667);
    matlab_matrix_entry(A, 21, 24, 0.0611111);
    matlab_matrix_entry(A, 22, 45, 0.05);
    matlab_matrix_entry(A, 22, 21, -0.068254);
    matlab_matrix_entry(A, 22, 46, -0.0722222);
    matlab_matrix_entry(A, 22, 22, 0.234921);
    matlab_matrix_entry(A, 22, 19, -0.266667);
    matlab_matrix_entry(A, 22, 23, -0.0833333);
    matlab_matrix_entry(A, 22, 43, 0.15);
    matlab_matrix_entry(A, 22, 47, 0.0722222);
    matlab_matrix_entry(A, 22, 48, 0.0166667);
    matlab_matrix_entry(A, 22, 20, 0.266667);
    matlab_matrix_entry(A, 22, 44, -0.15);
    matlab_matrix_entry(A, 22, 24, -0.00555556);
    matlab_matrix_entry(A, 23, 21, 0.0833333);
    matlab_matrix_entry(A, 23, 45, 0.0611111);
    matlab_matrix_entry(A, 23, 46, 0.05);
    matlab_matrix_entry(A, 23, 22, -0.0833333);
    matlab_matrix_entry(A, 23, 47, -0.05);
    matlab_matrix_entry(A, 23, 43, 0.116667);
    matlab_matrix_entry(A, 23, 19, 0.416667);
    matlab_matrix_entry(A, 23, 23, 0.366667);
    matlab_matrix_entry(A, 23, 48, 0.0166667);
    matlab_matrix_entry(A, 23, 24, 0.1);
    matlab_matrix_entry(A, 23, 44, 0.05);
    matlab_matrix_entry(A, 23, 20, -0.25);
    matlab_matrix_entry(A, 24, 45, 0.0166667);
    matlab_matrix_entry(A, 24, 21, 0.0611111);
    matlab_matrix_entry(A, 24, 46, 0.0166667);
    matlab_matrix_entry(A, 24, 22, -0.00555556);
    matlab_matrix_entry(A, 24, 47, -0.0166667);
    matlab_matrix_entry(A, 24, 19, 0.0833333);
    matlab_matrix_entry(A, 24, 23, 0.1);
    matlab_matrix_entry(A, 24, 43, 0.00555556);
    matlab_matrix_entry(A, 24, 20, -0.0833333);
    matlab_matrix_entry(A, 24, 44, -0.00555556);
    matlab_matrix_entry(A, 24, 24, 0.0568254);
    matlab_matrix_entry(A, 25, 25, 1.2);
    matlab_matrix_entry(A, 25, 29, 0.416667);
    matlab_matrix_entry(A, 25, 1, 0.25);
    matlab_matrix_entry(A, 25, 5, 0.316667);
    matlab_matrix_entry(A, 25, 30, 0.0833333);
    matlab_matrix_entry(A, 25, 26, -0.7);
    matlab_matrix_entry(A, 25, 2, -0.75);
    matlab_matrix_entry(A, 25, 6, -0.00555556);
    matlab_matrix_entry(A, 25, 27, 0.266667);
    matlab_matrix_entry(A, 25, 3, -0.116667);
    matlab_matrix_entry(A, 25, 28, -0.266667);
    matlab_matrix_entry(A, 25, 4, -0.15);
    matlab_matrix_entry(A, 26, 25, -0.7);
    matlab_matrix_entry(A, 26, 29, -0.25);
    matlab_matrix_entry(A, 26, 1, -0.75);
    matlab_matrix_entry(A, 26, 5, -0.316667);
    matlab_matrix_entry(A, 26, 30, -0.0833333);
    matlab_matrix_entry(A, 26, 26, 1.2);
    matlab_matrix_entry(A, 26, 2, 0.25);
    matlab_matrix_entry(A, 26, 6, 0.00555556);
    matlab_matrix_entry(A, 26, 27, -0.266667);
    matlab_matrix_entry(A, 26, 3, -0.05);
    matlab_matrix_entry(A, 26, 28, 0.266667);
    matlab_matrix_entry(A, 26, 4, 0.15);
    matlab_matrix_entry(A, 27, 25, 0.266667);
    matlab_matrix_entry(A, 27, 29, 0.0833333);
    matlab_matrix_entry(A, 27, 1, 0.316667);
    matlab_matrix_entry(A, 27, 5, 0.0722222);
    matlab_matrix_entry(A, 27, 26, -0.266667);
    matlab_matrix_entry(A, 27, 6, 0.0166667);
    matlab_matrix_entry(A, 27, 30, 0.0611111);
    matlab_matrix_entry(A, 27, 2, -0.316667);
    matlab_matrix_entry(A, 27, 3, 0.05);
    matlab_matrix_entry(A, 27, 27, 0.234921);
    matlab_matrix_entry(A, 27, 28, -0.068254);
    matlab_matrix_entry(A, 27, 4, -0.0722222);
    matlab_matrix_entry(A, 28, 25, -0.266667);
    matlab_matrix_entry(A, 28, 29, -0.0833333);
    matlab_matrix_entry(A, 28, 1, -0.15);
    matlab_matrix_entry(A, 28, 5, -0.0722222);
    matlab_matrix_entry(A, 28, 26, 0.266667);
    matlab_matrix_entry(A, 28, 6, -0.0166667);
    matlab_matrix_entry(A, 28, 30, -0.00555556);
    matlab_matrix_entry(A, 28, 2, 0.15);
    matlab_matrix_entry(A, 28, 3, -0.05);
    matlab_matrix_entry(A, 28, 27, -0.068254);
    matlab_matrix_entry(A, 28, 28, 0.234921);
    matlab_matrix_entry(A, 28, 4, 0.0722222);
    matlab_matrix_entry(A, 29, 5, 0.05);
    matlab_matrix_entry(A, 29, 25, 0.416667);
    matlab_matrix_entry(A, 29, 29, 0.366667);
    matlab_matrix_entry(A, 29, 1, -0.116667);
    matlab_matrix_entry(A, 29, 30, 0.1);
    matlab_matrix_entry(A, 29, 6, -0.0166667);
    matlab_matrix_entry(A, 29, 26, -0.25);
    matlab_matrix_entry(A, 29, 2, -0.05);
    matlab_matrix_entry(A, 29, 27, 0.0833333);
    matlab_matrix_entry(A, 29, 3, -0.0611111);
    matlab_matrix_entry(A, 29, 28, -0.0833333);
    matlab_matrix_entry(A, 29, 4, -0.05);
    matlab_matrix_entry(A, 30, 25, 0.0833333);
    matlab_matrix_entry(A, 30, 29, 0.1);
    matlab_matrix_entry(A, 30, 5, 0.0166667);
    matlab_matrix_entry(A, 30, 1, -0.00555556);
    matlab_matrix_entry(A, 30, 26, -0.0833333);
    matlab_matrix_entry(A, 30, 30, 0.0568254);
    matlab_matrix_entry(A, 30, 2, 0.00555556);
    matlab_matrix_entry(A, 30, 3, -0.0166667);
    matlab_matrix_entry(A, 30, 27, 0.0611111);
    matlab_matrix_entry(A, 30, 4, -0.0166667);
    matlab_matrix_entry(A, 30, 28, -0.00555556);
    matlab_matrix_entry(A, 31, 33, 0.416667);
    matlab_matrix_entry(A, 31, 9, -0.316667);
    matlab_matrix_entry(A, 31, 34, -0.266667);
    matlab_matrix_entry(A, 31, 10, 0.15);
    matlab_matrix_entry(A, 31, 35, 0.266667);
    matlab_matrix_entry(A, 31, 31, 1.2);
    matlab_matrix_entry(A, 31, 7, -0.25);
    matlab_matrix_entry(A, 31, 11, 0.116667);
    matlab_matrix_entry(A, 31, 36, 0.0833333);
    matlab_matrix_entry(A, 31, 32, -0.7);
    matlab_matrix_entry(A, 31, 8, 0.75);
    matlab_matrix_entry(A, 31, 12, 0.00555556);
    matlab_matrix_entry(A, 32, 33, -0.25);
    matlab_matrix_entry(A, 32, 9, 0.316667);
    matlab_matrix_entry(A, 32, 34, 0.266667);
    matlab_matrix_entry(A, 32, 10, -0.15);
    matlab_matrix_entry(A, 32, 35, -0.266667);
    matlab_matrix_entry(A, 32, 31, -0.7);
    matlab_matrix_entry(A, 32, 7, 0.75);
    matlab_matrix_entry(A, 32, 11, 0.05);
    matlab_matrix_entry(A, 32, 36, -0.0833333);
    matlab_matrix_entry(A, 32, 32, 1.2);
    matlab_matrix_entry(A, 32, 8, -0.25);
    matlab_matrix_entry(A, 32, 12, -0.00555556);
    matlab_matrix_entry(A, 33, 9, -0.05);
    matlab_matrix_entry(A, 33, 33, 0.366667);
    matlab_matrix_entry(A, 33, 34, -0.0833333);
    matlab_matrix_entry(A, 33, 10, 0.05);
    matlab_matrix_entry(A, 33, 35, 0.0833333);
    matlab_matrix_entry(A, 33, 31, 0.416667);
    matlab_matrix_entry(A, 33, 7, 0.116667);
    matlab_matrix_entry(A, 33, 11, 0.0611111);
    matlab_matrix_entry(A, 33, 36, 0.1);
    matlab_matrix_entry(A, 33, 12, 0.0166667);
    matlab_matrix_entry(A, 33, 32, -0.25);
    matlab_matrix_entry(A, 33, 8, 0.05);
    matlab_matrix_entry(A, 34, 33, -0.0833333);
    matlab_matrix_entry(A, 34, 9, 0.0722222);
    matlab_matrix_entry(A, 34, 34, 0.234921);
    matlab_matrix_entry(A, 34, 10, -0.0722222);
    matlab_matrix_entry(A, 34, 31, -0.266667);
    matlab_matrix_entry(A, 34, 11, 0.05);
    matlab_matrix_entry(A, 34, 35, -0.068254);
    matlab_matrix_entry(A, 34, 7, 0.15);
    matlab_matrix_entry(A, 34, 32, 0.266667);
    matlab_matrix_entry(A, 34, 12, 0.0166667);
    matlab_matrix_entry(A, 34, 36, -0.00555556);
    matlab_matrix_entry(A, 34, 8, -0.15);
    matlab_matrix_entry(A, 35, 33, 0.0833333);
    matlab_matrix_entry(A, 35, 9, -0.0722222);
    matlab_matrix_entry(A, 35, 34, -0.068254);
    matlab_matrix_entry(A, 35, 10, 0.0722222);
    matlab_matrix_entry(A, 35, 31, 0.266667);
    matlab_matrix_entry(A, 35, 11, -0.05);
    matlab_matrix_entry(A, 35, 35, 0.234921);
    matlab_matrix_entry(A, 35, 7, -0.316667);
    matlab_matrix_entry(A, 35, 32, -0.266667);
    matlab_matrix_entry(A, 35, 12, -0.0166667);
    matlab_matrix_entry(A, 35, 36, 0.0611111);
    matlab_matrix_entry(A, 35, 8, 0.316667);
    matlab_matrix_entry(A, 36, 33, 0.1);
    matlab_matrix_entry(A, 36, 9, -0.0166667);
    matlab_matrix_entry(A, 36, 10, 0.0166667);
    matlab_matrix_entry(A, 36, 34, -0.00555556);
    matlab_matrix_entry(A, 36, 31, 0.0833333);
    matlab_matrix_entry(A, 36, 11, 0.0166667);
    matlab_matrix_entry(A, 36, 35, 0.0611111);
    matlab_matrix_entry(A, 36, 7, 0.00555556);
    matlab_matrix_entry(A, 36, 32, -0.0833333);
    matlab_matrix_entry(A, 36, 36, 0.0568254);
    matlab_matrix_entry(A, 36, 8, -0.00555556);
    matlab_matrix_entry(A, 37, 37, 1.2);
    matlab_matrix_entry(A, 37, 41, 0.416667);
    matlab_matrix_entry(A, 37, 13, 0.25);
    matlab_matrix_entry(A, 37, 17, 0.316667);
    matlab_matrix_entry(A, 37, 42, 0.0833333);
    matlab_matrix_entry(A, 37, 38, -0.7);
    matlab_matrix_entry(A, 37, 14, -0.75);
    matlab_matrix_entry(A, 37, 18, -0.00555556);
    matlab_matrix_entry(A, 37, 39, 0.266667);
    matlab_matrix_entry(A, 37, 15, -0.116667);
    matlab_matrix_entry(A, 37, 40, -0.266667);
    matlab_matrix_entry(A, 37, 16, -0.15);
    matlab_matrix_entry(A, 38, 37, -0.7);
    matlab_matrix_entry(A, 38, 41, -0.25);
    matlab_matrix_entry(A, 38, 13, -0.75);
    matlab_matrix_entry(A, 38, 17, -0.316667);
    matlab_matrix_entry(A, 38, 42, -0.0833333);
    matlab_matrix_entry(A, 38, 38, 1.2);
    matlab_matrix_entry(A, 38, 14, 0.25);
    matlab_matrix_entry(A, 38, 18, 0.00555556);
    matlab_matrix_entry(A, 38, 39, -0.266667);
    matlab_matrix_entry(A, 38, 15, -0.05);
    matlab_matrix_entry(A, 38, 40, 0.266667);
    matlab_matrix_entry(A, 38, 16, 0.15);
    matlab_matrix_entry(A, 39, 37, 0.266667);
    matlab_matrix_entry(A, 39, 41, 0.0833333);
    matlab_matrix_entry(A, 39, 13, 0.316667);
    matlab_matrix_entry(A, 39, 17, 0.0722222);
    matlab_matrix_entry(A, 39, 38, -0.266667);
    matlab_matrix_entry(A, 39, 18, 0.0166667);
    matlab_matrix_entry(A, 39, 42, 0.0611111);
    matlab_matrix_entry(A, 39, 14, -0.316667);
    matlab_matrix_entry(A, 39, 15, 0.05);
    matlab_matrix_entry(A, 39, 39, 0.234921);
    matlab_matrix_entry(A, 39, 40, -0.068254);
    matlab_matrix_entry(A, 39, 16, -0.0722222);
    matlab_matrix_entry(A, 40, 37, -0.266667);
    matlab_matrix_entry(A, 40, 41, -0.0833333);
    matlab_matrix_entry(A, 40, 13, -0.15);
    matlab_matrix_entry(A, 40, 17, -0.0722222);
    matlab_matrix_entry(A, 40, 38, 0.266667);
    matlab_matrix_entry(A, 40, 18, -0.0166667);
    matlab_matrix_entry(A, 40, 42, -0.00555556);
    matlab_matrix_entry(A, 40, 14, 0.15);
    matlab_matrix_entry(A, 40, 15, -0.05);
    matlab_matrix_entry(A, 40, 39, -0.068254);
    matlab_matrix_entry(A, 40, 40, 0.234921);
    matlab_matrix_entry(A, 40, 16, 0.0722222);
    matlab_matrix_entry(A, 41, 17, 0.05);
    matlab_matrix_entry(A, 41, 37, 0.416667);
    matlab_matrix_entry(A, 41, 41, 0.366667);
    matlab_matrix_entry(A, 41, 13, -0.116667);
    matlab_matrix_entry(A, 41, 42, 0.1);
    matlab_matrix_entry(A, 41, 18, -0.0166667);
    matlab_matrix_entry(A, 41, 38, -0.25);
    matlab_matrix_entry(A, 41, 14, -0.05);
    matlab_matrix_entry(A, 41, 39, 0.0833333);
    matlab_matrix_entry(A, 41, 15, -0.0611111);
    matlab_matrix_entry(A, 41, 40, -0.0833333);
    matlab_matrix_entry(A, 41, 16, -0.05);
    matlab_matrix_entry(A, 42, 37, 0.0833333);
    matlab_matrix_entry(A, 42, 41, 0.1);
    matlab_matrix_entry(A, 42, 17, 0.0166667);
    matlab_matrix_entry(A, 42, 13, -0.00555556);
    matlab_matrix_entry(A, 42, 38, -0.0833333);
    matlab_matrix_entry(A, 42, 42, 0.0568254);
    matlab_matrix_entry(A, 42, 14, 0.00555556);
    matlab_matrix_entry(A, 42, 15, -0.0166667);
    matlab_matrix_entry(A, 42, 39, 0.0611111);
    matlab_matrix_entry(A, 42, 16, -0.0166667);
    matlab_matrix_entry(A, 42, 40, -0.00555556);
    matlab_matrix_entry(A, 43, 45, 0.416667);
    matlab_matrix_entry(A, 43, 21, -0.316667);
    matlab_matrix_entry(A, 43, 46, -0.266667);
    matlab_matrix_entry(A, 43, 22, 0.15);
    matlab_matrix_entry(A, 43, 47, 0.266667);
    matlab_matrix_entry(A, 43, 43, 1.2);
    matlab_matrix_entry(A, 43, 19, -0.25);
    matlab_matrix_entry(A, 43, 23, 0.116667);
    matlab_matrix_entry(A, 43, 48, 0.0833333);
    matlab_matrix_entry(A, 43, 44, -0.7);
    matlab_matrix_entry(A, 43, 20, 0.75);
    matlab_matrix_entry(A, 43, 24, 0.00555556);
    matlab_matrix_entry(A, 44, 45, -0.25);
    matlab_matrix_entry(A, 44, 21, 0.316667);
    matlab_matrix_entry(A, 44, 46, 0.266667);
    matlab_matrix_entry(A, 44, 22, -0.15);
    matlab_matrix_entry(A, 44, 47, -0.266667);
    matlab_matrix_entry(A, 44, 43, -0.7);
    matlab_matrix_entry(A, 44, 19, 0.75);
    matlab_matrix_entry(A, 44, 23, 0.05);
    matlab_matrix_entry(A, 44, 48, -0.0833333);
    matlab_matrix_entry(A, 44, 44, 1.2);
    matlab_matrix_entry(A, 44, 20, -0.25);
    matlab_matrix_entry(A, 44, 24, -0.00555556);
    matlab_matrix_entry(A, 45, 21, -0.05);
    matlab_matrix_entry(A, 45, 45, 0.366667);
    matlab_matrix_entry(A, 45, 46, -0.0833333);
    matlab_matrix_entry(A, 45, 22, 0.05);
    matlab_matrix_entry(A, 45, 47, 0.0833333);
    matlab_matrix_entry(A, 45, 43, 0.416667);
    matlab_matrix_entry(A, 45, 19, 0.116667);
    matlab_matrix_entry(A, 45, 23, 0.0611111);
    matlab_matrix_entry(A, 45, 48, 0.1);
    matlab_matrix_entry(A, 45, 24, 0.0166667);
    matlab_matrix_entry(A, 45, 44, -0.25);
    matlab_matrix_entry(A, 45, 20, 0.05);
    matlab_matrix_entry(A, 46, 45, -0.0833333);
    matlab_matrix_entry(A, 46, 21, 0.0722222);
    matlab_matrix_entry(A, 46, 46, 0.234921);
    matlab_matrix_entry(A, 46, 22, -0.0722222);
    matlab_matrix_entry(A, 46, 43, -0.266667);
    matlab_matrix_entry(A, 46, 23, 0.05);
    matlab_matrix_entry(A, 46, 47, -0.068254);
    matlab_matrix_entry(A, 46, 19, 0.15);
    matlab_matrix_entry(A, 46, 44, 0.266667);
    matlab_matrix_entry(A, 46, 24, 0.0166667);
    matlab_matrix_entry(A, 46, 48, -0.00555556);
    matlab_matrix_entry(A, 46, 20, -0.15);
    matlab_matrix_entry(A, 47, 45, 0.0833333);
    matlab_matrix_entry(A, 47, 21, -0.0722222);
    matlab_matrix_entry(A, 47, 46, -0.068254);
    matlab_matrix_entry(A, 47, 22, 0.0722222);
    matlab_matrix_entry(A, 47, 43, 0.266667);
    matlab_matrix_entry(A, 47, 23, -0.05);
    matlab_matrix_entry(A, 47, 47, 0.234921);
    matlab_matrix_entry(A, 47, 19, -0.316667);
    matlab_matrix_entry(A, 47, 44, -0.266667);
    matlab_matrix_entry(A, 47, 24, -0.0166667);
    matlab_matrix_entry(A, 47, 48, 0.0611111);
    matlab_matrix_entry(A, 47, 20, 0.316667);
    matlab_matrix_entry(A, 48, 45, 0.1);
    matlab_matrix_entry(A, 48, 21, -0.0166667);
    matlab_matrix_entry(A, 48, 22, 0.0166667);
    matlab_matrix_entry(A, 48, 46, -0.00555556);
    matlab_matrix_entry(A, 48, 43, 0.0833333);
    matlab_matrix_entry(A, 48, 23, 0.0166667);
    matlab_matrix_entry(A, 48, 47, 0.0611111);
    matlab_matrix_entry(A, 48, 19, 0.00555556);
    matlab_matrix_entry(A, 48, 44, -0.0833333);
    matlab_matrix_entry(A, 48, 48, 0.0568254);
    matlab_matrix_entry(A, 48, 20, -0.00555556);

    matlab_matrix_entry(b, 1, 5, 0.0277778);
    matlab_matrix_entry(b, 1, 1, 0.111111);
    matlab_matrix_entry(b, 1, 6, 0.00694444);
    matlab_matrix_entry(b, 1, 2, 0.0555556);
    matlab_matrix_entry(b, 1, 3, 0.0277778);
    matlab_matrix_entry(b, 1, 4, 0.0138889);
    matlab_matrix_entry(b, 2, 5, 0.0138889);
    matlab_matrix_entry(b, 2, 1, 0.0555556);
    matlab_matrix_entry(b, 2, 6, 0.00694444);
    matlab_matrix_entry(b, 2, 2, 0.111111);
    matlab_matrix_entry(b, 2, 3, 0.0277778);
    matlab_matrix_entry(b, 2, 4, 0.0277778);
    matlab_matrix_entry(b, 3, 5, 0.00694444);
    matlab_matrix_entry(b, 3, 1, 0.0277778);
    matlab_matrix_entry(b, 3, 6, 0.00277778);
    matlab_matrix_entry(b, 3, 2, 0.0277778);
    matlab_matrix_entry(b, 3, 3, 0.0111111);
    matlab_matrix_entry(b, 3, 4, 0.00694444);
    matlab_matrix_entry(b, 4, 5, 0.00555556);
    matlab_matrix_entry(b, 4, 1, 0.0138889);
    matlab_matrix_entry(b, 4, 6, 0.00277778);
    matlab_matrix_entry(b, 4, 2, 0.0277778);
    matlab_matrix_entry(b, 4, 3, 0.00694444);
    matlab_matrix_entry(b, 4, 4, 0.0111111);
    matlab_matrix_entry(b, 5, 5, 0.0111111);
    matlab_matrix_entry(b, 5, 1, 0.0277778);
    matlab_matrix_entry(b, 5, 6, 0.00277778);
    matlab_matrix_entry(b, 5, 2, 0.0138889);
    matlab_matrix_entry(b, 5, 3, 0.00694444);
    matlab_matrix_entry(b, 5, 4, 0.00555556);
    matlab_matrix_entry(b, 6, 5, 0.00277778);
    matlab_matrix_entry(b, 6, 1, 0.00694444);
    matlab_matrix_entry(b, 6, 6, 0.00111111);
    matlab_matrix_entry(b, 6, 2, 0.00694444);
    matlab_matrix_entry(b, 6, 3, 0.00277778);
    matlab_matrix_entry(b, 6, 4, 0.00277778);
    matlab_matrix_entry(b, 7, 9, 0.0138889);
    matlab_matrix_entry(b, 7, 10, 0.00694444);
    matlab_matrix_entry(b, 7, 11, 0.0138889);
    matlab_matrix_entry(b, 7, 7, 0.0555556);
    matlab_matrix_entry(b, 7, 12, 0.00347222);
    matlab_matrix_entry(b, 7, 8, 0.0277778);
    matlab_matrix_entry(b, 8, 9, 0.00694444);
    matlab_matrix_entry(b, 8, 10, 0.0138889);
    matlab_matrix_entry(b, 8, 11, 0.0138889);
    matlab_matrix_entry(b, 8, 7, 0.0277778);
    matlab_matrix_entry(b, 8, 12, 0.00347222);
    matlab_matrix_entry(b, 8, 8, 0.0555556);
    matlab_matrix_entry(b, 9, 9, 0.00555556);
    matlab_matrix_entry(b, 9, 10, 0.00277778);
    matlab_matrix_entry(b, 9, 11, 0.00347222);
    matlab_matrix_entry(b, 9, 7, 0.0138889);
    matlab_matrix_entry(b, 9, 12, 0.00138889);
    matlab_matrix_entry(b, 9, 8, 0.00694444);
    matlab_matrix_entry(b, 10, 9, 0.00277778);
    matlab_matrix_entry(b, 10, 10, 0.00555556);
    matlab_matrix_entry(b, 10, 11, 0.00347222);
    matlab_matrix_entry(b, 10, 7, 0.00694444);
    matlab_matrix_entry(b, 10, 12, 0.00138889);
    matlab_matrix_entry(b, 10, 8, 0.0138889);
    matlab_matrix_entry(b, 11, 9, 0.00347222);
    matlab_matrix_entry(b, 11, 10, 0.00347222);
    matlab_matrix_entry(b, 11, 11, 0.00555556);
    matlab_matrix_entry(b, 11, 7, 0.0138889);
    matlab_matrix_entry(b, 11, 12, 0.00138889);
    matlab_matrix_entry(b, 11, 8, 0.0138889);
    matlab_matrix_entry(b, 12, 9, 0.00138889);
    matlab_matrix_entry(b, 12, 10, 0.00138889);
    matlab_matrix_entry(b, 12, 11, 0.00138889);
    matlab_matrix_entry(b, 12, 7, 0.00347222);
    matlab_matrix_entry(b, 12, 12, 0.000555556);
    matlab_matrix_entry(b, 12, 8, 0.00347222);
    matlab_matrix_entry(b, 13, 17, 0.0277778);
    matlab_matrix_entry(b, 13, 13, 0.111111);
    matlab_matrix_entry(b, 13, 18, 0.00694444);
    matlab_matrix_entry(b, 13, 14, 0.0555556);
    matlab_matrix_entry(b, 13, 15, 0.0277778);
    matlab_matrix_entry(b, 13, 16, 0.0138889);
    matlab_matrix_entry(b, 14, 17, 0.0138889);
    matlab_matrix_entry(b, 14, 13, 0.0555556);
    matlab_matrix_entry(b, 14, 18, 0.00694444);
    matlab_matrix_entry(b, 14, 14, 0.111111);
    matlab_matrix_entry(b, 14, 15, 0.0277778);
    matlab_matrix_entry(b, 14, 16, 0.0277778);
    matlab_matrix_entry(b, 15, 17, 0.00694444);
    matlab_matrix_entry(b, 15, 13, 0.0277778);
    matlab_matrix_entry(b, 15, 18, 0.00277778);
    matlab_matrix_entry(b, 15, 14, 0.0277778);
    matlab_matrix_entry(b, 15, 15, 0.0111111);
    matlab_matrix_entry(b, 15, 16, 0.00694444);
    matlab_matrix_entry(b, 16, 17, 0.00555556);
    matlab_matrix_entry(b, 16, 13, 0.0138889);
    matlab_matrix_entry(b, 16, 18, 0.00277778);
    matlab_matrix_entry(b, 16, 14, 0.0277778);
    matlab_matrix_entry(b, 16, 15, 0.00694444);
    matlab_matrix_entry(b, 16, 16, 0.0111111);
    matlab_matrix_entry(b, 17, 17, 0.0111111);
    matlab_matrix_entry(b, 17, 13, 0.0277778);
    matlab_matrix_entry(b, 17, 18, 0.00277778);
    matlab_matrix_entry(b, 17, 14, 0.0138889);
    matlab_matrix_entry(b, 17, 15, 0.00694444);
    matlab_matrix_entry(b, 17, 16, 0.00555556);
    matlab_matrix_entry(b, 18, 17, 0.00277778);
    matlab_matrix_entry(b, 18, 13, 0.00694444);
    matlab_matrix_entry(b, 18, 18, 0.00111111);
    matlab_matrix_entry(b, 18, 14, 0.00694444);
    matlab_matrix_entry(b, 18, 15, 0.00277778);
    matlab_matrix_entry(b, 18, 16, 0.00277778);
    matlab_matrix_entry(b, 19, 21, 0.0138889);
    matlab_matrix_entry(b, 19, 22, 0.00694444);
    matlab_matrix_entry(b, 19, 23, 0.0138889);
    matlab_matrix_entry(b, 19, 19, 0.0555556);
    matlab_matrix_entry(b, 19, 24, 0.00347222);
    matlab_matrix_entry(b, 19, 20, 0.0277778);
    matlab_matrix_entry(b, 20, 21, 0.00694444);
    matlab_matrix_entry(b, 20, 22, 0.0138889);
    matlab_matrix_entry(b, 20, 23, 0.0138889);
    matlab_matrix_entry(b, 20, 19, 0.0277778);
    matlab_matrix_entry(b, 20, 24, 0.00347222);
    matlab_matrix_entry(b, 20, 20, 0.0555556);
    matlab_matrix_entry(b, 21, 21, 0.00555556);
    matlab_matrix_entry(b, 21, 22, 0.00277778);
    matlab_matrix_entry(b, 21, 23, 0.00347222);
    matlab_matrix_entry(b, 21, 19, 0.0138889);
    matlab_matrix_entry(b, 21, 24, 0.00138889);
    matlab_matrix_entry(b, 21, 20, 0.00694444);
    matlab_matrix_entry(b, 22, 21, 0.00277778);
    matlab_matrix_entry(b, 22, 22, 0.00555556);
    matlab_matrix_entry(b, 22, 23, 0.00347222);
    matlab_matrix_entry(b, 22, 19, 0.00694444);
    matlab_matrix_entry(b, 22, 24, 0.00138889);
    matlab_matrix_entry(b, 22, 20, 0.0138889);
    matlab_matrix_entry(b, 23, 21, 0.00347222);
    matlab_matrix_entry(b, 23, 22, 0.00347222);
    matlab_matrix_entry(b, 23, 23, 0.00555556);
    matlab_matrix_entry(b, 23, 19, 0.0138889);
    matlab_matrix_entry(b, 23, 24, 0.00138889);
    matlab_matrix_entry(b, 23, 20, 0.0138889);
    matlab_matrix_entry(b, 24, 21, 0.00138889);
    matlab_matrix_entry(b, 24, 22, 0.00138889);
    matlab_matrix_entry(b, 24, 23, 0.00138889);
    matlab_matrix_entry(b, 24, 19, 0.00347222);
    matlab_matrix_entry(b, 24, 24, 0.000555556);
    matlab_matrix_entry(b, 24, 20, 0.00347222);
    matlab_matrix_entry(b, 25, 29, 0.0277778);
    matlab_matrix_entry(b, 25, 25, 0.111111);
    matlab_matrix_entry(b, 25, 30, 0.00694444);
    matlab_matrix_entry(b, 25, 26, 0.0555556);
    matlab_matrix_entry(b, 25, 27, 0.0277778);
    matlab_matrix_entry(b, 25, 28, 0.0138889);
    matlab_matrix_entry(b, 26, 29, 0.0277778);
    matlab_matrix_entry(b, 26, 25, 0.0555556);
    matlab_matrix_entry(b, 26, 30, 0.00694444);
    matlab_matrix_entry(b, 26, 26, 0.111111);
    matlab_matrix_entry(b, 26, 27, 0.0138889);
    matlab_matrix_entry(b, 26, 28, 0.0277778);
    matlab_matrix_entry(b, 27, 29, 0.00694444);
    matlab_matrix_entry(b, 27, 25, 0.0277778);
    matlab_matrix_entry(b, 27, 30, 0.00277778);
    matlab_matrix_entry(b, 27, 26, 0.0138889);
    matlab_matrix_entry(b, 27, 27, 0.0111111);
    matlab_matrix_entry(b, 27, 28, 0.00555556);
    matlab_matrix_entry(b, 28, 29, 0.00694444);
    matlab_matrix_entry(b, 28, 25, 0.0138889);
    matlab_matrix_entry(b, 28, 30, 0.00277778);
    matlab_matrix_entry(b, 28, 26, 0.0277778);
    matlab_matrix_entry(b, 28, 27, 0.00555556);
    matlab_matrix_entry(b, 28, 28, 0.0111111);
    matlab_matrix_entry(b, 29, 29, 0.0111111);
    matlab_matrix_entry(b, 29, 25, 0.0277778);
    matlab_matrix_entry(b, 29, 30, 0.00277778);
    matlab_matrix_entry(b, 29, 26, 0.0277778);
    matlab_matrix_entry(b, 29, 27, 0.00694444);
    matlab_matrix_entry(b, 29, 28, 0.00694444);
    matlab_matrix_entry(b, 30, 29, 0.00277778);
    matlab_matrix_entry(b, 30, 25, 0.00694444);
    matlab_matrix_entry(b, 30, 30, 0.00111111);
    matlab_matrix_entry(b, 30, 26, 0.00694444);
    matlab_matrix_entry(b, 30, 27, 0.00277778);
    matlab_matrix_entry(b, 30, 28, 0.00277778);
    matlab_matrix_entry(b, 31, 33, 0.0138889);
    matlab_matrix_entry(b, 31, 34, 0.00694444);
    matlab_matrix_entry(b, 31, 35, 0.0138889);
    matlab_matrix_entry(b, 31, 31, 0.0555556);
    matlab_matrix_entry(b, 31, 36, 0.00347222);
    matlab_matrix_entry(b, 31, 32, 0.0277778);
    matlab_matrix_entry(b, 32, 33, 0.0138889);
    matlab_matrix_entry(b, 32, 34, 0.0138889);
    matlab_matrix_entry(b, 32, 35, 0.00694444);
    matlab_matrix_entry(b, 32, 31, 0.0277778);
    matlab_matrix_entry(b, 32, 36, 0.00347222);
    matlab_matrix_entry(b, 32, 32, 0.0555556);
    matlab_matrix_entry(b, 33, 33, 0.00555556);
    matlab_matrix_entry(b, 33, 34, 0.00347222);
    matlab_matrix_entry(b, 33, 35, 0.00347222);
    matlab_matrix_entry(b, 33, 31, 0.0138889);
    matlab_matrix_entry(b, 33, 36, 0.00138889);
    matlab_matrix_entry(b, 33, 32, 0.0138889);
    matlab_matrix_entry(b, 34, 33, 0.00347222);
    matlab_matrix_entry(b, 34, 34, 0.00555556);
    matlab_matrix_entry(b, 34, 35, 0.00277778);
    matlab_matrix_entry(b, 34, 31, 0.00694444);
    matlab_matrix_entry(b, 34, 36, 0.00138889);
    matlab_matrix_entry(b, 34, 32, 0.0138889);
    matlab_matrix_entry(b, 35, 33, 0.00347222);
    matlab_matrix_entry(b, 35, 34, 0.00277778);
    matlab_matrix_entry(b, 35, 35, 0.00555556);
    matlab_matrix_entry(b, 35, 31, 0.0138889);
    matlab_matrix_entry(b, 35, 36, 0.00138889);
    matlab_matrix_entry(b, 35, 32, 0.00694444);
    matlab_matrix_entry(b, 36, 33, 0.00138889);
    matlab_matrix_entry(b, 36, 34, 0.00138889);
    matlab_matrix_entry(b, 36, 35, 0.00138889);
    matlab_matrix_entry(b, 36, 31, 0.00347222);
    matlab_matrix_entry(b, 36, 36, 0.000555556);
    matlab_matrix_entry(b, 36, 32, 0.00347222);
    matlab_matrix_entry(b, 37, 41, 0.0277778);
    matlab_matrix_entry(b, 37, 37, 0.111111);
    matlab_matrix_entry(b, 37, 42, 0.00694444);
    matlab_matrix_entry(b, 37, 38, 0.0555556);
    matlab_matrix_entry(b, 37, 39, 0.0277778);
    matlab_matrix_entry(b, 37, 40, 0.0138889);
    matlab_matrix_entry(b, 38, 41, 0.0277778);
    matlab_matrix_entry(b, 38, 37, 0.0555556);
    matlab_matrix_entry(b, 38, 42, 0.00694444);
    matlab_matrix_entry(b, 38, 38, 0.111111);
    matlab_matrix_entry(b, 38, 39, 0.0138889);
    matlab_matrix_entry(b, 38, 40, 0.0277778);
    matlab_matrix_entry(b, 39, 41, 0.00694444);
    matlab_matrix_entry(b, 39, 37, 0.0277778);
    matlab_matrix_entry(b, 39, 42, 0.00277778);
    matlab_matrix_entry(b, 39, 38, 0.0138889);
    matlab_matrix_entry(b, 39, 39, 0.0111111);
    matlab_matrix_entry(b, 39, 40, 0.00555556);
    matlab_matrix_entry(b, 40, 41, 0.00694444);
    matlab_matrix_entry(b, 40, 37, 0.0138889);
    matlab_matrix_entry(b, 40, 42, 0.00277778);
    matlab_matrix_entry(b, 40, 38, 0.0277778);
    matlab_matrix_entry(b, 40, 39, 0.00555556);
    matlab_matrix_entry(b, 40, 40, 0.0111111);
    matlab_matrix_entry(b, 41, 41, 0.0111111);
    matlab_matrix_entry(b, 41, 37, 0.0277778);
    matlab_matrix_entry(b, 41, 42, 0.00277778);
    matlab_matrix_entry(b, 41, 38, 0.0277778);
    matlab_matrix_entry(b, 41, 39, 0.00694444);
    matlab_matrix_entry(b, 41, 40, 0.00694444);
    matlab_matrix_entry(b, 42, 41, 0.00277778);
    matlab_matrix_entry(b, 42, 37, 0.00694444);
    matlab_matrix_entry(b, 42, 42, 0.00111111);
    matlab_matrix_entry(b, 42, 38, 0.00694444);
    matlab_matrix_entry(b, 42, 39, 0.00277778);
    matlab_matrix_entry(b, 42, 40, 0.00277778);
    matlab_matrix_entry(b, 43, 45, 0.0138889);
    matlab_matrix_entry(b, 43, 46, 0.00694444);
    matlab_matrix_entry(b, 43, 47, 0.0138889);
    matlab_matrix_entry(b, 43, 43, 0.0555556);
    matlab_matrix_entry(b, 43, 48, 0.00347222);
    matlab_matrix_entry(b, 43, 44, 0.0277778);
    matlab_matrix_entry(b, 44, 45, 0.0138889);
    matlab_matrix_entry(b, 44, 46, 0.0138889);
    matlab_matrix_entry(b, 44, 47, 0.00694444);
    matlab_matrix_entry(b, 44, 43, 0.0277778);
    matlab_matrix_entry(b, 44, 48, 0.00347222);
    matlab_matrix_entry(b, 44, 44, 0.0555556);
    matlab_matrix_entry(b, 45, 45, 0.00555556);
    matlab_matrix_entry(b, 45, 46, 0.00347222);
    matlab_matrix_entry(b, 45, 47, 0.00347222);
    matlab_matrix_entry(b, 45, 43, 0.0138889);
    matlab_matrix_entry(b, 45, 48, 0.00138889);
    matlab_matrix_entry(b, 45, 44, 0.0138889);
    matlab_matrix_entry(b, 46, 45, 0.00347222);
    matlab_matrix_entry(b, 46, 46, 0.00555556);
    matlab_matrix_entry(b, 46, 47, 0.00277778);
    matlab_matrix_entry(b, 46, 43, 0.00694444);
    matlab_matrix_entry(b, 46, 48, 0.00138889);
    matlab_matrix_entry(b, 46, 44, 0.0138889);
    matlab_matrix_entry(b, 47, 45, 0.00347222);
    matlab_matrix_entry(b, 47, 46, 0.00277778);
    matlab_matrix_entry(b, 47, 47, 0.00555556);
    matlab_matrix_entry(b, 47, 43, 0.0138889);
    matlab_matrix_entry(b, 47, 48, 0.00138889);
    matlab_matrix_entry(b, 47, 44, 0.00694444);
    matlab_matrix_entry(b, 48, 45, 0.00138889);
    matlab_matrix_entry(b, 48, 46, 0.00138889);
    matlab_matrix_entry(b, 48, 47, 0.00138889);
    matlab_matrix_entry(b, 48, 43, 0.00347222);
    matlab_matrix_entry(b, 48, 48, 0.000555556);
    matlab_matrix_entry(b, 48, 44, 0.00347222);

    matlab_matrix_entry(c, 1, 21, -1); matlab_matrix_entry(c, 1, 12, 1);
    matlab_matrix_entry(c, 2, 10, 1); matlab_matrix_entry(c, 3, 8, 1);
    matlab_matrix_entry(c, 5, 19, -1); matlab_matrix_entry(c, 7, 21, 0.5);
    matlab_matrix_entry(c, 7, 16, 1); matlab_matrix_entry(c, 8, 18, 1);
    matlab_matrix_entry(c, 9, 19, 0.5); matlab_matrix_entry(c, 11, 14, 1);
    matlab_matrix_entry(c, 13, 3, 1); matlab_matrix_entry(c, 13, 16, -1);
    matlab_matrix_entry(c, 14, 18, -1); matlab_matrix_entry(c, 15, 14, -1);
    matlab_matrix_entry(c, 17, 1, 1); matlab_matrix_entry(c, 19, 3, -0.5);
    matlab_matrix_entry(c, 19, 12, -1); matlab_matrix_entry(c, 20, 10, -1);
    matlab_matrix_entry(c, 21, 1, -0.5); matlab_matrix_entry(c, 23, 8, -1);
    matlab_matrix_entry(c, 25, 22, -1); matlab_matrix_entry(c, 25, 11, -1);
    matlab_matrix_entry(c, 26, 24, -1); matlab_matrix_entry(c, 27, 7, -1);
    matlab_matrix_entry(c, 29, 20, -1); matlab_matrix_entry(c, 31, 22, 1);
    matlab_matrix_entry(c, 31, 15, -0.5); matlab_matrix_entry(c, 32, 24, 1);
    matlab_matrix_entry(c, 33, 20, 1); matlab_matrix_entry(c, 35, 13, -0.5);
    matlab_matrix_entry(c, 37, 15, 1); matlab_matrix_entry(c, 37, 4, 1);
    matlab_matrix_entry(c, 38, 6, 1); matlab_matrix_entry(c, 39, 13, 1);
    matlab_matrix_entry(c, 41, 2, 1); matlab_matrix_entry(c, 43, 11, 0.5);
    matlab_matrix_entry(c, 43, 4, -1); matlab_matrix_entry(c, 44, 6, -1);
    matlab_matrix_entry(c, 45, 2, -1); matlab_matrix_entry(c, 47, 7, 0.5);
  }

  void MaxwellTransmissionEVP::matlab_matrix_entry(concepts::Matrix<Real>& m,
						   uint i, uint j,
						   Real entry) {
    m(i-1,j-1) = entry;
  }

} // namespace test

// ****************************************************************** main() **

int main(int argc, char** argv) {
  concepts::Stacktrace::doit() = false;
  try {
    test::TestSuite t("Constrained Matrix Eigenvalue Problems");
    test::GolubExample t1;           t.addTest(&t1);
    test::MaxwellTransmissionEVP t2; t.addTest(&t2);
    test::GolubExampleSum t3;        t.addTest(&t3);
    t.run();
    return t.report();
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}
