// testsuite for ARPACK eigensolver

#include <iostream>
#include <unistd.h> // for command line parsing
#include "basics/fortranException.hh"
#include "space/space.hh"
#include "operator/sparseMatrix.hh"
#include "operator/cg.hh"
#include "eigensolver/ARPACKdsaupd.hh"

// debugging
#include "basics/debug.hh"

#define ARPACK_D 0

#if defined(__hpux) || defined(_IBMR2)
  #define FORTRAN(X) X
#else
  #define FORTRAN(X) X ## _
#endif

#ifdef __cplusplus
extern "C" {
#endif
  extern void FORTRAN(dsaupd)(int& IDO, const char& BMAT, const int& N,
			      const char WHICH[2], const int& NEV,
			      const double& TOL, const double* RESID,
			      const int& NCV, const double* V, const int& LDV,
			      int IPARAM[11], int IPNTR[11],
			      const double* WORKD, const double* WORKL,
			      const int& LWORKL, int& INFO);
  /*
  extern void FORTRAN(dgemv)(const char& TRANS, const int& M, const int& N,
			     const double& ALPHA, const double* const A,
			     const int& LDA, const double* const X,
			     const int& INCX, const double& BETA,
			     const double* Y, const int& INY);
  */
  extern void FORTRAN(dseupd)(const int& RVEC, const char& HOWMNY,
			      const int* SELECT, const double* D,
			      const double* Z, const int& LDZ,
			      const double& SIGMA,
			      const char& BMAT, const int& N,
			      const char WHICH[2], const int& NEV,
			      const double& TOL, const double* RESID,
			      const int& NCV, const double* V, const int& LDV,
			      int IPARAM[11], int IPNTR[11],
			      const double* WORKD, const double* WORKL,
			      const int& LWORKL, int& INFO);
#ifdef __cplusplus
}
#endif

using concepts::Real;

int main(int argc, char** argv) {

  try {
    uint dim = 10;
    int kmax_ = 4;
    int opt;
    while ((opt = getopt(argc, argv, "-d:k:")) != EOF)
      switch(opt) {
      case 'd': dim = atoi(optarg); break;
      case 'k': kmax_ = atoi(optarg); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-d DIM] [-k KMAX]" << std::endl
		  << "where" << std::endl
		  << "  DIM: dimension of problem (default: " << dim << ')'
		  << std::endl
		  << "  KMAX: number of eigenparis to be computed (default: "
		  << kmax_ << ')' << std::endl;
	exit(1); break;
      }
    conceptsAssert((uint)kmax_ < dim, concepts::Assertion());
    conceptsAssert(kmax_ > 0, concepts::Assertion());

    concepts::DummySpace<Real> spc(dim);
    std::cout << "Space: " << spc << std::endl;

    // create and fill mass and stiff
    concepts::SparseMatrix<Real> stiff_(spc, spc), mass_(spc, spc);
    const Real h = 1.0/((Real)dim+1.0);
    stiff_(0,0) = 2.0/h; stiff_(0,1) = -1.0/h;
    for (uint i = 1; i < dim-1; ++i) {
      stiff_(i, i-1) = -1.0/h; stiff_(i,i) = 2.0/h; stiff_(i,i+1) = -1.0/h;
    }
    stiff_(dim-1,dim-2) = -1.0/h; stiff_(dim-1,dim-1) = 2.0/h;
    mass_(0,0) = 4.0*h/6.0; mass_(0,1) = h/6.0;
    for (uint i = 1; i < dim-1; ++i) {
      mass_(i, i-1) = h/6.0; mass_(i,i) = 4.0*h/6.0; mass_(i,i+1) = h/6.0;
    }
    mass_(dim-1,dim-2) = h/6.0; mass_(dim-1,dim-1) = 4.0*h/6.0;
    std::cout << "Mass matrix: " << mass_ << std::endl;
    std::cout << "Stiffness matrix: " << stiff_ << std::endl;

    /*
    { // test for linking to correct BLAS routine
      // same routine as from dsaitr.f, line 573
      // with the wrong routine: result = nan
      // with the correct routine: result = 1222.7675453394
      concepts::Array<Real> v(10, 0), workd(31, 0);
      v[0] =  0.47333732179607;     v[1] = 0.38584472986338;
      v[2] = -0.49256020958245;     v[3] = -1.0958492464762;
      v[4] = 1.9124164589259;       v[5] = -2.369770126501;
      v[6] = 2.8945904290777;       v[7] = -2.4117396276622;
      v[8] = 1.745023804561;        v[9] = -1.0940559657072;
      workd[21] = 6.1691290510164;  workd[22] = 8.7000358226446;
      workd[23] = -3.0262749280734; workd[24] = -39.727102165254;
      workd[25] = 80.19497519912;   workd[26] = -105.01201855106;
      workd[27] = 116.2775967355;   workd[28] = -104.09402837859;
      workd[29] = 76.954275227406;  workd[30] = -43.26449309573;

      const int n = 10;
      const int j = 1;
      const int ldv = 256;
      FORTRAN(dgemv)('T', n, j, 1.0, (Real*)v, ldv, (Real*)workd+21, 1, 0.0,
		     (Real*)workd+11, 1);
      DEBUGL(1, workd);
    }
    */

    Real tol_ = 0.0;
    int maxiter_ = 300;
    concepts::CG<Real>* solver_ = new concepts::CG<Real>
      (mass_, 1e-12, 2*mass_.spaceX().dim()+100);

    int ido = 0;
    const char bmat = 'G';
    const int n = stiff_.spaceX().dim();
    const char which[2] = { 'S', 'M' };
    const int ncv = 3*kmax_ <= n ? 3*kmax_ : n; // suggestion by manual: >= 2*kmax
    const int lworkl = ncv*(8+ncv);
    concepts::Array<Real> resid(n, 0), v(ncv*n, 0), workd(3*n, 0),
      workl(lworkl, 0);
    const int ldv = n;
    concepts::Array<int> iparam(11, 0), ipntr(11, 0);
    iparam[0] = 1; // ishift = 1: use exact shifts
    iparam[1] = 0; // no longer referenced
    iparam[2] = maxiter_; // max number of Arnoldi iterations allowed
    iparam[3] = 0; // blocksize, only 1 is supported
    iparam[4] = 0; // number of converged eigenvalues
    iparam[5] = 0; // no longer referenced
    iparam[6] = 2; // mode: generalized EVP, mass_ SPD
    iparam[7] = 0; // not needed for ishift = 1
    int info = 0;
    DEBUGL(ARPACK_D, "bmat = " << bmat << ", n = " << n << ", which = "
	   << which[0] << which[1] << ", nev = " << kmax_ << ", tol = "
	   << tol_ << ", ncv = " << ncv << ", iparam = " << iparam);
    while (ido != 99) {
      FORTRAN(dsaupd)(ido, bmat, n, which, kmax_, tol_, (Real*)resid, ncv,
		      (Real*)v, ldv, (int*)iparam, (int*)ipntr, (Real*)workd,
		      (Real*)workl, lworkl, info);
      DEBUGL(ARPACK_D, " *** ido = " << ido << ", info = " << info
	     << ", ipntr = " << ipntr << ", iparam = " << iparam);
      switch (ido) {
      case -1: case 1: {
	concepts::Vector<Real> y(mass_.spaceX(), (Real*)workd+ipntr[1]-1),
	  x(mass_.spaceY(), (Real*)workd+ipntr[0]-1);
	DEBUGL(ARPACK_D, "y = OP*x -- " << ipntr[0] << ", "
	       << ipntr[1] << ", x = " << x);
	stiff_(x, y);
	x = y;
 	(*solver_)(x, y);
	DEBUGL(ARPACK_D, "x = " << x << ", y = " << y);
	DEBUGL(ARPACK_D, "iter of CG: " << solver_->iterations()
	       << ", eps: " << solver_->epsilon() << ", y = " << y);
      } break;
      case 2: {
	DEBUGL(ARPACK_D, "y = B*x -- " << ipntr[0] << ", " << ipntr[1]);
	concepts::Vector<Real> y(mass_.spaceX(), (Real*)workd+ipntr[1]-1),
	  x(mass_.spaceY(), (Real*)workd+ipntr[0]-1);
	mass_(x, y);
	DEBUGL(ARPACK_D, "x = " << x << ", y = " << y);
      } break;
      case 99: break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
    }
    DEBUGL(ARPACK_D, "resid = " << resid << ", iparam = " << iparam
	   << ", ipntr = " << ipntr << ", info = " << info << ", workl = "
	   << workl);
    delete solver_;

    std::cout << iparam[2] << " Arnoldi update iterations" << std::endl;
    std::cout << iparam[4] << " converged eigenvalues" << std::endl;
    std::cout << iparam[8] << " OP*x operations, OP = mass^-1 * stiff"
	      << std::endl;
    std::cout << iparam[9] << " B*x operations, B = mass" << std::endl;
    std::cout << iparam[10] << " steps of re-orthogonalization" << std::endl;
    switch (info) {
    case 0: break; // everything ok
    case 1: break; // not all eigenpairs have converged
    default: throw conceptsException(concepts::FortranException
				     (info, "dsaupd"));
    }

    concepts::Array<int> select(ncv, true);
    concepts::Array<Real> eigenvalues(kmax_, 0);
    concepts::Array<Real> eigenvectors(kmax_*n, 0);
    FORTRAN(dseupd)(1, 'A', select, (Real*)eigenvalues, (Real*)v,
		    ldv, 0.0,
		    bmat, n, which, kmax_, tol_, (Real*)resid, ncv,
		    (Real*)v, ldv, (int*)iparam, (int*)ipntr, (Real*)workd,
		    (Real*)workl, lworkl, info);
    if (info != 0) throw conceptsException(concepts::FortranException
					   (info, "dseupd"));
    DEBUGL(ARPACK_D, "info = " << info);
    DEBUGL(ARPACK_D, "eigenvectors = " << v);
    std::cout << "Eigenvalues = " << eigenvalues << std::endl;
    std::cout << "Eigenvectors: " << std::endl;
    for (int i = 0; i < kmax_; ++i)
      std::cout << "  " << i << "  "
		<< concepts::Vector<Real>(stiff_.spaceX(), v+i*n)
		<< std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
