// testsuite for inexact inverse iteration eigensolver

#include <cmath>
#include <limits>
#include <iostream>
#include <unistd.h> // for command line parsing
#include "space/space.hh"
#include "operator/sparseMatrix.hh"
#include "eigensolver/inexactInv.hh"

using concepts::Real;

int main(int argc, char** argv) {

  try {
    uint dim = 10;
    int opt;
    while ((opt = getopt(argc, argv, "-d:")) != EOF)
      switch(opt) {
      case 'd': dim = atoi(optarg); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-d DIM]" << std::endl
		  << "where" << std::endl
		  << "  DIM: dimension of problem (default: " << dim << ')'
		  << std::endl;
	exit(1); break;
      }

    concepts::DummySpace<Real> spc(dim);
    std::cout << "Space: " << spc << std::endl;

    // create and fill mass and stiff
    concepts::SparseMatrix<Real> stiff_(spc, spc), mass_(spc, spc);
    const Real h = 1.0/((Real)dim+1.0);
    stiff_(0,0) = 2.0/h; stiff_(0,1) = -1.0/h;
    for (uint i = 1; i < dim-1; ++i) {
      stiff_(i, i-1) = -1.0/h; stiff_(i,i) = 2.0/h; stiff_(i,i+1) = -1.0/h;
    }
    stiff_(dim-1,dim-2) = -1.0/h; stiff_(dim-1,dim-1) = 2.0/h;
    mass_(0,0) = 4.0*h/6.0; mass_(0,1) = h/6.0;
    for (uint i = 1; i < dim-1; ++i) {
      mass_(i, i-1) = h/6.0; mass_(i,i) = 4.0*h/6.0; mass_(i,i+1) = h/6.0;
    }
    mass_(dim-1,dim-2) = h/6.0; mass_(dim-1,dim-1) = 4.0*h/6.0;
    std::cout << "Mass matrix: " << mass_ << std::endl;
    std::cout << "Stiffness matrix: " << stiff_ << std::endl;

    // call Eigen solver
    concepts::Vector<Real> sol(spc);
    sol = 1;
    Real shift = 5;
    concepts::LiCo<Real> shifted(stiff_, mass_, 1.0, -shift);
    eigensolver::InexactInv<Real> evsolver(shifted, mass_, sol);
    std::cout << "Eigenvalues: " << evsolver.getEV() << std::endl;
    std::cout << "Eigenvectors: " << std::endl;
    const concepts::Array<concepts::Vector<Real>*>& ef = evsolver.getEF();
    for (int i = 0; i < 1; ++i)
      std::cout << "  " << i << "  " << *(ef[i]) << std::endl;
    std::cout << evsolver << std::endl;
    mass_(*(ef[0]), sol);
    std::cout << "Norm of Eigenvector: " << *(ef[0]) * sol << std::endl;

    std::cout << "Solve a second time with proper starting vectors:"
	      << std::endl;
    sol = *(ef[0]);
    eigensolver::InexactInv<Real> evsolver2(shifted, mass_, sol);
    std::cout << "Eigenvalues: " << evsolver2.getEV() << std::endl;
    std::cout << evsolver2 << std::endl;

    std::cout << "Solve a third time with a different shift:"
	      << std::endl;
    shift = M_PI*M_PI; sol = 1;
    concepts::LiCo<Real> shifted2(stiff_, mass_, 1.0, -shift);
    eigensolver::InexactInv<Real> evsolver3(shifted2, mass_, sol);
    std::cout << "Eigenvalues: " << evsolver3.getEV() << std::endl;
    std::cout << evsolver3 << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
