/* 2D Cells, ie. triangles and quads
 * Geometrical information (coordinates), brings element maps and
 * topology together
 */

#include <sstream>
#include "basics/exceptions.hh"
#include "basics/debug.hh"
#include "basics/level.hh"
#include "elementMaps.hh"
#include "cell2D.hh"

// debugging Triangle2d
#define Triangle2dConstr_D 0
#define Triangle2dChi_D 0
#define Triangle2djacobian_D 0
#define Quad2dChild_D 0

namespace concepts {

  // ************************************************************ Triangle2d **

  Triangle2d::Index::Index() : l_(0), i_(0), j_(0) {}

  Triangle2d::Index::Index(uchar l, ushort i, ushort j) :
    l_(l), i_(i), j_(j) {}

  std::ostream& Triangle2d::Index::info(std::ostream& os) const {
    return os << '(' << (uint)l_ << ", " << i_ << ", " << j_ << ')';
  }

  Triangle2d::Triangle2d(Triangle& cntr, MappingTriangle2d* map,
			 const Index& idx) :
    idx_(idx), cntr_(cntr), lnk_(0), chld_(0), map_(map) {
    DEBUGL(Triangle2dConstr_D, "map = " << *map_);
  }

  Triangle2d::Triangle2d(Triangle& cntr, 
			 const MappingTriangle2d& map) : 
    cntr_(cntr), lnk_(0), chld_(0), map_(map.clone()) {
    DEBUGL(Triangle2dConstr_D, "map = " << *map_);
  }

  Triangle2d::~Triangle2d() {

    if (idx_.l_ == 0) delete map_;

    while (chld_ != 0) {
      Triangle2d* foo = chld_; chld_ = chld_->lnk_;
      delete foo;
    }
  }

  Triangle2d* Triangle2d::child(uint i) {
    
    if (i > 3) return 0;
  
    if (chld_ == 0) {
      uint l = idx_.l_ + 1;
      uint i = idx_.i_ << 1;
      uint j = idx_.j_ << 1;
      int  d0 = (i < j) ? 0 : 1;
      int  d1 = d0 ^ 1;

      Triangle2d** chld = &chld_;

      *chld = new Triangle2d(*cntr_.child(0), map_, Index(l, i, j));
      chld = &(*chld)->lnk_;

      i += d0; j -= d1;
      *chld = new Triangle2d(*cntr_.child(1), map_, Index(l, i, j));
      chld = &(*chld)->lnk_;

      j += d0; i -= d1; 
      *chld = new Triangle2d(*cntr_.child(2), map_, Index(l, i, j));
      chld = &(*chld)->lnk_;

      i += d0; j -= d1;
      *chld = new Triangle2d(*cntr_.child(3), map_, Index(l, j, i));
    }

    Triangle2d* chld = chld_;
    while (i--) chld = chld->lnk_;

    return chld;
  }

  const Triangle2d* Triangle2d::child(uint i) const {
    if (i > 3) return 0;
    Triangle2d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;
    return chld;
  }

  std::ostream& Triangle2d::info(std::ostream& os) const {

    Real2d vtx0(chi(0.0, 0.0));
    Real2d vtx1(chi(1.0, 0.0));
    Real2d vtx2(chi(1.0, 1.0));

    os << "Triangle2d(idx = " << idx_;
    os << ", cntr = " << cntr_;
    os << ", vtx = [" << vtx0 << ", " << vtx1 << ", " << vtx2 << "], map " 
       << *map_ << ")"; 

    return os;
  }

  Real2d Triangle2d::chi(Real xi, Real eta) const {

    const Real h = 1 << idx_.l_;

    eta *= xi;

    if (idx_.i_ < idx_.j_) {
      xi = idx_.j_ - xi; eta = idx_.i_ - eta;
    } else {
      xi = idx_.i_ + xi; eta = idx_.j_ + eta;
    }

#if Triangle2dChi_D
    Real2d res((*map_)(xi / h, eta / h));
    if (fabs(res.y()) < 0.001)
      DEBUGL(Triangle2dChi_D, "x = " << res);
    return res;
#else
    return (*map_)(xi / h, eta / h);
#endif
  }

  Real Triangle2d::jacobianDeterminant(const Real x, const Real y) const {
    return jacobian(x, y).determinant();
  }

  MapReal2d Triangle2d::jacobian(Real x, Real y) const {
    const Real h = 1 << idx_.l_;
    DEBUGL(Triangle2djacobian_D, "x = " << x << ", y = " << y 
	   << ", h = " << h);
    y *= x;
    if (idx_.i_ < idx_.j_) {
      x = idx_.j_ - x; y = idx_.i_ - y;
    } else {
      x = idx_.i_ + x; y = idx_.j_ + y;
    }
    MapReal2d j(map_->jacobian(x/h, y/h));
    j *= (1.0/h);
    DEBUGL(Triangle2djacobian_D, "(" << x/h << ", " << y/h << ") -> " << j);
    return j;
  }

  MapReal2d Triangle2d::jacobianInverse(const Real x, const Real y) const {
    return jacobian(x, y).inverse();
  }

  Real Triangle2d::area() const {
    Real2d v0 = (*map_)(0.0, 0.0);
    Real2d v1 = (*map_)(1.0, 0.0);
    Real2d v2 = (*map_)(1.0, 1.0);
    Real fullArea = ( (v1[0]-v0[0]) * (v2[1]-v0[1]) - 
                  (v2[0]-v0[0]) * (v1[1]-v0[1]) ) / 2;
    return fullArea / (1 << (2*idx_.l_));
  }

  Real3d Triangle2d::vertex(uint i) const {

    switch(i) {
    case 0: return chi(0.0, 0.0);
    case 1: return chi(1.0, 0.0);
    case 2: return chi(1.0, 1.0);
    default: throw conceptsException(ExceptionBase());
    }
  }

  Real2d Triangle2d::duffyInv(const Real2d& x) {
    return Real2d(x[0], (x[0] != 0.0 ? x[1]/x[0] : 0));
  }

  std::ostream& operator<<(std::ostream& os, const Triangle2d::Index& i) {
    return i.info(os);
  }

  // ***************************************************** Quad2dSubdivision **

  void Quad2dSubdivision::removeChildren(Quad2d& q) const {
    while (q.chld_ != 0) {
      Quad2d* foo = q.chld_; q.chld_ = q.chld_->lnk_;
      delete foo;
    }
  }

  // ********************************************************* Quad2dSubdiv4 **

  std::auto_ptr<Quad2dSubdiv4> Quad2dSubdiv4::instance_(0);

  void Quad2dSubdiv4::createChildren(Quad2d& q) const {
    if (q.chld_ == 0) {
      const uint k = q.idx_.level_.l_[0] + 1;
      const uint l = q.idx_.level_.l_[1] + 1;
      const uint i = q.idx_.i_ << 1;
      const uint j = q.idx_.j_ << 1;

      Quad2d** chld = &(q.chld_);

      *chld = new Quad2d(*(q.cntr_.child(0)), q.map_,
			 Quad2d::Index(k, l, i, j));
      chld = &(*chld)->lnk_;

      *chld = new Quad2d(*(q.cntr_.child(1)), q.map_,
			 Quad2d::Index(k, l, i+1, j));
      chld = &(*chld)->lnk_;

      *chld = new Quad2d(*(q.cntr_.child(2)), q.map_,
			 Quad2d::Index(k, l, i+1, j+1));
      chld = &(*chld)->lnk_;

      *chld = new Quad2d(*(q.cntr_.child(3)), q.map_,
			 Quad2d::Index(k, l, i, j+1));
    }
  }

  const QuadSubdiv4* Quad2dSubdiv4::topologicalStrategy() const {
    return QuadSubdiv4::instance();
  }

  const Quad2dSubdiv4* Quad2dSubdiv4::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Quad2dSubdiv4());
    return instance_.get();
  }

  Quad2dSubdiv4::~Quad2dSubdiv4() {}

  std::ostream& Quad2dSubdiv4::info(std::ostream& os) const {
    return os << "Quad2dSubdiv4()";
  }

  // ******************************************************** Quad2dSubdiv2H **

  std::auto_ptr<Quad2dSubdiv2H> Quad2dSubdiv2H::instance_(0);

  void Quad2dSubdiv2H::createChildren(Quad2d& q) const {
    if (q.chld_ == 0) {
      const uint k = q.idx_.level_.l_[0];
      const uint l = q.idx_.level_.l_[1] + 1;
      const uint i = q.idx_.i_;
      const uint j = q.idx_.j_ << 1;

      Quad2d** chld = &(q.chld_);

      *chld = new Quad2d(*(q.cntr_.child(0)), q.map_,
			 Quad2d::Index(k, l, i, j));
      chld = &(*chld)->lnk_;

      *chld = new Quad2d(*(q.cntr_.child(1)), q.map_,
			 Quad2d::Index(k, l, i, j+1));
    }
  }

  const QuadSubdiv2H* Quad2dSubdiv2H::topologicalStrategy() const {
    return QuadSubdiv2H::instance();
  }

  const Quad2dSubdiv2H* Quad2dSubdiv2H::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Quad2dSubdiv2H());
    return instance_.get();
  }

  Quad2dSubdiv2H::~Quad2dSubdiv2H() {}

  std::ostream& Quad2dSubdiv2H::info(std::ostream& os) const {
    return os << "Quad2dSubdiv2H()";
  }

  // ******************************************************** Quad2dSubdiv2V **

  std::auto_ptr<Quad2dSubdiv2V> Quad2dSubdiv2V::instance_(0);

  void Quad2dSubdiv2V::createChildren(Quad2d& q) const {
    if (q.chld_ == 0) {
      const uint k = q.idx_.level_.l_[0] + 1;
      const uint l = q.idx_.level_.l_[1];
      const uint i = q.idx_.i_ << 1;
      const uint j = q.idx_.j_;

      Quad2d** chld = &(q.chld_);

      *chld = new Quad2d(*(q.cntr_.child(0)), q.map_,
			 Quad2d::Index(k, l, i, j));
      chld = &(*chld)->lnk_;

      *chld = new Quad2d(*(q.cntr_.child(1)), q.map_,
			 Quad2d::Index(k, l, i+1, j));
    }
  }

  const QuadSubdiv2V* Quad2dSubdiv2V::topologicalStrategy() const {
    return QuadSubdiv2V::instance();
  }

  const Quad2dSubdiv2V* Quad2dSubdiv2V::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Quad2dSubdiv2V());
    return instance_.get();
  }

  Quad2dSubdiv2V::~Quad2dSubdiv2V() {}

  std::ostream& Quad2dSubdiv2V::info(std::ostream& os) const {
    return os << "Quad2dSubdiv2V()";
  }

  // **************************************************************** Quad2d **

  std::ostream& Quad2d::Index::info(std::ostream& os) const {
    return os << level_ << "; " << i_ << ", " << j_ << ')';
  }

  Quad2d::Quad2d(Quad& cntr, MappingQuad2d* map, const Index& idx) :
    idx_(idx), cntr_(cntr), lnk_(0), chld_(0), map_(map), subdivStrategy_(0) {}

  Quad2d::Quad2d(Quad& cntr, const MappingQuad2d& map) :
    cntr_(cntr), lnk_(0), chld_(0), map_(map.clone()),
    subdivStrategy_(0) {}

  Quad2d::~Quad2d() {
    if (idx_.level_ == Level<2>(0)) delete map_;
    if (subdivStrategy_)
      subdivStrategy_->removeChildren(*this);
  }

  Quad2d* Quad2d::child(uint i) {
    if (i > 3) return 0;

    setStrategy();
    DEBUGL(Quad2dChild_D, *this  << " - i = " << i);
    if (chld_ == 0)
      subdivStrategy_->createChildren(*this);

    Quad2d* chld = chld_;
    while (i--) {
      if (chld) {
	chld = chld->lnk_;
      }
      else return 0;
    }

    return chld;
  }

  const Quad2d* Quad2d::child(uint i) const {
    if (i > 3) return 0;
    Quad2d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;
    return chld;
  }

  Real2d Quad2d::chi(Real xi, Real eta) const {
    conceptsAssert(0 <= xi && xi <= 1, Assertion());
    conceptsAssert(0 <= eta && eta <= 1, Assertion());
    const Real hx = 1.0 / (1 << idx_.level_.l_[0]);
    const Real hy = 1.0 / (1 << idx_.level_.l_[1]);
    xi += idx_.i_; eta += idx_.j_; 
    return (*map_)(hx * xi, hy * eta);
  }

  MapReal2d Quad2d::jacobian(const Real xi, const Real eta) const {
    conceptsAssert(0 <= xi && xi <= 1, Assertion());
    conceptsAssert(0 <= eta && eta <= 1, Assertion());
    const Real hx = 1.0 / (1 << idx_.level_.l_[0]);
    const Real hy = 1.0 / (1 << idx_.level_.l_[1]);
    const Real x(xi + idx_.i_), y(eta + idx_.j_);
    return map_->jacobian(x * hx, y * hy).scaleCols(Real2d(hx, hy));
  }

  Real3d Quad2d::vertex(uint i) const {
    switch(i) {
    case 0: return chi(0.0, 0.0);
    case 1: return chi(1.0, 0.0);
    case 2: return chi(1.0, 1.0);
    case 3: return chi(0.0, 1.0);
    default: throw conceptsException(ExceptionBase());
    }
  }

  void Quad2d::setStrategy(const Quad2dSubdivision* strategy)
    throw(StrategyChange) {
    if ( (subdivStrategy_) && (strategy == 0) )
      return;
    if ( (strategy) && (subdivStrategy_) && (chld_) ) {
      conceptsAssert3(false, Assertion(), "illegal change of strategy: "
		      << "old: " << *subdivStrategy_
		      << ", new: " << *strategy << " in " << *this);
    }
    if (! strategy) {
      // Assert that the topological strategy is not set yet.
      // FIXME: Later, if the topological strategy is already set
      // and cannot be changed, the default geometrical strategy
      // should respect this.
      conceptsAssert((connector().getStrategy() == 0) ||
		     (connector().getStrategy() == QuadSubdiv4::instance()),
		     Assertion());
      strategy = Quad2dSubdiv4::instance();
    }
    subdivStrategy_ = strategy;
    connector().setStrategy(subdivStrategy_->topologicalStrategy());
  }

  std::ostream& Quad2d::info(std::ostream& os) const {
    Real2d vtx0(chi(0.0, 0.0));
    Real2d vtx1(chi(1.0, 0.0));
    Real2d vtx2(chi(1.0, 1.0));
    Real2d vtx3(chi(0.0, 1.0));

    os << "Quad2d(idx = " << idx_;
    os << ", cntr = " << cntr_;
    os << ", vtx = [" << vtx0 << ", " << vtx1 << ", " << vtx2 << ", " << vtx3;
    os << "])"; 

    return os;
  }

  std::ostream& operator<<(std::ostream& os, const Quad2d::Index& i) {
    return i.info(os);
  }

  // ************************************************************ Triangle3d **

  Triangle3d::Index::Index() : l_(0), i_(0), j_(0) {}

  Triangle3d::Index::Index(uchar l, ushort i, ushort j) :
    l_(l), i_(i), j_(j) {}

  std::ostream& Triangle3d::Index::info(std::ostream& os) const {
    return os << '(' << (uint)l_ << ", " << i_ << ", " << j_ << ')';
  }

  Triangle3d::Triangle3d(Triangle& cntr, MapTriangle3d* map,
			 const Index& idx) :
    idx_(idx), cntr_(cntr), lnk_(0), chld_(0), map_(map) {}

  Triangle3d::Triangle3d(Triangle& cntr, const MapTriangle3d& map) :
    cntr_(cntr), lnk_(0), chld_(0), map_(new MapTriangle3d(map)) {}

  Triangle3d::~Triangle3d() {

    if (idx_.l_ == 0) delete map_;

    if (chld_ != 0) 
      for(uint i = 4; i--;) {
	Triangle3d* foo = chld_; chld_ = chld_->lnk_;
	delete foo;
      }
  }

  Triangle3d* Triangle3d::child(uint i) {

    if (i > 3) return 0;

    if (chld_ == 0) {
      uint l = idx_.l_ + 1;
      uint i = idx_.i_ << 1;
      uint j = idx_.j_ << 1;
      int  d0 = (i < j) ? 0 : 1;
      int  d1 = d0 ^ 1;

      Triangle3d** chld = &chld_;

      *chld = new Triangle3d(*cntr_.child(0), map_, Index(l, i, j));
      chld = &(*chld)->lnk_;

      i += d0; j -= d1;
      *chld = new Triangle3d(*cntr_.child(1), map_, Index(l, i, j));
      chld = &(*chld)->lnk_;

      j += d0; i -= d1; 
      *chld = new Triangle3d(*cntr_.child(2), map_, Index(l, i, j));
      chld = &(*chld)->lnk_;

      i += d0; j -= d1;
      *chld = new Triangle3d(*cntr_.child(3), map_, Index(l, j, i));
    }

    Triangle3d* chld = chld_;
    while (i--) chld = chld->lnk_;

    return chld;
  }

  const Triangle3d* Triangle3d::child(uint i) const {
    if (i > 3) return 0;
    Triangle3d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;
    return chld;
  }

  std::ostream& Triangle3d::info(std::ostream& os) const {

    Real3d vtx0 = chi(0.0, 0.0);
    Real3d vtx1 = chi(1.0, 0.0);
    Real3d vtx2 = chi(1.0, 1.0);

    os << "Triangle3d(idx = " << idx_;
    os << ", cntr = " << cntr_;
    os << ", vtx = [" << vtx0 << ", " << vtx1 << ", " << vtx2 << "])"; 

    return os;
  }

  Real3d Triangle3d::chi(Real xi, Real eta) const {

    Real h = 1 << idx_.l_;

    if (idx_.i_ < idx_.j_) {
      xi = idx_.j_ - xi; eta = idx_.i_ - eta;
    } else {
      xi = idx_.i_ + xi; eta = idx_.j_ + eta;
    }

    return xi == 0.0 ? (*map_)(0.0, 0.0) : (*map_)(xi / h, eta / xi);
  }

  Real3d Triangle3d::vertex(uint i) const {

    switch(i) {
    case 0: return chi(0.0, 0.0);
    case 1: return chi(1.0, 0.0);
    case 2: return chi(1.0, 1.0);
    default: throw conceptsException(ExceptionBase());
    }
  }

  std::ostream& operator<<(std::ostream& os, const Triangle3d::Index& i) {
    return i.info(os);
  }

} // namespace concepts
