/* 3D cells, ie. hexahedrons and tetrahedrons
 * Geometrical information (coordinates), brings element maps and
 * topology together
 */

#ifndef cell3D_hh
#define cell3D_hh

#include <iostream>
#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/level.hh"
#include "topology3D.hh"
#include "cell.hh"
#include "elementMaps3D.hh"

namespace concepts {

  // ****************************************************** NegativeJacobian **

  /** Exception which indicates that a negative Jacobian was found.

      @author Philipp Frauenfelder, 2001
  */
  class NegativeJacobian : public MissingFeature {
  public:
    /** Constructor.
	@param errMsg Name of the cell or map in which the
	              negative Jacobian occured
    */
    NegativeJacobian(const std::string& errMsg) throw();
  protected:
    virtual std::ostream& info(std::ostream& os) const throw();
  };

  // ********************************************************* Tetrahedron3d **

  /** A 3D cell: tetrahedron. If a tetrahedron is subdivided, eight
      new tetrahedra are created. This happens automatically, if a
      child is requested.
      @author Philipp Frauenfelder, 2000
      @ingroup geometry
  */
  class Tetrahedron3d : public Cell3 {
  public:
    /// Subclass of Tetrahedron3d representing its index.
    struct Index {
      /// Level of this element
      uchar  l_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from left to right.
      */
      ushort i_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from front to back.
      */
      ushort j_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from bottom to top.
      */
      ushort k_;

      inline Index() : l_(0), i_(0), j_(0), k_(0) {}
      inline Index(uchar l, ushort i, ushort j, ushort k) :
	l_(l), i_(i), j_(j), k_(k) {}
      inline std::ostream& info(std::ostream& os) const;
    };

    /** Constructor. Takes the connector \c cntr and the element map
	\c map and creates a cell.
	@param cntr Topological information of the tetrahedron
	@param map Element map of the tetrahedron
    */
    Tetrahedron3d(Tetrahedron& cntr, const MapTetrahedron3d& map);
    virtual ~Tetrahedron3d();
  
    virtual Tetrahedron3d* child(uint i);
    virtual const Tetrahedron3d* child(uint i) const;
    Tetrahedron& connector() const { return cntr_; }

    /** The element map. Maps a point from the unit cube $[0,1]^3$
	onto the element.

	@return Point in 3D in physical coordinates.
	@param xi \f$\in [0,1]\f$
	@param eta \f$\in [0,1]\f$
	@param zeta \f$\in [0,1]\f$
    */
    Real3d chi(Real xi, Real eta, Real zeta) const;
    Real3d center() const {return chi(0.5, 0.5, 0.5);};

    Real3d vertex(uint i) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Index of this element
    Index idx_;

    /// Reference to the tetrahedron connector (topology)
    Tetrahedron& cntr_;

    /// Pointer to a sibling
    Tetrahedron3d* lnk_;

    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Tetrahedron3d* chld_;

    /// Pointer to the element map
    MapTetrahedron3d* map_;

    /// Private constructor
    Tetrahedron3d(Tetrahedron& cntr, MapTetrahedron3d* map, 
		      const Index& idx);
  };

  inline std::ostream& Tetrahedron3d::Index::info(std::ostream& os) const {
    return os << '(' << (uint)l_ << ", " 
	      << i_ << ", " << j_ << ", " << k_ << ')';
  }

  // ****************************************************** Hex3dSubdivision **

  class Hexahedron3d;

  /** Interface for geometrical subdivision strategies for hexahedrons.

      Every specialization of Hex3dSubdivision should be a singleton
      whose instance can be get with a call to instance(). It
      implements the method how the children of a hexahedron are
      created. To have access to all members of the class Hexahedron3d,
      the specializations of Hex3dSubdivision should be friend to Hexahedron3d.

      @author Philipp Frauenfelder, 2002
      @ingroup geometry
  */
  class Hex3dSubdivision : public OutputOperator {
  public:
    /** If no children of h exist, they are created.
     */
    virtual void createChildren(Hexahedron3d& h) const = 0;
    /** If there are chilren, they are removed.
     */
    virtual void removeChildren(Hexahedron3d& h) const;
    /** Returns the respective topological subdivision strategy
     */
    virtual const HexSubdivision* topologicalStrategy() const = 0;
  };

  // ********************************************************** Hex3dSubdiv8 **

  /** Subdivision strategy for hexahedrons which generates 8 children.
      @author Philipp Frauenfelder, 2002
  */
  class Hex3dSubdiv8 : public Hex3dSubdivision {
  public:
    virtual void createChildren(Hexahedron3d& h) const;
    virtual const HexSubdiv8* topologicalStrategy() const;
    static const Hex3dSubdiv8* instance();
    virtual ~Hex3dSubdiv8();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Hex3dSubdiv8> instance_;
  };

  // ********************************************************* Hex3dSubdiv2x **

  /** Subdivision strategy for hexahedrons which generates 2 children.
      The cell is divided perpendicular to the x axis.
      @author Philipp Frauenfelder, 2002
  */
  class Hex3dSubdiv2x : public Hex3dSubdivision {
  public:
    virtual void createChildren(Hexahedron3d& h) const;
    virtual const HexSubdiv2x* topologicalStrategy() const;
    static const Hex3dSubdiv2x* instance();
    virtual ~Hex3dSubdiv2x();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Hex3dSubdiv2x> instance_;
  };

  // ********************************************************* Hex3dSubdiv2y **

  /** Subdivision strategy for hexahedrons which generates 2 children.
      The cell is divided perpendicular to the y axis.
      @author Philipp Frauenfelder, 2002
  */
  class Hex3dSubdiv2y : public Hex3dSubdivision {
  public:
    virtual void createChildren(Hexahedron3d& h) const;
    virtual const HexSubdiv2y* topologicalStrategy() const;
    static const Hex3dSubdiv2y* instance();
    virtual ~Hex3dSubdiv2y();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Hex3dSubdiv2y> instance_;
  };

  // ********************************************************* Hex3dSubdiv2z **

  /** Subdivision strategy for hexahedrons which generates 2 children.
      The cell is divided perpendicular to the z axis.
      @author Philipp Frauenfelder, 2002
  */
  class Hex3dSubdiv2z : public Hex3dSubdivision {
  public:
    virtual void createChildren(Hexahedron3d& h) const;
    virtual const HexSubdiv2z* topologicalStrategy() const;
    static const Hex3dSubdiv2z* instance();
    virtual ~Hex3dSubdiv2z();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Hex3dSubdiv2z> instance_;
  };

  // ********************************************************* Hex3dSubdiv4x **

  /** Subdivision strategy for hexahedrons which generates 4 children.
      @author Philipp Frauenfelder, 2002
  */
  class Hex3dSubdiv4x : public Hex3dSubdivision {
  public:
    virtual void createChildren(Hexahedron3d& h) const;
    virtual const HexSubdiv4x* topologicalStrategy() const;
    static const Hex3dSubdiv4x* instance();
    virtual ~Hex3dSubdiv4x();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Hex3dSubdiv4x> instance_;
  };

  // ********************************************************* Hex3dSubdiv4y **

  /** Subdivision strategy for hexahedrons which generates 4 children.
      @author Philipp Frauenfelder, 2002
  */
  class Hex3dSubdiv4y : public Hex3dSubdivision {
  public:
    virtual void createChildren(Hexahedron3d& h) const;
    virtual const HexSubdiv4y* topologicalStrategy() const;
    static const Hex3dSubdiv4y* instance();
    virtual ~Hex3dSubdiv4y();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Hex3dSubdiv4y> instance_;
  };

  // ********************************************************* Hex3dSubdiv4z **

  /** Subdivision strategy for hexahedrons which generates 4 children.
      @author Philipp Frauenfelder, 2002
  */
  class Hex3dSubdiv4z : public Hex3dSubdivision {
  public:
    virtual void createChildren(Hexahedron3d& h) const;
    virtual const HexSubdiv4z* topologicalStrategy() const;
    static const Hex3dSubdiv4z* instance();
    virtual ~Hex3dSubdiv4z();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Hex3dSubdiv4z> instance_;
  };

  // ********************************************************** Hexahedron3d **

  /** A 3D cell: hexahedron. This cell contains a reference to the topological
      information of the hexahedron in \c cntr_.
      If a hexahedron is subdivided, two, four or eight new hexahedra
      are created.  This happens automatically, if a child is
      requested, depending on the strategy that has been set.

      The faces of the hexahedron have the following local coordinates in
      the element:
      <ul>
      <li>face 0 (bottom): z = 0
      <li>face 1 (front): y = 0
      <li>face 2 (right): x = 1
      <li>face 3 (back): y = 1
      <li>face 4 (left): x = 0
      <li>face 5 (top): z = 1
      </ul>
      @ingroup geometry
  */
  class Hexahedron3d : public Cell3 {
    friend class Hex3dSubdivision;
    friend class Hex3dSubdiv8;
    friend class Hex3dSubdiv2x;
    friend class Hex3dSubdiv2y;
    friend class Hex3dSubdiv2z;
    friend class Hex3dSubdiv4x;
    friend class Hex3dSubdiv4y;
    friend class Hex3dSubdiv4z;
  public:
    /// Subclass of Hexahedron3d representing its index
    struct Index {
      /// Level of this element
      Level<3> level_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from left to right.
      */
      ushort i_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from front to back.
      */
      ushort j_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from bottom to top.
      */
      ushort k_;

      inline Index() : level_(0), i_(0), j_(0), k_(0) {}
      inline Index(uchar l, uchar m, uchar n, ushort i, ushort j, ushort k) :
	level_(l), i_(i), j_(j), k_(k) { level_.l_[1] = m; level_.l_[2] = n; }
      inline Index(const Index& i) : level_(i.level_), i_(i.i_), j_(i.j_),
				     k_(i.k_) {}
    };

    friend std::ostream& operator<<(std::ostream& os,
				    const Hexahedron3d::Index& i);

    /** Constructor. Takes the connector \c cntr and the element map
	\c map and creates a cell.
	@param cntr Topological information of the hexadron
	@param map Element map of the hexahedron
    */
    Hexahedron3d(Hexahedron& cntr, const MapHexahedron3d& map);
    virtual ~Hexahedron3d();
  
    virtual Hexahedron3d* child(uint i);
    virtual const Hexahedron3d* child(uint i) const;
    inline Hexahedron& connector() const { return cntr_; }

    /** The element map. Maps a point from the unit cube \f$[0,1]^3\f$
	onto the element.

	@return Point in 3D in physical coordinates.
	@param xi \f$\in [0,1]\f$
	@param eta \f$\in [0,1]\f$
	@param zeta \f$\in [0,1]\f$
    */
    Real3d chi(Real xi, Real eta, Real zeta) const;

    /** Computes the Jacobian for \c xi, \c eta, \c zeta \f$\in [0,1]\f$.
	The Jacobian of a cell in the initial mesh is computed by the
	element map. The Jacobian of a subdivided cell is computed as
	follows. Let \f$H : \hat K \rightarrow \hat K'\f$ be the subdivision
	map in the reference coordinates (0,1)<sup>3</sup> and
	\f$F_K : \hat K \rightarrow K\f$, \f$F_{K'} : \hat K \rightarrow K'\f$
	the element maps of K and K' respectively. Then,
	\f[F_K : \hat K' \rightarrow K' \quad \text{and} \quad
	F_K \circ H = F_{K'}. \f]

	The subdivision map H in reference coordinates is
	\f[H : \vec\xi \mapsto
	\begin{pmatrix} a & 0 & 0 \\ 0 & b & 0 \\ 0 & 0 & c \end{pmatrix}
	\vec \xi \f]
	where \f$a = 2^{-l_0}\f$, \f$b = 2^{-l_1}\f$, \f$c = 2^{-l_2}\f$
	\f$\vec l = (l_0, l_1, l_2)\f$ is the level of the cell with respect to
	the cell in the initial mesh. The derivative of H is
	\f$dH = \begin{pmatrix} a & 0 & 0 \\ 0 & b & 0 \\
	0 & 0 & c \end{pmatrix}\f$.

	Then, the Jacobian of \f$F_{K'}\f$ is
	\f[dF_{K'} = dF_K \circ H \cdot dH. \f]
	The part \f$dF_K \circ H\f$ (without H) is computed by the element map
	and the part \f$\cdot dH\f$ is computed in \c jacobian().
    */
    MapReal3d jacobian(const Real xi, const Real eta, const Real zeta) const;

    /** Returns the inverse of the Jacobian for
	\c xi, \c eta, \c zeta \f$\in [0,1]\f$
    */
    MapReal3d jacobianInverse(const Real xi, const Real eta,
			      const Real zeta) const {
      return jacobian(xi, eta, zeta).inverse();
    }

    /** Returns the determinant of the Jacobian for
	\c xi, \c eta, \c zeta \f$\in [0,1]\f$
    */
    Real jacobianDeterminant(const Real xi, const Real eta,
			     const Real zeta) const {
      return jacobian(xi, eta, zeta).determinant();
    }

    /// Returns the level of the cell
    const Level<3>& level() const { return idx_.level_; }

    Real3d vertex(uint i) const;
    Real3d center() const {return chi(0.5, 0.5, 0.5);};
    /// Returns the element map
    inline const MapHexahedron3d* map() const {return map_;};

    /** Sets the subdivision strategy of this hexahedron.

        If the parameter is set to 0 (or if the method is called
	without parameter) the strategy is set to the default (if not
	already set). The default subdivision strategy is subdivision
	into 8 children.

        @param strategy Pointer to an instance of a subdivision strategy.
	@throws StrategyChange if the change is not allowed (the change is
	        not allowed if there are children present or the topological
		strategy is set).
    */
    void setStrategy(const Hex3dSubdivision* strategy = 0)
      throw(StrategyChange);

    /** Returns the subdivision strategy of this hexahedron

        If you want to find check against another strategy use
	<pre>hex.getStrategy() == Hex3dSubdiv4x::instance()</pre>
    */
    const Hex3dSubdivision* getStrategy() const { return subdivStrategy_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Index of this element
    Index idx_;

    /// Reference to the hexahedron connector (topology)
    Hexahedron& cntr_;

    /// Pointer to a sibling
    Hexahedron3d* lnk_;

    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Hexahedron3d* chld_;

    /// Pointer to the element map
    MapHexahedron3d* map_;

    /// Subdivision strategy for the hexahedron
    const Hex3dSubdivision* subdivStrategy_;

    /// Private constructor
    Hexahedron3d(Hexahedron& cntr, MapHexahedron3d* map, 
		 const Index& idx);
  };

  // ****************************************************** Parallelepiped3d **

  /** A 3D cell: parallelepiped. If a parallelepiped is subdivided,
      eight new parallelepipeds are created.  This happens
      automatically, if a child is requested.

      @author Gregor Schmidlin, 2003
      @ingroup geometry
  */
  class Parallelepiped3d : public Cell3 {
  public:
    /// Subclass of Tetrahedron3d representing its index.
    struct Index {
      /// Level of this element
      uchar  l_;

      /** Index of this element on the local level with respect to the
	  element and the x axis. Counting starts at 0 and goes
	  from left to right.
      */
      ushort i_;

      /** Index of this element on the local level with respect to the
	  element and the y axis. Counting starts at 0 and goes
	  from front to back.
      */
      ushort j_;

      /** Index of this element on the local level with respect to the
	  element and the z axis. Counting starts at 0 and goes
	  from bottom to top.
      */
      ushort k_;

      inline Index() : l_(0), i_(0), j_(0), k_(0) {}
      inline Index(uchar l, ushort i, ushort j, ushort k) :
	l_(l), i_(i), j_(j), k_(k) {}
      inline std::ostream& info(std::ostream& os) const;
    };

    /** Constructor. Takes the connector \c cntr and the element map
	\c map and creates a cell.
	@param cntr Topological information of the tetrahedron
	@param map Element map of the tetrahedron
    */
    Parallelepiped3d(Hexahedron& cntr, const MapParallelepiped3d& map);
    virtual ~Parallelepiped3d();
  
    virtual Parallelepiped3d* child(uint c);
    virtual const Parallelepiped3d* child(uint c) const;
    Hexahedron& connector() const { return cntr_; }

    /** The element map. Maps a point from the unit cube $[0,1]^3$
	onto the element.

	@return Point in 3D in physical coordinates.
	@param xi \f$\in [0,1]\f$
	@param eta \f$\in [0,1]\f$
	@param zeta \f$\in [0,1]\f$
    */
    Real3d chi(Real xi, Real eta, Real zeta) const;
    Real3d center() const {return chi(0.5, 0.5, 0.5);};

    Real3d vertex(uint i) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Index of this element
    Index idx_;

    /// Reference to the tetrahedron connector (topology)
    Hexahedron& cntr_;

    /// Pointer to a sibling
    Parallelepiped3d* lnk_;

    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Parallelepiped3d* chld_;

    /// Pointer to the element map
    MapParallelepiped3d* map_;

    /// Private constructor
    Parallelepiped3d(Hexahedron& cntr, MapParallelepiped3d* map, 
		     const Index& idx);
  };

  inline std::ostream& Parallelepiped3d::Index::info(std::ostream& os) const {
    return os << '(' << (uint)l_ << ", " 
	      << i_ << ", " << j_ << ", " << k_ << ')';
  }

} // namespace concepts

#endif // cell3D_hh
