/* Connectors.
 * connectivity informations.
 */

#ifndef Connector_hh
#define Connector_hh

#include "basics/typedefs.hh"
#include "basics/outputOperator.hh"
#include <iostream>

namespace concepts {

  // ************************************************************* Attribute **

  /** Attributes for elements of the topology. These attributes can
      be used to fix boundary conditions or coefficients in certain
      areas.
      @ingroup geometry
  */
  class Attribute { 
    friend std::ostream& operator<<(std::ostream& os, const Attribute& a);
  public:
    /** Constructor.

	@param attrib Attribute, default value is 0.
    */
    inline Attribute(uint attrib = 0) : attrib_(attrib) {}

    /// Comparison operator
    int operator==(const Attribute& a) const { return attrib_ == a.attrib_; }

    /// Returns the attribute
    inline operator uint() const { return attrib_; }
         
    /// Returns the attribute
    inline uint attrib() const { return attrib_; }
  private:
    /// Attribute
    uint attrib_; 
  };

  // ******************************************************************* Key **

  /** Key class. Used for the enumeration of topological elements (connectors).
      It might be usefull to have topological elements with a key with more
      features...
  */
  class Key { 
    friend std::ostream& operator<<(std::ostream& os, const Key& k);
  public:
    /// Default constructor
    inline Key(uint key = 0) : key_(key) {}
         
    /// Comparison operator
    inline int operator==(const Key& key) const { return key_ == key.key_; }

    /// Returns the key
    inline operator uint() const { return key_; }
         
    /// Returns the key
    inline uint key() const { return key_; }

    /// Returns a key which never occurs
    inline static uint noKey() { return (uint)-1; }
  private:
    uint key_; 
  };

  // ************************************************************* Connector **

  /** An abstract class for elements of the topology.
      The topology contains the informations about the inter-element
      connections.

      This class prescribes the common interface for all topological 
      elements (connectors). This interface consists of a method to query 
      the children of a connector.
      @ingroup geometry
  */
  class Connector : public OutputOperator {
  public:
    /// Constructor
    Connector(const uint key, const Attribute attrib) :
      key_(key), attrib_(attrib) {}

    /** Returns the ith child of the connector.
	@param i Index of the child
	@param mode If mode = false(0) the child is created if it does not
      	            yet exist, if mode = true(1) no new children are created
		    (ie. 0 is returned if no children exist).
        @invariant Either all children or no children exist.
    */
    virtual Connector* child(uint i, bool mode = 0) = 0;
    virtual const Connector* child(uint i) const = 0;

    /// Comparison operator
    int operator==(const Connector& cntr) const { return key_ == cntr.key_; }

    /// Returns the key of the connector
    const Key& key() const { return key_; }

    /// Returns the attribute of the connector
    const Attribute& attrib() const { return attrib_; }

    /** Checks if this connector is related to \c conn. It is being checked if
	one of the two is an ancestor of the other.
	@param conn Other connector
	@return An integer i.<br>
	  i=0 if there is no relationship or the connectors are identical<br>
	  i>0 if \c conn is an ancestor of this connector<br>
	  i<0 if \c conn is a descendant of this connector<br>
	  |i| gives the number of generations between the connectors
    */
    int related(const Connector& conn) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;

    /// Unique key of the connector
    Key key_;

    /// Attribute
    Attribute attrib_;
  };

  // ************************************************************ Connector0 **

  /** A 0D element of the topology. Base class for a vertex.

      On this level a unique counter for all 0D connectors is 
      implemented.
  */
  class Connector0 : public Connector {
  public:
    /** Constructor.

	@param attrib The attribute
    */
    inline Connector0(Attribute attrib) : Connector(cnt_++, attrib) {}

    virtual Connector0* child(uint i, bool mode = 0) = 0;
    virtual const Connector0* child(uint i) const = 0;
  protected:
    /** Constructor. Used by the clone function of inherited classes.
	@param attrib The attribute
	@param key The key
    */
    inline Connector0(const Attribute attrib, const Key& key)
      : Connector(key, attrib) {};

    /** Counter for the connectors.
	This counter is used to create the unique key of the connector.
    */
    static uint cnt_;
  };

  // ************************************************************ Connector1 **

  /** A 1D element of the topology. Base class for an edge.
      1D elements of the topology are defined by 0D elements, ie. an edge
      is defined by two vertices.

      On this level a unique counter for all 1D connectors is
      implemented. In addition to the interface prescribed by
      connector a method to query the 0D components of this connector
      is declared.
  */
  class Connector1 : public Connector {
  public:
    /** Constructor.

	@param attrib Attribute
    */
    inline Connector1(Attribute attrib) : Connector(cnt_++, attrib) {}

    /** Returns a 0D component: vertex
	A 1D element consists of 0D elements, 
	the vertices can be queried here.

	@param i Index of the vertex
    */
    virtual Connector0* vertex(uint i) const = 0;

    virtual Connector1* child(uint i, bool mode = 0) = 0;
    virtual const Connector1* child(uint i) const = 0;
  protected:
    /** Constructor. Used by the clone function of inherited classes.
	@param attrib The attribute
	@param key The key
    */
    inline Connector1(const Attribute attrib, const Key& key)
      : Connector(key, attrib) {};

    /** Counter for the connectors.
	This counter is used to create the unique key of the connector.
    */
    static uint cnt_;
  };

  // ************************************************************ Connector2 **

  /** A 2D element of the topology.
      Base class for a face (a triangle or a quadrilateral).  2D
      elements of the topology are defined by 1D elements, ie. a
      triangle is defined by three edges.

      On this level a unique counter for all 2D connectors is
      implemented. In addition to the interface prescribed by
      connector methods to query the 0D and 1D components of this
      connector are declared.
  */
  class Connector2 : public Connector {
  public:
    /** Constructor.

	@param attrib The attribute
    */
    inline Connector2(Attribute attrib) : Connector(cnt_++, attrib) {}

    virtual Connector2* child(uint i, bool mode = 0) = 0;
    virtual const Connector2* child(uint i) const = 0;

    /** Returns a 1D component: edge.
	A 2D element consists of 1D and 0D elements, 
	the edges can be queried here.

	@param i Index of the edge
    */
    virtual Connector1* edge(uint i) const = 0;

    /** Returns a 0D component: vertex
	A 2D element consists of 1D and 0D elements, 
	the vertices can be queried here.

	@param i Index of the vertex
    */
    virtual Connector0* vertex(uint i) const = 0;
  protected: 
    /** Constructor. Used by the clone function of inherited classes.
	@param attrib The attribute
	@param key The key
    */
    inline Connector2(const Attribute attrib, const Key& key)
      : Connector(key, attrib) {};

    /** Counter for the connectors.
	This counter is used to create the unique key of the connector.
    */
    static  uint cnt_;
  };

  // ************************************************************ Connector3 **

  /** A 3D element of the topology. Base class for tetrahedrons etc.
      3D elements of the topology are defined by 2D elements, ie. a
      tetrahedron is defined by four triangles.

      On this level a unique counter for all 3D connectors is 
      implemented. In addition to the interface prescribed by 
      connector methods to query the 0D, 1D and 2D components of
      this connector are declared.
  */
  class Connector3 : public Connector {
  public:
    /** Constructor.

	@param attrib The attribute
    */
    inline Connector3(Attribute attrib) : Connector(cnt_++, attrib) {}

    virtual Connector3* child(uint i, bool mode = 0) = 0;
    virtual const Connector3* child(uint i) const = 0;

    /** Returns a 1D component: edge.
	A 3D element consists of 2D, 1D and 0D elements, 
	the edges can be queried here.

	@param i Index of the edge
    */
    virtual Connector1* edge(uint i) const = 0;

    /** Returns a 2D component: face.
	A 3D element consists of 2D, 1D and 0D elements, 
	the faces can be queried here.

	@param i Index of the face
    */
    virtual Connector2* face(uint i) const = 0;

    /** Returns a 0D component: vertex
	A 3D element consists of 2D, 1D and 0D elements, 
	the vertices can be queried here.

	@param i Index of the vertex
    */
    virtual Connector0* vertex(uint i) const = 0;
  protected: 
    /** Constructor. Used by the clone function of inherited classes.
	@param attrib The attribute
	@param key The key
    */
    inline Connector3(const Attribute attrib, const Key& key)
      : Connector(key, attrib) {};

    /** Counter for the connectors.
	This counter is used to create the unique key of the connector.
    */
    static uint cnt_;
  };

} // namespace concepts

#endif // Connector_hh
