/* Element maps.
 * Geometrical information (coordinates)
 */

#include <cmath>
#include "elementMaps.hh"
#include "basics/functionCompiler.hh"
#include "basics/exceptions.hh"
#include "basics/debug.hh"

// debugging MapQuad2d
#define MapQuad2dConstr_D 0

// debugging VertexQuad2d
#define VertexQuad2djacobian_D 0
#define VertexQuad2dlineElement_D 0

// debugging VertexTriangle2d
#define VertexTriangleConstr_D 0
#define VertexTriangleDestr_D 0
#define VertexTriangleClone_D 0

// debugging MapTriangle3d
#define MapTriangle3dConstr_D 0

namespace concepts {

  // ************************************************************* MapEdge1d **

  std::ostream& MapEdge1d::info(std::ostream& os) const {
    return os << "MapEdge1d(" << vtx0_ << ", " << vtx1_ << ')';
  }

  // ***************************************************** MappingTriangle2d **

  Real MappingTriangle2d::jacobianDeterminant(const Real x,
					      const Real y) const {
    return jacobian(x, y).determinant();
  }

  MapReal2d MappingTriangle2d::jacobianInverse(const Real x,
					       const Real y) const {
    return jacobian(x, y).inverse();
  }

  // ****************************************************** VertexTriangle2d **

  VertexTriangle2d::VertexTriangle2d(Real2d vtx0, Real2d vtx1, Real2d vtx2) :
    vtx0_(vtx0), vtx1_(vtx1), vtx2_(vtx2),
    jacobian_((Real2d)(vtx1_-vtx0_), (Real2d)(vtx2_-vtx1_)) {
    DEBUGL(VertexTriangleConstr_D, "done.");
  }

  VertexTriangle2d::VertexTriangle2d(const VertexTriangle2d &map) :
    vtx0_(map.vtx0_), vtx1_(map.vtx1_), vtx2_(map.vtx2_),
    jacobian_(map.jacobian_) {
    DEBUGL(VertexTriangleConstr_D, "done.");
  }

  VertexTriangle2d::~VertexTriangle2d() {
    DEBUGL(VertexTriangleDestr_D, "done.");
  }

  VertexTriangle2d* VertexTriangle2d::clone() const {
    DEBUGL(VertexTriangleClone_D, "start");
    return new VertexTriangle2d(*this);
  }

  std::ostream& VertexTriangle2d::info(std::ostream& os) const {
    return os << "VertexTriangle2d(vtx0 = " << vtx0_ << ", vtx1 = " << vtx1_ 
	      << ", vtx2 = " << vtx2_ << ")";
  }

  bool VertexTriangle2d::curved() const { 
    return false;
  }

  // ********************************************************* MapTriangle2d **

  MapTriangle2d::MapTriangle2d(char* map, const Real scx, const Real scy) :
    map_(0), sz_(0), scx_(scx), scy_(scy) {

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_];
      std::memcpy(map_, foo, sz_);
    }
  }

  MapTriangle2d::MapTriangle2d(const MapTriangle2d &map) :
    map_(0), sz_(map.sz_), scx_(map.scx_), scy_(map.scy_) {

    if (map.map_) {
      map_ = new uchar[sz_]; std::memcpy(map_, map.map_, sz_);
    }
  }

  MapTriangle2d::~MapTriangle2d() { delete[] map_; }

  MapTriangle2d* MapTriangle2d::clone() const {
    return new MapTriangle2d(*this);
  }

  std::ostream& MapTriangle2d::info(std::ostream& os) const {
    return os << "MapTriangle2d()";
  }

  bool MapTriangle2d::curved() const { 
    return false;
  }

  MapReal2d MapTriangle2d::jacobian(const Real, const Real) const {
    throw conceptsException(MissingFeature("Jacobian not implemented"));
  }

  // ********************************************************* MappingQuad2d **

  Real MappingQuad2d::jacobianDeterminant(const Real x, const Real y) const {
    return jacobian(x, y).determinant();
  }

  MapReal2d MappingQuad2d::jacobianInverse(const Real x, const Real y) const {
    return jacobian(x, y).inverse();
  }

  // general implementation of factor of differential element  
  // integrating of an edge.
  Real MappingQuad2d::lineElement(const Real x, const uint edge) const {
    conceptsAssert(x >= 0.0, Assertion());
    conceptsAssert(x <= 1.0, Assertion());
    conceptsAssert(edge < 4, Assertion());

    MapReal2d jac;
    switch(edge){
    case 0 : jac = jacobian(x,0.0);break;  // bottom edge
    case 1 : jac = jacobian(1.0,x);break;  // right
    case 2 : jac = jacobian(x,1.0);break;  // top
    case 3 : jac = jacobian(0.0,x);break;  // left
    default: throw conceptsException(concepts::ExceptionBase());
    }
    Real tmp;
    switch(edge){
    case 0 :
    case 2 :
      tmp = std::sqrt( std::pow(jac(0,0),2) + std::pow(jac(1,0),2) );
      break;
    case 1 :
    case 3 :
      tmp = std::sqrt( std::pow(jac(0,1),2) + std::pow(jac(1,1),2) );
      break;
    default: throw conceptsException(concepts::ExceptionBase());
    }
    return tmp;
  }

  // ************************************************************* MapQuad2d **

  MapQuad2d::MapQuad2d(char* map, Real scx, Real scy, 
			       Real2d org, Real stretch) :
    map_(0), sz_(0), scx_(scx), scy_(scy), org_(org),
    scaling_(Real2d(stretch, stretch)) {

    DEBUGL(MapQuad2dConstr_D, map << " parameter region (" << scx_ << ", " 
	   << scy_ << "), origin " << org_ << ", scaling " << scaling_);

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_]; std::memcpy(map_, foo, sz_);
    }
  }

  MapQuad2d::MapQuad2d(char* map, Real scx, Real scy, Real2d org,
			       Real2d scaling) :
    map_(0), sz_(0), scx_(scx), scy_(scy), org_(org), scaling_(scaling) {

    DEBUGL(MapQuad2dConstr_D, map << " parameter region (" << scx_ << ", " 
	   << scy_ << "), origin " << org_ << ", scaling " << scaling_);

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_]; std::memcpy(map_, foo, sz_);
    }
  }

  MapQuad2d::MapQuad2d(const MapQuad2d &map) :
    map_(0), sz_(map.sz_), scx_(map.scx_), scy_(map.scy_), 
    org_(map.org_), scaling_(map.scaling_) {

    map_ = new uchar[sz_]; std::memcpy(map_, map.map_, sz_);

    DEBUGL(MapQuad2dConstr_D, "parameter region (" << scx_ << ", " 
	   << scy_ << "), origin " << org_ << ", scaling " << scaling_);
  }

  MapQuad2d::~MapQuad2d() { delete[] map_; }

  MapReal2d MapQuad2d::jacobian(const Real x, const Real y) const {
    throw conceptsException(MissingFeature
			    ("Jacobian of map not implemented"));
  }

  // ********************************************************** VertexQuad2d **

  VertexQuad2d::VertexQuad2d(Real2d vtx0, Real2d vtx1,
			     Real2d vtx2, Real2d vtx3) {
    vtx_[0] = vtx0; vtx_[1] = vtx1; vtx_[1] -= vtx0;
    vtx_[3] = vtx3; vtx_[3] -= vtx0;
    vtx_[2] = vtx_[1]; vtx_[2] *= -1.0; vtx_[2] += vtx2; vtx_[2] -= vtx3;
    hasLength_ = false;
    conceptsAssert(jacobian(0.5, 0.5).determinant() > 0,
		   concepts::Assertion());
  }

  VertexQuad2d::VertexQuad2d(const VertexQuad2d& v) {
    std::memcpy(vtx_, v.vtx_, 4*sizeof(Real2d));
  }

  VertexQuad2d::~VertexQuad2d() {}

  // mapping is v0 + v1*x + v3*y + v2*x*y
  Real2d VertexQuad2d::operator()(Real x, Real y) const {
    conceptsAssert((x >= 0.0) && (x <= 1.0), concepts::Assertion());
    conceptsAssert((y >= 0.0) && (y <= 1.0), concepts::Assertion());
    Real2d res(vtx_[0]);
    res += vtx_[1]*x;    // v1 = x1-x0
    res += vtx_[3]*y;    // v3 = x3-x0
    res += vtx_[2]*x*y;  // v2 = x0+x2-x1-x3
    return res;
  }

  MapReal2d VertexQuad2d::jacobian(const Real x, const Real y) const {
    DEBUGL(VertexQuad2djacobian_D, "x = (" << x << ", " << y << ")");

    conceptsAssert(x >= 0.0, Assertion());
    conceptsAssert(x <= 1.0, Assertion());
    conceptsAssert(y >= 0.0, Assertion());
    conceptsAssert(y <= 1.0, Assertion());

    Real2d first, second;

    // first column v1 + v2*y with v1 = x1-x0, v2 = x0+x2-x1-x3
    first.lincomb(vtx_[2], vtx_[1], y);
    // second column v3 + v2*x with v3 = x3-x0
    second.lincomb(vtx_[2], vtx_[3], x);
    DEBUGL(VertexQuad2djacobian_D, "jac = "
	   << MapReal2d(first, second));
    return MapReal2d(first, second);
  }

  // differential element is just length of edge
  Real VertexQuad2d::lineElement(const Real x, const uint edge) {
    conceptsAssert(x >= 0.0, Assertion());
    conceptsAssert(x <= 1.0, Assertion());
    conceptsAssert(edge < 4, Assertion());

    if (!hasLength_){
      len_[0] = vtx_[1].l2();            // bottom: v1 = x1-x0
      len_[1] = (vtx_[2]-vtx_[3]).l2();  // right:  v2-v3 = x2-x1
      len_[2] = (vtx_[2]-vtx_[1]).l2();  // top:    v2-v1 = x2-x3
      len_[3] = vtx_[3].l2();            // left:   v3 = x3 - x0
    }

    DEBUGL(VertexQuad2dlineElement_D, "edge = " << edge <<
	   ", line element ds = " << len_[edge] << " d" << (edge%2?"x":"y"));
    return len_[edge];
  }

  VertexQuad2d* VertexQuad2d::clone() const {
    return new VertexQuad2d(*this);
  }

  // ********************************************************* MapTriangle3d **

  MapTriangle3d::MapTriangle3d(char* map, Real scx, Real scy) :
    map_(0), sz_(0), scx_(scx), scy_(scy), B1_(0.0, 0.0, 0.0), 
    B2_(0.0, 0.0, 0.0), b_(0.0, 0.0, 0.0) {

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_]; std::memcpy(map_, foo, sz_);
    }
  }

  MapTriangle3d::MapTriangle3d(Real3d vtx0, Real3d vtx1, Real3d vtx2) :
    map_(0), sz_(0), scx_(1.0), scy_(1.0), B1_(vtx1-vtx0), B2_(vtx2-vtx1), 
    b_(vtx0) {

    DEBUGL(MapTriangle3dConstr_D, "B1 = " << B1_ << ", B2 = " << B2_ 
	   << ", b = " << b_);
  }

  MapTriangle3d::MapTriangle3d(const MapTriangle3d &map) :
    map_(0), sz_(map.sz_), scx_(map.scx_), scy_(map.scy_), B1_(map.B1_),
    B2_(map.B2_), b_(map.b_) {

    if (map.map_) {
      map_ = new uchar[sz_]; std::memcpy(map_, map.map_, sz_);
    }
  }

} // namespace concepts
