/* Element maps in 3D
 * Geometrical information (coordinates)
 */

#ifndef elementMaps3D_hh
#define elementMaps3D_hh

#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/debug.hh"

// debugging
#define geoMapTetrahedron3dAppl_D 0

namespace concepts {

  // ***************************************************************** Map3d **

  /** An abstract class for a 3d map
      @ingroup geometry
   */
  class Map3d : public OutputOperator {
  public:
    std::ostream& info(std::ostream& os) const {return os << "Map3d()";};
  };


  // ****************************************************** MapTetrahedron3d **

  /** A 3D element map for a tetrahedron. The reference element is the
      unit cube.

      There are two ways to fix the map of the tetrahedron: give the
      mapping explicitly or give the four corners and the map is computed
      internally.

      Giving the map explicitly needs a few precautions: The reference
      element is the unit cube. For easier construction of the element
      map, a rectangle \f$[0,a] \times [0,b] \times [0,c]\f$ can be given as
      the parameter domain. To map this onto a tetrahedron, two sides
      have to degenerate using a so called Duffy transformation.

      With a arbitrary Duffy transformation, one can choose the side
      which sould collapse. Here, the top face of the cube and the back
      edge of the bottom face should degenerate to a point, ie. the unit
      cube maps onto the unit simplex. This can be done with three Duffy
      transformations:
      <ol>
      <li>Collapse the top face against the 2 direction.
      <li>Collapse the resulting egde against the 1 direction.
      <li>Collapse the back edge of the bottom face against the 1 direction.
      </ol>

      When giving the four corners, the first three have to be in the
      normal orientation of the first triangle of the tetrahedron, see
      Tetrahedron for the explanation of "normal orientation".

      The application operator maps a point from the unit square onto
      the physical domain taking into account the parameter domain given
      in the constructor.
  */
  class MapTetrahedron3d : public Map3d {
  public:
    /** Constructor. The values of scX, scY and scZ are only for the map,
	they are not needed by the user for mapping a point from the 
	reference element onto the Real element.

	@param map The element map for this tetrehdron as a string, 
	x, y and z are the allowed variables, the component 
	are separated by a comma.
	@param scX The range of x is [0, scX].
	@param scY The range of y is [0, scY].
	@param scZ The range of z is [0, scZ].
    */
    MapTetrahedron3d(char* map, Real scX, Real scY, Real scZ);

    /** Constructor. Takes the four physical corners of the tetrahedron and
	computes the element map: \f$F_K : S \rightarrow K\f$ with 
	\f$\vec x = F_K(\xi) = B \cdot \vec \xi + \vec b\f$, where
	\f$\vec b = \mbox{vtx0}\f$ and \f$B = [ (\mbox{vtx1} - \vec b)
	(\mbox{vtx2} - \vec b) (\mbox{vtx3} - \vec b) ]\f$.
    */
    MapTetrahedron3d(Real3d vtx0, Real3d vtx1, Real3d vtx2, Real3d vtx3);

    /// Copy constructor
    MapTetrahedron3d(const MapTetrahedron3d &map);
    ~MapTetrahedron3d() { delete[] map_; }

    /** Returns a point in 3D mapped from the unit cube [0,1]<sup>3</sup>
	onto the element in the original mesh.

	If the map was given as formula in the constructor, the parameters
	are directly inserted into the formula. If the map was given by the
	four corners of the physical tetrahedron, then the point is first
	mapped from the unit cube (\f$\eta\f$) onto the unit simplex
	(\f$\xi\f$) using
	\f$\xi_1 = \eta_1(1-\eta_3)(1-\eta_2(1-\eta_3))\f$, 
	\f$\xi_2 = \eta_2(1-\eta_3)\f$ and \f$\xi_3 = \eta_3\f$.

	@return \f$F_K(x,y,z)\f$
	@param x \f$\in [0, 1]\f$
	@param y \f$\in [0, 1]\f$
	@param z \f$\in [0, 1]\f$
    */
    Real3d operator()(Real x, Real y, Real z) const { 
      if (map_)
	return Real3d(map_, scx_ * x, scy_ * y, scz_ * z);
      else {
	Real3d xi(x*(1.0-y)*(1.0-z), y*(1.0-z), z);
	Real3d res(B_ * xi);
	res += b_;
	DEBUGL(geoMapTetrahedron3dAppl_D, '(' << x << ", " << y 
	       << ", " << z << ") -> " << xi << " -> " << res);
	return res;
      }
    }

    /// Returns the jacobian
    Real jacobian() const {
      if (map_)
        throw conceptsException(MissingFeature("jacobian not supported"));
      return B_.determinant(); }

    /// Returns a copy of the map
    inline MapTetrahedron3d* clone() const {
      return new MapTetrahedron3d(*this);
    };

private:
    /// Parsed formula for the map
    uchar* map_;

    /// Length of the parsed formula
    uint sz_;

    /// Right border of the x parameter domain
    Real scx_;

    /// Right border of the y parameter domain
    Real scy_;

    /// Right border of the z parameter domain
    Real scz_;

    /// Computed map, matrix part
    MapReal3d B_;

    /// Computed map, vector part
    Real3d b_;

    /// Private assignement operator
    void operator=(const MapTetrahedron3d &);
  };


  // ******************************************************* MapHexahedron3d **

  /** A 3D element map for a hexahedron. The reference element is the
      unit cube.

      There are two ways to fix the map of the hexahedron: give the
      mapping explicitly or give the eight corners and the map is computed
      internally.

      When giving the eight corners of the hexahedron, they must be
      given in the correct order, ie. first the corners of the floor (in
      right screwed orientation pointing into the hexahedron) and then
      the corners of the ceiling (in right screwed orientation pointing
      out of the hexahedron).

      The application operator maps a point from the unit square onto
      the physical domain taking into account the parameter domain given
      in the constructor.
  */
  class MapHexahedron3d : public Map3d {

    /// Parsed formula for the map
    uchar* map_;

    /// Length of the parsed formula
    uint sz_;

    /// Right border of the x parameter domain
    Real scx_;

    /// Right border of the y parameter domain
    Real scy_;

    /// Right border of the z parameter domain
    Real scz_;

    /// Computed map
    Real3d *comp_;

    /// Private assignement operator
    void operator=(const MapHexahedron3d &);

  public:
    /** Constructor. The values of scX, scY and scZ are only for the map,
	they are not needed by the user for mapping a point from the 
	reference element onto the Real element.

	@param map The element map for this triangle as a string, 
	x, y and z are the allowed variables, the component 
	are separated by a comma.
	@param scX The range of x is [0, scX].
	@param scY The range of y is [0, scY].
	@param scZ The range of z is [0, scZ].
    */
    MapHexahedron3d(char* map, Real scX, Real scY, Real scZ);

    /** Constructor. Takes the four physical corners of the hexahedron and
	computes the element map: \f$F_K : S \rightarrow K\f$ with 
	\f$\vec x = F_K(\xi) = B \cdot \vec \xi + \vec b\f$, where
	\f$\vec b = \mbox{vtx0}\f$ and \f$B = [ (\mbox{vtx1} - \vec b) 
	(\mbox{vtx2} - \vec b) (\mbox{vtx3} - \vec b) ]\f$.
    */
    MapHexahedron3d(const Real3d& vtx0, const Real3d& vtx1, 
		    const Real3d& vtx2, const Real3d& vtx3,
		    const Real3d& vtx4, const Real3d& vtx5,
		    const Real3d& vtx6, const Real3d& vtx7);

    /// Copy constructor
    MapHexahedron3d(const MapHexahedron3d &map);
    ~MapHexahedron3d() { delete[] map_;  delete[] comp_; };

    /** Returns a point in 3D mapped from the unit cube \f$[0,1]^3\f$
	onto the element in the original mesh.

	@return \f$F_K(x,y,z)\f$
	@param x \f$\in [0, 1]\f$
	@param y \f$\in [0, 1]\f$
	@param z \f$\in [0, 1]\f$
    */
    inline Real3d operator()(Real x, Real y, Real z) const { 
      conceptsAssert(x >= 0.0, Assertion());
      conceptsAssert(x <= 1.0, Assertion());
      conceptsAssert(y >= 0.0, Assertion());
      conceptsAssert(y <= 1.0, Assertion());
      conceptsAssert(z >= 0.0, Assertion());
      conceptsAssert(z <= 1.0, Assertion());
      if (map_)
	return Real3d(map_, scx_ * x, scy_ * y, scz_* z);
      else {
	Real3d res(comp_[0]*(x*y*z));
	res += comp_[1]*(x*y);
	res += comp_[2]*(x*z);
	res += comp_[3]*(y*z);
	res += comp_[4]*x;
	res += comp_[5]*y;
	res += comp_[6]*z;
	res += comp_[7];
	return res;
      }
    }

    /** Returns the jacobian of the element map
	\f[ \left(\frac{\partial F_K}{\partial\vec\xi}\right) =
	\left\{\frac{\partial x^i}{\partial \xi_j}\right\}_{i,j=1}^3 \f]
    */
    MapReal3d jacobian(const Real x, const Real y, const Real z) const;

    /** Computes the inverse of the jacobian:
	\f[ \left(\frac{\partial F_K}{\partial\vec\xi}\right)^{-1} =
	\left\{\frac{\partial \xi_i}{\partial x^j}\right\}_{i,j=1}^3 \f]
    */
    MapReal3d jacobianInverse(const Real x, const Real y, const Real z) const {
      return jacobian(x, y, z).inverse();
    }

    /// Returns the determinant of the Jacobian
    Real jacobianDeterminant(const Real x, const Real y, const Real z) const {
      return jacobian(x, y, z).determinant();
    }

    /// Returns a copy of the map
    MapHexahedron3d* clone() const { return new MapHexahedron3d(*this); }
  };


  // *************************************************** MapParallelepiped3d **

  /** A 3D element map for a Parallelepiped. The reference element is the
      unit cube.

      Given four corners the map is computed internally. The origin
      of the coordinate system is mapped to the first corner. The
      other three corners are the image of the points (1,0,0), (0,1,0)
      and (0,0,1).
  */
  class MapParallelepiped3d : public Map3d {
  public:
    /** Constructor. Takes four physical corners of the parallelepiped and
	computes the element map: \f$F_K : S \rightarrow K\f$ with
	\f$\vec x = F_K(\xi) = B \cdot \vec \xi + \vec b\f$, where
	\f$\vec b = \mbox{vtx0}\f$ and \f$B = [(\mbox{vtx1} - \vec b)
	(\mbox{vtx2} - \vec b) (\mbox{vtx3} - \vec b)]\f$.
    */
    MapParallelepiped3d(Real3d vtx0, Real3d vtx1, Real3d vtx2, Real3d vtx3);

    /// Copy constructor
    MapParallelepiped3d(const MapParallelepiped3d &map);

    /** Returns a point in 3D mapped from the unit cube [0,1]<sup>3</sup>
	onto the element in the original mesh.

	@return \f$F_K(x,y,z)\f$
	@param x \f$\in [0, 1]\f$
	@param y \f$\in [0, 1]\f$
	@param z \f$\in [0, 1]\f$
    */
    Real3d operator()(Real x, Real y, Real z) const {
      Real3d res(B1_*x + B2_*y + B3_*z);
      res += b_;
      return res;
    }

    /// Returns the jacobian
    Real jacobian() const {
      return (B1_ ^ B2_) * B3_;
    }

    /// Returns a copy of the map
    inline MapParallelepiped3d* clone() const {
      return new MapParallelepiped3d(*this);
    };

  private:
    Real3d b_;
    Real3d B1_;
    Real3d B2_;
    Real3d B3_;

    /// Private assignement operator
    void operator=(const MapParallelepiped3d &);
  };

} // namespace concepts

#endif // elementMaps3D_hh
