/** @file geometry/meshConverter.hh
    Mesh converter in 2D from triangular to quadrilateral
    @author Philipp Frauenfelder, 2004
*/

#ifndef meshConverter2D_hh
#define meshConverter2D_hh

#include <list>
#include "mesh.hh"
#include "cell2D.hh"

namespace concepts {

  // ****************************************************** ConvertMeshQuads **

  /** Mesh converter. Converts a mesh with triangles into a mesh with
      quadrilaterals only. Each triangle is split into three quads by
      joining the edge midpoints and the center of gravity.

      If the attribute of a triangle which is subdivided is 0, the
      newly created middle vertex, the edges and the quads inherit
      the key of the triangle as attribute. Otherwise, they inherit
      the attribute of the triangle.

      This class can handle triangular and quadrilateral input meshes.
      If a quad is met, there is nothing to do. At the moment, mixed meshes
      cannot be handled.
      @todo Implement methods to handle mixed meshes (quads and triangles)
      @author Philipp Frauenfelder, 2004
      @test test::ConvertMeshQuads
      @ingroup geometry
  */
  class ConvertMeshQuads : public Mesh2 {
  public:
    /** Constructor. Takes the original mesh \c m and converts all triangles
	to quadrilaterals to build a new mesh
	@param m Original mesh
	@param level Level in \c m which should be converted
    */
    ConvertMeshQuads(Mesh2& m, const uint level);
    ~ConvertMeshQuads();

    virtual uint ncell() const { return quadList_.size(); }
    virtual Scan2* scan() { return new S(quadList_); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Original mesh
    Mesh2& m_;
    /// List of new vertices (centers of gravity of the triangles)
    Joiner<Vertex*, 1>* vertexList_;
    /// List of new edges
    Joiner<Edge*, 1>* edgeList_;
    /// List of new quads
    Joiner<Quad2d*, 1>* newQuadList_;
    /// List of quads in the new mesh
    std::list<Quad2d*> quadList_;
    /// Checks for a mixed mesh
    bool hasQuads, hasTris;

    /// Scanner class for ConvertMeshQuads
    class S : public Scan2 {
    public:
      S(std::list<Quad2d*>& cell) :
        idx_(cell.begin()), cell_(cell) {}
      S(const S& scan) : idx_(scan.idx_), cell_(scan.cell_) {}
      bool eos() const { return idx_ == cell_.end(); }
      Quad2d& operator++(int) { return *(*idx_++); }
      Scan2* clone() const { return new S(*this); }
    private:
      std::list<Quad2d*>::iterator idx_;
      std::list<Quad2d*>& cell_;
    };

    /// Handles a cell if it is on the correct level, otherwise, recurses
    void convertCell_(Cell2& c, const uint level);
    void deleteStuff_();
  };

} // namespace concepts

#endif // meshConverter2D_hh
