/* Meshes
 * Combining cells to a mesh to represent a computational domain
 * Importing 2D meshes from Matlab files
 */

#include <fstream>
#include <sstream>
#include <algorithm>
#include "meshImport2D.hh"
#include "cell2D.hh"
#include "basics/fileexception.hh"
#include "basics/debug.hh"

// debugging

#define Import2dMeshConstr_D 0
#define Import2dMeshCrE_D 0
#define Import2dMeshReadInts_D 0
#define Import2dMeshReadLine_D 0
#define NodeCell_D 0

namespace concepts {

  // ************************************************ Import2dMesh::NodeCell **

  Import2dMesh::NodeCell::~NodeCell() {}

  // ******************************************** Import2dMesh::NodeTriangle **

  Import2dMesh::NodeTriangle::NodeTriangle(const uint one, const uint two,
					   const uint three) : NodeCell(3) {
    nodes_[0] = one; nodes_[1] = two; nodes_[2] = three;
  }

  Import2dMesh::NodeTriangle::~NodeTriangle() {}

  Cell2* Import2dMesh::NodeTriangle::cell
  (const std::vector<Real3d>& vertices, uint dim) const {
#ifdef DEBUG
    for (uint i = 0; i < size(); ++i)
      conceptsAssert(edges_[i] != 0, Assertion());
#endif
    DEBUGL(NodeCell_D, "dim = " << dim);
    conceptsAssert(dim == 2 || dim == 3, Assertion());

    Triangle* tri = new Triangle(*edges_[0], *edges_[1], *edges_[2]);
    if (dim == 2)
      return new Triangle2d(*tri, VertexTriangle2d(vertices[nodes_[0]],
						   vertices[nodes_[1]],
						   vertices[nodes_[2]]));
    else
      return new Triangle3d(*tri, MapTriangle3d(vertices[nodes_[0]],
						vertices[nodes_[1]],
						vertices[nodes_[2]]));
  }

  // ************************************************ Import2dMesh::NodeQuad **

  Import2dMesh::NodeQuad::NodeQuad(const uint one, const uint two,
				   const uint three, const uint four) :
    NodeCell(4) {
    nodes_[0] = one; nodes_[1] = two; nodes_[2] = three; nodes_[3] = four;
  }

  Import2dMesh::NodeQuad::~NodeQuad() {}

  Cell2* Import2dMesh::NodeQuad::cell
  (const std::vector<Real3d>& vertices, uint dim) const {
#ifdef DEBUG
    for (uint i = 0; i < size(); ++i)
      conceptsAssert(edges_[i] != 0, Assertion());
#endif
    conceptsAssert(dim == 2, Assertion());

    Quad* quad = new Quad(*edges_[0], *edges_[1], *edges_[2], *edges_[3]);
    Quad2d* cell = new Quad2d(*quad,
			      VertexQuad2d(vertices[nodes_[0]],
					   vertices[nodes_[1]],
					   vertices[nodes_[2]],
					   vertices[nodes_[3]]));
    return cell;
  }

  // ********************************************************** Import2dMesh **

  Import2dMesh::Import2dMesh(const std::string coord, const std::string elms,
			     const std::string boundary,
			     const uint idxStart) {
    DEBUGL(Import2dMeshConstr_D, "coord = " << coord << ", elms = " << elms
	   << ", boundary = " << boundary);

    // reading in vertices
    DEBUGL(Import2dMeshConstr_D, "reading in vertices");
    std::auto_ptr<std::ifstream> ifs(new std::ifstream(coord.c_str()));
    if (!(*ifs)) throw conceptsException(FileOpenError(coord));
    std::vector<Real3d> vertices;
    std::string line;
    uint dim = 0;
    while (ifs->good()) {
      std::getline(*ifs, line);
      DEBUGL(Import2dMeshConstr_D, "line = " << line);
      if (line.length() > 0) {
	Real3d crd;
	uint idx;
	if (dim == 0)  dim = readLine_(line, &idx, &crd);
	else {
#ifdef DEBUG
	  uint newdim =
#endif
	    readLine_(line, &idx, &crd);
	  conceptsAssert(dim == newdim, Assertion());
	}
	vertices.push_back(crd);
	vtx_.push_back(new Vertex());
      }
    }
    DEBUGL(Import2dMeshConstr_D, "# vertices = " << vertices.size());
    DEBUGL(Import2dMeshConstr_D, "spatial dimension = " << dim);

    // reading in boundary conditions
    DEBUGL(Import2dMeshConstr_D, "reading in boundaries");
    std::map<uint, std::map<uint, Edge*> > edges;
    ifs.reset(new std::ifstream(boundary.c_str()));
    if (!(*ifs)) throw conceptsException(FileOpenError(boundary));
    std::vector<int> nodes(5);
    while (ifs->good()) {
      nodes.resize(0);
      std::getline(*ifs, line);
      DEBUGL(Import2dMeshConstr_D, "line = " << line);
      uint n = readInts_(line, nodes);
      DEBUGL(Import2dMeshConstr_D, "nodes = " << n);
      switch(n) {
      case 3: { // vertex
	Vertex* foo = vtx_[nodes[1]-idxStart];
	vtx_[nodes[1]-idxStart] = new Vertex(Attribute(nodes[2]));
	delete foo;
	break;
      }
      case 4: // edge
	createEdge_(nodes[1]-idxStart, nodes[2]-idxStart, edges,
		    Attribute(nodes[3])); break;
      case 0: break; // last line
      default:
	for (std::vector<Vertex*>::iterator i = vtx_.begin();
	     i != vtx_.end(); ++i) { delete *i; *i = 0; }
	throw conceptsException
	  (MissingFeature("could not interpret as boundary "
			  "(not vertex, edge, quad or triangle)"));
      } // switch n
    } // while ifs is good
    DEBUGL(Import2dMeshConstr_D, "# boundary edges = " << edg_.size());

    // reading in cells
    DEBUGL(Import2dMeshConstr_D, "reading in cells");
    ifs.reset(new std::ifstream(elms.c_str()));
    if (!(*ifs)) throw conceptsException(FileOpenError(elms));
    std::vector<NodeCell*> cells;
    while (ifs->good()) {
      std::getline(*ifs, line);
      DEBUGL(Import2dMeshConstr_D, "line = " << line);
      nodes.resize(0);
      uint n = readInts_(line, nodes);
      DEBUGL(Import2dMeshConstr_D, "nodes = " << n);
      switch(n) {
      case 4: // triangle
	cells.push_back(new NodeTriangle(nodes[1]-idxStart, nodes[2]-idxStart,
					 nodes[3]-idxStart));
	break;
      case 5: // quad
	cells.push_back(new NodeQuad(nodes[1]-idxStart, nodes[2]-idxStart,
				     nodes[3]-idxStart, nodes[4]-idxStart));
	break;
      case 0: break; // last line
      default:
	clear_(cells);
	for (std::vector<Vertex*>::iterator i = vtx_.begin();
	     i != vtx_.end(); ++i) { delete *i; *i = 0; }
	throw conceptsException(MissingFeature
				("just quads and triangles allowed"));
      }
    }
    DEBUGL(Import2dMeshConstr_D, "# cells = " << cells.size());

    // creating edges
    DEBUGL(Import2dMeshConstr_D, "creating edges");
    for (std::vector<NodeCell*>::iterator i = cells.begin(); i != cells.end();
	 ++i) {
      for (uint eIdx = 0; eIdx < (*i)->size(); ++eIdx) {
	createEdge_((*i)->estart(eIdx), (*i)->eend(eIdx), edges, 0,
		    &((*i)->edge(eIdx)));
      }
    }
    DEBUGL(Import2dMeshConstr_D, "total # edges = " << edg_.size());

    // creating cells
    DEBUGL(Import2dMeshConstr_D, "creating cells");
    for (std::vector<NodeCell*>::iterator i = cells.begin(); i != cells.end();
	 ++i) {
      cell_.push_back((*i)->cell(vertices, dim));
    }
    DEBUGL(Import2dMeshConstr_D, "# cells = " << cell_.size());

    clear_(cells);
  }

  //                         *************

  Import2dMesh::Import2dMesh(const std::string coord, const std::string elms,
			     const uint idxStart) {
    DEBUGL(Import2dMeshConstr_D, "coord = " << coord << ", elms = " << elms);

    // reading in vertices
    DEBUGL(Import2dMeshConstr_D, "reading in vertices");
    std::auto_ptr<std::ifstream> ifs(new std::ifstream(coord.c_str()));
    if (!(*ifs)) throw conceptsException(FileOpenError(coord));
    std::vector<Real3d> vertices;
    std::string line;
    uint dim = 0;
    while (ifs->good()) {
      std::getline(*ifs, line);
      DEBUGL(Import2dMeshConstr_D, "line = " << line);
      if (line.length() > 0) {
	Real3d crd;
	uint idx;
	if (dim == 0)  dim = readLine_(line, &idx, &crd);
	else {
#ifdef DEBUG
	  uint newdim =
#endif
	    readLine_(line, &idx, &crd);
	  conceptsAssert(dim == newdim, Assertion());
	}
	vertices.push_back(crd);
	vtx_.push_back(new Vertex());
      }
    }
    DEBUGL(Import2dMeshConstr_D, "# vertices = " << vertices.size());
    DEBUGL(Import2dMeshConstr_D, "spatial dimension = " << dim);

    // reading in cells
    DEBUGL(Import2dMeshConstr_D, "reading in cells");
    ifs.reset(new std::ifstream(elms.c_str()));
    if (!(*ifs)) throw conceptsException(FileOpenError(elms));
    std::map<uint, std::map<uint, Edge*> > edges;
    std::vector<NodeCell*> cells;
    while (ifs->good()) {
      std::getline(*ifs, line);
      DEBUGL(Import2dMeshConstr_D, "line = " << line);
      std::vector<int> nodes(5);
      nodes.resize(0);
      uint n = readInts_(line, nodes);
      DEBUGL(Import2dMeshConstr_D, "nodes = " << n);
      switch(n) {
      case 4: // triangle
	createEdge_(nodes[1]-idxStart, nodes[2]-idxStart, edges, 0);
	createEdge_(nodes[2]-idxStart, nodes[3]-idxStart, edges, 0);
	createEdge_(nodes[3]-idxStart, nodes[1]-idxStart, edges, 0);
	cells.push_back(new NodeTriangle(nodes[1]-idxStart, nodes[2]-idxStart,
					 nodes[3]-idxStart));
	break;
      case 5: // quad
	createEdge_(nodes[1]-idxStart, nodes[2]-idxStart, edges, 0);
	createEdge_(nodes[2]-idxStart, nodes[3]-idxStart, edges, 0);
	createEdge_(nodes[3]-idxStart, nodes[4]-idxStart, edges, 0);
	createEdge_(nodes[4]-idxStart, nodes[1]-idxStart, edges, 0);
	cells.push_back(new NodeQuad(nodes[1]-idxStart, nodes[2]-idxStart,
				     nodes[3]-idxStart, nodes[4]-idxStart));
	break;
      case 0: break; // last line
      default:
	clear_(cells);
	for (std::vector<Vertex*>::iterator i = vtx_.begin();
	     i != vtx_.end(); ++i) { delete *i; *i = 0; }
	throw conceptsException(MissingFeature
				("just quads and triangles allowed"));
      }
    }
    DEBUGL(Import2dMeshConstr_D, "# cells = " << cells.size());

    // creating edges
    DEBUGL(Import2dMeshConstr_D, "creating edges");
    for (std::vector<NodeCell*>::iterator i = cells.begin(); i != cells.end();
	 ++i) {
      for (uint eIdx = 0; eIdx < (*i)->size(); ++eIdx) {
	createEdge_((*i)->estart(eIdx), (*i)->eend(eIdx), edges, 0,
		    &((*i)->edge(eIdx)));
      }
    }
    DEBUGL(Import2dMeshConstr_D, "total # edges = " << edg_.size());

    // creating cells
    DEBUGL(Import2dMeshConstr_D, "creating cells");
    for (std::vector<NodeCell*>::iterator i = cells.begin(); i != cells.end();
	 ++i) {
      cell_.push_back((*i)->cell(vertices, dim));
    }
    DEBUGL(Import2dMeshConstr_D, "# cells = " << cell_.size());

    clear_(cells);
  }

  //                         *************

  Import2dMesh::~Import2dMesh() {
    for (std::vector<Cell2*>::iterator i = cell_.begin();
	 i != cell_.end(); ++i) {
      delete &((*i)->connector());
      delete *i; *i = 0;
    }
    for (std::vector<Edge*>::iterator i = edg_.begin();
	 i != edg_.end(); ++i) {
      delete *i; *i = 0;
    }
    for (std::vector<Vertex*>::iterator i = vtx_.begin();
	 i != vtx_.end(); ++i) {
      delete *i; *i = 0;
    }
  }

  //                         *************

  std::ostream& Import2dMesh::info(std::ostream& os) const {
    return os << "Import2dMesh(ncell = " << ncell() << ')';
  }

  //                         *************

  uint Import2dMesh::readInts_(const std::string& i,
			       std::vector<int>& v) const {
    std::istringstream input(i.c_str());
    int number = 0;
    DEBUGL(Import2dMeshReadInts_D, number);
    while (input >> number) {
      v.push_back(number);
      DEBUGL(Import2dMeshReadInts_D, "inserted " << v.back());
    }
    return v.size();
  }

  //                         *************

  uint Import2dMesh::readLine_(const std::string& i, uint* idx,
			       Real3d* coord) const {
    std::istringstream input(i.c_str());
    input >> *idx;
    int number = 0;
    while (input.good()) input >> (*coord)[number++];
    DEBUGL(Import2dMeshReadLine_D, "inserted " << *idx << ", " << *coord
	   << ", number = " << number);
#if defined (__GNUC__) && (__GNUC__ < 3)
    return (uint) number-1;
#else
    return (uint) number;
#endif
  }

  //                         *************

  void Import2dMesh::clear_(std::vector<NodeCell*>& cells) const {
    for (std::vector<NodeCell*>::iterator i = cells.begin();
	 i != cells.end(); ++i) {
      delete *i; *i = 0;
    }
  }

  //                         *************

  void Import2dMesh::createEdge_(uint ix, uint iy,
				 std::map<uint, std::map<uint, Edge*> >& edges,
				 Attribute attr, Edge** e) {
    if (iy < ix) std::swap(ix, iy);
    conceptsAssert(ix < iy, Assertion());
    conceptsAssert(iy < vtx_.size(), Assertion());
    std::map<uint, std::map<uint, Edge*> >::iterator j =
      edges.find(ix);
    if (j == edges.end()) {
      std::pair<std::map<uint, std::map<uint, Edge*> >::iterator, bool> k =
	edges.insert(std::pair<uint, std::map<uint, Edge*> >
		     (ix, std::map<uint, Edge*>()));
      j = k.first;
    }
    std::map<uint, Edge*>::iterator k = j->second.find(iy);
    Edge* edg;
    if (k == j->second.end()) {
      DEBUGL(Import2dMeshCrE_D, "Creating edge " << ix << " - " << iy);
      edg = new Edge(*vtx_[ix], *vtx_[iy], Attribute(attr));
      DEBUGL(Import2dMeshCrE_D, "# before: " << j->second.size());
      k = j->second.insert(std::pair<uint, Edge*>(iy, edg)).first;
      DEBUGL(Import2dMeshCrE_D, "# after: " << j->second.size());
      edg_.push_back(edg);
    } else {
      DEBUGL(Import2dMeshCrE_D, "Found edge " << ix << " - " << iy);
      edg = k->second;
    }

    if (e != 0) *e = edg;
  }


} // namespace concepts
