/* Meshes
 * Combining cells to a mesh to represent a computational domain
 * Importing 3D meshes from Matlab files
 */

#include <fstream>
#include <sstream>
#include <algorithm>
#include <cstdlib>
#include "meshImport3D.hh"
#include "cell3D.hh"
#include "basics/fileexception.hh"
#include "basics/debug.hh"

// debugging

#define IndexSort_D 0
#define IndexCompare_D 0

#define Import3dMeshConstr_D 0
#define Import3dMeshConstrCheck_D 0
#define Import3dMeshCrE_D 0
#define Import3dMeshCreateTri_D 0
#define Import3dMeshCreateQuad_D 0
#define Import3dMeshReadInts_D 0

#define ListEdges_D 0
#define ListFaces_D 0

#define NodeCellCell_D 0

namespace __gnu_cxx {

  size_t hash_value(const concepts::Index& i) {
    const uint size = sizeof(size_t)*8;
    const uint bit = size/i.n_.size();
    size_t res = 0;
    for (uint j = 0; j < i.n_.size(); ++j)
      res = (res << bit) + i[j];
    return res;
  }

} // namespace __gnu_cxx

namespace concepts {

  // ************************************************ Import3dMesh::NodeCell **

  Import3dMesh::NodeCell::~NodeCell() {}

  // ***************************************** Import3dMesh::NodeTetrahedron **

  Import3dMesh::NodeTetrahedron::NodeTetrahedron
  (const uint one, const uint two, const uint three, const uint four,
   const bool leftHand) :
    NodeCell(4), faces_(4, 0) {
    if (leftHand) {
      nodes_[0] = two; nodes_[1] = one;
    } else {
      nodes_[0] = one; nodes_[1] = two;
    }
    nodes_[2] = three; nodes_[3] = four;
  }

  Import3dMesh::NodeTetrahedron::~NodeTetrahedron() {}

  Index Import3dMesh::NodeTetrahedron::faceIndex(const uint i) {
    switch(i) {
    case 0: return Index(nodes_[0], nodes_[1], nodes_[2]);
    case 1: return Index(nodes_[0], nodes_[1], nodes_[3]);
    case 2: return Index(nodes_[1], nodes_[2], nodes_[3]);
    case 3: return Index(nodes_[2], nodes_[0], nodes_[3]);
    default: conceptsAssert(false, Assertion()); throw 0;
    }
  }

  Tetrahedron3d* Import3dMesh::NodeTetrahedron::cell
  (const std::vector<Real3d>& vertices) const {
    DEBUGL(NodeCellCell_D, "start.");
#ifdef DEBUG
    for (uint i = 0; i < size(); ++i)
      conceptsAssert(faces_[i] != 0, Assertion());
#endif
    DEBUGL(NodeCellCell_D, "faces are: " << *faces_[0] << ", " << *faces_[1]
	   << ", " << *faces_[2] << ", " << *faces_[3]);
    Tetrahedron* tet = new Tetrahedron(*faces_[0], *faces_[1], *faces_[2],
				       *faces_[3]);
    Tetrahedron3d* cell = 0;
    cell = new Tetrahedron3d
      (*tet, MapTetrahedron3d(vertices[nodes_[0]],
			      vertices[nodes_[1]],
			      vertices[nodes_[2]],
			      vertices[nodes_[3]]));
    DEBUGL(NodeCellCell_D, "end: " << *cell);
    return cell;
  }

  // ****************************************** Import3dMesh::NodeHexahedron **

  Import3dMesh::NodeHexahedron::NodeHexahedron
  (const uint one, const uint two, const uint three, const uint four,
   const uint five, const uint six, const uint seven, const uint eight,
   const uint elmType) :
    NodeCell(8), faces_(6, 0), elmType_(elmType) {
    nodes_[0] = one; nodes_[1] = two; nodes_[2] = three; nodes_[3] = four;
    nodes_[4] = five; nodes_[5] = six; nodes_[6] = seven; nodes_[7] = eight;
  }

  Import3dMesh::NodeHexahedron::~NodeHexahedron() {}

  Index Import3dMesh::NodeHexahedron::faceIndex(const uint i) {
    switch(i) {
    case 0: return Index(nodes_[0], nodes_[1], nodes_[2], nodes_[3]);
    case 1: return Index(nodes_[0], nodes_[1], nodes_[5], nodes_[4]);
    case 2: return Index(nodes_[1], nodes_[2], nodes_[6], nodes_[5]);
    case 3: return Index(nodes_[2], nodes_[3], nodes_[7], nodes_[6]);
    case 4: return Index(nodes_[3], nodes_[0], nodes_[4], nodes_[7]);
    case 5: return Index(nodes_[4], nodes_[5], nodes_[6], nodes_[7]);
    default: conceptsAssert(false, Assertion()); throw 0;
    }
  }

  Cell3* Import3dMesh::NodeHexahedron::cell
  (const std::vector<Real3d>& vertices) const {
    DEBUGL(NodeCellCell_D, "start.");
#ifdef DEBUG
    for (uint i = 0; i < faces_.size(); ++i)
      conceptsAssert(faces_[i] != 0, Assertion());
#endif
    DEBUGL(NodeCellCell_D, "faces are: " << *faces_[0] << ", " << *faces_[1]
	   << ", " << *faces_[2] << ", " << *faces_[3] << ", " << *faces_[4]
	   << ", " << *faces_[5]);
    Hexahedron* hex =
      new Hexahedron(*faces_[0], *faces_[1], *faces_[2],
		     *faces_[3], *faces_[4], *faces_[5]);
    if (elmType_ == 0) {
      Hexahedron3d* cell = new Hexahedron3d
	(*hex, MapHexahedron3d(vertices[nodes_[0]], vertices[nodes_[1]],
			       vertices[nodes_[2]], vertices[nodes_[3]],
			       vertices[nodes_[4]], vertices[nodes_[5]],
			       vertices[nodes_[6]], vertices[nodes_[7]]));
      DEBUGL(NodeCellCell_D, "end: " << *cell);
      return cell;
    }
    else {
      Parallelepiped3d* cell =
	new Parallelepiped3d(*hex, MapParallelepiped3d(vertices[nodes_[0]],
						       vertices[nodes_[1]],
						       vertices[nodes_[3]],
						       vertices[nodes_[4]]));
      DEBUGL(NodeCellCell_D, "end: " << *cell);
      return cell;
    }
  }

  // ***************************************************************** Index **

  Index::Index(const uint n0, const uint n1, const uint n2) :
    n_(3) {
    n_[0] = n0; n_[1] = n1; n_[2] = n2;
    sort_();
  }

  Index::Index(const uint n0, const uint n1, const uint n2,
			     const uint n3) :
    n_(4) {
    n_[0] = n0; n_[1] = n1; n_[2] = n2; n_[3] = n3;
    sort_();
  }

  void Index::sort_() {
    Array<uint> m(n_);
    DEBUGL(IndexSort_D, "n = " << n_);
    uint* min = std::min_element((uint*)n_, (uint*)n_+n_.size());
    DEBUGL(IndexSort_D, "min is " << *min << " at " << min-(uint*)n_);
    const int inc = n_.operator[]( (min-(uint*)n_-1+n_.size())%n_.size() ) <
      n_.operator[]( (min-(uint*)n_+1)%n_.size() ) ? -1 : +1;
    DEBUGL(IndexSort_D, "inc is " << inc);
    for (int i = 0; i < (int)n_.size(); ++i)
      n_[i] = m.operator[]( (min-(uint*)n_+i*inc+n_.size())%n_.size() );
    DEBUGL(IndexSort_D, "n sorted = " << n_);
  }

  bool Index::operator<(const concepts::Index& y) const {
    const uint n = n_.size() < y.n_.size() ? y.n_.size() : n_.size();
    for (uint i = 0; i < n; ++i) {
      if (operator[](i) < y[i]) {
	DEBUGL(IndexCompare_D, "x = " << n_ << " < y = " << y.n_ << " -> "
	       << "true");
	return true;
      }
    }
    DEBUGL(IndexCompare_D, "x = " << n_ << " < y = " << y.n_ << " -> "
	   << "false");
    return false;
  }

  bool Index::operator==(const concepts::Index& y) const {
    if (n_.size() != y.n_.size()) return false;
    bool equal = true;
    for (uint i = 0; i < n_.size(); ++i)
      equal &= (operator[](i) == y[i]);
    DEBUGL(IndexCompare_D, "x = " << n_ << " == y = " << y.n_ << " -> "
	   << (equal ? "true" : "false"));
    return equal;
  }

  std::ostream& operator<<(std::ostream& os, const Index& i) {
    return os << i.n_;
  }

  // ********************************************************** Import3dMesh **

  Import3dMesh::Import3dMesh(const std::string coord, const std::string elms,
			     const std::string boundary,
			     const uint idxStart, const bool leftHand,
			     const uint elmtyp) {
    DEBUGL(Import3dMeshConstr_D, "coord = " << coord << ", elms = " << elms
	   << ", boundary = " << boundary);

    // reading in vertices
    DEBUGL(Import3dMeshConstr_D, "reading in vertices");
    std::auto_ptr<std::ifstream> ifs(new std::ifstream(coord.c_str()));
    if (!(*ifs)) throw conceptsException(FileOpenError(coord));
    std::vector<Real3d> vertices;
    std::string idx, vx, vy, vz;
    *ifs >> idx;
    *ifs >> vx;
    *ifs >> vy;
    *ifs >> vz;
    while (ifs->good()) {
      vertices.push_back( Real3d(std::atof(vx.c_str()), std::atof(vy.c_str()),
				 std::atof(vz.c_str())) );
      vtx_.push_back(new Vertex());
      *ifs >> idx;
      *ifs >> vx;
      *ifs >> vy;
      *ifs >> vz;
    }
    DEBUGL(Import3dMeshConstr_D, "# vertices = " << vertices.size());

    // reading in boundary conditions
    DEBUGL(Import3dMeshConstr_D, "reading in boundaries");
    std::map<uint, std::map<uint, Edge*> > edges;
    ifs.reset(new std::ifstream(boundary.c_str()));
    if (!(*ifs)) throw conceptsException(FileOpenError(boundary));
    std::string line;
    std::vector<int> nodes(9);
    while (ifs->good()) {
      nodes.resize(0);
      std::getline(*ifs, line);
      DEBUGL(Import3dMeshConstr_D, "line = " << line);
      uint n = readInts_(line, nodes);
      DEBUGL(Import3dMeshConstr_D, "nodes = " << n);
      switch(n) {
      case 3: { // vertex
	Vertex* foo = vtx_[nodes[1]-idxStart];
	vtx_[nodes[1]-idxStart] = new Vertex(Attribute(nodes[2]));
	delete foo;
	break;
      }
      case 4: // edge
	createEdge_(nodes[1]-idxStart, nodes[2]-idxStart, edges,
		    Attribute(nodes[3])); break;
      case 5: // triangle
	createTriangle_(Index(nodes[1]-idxStart, nodes[2]-idxStart,
			nodes[3]-idxStart), edges, Attribute(nodes[4])); break;
      case 6: // quad
	createQuad_(Index(nodes[1]-idxStart, nodes[2]-idxStart,
		    nodes[3]-idxStart, nodes[4]-idxStart), edges,
		    Attribute(nodes[5])); break;
      case 0: break; // last line
      default:
	for (std::vector<Vertex*>::iterator i = vtx_.begin();
	     i != vtx_.end(); ++i) { delete *i; *i = 0; }
	throw conceptsException
	  (MissingFeature("could not interpret as boundary "
			  "(not vertex, edge, quad or triangle)"));
      }
    }
    DEBUGL(Import3dMeshConstr_D, "# boundary faces = " << tri_.size()
	   << " triangles + " << quad_.size() << " quads");

    // reading in cells
    DEBUGL(Import3dMeshConstr_D, "reading in cells");
    ifs.reset(new std::ifstream(elms.c_str()));
    if (!(*ifs)) throw conceptsException(FileOpenError(elms));
    std::vector<NodeCell*> cells;
    while (ifs->good()) {
      nodes.resize(0);
      std::getline(*ifs, line);
      DEBUGL(Import3dMeshConstr_D, "line = " << line);
      uint n = readInts_(line, nodes);
      DEBUGL(Import3dMeshConstr_D, "nodes = " << n);
      switch(n) {
      case 5: // tetrahedron
	cells.push_back(new NodeTetrahedron(nodes[1]-idxStart,
					    nodes[2]-idxStart,
					    nodes[3]-idxStart,
					    nodes[4]-idxStart, leftHand));
	break;
      case 9: // hexahedron
	cells.push_back(new NodeHexahedron
			(nodes[1]-idxStart, nodes[2]-idxStart,
			 nodes[3]-idxStart, nodes[4]-idxStart,
			 nodes[5]-idxStart, nodes[6]-idxStart,
			 nodes[7]-idxStart, nodes[8]-idxStart, elmtyp));
	break;
      case 0: break; // last line
      default:
	clear_(cells);
	for (__gnu_cxx::hash_map<Index, Triangle*, __gnu_cxx::hash<Index>,
	       __gnu_cxx::IndexEqual>::iterator
	       i = tri_.begin(); i != tri_.end(); ++i)
	  { delete i->second; i->second = 0; }
	for (__gnu_cxx::hash_map<Index, Quad*, __gnu_cxx::hash<Index>,
	       __gnu_cxx::IndexEqual>::iterator
	       i = quad_.begin(); i != quad_.end(); ++i)
	  { delete i->second; i->second = 0; }
	for (std::vector<Vertex*>::iterator i = vtx_.begin();
	     i != vtx_.end(); ++i) { delete *i; *i = 0; }
	throw conceptsException
	  (MissingFeature("just tetrahedra and hexahedra allowed"));
      }
    }
    DEBUGL(Import3dMeshConstr_D, "# cells = " << cells.size());

    // creating faces
    DEBUGL(Import3dMeshConstr_D, "creating faces");
    for (std::vector<NodeCell*>::iterator i = cells.begin(); i != cells.end();
	 ++i) {
      for (uint j = 0; j < (*i)->nofaces(); ++j) {
	Index I = (*i)->faceIndex(j);
	switch (I.size()) {
	case 3: // triangle
	  *((*i)->triangle(j)) = createTriangle_(I, edges);
	  break;
	case 4: // quad
	  *((*i)->quad(j)) = createQuad_(I, edges);
	  break;
	default:
	  clear_(cells);
	  for (__gnu_cxx::hash_map<Index, Triangle*, __gnu_cxx::hash<Index>,
		 __gnu_cxx::IndexEqual>::iterator
		 i = tri_.begin(); i != tri_.end(); ++i)
	    { delete i->second; i->second = 0; }
	  for (__gnu_cxx::hash_map<Index, Quad*, __gnu_cxx::hash<Index>,
		 __gnu_cxx::IndexEqual>::iterator i = quad_.begin();
	       i != quad_.end(); ++i)
	    { delete i->second; i->second = 0; }
	  for (std::vector<Vertex*>::iterator i = vtx_.begin();
	       i != vtx_.end(); ++i) { delete *i; *i = 0; }
	  throw conceptsException
	    (MissingFeature("just quads and triangles allowed as boundary"));
	}
      }
    }
    DEBUGL(Import3dMeshConstr_D, "total # faces = " <<  tri_.size()
	   << " triangles + " << quad_.size() << " quads");

    // creating cells
    DEBUGL(Import3dMeshConstr_D, "creating cells");
    for (std::vector<NodeCell*>::iterator i = cells.begin(); i != cells.end();
	 ++i) {
      cell_.push_back((*i)->cell(vertices));
    }
    DEBUGL(Import3dMeshConstr_D, "# cells = " << cell_.size());

#if ListEdges_D
    // list edges in MATLAB compatible way
    DEBUGL(ListEdges_D, "number of edges = " << edg_.size());
    uint k = 1;
    for (std::vector<Edge*>::iterator i = edg_.begin();
	 i != edg_.end(); ++i, ++k) {
      std::cout << k << " " << (*i)->vertex(0)->key().key()+1 << " "
		<< (*i)->vertex(1)->key().key()+1 << std::endl;
    }
#endif

#if ListFaces_D
    // list triangles in MATLAB compatible way
    DEBUGL(ListFaces_D, "number of triangles = " << tri_.size());
    uint l = 1;
    for (__gnu_cxx::hash_map<Index, Triangle*, __gnu_cxx::hash<Index>,
	   __gnu_cxx::IndexEqual>::iterator i = tri_.begin();
	 i != tri_.end(); ++i, ++l) {
      std::cout << l << " " << i->second->vertex(0)->key().key()+1 << " "
		<< i->second->vertex(1)->key().key()+1 << " "
		<< i->second->vertex(2)->key().key()+1 << std::endl;
    }
#endif

    clear_(cells);

    // check if all elements fit geometrically
#ifdef DEBUG
    std::map<uint, Real3d> coordinates;
    std::auto_ptr<Scan3> sc(scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      Connector0* vtx = 0;
      DEBUGL(Import3dMeshConstrCheck_D, "checking " << cell);
      for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j) {
	std::map<uint, Real3d>::iterator hit =
	  coordinates.find(vtx->key().key());
	if (hit != coordinates.end()) {
	  DEBUGL(Import3dMeshConstrCheck_D, "checking " << vtx->key() << " at "
		 << cell.vertex(j));
	  conceptsAssert3( (hit->second - cell.vertex(j)).l2_2() < 1e-10,
			   Assertion(),
			   "problem with vertex " << j << " in " << cell
			   << ": " << hit->second << " != " << cell.vertex(j));
	} else {
	  coordinates.insert(std::pair<uint, Real3d>(vtx->key().key(),
						     cell.vertex(j)));
	  DEBUGL(Import3dMeshConstrCheck_D, "adding " << vtx->key() << " at "
		 << cell.vertex(j));
	}
      } // for j
    }
#endif
  }

  Import3dMesh::~Import3dMesh() {
    for (std::vector<Cell3*>::iterator i = cell_.begin();
	 i != cell_.end(); ++i) {
      delete &((*i)->connector());
      delete *i; *i = 0;
    }
    for (__gnu_cxx::hash_map<Index, Triangle*, __gnu_cxx::hash<Index>,
	   __gnu_cxx::IndexEqual>::iterator i = tri_.begin();
	 i != tri_.end(); ++i)
      { delete i->second; i->second = 0; }
    for (__gnu_cxx::hash_map<Index, Quad*, __gnu_cxx::hash<Index>,
	   __gnu_cxx::IndexEqual>::iterator i = quad_.begin();
	 i != quad_.end(); ++i)
      { delete i->second; i->second = 0; }
    for (std::vector<Edge*>::iterator i = edg_.begin();
	 i != edg_.end(); ++i) { delete *i; *i = 0; }
    for (std::vector<Vertex*>::iterator i = vtx_.begin();
	 i != vtx_.end(); ++i) { delete *i; *i = 0; }
  }

  std::ostream& Import3dMesh::info(std::ostream& os) const {
    return os << "Import3dMesh(ncell = " << ncell() << ')';
  }

  uint Import3dMesh::readInts_(const std::string& i,
			       std::vector<int>& v) const {
    std::istringstream input(i.c_str());
    int number = 0;
    DEBUGL(Import3dMeshReadInts_D, number);
    while (input >> number) {
      v.push_back(number);
      DEBUGL(Import3dMeshReadInts_D, "inserted " << v.back());
    }
    return v.size();
  }

  void Import3dMesh::clear_(std::vector<NodeCell*>& cells) const {
    for (std::vector<NodeCell*>::iterator i = cells.begin();
	 i != cells.end(); ++i) {
      delete *i; *i = 0;
    }
  }

  void Import3dMesh::createEdge_(uint ix, uint iy,
				 std::map<uint, std::map<uint, Edge*> >& edges,
				 Attribute attr, Edge** e) {
    if (iy < ix) std::swap(ix, iy);
    conceptsAssert(ix < iy, Assertion());
    conceptsAssert(iy < vtx_.size(), Assertion());
    std::map<uint, std::map<uint, Edge*> >::iterator j =
      edges.find(ix);
    if (j == edges.end()) {
      std::pair<std::map<uint, std::map<uint, Edge*> >::iterator, bool> k =
	edges.insert(std::pair<uint, std::map<uint, Edge*> >
		     (ix, std::map<uint, Edge*>()));
      j = k.first;
    }
    std::map<uint, Edge*>::iterator k = j->second.find(iy);
    Edge* edg;
    if (k == j->second.end()) {
      DEBUGL(Import3dMeshCrE_D, "Creating edge " << ix << " - " << iy);
      edg = new Edge(*vtx_[ix], *vtx_[iy], attr);
      DEBUGL(Import3dMeshCrE_D, "# before: " << j->second.size());
      k = j->second.insert(std::pair<uint, Edge*>(iy, edg)).first;
      DEBUGL(Import3dMeshCrE_D, "# after: " << j->second.size());
      edg_.push_back(edg);
    } else {
      DEBUGL(Import3dMeshCrE_D, "Found edge " << ix << " - " << iy);
      edg = k->second;
    }

    if (e != 0) *e = edg;
  }

  std::ostream& operator<<(std::ostream& os,
			   const __gnu_cxx::hash_map<Index, Triangle*,
			   __gnu_cxx::hash<Index>,
			   __gnu_cxx::IndexEqual>& tris) {
    os << "Triangles(" << tris.size() << ": " << std::endl;
    for (__gnu_cxx::hash_map<Index, Triangle*, __gnu_cxx::hash<Index>,
	   __gnu_cxx::IndexEqual>::const_iterator
	   i = tris.begin(); i != tris.end(); ++i)
      os << "  " << i->first << " - " << *(i->second) << std::endl;
    return os << ')';
  }

  Triangle* Import3dMesh::createTriangle_
  (const Index& I,
   std::map<uint, std::map<uint, Edge*> >& edges, Attribute attr) {
    DEBUGL(Import3dMeshCreateTri_D, "creating triangle for " << I);
    DEBUGL(Import3dMeshCreateQuad_D, tri_);
    conceptsAssert(I.size() == 3, Assertion());
    Edge* triEdges[3] = { 0, 0, 0 };
    __gnu_cxx::hash_map<Index, Triangle*, __gnu_cxx::hash<Index>,
      __gnu_cxx::IndexEqual>::iterator k = tri_.find(I);
    Triangle* tri;
    if (k == tri_.end()) {
      DEBUGL(Import3dMeshCreateTri_D, "triangle not found");
      for (uint i = 0; i < 3; ++i)
	createEdge_(I[i], I[ (i+1)%3 ], edges, 0, &(triEdges[i]));
      k = tri_.insert
	(std::pair<Index, Triangle*>
	 (I, new Triangle(*triEdges[0], *triEdges[1], *triEdges[2],
			  attr))).first;
    } else
      DEBUGL(Import3dMeshCreateTri_D, "triangle found");
    tri = k->second;
    DEBUGL(Import3dMeshCreateTri_D, "triangle is " << *tri);
    return tri;
  }

  std::ostream& operator<<(std::ostream& os,
			   const __gnu_cxx::hash_map<Index, Quad*,
			   __gnu_cxx::hash<Index>,
			   __gnu_cxx::IndexEqual>& quads) {
    os << "Quads(" << quads.size() << ": " << std::endl;
    for (__gnu_cxx::hash_map<Index, Quad*, __gnu_cxx::hash<Index>,
	   __gnu_cxx::IndexEqual>::const_iterator i = quads.begin();
	 i != quads.end(); ++i)
      os << "  " << i->first << " - " << *(i->second) << std::endl;
    return os << ')';
  }

  Quad* Import3dMesh::createQuad_
  (const Index& I,
   std::map<uint, std::map<uint, Edge*> >& edges, Attribute attr) {
    DEBUGL(Import3dMeshCreateQuad_D, "creating quad for " << I);
    DEBUGL(Import3dMeshCreateQuad_D, quad_);
    conceptsAssert(I.size() == 4, Assertion());
    Edge* quadEdges[4] = { 0, 0, 0, 0 };
    __gnu_cxx::hash_map<Index, Quad*, __gnu_cxx::hash<Index>,
      __gnu_cxx::IndexEqual>::iterator k = quad_.find(I);
    Quad* quad;
    if (k == quad_.end()) {
      DEBUGL(Import3dMeshCreateQuad_D, "quad not found");
      for (uint i = 0; i < 4; ++i)
	createEdge_(I[i], I[ (i+1)%4 ], edges, 0, &(quadEdges[i]));
      k = quad_.insert
	(std::pair<Index, Quad*>
	 (I, new Quad(*quadEdges[0], *quadEdges[1], *quadEdges[2],
		      *quadEdges[3], attr))).first;
    } else
      DEBUGL(Import3dMeshCreateQuad_D, "quad found");
    quad = k->second;
    DEBUGL(Import3dMeshCreateQuad_D, "quad is " << *quad);
    return quad;
  }

} // namespace concepts
