/* testsuite for 2D mesh classes in geometry
 */

#include <iostream>
#include "basics.hh"
#include "geometry/mesh.hh"
#include "geometry/elementMaps.hh"
#include "geometry/cell2D.hh"
#include "testJacobian.hh"

using namespace concepts;

// **************************************************************** TestMesh **

class TestMesh : public Mesh2 {
private:
  class S : public Scan<Cell2> {
    unsigned int idx_;
    Cell2 *(&cell_)[1];
  public:
    inline S(Cell2 *(&cell)[1]) : idx_(0), cell_(cell) {}
    inline S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    inline bool eos() const { return idx_ == 1; }
    inline Cell2& operator++(int) { return *cell_[idx_++]; }
    inline Scan2* clone() const { return new S(*this); }
  };
  Vertex *vtx_[4];
  Edge *edg_[4];
  Quad *quad_[1];
  Cell2 *cell_[1];
public:
  TestMesh(uint rotation = 0);
  virtual ~TestMesh();
  inline unsigned int ncell() const { return 1; }
  inline Scan2* scan() { return new S(cell_); }
  virtual std::ostream& info(std::ostream& os) const;
};

TestMesh::TestMesh(uint rotation) {
  for (int i = 0; i < 4; ++i)
    vtx_[i] = new Vertex;

  edg_[0] = new Edge(*vtx_[0], *vtx_[1]);
  edg_[1] = new Edge(*vtx_[1], *vtx_[2]);
  edg_[2] = new Edge(*vtx_[2], *vtx_[3]);
  edg_[3] = new Edge(*vtx_[3], *vtx_[0]);

  quad_[0] = new Quad(*edg_[0], *edg_[1], *edg_[2], *edg_[3]);
  switch (rotation) {
  case 0: {
    VertexQuad2d m(Real2d(0.0, 0.0), Real2d(1.0, 0.0), 
		   Real2d(1.0, 1.0), Real2d(0.0, 1.0));
    cell_[0] = new Quad2d(*quad_[0], m); break; }
  case 1: {
    VertexQuad2d m(Real2d(1.0, 0.0), Real2d(1.0, 1.0), 
		   Real2d(0.0, 1.0), Real2d(0.0, 0.0));
    cell_[0] = new Quad2d(*quad_[0], m); break; }
  case 2: {
    VertexQuad2d m(Real2d(1.0, 1.0), Real2d(0.0, 1.0), 
		   Real2d(0.0, 0.0), Real2d(1.0, 0.0));
    cell_[0] = new Quad2d(*quad_[0], m); break; }
  case 3: {
    VertexQuad2d m(Real2d(0.0, 1.0), Real2d(0.0, 0.0), 
		   Real2d(1.0, 0.0), Real2d(1.0, 1.0));
    cell_[0] = new Quad2d(*quad_[0], m); break; }
  }
}

TestMesh::~TestMesh() {
  delete cell_[0];
  delete quad_[0];
  for (int i = 0; i < 4; ++i)
    delete edg_[i];
  for (int i = 0; i < 4; ++i)
    delete vtx_[i];
}

std::ostream& TestMesh::info(std::ostream& os) const {
  os << "TestMesh(" << *cell_[0] << ')';
  return os;
}

namespace test {

  // ******************************************************** TestJacobian2D **

  TestJacobian2D::~TestJacobian2D() {}

  void TestJacobian2D::run() {
    testRot0Isotropic(); testRot0Horizontal(); testRot0Vertical();
    testRot1Isotropic(); testRot1Horizontal(); testRot1Vertical();
    testRot2Isotropic(); testRot2Horizontal(); testRot2Vertical();
    testRot3Isotropic(); testRot3Horizontal(); testRot3Vertical();
  }

  void TestJacobian2D::testRot0Isotropic() {
    TestMesh t;
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/4.0, 4.0,
	       1.0/2.0, 2.0, 0, 1, 0, 1, 1.0/2.0, 2.0,
	       3);
    }
  }

  void TestJacobian2D::testRot0Horizontal() {
    TestMesh t;
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/2.0, 2.0,
	       1.0, 1.0, 0, 1, 0, 1, 1.0/2.0, 2.0,
	       3, Quad2dSubdiv2H::instance());
    }
  }

  void TestJacobian2D::testRot0Vertical() {
    TestMesh t;
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/2.0, 2.0,
	       1.0/2.0, 2.0, 0, 1, 0, 1, 1.0, 1.0,
	       3, Quad2dSubdiv2V::instance());
    }
  }

  void TestJacobian2D::testRot1Isotropic() {
    TestMesh t(1);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/4.0, 4.0,
	       0, 1, -1.0/2.0, 2.0, 1.0/2.0, 2.0, 0, 1,
	       3);
    }
  }

  void TestJacobian2D::testRot1Horizontal() {
    TestMesh t(1);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/2.0, 2.0,
	       0, 1, -1.0/2.0, 2.0, 1.0, 1.0, 0, 1,
	       3, Quad2dSubdiv2H::instance());
    }
  }

  void TestJacobian2D::testRot1Vertical() {
    TestMesh t(1);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/2.0, 2.0,
	       0, 1, -1.0, 1.0, 1.0/2.0, 2.0, 0, 1,
	       3, Quad2dSubdiv2V::instance());
    }
  }

  void TestJacobian2D::testRot2Isotropic() {
    TestMesh t(2);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/4.0, 4.0,
	       -1.0/2.0, 2.0, 0, 1, 0, 1, -1.0/2.0, 2.0,
	       3);
    }
  }

  void TestJacobian2D::testRot2Horizontal() {
    TestMesh t(2);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/2.0, 2.0,
	       -1.0, 1.0, 0, 1, 0, 1, -1.0/2.0, 2.0,
	       3, Quad2dSubdiv2H::instance());
    }
  }

  void TestJacobian2D::testRot2Vertical() {
    TestMesh t(2);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/2.0, 2.0,
	       -1.0/2.0, 2.0, 0, 1, 0, 1, -1.0, 1.0,
	       3, Quad2dSubdiv2V::instance());
    }
  }

  void TestJacobian2D::testRot3Isotropic() {
    TestMesh t(3);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/4.0, 4.0,
	       0, 1, 1.0/2.0, 2.0, -1.0/2.0, 2.0, 0, 1,
	       3);
    }
  }

  void TestJacobian2D::testRot3Horizontal() {
    TestMesh t(3);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/2.0, 2.0,
	       0, 1, 1.0/2.0, 2.0, -1.0, 1.0, 0, 1,
	       3, Quad2dSubdiv2H::instance());
    }
  }

  void TestJacobian2D::testRot3Vertical() {
    TestMesh t(3);
    std::auto_ptr<Scan2> sc(t.scan());
    while (*sc) {
      Cell2& cell = (*sc)++;
      Quad2d* q = dynamic_cast<Quad2d*>(&cell);
      _numtest(q->jacobianDeterminant(0.5, 0.5), 1);
      recTest_(q, 1.0/2.0, 2.0,
	       0, 1, 1.0, 1.0, -1.0/2.0, 2.0, 0, 1,
	       3, Quad2dSubdiv2V::instance());
    }
  }

  void TestJacobian2D::recTest_(Quad2d* q, Real guess, Real div,
				Real guess00, Real div00,
				Real guess01, Real div01,
				Real guess10, Real div10,
				Real guess11, Real div11,
				uint rec, const Quad2dSubdivision* ref) {
    q->setStrategy(ref);
    Quad2d* qChild = 0;
    for (uint i = 0; (qChild = q->child(i)) != 0; ++i) {
      MapReal2d j = qChild->jacobian(0.5, 0.5);
      _numtest(j(0,1), guess01); _numtest(j(1,0), guess10);
      _numtest(j(0,0), guess00); _numtest(j(1,1), guess11);
      _numtest(qChild->jacobianDeterminant(0.5, 0.1), guess);
      _numtest(qChild->jacobianDeterminant(0.1, 0.5), guess);
      _numtest(qChild->jacobianDeterminant(0.1, 0.1), guess);
      _numtest(qChild->jacobianDeterminant(0.9, 0.9), guess);
      _numtest(qChild->jacobianDeterminant(0.9, 0.5), guess);
      _numtest(qChild->jacobianDeterminant(0.5, 0.9), guess);
      _numtest(qChild->jacobianDeterminant(0.5, 0.5), guess);
      _numtest(qChild->jacobianDeterminant(0.1, 0.9), guess);
      _numtest(qChild->jacobianDeterminant(0.9, 0.1), guess);
      if (rec > 0)
	recTest_(qChild, guess/div, div,
		 guess00/div00, div00,
		 guess01/div01, div01,
		 guess10/div10, div10,
		 guess11/div11, div11,
		 rec-1, ref);
    }
  }

} // namespace test

int main(int, char**) {
  concepts::Stacktrace::doit() = false;
  try {
    test::TestJacobian2D t;
    t.run();
    return t.report();
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
  return 1;
}
