/* Topology.
 * how to subdivide
 */

#ifndef topology_hh
#define topology_hh

#include <memory>
#include <map>
#include "basics/typedefs.hh"
#include "basics/Zm.hh"
#include "toolbox/scannerConnectors.hh"
#include "connector.hh"
#include "quadsubdiv.hh"
#include "strategyChange.hh"

namespace concepts {

  // **************************************************************** Vertex **

  /** A vertex in the topology. Implements the abstract methods from the
      parent classes to query a child.
      @ingroup geometry
  */
  class Vertex : public Connector0 {
  public:
    /// Default constructor
    Vertex(const Attribute attrib = Attribute()) : Connector0(attrib) {}
    virtual ~Vertex();

    virtual Vertex* child(uint i, bool mode = 0) { return this; }
    virtual const Vertex* child(uint) const { return this; }

    /** Returns the pointer of a new vertex with the same attribute,
	the same key but NO children
    */
    inline Vertex* clone() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Constructor used by clone()
    Vertex(const Attribute attrib, const Key& key) : Connector0(attrib, key) {}
  };

  Vertex* Vertex::clone() const {
    return new Vertex(attrib(), key());
  }

  // ****************************************************************** Edge **

  /** An edge in the topology. Implements the abstract methods from the
      parent classes to query the children and the vertices of the edge.
      @ingroup geometry
  */
  class Edge : public Connector1 {

  public:
    /** Constructor.
	@param vtx0 Starting vertex of this edge
	@param vtx1 Ending vertex of this edge
	@param attrib Attribute of the edge
    */
    Edge(Vertex& vtx0, Vertex& vtx1, const Attribute attrib = Attribute());
    virtual ~Edge();

    /** Returns a child. 

        When creating new children, the starting vertex of the first
        edge is the child of the starting vertex of this edge, the
        ending Vertex is new. It is also the starting vertex for the
        second child. The ending vertex of the second child is the
        child of the ending Vertex of this edge.

	@param i Index of the child to be returned.
	@param mode mode = 1: No children are created<br>
	            mode = 0: If no child exists, two new children are
		              created (with the same attributes as this one).
    */
    virtual Edge* child(uint i, bool mode = 0);
    virtual const Edge* child(uint i) const;

    Vertex* vertex(uint i) const { return i < 2 ? vtx_[i] : 0; }

    /** Returns the pointer of a new Edge with the same attribute,
       the same key but NO children and sibling
    */
    inline Edge* clone(Vertex& vtx0, Vertex& vtx1) const;
  protected:
    /// Constructor used by clone()
    Edge(Vertex& vtx0, Vertex& vtx1, const Attribute attrib, const Key& key);

    virtual std::ostream& info(std::ostream& os) const;
    /// Pointer to a sibling
    Edge* lnk_;

    /** Pointer to the first child. The children are stored in a linked
	list.
    */
    Edge* chld_;

    /// Array of the vertices
    Vertex* vtx_[2];
  };

  Edge* Edge::clone(Vertex& vtx0, Vertex& vtx1) const {
    return new Edge(vtx0, vtx1, attrib(), key());
  }

  // ************************************************************** Triangle **

  /** A triangle in the topology. Implements the abstract methods from the
      parent classes to query the children, the vertices and the edges of
      the triangle.

      Every edge can have two different orientations in the triangle,
      this orientation is stored in the array rho.
      @ingroup geometry
  */
  class Triangle : public Connector2 {
  public:
    /** Constructor. Creates a triangle out of three edges. The edges
	must form a triangle: the 1st and the 2nd edge, the 2nd and 
	the 3rd edge and the 3rd and the 1st edge must have a 
	common vertex. The edges can have an arbitrary orientation. 
	The orientation is detected and stored in rho_.

	@param edg0 1st edge.
	@param edg1 2nd edge.
	@param edg2 3rd edge.
	@param attrib Attribute of the triangle
    */
    Triangle(Edge& edg0, Edge& edg1, Edge& edg2, 
	     const Attribute attrib = Attribute());
    virtual ~Triangle();

    /** Returns a child.

        If children are created, four new children are created (with
	the same attributes as this one). The triangle is cut into
	four new triangles by joining the midpoints of its edges.

	@param i Index of the child to be returned.
	@param mode mode = 1: No children are created<br>
	            mode = 0: If no children exist, four new children are
		              created.
    */
    virtual Triangle* child(uint i, bool mode = 0);
    virtual const Triangle* child(uint i) const;

    Edge* edge(uint i) const { return i < 3 ? edg_[i] : 0; }
    Vertex* vertex(uint i) const { return i<3 ? edg_[i]->vertex(rho_[i]) : 0; }

    /** Returns the orientation of an edge.
	@param i Index of the edge.
    */
    Z2 rho(int i) const { return rho_[i]; }

    /** Returns the pointer of a new Triangle with the same attribute,
       the same key and the same rho but NO children and sibling
    */
    inline Triangle* clone(Edge& edg0, Edge& edg1, Edge& edg2) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Constructor used by clone()
    Triangle(Edge& edg0, Edge& edg1, Edge& edg2,
	     const Attribute attrib, const Key& key, const Z2 rho[]);

    /// Pointer to a sibling
    Triangle* lnk_;

    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Triangle* chld_;

    /// Array of the edges
    Edge* edg_[3];

    /// Array of the orientation flags for the edges
    Z2 rho_[3];
  };

  Triangle* Triangle::clone(Edge& edg0, Edge& edg1, Edge& edg2) const {
    return new Triangle(edg0, edg1, edg2, attrib(), key(), rho_);
  }

  // ****************************************************************** Quad **

  /** A quadrilateral in the topology. Implements the abstract methods
      from the parent classes to query the children, the vertices and
      the edges of the quadrilateral.

      Every edge can have two different orientations in the quadrilateral,
      this orientation is stored in the array rho.
      @ingroup geometry
  */
  class Quad : public Connector2 {
    friend class QuadSubdiv4;
    friend class QuadSubdiv2H;
    friend class QuadSubdiv2V;
  public:
    /** Constructor. Creates a quadrilateral out of four edges. The
	edges must form a quadrilateral: the 1st and the 2nd edge, the
	2nd and the 3rd edge, the 3rd and the 4th edge and the 4th and
	the 1st edge must have a common vertex. The edges can have an
	arbitrary orientation.  The orientation is detected and stored
	in rho_.

	@param edg0 1st edge.
	@param edg1 2nd edge.
	@param edg2 3rd edge.
	@param edg3 4th edge.
	@param attrib Attribute of the edge
    */
    Quad(Edge& edg0, Edge& edg1, Edge& edg2, Edge& edg3, 
	 const Attribute attrib = Attribute());
    /** Constructor for coalescing two quadrilaterials with common edge.
	The subdivision strategy has to be given, this is neccessary
	to use the right S matrices.
     */
    Quad(Quad& quad0, Quad& quad1,
	 const QuadSubdivision* strategy = 0, const Quad* bigquad = 0);
    /// Destructor
    virtual ~Quad();

    /** Returns a child.

        If children are created, four new children are created (with
	the same attributes as this one). All edges of the
	quadrilateral are refined and four new edges introduced.

	@param i Index of the child to be returned.
	@param mode mode = 1: No children are created<br>
	            mode = 0: If no childs exist, four new children are
		              created. setStrategy() is called before
			      creating the children.
    */
    virtual Quad* child(uint i, bool mode = 0);
    virtual const Quad* child(uint i) const;

    Edge* edge(uint i) const { return i < 4 ? edg_[i] : 0; }
    Vertex* vertex(uint i) const { return i<4 ? edg_[i]->vertex(rho_[i]) : 0; }
    /** Returns the orientation of an edge.

	@param i Index of the edge.
    */
    Z2 rho(int i) const { return rho_[i]; }

    /** Returns the pointer of a new Quad with the same attribute,
	the same key and the same rho but NO children and sibling
    */
    inline Quad* clone(Edge& edg0, Edge& edg1, Edge& edg2, Edge& edg3) const;

    /** Sets the subdivision strategy of this quad.

        If the parameter is set to 0 (or if the method is called
	without parameter) the strategy is set to the default (if not
	already set). The default subdivision strategy is subdivision
	into 4 children.

        @param strategy Pointer to an instance of a subdivision strategy.
	@throws StrategyChange if the change is not allowed (the change is
	        not allowed if there are children present)
    */
    void setStrategy(const QuadSubdivision* strategy = 0)
      throw(StrategyChange);

    /** Returns the subdivision strategy of this quad.

        If you want to find check against another strategy use
	<pre>quad.getStrategy() == QuadSubdiv2V::instance()</pre>
    */
    const QuadSubdivision* getStrategy() const { return subdivStrategy_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;

    /// Constructor used by clone()
    Quad(Edge& edg0, Edge& edg1, Edge& edg2, Edge& edg3,
	 const Attribute attrib, const Key& key, const Z2 rho[]);

    /// List of Pointers to the children
    concepts::Joiner<Quad*, 1>* chld_;

    /// Array of the edges
    Edge* edg_[4];

  /// Array of the orientation flags for the edges
    Z2 rho_[4];

  /// Subdivision strategy for the quadrilateral
    const QuadSubdivision* subdivStrategy_;

    /// Scanner over the children
    inline Scan<Quad>* children_() const {
      return new concepts::PListScan<Quad>(*chld_);
    }
};

  Quad* Quad::clone(Edge& edg0, Edge& edg1, Edge& edg2, Edge& edg3) const {
    return new Quad(edg0, edg1, edg2, edg3, attrib(), key(), rho_);
  }

} // namespace concepts

#endif
