/* graphics of the data in Gnuplot
 */

#ifndef graphDataGnuplot_hh
#define graphDataGnuplot_hh

#include <string>
#include <fstream>
#include <memory>
#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "basics/exceptions.hh"
#include "function/vector.hh"
#include "space/postProcess.hh"
#include "toolbox/formula.hh"

namespace graphics {
  using concepts::Real;

  // ******************************************************* DataGnuplotCell **

  /** Writes the data which describe the elements to the given streams.
      @author Philipp Frauenfelder, 2002.
  */
  template<typename F>
  class DataGnuplotCell : public concepts::CellPostprocess<F> {
  public:
    /** Constructor
	@param ofs Stream to write to
	@param solution Solution vector to plot
    */
    DataGnuplotCell(std::ofstream* ofs,
		    const concepts::Vector<Real>& solution);
    /** Constructor
	@param ofs Stream to write to
	@param frm Formula to plot
    */
    DataGnuplotCell(std::ofstream* ofs, const concepts::Formula& frm);
    /** Constructor
	@param ofs Stream to write to
	@param frm Formula to plot (given elementwise)
    */
    DataGnuplotCell(std::ofstream* ofs,
		    const concepts::ElementFormula<F>& frm);
    virtual ~DataGnuplotCell();
    virtual void operator() (const concepts::Element<F>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature);
    /// Returns the stream to write the data to
    std::ofstream* write() { return ofs_; }
    /// Returns the coefficients of the elements shape functions
    const concepts::Array<Real>& coeff() const { return coeff_; }
    /// Returns the formula to plot
    const concepts::Formula* formula() const { return frm_.get(); }
    /// Returns the formula to plot (given elementwise)
    const concepts::ElementFormula<F>* elementFormula() const {
      return frmElm_;
    }
  private:
    /// Stream to write the data to
    std::ofstream* ofs_;
    /// Elementwise coefficients of the shape functions
    concepts::Array<Real> coeff_;
    /// Vector of the variable to plot
    const concepts::Vector<Real>* solution_;
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;
    /// Formula (given elementwise)
    const concepts::ElementFormula<F>* frmElm_;
  };

  // *********************************************************** DataGnuplot **

  /** Creates a data file for viewing the data with Gnuplot using
      DataGnuplotCell.

      The data can be plotted in gnuplot with
      <pre>set data style lines
splot 'file.gnuplot'</pre>

      @author Philipp Frauenfelder, 2002
  */
  template<typename T, typename S>
  void drawDataGnuplot(T& msh_spc, std::string filename, const S& sol) {
    std::ofstream ofs(filename.c_str());
    concepts::GlobalPostprocess<Real> postProcess(msh_spc);
    DataGnuplotCell<Real> gnu(&ofs, sol);
    postProcess(gnu);
  }

} // namespace graphics

#endif // graphDataGnuplot_hh
