/* graphics in Matlab
 */

#ifndef graphDataMatlab_hh
#define graphDataMatlab_hh

#include <map>
#include <string>
#include <fstream>
#include <sstream>
#include <memory>
#include <vector>
#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "basics/exceptions.hh"
#include "toolbox/dynArray.hh"
#include "function/vector.hh"
#include "space/postProcess.hh"

namespace concepts {
  template<class F>
  class Vector;
}

namespace graphics {

  using concepts::Real;

  // typedef MapToPtrOstream std::map<std::string, std::ostringstream*>;
//   typedef PairPtrOstream std::pair<std::string, std::ostringstream*>;

  // ******************************************************** DataMatlabCell **

  /** Writes the data which describe the elements to the given streams.
      @author Gregor Schmidlin, 2002
  */
  class DataMatlabCell : public concepts::CellPostprocess<Real> {
  public:
    /** Constructor
	@param ofsmsh Stream for the mesh
	@param ofs1   Stream for x-Coordinates
	@param ofs2   Stream for y-Coordinates
	@param ofs3   Stream for z-Coordinates (dim equal 3) or
	              the data of the vector (dim equal 2)
	@param dim    Spatial dimension of the elements
	@param scl    Data scaling factor
    */
    DataMatlabCell(std::ofstream* ofsmsh, std::ofstream* ofs1,
		   std::ofstream* ofs2, std::ofstream* ofs3,
		   uint dim, Real scl);

    /** Constructor
	@param ofs    Stream for the mesh
	@param dim    Spatial dimension of the elements
	@param scl    Data scaling factor
    */
    DataMatlabCell(std::ofstream* ofs, uint dim, Real scl);

    /** Output of the mesh and vector data of the element
     */
    virtual void operator()(const concepts::Element<Real>& elm)
      throw(concepts::MissingFeature);
    /// not operational
    virtual void operator()(const concepts::Cell& cell)
      throw(concepts::MissingFeature);

    /// Returns the streams to write the data to
    void write(std::ofstream** ofsmsh, std::ofstream** ofs1,
	       std::ofstream** ofs2, std::ofstream** ofs3);
    /// Sets the vector of the variable to plot
    void setVec(const concepts::Vector<Real>* sol);
    /// Returns the coefficients of the elements shape functions
    const concepts::Array<Real>& coeff() const {return coeff_;}
    /// Returns pointer to vector to plot
    const concepts::Vector<Real>* solution() const {return sol_;}
    Real offset() const {return offset_;}
    Real range() const {return range_;}
    /// Number of points written to the files, used for mesh representation
    uint& count() {return cnt_;}
    uint dim() const {return dim_;}
    Real scale() const {return scl_;}
    concepts::DynArray<uint>& map() {return *map_;}
  private:
    /// Spatial dimension of the space
    uint dim_;
    /// Offset and range for color
    Real offset_, range_;
    /// Streams to write the data to
    std::ofstream *ofsmsh_, *ofs1_, *ofs2_, *ofs3_;
    /// Elementwise coefficients of the shape functions
    concepts::Array<Real> coeff_;
    /// Vector of the variable to plot
    const concepts::Vector<Real>* sol_;
    /// Counter
    uint cnt_;
    /// Data scaling factor
    Real scl_;

    /// Map for vertices of the elements
    std::auto_ptr<concepts::DynArray<uint> > map_;
  };

  // ************************************************************ DataMatlab **

  /** Draws a picture of data in Matlab format and stores the result in
      files.

      The data is stored in four files by appending '_1.m', '_2.m',
      '_3.m' and '_msh.m'. The data can be read into Matlab by giving
      the three four file names as commands (and ommiting the ending '_*.m').
      The data is plotted by
      <pre>trimesh(msh, x, y, z0)</pre>
      or
      <pre>trimesh(msh, x, y, z0, ones(size(x,1),1))</pre>

      This class opens and closes the four files and hands the
      streams to DataMatlabCell.
  */
  class DataMatlab {
  public:
    /** Constructor
	@param spc Space on which the data should be plotted
	@param filename Name base for the files to be written
	@param dim Spatial dimension of the data (only dim=2 supported)
	@param scl Data scaling factor
     */
    DataMatlab(concepts::Space<Real>& spc, const std::string filename,
	       uint dim = 2, Real scl = 1.0);
#if 0
    /** Constructor for edge elements (vector fields)
	@param spc Space on which the data should be plotted
	@param filename Name base for the files to be written
	@param sol The solution which should be plotted
	@param scl Data scaling factor
     */
    DataMatlab(hp2Dedge::Space& spc, const std::string filename,
	       const concepts::Vector<Real>& sol, Real scl = 1.0);
#endif

    ~DataMatlab() {closeStreams_();}

    /** Application operator.
	@param sol The solution which should be plotted
    */
    void operator()(const concepts::Vector<Real>& sol);

  private:
    /// Create the streams
    void createStreams_(std::string filename);
    /// Closes the streams
    void closeStreams_();

    /// The space
    concepts::Space<Real>* spc_;
    /// Spatial dimension of the space
    uint dim_;
    /// Number of vectors written to the files
    uint cnt_;
    /// Stream for connections
    std::auto_ptr<std::ofstream> ofsmsh_;
    /// Streams for coordinates
    std::auto_ptr<std::ofstream> ofs1_, ofs2_, ofs3_;

    /// Post processor
    concepts::GlobalPostprocess<Real> postProcess_;
    std::auto_ptr<DataMatlabCell> mtlb_;
  };

  // ******************************************************* DataMatlabCell2 **

  /** Writes the data which describe the elements to the given streams.

      @author Kersten Schmidt, 2004
  */
  class DataMatlabCell2 : public concepts::CellPostprocess<Real> {
  public:
    /** Constructor (for single and repeated output)
	@param os     Streams
	@param dim    Spatial dimension of the elements
	@param sol    The first solution set which should be plotted
    */
    DataMatlabCell2(std::map<std::string, std::ostream*>& os, 
		    uint dim, const concepts::Vector<Real>* sol = 0);

    /** Output of the mesh and vector data of the element
     */
    virtual void operator()(const concepts::Element<Real>& elm)
      throw(concepts::MissingFeature);
    /// not operational
    virtual void operator()(const concepts::Cell& cell)
      throw(concepts::MissingFeature);
    /// Returns the streams to write the data to
    std::map<std::string, std::ostream*> write() {return os_;}
    /// Returns the coefficients of the elements shape functions
    const concepts::Array<Real>& coeff() const {return coeff_;}
    /// Sets the vector of the variable to plot
    void setVec(const concepts::Vector<Real>* sol);
    /// Returns pointer to vector to plot
//     const concepts::Vector<Real>* solution() const {return sol_;}
    /// Number of points written to the files, used for mesh representation
    uint& count() {return cnt_;}
    /// dimension
    uint dim() const {return dim_;}
    /// Map for vertices of the elements
    concepts::DynArray<uint>& map() {return *map_;}

  private:
    /// Spatial dimension of the space
    uint dim_;
    /// Streams to write the data to
    std::map<std::string, std::ostream*>& os_;
    /// Elementwise coefficients of the shape functions
    concepts::Array<Real> coeff_;
    /// Vector of the variable to plot
    const concepts::Vector<Real>* sol_;
    /// Counter
    uint cnt_;
    /// Map for vertices of the elements
    std::auto_ptr<concepts::DynArray<uint> > map_;
  };

  // *********************************************************** DataMatlab2 **

  /** Draws a picture of data in Matlab format and stores the result in
      a single file.

      The data is stored in a Matlab M-file by appending '.m'.
      The data can be read into Matlab by giving the file name as commands
      (and ommiting the ending '.m').

      In general there is the FE-mesh (cells) and the output mesh, which in
      hp-FEM is given by the quadrature points. The data points are given
      in the vertices of the output mesh (<tt>x</tt>, <tt>y</tt>).

      The <em>FE-mesh</em> of triangles or quadrilaterials is given by
      <tt>edgemsh</tt>, while <tt>vtxmsh</tt> consists of the vertices.
      For a FE-mesh with cells with only straight edges its more efficient
      to use <tt>vtxmsh</tt> instead of <tt>edgemesh</tt>.
      In each column there are the numbers of the vertices of one cell given,
      and can be plotted by

      <pre>
      Msh = edgemsh([1:4 1],:);
      plot(x(Msh),y(Msh),'k');
      </pre>

      For <em>scalar spaces</em>, the variables of the data values are
      <tt>z</tt> (single set) or <tt>z{1}</tt> to <tt>z{n}</tt>
      (repeated set).

      <em>Scalar data</em> is plotted in the <em>view from above</em> with
      <pre>
      patch(x(msh),y(msh),z(msh))
      </pre>
      whereby in <tt>msh</tt> the output mesh is given.

      Its possible to emphasize the FE-mesh in this diagramme with
      <pre>
      patch(x(msh),y(msh),z(msh));
      hold on
      Msh = edgemsh([1:4 1],:);
      plot(x(Msh),y(Msh),'k');
      </pre>

      As the command <tt>contour</tt> is only applicable to tensor product
      meshes, there is in general no contour plot possible.

      <em>Scalar data</em> is plotted in the <em>side view</em> with
      <pre>fill3(x(msh),y(msh),z(msh),z(msh));</pre>

      Its possible to emphasize the FE-mesh in this diagramme with
      <pre>
      fill3(x(msh),y(msh),z(msh),z(msh));
      hold on
      plot3(x(edgemsh),y(edgemsh),p(edgemsh),'k','linewidth',2)
      </pre>
      whereby in <tt>edgemsh</tt> represents the points in the output mesh
      lying on the FE mesh.

      The data can be shown unicoloured in side view with
      <pre>
      fill3(x(msh),y(msh),z(msh),ones(size(msh)));
      </pre>

      For <em>vector valued spaces</em>, the variables of the data values are
      <tt>px</tt>,<tt>py</tt> and possibly <tt>pz</tt> (single set) or 
      <tt>px{1}</tt> to <tt>px{n}</tt>, <tt>py{1}</tt> to <tt>py{n}</tt>
      and possibly <tt>pz{1}</tt> to <tt>pz{n}</tt> (repeated set).

      <em>Vectorial data</em> are shown with
      <pre>
      quiver(x,y,px,py);
      </pre>

      Its possible to emphasize the FE-mesh in this diagramme with
      <pre>
      quiver(x,y,px,py);
      hold on
      Msh = edgemsh([1:4 1],:);
      plot(x(Msh),y(Msh),'k');
      </pre>

      This class opens and closes the files and hands the
      streams to DataMatlabCell2.
  */
  class DataMatlab2 {
  public:
    /** Constructor for scalar space
	@param spc Space on which the data should be plotted
	@param filename Name base for the files to be written
	@param dim Spatial dimension of the data (only dim=2 supported)
	@param sol The first solution set which should be plotted
     */
    DataMatlab2(concepts::Space<Real>& spc, const std::string filename,
	       uint dim = 2, const concepts::Vector<Real>* sol = 0);
#if 0
    /** Constructor for edge elements (vector fields)
	@param spc Space on which the data should be plotted
	@param filename Name base for the files to be written
	@param sol The solution which should be plotted

	Without declaration of sol, only the FE mesh will be given.
     */
    DataMatlab2(hp2Dedge::Space& spc, const std::string filename,
	       const concepts::Vector<Real>* sol = 0);
#endif
    
    ~DataMatlab2() {closeStreams_();}

    /** Application operator (repeated output possible)
	@param sol The solution set which should be plotted
    */
    void operator()(const concepts::Vector<Real>& sol){}

  private:
    /// Create the streams
    void createStreams_(std::string filename);
    /// Closes the streams
    void closeStreams_();

    /// The space
    concepts::Space<Real>* spc_;
    /// Spatial dimension of the space
    uint dim_;
    /// Number of vectors written to the files
    uint cnt_;
    /// Stream for output file
    std::auto_ptr<std::ofstream> ofs_;
    /// Pointer to the streams, i.e. *(s_ptr_["vtxmsh"]) for FE mesh
    std::map<std::string, std::ostream*> s_ptr_;
    /// Post processor
    concepts::GlobalPostprocess<Real> postProcess_;
    std::auto_ptr<DataMatlabCell2> mtlb_;
  };


} // namespace graphics

#endif // graphDataMatlab_hh
