/* graphics of the mesh in OpenDX
 */

#ifndef graphMeshDX_hh
#define graphMeshDX_hh

#include <iostream>
#include <string>
#include <list>
#include "space/postProcess.hh"
#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "spaceTraits.hh"
#include "geometry/topology.hh"

namespace hp2D {
  // forward declaration
  class Space;
}

namespace hp3D {
  // forward declaration
  class Space;
}

namespace graphics {

  using concepts::Real;

  // forward declartions
  class VertexList;

  // ********************************************************** Space traits **

  /** Determines the traits class belonging to a space by a default
      template parameter. This class is used by spaceDimension.
      @see spaceDimension
      @author Philipp Frauenfelder, 2002
   */
  template<typename T, typename traits = spaceTraits<T> >
  struct spcDim {
    static inline uint dim() { return traits::dim(); }
  };

  /** Trampoline function for the class spcDim. There, the actual work is
      done: the class calls the respective function in the traits class
      and it determines (by a default template argument) which traits class
      should be used.

      @return the dimension of the space. In a 2D and 3D space,
              quads and hexahedrons are depicted, respectively.

      @param spc Space whose dimension should be returned. This parameter
                 is only used to determine the parameter type.
      @author Philipp Frauenfelder, 2002
  */
  template<typename T>
  uint spaceDimension(const T& spc) {
    return spcDim<T>::dim();
  }

  // **************************************************************** MeshDX **

  /** Draws a picture of a mesh in DX format and stores the
      result in a file. This class is just an executor class for 
      \c MeshDXCell and \c BoundaryDXCell, the real work to draw
      the picture is done there.

      The created file contains information on the cells of the mesh and the
      attribute of the boundary cells (if it is set).
      \c graphics/dx/mesh-boundary.net contains a Visual Program for OpenDX
      to display this information.

      The class has a template parameter which is not very handy. For this
      purpose, there are two trampoline functions \c drawMeshDX which
      create this class and handle the template parameter corretctly.

      @see \c drawMeshDX the trampoline function of this class
      @see \c MeshDXCell on of the classes which does the real work
      @see \c BoundaryDXCell on of the classes which does the real work
      @see <a href="http://www.opendx.org/">Homepage of OpenDX</a>
      @see <a href="http://www.research.ibm.com/dx/">Homepage of OpenDX at IBM research</a>
      @author Philipp Frauenfelder, 2001
      @todo Include more information: attribute for cells, diameter etc.
  */
  template<class F>
  class MeshDX {
  public:
    /** Constructor. The coarsest grid will be depicted. To store the results,
	a new file is created.
	@param msh Mesh which should be drawn
	@param filename File where the data for OpenDX should be stored
    */
    MeshDX(concepts::Mesh& msh, std::string filename)
      throw(concepts::Assertion);

    /** Constructor. The currently used grid in the space will be depicted.
	To store the results, a new file is created.
	@param spc Space which should be drawn
	@param filename File where the data for OpenDX should be stored
	@param dim Spatial dimensions of the space, this parameter
	           can be determined by the spaceDimensionTraits by
		   using spaceDimension
	@see spaceDimension
    */
    MeshDX(concepts::Space<F>& spc, std::string filename, uint dim)
      throw(concepts::Assertion);
  private:
    /// The space
    concepts::Space<F>* spc_;

    /// The mesh
    concepts::Mesh* msh_;

    /// Filename
    std::string filename_;

    /// Private constructor
    void constructor_(concepts::GlobalPostprocess<F>& postProcess,
		      std::string& filename, uint nelm, uint dim);
  };

  /** Trampoline function to create a \c MeshDX. The purpose of this
      function is to hide the template parameters of \c MeshDX.

      @param msh Mesh which should be drawn
      @param filename File where the data for OpenDX should be stored
      @see \c MeshDX for the complete description
      @author Philipp Frauenfelder, 2001
  */
  MeshDX<Real> drawMeshDX(concepts::Mesh& msh, std::string filename);

  /** Trampoline function to create a \c MeshDX. The purpose of this
      function is to hide the template parameters of \c MeshDX.

      This function uses the space traits spaceDimensionTraits to
      fill in the dimension argument of the constructor of \c MeshDX
      with the correct value.
      @param spc Space which should be drawn
      @param filename File where the data for OpenDX should be stored
      @see \c MeshDX for the complete description
      @author Philipp Frauenfelder, 2001
  */
  template<typename T>
  MeshDX<Real> drawMeshDX(T& spc, std::string filename) {
    return MeshDX<Real>(spc, filename, spaceDimension(spc));
  }

  // ************************************************************ MeshDXCell **

  /** Draws a picture of a mesh in DX. The application operator of this class 
      sketches the mesh in DX format for each cell or element.

      You should not use this class directly. Instead, use the executor class
      \c MeshDX or (even simpler) its trampoline function 
      \c drawMeshDX (which hides the template parameter of \c MeshDX).

      @see \c MeshDX the executor class
      @see \c drawMeshDX the trampoline function of \c MeshDX
      @author Philipp Frauenfelder, 2000
  */
  template<class F>
  class MeshDXCell : public concepts::CellPostprocess<F> {
  public:
    /** Constructor. Creates the header of the file and writes the coordinates
	of the vertices to the file. The order of the vertices is stored in
	the vertex list.

	@param os Output stream where the data should be written to.
	@param vtxList List of vertices with coordinates and connectivity
	information
	@param nelms Number of elements
	@param nfaces Number of boundary faces
	@param dim Number of spatial dimensions of the mesh or space

	@pre dim = 2 or dim = 3 corresponds to the spatial dimension
	of the mesh or space.
    */
    MeshDXCell(std::ostream& os, VertexList& vtxList, uint nelms, uint nfaces,
	       uint dim = 2);

    virtual ~MeshDXCell();

    virtual void operator() (const concepts::Element<F>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature, concepts::Assertion);
  private:
    /// Output stream
    std::ostream& os_;

    /// List of vertices
    VertexList& vtxList_;

    /// Spatial dimension of the mesh or space
    uint dim_;
  };

  // ******************************************************** BoundaryDXCell **

  /** Gathers information about boundary faces (or edges in 2D)
      and prepares information to be written to an OpenDX file.

      You should not use this class directly. Instead, use the executor class
      \c MeshDX or (even simpler) its trampoline function 
      \c drawMeshDX (which hides the template parameter of \c MeshDX).
      @see \c MeshDX the executor class
      @see \c drawMeshDX the trampoline function of \c MeshDX
      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class BoundaryDXCell : public concepts::CellPostprocess<F> {
  public:
    BoundaryDXCell(VertexList& vtxList) : vtxList_(vtxList) {}
    virtual ~BoundaryDXCell();

    virtual void operator() (const concepts::Element<F>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature, concepts::Assertion);

    /// Returns the list of faces
    const std::list<const concepts::Connector*>& faces() const
    { return faces_; }
  private:
    /// List of vertices
    VertexList& vtxList_;

    /// List of faces (or edges) which have boundary attribute != 0
    std::list<const concepts::Connector*> faces_;
  };

} // namespace graphics

#endif // graphMeshDX_hh
