/* different lists of the topology
 */

#include "vertexList.hh"
#include "basics/exceptions.hh"
#include "geometry/cell.hh"
#include "space/element.hh"
#include "toolbox/elementGraphics.hh"
#include "basics/debug.hh"

// debugging VertexList
#define geoVertexListAppl_D 0

namespace graphics {

  // ************************************************************ VertexInfo **

  std::ostream& operator<< (std::ostream& os, const VertexInfo& v) {

    __gnu_cxx::hash_set<uint>::const_iterator i;

    os << "VertexInfo(index = " << v.index_ << ", " << v.coord_ 
       << ", Edges: ";
    for (i = v.edges_.begin(); i != v.edges_.end(); ++i) os << *i << ", ";
    os << "Faces: ";
    for (i = v.faces_.begin(); i != v.faces_.end(); ++i) os << *i << ", ";
    os << "Volumes: ";
    for (i = v.volumes_.begin(); i != v.volumes_.end(); ++i) os << *i << ", ";
    os << ')';
    return os;
  }

  // ************************************************************ VertexList **

  VertexList::VertexList(const VertexList& vl) : 
    vertices_(vl.vertices_), vtxInfo_(vl.vtxInfo_) {}

  std::ostream& VertexList::info(std::ostream& os) const {

    __gnu_cxx::hash_map<uint, VertexInfo>::const_iterator i;

    os << "VertexList(" << std::endl;
    for (i = vertices_.begin(); i != vertices_.end(); ++i)
      os << i->first << ": " << i->second << std::endl;
    os << ')';
    return os;
  }

  VertexInfo* VertexList::vertex(uint v) {

    __gnu_cxx::hash_map<uint, VertexInfo>::iterator i = vertices_.find(v);
    if (i != vertices_.end())
      return &(i->second);
    else
      return 0;
  }

  VertexInfo* VertexList::addVertex(uint v, concepts::Real3d coord) {

    __gnu_cxx::hash_map<uint, VertexInfo>::iterator i;

#ifndef NDEBUG
    i = vertices_.find(v);
    assert(i == vertices_.end());
#endif

    i = vertices_.insert(std::pair<uint, 
			 VertexInfo>(v, VertexInfo(coord))).first;
    return &(i->second);
  }

  void VertexList::operator() (const concepts::Cell& cell) {
    DEBUGL(geoVertexListAppl_D, "cell = " << cell);

    const concepts::Cell1* cell1 = dynamic_cast<const concepts::Cell1*>(&cell);
    if (cell1) {
      concepts::Connector0* V;
      concepts::Connector1& cntr = cell1->connector();

      // loop over vertices
      uint i = 0;
      while ( (V = cntr.vertex(i)) ) {
	uint k = V->key();
	concepts::Real3d x = cell1->vertex(i);
	VertexInfo* inf = vertex(k);
	DP(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	   << __LINE__ << "] VertexList::operator() -- vertex " 
	   << i << " key " << k << " at", x);
	if (!inf) {
	  inf = addVertex(k, concepts::Real3d(x));
	  DPL(geoVertexListAppl_D, " not found,", "adding");
	} else
	  DPL(geoVertexListAppl_D, " found,", "not adding");
	if (vtxInfo_ & VOLUMES) {
	  DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	      << __LINE__ << "] VertexList::operator() -- adding volume", 
	      cntr);
	  inf->addVolume(cntr.key());
	}
	++i;
      } // loop over vertices
      return;
    } // if cell1    

    const concepts::Cell2* cell2 = dynamic_cast<const concepts::Cell2*>(&cell);
    if (cell2) {
      concepts::Connector0* V;
      concepts::Connector1* E;
      concepts::Connector2& cntr = cell2->connector();

      // loop over vertices
      uint i = 0;
      while ( (V = cntr.vertex(i)) ) {
	uint k = V->key();
	concepts::Real3d x = cell2->vertex(i);
	VertexInfo* inf = vertex(k);
	DP(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	   << __LINE__ << "] VertexList::operator() -- vertex " 
	   << i << " key " << k << " at", x);
	if (!inf) {
	  inf = addVertex(k, concepts::Real3d(x));
	  DPL(geoVertexListAppl_D, " not found,", "adding");
	} else
	  DPL(geoVertexListAppl_D, " found,", "not adding");
	if (vtxInfo_ & VOLUMES) {
	  DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	      << __LINE__ << "] VertexList::operator() -- adding volume", 
	      cntr);
	  inf->addVolume(cntr.key());
	}
	++i;
      } // loop over vertices

      if (vtxInfo_ & EDGES) {
	DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	    << __LINE__ << "] VertexList::operator() -- adding", 
	    "edges");
	// loop over edges
	i = 0;
	while ( (E = cntr.edge(i)) ) {
	  DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	      << __LINE__ << "] VertexList::operator() -- adding edge", *E);
	  uint e = E->key();

	  for (uint j = 0; j < 2; ++j) {
	    uint k = E->vertex(j)->key();
	    VertexInfo* inf = vertex(k);
	    inf->addEdge(e);
	  }
	  ++i;
	} // loop over edges
      }
      return;
    }

    const concepts::Cell3* cellC = dynamic_cast<const concepts::Cell3*>(&cell);
    if (cellC) {
      uint i, j;
      concepts::Connector0* V;
      concepts::Connector1* E;
      concepts::Connector2* F;
      concepts::Connector3& cntr = cellC->connector();

      // loop over vertices
      i = 0;
      while ( (V = cntr.vertex(i)) ) {
	uint k = V->key();
	concepts::Real3d x = cellC->vertex(i);
	VertexInfo* inf = vertex(k);
	DP(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	   << __LINE__ << "] VertexList::operator() -- vertex " 
	   << i << " key " << k << " at", x);
	if (!inf) {
	  inf = addVertex(k, concepts::Real3d(x));
	  DPL(geoVertexListAppl_D, " not found,", "adding");
	} else
	  DPL(geoVertexListAppl_D, " found,", "not adding");
	if (vtxInfo_ & VOLUMES) {
	  DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	      << __LINE__ << "] VertexList::operator() -- adding volume", 
	      cntr);
	  inf->addVolume(cntr.key());
	}
	++i;
      } // loop over vertices

      if (vtxInfo_ & EDGES) {
	DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	    << __LINE__ << "] VertexList::operator() -- adding", 
	    "edges");
	// loop over edges
	i = 0;
	while ( (E = cntr.edge(i)) ) {
	  DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	      << __LINE__ << "] VertexList::operator() -- adding edge", *E);
	  uint e = E->key();

	  for (uint j = 0; j < 2; ++j) {
	    uint k = E->vertex(j)->key();
	    VertexInfo* inf = vertex(k);
	    inf->addEdge(e);
	  }
	  ++i;
	} // loop over edges
      }

      if (vtxInfo_ & FACES) {
	DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	    << __LINE__ << "] VertexList::operator() -- adding", 
	    "faces");
	// loop over faces
	i = 0;
	while ( (F = cntr.face(i)) ) {
	  DPL(geoVertexListAppl_D, '[' <<  __FILE__ << ", line " 
	      << __LINE__ << "] VertexList::operator() -- adding face", *F);
	  uint f = F->key();

	  j = 0;
	  while ( (V = F->vertex(j)) ) {
	    uint k = V->key();
	    VertexInfo* inf = vertex(k);
	    inf->addFace(f);
	    ++j;
	  }
	  ++i;
	} // loop over faces
      }

      return;
    }
    throw conceptsException(concepts::MissingFeature("cell not supported"));
  }

  void VertexList::operator() (const concepts::Element<Real>& elm) {
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<Real>::VERTEX_LIST,
			       *this);
  }

} // namespace graphics
