/* different lists of the topology
 */

#ifndef geoLists_hh
#define geoLists_hh

#include <iostream>

#if __ICC == 800
#  include <hash_map>
#  include <hash_set>
#  define __gnu_cxx std
#else
#  if __GNUC__ == 3
#    include <ext/hash_map>
#    include <ext/hash_set>
#    if __GNUC_MINOR__ < 1
#      define __gnu_cxx std
#    endif
#  else
#    include <hash_map>
#    include <hash_set>
#    define __gnu_cxx std
#  endif
#endif

#include "space/postProcess.hh"
#include "basics/vectorsMatrices.hh"

namespace concepts {
  // forward declartions
} // namespace concepts

namespace graphics {

  using concepts::Real;

  // ************************************************************ VertexInfo **

  /** Class containing information about one single vertex. Included are
      connectivity and coordinates. A whole list of vertices can be managed
      with VertexList.

      A vertex is create by giving its coordinates. After creation, the
      connectivity information for the edges, faces and volumes can be added.

      @see VertexList
      @author Philipp Frauenfelder, 2001
  */
  class VertexInfo {
    friend std::ostream& operator<< (std::ostream& os, const VertexInfo& v);
  public:
    /** Constructor.

    Creates a vertex with the given coordinates and no edges, faces etc.
    */
    VertexInfo(concepts::Real3d coord, uint index = 0) : 
      index_(index), coord_(coord) {}

    /// Adds an edge to the list of edges
    void addEdge(uint edge) { edges_.insert(edge); }

    /// Adds a face to the list of faces
    void addFace(uint face) { faces_.insert(face); }

    /// Adds a volume to the list of volumes
    void addVolume(uint volume) { volumes_.insert(volume); }

    /// Sets the index of the vertex
    uint& index() { return index_; }
    /// Returns the index of the vertex
    uint index() const { return index_; }

    /// Returns the coordinates
    concepts::Real3d coordinates() const { return coord_; }
  private:
    /** Index of the vertex. This number can be different from 
	the key of the vertex in the list and in the topology.

	@see VertexList
    */
    uint index_;

    /// Coordinates of the vertex
    concepts::Real3d coord_;

    /// List of edges meeting at this vertex
    __gnu_cxx::hash_set<uint> edges_;

    /// List of faces meeting at this vertex
    __gnu_cxx::hash_set<uint> faces_;

    /// List of volumes meeting at this vertex (only useful for 3D)
    __gnu_cxx::hash_set<uint> volumes_;
  };

  // ************************************************************ VertexList **

  /** Class containing a list of vertices.

      Adding a vertex should be done in three steps:
      <ol>
      <li>Check if the vertex already exists (with vertex)
      <li>Add a new vertex if it does not yet exist (with addVertex)
      <li>If needed: fill in the connectivity information for the vertex.
      This is done using the methods of VertexInfo.
      </ol>

      @see VertexInfo
      @author Philipp Frauenfelder, 2001
  */
  class VertexList : public concepts::CellPostprocess<Real> {
  public:
    typedef __gnu_cxx::hash_map<uint, VertexInfo>::iterator iterator;
    typedef __gnu_cxx::hash_map<uint, VertexInfo>::const_iterator
    const_iterator;

    /** Allows to chose if all information should be stored or only some
	parts. To generate an image of the mesh using MeshDX, only
	the coordinates are needed.

	@see MeshDX
    */
    enum vtxInfo { COORD = 0, EDGES = 1, FACES = 2, VOLUMES = 4, ALL = 7 };

    /** Constructor.
     */
    VertexList(enum vtxInfo info = ALL) : vtxInfo_(info) {}

    /// Copy constructor
    VertexList(const VertexList& vl);

    virtual ~VertexList() {}

    /** Looks for the vertex with the given key.
	@return A reference to the vertex, 0 if it does not yet exist
	@param v Key of the vertex
    */
    VertexInfo* vertex(uint v);

    /** Add a new vertex.
	@pre The vertex does not yet exist in the list.
	@param v Key of the vertex
	@param coord Coordinates of the vertex
    */
    VertexInfo* addVertex(uint v, concepts::Real3d coord);

    /// Returns the size of the list
    uint size() const { return vertices_.size(); }

    /// Beginning of the list
    iterator begin() { return vertices_.begin(); }
    const_iterator begin() const { return vertices_.begin(); }

    /// End of the list
    iterator end() { return vertices_.end(); }
    const_iterator end() const { return vertices_.end(); }

    /// Finds a vertex in the list
    iterator find(uint k) { return vertices_.find(k); }
    const_iterator find(uint k) const { return vertices_.find(k); }

    /** Application operator. Adds the vertex of this element to the list
	and updates the connectivity information of the vertex.

	@param cell Element to be considered
    */
    virtual void operator() (const concepts::Cell& cell);

    /** Application operator. Adds the vertex of this element to the list
	and updates the connectivity information of the vertex.

	@param elm Element to be considered
    */
    virtual void operator() (const concepts::Element<Real>& elm);

    /// Returns vtxInfo_
    enum vtxInfo vertexInfo() const { return vtxInfo_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// List of vertices
    __gnu_cxx::hash_map<uint, VertexInfo> vertices_;

    /// What information about the vertices should be stored
    enum vtxInfo vtxInfo_;
  };

} // namespace graphics

#endif // geoLists_hh
