/* hp extension for 1D FEM: ie. determine wether to refine or raise p
 * in an element
 */

#ifndef hpExtension1D_hh
#define hpExtension1D_hh

#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "space/postProcess.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class Vector;

  template<typename F>
  class Array;
}

namespace hp1D {
  using concepts::Real;

  // ******************************************************** RefineOrRaise  **

  /** Refines element or raises its polynomial degree. The decision is based
      on a log linear regression to find out if the Legendre coefficients
      of the solution decrease exponentially.

      The parameter \c theta is used to control this decision. If \f$e^m\f$
      is less than \c theta, the element's polynomial degree is raised,
      otherwise it is refined. m is the slope of the fitted line in
      the regression.

      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class RefineOrRaise : public concepts::CellPostprocess<F> {
  public:
    /** Constructor.
	@param solution Solution vector
	@param theta Decision parameter
	@param spc Space (is used for the refinements)
    */
    RefineOrRaise(const concepts::Vector<F>& solution, const Real theta,
		  concepts::AdaptiveSpace<F,
		  concepts::AdaptiveAdjustP<1> >& spc);
    virtual void operator() (const concepts::Element<F>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature) {}
  private:
    /// Elementwise coefficients of the shape functions
    concepts::Array<F> coeff_;
    /// Legendre coefficients
    concepts::Array<F> legendre_;
    /// Vector of the variable to plot
    const concepts::Vector<F>& solution_;
    /// Space
    concepts::AdaptiveSpace<F, concepts::AdaptiveAdjustP<1> >& spc_;
    /// Decision parameter
    const Real theta_;
  };

} // namespace hp1D

#endif // hpExtension1D_hh
