/* class for the hp FEM space in 1D
 */

#ifndef Space1D_hh
#define Space1D_hh

#include "basics/exceptions.hh"
#include "basics/typedefs.hh"
#include "geometry/mesh.hh"
#include "geometry/boundaryConditions.hh"
#include "space/space.hh"
#include "space/hpMethod.hh"
#include "toolbox/scannerConnectors.hh"
#include "toolbox/dynArray.hh"
#include "element.hh"

namespace concepts {

  // ****************************************************************** Scan **

  template<>
  class Scan<hp1D::Element<Real> > :
    public concepts::Scan<concepts::Element<Real> > {
  public:  
    hp1D::Element<Real>& operator++(int) = 0;
  };

} // namespace concepts

namespace hp1D {
  using concepts::Real;

  // ***************************************************************** Space **

  /** A 3D hp FEM space with continuous, picewise polynomial basis functions.
      @author Philipp Frauenfelder, 2002
  */
  class Space : public concepts::AdaptiveSpace<Real,
		concepts::AdaptiveAdjustP<1> > {
  public:
    typedef concepts::Scan<hp1D::Element<Real> > Scan;

    /** Constructor. Scans the mesh and sets the cells in the mesh active
	and the level of refinement and the polynomial degree in
	all cells to the given values. \c rebuild_ is set to true, ie. 
	if the mesh is used it will firstly be rebuilt.

	@param msh The domain of interest partitioned into a mesh.
	@param l Level of refinement
	@param p Degree of the polynomials to be used.
	@param bc Boundary conditions
    */
    Space(concepts::Mesh1& msh, uint l, uint p,
	  concepts::BoundaryConditions* bc = 0);
    virtual ~Space();
    
    inline uint dim() const;
    inline uint nelm() const;
    inline Scan* scan() const;

    inline uint dim();
    inline uint nelm();
    inline Scan* scan();

    virtual void adjust(const concepts::Element<Real>& elm,
			const concepts::AdaptiveAdjustP<1>& a);

    /// Rebuilds the space after an adjustment with \c adjust.
    void rebuild();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Mesh
    concepts::Mesh1& msh_;
    /// Boundary conditions
    concepts::BoundaryConditions* bc_;

    /** If true: the elements have to be rebuilt. Some elements have been
	marked for refinement or for coarsening and the elements are
	outdated.
    */
    bool rebuild_;

    /// Dimension of the FE space
    uint dim_;

    /// Number of elements currently active in the mesh
    uint nelm_;

    /// Linked list of the elements
    concepts::Joiner<Element<Real>*, 1>* elm_;

    /// Hash table of the control information for the vertices
    concepts::DynArray<concepts::AdaptiveControl> ctrl0_;
    /// Hash table of the control information for the 1D elements
    concepts::DynArray<concepts::AdaptiveControlP<1> > ctrl1_;

    /// Hash table of the adjustment information of the elements
    concepts::DynArray<concepts::AdaptiveAdjustP<1> > adj_;

    void meshAndPoly_(concepts::Edge1d& cell, int l, int& L, int& P);
    void buildElements_(concepts::Edge1d& cell,
			concepts::TColumn<Real>* T0 = 0);
    void enforceBC_(concepts::Edge1d& cell);
  };

  uint Space::dim() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return dim_; 
  }

  uint Space::nelm() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return nelm_; 
  }

  Space::Scan* Space::scan() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return new concepts::PListScan<Element<Real> >(*elm_);
  }

  uint Space::dim() {
    if (rebuild_) rebuild();
    return dim_; 
  }

  uint Space::nelm() {
    if (rebuild_) rebuild();
    return nelm_; 
  }

  Space::Scan* Space::scan() {
    if (rebuild_) rebuild();
    return new concepts::PListScan<Element<Real> >(*elm_);
  }

} // namespace hp1D

#endif // Space1D_hh
