/* Bilinear forms for hp 2D FEM
 */

#ifndef hp2dbilinearform_hh
#define hp2dbilinearform_hh

#include <memory>
#include "operator/bilinearForm.hh"
#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "toolbox/formula.hh"

namespace concepts {
  // forward declarations
  template<class F>
  class Element;

  template<class F>
  class ElementMatrix;

  template<class F>
  class Array;

  template<typename F>
  class DynArray;
}

namespace hp2D {
  // forward declarations
  class Quad;

  using concepts::Real;

  // *************************************************************** Laplace **

  /** A function class to calculate element matrices for the Laplacian.

      @test test::TestMatrices2D
      @author Kersten Schmidt, 2003
  */
  class Laplace : public concepts::BilinearForm<Real> {
  public:
    /// Default constructor
    Laplace() : frm_(0), cellwise_(0) {}
    /** Constructor. The formula \c frm is evaluated in each quadrature
	point.
    */
    Laplace(const concepts::Formula& frm) : frm_(frm.clone()), cellwise_(0) {}
    /** Constructor. The results of \c frm and \c cellwise are summed up
	at each quadrature point and taken as coefficient.
	@param frm Fomula (evaluated at each quadrature point)
	@param cellwise Cellwise coefficients. The attribute of the cell
	is the index of the array \c cellwise
     */
    Laplace(const concepts::Formula& frm,
	    const concepts::DynArray<Real>& cellwise);
    virtual ~Laplace();

    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY, 
			    concepts::ElementMatrix<Real>& em); 
    void operator()(const Quad& elmX, const Quad& elmY, 
		    concepts::ElementMatrix<Real>& em);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Intermediate data for element matrix computation
    concepts::Array<Real> jacobian_;
    concepts::Array<concepts::MapReal2d> jacobianInv_;
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;
    /// Cellwise values
    const concepts::DynArray<Real>* cellwise_;
  };

  // ************************************************************** Identity **

  /** A function class to calculate element matrices for the mass matrix.

      @test test::TestMatrices2D
      @author Kersten Schmidt, 2003
  */
  class Identity : public concepts::BilinearForm<Real> {
  public:
    /// Constructor
    Identity(const concepts::Formula* frm = 0);
    virtual ~Identity();

    void operator()(const concepts::Element<Real>& elmX,
		    const concepts::Element<Real>& elmY, 
		    concepts::ElementMatrix<Real>& em); 
    void operator()(const Quad& elmX, const Quad& elmY, 
		    concepts::ElementMatrix<Real>& em);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Intermediate data for element matrix computation
    concepts::Array<Real> jacobian_;
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;
  };

} // namespace hp2D

#endif // hp2dbilinearform_hh
