/* discretization of div * div for Maxwell equations in frequency domain
 */

#include "divdiv.hh"
#include "quad.hh"
#include "shapeFunction2d.hh"
#include "sumfact.hh"
#include "transmissionWeight.hh"
#include "toolbox/coeffIterator.hh"
#include "space/element.hh"
#include "vectorial/vectorial.hh"

// debugging
#include "basics/debug.hh"

#define hpDivDivAppl_D 0
#define hpDivDivApplTime_D 0
#define hpDivDivApplMem_D 0
#define hpDivDivApplCoeff_D 0

namespace hp2D {

  // **************************************************************** DivDiv **

  template<typename Weight>
  concepts::Array<Real> DivDiv<Weight>::jacobianDiv_;
  template<typename Weight>
  concepts::Array<concepts::MapReal2d> DivDiv<Weight>::jacobianInvDiv_;
  template<typename Weight>
  const hp2D::Quad* DivDiv<Weight>::elmDiv_(0);
  template<typename Weight>
  concepts::Array<std::auto_ptr<DivDiv<Weight> > > DivDiv<Weight>::divdiv_(0);

  template<typename Weight>
  DivDiv<Weight>::DivDiv(concepts::Array<Real>& jacobian,
			 concepts::Array<concepts::MapReal2d>& jacobianInv,
			 const Quad** oldElm, SingularSet& singularities,
			 const uint i, const uint j,
			 const concepts::Formula* frm) :
    i_(i), j_(j), jacobian_(jacobian), jacobianInv_(jacobianInv),
    oldElm_(oldElm), singularities_(singularities), weight_(), frm_(0) {
    if (frm)
      frm_.reset(frm->clone());
  }

  template<typename Weight>
  DivDiv<Weight>::~DivDiv() {}

  template<typename Weight>
  void DivDiv<Weight>::operator()(const concepts::Element<Real>& elmX, 
			  const concepts::Element<Real>& elmY, 
			  concepts::ElementMatrix<Real>& em) {
    const Quad* elmYH = dynamic_cast<const Quad*>(&elmY);
    if (elmYH) {
      const Quad* elmXH = dynamic_cast<const Quad*>(&elmX);
      if (elmXH) { (*this)(*elmXH, *elmYH, em); return; }
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  template<typename Weight>
  void DivDiv<Weight>::operator()(const Quad& elmX, const Quad& elmY, 
			  concepts::ElementMatrix<Real>& em) {
    uint n = elmX.T().m();
    em.resize(n, elmY.T().m());
    em.zeros();

    if (elmX.support() == elmY.support()) {
      DEBUGL(hpDivDivAppl_D, elmX);
      DEBUGL(hpDivDivApplMem_D, "em is (" << n << 'x' << elmY.T().m() << ')');

      em.zeros();

      // quadrature points
      const uint nqx = elmX.integrationX()->n();
      const uint nqy = elmX.integrationY()->n();
      const Real* const px = elmX.integrationX()->abscissas();
      const Real* const py = elmX.integrationY()->abscissas();

#if hpDivDivApplTime_D
      concepts::ResourceMonitor timer;
#endif
      DEBUGL(hpDivDivAppl_D, "elmX = " << &elmX << ", oldElm = "
	     << *oldElm_  << ", i = " << i_ << ", j = " << j_);
      if (&elmX != *oldElm_){
	// Jacobian
	jacobian_.resize(nqx*nqy);
        DEBUGL(hpDivDivApplMem_D, "reserved " << (nqx*nqy)
	       << " for jacobian in DivDiv-BF");
	Real* dx = (Real*)jacobian_;
	jacobianInv_.resize(nqx*nqy);
	DEBUGL(hpDivDivApplMem_D, "reserved " << (nqx*nqy)
	       << " for jacobianInv in DivDiv-BF");
	concepts::MapReal2d* jacInv = (concepts::MapReal2d*)jacobianInv_;
	const Real* pxAcc = px;
	for (uint i = 0; i < nqx; ++i, ++pxAcc) {
	  const Real* pyAcc = py;
	  for (uint j = 0; j < nqy; ++j, ++pyAcc) {
	    const concepts::Real2d refPoint( (*pxAcc + 1.0)/2.0,
					     (*pyAcc + 1.0)/2.0 );
	    concepts::MapReal2d jacobian(elmX.jacobian(refPoint[0],
						       refPoint[1]));
	    Real ws = weight_(singularities_, refPoint, &elmX);
	    Real f = 1.0;
	    if (frm_.get())
	      f = (*frm_)(elmX.cell().chi(refPoint[0], refPoint[1]));
	    *dx++ = jacobian.determinant() * ws * f*f;
	    
	    // inverse of the jacobian with three columns (a1, a2, a3)
	    concepts::MapReal2d jacinv(jacobian.inverse());
	    // store it in
	    *jacInv++ = jacinv;
	  }
	}
#if hpDivDivApplTime_D
	DEBUGL(hpDivDivApplTime_D, "TIME for Jacobian: " << timer);
#endif
      } else {
	DEBUGL(hpDivDivApplMem_D, "took already computed jacobian");
      }

      DEBUGL(hpDivDivAppl_D, "Jacobian: " << jacobian_);

      concepts::Array<concepts::MapReal2d> coeffMatrix_(nqx*nqy);
      //     coeffMatrix_.resize(nqx*nqy);
      concepts::MapReal2d* jacInv = (concepts::MapReal2d*)jacobianInv_;
      concepts::MapReal2d* coeffMatrix = (concepts::MapReal2d*)coeffMatrix_;
      for (uint i = 0; i < nqx; ++i) {
	for (uint j = 0; j < nqy; ++j) {
	  concepts::MapReal2d jacinv(*jacInv++);

	  DEBUGL(hpDivDivApplCoeff_D,"jacinv = " << jacinv);

	  // calculate a_i*a_j'
	  concepts::Real2d firstColumn(jacinv.column(i_));
	  concepts::Real2d secondColumn(jacinv.column(j_));
	  coeffMatrix->zeros();

	  DEBUGL(hpDivDivApplCoeff_D,"firstColumn = " << firstColumn
		 << ", secondColumn = " << secondColumn);

	  coeffMatrix->rank1Product(firstColumn, secondColumn);
	  DEBUGL(hpDivDivApplCoeff_D,"coeffMatrix = " << *coeffMatrix);
	  ++coeffMatrix;
	}
      }

      const Real* const wx = elmX.integrationX()->weights();
      const Real* const wy = elmX.integrationY()->weights();

      // shapefunctions
      const uint nsx = elmX.shpfctX()->n();
      const uint nsy = elmX.shpfctY()->n();
      const Real* const sx = elmX.shpfctX()->values();
      const Real* const sy = elmX.shpfctY()->values();
      const Real* const sdx = elmX.shpfctDX()->values();
      const Real* const sdy = elmX.shpfctDY()->values();
    
      ShapeFunction2D<Real> u;
      u.n(0, nsx);        u.n(1, nsy);
      u.nq(0, nqx);       u.nq(1, nqy);
      u.values(0, sx);    u.values(1, sy);
      u.abscissas(0, px); u.abscissas(1, py);
      u.weights(0, wx);   u.weights(1, wy);
      u.T(&(elmX.T()));

      ShapeFunction2D<Real> v;
      v.n(0, nsx);        v.n(1, nsy);
      v.nq(0, nqx);       v.nq(1, nqy);
      v.values(0, sx);    v.values(1, sy);
      v.abscissas(0, px); v.abscissas(1, py);
      v.weights(0, wx);   v.weights(1, wy);
      v.T(&(elmY.T()));
    
      SumFactorization sumfact;
      em.zeros();
      for (uint i = 0; i < 2; ++i) {
	switch (i) {
	case 0: u.values(i, sdx); break;
	case 1: u.values(i, sdy); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
	for (uint j = 0; j < 2; ++j) {
	  switch (j) {
	  case 0: v.values(j, sdx); break;
	  case 1: v.values(j, sdy); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	  concepts::CoeffIterator<Real, concepts::MapReal2d>
	    coeff((concepts::MapReal2d*)coeffMatrix_, i, j);

	  DEBUGL(hpDivDivAppl_D, "Coeff-Matrix ( " << i << ", " << j << "): ");

	  sumfact(em, jacobian_, &coeff, u, v);

	  DEBUGL(hpDivDivAppl_D, "em = " << em);

	  switch (j) {
	  case 0: v.values(j, sx); break;
	  case 1: v.values(j, sy); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	}
	switch (i) {
	case 0: u.values(i, sx); break;
	case 1: u.values(i, sy); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
      }


#if hpDivDivApplTime_D
      DEBUGL(hpDivDivApplTime_D, "TIME for integration: " << timer);
#endif

      DEBUGL(hpDivDivAppl_D, "em = " << em);
#if hpDivDivApplTime_D
      DEBUGL(hpDivDivApplTime_D, "TIME for symmetric em: " << timer);
#endif
      *oldElm_ =  &elmX;

      DEBUGL(hpDivDivAppl_D, "elmX = " << &elmX << ", oldElm = " << *oldElm_
	     << ", &oldElm = " << oldElm_ );
    } // elmX == elmY
  }
 
  template<typename Weight>
  std::ostream& DivDiv<Weight>::info(std::ostream& os) const{
    return os << "DivDiv(weight = " << weight_
// 	      << ", i = " << i_ << ", j = " << j_
// 	      << ", jacobian = " << jacobian_ << ", jacobianInv = "
// 	      << jacobianInv_ << ", oldElm = " << *oldElm_
	      << ")";
  }

  template<typename Weight>
  void DivDiv<Weight>::setup(vectorial::BilinearForm<Real>& bf,
			     hp2D::SingularSet& singularities,
			     const concepts::Formula* frm) {
    divdiv_.resize(4);
    conceptsAssert(bf.vdim() == 2, concepts::Assertion());
    for (uint k = 0; k < 4; ++k){
      uint i = (uint)k/2, j = k%2;

      divdiv_[k].reset
	(new DivDiv<Weight>(jacobianDiv_, jacobianInvDiv_, &elmDiv_,
			    singularities, i, j, frm));
      bf.put(*divdiv_[k], i, j);
    }
  }

  // Weight 1.0
  template class DivDiv<TrivialWeight>;
  // Weight of squared distance
  template class DivDiv<ShortestDist>;
  template class DivDiv<DistancePost<ShortestDist, Postprocess4> >;
  template class DivDiv<DistancePost<ShortestDist, Postprocess9> >;
  template class DivDiv<DistancePost<ShortestDist, PostprocessSqrt> >;
  // Weight for transmission problem: distance to coordinate axis
  template class DivDiv<TransmissionWeight>;
  template class DivDiv<DistancePost<TransmissionWeight, Postprocess4> >;
  template class DivDiv<DistancePost<TransmissionWeight, Postprocess7> >;
  template class DivDiv<DistancePost<TransmissionWeight, Postprocess8> >;
  template class DivDiv<DistancePost<TransmissionWeight, Postprocess9> >;
  template class DivDiv<DistancePost<TransmissionWeight, PostprocessSqrt> >;
  // Weight for transmission problem: product of distances to coordinate axis
  template class DivDiv<TransmissionWeightProd>;
  template class DivDiv<DistancePost<TransmissionWeightProd, Postprocess4> >;
  template class DivDiv<DistancePost<TransmissionWeightProd, Postprocess7> >;
  template class DivDiv<DistancePost<TransmissionWeightProd, Postprocess8> >;
  template class DivDiv<DistancePost<TransmissionWeightProd, Postprocess9> >;
  template class DivDiv<DistancePost<TransmissionWeightProd, PostprocessSqrt> >;
} // namespace hp2D
