/* discretization of div * div for Maxwell equations in frequency domain
 */

#ifndef divdiv2D_hh
#define divdiv2D_hh

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "toolbox/array.hh"
#include "toolbox/formula.hh"
#include "operator/bilinearForm.hh"
#include "toolbox/resourceMonitor.hh"
#include "hp2D/singularSet.hh"
#include "hp2D/shortestDist.hh"

namespace vectorial {
  template<typename F>
  class BilinearForm;
} // namespace vectorial

namespace hp2D {
  // forward declarations
  class Quad;

  using concepts::Real;

  // **************************************************************** DivDiv **

  /** A function class to calculate element matrices 
      for the Div u*Div v Bilinearform

      The Jacobian, its inverse and additional quantities only depending on
      the physical coordinates are only computed once for both
      components of the bilinear form. To this end, several additional
      parameters have to be given to the constructor (see there).

      @author Kersten Schmidt, 2003
  */
  template<typename Weight = TrivialWeight>
  class DivDiv : public concepts::BilinearForm<Real> {
  public:
    /** Constructor
	@param jacobian Reference to the storage space for the Jacobian.
	Here, also other quantities are stored (premultiplied).
	@param jacobianInv Reference to the storage space for the inverse
	of the Jacobian.
	@param oldElm Pointer to the element of the last invocation to
	detect if the information in \c jacobian and \c jacobianInv is
	still valid.
	@param singularities List of singular edges and vertices.
	@param i Index of the bilinear form in the vector valued bilinear form
	@param j Index of the bilinear form in the vector valued bilinear form
	@param frm Formula for the dielectic \f$\epsilon\f$
    */
    DivDiv(concepts::Array<Real>& jacobian,
	   concepts::Array<concepts::MapReal2d>& jacobianInv,
	   const Quad** oldElm, SingularSet& singularities,
	   const uint i = 0, const uint j = 0,
	   const concepts::Formula* frm = 0);
    virtual ~DivDiv();

    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY, 
			    concepts::ElementMatrix<Real>& em); 
    void operator()(const Quad& elmX, const Quad& elmY, 
		    concepts::ElementMatrix<Real>& em);

    /** Sets up the bilinear form for the div-div part for Maxwell's
	equation using weighted regularization.
	@param bf Resulting bilinear form (use this to compute the matrix)
	@param singularities Singularities needed for computing the weight
	@param frm Formula for the dielectricity (if none is given, 1 is
	       assumed)
	@pre The number of components of \c bf needs to be 2 x 2.
    */
    static void setup(vectorial::BilinearForm<Real>& bf,
		      hp2D::SingularSet& singularities,
		      const concepts::Formula* frm = 0);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Row and column inside the vectorial bilinear form
    uint i_, j_;

    /// Intermediate data for element matrix computation
    concepts::Array<Real> &jacobian_;
    concepts::Array<concepts::MapReal2d> &jacobianInv_;

    const Quad** oldElm_;

    /// Set of singular points
    SingularSet& singularities_;
    /// Weight of regularization
    Weight weight_;

    /// Formula for dielectricity \f$\epsilon\f$
    std::auto_ptr<const concepts::Formula> frm_;

    static concepts::Array<Real> jacobianDiv_;
    static concepts::Array<concepts::MapReal2d> jacobianInvDiv_;
    static const hp2D::Quad* elmDiv_;
    static concepts::Array<std::auto_ptr<DivDiv<Weight> > > divdiv_;
  };

} // namespace hp2D

#endif // divdiv2D_hh
