/* Linear forms for hp 2D FEM
 */

#include <memory>
#include <cstring>
#include <sstream>
#include "basics/vectorsMatrices.hh"
#include "basics/exceptions.hh"
#include "toolbox/array.hh"
#include "toolbox/formula.hh"
#include "toolbox/boundary.hh"
#include "geometry/boundaryConditions.hh"
#include "space/element.hh"
#include "hp2D/quad.hh"
#include "hp2D/linearForm.hh"
#include "basics/debug.hh"

// debugging
#define hp2DRieszConstr_D 0
#define hp2DRieszAppl_D 0
#define hp2DRieszApplBC_D 0

namespace hp2D {

  // ***************************************************************** Riesz **

  Riesz::Riesz(const concepts::Formula& frm,
	       concepts::BoundaryConditions* bc) :
    frm_(frm.clone()), jacobian_(0), bc_(bc) {
    DEBUGL(hp2DRieszConstr_D, *frm_);
    DEBUGL(hp2DRieszConstr_D, *bc_);
  }

  Riesz::~Riesz() {}

  void Riesz::operator()(const concepts::Element<Real>& elm,
			 concepts::ElementMatrix<Real>& em) {
    const Quad* elmH = dynamic_cast<const Quad*>(&elm);
    if (elmH) { this->operator()(*elmH, em); return; }
    std::stringstream error;
    error << "element " << elm << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }

  void Riesz::operator()(const Quad& elm,
			 concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();
    DEBUGL(hp2DRieszAppl_D, "elm = " << elm);
    DEBUGL(hp2DRieszAppl_D, "n = " << n);

    // quadrature points
    const uint nqx = elm.integrationX()->n();
    const uint nqy = elm.integrationY()->n();
    const Real * const wx = elm.integrationX()->weights();
    const Real * const wy = elm.integrationY()->weights();
    const Real * px = elm.integrationX()->abscissas();
    const Real * py = elm.integrationY()->abscissas();

    // Jacobian
    jacobian_.resize(nqx*nqy);
    DEBUGL(hp2DRieszAppl_D, "reserved " << (nqx*nqy) << " for jacobian_");
    Real* dx = (Real*)jacobian_;
    const Real * pxAcc = px;
    for (uint i = 0; i < nqx; ++i) {
      const Real * pyAcc = py;
      for (uint j = 0; j < nqy; ++j) {
	*dx++ = elm.jacobianDeterminant( (*pxAcc + 1.0)/2.0, 
					 (*pyAcc + 1.0)/2.0 )
	    * (*frm_)(elm.chi( (*pxAcc + 1.0)/2.0, (*pyAcc++ + 1.0)/2.0) );
      }
      pxAcc++;
    }

    // shapefunctions
    const uint nsx = elm.shpfctX()->n();
    const uint nsy = elm.shpfctY()->n();
    const Real * sx = elm.shpfctX()->values();
    const Real * sy = elm.shpfctY()->values();

    // integration
    Real* intermed = (Real*)em;
    std::memset(intermed, 0, sizeof(Real)*nsx*nsy);
    for (uint iy = 0; iy < nsy; ++iy) {
      for (uint ix = 0; ix < nsx; ++ix) {
	const Real * wxAcc = wx;
	const Real * six = sx + ix*nqx;
	dx = (Real*)jacobian_;
	for (uint I = 0; I < nqx; ++I) {
	  const Real * wyAcc = wy;
	  const Real * siy = sy + iy*nqy;
	  for (uint J = 0; J < nqy; ++J) {
	    *intermed += *dx++ 
	      * *wxAcc * *wyAcc++
	      * *six * *siy++;
	  } // J
	  wxAcc++;
	  six++;
	} // I
	intermed++;
      } // ix
    } // iy

    DEBUGL(hp2DRieszAppl_D, "volume integration: em = " << em);

    // treating Neumann boundary conditions
    if (bc_) {
      DEBUGL(hp2DRieszApplBC_D, "bc = " << *bc_);

      // loop over the edges
      concepts::Connector1* edge = 0;
      for (uint j = 0; (edge = elm.support().edge(j)) != 0; ++j) {
	uint attrib = edge->attrib();
	DEBUGL(hp2DRieszApplBC_D, *edge << " has attribute " << attrib);

	if ( (*bc_)(attrib).type() == concepts::Boundary::NEUMANN) {
	  DEBUGL(hp2DRieszApplBC_D, *edge << " has NEUMANN conditions " << 
		 (*bc_)(attrib));
	  // Jacobian: length of edge / 2 (integration intervall is [-1,1])
	  const Real jac = (elm.vertex(j)-elm.vertex( (j+1)%4 )).l2() / 2.0;

	  // store points and shapefunctions
	  px = elm.integrationX()->abscissas();
	  py = elm.integrationY()->abscissas();
	  sx = elm.shpfctX()->values();
	  sy = elm.shpfctY()->values();

	  // dummy points and shape functions which lie on the end points
	  std::auto_ptr<concepts::Quadrature<0> > qDummy(0);
	  std::auto_ptr<concepts::Karniadakis<1,0> > sDummy(0);

	  uint ni = 0; // number of quadrature points
	  const Real* wi = 0; // weights of the quadrature points
	  uint I = 100; // counter variables in the loops
	  uint J = 300; // is changed and used as a constant
	  uint* xcnt = 0, * ycnt = 0; // indices in the 3 directions

	  switch(j) {
	  case 0:
	    xcnt = &I; ycnt = &J; J = 0;
	    ni = elm.integrationX()->n(); wi = elm.integrationX()->weights();
	    DEBUGL(hp2DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationY()) << ", " << *(elm.shpfctY()));
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationY()->n()));
	    py = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctY()->n()-1, qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sy = sDummy->values();
	    break;
	  case 1:
	    xcnt = &J; ycnt = &I; J = elm.integrationX()->n()-1;
	    ni = elm.integrationY()->n(); wi = elm.integrationY()->weights();
	    DEBUGL(hp2DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationX()) << ", " << *(elm.shpfctX()));
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationX()->n()));
	    px = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctX()->n()-1, qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sx = sDummy->values();
	    break;
	  case 2:
	    xcnt = &I; ycnt = &J; J = elm.integrationY()->n()-1;
	    ni = elm.integrationX()->n(); wi = elm.integrationX()->weights();
	    DEBUGL(hp2DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationY()) << ", " << *(elm.shpfctY()));
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationY()->n()));
	    py = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctY()->n()-1, qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sy = sDummy->values();
	    break;
	  case 3:
	    xcnt = &J; ycnt = &I; J = 0;
	    ni = elm.integrationY()->n(); wi = elm.integrationY()->weights();
	    DEBUGL(hp2DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationX()) << ", " << *(elm.shpfctX()));
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationX()->n()));
	    px = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctX()->n()-1, qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sx = sDummy->values();
	    break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  } // switch edge
	  DEBUGL(hp2DRieszApplBC_D, "after new integration rule: " <<
		 *qDummy << ", " << *sDummy);
	  DEBUGL(hp2DRieszApplBC_D, "xcnt = " << *xcnt << ", ycnt = " 
		 << *ycnt << ", ni = " << ni);

	  // Evaluation of formula on boundary
	  jacobian_.resize(ni);
	  dx = (Real*)jacobian_;
	  for (I = 0; I < ni; ++I)
	    *dx++ = jac * (*bc_)(attrib)(elm.chi( (px[*xcnt] + 1.0)/2.0,
						  (py[*ycnt] + 1.0)/2.0));

	  // integration
	  Real* intermed = (Real*)em;
	  for (uint iy = 0; iy < nsy; ++iy) {
	    for (uint ix = 0; ix < nsx; ++ix) {
	      const Real * six = sx + ix*nqx;
	      const Real * siy = sy + iy*nqy;
	      const Real * wiAcc = wi;
	      dx = (Real*)jacobian_;
	      for (I = 0; I < ni; ++I) {
		*intermed += *dx++ * *wiAcc++ // jacobian and weight
		  * six[*xcnt] * siy[*ycnt]; // shape function at correct point
	      } // for I
	      intermed++;
	    } // ix
	  } // iy
	} // if NEUMANN
      } // for edges
    } // if bc_

    DEBUGL(hp2DRieszAppl_D, "em = " << em);
  }

  std::ostream& Riesz::info(std::ostream& os) const {
    return os << "Ries(" << *frm_ << ')';
  }

} // namespace hp2D
