/* hexahedral element for 2D hp FEM
 */

#include <iostream>
#include "quad.hh"
#include "basics/debug.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/typedefs.hh"

// debugging
#define QuadConstr_D 0
#define QuadConstrT_D 0
#define QuadDestr_D 0

namespace hp2D {

  // ****************************************************************** Quad **

  std::auto_ptr<QuadGraphics> Quad::graphics_(0);

  enum Quad::intRule Quad::integrationType_ = Quad::GAUSS_JACOBI;
  uint Quad::constNumerOfPoints_ = 10;
  uint Quad::addNumberOfPoints_ = 2;
  bool Quad::useConstantNumberOfPoints_ = false;

  Quad::Quad(concepts::Quad2d& cell, ushort* p,
	     concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1) :
    Element<Real>(T1, p), cell_(cell),
    shpfctX_(0), shpfctY_(0), intX_(0), intY_(0) {
    if (T0)
      T_.append(T0);

    DEBUGL(QuadConstrT_D, "T = " << T_);

    DEBUGL(QuadConstr_D, "computing shape functions for x");
    uint k = addNumberOfPoints_ + p[0];
    if (useConstantNumberOfPoints_) k = constNumerOfPoints_;
    switch (integrationType_) {
    case GAUSS_LOBATTO:
      intX_.reset(new concepts::QuadratureRuleGaussLobatto(k)); break;
    case GAUSS_JACOBI:
      intX_.reset(new concepts::QuadratureRuleGaussJacobi(k)); break;
    default: throw conceptsException
	       (concepts::MissingFeature("quadrature rule not defined"));
    }
    // shape functions
    concepts::Karniadakis<1,0>* tmp_Karn
      = new concepts::Karniadakis<1,0>(p[0], intX_->abscissas(), intX_->n());
    shpfctX_.reset(tmp_Karn);
    // derivatives of shape functions
    concepts::Karniadakis<1,1>* tmp_KarnD
      = new concepts::Karniadakis<1,1>(p[0], intX_->abscissas(), intX_->n());
    shpfctDX_.reset(tmp_KarnD);
    DEBUGL(QuadConstr_D, "computed shape functions for x: " << *shpfctX_);

    DEBUGL(QuadConstr_D, "computing shape functions for y");
    k = addNumberOfPoints_ + p[1];
    if (useConstantNumberOfPoints_) k = constNumerOfPoints_;
    switch (integrationType_) {
    case GAUSS_LOBATTO:
      intY_.reset(new concepts::QuadratureRuleGaussLobatto(k)); break;
    case GAUSS_JACOBI:
      intY_.reset(new concepts::QuadratureRuleGaussJacobi(k)); break;
    default: throw conceptsException
	       (concepts::MissingFeature("quadrature rule not defined"));
    }
    // shape functions
    tmp_Karn = new concepts::Karniadakis<1,0>(p[1], intY_->abscissas(),
					      intY_->n());
    shpfctY_.reset(tmp_Karn);
    // derivatives of shape functions
    tmp_KarnD = new concepts::Karniadakis<1,1>(p[1], intY_->abscissas(),
					       intY_->n());
    shpfctDY_.reset(tmp_KarnD);
  }

  Quad::~Quad() {
    DEBUGL(QuadDestr_D, "done.");
  }

  concepts::Real3d Quad::vertex(uint i) const {
    return cell_.vertex(i);
  }

  const concepts::Quad2d& Quad::cell() const {
    return cell_;
  }

  const concepts::ElementGraphics<Real>* Quad::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new QuadGraphics());
    return graphics_.get();
  }

  void Quad::setIntegrationRule(enum Quad::intRule rule, bool constant,
				uint points) {
    integrationType_ = rule;
    useConstantNumberOfPoints_ = constant;
    if (constant) constNumerOfPoints_ = points;
    else addNumberOfPoints_ = points;
  }

  std::string Quad::integrationRule() {
    std::stringstream res;
    res << "rule = " << integrationType_ << ", with "
	<< (useConstantNumberOfPoints_ ? "constant" : "dynamic")
	<< " number of points = "
	<< (useConstantNumberOfPoints_ ? constNumerOfPoints_ :
	    addNumberOfPoints_)
	<< std::ends;
    return res.str();
  }

  std::ostream& Quad::info(std::ostream& os) const {
    const ushort* P = p();
    return os << "Quad(" << cell_ << ", p = (" << P[0] << ", " << P[1]
	      << "), " << T_ << ')';
  }

} // namespace hp2D
