/* quadrilateral element for 2D hp FEM
 */

#ifndef quad_hh
#define quad_hh

#include <iostream>
#include <memory>
#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "geometry/cell2D.hh"
#include "integration/quadRule.hh"
#include "space/tmatrix.hh"
#include "space/hpMethod.hh"
#include "element.hh"
#include "quadGraphics.hh"
#include "integration/karniadakis.hh"

namespace hp2D {

  using concepts::Real;

  // ****************************************************************** Quad **

  /** A 2D FEM element: a quad.
   */
  class Quad : public Element<Real> {
  public:
    /** Constructor
	@param cell Cell on which the element is defined
	@param p Polynomial degree (might be anisotropic)
	@param T0 Part of the T matrix
	@param T1 Part of the T matrix
    */
    Quad(concepts::Quad2d& cell, ushort* p, 
	 concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1);
    virtual ~Quad();

    virtual const concepts::Quad& support() const { return cell_.connector(); }
    virtual concepts::Real3d vertex(uint i) const;
    virtual const concepts::Quad2d& cell() const;

    /// Returns the shape functions in x direction
    inline const concepts::Karniadakis<1,0>* shpfctX() const {
      return shpfctX_.get(); }
    /// Returns the shape functions in y direction
    inline const concepts::Karniadakis<1,0>* shpfctY() const {
      return shpfctY_.get(); }

    /// Returns the derivatives of the shape functions in x direction
    inline const concepts::Karniadakis<1,1>* shpfctDX() const {
      return shpfctDX_.get(); }
    /// Returns the shape functions in y direction
    inline const concepts::Karniadakis<1,1>* shpfctDY() const {
      return shpfctDY_.get(); }

    /// Returns the integration rule in x direction
    inline const concepts::QuadratureRule* integrationX()
      const { return intX_.get(); }
    /// Returns the integration rule in y direction
    inline const concepts::QuadratureRule* integrationY()
      const { return intY_.get(); }

    /** Computes the element map. The reference element is the unit quad.
     */
    inline concepts::Real2d chi(const Real x, const Real y) const {
      return cell_.chi(x, y); }

    /// Computes the Jacobian
    inline concepts::MapReal2d jacobian(const Real x, const Real y) const {
      concepts::MapReal2d jac = cell_.jacobian(x, y);
      jac *= 0.5;
      return jac; }

    /// Computes the inverse of the Jacobian
    inline concepts::MapReal2d jacobianInverse(const Real x,
					       const Real y) const {
      return jacobian(x, y).inverse(); }

    /// Computes the determinant of the Jacobian
    inline Real jacobianDeterminant(const Real x, const Real y) const {
      return jacobian(x, y).determinant(); }

    /** Sets the subdivision strategy of the underlying cell of this element.
	It calls Quad2d::setStrategy.

        @param strategy Pointer to an instance of a subdivision strategy.
	@throws StrategyChange if the change is not allowed (the change is
	        not allowed if there are children present)
    */
    void setStrategy(const concepts::Quad2dSubdivision* strategy = 0)
      throw(concepts::StrategyChange) { cell_.setStrategy(strategy); }

    /** Returns the subdivision strategy of the underlying cell of
	this element.
    */
    const concepts::Quad2dSubdivision* getStrategy() const
    { return cell_.getStrategy(); }

    virtual const concepts::ElementGraphics<Real>* graphics() const;

    /// Set polynomial degree of edge \c i to \c p
    void edgeP(const uint i, const concepts::AdaptiveControlP<1>& p) {
      conceptsAssert(i < 4, concepts::Assertion());
      edges_[i] = &p; }
    /// Get polynomial degree of edge \c i
    const concepts::AdaptiveControlP<1>& edgeP(const uint i) const {
      conceptsAssert(i < 4, concepts::Assertion());
      conceptsAssert(edges_[i], concepts::Assertion());
      return *(edges_[i]); }

    /// Types of integration rules to choose from
    enum intRule { GAUSS_LOBATTO = 0, GAUSS_JACOBI = 4 };

    /** Sets the type of integration rule to use.
	@param type Determines the integration rule
	@param constant Use constant number of points
	@param points Number of points to use (for <tt>constant = true</tt>)
	or points to add to approximation order (for <tt>constant = false</tt>)
    */
    static void setIntegrationRule(enum intRule rule, bool constant,
				   uint points);
    /// Returns information on the settings of the quadrature rule
    static std::string integrationRule();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// The cell
    concepts::Quad2d& cell_;
    /// The shape functions
    std::auto_ptr<concepts::Karniadakis<1,0> > shpfctX_, shpfctY_;
    /// The derivatives of the shape functions
    std::auto_ptr<concepts::Karniadakis<1,1> > shpfctDX_, shpfctDY_;
    /// The integration rules
    std::auto_ptr<concepts::QuadratureRule > intX_, intY_;

    /// Polynomial degree of edges
    const concepts::AdaptiveControlP<1>* edges_[4];

    static std::auto_ptr<QuadGraphics> graphics_;

    /** Default behaviour: integration rule Gauss Jacobi (highest order).
	The default value is 4 (Gauss Jacobi). Other possible values are
	0 (Gauss Lobatto) for better graphics (connected elements). Use
	setIntegrationRule() to change it.
    */
    static enum intRule integrationType_;
    /** Number of integration points to use when constant number is requested.
	Default value is 10.
    */
    static uint constNumerOfPoints_;
    /** Number of integration points to add to approximation order when
	varying number of integration points is requested. Default is 2.
    */
    static uint addNumberOfPoints_;
    /** Use constant number of integration points (true) or not (false).
	Default is false (ie. varying number of points).
    */
    static bool useConstantNumberOfPoints_;
  };

} // namespace hp2D

#endif // quad_hh
