/* Graphics for 2D hp FEM
 */

#include <sstream>
#include "quad.hh"
#include "quadGraphics.hh"
#include "basics/exceptions.hh"
#include "graphics/meshEPS.hh"
#include "graphics/dataGnuplot.hh"

// debugging
#include "basics/debug.hh"

#define graphGeneral_D 0
#define graphGnuplot_D 0

namespace hp2D {

  // ********************************************************** QuadGraphics **

  void QuadGraphics::operator()(const concepts::Element<Real>& elm,
				enum graphicsType type,
				concepts::CellPostprocess<Real>& post) const {
    DEBUGL(graphGeneral_D, elm);
    const Quad* e = dynamic_cast<const Quad*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    switch(type) {

    case TOPOLOGY_DOTTY:
    case VERTEX_LIST:
    case MESH_DX:
    case MESH_MATLAB:
    case MESH_EPS:
      post(e->cell()); break;

    case DATA_GNUPLOT: {
      graphics::DataGnuplotCell<Real>* gnuplot =
	dynamic_cast<graphics::DataGnuplotCell<Real>*>(&post);
      conceptsAssert(gnuplot, concepts::Assertion());

      // coefficients
      const concepts::Array<Real>& coeff = gnuplot->coeff();
      DEBUGL(graphGnuplot_D, coeff);
      std::ofstream* ofs = gnuplot->write();
      concepts::Level<2> corner[4];
      corner[0].l_[0] = 0;
      corner[0].l_[1] = 0;
      corner[1].l_[0] = 1;
      corner[1].l_[1] = 0;
      corner[2].l_[0] = 0;
      corner[2].l_[1] = 1;
      corner[3].l_[0] = 1;
      corner[3].l_[1] = 1;

      const uint npx = e->integrationX()->n();
      const uint npy = e->integrationY()->n();
      const Real* px = e->integrationX()->abscissas();
      const Real* py = e->integrationY()->abscissas();

      const uint nsx = e->shpfctX()->n();
      const uint nsy = e->shpfctY()->n();
      const Real* sx = e->shpfctX()->values();
      const Real* sy = e->shpfctY()->values();

      // formula
      const concepts::Formula* frm = gnuplot->formula();

      // formula, elementwise
      const concepts::ElementFormula<Real>* frmE = gnuplot->elementFormula();

      for (uint j = 0; j < npy-1; ++j) {
	for (uint i = 0; i < npx-1; ++i) {
	  for (uint c = 0; c < 4; ++c) {
	    const Real* coeffTmp = (const Real*)coeff;
	    concepts::Real2d x(e->chi((1.0 + px[i+corner[c].l_[0]])/2.0,
				      (1.0 + py[j+corner[c].l_[1]])/2.0));
	    Real f = 0;
	    if (coeff.size() > 0) {
	      for (uint k = 0; k < nsy; ++k)
		for (uint l = 0; l < nsx; ++l)
		  f += sy[k*npy+j+corner[c].l_[1]]
		    *sx[l*npx+i+corner[c].l_[0]]* *coeffTmp++;
	    }
	    if (frm) {
	      f = (*frm)(x);
	    }
	    if (frmE) {
	      f = (*frmE)(*e, concepts::Real2d
			  ((1.0 + px[i+corner[c].l_[0]])/2.0,
			   (1.0 + py[j+corner[c].l_[1]])/2.0));
	    }
	    *ofs << x[0] << " " << x[1] << " " << f << std::endl;

	    if (c == 1) *ofs << std::endl;
	  }
	  *ofs << std::endl;
	  *ofs << std::endl;
	}
      }
      break;
    }

    case MESH_EPS_INIT: {
      graphics::MeshEPSCell<Real>* eps =
	dynamic_cast<graphics::MeshEPSCell<Real>*>(&post);
      conceptsAssert(eps, concepts::Assertion());
      concepts::Connector0* vtx = 0;
      for (uint i = 0; (vtx = e->support().vertex(i)) != 0; ++i) {
	concepts::Real2d x(e->vertex(i));
	eps->offsetMin(x);
	eps->maxMax(x);
      }
      break;
    }

    default: 
      std::stringstream error;
      error << "graphics type " << type << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str())); break;
    } // switch
  }

  std::ostream& QuadGraphics::info(std::ostream& os) const {
    return os << "QuadGraphics()";
  }


} // namespace hp2D
