/* weight functions for weighted regularization
 */

#ifndef shortestDist2D_hh
#define shortestDist2D_hh

#include <cmath>
#include "basics/vectorsMatrices.hh"
#include "hp2D/singularSet.hh"

namespace hp2D {

  // forward declarations
  class Quad;
  using concepts::Real;

  // ********************************************************* TrivialWeight **

  /** A function class for trivial (constant equal 1.0) weight function.
      @author Kersten Schmidt, 2002
  */
  class TrivialWeight {
  public:
    /// Returns 1
    Real operator()(const SingularSet&, concepts::Real2d,
		    const Quad*) const { return 1.0; }
  };

  std::ostream& operator<<(std::ostream& os, const TrivialWeight& p);

  // ********************************************************** ShortestDist **

  /** A function class for weighted regularization, which returns
      the square of the shortest distance of a point to the singular
      vertices.
      @author Kersten Schmidt, 2002
  */
  class ShortestDist {
  public:
    /** Returns the square of the shortest distance of a point to the singular
	vertices.
	@param point Reference coordinates of the point to which the shortest
	       distance has to be computed
	@param elm Element the point is belonging to
	@param singularities Set of singular vertices
    */
    Real operator()(const SingularSet& singularities,
		    concepts::Real2d point, const Quad* elm) const;
  };

  std::ostream& operator<<(std::ostream& os, const ShortestDist& p);

  // ********************************************************** DistancePost **

  /** A function class for weighted regularization. It uses a a distance
      function (see template parameter) and simply exponentiates the result
      it gets from there.
      @param DistClass Function class which is able to evaluate a distance
      @param Function Function class which is able to post process the result
             from DistClass (has a static function <tt>post(Real, Real)</tt>
      @author Philipp Frauenfelder, 2004
  */
  template<typename DistClass, typename Function>
  class DistancePost {
  public:
    /** Uses \c DistClass to compute the weight function and \c Function to
	post process the weight.
    */
    Real operator()(const SingularSet& singularities,
		    concepts::Real2d point, const Quad* elm) const {
      return Function::post(dist_(singularities, point, elm));
    }
  private:
    DistClass dist_;
  };

  template<typename DistClass, typename Function>
  inline std::ostream& operator<<(std::ostream& os,
				  const DistancePost<DistClass, Function>& p) {
    return os << Function() << '(' << DistClass() << ')';
  }


  // ********************************************************** Postprocess4 **

  /** Computes \c x to the power of 0.4.
      @author Philipp Frauenfelder, 2004
  */
  class Postprocess4 {
  public:
    static Real post(Real x) { return std::pow(x, 0.4); }
  };

  std::ostream& operator<<(std::ostream& os, const Postprocess4& p);

  // ********************************************************** Postprocess7 **

  /** Computes \c x to the power of 0.7.
      @author Philipp Frauenfelder, 2004
  */
  class Postprocess7 {
  public:
    static Real post(Real x) { return std::pow(x, 0.7); }
  };

  std::ostream& operator<<(std::ostream& os, const Postprocess7& p);

  // ********************************************************** Postprocess8 **

  /** Computes \c x to the power of 0.8.
      @author Philipp Frauenfelder, 2004
  */
  class Postprocess8 {
  public:
    static Real post(Real x) { return std::pow(x, 0.8); }
  };

  std::ostream& operator<<(std::ostream& os, const Postprocess8& p);

  // ********************************************************** Postprocess9 **

  /** Computes \c x to the power of 0.9.
      @author Philipp Frauenfelder, 2004
  */
  class Postprocess9 {
  public:
    static Real post(Real x) { return std::pow(x, 0.9); }
  };

  std::ostream& operator<<(std::ostream& os, const Postprocess9& p);

  // ******************************************************* PostprocessSqrt **

  /** Computes to square root of \c x
      @author Philipp Frauenfelder, 2004
  */
  class PostprocessSqrt {
  public:
    static Real post(Real x) { return std::sqrt(x); }
  };

  std::ostream& operator<<(std::ostream& os, const PostprocessSqrt& p);

} // namespace hp2D

#endif // shortestDist2D_hh
