/* singular vertices
 */

#ifndef singularSet2D_hh
#define singularSet2D_hh

#include "basics/vectorsMatrices.hh"
#include "toolbox/scannerConnectors.hh"
#include "geometry/topology.hh"
#include "geometry/mesh.hh"
#include "hp2D/quad.hh"

namespace hp2D {

  using concepts::Real;

  // ******************************************************** SingularVertex **

  /** Class for storing a singular vertex with the coordinates. The
      class also provides methods to compute a distance from a given
      point (in local coordinates in an element) to the vertex. The
      returned result is the square of the Euclidiean distance.
      @author Kersten Schmidt, 2002
  */
  class SingularVertex {
    friend std::ostream& operator<<(std::ostream& os, const SingularVertex& s);
  public:
    /** Constructor.
	@param vertex Vertex object
	@param vtx0 Coordinates of the vertex
    */
    SingularVertex(const concepts::Connector0* vertex,
		   const concepts::Real2d vtx0) :
      vertex_(vertex), vtx0_(vtx0) {}
    /// Returns a pointer to the vertex object (topological, connector)
    const concepts::Connector0* vertex() const { return vertex_; }
    /// Returns the square of the distance from the vertex to the \c
    /// point (in physical coordinates)
    Real distance(concepts::Real2d point) const;
    /** Returns the distance from the vertex to a \c point inside an
	Hexahedron (point in reference coordinates)
     */
    Real distance(concepts::Real2d point, const Quad* elm) const {
      return distance(elm->chi(point[0], point[1]));
    }
  private:
    /// Pointer to the vertex object (topological, connector)
    const concepts::Connector0* vertex_;
    /// Coordinates of the vertex
    const concepts::Real2d vtx0_;
  };

  // *********************************************************** SingularSet **

  /** Class for handling a set of singular vertices
      @author Kersten Schmidt, 2002
  */
  class SingularSet {
    friend std::ostream& operator<<(std::ostream& os, const SingularSet& s);
  public:
    SingularSet() : vertices_(0) {}
    ~SingularSet();
    /** Add all vertices and edges in the mesh with \c attrib to the
	set of singular vertices and edges
    */
    void add(const concepts::Attribute& attrib, concepts::Mesh2& msh);
    /// Scanner over all singular vertices
    concepts::Scan<SingularVertex>* verticesScan() const {
      return new concepts::PListScan<SingularVertex>(*vertices_);
    }
  private:
    /// List of singular vertices
    concepts::Joiner<SingularVertex*, 1>* vertices_;
  };

} // namespace hp2D

#endif // singularSet2D_hh
