/* space class for 2D hp FEM
 */

#include <memory>
#include <sstream>
#include <algorithm>
#include "hp2D/space.hh"
#include "hp2D/element.hh"
#include "hp2D/quad.hh"
#include "basics/exceptions.hh"
#include "basics/Zm.hh"
#include "toolbox/boundary.hh"
#include "space/tmatrix.hh"
#include "space/smatrix.hh"

#include "basics/debug.hh"

#define hpSpace2DConstr_D 0
#define hpSpace2DDestr_D 0
#define hpSpace2DAdj_D 0
#define hpSpace2DRebuild_D 0

namespace hp2D {

  // ***************************************************************** Space **

  Space::Space(concepts::Mesh2& msh, uint l, uint p,
	       concepts::BoundaryConditions* bc, bool trunk) :
    msh_(msh), bc_(bc), trunk_(trunk), rebuild_(true), dim_(0), nelm_(0),
    elm_(0), ctrl0_(2, 4), ctrl1_(2, 3), ctrl2_(2, 2), adj_(2, 2),
    S1left_(0), S1right_(0) {
    std::auto_ptr<concepts::Scan2> sc(msh_.scan());
    while (*sc) {
      concepts::Key key(((*sc)++).connector().key());
      ctrl2_[key].tag_.data_ = concepts::AdaptiveControlTag::MEMBER;
      adj_[key].level_ = concepts::Level<2>(l);
      for (uint i = 0; i < 2; ++i)
	adj_[key].p_[i] = p;
    }
    DEBUGL(hpSpace2DConstr_D, "done.");
  }

  Space::~Space() {
    concepts::Joiner<Element<Real>*, 1>::destructor(elm_);
    DEBUGL(hpSpace2DDestr_D, "done.");
  }

  void Space::adjust(const concepts::Element<Real>& elm,
		     const concepts::AdaptiveAdjustP<2>& a) {
    DEBUGL(hpSpace2DAdj_D, "start. elm = " << elm << ", a = " << a);
    const Element<Real>* e = dynamic_cast<const Element<Real>*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    adj_[e->support().key()] = a;
    rebuild_ = true;
    DEBUGL(hpSpace2DAdj_D, "done.");
  }

  void Space::rebuild() {
    DEBUGL(hpSpace2DRebuild_D, "start.");

    // remove old elements
    concepts::Joiner<Element<Real>*, 1>::destructor(elm_);
    ctrl1_.reset();
    elm_ = 0;
    dim_ = 0;
    nelm_ = 0;

    DEBUGL(hpSpace2DRebuild_D, "adj = " << adj_);

    // prepare for rebuild of mesh by marking level and polynomial degree
    // sort out boundary conditions (ie. mark DIRICHLET nodes as passive)
    DEBUGL(hpSpace2DRebuild_D, "scanning mesh with meshAndPoly_");
    std::auto_ptr<concepts::Scan2> sc(msh_.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;
      concepts::Level<2> L(-1);
      int P[2] = { -1, -1 };

      concepts::Quad2d* cellQ = dynamic_cast<concepts::Quad2d*>(&cell);
      if (cellQ) {
	meshAndPoly_(*cellQ, L, P);
	continue;
      }

      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }
    DEBUGL(hpSpace2DRebuild_D, "ctrl0 = " << ctrl0_);

    // create list of cells, edges and so on
    DEBUGL(hpSpace2DRebuild_D, "creating cell list");
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;
      createCellList_(cell.connector(), 0);
    } // loop over mesh
    DEBUGL(hpSpace2DRebuild_D, "registered " << cellList_.size() << " cells");
    DEBUGL(hpSpace2DRebuild_D, "creating vertex and edge list");
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;
      createVtxEdgList_(cell.connector());
    } // loop over mesh
    DEBUGL(hpSpace2DRebuild_D, "registered " << edgeList_.size() << " edges");
    DEBUGL(hpSpace2DRebuild_D, "registered " << vertexList_.size()
	   << " vertices");

    concepts::EdgeData::checkRelatedEdges(edgeList_);
    DEBUGL(hpSpace2DRebuild_D, "updated edgeList_");

    bool notChanged = false;
    while (! notChanged) {
      notChanged = true;
      for (std::map<uint, concepts::VertexData>::iterator i
	     = vertexList_.begin(); i != vertexList_.end(); ++i) {
	notChanged &= !(i->second.checkRelations());
      }
    }
    DEBUGL(hpSpace2DRebuild_D, "updated vertexList_");

    // enforce minimum rule
    for (std::map<uint, concepts::EdgeData>::iterator i =
	   edgeList_.begin(); i != edgeList_.end(); ++i) {
      for (concepts::EdgeData::iterator j = i->second.begin();
	   j != i->second.end(); ++j) {
	const concepts::Connector2* cell =
	  dynamic_cast<const concepts::Connector2*>(&(j->second->cell()));
	conceptsAssert(cell, concepts::Assertion());
	minimumRule_(*cell, i->second.edge());
      } // for cells in edge
    } // for edges

    // enrich elements
    DEBUGL(hpSpace2DRebuild_D, "enriching elements with enrichElm_");
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;
      int P[6];
      for (uint i = 0; i < 6; ++i) P[i] = -1;

      concepts::Quad2d* cellQ = dynamic_cast<concepts::Quad2d*>(&cell);
      if (cellQ) {
	enrichElm_(*cellQ, P);
	continue;
      }

      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }

    // rebuild the elements
    DEBUGL(hpSpace2DRebuild_D, "rebuilding elements with rebuild1_");
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;

      concepts::Quad2d* cellQ = dynamic_cast<concepts::Quad2d*>(&cell);
      if (cellQ) {
	ushort p[2] = { 0, 0 };
	buildElements_(*cellQ, p);
	continue;
      }

      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }

    // clear pre rebuild information
    DEBUGL(hpSpace2DRebuild_D, "clear pre rebuild information");
    rebuild_ = false;
    adj_.reset();
    // clear the control information of vertices and edges
    ctrl0_.reset();
    cellList_.clear();
    edgeList_.clear();
    vertexList_.clear();
    DEBUGL(hpSpace2DRebuild_D, "done.");
  }

  std::ostream& Space::info(std::ostream& os) const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return os << "Space(dim = " << dim_ << ", nelm = " << nelm_ << ')';
  }

  void Space::deactivate_(concepts::Edge& edg) {
    const concepts::Connector1* edge = 0;
    for (uint i = 0; (edge = edg.child(i)) != 0; ++i)
      ctrl1_[edge->key()].tag_.deactivate();
    if ((edge = edg.child(1)) != 0)
      ctrl0_[edge->vertex(0)->key()].tag_.deactivate();
  }

  void Space::deactivate_(concepts::Vertex&) {}

} // namespace hp2d
