/* space class for 2D hp FEM
 */

#ifndef hpSpace2D_hh
#define hpSpace2D_hh

#include <memory>
#include <cstring>
#include "basics/exceptions.hh"
#include "basics/typedefs.hh"
#include "space/space.hh"
#include "space/hpMethod.hh"
#include "geometry/continuityData.hh"
#include "geometry/cell2D.hh"
#include "geometry/mesh.hh"
#include "toolbox/scannerConnectors.hh"
#include "toolbox/dynArray.hh"
#include "geometry/boundaryConditions.hh"
#include "element.hh"

namespace concepts {

  // forward declarations
  template<class F>
  class SMatrixBase;

  class SMatrix1D;

  // ****************************************************************** Scan **

  template<>
  class Scan<hp2D::Element<Real> > :
    public concepts::Scan<concepts::Element<Real> > {
  public:  
    hp2D::Element<Real>& operator++(int) = 0;
  };

} // namespace concepts

namespace hp2D {
  using concepts::Real;

  // forward declarations
  class Quad;

  // ***************************************************************** Space **

  /** A 2D hp FEM space with continuous, piecewise polynomial basis functions.
      Currently, only hexahedrons and conforming meshes are possible.

      @author Philipp Frauenfelder, 2001
  */
  class Space : public concepts::AdaptiveSpace<Real,
		concepts::AdaptiveAdjustP<2> > {
  public:
    typedef concepts::Scan<hp2D::Element<Real> > Scan;

    /** Constructor. Scans the mesh and sets the cells in the mesh active
	and the level of refinement and the polynomial degree in
	all cells to the given values. rebuild_ is set to true, ie. 
	if the mesh is used it will firstly be rebuilt.

	@param msh The domain of interest partitioned into a mesh.
	@param l Level of refinement
	@param p Degree of the polynomials to be used.
	@param bc Boundary conditions
	@param trunk Create trunk space ie., use the the rules
	       <tt>p+q <= pIntMax + 1</tt> in the interior instead of
	       allowing all <tt>p, q</tt> combinations.
    */
    Space(concepts::Mesh2& msh, uint l, uint p,
	  concepts::BoundaryConditions* bc = 0, bool trunk = true);
    virtual ~Space();
    
    inline virtual uint dim() const;
    inline virtual uint nelm() const;
    inline virtual Scan* scan() const;

    inline uint dim();
    inline uint nelm();
    inline Scan* scan();

    virtual void adjust(const concepts::Element<Real>& elm,
			const concepts::AdaptiveAdjustP<2>& a);

    /** Rebuilds the space. First the old list of Finite Elements in the space
	is removed (not the elements in the topology!).
	Then the control information for the vertices, edges and faces is
	cleared (the information on the 2D elements is retained).
	For every cell in the original mesh, rebuild0_ and 
	rebuild1_ are called. From the mesh, they build the space and
	store the elements in elm_. An essential part of the generated
	information are the T matrices. They describe how the local shape
	functions are glued together to form the global shape functions.

	rebuild0_ marks the elements with the right values and 
	rebuild1_ really does the work.

	@see TMatrix for more information on T matrices
	     and how they are stored.
    */
    void rebuild();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Mesh
    concepts::Mesh2& msh_;
    /// Boundary conditions
    concepts::BoundaryConditions* bc_;
    /// Build trunk space
    bool trunk_;

    /** If true: the elements have to be rebuilt. Some elements have been
	marked for refinement or for coarsening and the elements are
	outdated.
    */
    bool rebuild_;

    /// Dimension of the FE space
    uint dim_;

    /// Number of elements currently active in the mesh
    uint nelm_;

    /// Linked list of the elements
    concepts::Joiner<Element<Real>*, 1>* elm_;

    /// Hash table of the control information for the vertices
    concepts::DynArray<concepts::AdaptiveControl> ctrl0_;
    /// Hash table of the control information for the edges
    concepts::DynArray<concepts::AdaptiveControlP<1> > ctrl1_;
    /// Hash table of the control information for the 2D elements
    concepts::DynArray<concepts::AdaptiveControlP<2> > ctrl2_;

    /// List of cells
    std::map<uint, concepts::CellData> cellList_;
    /// List of edges
    std::map<uint, concepts::EdgeData> edgeList_;
    /// List of vertices
    std::map<uint, concepts::VertexData> vertexList_;

    /// Hash table of the adjustment information of the elements
    concepts::DynArray<concepts::AdaptiveAdjustP<2> > adj_;

    /// S matrices in 1D
    std::auto_ptr<concepts::SMatrix1D> S1left_, S1right_;
    /** S matrices for horizontal subdivision
	@see concepts::Quad2dSubdiv2H
    */
    std::auto_ptr<concepts::SMatrixBase<Real> > Smatrices2H_[2];
    /** S matrices for vertical subdivision
	@see concepts::Quad2dSubdiv2V
    */
    std::auto_ptr<concepts::SMatrixBase<Real> > Smatrices2V_[2];
    /** S matrices for subdivision into 4 quads
	@see concepts::Quad2dSubdiv4
    */
    std::auto_ptr<concepts::SMatrixBase<Real> > Smatrices4_[4];

    /** Adjusts the mesh and the polynomial degree.

        On the old mesh, the adjustment information is taken. This leads
	to mesh refinements or coarsening and a certain polynomial degree
	distribution.

	The polynomial degree is fixed on the finest level (where the
	elements are eventually built) and given to the
	coarser levels (for edges and the interior of the cell).
	This is done <b>bottom up</b>.

	Finally, the boundary conditions are enforced on every cell using
	\c enforceBC_

        @param cell Cell
	@param L The desired level
	@param P The desired polynomial degree
	@pre \c P is an array of 3 integers
	@see enforceBC_
     */
    void meshAndPoly_(concepts::Quad2d& cell, concepts::Level<2>& L, int* P);

    /** Enforce Dirichlet boundary conditions.
	This routine is called by \c meshAndPoly_ on every cell.
	@param cell Cell
    */
    void enforceBC_(concepts::Quad2d& cell);

    /// Initially fills \c cellList_ (recusively)
    void createCellList_(const concepts::Connector& cntr,
			 const concepts::CellData* father);
    /// Initially fills \c edgeListVtx_ and \c vertexList_
    void createVtxEdgList_(const concepts::Connector2& cell);

    /** Enriches the polynomial degrees. This enrichement is necessary for the
	edges to be able to create conforming global basis functions.

	In refined cells, some of the small edges might have
	a polynomial degree which is too low to represent a part
	of a basis function which is defined on a larger element.

	To achieve this, the cells are traversed <b>top down</b> until
	the level of the elements is reached. During the traversal,
	the needed polynomial degree for the small elements is
	computed and enforced on the element level (only on the edges).

	@pre \c p is an array of 6 integers
    */
    void enrichElm_(const concepts::Quad2d& cell, int* p);

    /** Used to enforce the minimum rule. The minimum rule states that the
	polynomial degree on a edge has to be the minimum of the adjacent
	elements.

	This routine recursively calls itself and asserts that all the elements
	adjacent to \c edge enforce their polynomial degree on <tt>edge</tt>.
	This is done if the \c cell is member of the space. Otherwise,
	\c minimumRule_ is called with the children of <tt>cell</tt>.
    */
    void minimumRule_(const concepts::Connector2& cntr,
		      const concepts::Connector1& edge);

    /** Creates the elements and their T matrices.

        This routine is called recursively until the finest level is
	reached. If there have degrees of freedom to be assigned on
	higher level, the according T columns are adapted using the S
	matrices. On the finest level, the elements are created and
	the T matrix is created from the T columns.

	@param cell The current cell
	@param Pmax Request for maximal polynomial degree, array with 2 entries
	@param T0 Precomputed T columns which belong to a parent of the cell
	@pre \c Pmax is an array of 2 integers
     */
    void buildElements_(concepts::Quad2d& cell, ushort* Pmax,
			concepts::TColumn<Real>* T0 = 0);

    /** Computes the maximal polynomial degree of \c cell in each direction.
	Takes into account higher polynomial degrees on faces and edges
	compared to the interior.
    */
    void computePmax_(const concepts::Quad2d& cell,
		      ushort Pmax[2]) const;

    /** Checks if the S matrices need to be recomputed and does so if
	necessary.

	@param elm Element which the S matrices are needed for (the
	           polynomial degrees are important)
    */
    void recomputeSmatrices_(const Quad& elm);

    void deactivate_(concepts::Edge& edg);
    void deactivate_(concepts::Vertex& vtx);
  };

  uint Space::dim() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return dim_; 
  }

  uint Space::nelm() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return nelm_; 
  }

  Space::Scan* Space::scan() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return new concepts::PListScan<Element<Real> >(*elm_);
  }

  uint Space::dim() {
    if (rebuild_) rebuild();
    return dim_; 
  }

  uint Space::nelm() {
    if (rebuild_) rebuild();
    return nelm_; 
  }

  Space::Scan* Space::scan() {
    if (rebuild_) rebuild();
    return new concepts::PListScan<Element<Real> >(*elm_);
  }

} // namespace hp2D

#endif // hpSpace2D_h
