/* Functor class for sum factorization
 */

#include "sumfact.hh"
#include "shapeFunction2d.hh"
#include "space/element.hh"
#include "toolbox/coeffIterator.hh"
#include "basics/debug.hh"

#define SumFactAppl_D 0

namespace hp2D {

  // ****************************************************** SumFactorization **

  template<typename T>
  void SumFactorization::operator()(concepts::ElementMatrix<Real>& em,
				    const concepts::Array<Real>& jacobian,
				    T* coeff,
				    const ShapeFunction2D<Real>& u,
				    const ShapeFunction2D<Real>& v) {
    // quadrature points
    const uint nq[2] = { u.nq(0), u.nq(1) };
    const Real* const w[2] = { u.weights(0), u.weights(1) };

    // assert that u and v use the same quadrature points
    conceptsAssert(nq[0] == v.nq(0), concepts::Assertion());
    conceptsAssert(nq[1] == v.nq(1), concepts::Assertion());
    conceptsAssert(w[0] == v.weights(0), concepts::Assertion());
    conceptsAssert(w[1] == v.weights(1), concepts::Assertion());

    // shapefunctions for u
    const uint uns[2] = { u.n(0), u.n(1) };
    DEBUGL(SumFactAppl_D, "uuns = " << uns[0] << ", " << uns[1]);
    const Real* const us[2] = { u.values(0), u.values(1) };
    DEBUGL(SumFactAppl_D, "u = " << us[0][0] << ", " << us[0][1] << "; "
	   << us[1][0] << ", " << us[1][1]);

    // shapefunctions for v
    const uint vns[2] = { v.n(0), v.n(1) };
    const Real* const vs[2] = { v.values(0), v.values(1) };
    DEBUGL(SumFactAppl_D, "v = " << vs[0][0] << ", " << vs[0][1] << "; "
	   << vs[1][0] << ", " << vs[1][1]);

    // get needed indices information
    concepts::TColumnTensor<bool, 2> uneededIdx(uns, 0);
    u.T()->usedIdx(uneededIdx);
    DEBUGL(SumFactAppl_D, "needed indices for u: " << uneededIdx);
    concepts::TColumnTensor<bool, 2> vneededIdx(vns, 0);
    v.T()->usedIdx(vneededIdx);
    DEBUGL(SumFactAppl_D, "needed indices for v: " << vneededIdx);


    // swap order
    int a = 1, b = 0; // original
    // add a method to set a, b and c to get the fastest method possible

    // prepare matrices for sum factorization
    sumfact_.resize(nq[b]*uns[a]*vns[a]);
    sumfact_.zeros();
    DEBUGL(SumFactAppl_D, "reserved " << sumfact_.size()
	   << " for sumfact_");

    // fill sumfact_
    Real* sumfact = (Real*)sumfact_;
    if (coeff != 0) {
      for (uint iy = 0; iy < uns[a]; ++iy) {
	for (uint jy = 0; jy < vns[a]; ++jy) {
	  const Real* wxAcc = w[b];
	  const Real* dx = (const Real*)jacobian;
	  T coeffTmp = *coeff;
	  T c = *coeff;
	  for (uint I = 0; I < nq[b]; ++I) {
	    const Real* wyAcc = w[a];
	    const Real* siy = us[a] + iy*nq[a];
	    const Real* sjy = vs[a] + jy*nq[a];
	    for (uint K = 0; K < nq[a]; ++K) {
	      DEBUGL(SumFactAppl_D, "siy = " << *siy << ", sjy = " << *sjy <<
		     ", coeff(" << I*nq[a]+K << ") = " << *c++ << ", dx = " << *dx <<
		     ", wxAcc = " << *wxAcc << ", wyAcc = " << *wyAcc);
	      *sumfact += *siy++ * *sjy++ * *coeffTmp++ *
		*dx++ * *wxAcc * *wyAcc++;
	    } // K
	    ++sumfact;
	    ++wxAcc;
	  } // I
	} // jy
      } // iy
    } else {
      for (uint iy = 0; iy < uns[a]; ++iy) {
	for (uint jy = 0; jy < vns[a]; ++jy) {
	  const Real* wxAcc = w[b];
	  const Real* dx = (const Real*)jacobian;
	  for (uint I = 0; I < nq[b]; ++I) {
	    const Real* wyAcc = w[a];
	    const Real* siy = us[a] + iy*nq[a];
	    const Real* sjy = vs[a] + jy*nq[a];
	    for (uint K = 0; K < nq[a]; ++K) {
	      *sumfact += *siy++ * *sjy++ *
		*dx++ * *wxAcc * *wyAcc++;
	    } // K
	    ++wxAcc;
	    ++sumfact;
	  } // I
	} // jy
      } // iy
    }

    // fill em
    Real* intermed = (Real*)em;
    for (uint iy = 0; iy < uns[1]; ++iy) {
      for (uint ix = 0; ix < uns[0]; ++ix) {
	uint i2[2] = { ix, iy };
	if (uneededIdx[i2]) {
	  for (uint jy = 0; jy < vns[1]; ++jy) {
	    for (uint jx = 0; jx < vns[0]; ++jx) {
	      uint j2[2] = { jx, jy};
	      if (vneededIdx[j2]) {
// 		if (sameMatrix &&
// 		    neededIdxOldElm[i2] && neededIdxOldElm[j2]) {
// 		  // no computation needed, copy of old value used
// 		} else {
// 		  if (!((jx < ix) && (jy < iy))) {
// 		    // exploiting symmetry of matrix
// 		    *intermed = 0.0;
		sumfact = (Real*)sumfact_
		  + i2[a]*vns[a]*nq[b]
		  + j2[a]*nq[b];
// 		Real* intermed = (Real*)em + j[0] + vns[0]*(j[1] + vns[1]);
		const Real* six = us[b] + i2[b]*nq[b];
		const Real* sjx = vs[b] + j2[b]*nq[b];
		for (uint I = 0; I < nq[b]; ++I) {
		  DEBUGL(SumFactAppl_D, "six = " << *six << ", sjx = " << *sjx <<
			 ", R(" << i2[a]*vns[a]*nq[b] + j2[a]*nq[b] << ") = " << *sumfact);
		  *intermed += *six++ * *sjx++ * *sumfact++;
		  DEBUGL(SumFactAppl_D, "em = " << em);
		} // I
// 		      } // symmetric
// 		    } // computation needed
	      } // if neededId[j2]
	      ++intermed;
	    } // jx
	  } // jy
	}
	else // if neededIdx[i2]
	  intermed += vns[a]*vns[b];
      } // ix
    } // iy
  }

  template
  void SumFactorization::operator()(concepts::ElementMatrix<Real>& em,
				    const concepts::Array<Real>& jacobian,
				    concepts::CoeffIterator<Real, Real>* coeff,
				    const ShapeFunction2D<Real>& u,
				    const ShapeFunction2D<Real>& v);

  template
  void SumFactorization::operator()(concepts::ElementMatrix<Real>& em,
				    const concepts::Array<Real>& jacobian,
				    concepts::CoeffIterator<Real,
				    concepts::MapReal2d>* coeff,
				    const ShapeFunction2D<Real>& u,
				    const ShapeFunction2D<Real>& v);

} // namespace hp2D
