/* testsuite for 2D stiffness matrix for needle elements
 */

#include <iostream>
#include "geometry/mesh.hh"
#include "geometry/elementMaps.hh"
#include "geometry/cell2D.hh"
#include "hp2D/space.hh"
#include "hp2D/bilinearForm.hh"
#include "hp2D/aprioriRef2D.hh"
#include "operator/denseMatrix.hh"
#include "matrices.hh"

using namespace concepts;

// **************************************************************** TestMesh **

class TestMesh : public Mesh2 {
private:
  class S : public Scan<Cell2> {
    unsigned int idx_;
    Cell2 *(&cell_)[1];
  public:
    inline S(Cell2 *(&cell)[1]) : idx_(0), cell_(cell) {}
    inline S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    inline bool eos() const { return idx_ == 1; }
    inline Cell2& operator++(int) { return *cell_[idx_++]; }
    inline Scan2* clone() const { return new S(*this); }
  };
  Vertex *vtx_[4];
  Edge *edg_[4];
  Quad *quad_[1];
  Cell2 *cell_[1];
public:
  TestMesh(const Real x = 1.0, const Real y = 1.0);
  virtual ~TestMesh();
  inline unsigned int ncell() const { return 1; }
  inline Scan2* scan() { return new S(cell_); }
  virtual std::ostream& info(std::ostream& os) const;
};

TestMesh::TestMesh(const Real x, const Real y) {
  for (int i = 0; i < 4; ++i)
    vtx_[i] = new Vertex;

  edg_[0] = new Edge(*vtx_[0], *vtx_[1]);
  edg_[1] = new Edge(*vtx_[1], *vtx_[2]);
  edg_[2] = new Edge(*vtx_[3], *vtx_[2]);
  edg_[3] = new Edge(*vtx_[0], *vtx_[3]);

  quad_[0] = new Quad(*edg_[0], *edg_[1], *edg_[2], *edg_[3]);
  VertexQuad2d m(Real2d(0.0, 0.0), Real2d(x, 0.0), 
		 Real2d(x, y), Real2d(0.0, y));
  cell_[0] = new Quad2d(*quad_[0], m);
}

TestMesh::~TestMesh() {
  delete cell_[0];
  delete quad_[0];
  for (int i = 0; i < 4; ++i)
    delete edg_[i];
  for (int i = 0; i < 4; ++i)
    delete vtx_[i];
}

std::ostream& TestMesh::info(std::ostream& os) const {
  os << "TestMesh(" << *cell_[0] << ')';
  return os;
}

namespace test {

  // ******************************************************** TestMatrices2D **

  TestMatrices2D::~TestMatrices2D() {}

  void TestMatrices2D::run() {
    testMassMatrix0101();
    testMassMatrix010_01();
    testStiffnessMatrix0101();
    testStiffnessMatrix010_01();
  }

  void TestMatrices2D::testMassMatrix0101() {
    const Real m[20][20] =
      { {1.0/9, 1.0/18, 1.0/36, 1.0/18, 1.0/36, -1.0/90,
	 0, 1.0/72, -1.0/180, 1.0/72, -1.0/180, 0, 1.0/36,
	 -1.0/90, 1.0/144, -1.0/360, -1.0/360, 1.0/900, 0, 0},
	{1.0/18, 1.0/9, 1.0/18, 1.0/36, 1.0/36, 1.0/90, 0,
	 1.0/36, -1.0/90, 1.0/72, 1.0/180, 0, 1.0/72, -1.0/180,
	 1.0/144, -1.0/360, 1.0/360, -1.0/900, 0, 0},
	{1.0/36, 1.0/18, 1.0/9, 1.0/18, 1.0/72, 1.0/180, 0,
	 1.0/36, 1.0/90, 1.0/36, 1.0/90, 0, 1.0/72, 1.0/180,
	 1.0/144, 1.0/360, 1.0/360, 1.0/900, 0, 0},
	{1.0/18, 1.0/36, 1.0/18, 1.0/9, 1.0/72, -1.0/180, 0,
	 1.0/72, 1.0/180, 1.0/36, -1.0/90, 0, 1.0/36, 1.0/90,
	 1.0/144, 1.0/360, -1.0/360, -1.0/900, 0, 0},
	{1.0/36, 1.0/36, 1.0/72, 1.0/72, 1.0/90, 0, -1.0/420,
	 1.0/144, -1.0/360, 1.0/180, 0, -1.0/840, 1.0/144,
	 -1.0/360, 1.0/360, -1.0/900, 0, 0, -1.0/1680,
	 1.0/4200},
	{-1.0/90, 1.0/90, 1.0/180, -1.0/180, 0, 2.0/315, 0,
	 1.0/360, -1.0/900, 0, 1.0/315, 0, -1.0/360, 1.0/900,
	 0, 0, 1.0/630, -1.0/1575, 0, 0},
	{0, 0, 0, 0, -1.0/420, 0, 1.0/210, 0, 0, -1.0/840, 0,
	 1.0/420, 0, 0, -1.0/1680, 1.0/4200, 0, 0, 1.0/840,
	 -1.0/2100},
	{1.0/72, 1.0/36, 1.0/36, 1.0/72, 1.0/144, 1.0/360, 0,
	 1.0/90, 0, 1.0/144, 1.0/360, 0, 1.0/180, 0, 1.0/360,
	 0, 1.0/900, 0, 0, 0},
	{-1.0/180, -1.0/90, 1.0/90, 1.0/180, -1.0/360, -1.0/900,
	 0, 0, 2.0/315, 1.0/360, 1.0/900, 0, 0, 1.0/315,0,
	 1.0/630, 0, 1.0/1575, 0, 0},
	{1.0/72, 1.0/72, 1.0/36, 1.0/36, 1.0/180, 0, -1.0/840,
	 1.0/144, 1.0/360, 1.0/90,0, -1.0/420, 1.0/144,
	 1.0/360, 1.0/360, 1.0/900,0, 0, -1.0/1680, -1.0/4200},
	{-1.0/180, 1.0/180, 1.0/90, -1.0/90,0, 1.0/315, 0,
	 1.0/360, 1.0/900, 0, 2.0/315,0, -1.0/360, -1.0/900,
	 0,0, 1.0/630, 1.0/1575, 0, 0},
	{0, 0, 0, 0, -1.0/840, 0, 1.0/420, 0, 0, -1.0/420, 0,
	 1.0/210, 0, 0, -1.0/1680, -1.0/4200, 0,0, 1.0/840,
	 1.0/2100},
	{1.0/36, 1.0/72, 1.0/72, 1.0/36, 1.0/144,-1.0/360, 0,
	 1.0/180, 0, 1.0/144, -1.0/360,0, 1.0/90, 0, 1.0/360,
	 0, -1.0/900, 0, 0,0},
	{-1.0/90, -1.0/180, 1.0/180, 1.0/90,-1.0/360, 1.0/900,
	 0,0, 1.0/315, 1.0/360, -1.0/900, 0, 0, 2.0/315, 0,
	 1.0/630, 0, -1.0/1575, 0, 0},
	{1.0/144, 1.0/144, 1.0/144, 1.0/144, 1.0/360, 0,
	 -1.0/1680, 1.0/360, 0, 1.0/360,0, -1.0/1680, 1.0/360,
	 0, 1.0/900, 0, 0,0, -1.0/4200, 0},
	{-1.0/360, -1.0/360, 1.0/360, 1.0/360,-1.0/900, 0,
	 1.0/4200, 0, 1.0/630, 1.0/900,0, -1.0/4200, 0, 1.0/630,
	 0, 1.0/1575, 0, 0,0, -1.0/7350},
	{-1.0/360, 1.0/360, 1.0/360, -1.0/360,0, 1.0/630, 0,
	 1.0/900, 0, 0, 1.0/630, 0, -1.0/900, 0, 0, 0,
	 1.0/1575, 0, 0, 0},
	{1.0/900, -1.0/900, 1.0/900, -1.0/900,0, -1.0/1575, 0,
	 0, 1.0/1575, 0, 1.0/1575, 0, 0, -1.0/1575, 0, 0, 0,
	 4.0/11025, 0, 0},
	{0, 0, 0, 0, -1.0/1680, 0, 1.0/840, 0,0, -1.0/1680, 0,
	 1.0/840, 0, 0, -1.0/4200,0, 0, 0, 1.0/2100, 0},
	{0, 0, 0, 0, 1.0/4200, 0, -1.0/2100, 0,0, -1.0/4200, 0,
	 1.0/2100, 0, 0, 0, -1.0/7350, 0, 0, 0, 1.0/3675} };

    TestMesh mesh;
    _test(mesh.ncell() == 1);
    hp2D::Space space(mesh, 0, 1, 0, false);
    space.rebuild();
    _test(space.nelm() == 1);
    _test(space.dim() == 4);

    const int pMax[] = { 3, 2 };
    hp2D::APrioriRefinement refine(space, 1, 1, pMax);
    GlobalPostprocess<Real> post(space);
    post(refine);

    space.rebuild();
    _test(space.nelm() == 1);
    _test(space.dim() == 20);

    hp2D::Identity mbf;
    DenseMatrix<Real> mass(space, mbf);
    for (uint i = 0; i < space.dim(); ++i) {
      for (uint j = 0; j < space.dim(); ++j) {
	_numtest(mass(i,j), m[i][j]);
      }
    }
  }

  void TestMatrices2D::testMassMatrix010_01() {
    const Real m[20][20] =
      { {1.0/900, 1.0/1800, 1.0/3600, 1.0/1800, 1.0/3600, -1.0/9000, 0,
	 1.0/7200, -1.0/18000, 1.0/7200, -1.0/18000, 0, 1.0/3600, -1.0/9000,
	 1.0/14400, -1.0/36000, -1.0/36000, 1.0/90000, 0, 0},
	{1.0/1800, 1.0/900, 1.0/1800, 1.0/3600, 1.0/3600, 1.0/9000, 0,
	 1.0/3600, -1.0/9000, 1.0/7200, 1.0/18000, 0, 1.0/7200, -1.0/18000,
	 1.0/14400, -1.0/36000, 1.0/36000, -1.0/90000, 0, 0},
	{1.0/3600, 1.0/1800, 1.0/900, 1.0/1800, 1.0/7200, 1.0/18000, 0,
	 1.0/3600, 1.0/9000, 1.0/3600, 1.0/9000, 0, 1.0/7200, 1.0/18000,
	 1.0/14400, 1.0/36000, 1.0/36000, 1.0/90000, 0, 0},
	{1.0/1800, 1.0/3600, 1.0/1800, 1.0/900, 1.0/7200, -1.0/18000, 0,
	 1.0/7200, 1.0/18000, 1.0/3600, -1.0/9000, 0, 1.0/3600, 1.0/9000,
	 1.0/14400, 1.0/36000, -1.0/36000, -1.0/90000, 0, 0},
	{1.0/3600, 1.0/3600, 1.0/7200, 1.0/7200, 1.0/9000, 0, -1.0/42000,
	 1.0/14400, -1.0/36000, 1.0/18000, 0, -1.0/84000, 1.0/14400,
	 -1.0/36000, 1.0/36000, -1.0/90000, 0, 0, -1.0/168000, 1.0/420000},
	{-1.0/9000, 1.0/9000, 1.0/18000, -1.0/18000, 0, 1.0/15750, 0,
	 1.0/36000, -1.0/90000, 0, 1.0/31500, 0, -1.0/36000, 1.0/90000, 0, 0,
	 1.0/63000, -1.0/157500, 0, 0},
	{0, 0, 0, 0, -1.0/42000, 0, 1.0/21000, 0, 
	 0, -1.0/84000, 0, 1.0/42000, 0, 
	 0, -1.0/168000, 1.0/420000, 0, 0, 1.0/84000, -1.0/210000},
	{1.0/7200, 1.0/3600, 1.0/3600, 1.0/7200, 1.0/14400, 
	 1.0/36000, 0, 1.0/9000, 0, 1.0/14400, 1.0/36000, 
	 0, 1.0/18000, 0, 1.0/36000, 0, 1.0/90000, 0, 0, 0},
	{-1.0/18000, -1.0/9000, 1.0/9000, 1.0/18000, -1.0/36000, -1.0/90000, 0,
	 0, 1.0/15750, 1.0/36000, 1.0/90000, 0, 0, 1.0/31500, 0, 1.0/63000, 0,
	 1.0/157500, 0, 0},
	{1.0/7200, 1.0/7200, 1.0/3600, 1.0/3600, 1.0/18000, 0, -1.0/84000,
	 1.0/14400, 1.0/36000, 1.0/9000, 0, -1.0/42000, 1.0/14400, 1.0/36000,
	 1.0/36000, 1.0/90000, 0, 0, -1.0/168000, -1.0/420000},
	{-1.0/18000, 1.0/18000, 1.0/9000, -1.0/9000, 
	 0, 1.0/31500, 0, 1.0/36000, 1.0/90000, 0, 1.0/15750,
         0, -1.0/36000, -1.0/90000, 0, 0, 1.0/63000, 1.0/157500, 0, 0},
	{0, 0, 0, 0, -1.0/84000, 0, 1.0/42000, 0, 0, -1.0/42000, 0, 1.0/21000,
	 0, 0, -1.0/168000, -1.0/420000, 0, 0, 1.0/84000, 1.0/210000},
	{1.0/3600, 1.0/7200, 1.0/7200, 1.0/3600, 1.0/14400, 
	 -1.0/36000, 0, 1.0/18000, 0, 1.0/14400, -1.0/36000, 
	 0, 1.0/9000, 0, 1.0/36000, 0, -1.0/90000, 0, 0, 0},
	{-1.0/9000, -1.0/18000, 1.0/18000, 1.0/9000, -1.0/36000, 1.0/90000, 0, 
	 0, 1.0/31500, 1.0/36000, -1.0/90000, 0, 0, 1.0/15750, 0, 1.0/63000, 0,
	 -1.0/157500, 0, 0},
	{1.0/14400, 1.0/14400, 1.0/14400, 1.0/14400, 1.0/36000, 
	 0, -1.0/168000, 1.0/36000, 0, 1.0/36000, 
	 0, -1.0/168000, 1.0/36000, 0, 1.0/90000, 0, 0, 0, -1.0/420000, 0},
	{-1.0/36000, -1.0/36000, 1.0/36000, 1.0/36000, 
	 -1.0/90000, 0, 1.0/420000, 0, 1.0/63000, 1.0/90000, 
	 0, -1.0/420000, 0, 1.0/63000, 0, 1.0/157500, 0, 0, 0, -1.0/735000},
	{-1.0/36000, 1.0/36000, 1.0/36000, -1.0/36000, 
	 0, 1.0/63000, 0, 1.0/90000, 0, 0, 1.0/63000, 
	 0, -1.0/90000, 0, 0, 0, 1.0/157500, 0, 0, 0},
	{1.0/90000, -1.0/90000, 1.0/90000, -1.0/90000, 0, -1.0/157500, 0, 0,
	 1.0/157500, 0, 1.0/157500, 0, 0, -1.0/157500, 0, 0, 0, 1.0/275625, 0,
	 0},
	{0, 0, 0, 0, -1.0/168000, 0, 1.0/84000, 0, 0, -1.0/168000, 0,
	 1.0/84000, 0, 0, -1.0/420000, 0, 0, 0, 1.0/210000, 0},
	{0, 0, 0, 0, 1.0/420000, 0, -1.0/210000, 0, 0, -1.0/420000,
	 0, 1.0/210000, 0, 0, 0, -1.0/735000, 0, 0, 0, 1.0/367500} };

    TestMesh mesh(1.0, 0.01);
    _test(mesh.ncell() == 1);
    hp2D::Space space(mesh, 0, 1, 0, false);
    space.rebuild();
    _test(space.nelm() == 1);
    _test(space.dim() == 4);

    const int pMax[] = { 3, 2 };
    hp2D::APrioriRefinement refine(space, 1, 1, pMax);
    GlobalPostprocess<Real> post(space);
    post(refine);

    space.rebuild();
    _test(space.nelm() == 1);
    _test(space.dim() == 20);

    hp2D::Identity mbf;
    DenseMatrix<Real> mass(space, mbf);
    for (uint i = 0; i < space.dim(); ++i) {
      for (uint j = 0; j < space.dim(); ++j) {
	_numtest(mass(i,j), m[i][j]);
      }
    }
  }

  void TestMatrices2D::testStiffnessMatrix0101() {
    const Real s[20][20] =
      { {2.0/3, -1.0/6, -1.0/3, -1.0/6, 1.0/12, -1.0/30, 0, -1.0/12, 1.0/30,
	 -1.0/12, 1.0/30, 0, 1.0/12, -1.0/30, 0, 0, 0, 0, 0, 0},
	{-1.0/6, 2.0/3, -1.0/6, -1.0/3, 1.0/12, 1.0/30, 0, 1.0/12, -1.0/30,
	 -1.0/12, -1.0/30, 0, -1.0/12, 1.0/30, 0, 0, 0, 0, 0, 0},
	{-1.0/3, -1.0/6, 2.0/3, -1.0/6, -1.0/12, -1.0/30, 0, 1.0/12, 1.0/30,
	 1.0/12, 1.0/30, 0, -1.0/12, -1.0/30, 0, 0, 0, 0, 0, 0},
	{-1.0/6, -1.0/3, -1.0/6, 2.0/3, -1.0/12, 1.0/30, 0, -1.0/12, -1.0/30,
	 1.0/12, -1.0/30, 0, 1.0/12, 1.0/30, 0, 0, 0, 0, 0, 0},
	{1.0/12, 1.0/12, -1.0/12, -1.0/12, 13.0/90, 0, -1.0/140, 0, 0, 1.0/45,
	 0, 1.0/140, 0, 0, 1.0/36, -1.0/90, 0, 0, 0, 0},
	{-1.0/30, 1.0/30, -1.0/30, 1.0/30, 0, 2.0/7, 
	 0, 0, 0, 0, 4.0/35, 0, 0, 0, 0, 0, 1.0/15, -2.0/75, 0, 0},
	{0, 0, 0, 0, -1.0/140, 0, 31.0/70, 0, 
	 0, 1.0/140, 0, 1.0/5, 0, 0, 0, 0, 0, 0, 3.0/28, -3.0/70},
	{-1.0/12, 1.0/12, 1.0/12, -1.0/12, 0, 0, 
	 0, 13.0/90, 0, 0, 0, 0, 1.0/45, 0, 1.0/36, 0, 1.0/90, 0, 0, 0},
	{1.0/30, -1.0/30, 1.0/30, -1.0/30, 0, 0, 0, 
	 0, 2.0/7, 0, 0, 0, 0, 4.0/35, 0, 1.0/15, 0, 2.0/75, 0, 0},
	{-1.0/12, -1.0/12, 1.0/12, 1.0/12, 1.0/45, 0, 1.0/140, 0, 0,
	 13.0/90, 0, -1.0/140, 0, 0, 1.0/36, 1.0/90, 0, 0, 0, 0},
	{1.0/30, -1.0/30, 1.0/30, -1.0/30, 0, 4.0/35, 
	 0, 0, 0, 0, 2.0/7, 0, 0, 0, 0, 0, 1.0/15, 2.0/75, 0, 0},
	{0, 0, 0, 0, 1.0/140, 0, 1.0/5, 0, 
	 0, -1.0/140, 0, 31.0/70, 0, 0, 0, 0, 0, 0, 3.0/28, 3.0/70},
	{1.0/12, -1.0/12, -1.0/12, 1.0/12, 0, 0, 
	 0, 1.0/45, 0, 0, 0, 0, 13.0/90, 0, 1.0/36, 0, -1.0/90, 0, 0, 0},
	{-1.0/30, 1.0/30, -1.0/30, 1.0/30, 0, 0, 0, 
	 0, 4.0/35, 0, 0, 0, 0, 2.0/7, 0, 1.0/15, 0, -2.0/75, 0, 0},
	{0, 0, 0, 0, 1.0/36, 0, 0, 1.0/36, 0, 1.0/36, 
	 0, 0, 1.0/36, 0, 1.0/45, 0, 0, 0, -1.0/420, 0},
	{0, 0, 0, 0, -1.0/90, 0, 0, 
	 0, 1.0/15, 1.0/90, 0, 0, 0, 1.0/15, 0, 52.0/1575, 0, 0, 0, -1.0/175},
	{0, 0, 0, 0, 0, 1.0/15, 0, 1.0/90, 0, 
	 0, 1.0/15, 0, -1.0/90, 0, 0, 0, 52.0/1575, 0, 0, 0},
	{0, 0, 0, 0, 0, -2.0/75, 0, 0, 2.0/75, 
	 0, 2.0/75, 0, 0, -2.0/75, 0, 0, 0, 16.0/525, 0, 0},
	{0, 0, 0, 0, 0, 0, 3.0/28, 0, 0, 0, 0, 3.0/28,
	 0, 0, -1.0/420, 0, 0, 0, 1.0/21, 0},
	{0, 0, 0, 0, 0, 0, -3.0/70, 0, 0, 0, 
	 0, 3.0/70, 0, 0, 0, -1.0/175, 0, 0, 0, 44.0/1225} };

    TestMesh mesh;
    _test(mesh.ncell() == 1);
    hp2D::Space space(mesh, 0, 1, 0, false);
    space.rebuild();
    _test(space.nelm() == 1);
    _test(space.dim() == 4);

    const int pMax[] = { 3, 2 };
    hp2D::APrioriRefinement refine(space, 1, 1, pMax);
    GlobalPostprocess<Real> post(space);
    post(refine);

    space.rebuild();
    _test(space.nelm() == 1);
    _test(space.dim() == 20);

    hp2D::Laplace sbf;
    DenseMatrix<Real> stiff(space, sbf);
    for (uint i = 0; i < space.dim(); ++i) {
      for (uint j = 0; j < space.dim(); ++j) {
 	_numtest(stiff(i,j), s[i][j]);
      }
    }
  }

  void TestMatrices2D::testStiffnessMatrix010_01() {
    const Real s[20][20] =
      { {10001.0/300, 4999.0/300, -10001.0/600, -19999.0/600, 25.0/3, -10.0/3,
	 0, -1.0/1200, 1.0/3000, -25.0/3, 10.0/3, 0, 1.0/1200, -1.0/3000, 0, 0,
	 0, 0, 0, 0},
	{4999.0/300, 10001.0/300, -19999.0/600, -10001.0/600, 25.0/3, 10.0/3,
	 0, 1.0/1200, -1.0/3000, -25.0/3, -10.0/3, 0, -1.0/1200, 1.0/3000, 0,
	 0, 0, 0, 0, 0},
	{-10001.0/600, -19999.0/600, 10001.0/300, 4999.0/300, -25.0/3,
	 -10.0/3, 0, 1.0/1200, 1.0/3000, 25.0/3, 10.0/3, 0, -1.0/1200,
	 -1.0/3000, 0, 0, 0, 0, 0, 0},
	{-19999.0/600, -10001.0/600, 4999.0/300, 
	 10001.0/300, -25.0/3, 10.0/3, 0, -1.0/1200, -1.0/3000, 25.0/3,
	 -10.0/3, 0, 1.0/1200, 1.0/3000, 0, 0, 0, 0, 0, 0},
	{25.0/3, 25.0/3, -25.0/3, -25.0/3, 3001.0/900, 0, -5.0/7, 0, 0,
	 -5999.0/1800, 0, 5.0/7, 0, 0, 1.0/3600, -1.0/9000, 0, 0, 0, 0},
	{-10.0/3, 10.0/3, -10.0/3, 10.0/3, 0, 1669.0/875, 0, 0, 0, 0,
	 -3331.0/1750, 0, 0, 0, 0, 0, 1.0/1500, -1.0/3750, 0, 0},
	{0, 0, 0, 0, -5.0/7, 0, 1003.0/700, 0, 0, 5.0/7, 0, -1997.0/1400, 0,
	 0, 0, 0, 0, 0, 3.0/2800, -3.0/7000},
	{-1.0/1200, 1.0/1200, 1.0/1200, -1.0/1200, 0, 0, 0, 100003.0/9000, 0,
	 0, 0, 0, 49997.0/9000, 0, 25.0/9, 0, 10.0/9, 0, 0, 0},
	{1.0/3000, -1.0/3000, 1.0/3000, -1.0/3000, 0, 0, 0, 0, 46667.0/1750,
	 0, 0, 0, 0, 23333.0/1750, 0, 20.0/3, 0, 8.0/3, 0, 0},
	{-25.0/3, -25.0/3, 25.0/3, 25.0/3, -5999.0/1800, 0, 5.0/7, 0, 0,
	 3001.0/900, 0, -5.0/7, 0, 0, 1.0/3600, 1.0/9000, 0, 0, 0, 0},
	{10.0/3, -10.0/3, 10.0/3, -10.0/3, 0, -3331.0/1750, 0, 0, 0, 0,
	 1669.0/875, 0, 0, 0, 0, 0, 1.0/1500, 1.0/3750, 0, 0},
	{0, 0, 0, 0, 5.0/7, 0, -1997.0/1400, 0, 0, -5.0/7, 0, 1003.0/700,
	 0, 0, 0, 0, 0, 0, 3.0/2800, 3.0/7000},
	{1.0/1200, -1.0/1200, -1.0/1200, 1.0/1200, 0, 0, 0, 49997.0/9000,
	 0, 0, 0, 0, 100003.0/9000, 0, 25.0/9, 0, -10.0/9, 0, 0, 0},
	{-1.0/3000, 1.0/3000, -1.0/3000, 1.0/3000, 0, 0, 0, 0, 23333.0/1750,
	 0, 0, 0, 0, 46667.0/1750, 0, 20.0/3, 0, -8.0/3, 0, 0},
	{0, 0, 0, 0, 1.0/3600, 0, 0, 25.0/9, 0, 1.0/3600,
         0, 0, 25.0/9, 0, 10001.0/9000, 0, 0, 0, -5.0/21, 0},
	{0, 0, 0, 0, -1.0/9000, 0, 0, 0, 20.0/3, 1.0/9000, 0, 0, 0,
	 20.0/3, 0, 42001.0/15750, 0, 0, 0, -4.0/7},
	{0, 0, 0, 0, 0, 1.0/1500, 0, 10.0/9, 0, 
	 0, 1.0/1500, 0, -10.0/9, 0, 0, 0, 50021.0/78750, 0, 0, 0},
	{0, 0, 0, 0, 0, -1.0/3750, 0, 0, 8.0/3, 
	 0, 1.0/3750, 0, 0, -8.0/3, 0, 0, 0, 20002.0/13125, 0, 0},
	{0, 0, 0, 0, 0, 0, 3.0/2800, 0, 0, 0, 
	 0, 3.0/2800, 0, 0, -5.0/21, 0, 0, 0, 10009.0/21000, 0},
	{0, 0, 0, 0, 0, 0, -3.0/7000, 0, 0, 0, 
	 0, 3.0/7000, 0, 0, 0, -4.0/7, 0, 0, 0, 14003.0/12250} };

    TestMesh mesh(1.0, 0.01);
    _test(mesh.ncell() == 1);
    hp2D::Space space(mesh, 0, 1, 0, false);
    space.rebuild();
    _test(space.nelm() == 1);
    _test(space.dim() == 4);

    const int pMax[] = { 3, 2 };
    hp2D::APrioriRefinement refine(space, 1, 1, pMax);
    GlobalPostprocess<Real> post(space);
    post(refine);

    space.rebuild();
    _test(space.nelm() == 1);
    _test(space.dim() == 20);

    hp2D::Laplace sbf;
    DenseMatrix<Real> stiff(space, sbf);
    for (uint i = 0; i < space.dim(); ++i) {
      for (uint j = 0; j < space.dim(); ++j) {
 	_numtest(stiff(i,j), s[i][j]);
      }
    }
  }

} // namespace test

int main(int, char**) {
  test::TestMatrices2D t;
  t.run();
  return t.report();
}
