/* a priori refinement for hexahedra
 */

#ifndef APRIORIRef3D_hh
#define APRIORIRef3D_hh

#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "geometry/connector.hh"
#include "space/postProcess.hh"
#include "space/hpMethod.hh"

// forward declaration
namespace concepts {
  class Hexahedron;
}

namespace hp3D {

  using concepts::Real;

  // ***************************************************** APrioriRefinement **

  /** Carries out a-priori given refinements. The mesh is geometrically
      refined towards the given vertices, edges or faces. These can be
      determined by their attribute. If a refinement in one direction occurs,
      p is not increased in this direction. Otherwise it is increased by the
      given amount. This makes it possible to generate geometrically refined
      meshes to corners and edges with linear polynomial degree distribution.

      Currently, methods for hp3D::Hexahedron are implemented.
      @author Philipp Frauenfelder, 2003
      @bug The parameter \c p of the constructor can be set to a pointer to
      0. This is not detected and a run-time error occurs.
  */
  class APrioriRefinement : public concepts::CellPostprocess<Real> {
  public:
    /** Constructor. If a given attribute for the vertices or edges is 0,
	it is not considered.
	@param spc Space to be refined
	@param attribVtx Refine towards vertices with this attribute
	@param attribEdge Refine towards edges with this attribute
	@param attribFace Refine towards faces with this attribute
	@param p If there is no refinement,
	p will be increased by this amount (anisotropically)
    */
    APrioriRefinement
    (concepts::AdaptiveSpace<Real, concepts::AdaptiveAdjustP<3> >& spc,
     concepts::Attribute attribVtx, concepts::Attribute attribEdge,
     concepts::Attribute attribFace, const int* const p) :
      spc_(spc), attribVtx_(attribVtx), attribEdge_(attribEdge),
      attribFace_(attribFace), p_(p) {}
    virtual void operator() (const concepts::Element<Real>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature) {}
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Space to be refined
    concepts::AdaptiveSpace<Real, concepts::AdaptiveAdjustP<3> >& spc_;
    /// Attribute of entity to which should be refined
    concepts::Attribute attribVtx_, attribEdge_, attribFace_;
    /// Maximal p in a refinement step in each direction
    const int* const p_;
    /// Computes the refinement solely using topological information
    concepts::AdaptiveAdjustP<3> computeRef_
    (const concepts::Hexahedron& h) const;
  };


} // namespace hp3D

#endif // APRIORIRef3D_hh

