/* discretization of div * div for Maxwell equations in frequency domain
 */

#include "divdiv.hh"
#include "hexahedron.hh"
#include "shapeFunction3d.hh"
#include "sumfact.hh"
#include "toolbox/coeffIterator.hh"
#include "space/element.hh"
#include "vectorial/vectorial.hh"

// debugging
#include "basics/debug.hh"

#define hpDivDivAppl_D 0
#define hpDivDivApplTime_D 0
#define hpDivDivApplMem_D 0
#define hpDivDivApplCoeff_D 0

namespace hp3D {

  // **************************************************************** DivDiv **

  template<typename Weight>
  concepts::Array<Real> DivDiv<Weight>::jacobianDiv_;
  template<typename Weight>
  concepts::Array<concepts::MapReal3d> DivDiv<Weight>::jacobianInvDiv_;
  template<typename Weight>
  const Hexahedron* DivDiv<Weight>::elmDiv_(0);
  template<typename Weight>
  concepts::Array<std::auto_ptr<DivDiv<Weight> > > DivDiv<Weight>::divdiv_(0);

  template<class Weight>
  DivDiv<Weight>::DivDiv(concepts::Array<Real>& jacobian,
			 concepts::Array<concepts::MapReal3d>& jacobianInv,
			 const Hexahedron** oldElm, SingularSet& singularities,
			 const uint i, const uint j,
			 const concepts::Formula* frm) :
    i_(i), j_(j), jacobian_(jacobian), jacobianInv_(jacobianInv),
    oldElm_(oldElm), singularities_(singularities), weight_(), frm_(0) {
    if (frm)
      frm_.reset(frm->clone());
  }

  template<class Weight>
  DivDiv<Weight>::~DivDiv() {}

  template<class Weight>
  void DivDiv<Weight>::operator()(const concepts::Element<Real>& elmX, 
				  const concepts::Element<Real>& elmY, 
				  concepts::ElementMatrix<Real>& em) {

    const Hexahedron* elmYH = dynamic_cast<const Hexahedron*>(&elmY);
    if (elmYH) {
      const Hexahedron* elmXH = dynamic_cast<const Hexahedron*>(&elmX);
      if (elmXH) { (*this)(*elmXH, *elmYH, em); return; }
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  template<class Weight>
  void DivDiv<Weight>::operator()(const Hexahedron& elmX,
				  const Hexahedron& elmY, 
				  concepts::ElementMatrix<Real>& em) {
    uint n = elmX.T().m();
    em.resize(n, elmY.T().m());
    em.zeros();

    if (elmX.support() == elmY.support()) {
      DEBUGL(hpDivDivAppl_D, elmX);
      DEBUGL(hpDivDivApplMem_D, "em is (" << n << 'x' << elmY.T().m() << ')');

      em.zeros();

      // quadrature points
      const uint nqx = elmX.integrationX()->n();
      const uint nqy = elmX.integrationY()->n();
      const uint nqz = elmX.integrationZ()->n();
      const Real* const px = elmX.integrationX()->abscissas();
      const Real* const py = elmX.integrationY()->abscissas();
      const Real* const pz = elmX.integrationZ()->abscissas();

#if hpDivDivApplTime_D
      concepts::ResourceMonitor timer;
#endif
      DEBUGL(hpDivDivAppl_D, "elmX = " << &elmX << ", oldElm = " << *oldElm_
	     << ", &oldElm = " << oldElm_<< ", i = " << i_ << ", j = " << j_);
      if (&elmX != *oldElm_) {
	// Jacobian
	jacobian_.resize(nqx*nqy*nqz);
	DEBUGL(hpDivDivApplMem_D, "reserved " << (nqx*nqy*nqz)
	       << " for jacobian_");
	Real* dx = (Real*)jacobian_;
	jacobianInv_.resize(nqx*nqy*nqz);
	DEBUGL(hpDivDivApplMem_D, "reserved " << (nqx*nqy*nqz)
	       << " for jacobianInv_");
	concepts::MapReal3d* jacInv = (concepts::MapReal3d*)jacobianInv_;
	const Real* pxAcc = px;
	for (uint i = 0; i < nqx; ++i, ++pxAcc) {
	  const Real* pyAcc = py;
	  for (uint j = 0; j < nqy; ++j, ++pyAcc) {
	    const Real* pzAcc = pz;
	    for (uint k = 0; k < nqz; ++k) {
	      const concepts::Real3d refPoint( (*pxAcc + 1.0)/2.0,
					       (*pyAcc + 1.0)/2.0,
					       (*pzAcc++ + 1.0)/2.0 );
	      concepts::MapReal3d jacobian
		(elmX.jacobian(refPoint[0], refPoint[1], refPoint[2]));
	      Real ws = weight_(singularities_, refPoint, &elmX);
	      DEBUGL(hpDivDivApplMem_D, "i = " << i << ", j = " << j
		     << ", k = " << k << ", weight = " << ws);
	      Real f = 1.0;
	      if (frm_.get())
		f = (*frm_)(elmX.cell().chi(refPoint[0],
					    refPoint[1], refPoint[2]));
	      *dx++ = jacobian.determinant() * ws * f * f;

	      // inverse of the jacobian with three columns (a1, a2, a3)
	      concepts::MapReal3d jacinv(jacobian.inverse());
	      // store it in
	      *jacInv++ = jacinv;
	    }
	  }
	}
#if hpDivDivApplTime_D
	DEBUGL(hpDivDivApplTime_D, "TIME for Jacobian: " << timer);
#endif
      } else {
	DEBUGL(hpDivDivApplMem_D, "took already computed jacobian");
      }
      DEBUGL(hpDivDivAppl_D, "Jacobian: " << jacobian_);
      concepts::Array<concepts::MapReal3d> coeffMatrix_(nqx*nqy*nqz);
      concepts::MapReal3d* jacInv = (concepts::MapReal3d*)jacobianInv_;
      concepts::MapReal3d* coeffMatrix = (concepts::MapReal3d*)coeffMatrix_;

      for (uint i = 0; i < nqx; ++i) {
	for (uint j = 0; j < nqy; ++j) {
	  for (uint k = 0; k < nqz; ++k) {
	    concepts::MapReal3d jacinv(*jacInv++);

	    DEBUGL(hpDivDivApplCoeff_D,"jacinv = " << jacinv);

	    // calculate a_i*a_j'
	    concepts::Real3d firstColumn(jacinv.column(i_));
	    concepts::Real3d secondColumn(jacinv.column(j_));
	    coeffMatrix->zeros();

	    DEBUGL(hpDivDivApplCoeff_D,"firstColumn = " << firstColumn
		   << ", secondColumn = " << secondColumn);

	    coeffMatrix->rank1Product(firstColumn, secondColumn);
	    DEBUGL(hpDivDivApplCoeff_D,"jacinv = " << jacinv);
	    DEBUGL(hpDivDivApplCoeff_D,"coeffMatrix = " << *coeffMatrix);
	    ++coeffMatrix;
	  }
	}
      }


      const Real* const wx = elmX.integrationX()->weights();
      const Real* const wy = elmX.integrationY()->weights();
      const Real* const wz = elmX.integrationZ()->weights();

      // shapefunctions
      const uint nsx = elmX.shpfctX()->n();
      const uint nsy = elmX.shpfctY()->n();
      const uint nsz = elmX.shpfctZ()->n();
      const Real* const sx = elmX.shpfctX()->values();
      const Real* const sy = elmX.shpfctY()->values();
      const Real* const sz = elmX.shpfctZ()->values();
      const Real* const sdx = elmX.shpfctDX()->values();
      const Real* const sdy = elmX.shpfctDY()->values();
      const Real* const sdz = elmX.shpfctDZ()->values();

      ShapeFunction3D<Real> u;
      u.n(0, nsx);        u.n(1, nsy);        u.n(2, nsz);
      u.nq(0, nqx);       u.nq(1, nqy);       u.nq(2, nqz);
      u.values(0, sx);    u.values(1, sy);    u.values(2, sz);
      u.abscissas(0, px); u.abscissas(1, py); u.abscissas(2, pz);
      u.weights(0, wx);   u.weights(1, wy);   u.weights(2, wz);
      u.T(&(elmX.T()));

      ShapeFunction3D<Real> v;
      v.n(0, nsx);        v.n(1, nsy);        v.n(2, nsz);
      v.nq(0, nqx);       v.nq(1, nqy);       v.nq(2, nqz);
      v.values(0, sx);    v.values(1, sy);    v.values(2, sz);
      v.abscissas(0, px); v.abscissas(1, py); v.abscissas(2, pz);
      v.weights(0, wx);   v.weights(1, wy);   v.weights(2, wz);
      v.T(&(elmY.T()));

      SumFactorization sumfact;
      em.zeros();
      for (uint i = 0; i < 3; ++i) {
	switch (i) {
	case 0: u.values(i, sdx); break;
	case 1: u.values(i, sdy); break;
	case 2: u.values(i, sdz); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
	for (uint j = 0; j < 3; ++j) {
	  switch (j) {
	  case 0: v.values(j, sdx); break;
	  case 1: v.values(j, sdy); break;
	  case 2: v.values(j, sdz); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	  concepts::CoeffIterator<Real, concepts::MapReal3d>
	    coeff((concepts::MapReal3d*)coeffMatrix_, i, j);
	  sumfact(em, jacobian_, &coeff, u, v);
	  switch (j) {
	  case 0: v.values(j, sx); break;
	  case 1: v.values(j, sy); break;
	  case 2: v.values(j, sz); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	}
	switch (i) {
	case 0: u.values(i, sx); break;
	case 1: u.values(i, sy); break;
	case 2: u.values(i, sz); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
      }
#if hpDivDivApplTime_D
      DEBUGL(hpDivDivApplTime_D, "TIME for integration: " << timer);
#endif

      DEBUGL(hpDivDivAppl_D, "em = " << em);
#if hpDivDivApplTime_D
      DEBUGL(hpDivDivApplTime_D, "TIME for symmetric em: " << timer);
#endif
      *oldElm_ = &elmX;

      DEBUGL(hpDivDivAppl_D, "elmX = " << &elmX << ", oldElm = " << *oldElm_
	     << ", &oldElm = " << oldElm_);
    } // elmX == elmY
  }

  template<class Weight>
  std::ostream& DivDiv<Weight>::info(std::ostream& os) const{
    return os << "DivDiv(weight = " << weight_
// 	      << ", i = " << i_ << ", j = " << j_ 
// 	      << ", jacobian = " << jacobian_ << ", jacobianInv = "
// 	      << jacobianInv_ << ", oldElm = " << oldElm_
	      << ")";
  }

  template<typename Weight>
  void DivDiv<Weight>::setup(vectorial::BilinearForm<Real>& bf,
			     hp3D::SingularSet& singularities,
			     const concepts::Formula* frm) {
    divdiv_.resize(9);
    conceptsAssert(bf.vdim() == 3, concepts::Assertion());
    for (uint k = 0; k < 9; ++k){
      uint i = (uint)k/3, j = k%3;

      divdiv_[k].reset
	(new DivDiv<Weight>(jacobianDiv_, jacobianInvDiv_, &elmDiv_,
			    singularities, i, j, frm));
      bf.put(*divdiv_[k], i, j);
    }
  }

  /// Weight 1.0
  template class DivDiv<TrivialWeight>;
  /// Weight of squared Distance
  template class DivDiv<ShortestDist>;
  /// Weight of Distance
  template class DivDiv<DistancePost<ShortestDist, PostprocessRoot> >;
  /// Weight of Distance to power of 4
  template class DivDiv<DistancePost<ShortestDist, PostprocessSquare> >;
  /// Weight of root of Distance
  template class DivDiv<DistancePost<ShortestDist, PostprocessRoot4> >;

  template class DivDiv<DistancePost<ShortestDist, Postprocess34> >;
  template class DivDiv<DistancePost<ShortestDist, Postprocess14> >;
  template class DivDiv<DistancePost<ShortestDist, Postprocess4> >;
  template class DivDiv<DistancePost<ShortestDist, Postprocess3> >;

//   template class DivDiv<ShortestDistLimited>
  template class DivDiv<ProductOfAll>;
  template class DivDiv<DaugeWeight>;

} // namespace hp3D
