/* abstract base class for hp elements in 3D
 */

#ifndef hpElement3d_h
#define hpElement3d_h 

#include <cstring>
#include "basics/vectorsMatrices.hh"
#include "geometry/connector.hh"
#include "geometry/cell.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"

namespace hp3D {

  using concepts::Real;

  // *************************************************************** Element **

  /** Abstract class for a 3D FEM element. All FEM elements have to be
      derived from this class. It prescribes the interface for querying
      the support of the element and it has a polynomial degree (ie. hp).

      @author Philipp Frauenfelder, 2001
  */
  template<class F>
  class Element : public concepts::Element<F> {
  public:
    /** Constructor. This constructor initializes an isotropic polynomial
	degree.
	@param T List of T columns for this element
	@param p Polynomial degree of this element
    */
    inline Element(concepts::TColumn<Real>* T, uint p) : T_(T) {
      p_[0] = p_[1] = p_[2] = p;
    }

    /** Constructor. This constructor can initialize an anisotropic
	polynomial degree.
	@param T List of T columns for this element
	@param p Polynomial degree in the three spatial directions
    */
    inline Element(concepts::TColumn<Real>* T, const ushort* p) : T_(T) {
      std::memcpy(p_, p, 3*sizeof(ushort));
    }

    /** Returns the polynomial degree. The returned array has 3 elements.
     */
    inline const ushort* p() const { return p_; }

    /** Returns the topolgical support of the element. 
	Possible supports for an element are hexahedrons, tetrahedrons,
	prisms and pyramids.
    */
    virtual const concepts::Connector3& support() const = 0;

    virtual const concepts::TMatrixBase<Real>& T() const { return T_; }

    /// Appends the T columns to the T matrix
    void appendT(concepts::TColumn<Real>* T) { T_.append(T); }

    /** Returns the coordinates of the ith vertex of this element.

	@param i Index of the vertex
    */
    virtual concepts::Real3d vertex(uint i) const = 0;

    /** Returns the cell on which the element is built.
	Possible are tetrahedrons, hexahedron, prims and pyramids.
    */
    virtual const concepts::Cell3& cell() const = 0;

    /// Comparison operator for elements
    virtual bool operator<(const Element<F>& elm) const = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const;

    /// T matrix of the element
    concepts::TMatrix<Real> T_;
  private:
    /// Polynomial degree
    ushort p_[3];
  };

} // namespace hp3d

#endif // hpElement3d_h
