/* hexahedral element for 3D hp FEM
 */

#include <iostream>
#include "hexahedron.hh"
#include "basics/debug.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/typedefs.hh"

#define CONST_POINTS 0

#if CONST_POINTS
#define NUM_POINTS 10
#else
#define NUM_POINTS_ADD 2
#endif

// debugging
#define HexahedronConstr_D 0
#define HexahedronConstrT_D 0
#define HexahedronDestr_D 0
#define HexahedronComp_D 0
#define HexahedronCompEqual_D 0

namespace hp3D {

  // ************************************************************ Hexahedron **

  std::auto_ptr<HexahedronGraphics> Hexahedron::graphics_(0);

  Hexahedron::Hexahedron(concepts::Hexahedron3d& cell, const ushort* p,
			 concepts::TColumn<Real>* T0,
			 concepts::TColumn<Real>* T1) :
    Element<Real>(T1, p), cell_(cell),
    shpfctX_(0), shpfctY_(0), shpfctZ_(0), intX_(0), intY_(0), intZ_(0) {
    for (uint i = 0; i < 12; ++i) edges_[i] = 0;
    for (uint i = 0; i < 6; ++i) faces_[i] = 0;
    if (T0)
      T_.append(T0);

    DEBUGL(HexahedronConstrT_D, "T = " << T_);

    DEBUGL(HexahedronConstr_D, "computing shape functions for x");
    concepts::Quadrature<INTEGRATION_RULE>* tmp_Int =
      new concepts::Quadrature<INTEGRATION_RULE>(
#if CONST_POINTS
						 NUM_POINTS
#else
						 p[0]+NUM_POINTS_ADD
#endif
						 );
    intX_.reset(tmp_Int);
    // shape functions
    concepts::Karniadakis<1,0>* tmp_Karn
      = new concepts::Karniadakis<1,0>(p[0], intX_->abscissas(), intX_->n());
    shpfctX_.reset(tmp_Karn);
    // derivatives of shape functions
    concepts::Karniadakis<1,1>* tmp_KarnD
      = new concepts::Karniadakis<1,1>(p[0], intX_->abscissas(), intX_->n());
    shpfctDX_.reset(tmp_KarnD);
    DEBUGL(HexahedronConstr_D, "computed shape functions for x:" << *shpfctX_);

    DEBUGL(HexahedronConstr_D, "computing shape functions for y");
    tmp_Int = new concepts::Quadrature<INTEGRATION_RULE>(
#if CONST_POINTS
							 NUM_POINTS
#else
							 p[1]+NUM_POINTS_ADD
#endif
							 );
    intY_.reset(tmp_Int);
    // shape functions
    tmp_Karn = new concepts::Karniadakis<1,0>(p[1], intY_->abscissas(),
					      intY_->n());
    shpfctY_.reset(tmp_Karn);
    // derivatives of shape functions
    tmp_KarnD = new concepts::Karniadakis<1,1>(p[1], intY_->abscissas(),
					       intY_->n());
    shpfctDY_.reset(tmp_KarnD);

    DEBUGL(HexahedronConstr_D, "computing shape functions for z");
    tmp_Int = new concepts::Quadrature<INTEGRATION_RULE>(
#if CONST_POINTS
							 NUM_POINTS
#else
							 p[2]+NUM_POINTS_ADD
#endif
							 );
    intZ_.reset(tmp_Int);
    // shape functions
    tmp_Karn = new concepts::Karniadakis<1,0>(p[2], intZ_->abscissas(),
					      intZ_->n());
    shpfctZ_.reset(tmp_Karn);
    // derivatives of shape functions
    tmp_KarnD = new concepts::Karniadakis<1,1>(p[2], intZ_->abscissas(),
					       intZ_->n());
    shpfctDZ_.reset(tmp_KarnD);
  }

  Hexahedron::~Hexahedron() {
    DEBUGL(HexahedronDestr_D, "done.");
  }

  const concepts::ElementGraphics<Real>* Hexahedron::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new HexahedronGraphics());
    return graphics_.get();
  }

  bool Hexahedron::operator<(const Element<Real>& elm) const {
    const Hexahedron* elmH = dynamic_cast<const Hexahedron*>(&elm);
    if (elmH) {
      if (support() == elmH->support()) return false;
      if (hasSameMatrix(*elmH)) return false;

      // compare polynomial degrees
      const ushort* P = p();
      int p_size = (P[0]+1)*(P[1]+1)*(P[2]+1);
      P = elmH->p();
      if (p_size < (P[0]+1)*(P[1]+1)*(P[2]+1)) {
	DEBUGL(HexahedronComp_D, this->support() << " < " << elm.support());
	return true;
      }
    }
    conceptsAssert(elmH, concepts::Assertion());
    return false;
  }

  bool Hexahedron::hasSameMatrix(const Hexahedron& elm) const {
    // compare polynomial degrees
    const ushort* P = p();
    const ushort* P2 = elm.p();
    for (uint i = 0; i < 3; ++i)
      if (P[i] != P2[i]) return false;

    // compare element size
    concepts::Real3d e[12];
    e[0] = vertex(1)-vertex(0);
    e[1] = vertex(2)-vertex(1);
    e[2] = vertex(3)-vertex(2);
    e[3] = vertex(0)-vertex(3);
    e[4] = vertex(4)-vertex(0);
    e[5] = vertex(5)-vertex(1);
    e[6] = vertex(6)-vertex(2);
    e[7] = vertex(7)-vertex(3);
    e[8] = vertex(5)-vertex(4);
    e[9] = vertex(6)-vertex(5);
    e[10] = vertex(7)-vertex(6);
    e[11] = vertex(4)-vertex(7);

    concepts::Real3d e2[12];
    e2[0] = elm.vertex(1)-elm.vertex(0);
    e2[1] = elm.vertex(2)-elm.vertex(1);
    e2[2] = elm.vertex(3)-elm.vertex(2);
    e2[3] = elm.vertex(0)-elm.vertex(3);
    e2[4] = elm.vertex(4)-elm.vertex(0);
    e2[5] = elm.vertex(5)-elm.vertex(1);
    e2[6] = elm.vertex(6)-elm.vertex(2);
    e2[7] = elm.vertex(7)-elm.vertex(3);
    e2[8] = elm.vertex(5)-elm.vertex(4);
    e2[9] = elm.vertex(6)-elm.vertex(5);
    e2[10] = elm.vertex(7)-elm.vertex(6);
    e2[11] = elm.vertex(4)-elm.vertex(7);

    for (uint i = 0; i < 12; ++i) {
      DEBUGL(HexahedronCompEqual_D, i << ": "
	     << e[i].l2() << " != " << e2[i].l2() << "?");
      if (e[i].l2() != e2[i].l2()) return false;
    }

    // check some angles
    DEBUGL(HexahedronCompEqual_D, e[0]*e[1] << " != " << e2[0]*e2[1]);
    if (e[0]*e[1] != e2[0]*e2[1]) return false;
    DEBUGL(HexahedronCompEqual_D, e[0]*e[4] << " != " << e2[0]*e2[4]);
    if (e[0]*e[4] != e2[0]*e2[4]) return false;
    DEBUGL(HexahedronCompEqual_D, e[3]*e[7] << " != " << e2[3]*e2[7]);
    if (e[3]*e[7] != e2[3]*e2[7]) return false;
    DEBUGL(HexahedronCompEqual_D, e[8]*e[9] << " != " << e2[8]*e2[9]);
    if (e[8]*e[9] != e2[8]*e2[9]) return false;

    DEBUGL(HexahedronCompEqual_D, *this << " == " << elm);
    DEBUGL(HexahedronComp_D, this->support() << " == " << elm.support());

    return true;
  }

  std::ostream& Hexahedron::info(std::ostream& os) const {
    const ushort* P = p();
    return os << "Hexahedron(" << cell_ << ", p = ("
	      << P[0] << ", " << P[1] << ", " << P[2] << "), " << T_ << ')';
  }

} // namespace hp3D
