/* hexahedral element for 3D hp FEM
 */

#ifndef hexahedron_hh
#define hexahedron_hh

#include <iostream>
#include <memory>
#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "geometry/cell3D.hh"
#include "integration/quadrature.hh"
#include "integration/karniadakis.hh"
#include "space/tmatrix.hh"
#include "space/hpMethod.hh"
#include "element.hh"
#include "hexahedronGraphics.hh"

// 0: Gauss Lobatto (good for graphics)
// 4: Gauss Jacobi (highest order)
#ifdef INTEGRATION_RULE
#undef INTEGRATION_RULE
#endif
#define INTEGRATION_RULE 4

namespace hp3D {

  using concepts::Real;

  // ************************************************************ Hexahedron **

  /** A 3D FEM element: a hexahedron.

      @author Philipp Frauenfelder, 2000
  */
  class Hexahedron : public Element<Real> {
  public:
    /** Constructor
	@param cell Cell on which the element is defined
	@param p Polynomial degree (can be anisotropic), this array must
	         have at least 3 entries.
	@param T0 Part of the T matrix
	@param T1 Part of the T matrix
    */
    Hexahedron(concepts::Hexahedron3d& cell, const ushort* p, 
	       concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1);
    virtual ~Hexahedron();

    virtual const concepts::Hexahedron& support() const {
      return cell_.connector(); }
    virtual concepts::Real3d vertex(uint i) const { return cell_.vertex(i); }
    virtual const concepts::Hexahedron3d& cell() const { return cell_; }

    /// Returns the shape functions in x direction
    inline const concepts::Karniadakis<1,0>* shpfctX() const {
      return shpfctX_.get(); }
    /// Returns the shape functions in y direction
    inline const concepts::Karniadakis<1,0>* shpfctY() const {
      return shpfctY_.get(); }
    /// Returns the shape functions in z direction
    inline const concepts::Karniadakis<1,0>* shpfctZ() const {
      return shpfctZ_.get(); }

    /// Returns the derivatives of the shape functions in x direction
    inline const concepts::Karniadakis<1,1>* shpfctDX() const {
      return shpfctDX_.get(); }
    /// Returns the shape functions in y direction
    inline const concepts::Karniadakis<1,1>* shpfctDY() const {
      return shpfctDY_.get(); }
    /// Returns the shape functions in z direction
    inline const concepts::Karniadakis<1,1>* shpfctDZ() const {
      return shpfctDZ_.get(); }

    /// Returns the integration rule in x direction
    inline const concepts::Quadrature<INTEGRATION_RULE>* integrationX() const {
      return intX_.get(); }
    /// Returns the integration rule in y direction
    inline const concepts::Quadrature<INTEGRATION_RULE>* integrationY() const {
      return intY_.get(); }
    /// Returns the integration rule in z direction
    inline const concepts::Quadrature<INTEGRATION_RULE>* integrationZ() const {
      return intZ_.get(); }

    /** Computes the element map. The reference element is the unit cube.
     */
    inline concepts::Real3d chi(const Real x, const Real y,
				const Real z) const {
      return cell_.chi(x, y, z); }

    /// Computes the Jacobian
    inline concepts::MapReal3d jacobian(const Real x, const Real y,
					const Real z) const {
      concepts::MapReal3d jac = cell_.jacobian(x, y, z);
      jac *= 0.5;
      return jac; }

    /// Computes the inverse of the Jacobian
    inline concepts::MapReal3d jacobianInverse(const Real x, const Real y, 
					       const Real z) const {
      return jacobian(x, y, z).inverse(); }

    /// Computes the determinant of the Jacobian
    inline Real jacobianDeterminant(const Real x, const Real y,
				    const Real z) const {
      return jacobian(x, y, z).determinant(); }

    /** Sets the subdivision strategy of the underlying cell of this element.
	It calls Quad2d::setStrategy.

        @param strategy Pointer to an instance of a subdivision strategy.
	@throws StrategyChange if the change is not allowed (the change is
	        not allowed if there are children present)
    */
    void setStrategy(const concepts::Hex3dSubdivision* strategy = 0)
      throw(concepts::StrategyChange) { cell_.setStrategy(strategy); }

    /** Returns the subdivision strategy of the underlying cell of
	this element.
    */
    const concepts::Hex3dSubdivision* getStrategy() const {
      return cell_.getStrategy(); }

    virtual const concepts::ElementGraphics<Real>* graphics() const;

    virtual bool operator<(const Element<Real>& elm) const;

    /** Returns true if element matrix is the same. This method returns
	true if the element matrix of this element and \c elm are the same.
	To find this out, the length of the edges and their angles and the
	polynomial degrees are compared.
    */
    bool hasSameMatrix(const Hexahedron& elm) const;

    /// Set polynomial degree of edge \c i to \c p
    void edgeP(const uint i, const concepts::AdaptiveControlP<1>& p) {
      conceptsAssert(i < 12, concepts::Assertion());
      edges_[i] = &p; }
    /// Get polynomial degree of edge \c i
    const concepts::AdaptiveControlP<1>& edgeP(const uint i) const {
      conceptsAssert(i < 12, concepts::Assertion());
      conceptsAssert(edges_[i], concepts::Assertion());
      return *(edges_[i]); }
    /// Set polynomial degree of face \c i to \c p
    void faceP(const uint i, const concepts::AdaptiveControlP<2>& p) {
      conceptsAssert(i < 6, concepts::Assertion());
      faces_[i] = &p; }
    /// Get polynomial degree of face \c i
    const concepts::AdaptiveControlP<2>& faceP(const uint i) const {
      conceptsAssert(i < 6, concepts::Assertion());
      conceptsAssert(faces_[i], concepts::Assertion());
      return *(faces_[i]); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// The cell
    concepts::Hexahedron3d& cell_;
    /// The shape functions
    std::auto_ptr<concepts::Karniadakis<1,0> > shpfctX_, shpfctY_, shpfctZ_;
    /// The derivatives of the shape functions
    std::auto_ptr<concepts::Karniadakis<1,1> > shpfctDX_, shpfctDY_, shpfctDZ_;
    /// The integration rules
    std::auto_ptr<concepts::Quadrature<INTEGRATION_RULE> > intX_, intY_, intZ_;

    /// Polynomial degree of edges
    const concepts::AdaptiveControlP<1>* edges_[12];
    /// Polynomial degree of faces
    const concepts::AdaptiveControlP<2>* faces_[6];

    static std::auto_ptr<HexahedronGraphics> graphics_;
  };

} // namespace hp3D

#endif // hexahedron_hh
