/* Graphics for hexahedra
 */

#include <sstream>
#include "hexahedronGraphics.hh"
#include "hexahedron.hh"
#include "basics/exceptions.hh"
#include "graphics/dataDX.hh"

// debugging
#include "basics/debug.hh"

#define DX_DATA_D 0

namespace hp3D {

  // **************************************************** HexahedronGraphics **

  void HexahedronGraphics::operator()
    (const concepts::Element<Real>& elm, enum graphicsType type,
     concepts::CellPostprocess<Real>& post) const {

    const Hexahedron* elmH = dynamic_cast<const Hexahedron*>(&elm);
    conceptsAssert(elmH, concepts::Assertion());

    switch(type) {

    case TOPOLOGY_DOTTY:
    case VERTEX_LIST:
    case MESH_MATLAB:
    case MESH_DX:
      post(elmH->cell()); break;

    case DATA_DX_COUNTS: {
      graphics::DataDXCounts* counter =
	dynamic_cast<graphics::DataDXCounts*>(&post);
      conceptsAssert(counter, concepts::Assertion());
      counter->add(elmH->integrationX()->n()*
		   elmH->integrationY()->n()*elmH->integrationZ()->n());
      break;
    }

    case DATA_DX_COORDINATES: {
      graphics::DataDXCoordinates* coord = 
	dynamic_cast<graphics::DataDXCoordinates*>(&post);
      conceptsAssert(coord, concepts::Assertion());

      // polynomial degrees of the hexahedron
      coord->add((elmH->integrationX()->n()-1)*
		 (elmH->integrationY()->n()-1)*(elmH->integrationZ()->n()-1));

      const Real* px = elmH->integrationX()->abscissas();
      const Real* py = elmH->integrationY()->abscissas();
      const Real* pz = elmH->integrationZ()->abscissas();

      const concepts::Hexahedron3d& cell = elmH->cell();

      for (uint k = 0; k < elmH->integrationZ()->n(); ++k) {
	for (uint j = 0; j < elmH->integrationY()->n(); ++j) {
	  for (uint i = 0; i < elmH->integrationX()->n(); ++i, coord->inc() ) {
	    concepts::Real3d x = cell.chi((1.0+px[i])/2.0,
					  (1.0+py[j])/2.0,
					  (1.0+pz[k])/2.0);
	    coord->write(x[0]);
	    coord->write(x[1]);
	    coord->write(x[2]);
	  } // for i
	} // for j
      } // for k
      break;
    }

    case DATA_DX_CONNECTIONS: {
      graphics::DataDXConnections* conn = 
	dynamic_cast<graphics::DataDXConnections*>(&post);
      conceptsAssert(conn, concepts::Assertion());
      uint l = conn->counts();
      for (uint k = 0; k < elmH->integrationZ()->n()-1; ++k, 
	     l += elmH->integrationX()->n()) {
	for (uint j = 0; j < elmH->integrationY()->n()-1; ++j, ++l) {
	  for (uint i = 0; i < elmH->integrationX()->n()-1; ++i, ++l) {
	    conn->write(l);
	    conn->write(l+1);
	    conn->write(l+elmH->integrationX()->n());
	    conn->write(l+elmH->integrationX()->n()+1);
	    conn->write(l
			+elmH->integrationY()->n()*elmH->integrationX()->n());
	    conn->write(l+1
			+elmH->integrationY()->n()*elmH->integrationX()->n());
	    conn->write(l+elmH->integrationX()->n()
			+elmH->integrationY()->n()*elmH->integrationX()->n());
	    conn->write(l+elmH->integrationX()->n()+1
			+elmH->integrationY()->n()*elmH->integrationX()->n());
	  } // for i
	} // for j
      } // for k

      conn->add(elmH->integrationX()->n()*
		elmH->integrationY()->n()*elmH->integrationZ()->n());
      break;
    }

    case DATA_DX_DATA: {
      graphics::DataDXData* data = 
	dynamic_cast<graphics::DataDXData*>(&post);
      conceptsAssert(data, concepts::Assertion());

      DEBUGL(DX_DATA_D, elm);

      const uint nqx = elmH->integrationX()->n();
      const uint nqy = elmH->integrationY()->n();
      const uint nqz = elmH->integrationZ()->n();

      const concepts::Formula* formula = data->formula();
      const concepts::Array<Real>& coeff = data->coeff();
      const uint coeffSize = coeff.size();
      const concepts::ElementFormula<Real>* elmForm = data->elementFormula();

      DEBUGL(DX_DATA_D, "coeff = " << coeff);

      for (uint k = 0; k < nqz; ++k) {
	for (uint j = 0; j < nqy; ++j) {
	  for (uint i = 0; i < nqx; ++i) {
	    Real res = 0.0;
	    if (formula) {
	      const Real* px = elmH->integrationX()->abscissas();
	      const Real* py = elmH->integrationY()->abscissas();
	      const Real* pz = elmH->integrationZ()->abscissas();

	      concepts::Real3d x = elmH->cell().chi((1.0+px[i])/2.0,
						    (1.0+py[j])/2.0,
						    (1.0+pz[k])/2.0);
	      res = (*formula)(x);
	    } else {
	      if (elmForm) {
		const Real* px = elmH->integrationX()->abscissas();
		const Real* py = elmH->integrationY()->abscissas();
		const Real* pz = elmH->integrationZ()->abscissas();
		res = (*elmForm)(*elmH, concepts::Real3d((1.0+px[i])/2.0,
							 (1.0+py[j])/2.0,
							 (1.0+pz[k])/2.0));
	      } else {
		// musst have coeff!
		if (coeffSize == 0)
		  res = 0;
		else {
		  const Real * coeffAcc = coeff;
		  const uint nsx = elmH->shpfctX()->n();
		  const uint nsy = elmH->shpfctY()->n();
		  const uint nsz = elmH->shpfctZ()->n();
		  const Real * const sx = elmH->shpfctX()->values();
		  const Real * const sy = elmH->shpfctY()->values();
		  const Real * const sz = elmH->shpfctZ()->values();

		  for (uint iz = 0; iz < nsz; ++iz) {
		    for (uint iy = 0; iy < nsy; ++iy) {
		      for (uint ix = 0; ix < nsx; ++ix) {

			res += *coeffAcc++ * 
			  ( sx[ix*nqx + i] * sy[iy*nqy + j] * sz[iz*nqz + k] );
		      } // ix
		    } // iy
		  } // iz
		} // coeffSize is 0
	      } // has coeff
	    } // has no formula
	    data->write(res);
	  } // for i
	} // for j
      } // for k
      break;
    }

    default: 
      std::stringstream error;
      error << "graphics type " << type << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str())); break;
    }

  }

  std::ostream& HexahedronGraphics::info(std::ostream& os) const {
    return os << "HexahedronGraphics()";
  }

} // namespace hp3D
