/* Linear forms for hp 3D FEM
 */

#include <memory>
#include <cstring>
#include <sstream>
#include "basics/vectorsMatrices.hh"
#include "basics/exceptions.hh"
#include "toolbox/array.hh"
#include "toolbox/formula.hh"
#include "toolbox/boundary.hh"
#include "geometry/boundaryConditions.hh"
#include "space/element.hh"
#include "hp3D/hexahedron.hh"
#include "hp3D/linearForm.hh"

#include "basics/debug.hh"

// debugging
#define hp3DRieszConstr_D 0
#define hp3DRieszAppl_D 0
#define hp3DRieszApplBC_D 0

namespace hp3D {

  // ***************************************************************** Riesz **

  Riesz::Riesz(const concepts::Formula& frm,
	       concepts::BoundaryConditions* bc) :
    frm_(frm.clone()), jacobian_(0), bc_(bc) {
    DEBUGL(hp3DRieszConstr_D, *frm_);
    DEBUGL(hp3DRieszConstr_D, *bc_);
  }

  Riesz::~Riesz() {}

  void Riesz::operator()(const concepts::Element<Real>& elm,
			 concepts::ElementMatrix<Real>& em) {
    const Hexahedron* elmH = dynamic_cast<const Hexahedron*>(&elm);
    if (elmH) { this->operator()(*elmH, em); return; }
    std::stringstream error;
    error << "element " << elm << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }

  void Riesz::operator()(const Hexahedron& elm,
			 concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();
    DEBUGL(hp3DRieszAppl_D, "elm = " << elm);
    DEBUGL(hp3DRieszAppl_D, "n = " << n);

    // quadrature points
    const uint nqx = elm.integrationX()->n();
    const uint nqy = elm.integrationY()->n();
    const uint nqz = elm.integrationZ()->n();
    const Real * const wx = elm.integrationX()->weights();
    const Real * const wy = elm.integrationY()->weights();
    const Real * const wz = elm.integrationZ()->weights();
    const Real * px = elm.integrationX()->abscissas();
    const Real * py = elm.integrationY()->abscissas();
    const Real * pz = elm.integrationZ()->abscissas();

    // Jacobian
    jacobian_.resize(nqx*nqy*nqz);
    DEBUGL(hp3DRieszAppl_D, "reserved " << (nqx*nqy*nqz) << " for jacobian_");
    Real* dx = (Real*)jacobian_;
    const Real * pxAcc = px;
    for (uint i = 0; i < nqx; ++i) {
      const Real * pyAcc = py;
      for (uint j = 0; j < nqy; ++j) {
	const Real * pzAcc = pz;
	for (uint k = 0; k < nqz; ++k) {
	  *dx++ = elm.jacobianDeterminant( (*pxAcc + 1.0)/2.0, 
					   (*pyAcc + 1.0)/2.0,
					   (*pzAcc + 1.0)/2.0 )
	    * (*frm_)(elm.chi( (*pxAcc + 1.0)/2.0, (*pyAcc + 1.0)/2.0,
			       (*pzAcc++ + 1.0)/2.0 ));
	}
	pyAcc++;
      }
      pxAcc++;
    }

    // shapefunctions
    const uint nsx = elm.shpfctX()->n();
    const uint nsy = elm.shpfctY()->n();
    const uint nsz = elm.shpfctZ()->n();
    const Real * sx = elm.shpfctX()->values();
    const Real * sy = elm.shpfctY()->values();
    const Real * sz = elm.shpfctZ()->values();

    // integration
    Real* intermed = (Real*)em;
    std::memset(intermed, 0, sizeof(Real)*nsx*nsy*nsz);
    for (uint iz = 0; iz < nsz; ++iz) {
      for (uint iy = 0; iy < nsy; ++iy) {
	for (uint ix = 0; ix < nsx; ++ix) {
	  const Real * wxAcc = wx;
	  const Real * six = sx + ix*nqx;
	  dx = (Real*)jacobian_;
	  for (uint I = 0; I < nqx; ++I) {
	    const Real * wyAcc = wy;
	    const Real * siy = sy + iy*nqy;
	    for (uint J = 0; J < nqy; ++J) {
	      const Real * wzAcc = wz;
	      const Real * siz = sz + iz*nqz;
	      for (uint K = 0; K < nqz; ++K) {
		*intermed += *dx++ 
		  * *wxAcc * *wyAcc * *wzAcc++
		  * *six * *siy * *siz++;
	      } // K
	      wyAcc++;
	      siy++;
	    } // J
	    wxAcc++;
	    six++;
	  } // I
	  intermed++;
	} // ix
      } // iy
    } // iz

    DEBUGL(hp3DRieszAppl_D, "volume integration: em = " << em);

    // treating Neumann boundary conditions
    if (bc_) {
      DEBUGL(hp3DRieszApplBC_D, "bc = " << *bc_);

      // loop over the faces (triangles or quads)
      uint i = 0; 
      for (concepts::Connector2* F = 0; (F = elm.support().face(i)); ++i ) {
	uint attrib = F->attrib();
	DEBUGL(hp3DRieszApplBC_D, *F << " has attribute " << attrib);

	if ( (*bc_)(attrib).type() == concepts::Boundary::NEUMANN) {
	  DEBUGL(hp3DRieszApplBC_D, *F << " has NEUMANN conditions " << 
		 (*bc_)(attrib));

	  // restore points and shapefunctions
	  px = elm.integrationX()->abscissas();
	  py = elm.integrationY()->abscissas();
	  pz = elm.integrationZ()->abscissas();
	  sx = elm.shpfctX()->values();
	  sy = elm.shpfctY()->values();
	  sz = elm.shpfctZ()->values();

	  // dummy points and shape functions which lie on the end points
	  std::auto_ptr<concepts::Quadrature<0> > qDummy(0);
	  std::auto_ptr<concepts::Karniadakis<1,0> > sDummy(0);

	  uint ni = 0, nj = 0; // number of quadrature points
	  const Real* wi = 0, * wj = 0; // weights of the quadrature points
	  uint I = 100, J = 200; // counter variables in the loops
	  uint K = 300; // is changed and used as a constant
	  uint* xcnt = 0, * ycnt = 0,
	    * zcnt = 0; // indices in the 3 directions

	  // rows of the Jacobian which have to be inserted into the
	  // Gram determinant
	  uint jacRow1 = 0, jacRow2 = 0;
	  switch(i) {
	  case 0:
	    DEBUGL(hp3DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationZ()) << ", " << *(elm.shpfctZ()));
	    xcnt = &I; ycnt = &J; zcnt = &K; K = 0;
	    ni = elm.integrationX()->n();
	    nj = elm.integrationY()->n();
	    wi = elm.integrationX()->weights();
	    wj = elm.integrationY()->weights();
	    jacRow1 = 0; jacRow2 = 1;
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationZ()->n()));
	    pz = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctZ()->n(), qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sz = sDummy->values();
	    break;
	  case 1:
	    DEBUGL(hp3DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationY()) << ", " << *(elm.shpfctY()));
	    xcnt = &I; ycnt = &K; zcnt = &J; K = 0;
	    ni = elm.integrationX()->n();
	    nj = elm.integrationZ()->n();
	    wi = elm.integrationX()->weights();
	    wj = elm.integrationZ()->weights();
	    jacRow1 = 0; jacRow2 = 2;
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationY()->n()));
	    py = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctY()->n(), qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sy = sDummy->values();
	    break;
	  case 2:
	    DEBUGL(hp3DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationX()) << ", " << *(elm.shpfctX()));
	    xcnt = &K; ycnt = &I; zcnt = &J; K = elm.integrationX()->n()-1;
	    ni = elm.integrationY()->n();
	    nj = elm.integrationZ()->n();
	    wi = elm.integrationY()->weights();
	    wj = elm.integrationZ()->weights();
	    jacRow1 = 1; jacRow2 = 2;
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationX()->n()));
	    px = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctX()->n(), qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sx = sDummy->values();
	    break;
	  case 3:
	    DEBUGL(hp3DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationY()) << ", " << *(elm.shpfctY()));
	    xcnt = &I; ycnt = &K; zcnt = &J; K = elm.integrationY()->n()-1;
	    ni = elm.integrationX()->n();
	    nj = elm.integrationZ()->n();
	    wi = elm.integrationX()->weights();
	    wj = elm.integrationZ()->weights();
	    jacRow1 = 0; jacRow2 = 2;
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationY()->n()));
	    py = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctY()->n(), qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sy = sDummy->values();
	    break;
	  case 4:
	    DEBUGL(hp3DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationX()) << ", " << *(elm.shpfctX()));
	    xcnt = &K; ycnt = &I; zcnt = &J; K = 0;
	    ni = elm.integrationY()->n();
	    nj = elm.integrationZ()->n();
	    wi = elm.integrationY()->weights();
	    wj = elm.integrationZ()->weights();
	    jacRow1 = 1; jacRow2 = 2;
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationX()->n()));
	    px = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctX()->n(), qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sx = sDummy->values();
	    break;
	  case 5:
	    DEBUGL(hp3DRieszApplBC_D, "before new integration rule: " <<
		   *(elm.integrationZ()) << ", " << *(elm.shpfctZ()));
	    xcnt = &I; ycnt = &J; zcnt = &K; K = elm.integrationZ()->n()-1;
	    ni = elm.integrationX()->n();
	    nj = elm.integrationY()->n();
	    wi = elm.integrationX()->weights();
	    wj = elm.integrationY()->weights();
	    jacRow1 = 0; jacRow2 = 1;
	    qDummy.reset(new concepts::Quadrature<0>(elm.integrationZ()->n()));
	    pz = qDummy->abscissas();
	    sDummy.reset(new concepts::Karniadakis<1,0>
			 (elm.shpfctZ()->n(), qDummy->abscissas(),
			  qDummy->n(), 0, 0, false));
	    sz = sDummy->values();
	    break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	  DEBUGL(hp3DRieszApplBC_D, "after new integration rule: " <<
		 *qDummy << ", " << *sDummy);
	  DEBUGL(hp3DRieszApplBC_D, "xcnt = " << *xcnt << ", ycnt = " 
		 << *ycnt << ", zcnt = " << *zcnt << ", ni = " << ni 
		 << ", nj = " << nj);

	  jacobian_.resize(ni*nj);
	  DEBUGL(hp3DRieszApplBC_D, "reserved " << ni*nj << " for Jacobian");
	  dx = (Real*)jacobian_;
	  for (I = 0; I < ni; ++I) {
	    for (J = 0; J < nj; ++J) {
	      concepts::MapReal3d jac = elm.jacobian((px[*xcnt] + 1.0)/2.0,
					   (py[*ycnt] + 1.0)/2.0,
					   (pz[*zcnt] + 1.0)/2.0);
	      concepts::Real3d first(jac.column(jacRow1));
	      concepts::Real3d second(jac.column(jacRow2));
	      *dx++ = (first ^ second).l2()
		* (*bc_)(attrib)(elm.chi( (px[*xcnt] + 1.0)/2.0,
					  (py[*ycnt] + 1.0)/2.0,
					  (pz[*zcnt] + 1.0)/2.0));
	    } // for J
	  } // for I

	  // integration
	  Real* intermed = (Real*)em;
	  for (uint iz = 0; iz < nsz; ++iz) {
	    for (uint iy = 0; iy < nsy; ++iy) {
	      for (uint ix = 0; ix < nsx; ++ix) {
		const Real * six = sx + ix*nqx;
		const Real * siy = sy + iy*nqy;
		const Real * siz = sz + iz*nqz;
		dx = (Real*)jacobian_;
		const Real * wiAcc = wi;
		for (I = 0; I < ni; ++I) {
		  const Real * wjAcc = wj;
		  for (J = 0; J < nj; ++J) {
		    *intermed += *dx++
		      * *wiAcc * *wjAcc++
		      * six[*xcnt] * siy[*ycnt] * siz[*zcnt];
		  } // for J
		  wiAcc++;
		} // for I
		intermed++;
	      } // iz
	    } // iy
	  } // ix

	} // bc is NEUMANN
      } // for F: faces
    } // if bc_

    DEBUGL(hp3DRieszAppl_D, "em = " << em);
  }

  std::ostream& Riesz::info(std::ostream& os) const {
    return os << "Ries(" << *frm_ << ')';
  }

} // namespace concepts
