/* Linear forms for hp 3D FEM
 */

#ifndef hp3dlinearform_hh
#define hp3dlinearform_hh

#include <memory>
#include "function/linearForm.hh"
#include "basics/typedefs.hh"

namespace concepts {
  // forward declarations
  class Formula;
  class BoundaryConditions;

  template<class F>
  class Element;

  template<class F>
  class ElementMatrix;

  template<class F>
  class Array;
}

namespace hp3D {
  // forward declarations
  class Hexahedron;

  using concepts::Real;

  // ***************************************************************** Riesz **

  /** Linear form in 3D.

      This linear form computes
      \f[ \int_K f v \, dx. \f]
      Currently only on hexahedrons.
      @author Philipp Frauenfelder, 2000
  */
  class Riesz : public concepts::LinearForm<Real> {
  public:
    /** Constructor. Parses the formula.
	@param frm The formula
	@param bc Boundary conditions, defaults to homogeneous
    */
    Riesz(const concepts::Formula& frm, concepts::BoundaryConditions* bc = 0); 
    virtual ~Riesz();

    /** Computes the element load vector. As for the computation of an
	element stiffness matrix, there are the loops over all quadrature
	points and the loops over all shape functions.
	@param elm The element for which the load vector should be computed.
	@param em The load vector
    */
    void operator()(const concepts::Element<Real>& elm,
		    concepts::ElementMatrix<Real>& em);
    void operator()(const Hexahedron& elm, concepts::ElementMatrix<Real>& em);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;

    /// Intermediate data for element matrix computation
    concepts::Array<Real> jacobian_;

    /** Reference to the boundary conditions.
     */
    concepts::BoundaryConditions* bc_;
  }; 

} // namespace hp3D

#endif // hp3dlinearform_hh
