/* discretization of rot * rot for Maxwell equations in frequency domain
 */

#ifndef rotrot3D_hh
#define rotrot3D_hh

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "toolbox/array.hh"
#include "operator/bilinearForm.hh"

namespace vectorial {
  template<typename F>
  class BilinearForm;
} // namespace vectorial

namespace hp3D {
  // forward declarations
  class Hexahedron;

  using concepts::Real;

  // **************************************************************** RotRot **

  /** A function class to calculate element matrices 
      for the Rot u*Rot v Bilinearform

      @author Kersten Schmidt, 2002
  */
  class RotRot : public concepts::BilinearForm<Real> {
  public:
    /// Constructor of the i.th row and j.th column of vectorial Bilinearform
    RotRot(concepts::Array<Real>& jacobian,
	   concepts::Array<concepts::MapReal3d>& jacobianInv,
	   const Hexahedron** oldElm, const uint i = 0, const uint j = 0) :
      i_(i), j_(j), jacobian_(jacobian), jacobianInv_(jacobianInv),
      coeffMatrix_(0), oldElm_(oldElm) {}

    virtual ~RotRot();

    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY, 
			    concepts::ElementMatrix<Real>& em); 
    void operator()(const Hexahedron& elmX, const Hexahedron& elmY, 
		    concepts::ElementMatrix<Real>& em);

    /** Sets up the bilinear form for the rot-rot part for Maxwell's
	equation using weighted regularization.
	@param bf Resulting bilinear form (use this to compute the matrix)
	@pre The number of components of \c bf needs to be 3 x 3.
    */
    static void setup(vectorial::BilinearForm<Real>& bf);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// row and column inside the vectorial Bilinearform
    uint i_, j_;

    //@{
    /// Intermediate data for element matrix computation
    concepts::Array<Real> &jacobian_;
    concepts::Array<concepts::MapReal3d> &jacobianInv_;
    concepts::Array<concepts::MapReal3d> coeffMatrix_;
    //@}

    const Hexahedron** oldElm_;

    //@{
    /** Data shared between components of the rot-rot bilinear forms,
	needed by setup()
    */
    static concepts::Array<Real> jacobianRot_;
    static concepts::Array<concepts::MapReal3d> jacobianInvRot_;
    static const Hexahedron* elmRot_;
    static concepts::Array<std::auto_ptr<RotRot> > rotrot_;
    //@}
  };

} // namespace hp3D

#endif // rotrot3D_hh
