/* weighted functions for regularization, which takes
   the vertex or edge with shortest distance
*/

#ifndef shortestDist_hh
#define shortestDist_hh

#include <cmath>
#include "basics/vectorsMatrices.hh"
#include "singularSet.hh"

namespace hp3D {

  // forward declarations
  class Hexahedron;
  using concepts::Real;

  // ********************************************************* TrivialWeight **

  /** A function class for trivial (constant equal 1.0) weight function.
      @author Kersten Schmidt, 2002
  */
  class TrivialWeight {
  public:
    Real operator()(const SingularSet&, concepts::Real3d,
		    const Hexahedron*) const { return 1.0; }
  };

  std::ostream& operator<<(std::ostream& os, const TrivialWeight& p);

  // ********************************************************** ShortestDist **

  /** A function class for weighted regularization, which returns
      the square of the shortest distance of a point to the singular
      edges and vertices
      @author Kersten Schmidt, 2002
  */
  class ShortestDist {
  public:
    /** returns the square of the shortest distance of a point to the singular
      edges and vertices

      @param point Reference coordinates of the point, for whom the shortest
                   distance is asked
      @param elm   Element the point is belonging to
      @param singularities Storage of singular edges and vertices
    */
    Real operator()(const SingularSet& singularities,
		    concepts::Real3d point, const Hexahedron* elm) const;
  };

  std::ostream& operator<<(std::ostream& os, const ShortestDist& p);

  // ***************************************************** PostprocessSquare **

  /** Squares the given \c x.
      @author Philipp Frauenfelder, 2003
  */
  class PostprocessSquare {
  public:
    static Real post(Real x) { return x*x; }
  };

  std::ostream& operator<<(std::ostream& os, const PostprocessSquare& p);

  // ******************************************************* PostprocessRoot **

  /** Returns the square root of \c x.
      @author Philipp Frauenfelder, 2003
  */
  class PostprocessRoot {
  public:
    static Real post(Real x) { return std::sqrt(x); }
  };

  std::ostream& operator<<(std::ostream& os, const PostprocessRoot& p);

  // ****************************************************** PostprocessRoot4 **

  /** Returns the forth root of \c x.
      @author Kersten Schmidt, 2003
  */
  class PostprocessRoot4 {
  public:
    static Real post(Real x) { return std::sqrt(std::sqrt(x)); }
  };

  std::ostream& operator<<(std::ostream& os, const PostprocessRoot4& p);

  // ********************************************************* Postprocess34 **

  /** Computes \c x to the power of 3/4.
      @author Philipp Frauenfelder, 2003
  */
  class Postprocess34 {
  public:
    static Real post(Real x) { return std::pow(x, 0.75); }
  };

  std::ostream& operator<<(std::ostream& os, const Postprocess34& p);

  // ********************************************************* Postprocess14 **

  /** Computes \c x to the power of 1/4.
      @author Philipp Frauenfelder, 2003
  */
  class Postprocess14 {
  public:
    static Real post(Real x) { return std::pow(x, 0.25); }
  };

  std::ostream& operator<<(std::ostream& os, const Postprocess14& p);

  // ********************************************************** Postprocess4 **

  /** Computes \c x to the power of 4.
      @author Philipp Frauenfelder, 2003
  */
  class Postprocess4 {
  public:
    static Real post(Real x) { return x*x*x*x; }
  };

  std::ostream& operator<<(std::ostream& os, const Postprocess4& p);

  // ********************************************************** Postprocess3 **

  /** Computes \c x to the power of 3.
      @author Philipp Frauenfelder, 2003
  */
  class Postprocess3 {
  public:
    static Real post(Real x) { return x*x*x; }
  };

  std::ostream& operator<<(std::ostream& os, const Postprocess3& p);

  // ********************************************************** DistancePost **

  /** A function class for weighted regularization. It uses a a distance
      function (see template parameter) and simply exponentiates the result
      it gets from there.
      @param DistClass Function class which is able to evaluate a distance
      @param Function Function class which is able to post process the result
             from DistClass (has a static function <tt>post(Real, Real)</tt>
      @author Philipp Frauenfelder, 2003
  */
  template<typename DistClass, typename Function>
  class DistancePost {
  public:
    /** Uses \c DistClass to compute the weight function and \c Function to
	post process the weight.
    */
    Real operator()(const SingularSet& singularities,
		    concepts::Real3d point, const Hexahedron* elm) const {
      return Function::post(dist_(singularities, point, elm));
    }
  private:
    DistClass dist_;
  };

  template<typename DistClass, typename Function>
  inline std::ostream& operator<<(std::ostream& os,
				  const DistancePost<DistClass, Function>& p) {
    return os << Function() << '(' << DistClass() << ')';
  }

  // ********************************************************** ProductOfAll **

  /** Computes the product of all distances in the singular set
      \c singularities.
      \f[ w = \left( \prod_{c\in C} r_c^2 \right)
              \left( \prod_{e\in E} \rho_e^2 \right) \f]
      where \f$C\f$ is the set of singular corners, \f$E\f$ the set of singular
      edges and \f$\rho_e\f$ is defined by \f$r_e = \rho_e r_c r_{c'}\f$ where
      \f$c\f$ and \f$c'\f$ are the corners of the edge.
      @author Philipp Frauenfelder, 2003
  */
  class ProductOfAll {
  public:
    Real operator()(const SingularSet& singularities,
		    concepts::Real3d point, const Hexahedron* elm) const;
  };  

  std::ostream& operator<<(std::ostream& os, const ProductOfAll& p);

  // *********************************************************** DaugeWeight **

  /** Weight implemented by
      <a href="http://perso.univ-rennes1.fr/monique.dauge/">Monique Dauge</a>
      in
      <a href="http://perso.univ-rennes1.fr/daniel.martin/melina/www/">Melina</a>.
      Her Fortran code:
      <pre>      XX= POINT(1)
      YY= POINT(2)
      ZZ= POINT(3)
      R1=SQRT (ZZ*ZZ+YY*YY)
      R2=SQRT (ZZ*ZZ+XX*XX)
      R3=SQRT (XX*XX+YY*YY)
      IF (XX+YY+ZZ.GE.0.) THEN
         R=SQRT (XX*XX+YY*YY+ZZ*ZZ)
      ELSE
         R=MIN(MIN(R1,R2),R3)
      ENDIF
      IF (R.GE.1.E-12) THEN
         RHO=S*EXP(ALPHA*LOG(R))
      ELSE
         RHO=0.
      ENDIF</pre>
      This computes a correct weight for the Fichera corner.
      @see <a href="http://perso.univ-rennes1.fr/monique.dauge/">Homepage of Monique Dauge</a>
      @see <a href="http://perso.univ-rennes1.fr/daniel.martin/melina/www/">Homepage of Melina</a>
      @author Philipp Frauenfelder, 2003
  */
  class DaugeWeight {
  public:
    Real operator()(const SingularSet& singularities,
		    concepts::Real3d point, const Hexahedron* elm) const;
  };

  std::ostream& operator<<(std::ostream& os, const DaugeWeight& p);

  // *************************************************** ShortestDistLimited **

  /** A function class for weighted regularization, which returns the minimum
      of a value and the square of the shortest distance of a point to the
      singular edges and vertices
      @bug Not correctly implemented
      @author Kersten Schmidt, 2002
  */
  class ShortestDistLimited {
  public:
    ShortestDistLimited() : shortestdist_() {}
    /** returns the minimum of a value and the square of the shortest distance
	of a point to the singular edges and vertices (not implemented)

      @param point Reference coordinates of the point, for whom the shortest
                   distance is asked
      @param elm   Element the point is belonging to
      @param singularities Storage of singular edges and vertices
    */
    Real operator()(const SingularSet& singularities,
		    concepts::Real3d point, const Hexahedron* elm) const {
      return shortestdist_(singularities, point, elm);
    }
  private:
    const ShortestDist shortestdist_;
  };

  std::ostream& operator<<(std::ostream& os, const ShortestDistLimited& p);

} // namespace hp3D

#endif // shortestDist_hh
