/* singular Edges and classes for distances
*/

#ifndef singularSet_hh
#define singularSet_hh

#include "basics/vectorsMatrices.hh"
#include "toolbox/scannerConnectors.hh"
#include "geometry/topology.hh"
#include "geometry/mesh.hh"

namespace hp3D {
// forward declarations
  class Hexahedron;
  class TrivialWeight;

  using concepts::Real;

  // ********************************************************** SingularEdge **

  /** Class for storing a singular edge with coordinates of its
      corners. The class also provides methods to compute a distance
      from a given point (in local coordinates in an element) to the
      edge. The returned results are the Euclidiean distance squared.
      @author Kersten Schmidt, 2002
  */
  class SingularEdge {
    friend std::ostream& operator<<(std::ostream& os, const SingularEdge& s);
  public:
    /** Constructor.
	Precomputes \c diffVector_, \c diffVectorNormed_
	and \c diffVectorTimesVtx0Normed_ to allow fast computations
	in the \c distance function.
	@param edge Edge object
	@param vtx0, vtx1 Coordinates of the corners of the edge
    */
    SingularEdge(const concepts::Connector1* edge, const concepts::Real3d vtx0,
		 const concepts::Real3d vtx1) :
      edge_(edge), vtx0_(vtx0), vtx1_(vtx1) {
      diffVector_ = diffVectorNormed_ = (vtx1_-vtx0_);
      diffVectorNormed_ *= 1.0 / diffVectorNormed_.l2_2();
      diffVectorTimesVtx0Normed_ = diffVectorNormed_ * vtx0_;
    }
    /// Returns a pointer to the edge object (topological, connector)
    const concepts::Connector1* edge() const { return edge_; }
    /** Returns the square of the distance from the edge to a \c point
	inside \c elm
	@param point Point in reference coordinates
	       \f$\xi \in [0,1]^3\f$
	@param elm Hexahedron in which the point is located

	Let \f$k\f$ be a straight line along the edge, in parameters:
	\f[
	  k(t) := x_c + (x_{c'}- x_c) \cdot t = x_c + x_D \cdot t
	\f]
	where \f$x_c\f$ and \f$x_{c'}\f$ are the corners of the edge.

	To determine the root point \f$k(t)\f$ of \f$x\f$ (\f$x\f$ is
	\f$\xi\f$ in physical coordinates) on \f$k\f$, compute
	\f[
	  t = x \cdot\frac{x_D}{|x_D|^2} -
	  \frac{x_c\cdot x_D}{|x_D|^2} = x \cdot a - b
	\f]
	where \f$a\f$ = \c diffVectorNormed_ and \f$b\f$ =
	\c diffVectorTimesVtx0Normed_ respectively.

	If \f$0 < t < 1\f$, the closest point on the edge is
	not one of the corners and the distance \f$x - k(t)\f$ is
	computed. Otherwise, the distance to the closer corner is
	computed.
    */
    Real distance(const concepts::Real3d point, const Hexahedron* elm) const;
    /// Returns the the square of the distance of \c point to \c vtx0_
    Real distanceV0(const concepts::Real3d point, const Hexahedron* elm) const;
    /// Returns the the square of the distance of \c point to \c vtx1_
    Real distanceV1(const concepts::Real3d point, const Hexahedron* elm) const;
  private:
    /// Pointer to the edge object (topological, connector)
    const concepts::Connector1* edge_;
    /// Coordinates of the vertices of the edge (\f$x_c\f$ and \f$x_{c'}\f$)
    const concepts::Real3d vtx0_, vtx1_;
    /// Difference vector between the 2 vertices: \f$x_c - x_{c'}\f$
    concepts::Real3d diffVector_;
    /// Scaled difference vector: \f$(x_c - x_{c'})/|x_c - x_{c'}|^2\f$
    concepts::Real3d diffVectorNormed_;
    /// Scalar product of the scaled difference vector and
    /// the vector to the first vertex: 
    /// \f$(x_c - x_{c'}) \cdot x_c /|x_c - x_{c'}|^2\f$ 
    concepts::Real diffVectorTimesVtx0Normed_;
  };

  // ******************************************************** SingularVertex **

  /** Class for storing a singular vertex with the coordinates. The
      class also provides methods to compute a distance from a given
      point (in local coordinates in an element) to the vertex. The
      returned results are the Euclidiean distance squared.
      @author Kersten Schmidt, 2002
  */
  class SingularVertex {
    friend std::ostream& operator<<(std::ostream& os, const SingularVertex& s);
  public:
    /** Constructor.
	@param vertex Vertex object
	@param vtx0 Coordinates of the vertex
    */
    SingularVertex(const concepts::Connector0* vertex,
		   const concepts::Real3d vtx0) :
      vertex_(vertex), vtx0_(vtx0) {}
    /// Returns a pointer to the vertex object (topological, connector)
    const concepts::Connector0* vertex() const { return vertex_; }
    /// Returns the square of the distance from the vertex to the \c
    /// point (in physical coordinates)
    Real distance(concepts::Real3d point) const;
    /** Returns the distance from the vertex to a \c point inside an
	Hexahedron (point in reference coordinates)
     */
    Real distance(concepts::Real3d point, const Hexahedron* elm) const;
  private:
    /// Pointer to the vertex object (topological, connector)
    const concepts::Connector0* vertex_;
    /// Coordinates of the vertex
    const concepts::Real3d vtx0_;
  };

  // *********************************************************** SingularSet **

  /** Class for handling a set of singular edges and vertices
      @author Kersten Schmidt, 2002
  */
  class SingularSet {
    friend std::ostream& operator<<(std::ostream& os, const SingularSet& s);
  public:
    SingularSet() : vertices_(0), edges_(0) {}
    ~SingularSet();
    /** Add all vertices and edges in the mesh with \c attrib to the
	set of singular vertices and edges
    */
    void add(const concepts::Attribute& attrib, concepts::Mesh3& msh);
    /// Scanner over all singular vertices
    concepts::Scan<SingularVertex>* verticesScan() const {
      return new concepts::PListScan<SingularVertex>(*vertices_);
    }
    /// Scanner over all singular edges 
    concepts::Scan<SingularEdge>* edgesScan() const {
      return new concepts::PListScan<SingularEdge>(*edges_);
    }
  private:
    /// List of singular vertices
    concepts::Joiner<SingularVertex*, 1>* vertices_;
    /// List of singular edges
    concepts::Joiner<SingularEdge*, 1>* edges_;
  };

} // namespace hp3D

#endif // singularSet_hh
