/* space class for 3D hp FEM
 */

#ifndef hpSpace3D_hh
#define hpSpace3D_hh

#if __ICC == 800
#  include <hash_map>
#  define __gnu_cxx std
#else
#  if __GNUC__ == 3
#    include <ext/hash_map>
#    if __GNUC_MINOR__ < 1
#      define __gnu_cxx std
#    endif
#  else
#    include <hash_map>
#    define __gnu_cxx std
#  endif
#endif

#include <memory>
#include <map>
#include <cstring>
#include "basics/exceptions.hh"
#include "basics/typedefs.hh"
#include "basics/level.hh"
#include "space/space.hh"
#include "space/hpMethod.hh"
#include "space/smatrix.hh"
#include "geometry/continuityData.hh"
#include "geometry/cell3D.hh"
#include "geometry/mesh.hh"
#include "geometry/boundaryConditions.hh"
#include "toolbox/scannerConnectors.hh"
#include "element.hh"

namespace concepts {

  // forward declaration
  class InOutParameters;

  // ****************************************************************** Scan **

  template<>
  class Scan<hp3D::Element<Real> > :
    public concepts::Scan<concepts::Element<Real> > {
  public:  
    hp3D::Element<Real>& operator++(int) = 0;
  };

} // namespace concepts

namespace hp3D {
  using concepts::Real;

  // forward declarations
  class Hexahedron;

  // ***************************************************************** Space **

  /** A 3D hp FEM space with continuous, picewise polynomial basis functions.
      Currently, only hexahedra are possible.

      This class is equiped with an interface to get timings of internal
      computations if compiled accordingly (see \c space.cc file), see
      \c setTimings() and \c timings().

      @author Philipp Frauenfelder, 2001
      @test test::DeadLock2
      @test test::DeadLock3
  */
  class Space : public concepts::AdaptiveSpace<Real,
		concepts::AdaptiveAdjustP<3> > {
  public:
    typedef concepts::Scan<hp3D::Element<Real> > Scan;

    /** Constructor. Scans the mesh and sets the cells in the mesh active
	and the level of refinement and the polynomial degree in
	all cells to the given values. \c rebuild_ is set to true, ie. 
	if the mesh is used it will firstly be rebuilt.

	@param msh The domain of interest partitioned into a mesh.
	@param l Level of refinement
	@param p Degree of the polynomials to be used.
	@param bc Boundary conditions
	@param trunk Create trunk space ie., use the the rules
	       <tt>p+q <= pFaceMax + 1</tt> on the faces and
	       <tt>p+q+r <= pIntMax + 1</tt> in the interior instead of
	       allowing all <tt>p, q</tt> and <tt>p, q, r</tt> combinations
	       resp.
    */
    Space(concepts::Mesh3& msh, uint l, uint p,
	  concepts::BoundaryConditions* bc = 0, bool trunk = true);
    /// Copy constructor
    Space(const Space& spc);
    virtual ~Space();
    
    inline virtual uint dim() const;
    inline virtual uint nelm() const;
    inline virtual Scan* scan() const;

    inline uint dim();
    inline uint nelm();
    inline Scan* scan();

    virtual void adjust(const concepts::Element<Real>& elm,
			const concepts::AdaptiveAdjustP<3>& a);

    /** Rebuilds the space after an adjustment with \c adjust.

        Steps:
	<ul>
	<li>Calls \c meshAndPoly_ for every cell in the coarse mesh.
	<li>Creates the lists \c cellList_, \c faceList_, \c faceListEdg_,
	    \c edgeList_, \c edgeListVtx_, and \c vertexList_ by calling
	    \c createCellList_, \c createEdgFaceList_
	    and \c createVtxEdgList_ for every cell in the coarse mesh.
	<li>The lists which were created in the previous step are update and
	    changed using \c FaceData::checkRelatedFaces,
	    \c EdgeData::checkRelatedEdges,
	    \c EdgeData::checkRelations,
	    \c VertexData::checkRelations in order to find the
	    supports of the basis functions (which have to be continuous).
	<li>Calls \c enrichElm_ for every cell in the coarse mesh.
	<li>Calls \c buildElements_ for every cell in the coarse mesh.
	</ul>
	They all work recursively and rebuild the space.

	At the end of this routine, the information in \c adj_, \c ctrl0_,
	\c ctrl1_, \c ctrl2_, \c cellList_, \c faceList_, \c faceListEdg_,
	\c edgeList_, \c edgeListVtx_ and \c vertexList_ is deleted. Just
	the list of elements in \c elm_ and \c ctrl3_ is kept.

	@see concepts::TMatrixBase for more information on T matrices
	     and how they are stored
	@see createCellList_
	@see createVtxEdgList_
	@see concepts::FaceData::checkRelatedFaces
	@see concepts::EdgeData::checkRelatedEdges
	@see concepts::EdgeData::checkRelations
	@see concepts::VertexData::checkRelations
	@see enrichElm_
	@see buildElements_
	@return Copy of the space on the old level
    */
    Space rebuild();

    /// Returns the list of matrices matching the subdivision type \c subdiv
    const std::auto_ptr<concepts::SMatrixBase<Real> >*
    getSmatrices(const concepts::Hex3dSubdivision* subdiv) const;

    /** Checks if the S matrices need to be recomputed and does so if
	necessary.

	@param elm Element which the S matrices are needed for (the
	           polynomial degrees are important)
    */
    void recomputeSmatrices(const Hexahedron* elm) const;

    /** @name Timing Interface
     *  These functions are used to get timings from class internal
     *  computations. The values are stored in a user defined
     *  concepts::InOutParameters structure in different arrays (see 
     *  \c setTimings). These arrays can be grouped into a table for
     *  easier postprocessing with
     *  \code
    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::DOUBLE, "jacobian", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "whole_sumfact", output);
    std::ofstream ofs("table.gnuplot");
    ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(ofs);
        \endcode
     */
    //@{
    /** Sets the class to store the timing values in. Additionally,
	the \c timeCntr_ is reset to 0. This counter is used to fill
	in the values into the arrays listed below in subsequent calls.
	The following timings are taken and stored in \c timings:
	<ul>
	<li>subroutine \c meshAndPoly_ in \c meshAndPoly
	<li>creation of the lists of vertices, edges and faces in
	\c createTopologyLists
	<li>finding the support of faces modes in \c checkRelatedFaces
	<li>finding the support of edge modes in \c checkRelatedEdges
	<li>finding the support of vertex modes in \c checkRelatedVertices
	<li>enriching the degree of elements in \c enrichElements
	<li>building the elements and the T matrices in \c buildElements
	</ul>
    */
    static void setTimings(concepts::InOutParameters* timings);
    /** Returns true if the class is able to do timings.  The ability
	to do timings depends on a compiler switch in \c space.cc file.
    */
    static bool timings();
    //@}
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Mesh
    concepts::Mesh3& msh_;
    /// Boundary conditions
    concepts::BoundaryConditions* bc_;
    /// Build trunk space
    bool trunk_;
    /** If true: the elements have to be rebuilt. Some elements have been
	marked for refinement or for coarsening and the elements are
	outdated. */
    bool rebuild_;

    /// Dimension of the FE space
    uint dim_;
    /// Number of elements currently active in the mesh
    uint nelm_;
    /// Linked list of the elements
    concepts::Joiner<Element<Real>*, 1>* elm_;
    concepts::Joiner<Element<Real>*, 1>* endElm_;
    /// Hash table of the control information for the vertices
    __gnu_cxx::hash_map<uint, concepts::AdaptiveControl > ctrl0_;
    /// Hash table of the control information for the edges
    std::auto_ptr<__gnu_cxx::hash_map<uint,
				      concepts::AdaptiveControlP<1> > > ctrl1_;
    /// Hash table of the control information for the faces
    std::auto_ptr<__gnu_cxx::hash_map<uint,
				      concepts::AdaptiveControlP<2> > > ctrl2_;
    /// Hash table of the control information for the 3D elements
    __gnu_cxx::hash_map<uint, concepts::AdaptiveControlP<3> > ctrl3_;

    /// List of cells
    std::map<uint, concepts::CellData> cellList_;
    /// List of faces
    std::map<uint, concepts::FaceData> faceList_;
    /// List of edges
    std::map<uint, concepts::EdgeData> edgeList_;
    /// List of vertices
    std::map<uint, concepts::VertexData> vertexList_;

    /// Hash table of the adjustment information of the elements
    __gnu_cxx::hash_map<uint, concepts::AdaptiveAdjustP<3> > adj_;

    /// The S matrices in 1D
    mutable std::auto_ptr<concepts::SMatrix1D> S1left_, S1right_;
    /** The S matrices in 3D for the subdivision into 8 children
        @see concepts::Hex3dSubdiv8
    */
    mutable std::auto_ptr<concepts::SMatrixBase<Real> > Smatrices8_[8];
    /** The S matrices in 3D for the subdivision into 4 children
        @see concepts::Hex3dSubdiv4x concepts::Hex3dSubdiv4y concepts::Hex3dSubdiv4z
    */
    mutable std::auto_ptr<concepts::SMatrixBase<Real> > Smatrices4x_[4],
      Smatrices4y_[4], Smatrices4z_[4];
    /** The S matrices in 3D for the subdivision into 2 children
        @see concepts::Hex3dSubdiv2x concepts::Hex3dSubdiv2y concepts::Hex3dSubdiv2z
    */
    mutable std::auto_ptr<concepts::SMatrixBase<Real> > Smatrices2x_[2],
      Smatrices2y_[2], Smatrices2z_[2];

    /// Place to store timing values
    static concepts::InOutParameters* timings_;
    /// Counter for timing table
    static uint timeCntr_;

    /** Adjusts the mesh and the polynomial degree.

        On the old mesh, the adjustment information is taken. This leads
	to mesh refinements or coarsening and a certain polynomial degree
	distribution.

	The polynomial degree is fixed on the finest level (where the
	elements are eventually built) and given to the
	coarser levels (for faces, edges and the interior of the cell).
	This is done <b>bottom up</b>.

	Finally, the boundary conditions are enforced on every cell using
	\c enforceBC_

        @param cell Cell
	@param L The desired level
	@param P The desired polynomial degree
	@pre \c P is an array of 3 integers
	@see enforceBC_
     */
    void meshAndPoly_(concepts::Hexahedron3d& cell,
		      concepts::Level<3>& L, int* P);

    /** Enforce Dirichlet boundary conditions.
	This routine is called by \c meshAndPoly_ on every cell.
	@param cell Cell to work on
    */
    void enforceBC_(concepts::Hexahedron3d& cell);

    /** Initially fills \c cellList_ (recusively). Up to the active
	level, all cells are stored in the list with their key as list
	key. The father pointer in the \c CellData structure is set
	during this process.
	@param cell Cell to work on
	@param father Parent cell of \c cell
    */
    void createCellList_(const concepts::Connector& cntr,
			 const concepts::CellData* father);
    /** Initially fills \c faceList_ and \c edgeList_. In \c faceList_
	every face of all active cells is stored together with a
	pointer to its cells. In \c edgeList_ every edge of all active
	cells is stored together with pointers to its cells and
	faces.
	@param cell Cell to work on
    */
    void createEdgFaceList_(const concepts::Cell3& cell);
    /** Initially fills \c vertexList_. In \c vertexLIst_ every vertex
	of all active cells is stored together with pointers to its
	cells and edges.
	@param cell Cell to work on
    */
    void createVtxEdgList_(const concepts::Cell3& cell);

    /** Enriches the polynomial degrees. This enrichement is necessary for the
	edges and faces to be able to create conforming global basis functions.

	In refined cells, some of the small edges and faces might have
	a polynomial degree which is too low to represent a part
	of a basis function which is defined on a larger element.

	To achieve this, the cells are traversed <b>top down</b> until
	the level of the elements is reached. During the traversal,
	the needed polynomial degree for the small elements is
	computed and enforced on the element level (only on the edges
	and faces!).

	@pre \c p is an array of 27 integers
    */
    void enrichElm_(const concepts::Hexahedron3d& cell, int* p);

    /** Creates the elements and their T matrices.

        This routine is called recursively until the finest level is
	reached. If there have degrees of freedom to be assigned on
	higher level, the according T columns are adapted using the S
	matrices. On the finest level, the elements are created and
	the T matrix is created from the T columns.

	@param cell The current cell
	@param Pmax Request for maximal polynomial degree, array with 3 entries
	@param T0 Precomputed T columns which belong to a parent of the cell
	@pre \c Pmax is an array of 3 integers
     */
    void buildElements_(concepts::Hexahedron3d& cell,
			ushort* Pmax,
			concepts::TColumn<Real>* T0 = 0);

    /** Computes the maximal polynomial degree of \c cell in each direction.
	Takes into account higher polynomial degrees on faces and edges
	compared to the interior.
    */
    void computePmax_(const concepts::Hexahedron3d& cell,
		      ushort Pmax[3]) const;

    /** Determines wether a shape function direction has to flip the
	sign for the odd degree polynomials.
	@param cntr Connector
	@param i Index of the face [0..5]
	@param flip0 Return value, the first direction has to flip
	@param flip1 Return value, the second direction has to flip
    */
    void faceShapefctOrientation_(const concepts::Hexahedron& cntr,
				  const uint i,
				  bool& flip0, bool& flip1) const;

    /** Deactivates the children of \c face
	(including the new edges and vertices).
    */
    void deactivate_(const concepts::Connector2& face);
    /** Deactivates the children of the edge \c edg
	(including the middle vertex).
    */
    void deactivate_(const concepts::Connector1& edg);
    /// Deactivates the children of the vertex \c vtx
    void deactivate_(const concepts::Connector0& vtx);
    /** Does apply the correct refinement based on the requested level
	change. If the refinement is not readly applicable, upgraded
	refinements are tested to circumvent a geometric deadlock problem.

	This routine is called by <tt>meshAndPoly_</tt>.
    */
    void doCorrectRefinement_(concepts::Hexahedron3d& cell,
			      concepts::Level<3>& L) const;
  };

  uint Space::dim() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return dim_; 
  }

  uint Space::nelm() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return nelm_; 
  }

  Space::Scan* Space::scan() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return new concepts::PListScan<Element<Real> >(*elm_);
  }

  uint Space::dim() {
    if (rebuild_) rebuild();
    return dim_; 
  }

  uint Space::nelm() {
    if (rebuild_) rebuild();
    return nelm_; 
  }

  Space::Scan* Space::scan() {
    if (rebuild_) rebuild();
    return new concepts::PListScan<Element<Real> >(*elm_);
  }

} // namespace hp3d

#endif // hpElement3d_h
