/* space class for 3D hp FEM
 * buildElements_
 */

#include <memory>
#include "hp3D/space.hh"
#include "hp3D/element.hh"
#include "hp3D/hexahedron.hh"
#include "hexFunctions.hh"
#include "basics/debug.hh"

#define hpSpace3DBuildElements_D 0
#define hpSpace3DPmax_D 0
#define hpSpace3Ddof_D 0

// careful with those! (there are dependencies faces -> edges)
#define buildVertexModes 1
#define buildEdgeModes 1
#define buildFaceModes 1
#define buildInnerModes 1

namespace hp3D {

  // ***************************************************************** Space **

  void Space::computePmax_(const concepts::Hexahedron3d& cell,
			   ushort Pmax[3]) const {
    const concepts::Hexahedron& cntr = cell.connector();
    const concepts::AdaptiveControlP<3>&
      ctrl3(ctrl3_.find(cntr.key())->second);
    DEBUGL(hpSpace3DPmax_D, "start - " << cell << ", Pmax = ("
	   << Pmax[0] << ", " << Pmax[1] << ", " << Pmax[2] << ')');

    // start with internal polynomial degree
    for (uint j = 0; j < 3; ++j)
      if (ctrl3.p_[j] > Pmax[j]) Pmax[j] = ctrl3.p_[j];
    DEBUGL(hpSpace3DPmax_D, "int , Pmax = (" << Pmax[0] << ", " << Pmax[1]
	   << ", " << Pmax[2] << ')');

    // get maximum of edges
    const concepts::Edge* cntr1 = 0;
    for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
      const concepts::AdaptiveControlP<1>&
	ctrl1(ctrl1_->find(cntr1->key())->second);
      uint pIndex = 0, qIndex = 0, rIndex = 0;
      HexFunctions::edgeOrientation(i, pIndex, qIndex, rIndex);
      if (Pmax[pIndex] < ctrl1.p_[0]) Pmax[pIndex] = ctrl1.p_[0];
    } // for edges (i)
    DEBUGL(hpSpace3DPmax_D, "edge, Pmax = (" << Pmax[0] << ", " << Pmax[1]
	   << ", " << Pmax[2] << ')');

    // get maximum of faces
    const concepts::Connector2* cntr2 = 0;
    for (uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i) {
      const concepts::AdaptiveControlP<2>&
	ctrl2(ctrl2_->find(cntr2->key())->second);
      DEBUGL(hpSpace3DPmax_D, "face " << i << " has " << ctrl2);
      uint pIndex = 0, qIndex = 0, rIndex = 0;
      HexFunctions::faceOrientation(cntr, i, pIndex, qIndex, rIndex);
      if (Pmax[pIndex] < ctrl2.p_[0]) Pmax[pIndex] = ctrl2.p_[0];
      if (Pmax[qIndex] < ctrl2.p_[1]) Pmax[qIndex] = ctrl2.p_[1];
      DEBUGL(hpSpace3DPmax_D, "orientation = (" << pIndex << ", " << qIndex
	     << "), Pmax = (" << Pmax[0] << ", " << Pmax[1]
	     << ", " << Pmax[2] << ')');
    } // for faces (i)
    DEBUGL(hpSpace3DPmax_D, "face, Pmax = (" << Pmax[0] << ", " << Pmax[1]
	   << ", " << Pmax[2] << ')');
    DEBUGL(hpSpace3DPmax_D, "done.");
  }

  void Space::buildElements_(concepts::Hexahedron3d& cell,
			     ushort* Pmax,
			     concepts::TColumn<Real>* T0) {
//     uint hpSpace3DBuildElements_D = 0;
//     if ((int)cell.connector().key() == 2)
//       hpSpace3DBuildElements_D = 1;
    DEBUGL(hpSpace3DBuildElements_D, cell.connector().key() << " - dim = "
	   << dim_ << ", start.");

    concepts::Hexahedron& cntr = cell.connector();
    concepts::AdaptiveControlP<3>& ctrl3(ctrl3_[cntr.key()]);
    DEBUGL(hpSpace3Ddof_D, cntr);

    // compute p max in this element
    computePmax_(cell, Pmax);
    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - p max = ("
	   << Pmax[0] << ", " << Pmax[1] << ", " << Pmax[2] << ')');

    // maximal number of degrees of freedom in this element
    uint nn[3] = { Pmax[0]+1, Pmax[1]+1, Pmax[2]+1 }; // p max
    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - n = "
	   << nn[0]*nn[1]*nn[2] << ", cell tag = " << ctrl3.tag_);
    concepts::TColumn<Real>* T1 = 0;
    concepts::TColumnTensor<Real, 3>* TTensor = 0;

    Element<Real>* elm = 0;
    Hexahedron* elmH = 0;
    std::auto_ptr<Element<Real> > elmTmp(0); // used for temporary elements
    if (ctrl3.tag_.isMember()) {
      elm = elmH = new Hexahedron(cell, Pmax, T0, T1); // p max

      // connect polynomial degrees with element
      const concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i)
	elmH->edgeP(i, (*ctrl1_)[cntr1->key()]);
      concepts::Connector2* cntr2 = 0;
      for(uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i)
	elmH->faceP(i, (*ctrl2_)[cntr2->key()]);

      DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - creating element "
	     << *elm << ", p int = (" << ctrl3.p_[0] << ", " << ctrl3.p_[1]
	     << ", " << ctrl3.p_[2] << ')');
      ++nelm_;
      if (elm_ == 0) {
	// first element in the list
	endElm_ = elm_ =
	  new concepts::Joiner<Element<Real>*, 1>(elm, elm_);
      } else {
	// list already has elements, add to the end of the list
	concepts::Joiner<Element<Real>*, 1>* newElm =
	  new concepts::Joiner<Element<Real>*, 1>(elm, 0);
	(*endElm_)[0] = newElm;
	endElm_ = newElm;
      }
    } else {
      elmTmp.reset(elmH = new Hexahedron(cell, Pmax, T0, T1)); // p max
      elm = elmTmp.get();
      DEBUGL(hpSpace3DBuildElements_D, cntr.key()
	     << " - creating temporary element " << *elm);
    }

    // ll denotes the local degree of freedom in tensor product notation
    uint ll[3] = { 0, 0, 0 };
    // k counts the local degrees of freedom in topological notation
    uint k = 0;
    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - vertices starting, k = "
	   << k);
    // loop over vertices
    concepts::Vertex* cntr0 = 0;
    for(uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i, ++k) {
#if buildVertexModes
      concepts::AdaptiveControl& ctrl0(ctrl0_[cntr0->key()]);
      DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - " << *cntr0
	     << ", tag = " << ctrl0.tag_);
      if (ctrl0.tag_.isMember() || ctrl0.tag_.isPassive()) {
	// is member of the space or passive
	std::map<uint, concepts::VertexData>::const_iterator j =
	  vertexList_.find(cntr0->key().key());
	if (ctrl0.tag_.isMember() && ctrl0.tag_.isActive() &&
	    (j != vertexList_.end()) && (j->second.hasCell(cell.connector()))) {
	  // active member of the space
	  if (! ctrl0.tag_.idxValid()) { 
	    // not yet counted
	    ctrl0.idx_ = dim_; ++dim_;
	    ctrl0.tag_.validateIdx();
	    DEBUGL(hpSpace3Ddof_D, *cntr0 << " has dof " << ctrl0.idx_);
	  }
	  T1 = TTensor
	    = new concepts::TColumnTensor<Real, 3>(nn, ctrl0.idx_, T1);

	  HexFunctions::vertexIndex(i, ll);
	  conceptsAssert(ll[0] < 2, concepts::Assertion());
	  conceptsAssert(ll[1] < 2, concepts::Assertion());
	  conceptsAssert(ll[2] < 2, concepts::Assertion());

 	  DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - active dof: k = "
		 << k << ", ll = (" << ll[0] << ", " << ll[1]
		 << ", " << ll[2] << "), vtx = " << cntr0->key()
		 << ", idx = " << ctrl0.idx_);
	  (*TTensor)[ll] = 1.0;
	} // if member of the space and active
	if (! ctrl3.tag_.isMember()) deactivate_(*cntr0);
      } // if member of the space or passive
#endif
    } // for vertices

    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - edges starting, k = "
	   << k);
    // loop over edges
    concepts::Edge* cntr1 = 0;
    for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
#if buildEdgeModes
      concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[cntr1->key()]);
      DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - " << *cntr1
	     << ", tag = " << ctrl1.tag_);
      if (ctrl1.tag_.isMember() || ctrl1.tag_.isPassive()) {
	std::map<uint, concepts::EdgeData>::const_iterator j =
	  edgeList_.find(cntr1->key().key());
	if (ctrl1.tag_.isMember() && ctrl1.tag_.isActive() &&
	    (j != edgeList_.end()) && (j->second.hasCell(cell.connector()))) {
	  // active member
	  if (ctrl1.p_[0] > 1) {
	    // has dofs
	    if (! ctrl1.tag_.idxValid()) {
	      // not yet counted
	      ctrl1.idx_ = dim_; dim_ += ctrl1.p_[0] - 1;
	      ctrl1.tag_.validateIdx();
	      DEBUGL(hpSpace3Ddof_D, *cntr1 << " has dof " << ctrl1.idx_);
	    }
	    Real s = 0.0;

	    // assume we have a hex
	    // get sign flip for the entry in the T matrix
	    conceptsAssert(i < 12, concepts::Assertion());
	    concepts::Vertex* vtx = cntr1->vertex(0);
	    switch (i) {
	    case 0: case 3: case 4:
	      if (cntr.vertex(0) == vtx) s = 1.0; else s = -1.0; break;
	    case 1: case 5:
	      if (cntr.vertex(1) == vtx) s = 1.0; else s = -1.0; break;
	    case 2: case 7:
	      if (cntr.vertex(3) == vtx) s = 1.0; else s = -1.0; break;
	    case 6:
	      if (cntr.vertex(2) == vtx) s = 1.0; else s = -1.0; break;
	    case 8: case 11:
	      if (cntr.vertex(4) == vtx) s = 1.0; else s = -1.0; break;
	    case 9:
	      if (cntr.vertex(5) == vtx) s = 1.0; else s = -1.0; break;
	    case 10:
	      if (cntr.vertex(7) == vtx) s = 1.0; else s = -1.0; break;
	    default: throw conceptsException(concepts::ExceptionBase());
	    }

	    // get index information for the T matrix
	    uint pIndex = 0, qIndex = 0, rIndex = 0;
	    HexFunctions::edgeOrientation(i, pIndex, qIndex, rIndex);
	    HexFunctions::edgeIndex(i, qIndex, rIndex, ll);
	    conceptsAssert(pIndex != rIndex, concepts::Assertion());
	    conceptsAssert(qIndex != rIndex, concepts::Assertion());
	    conceptsAssert(pIndex != qIndex, concepts::Assertion());

	    Real t = 1.0;
	    uint idx = ctrl1.idx_;
 	    for (int j = 0; j < ctrl1.p_[0] - 1; ++j) {
	      T1 = TTensor
		= new concepts::TColumnTensor<Real, 3>(nn, idx++, T1);
	      ll[pIndex] = j + 2;
	      (*TTensor)[ll] = t;
	      t *= s;
	      DEBUGL(hpSpace3DBuildElements_D, cntr.key()
		     << " - active dof: k = "
		     << k+j << ", ll = (" << ll[0] << ", " << ll[1]
		     << ", " << ll[2] << "), p = " << ctrl1.p_[0]
		     << ", idx = " << ctrl1.idx_+j);
	    } // for j
	  } // p > 1
	} // active
	if (! ctrl3.tag_.isMember()) deactivate_(*cntr1);
      }
#endif
      uint pIndex = 0, qIndex = 0, rIndex = 0;
      HexFunctions::edgeOrientation(i, pIndex, qIndex, rIndex);
      k += Pmax[pIndex] - 1; // p max
    } // for i

    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - faces starting, k = "
	   << k);
    // loop over faces
    concepts::Connector2* cntr2 = 0;
    for (uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i) {
#if buildFaceModes
      concepts::AdaptiveControlP<2>& ctrl2((*ctrl2_)[cntr2->key()]);
      if (ctrl2.tag_.isMember() || ctrl2.tag_.isPassive()) { 
	std::map<uint, concepts::FaceData>::const_iterator j =
	  faceList_.find(cntr2->key().key());
	if (ctrl2.tag_.isMember() && ctrl2.tag_.isActive() &&
	    (j != faceList_.end()) && (j->second.hasCell(cell.connector()))) {
	  // active member
	  DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - treating face "
		 << i << " = " << *cntr2 << ", ctrl = " << ctrl2);
	  if ((ctrl2.p_[0] > 1) && (ctrl2.p_[1] > 1)) {
	    // has dofs
	    int p0_1 = ctrl2.p_[0] - 1;
	    int p1_1 = ctrl2.p_[1] - 1;
	    bool addToDim = false;
	    if (! ctrl2.tag_.idxValid()) {
	      // not yet counted
	      ctrl2.idx_ = dim_;
	      addToDim = true;
	      ctrl2.tag_.validateIdx();
	      DEBUGL(hpSpace3Ddof_D, *cntr2 << " has dof " << ctrl2.idx_);
	    }

	    // check if signs need to be flipped
	    bool flip0, flip1;
	    faceShapefctOrientation_(cntr, i, flip0, flip1);
// 	    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - flip0 = "
// 		   << flip0 << ", flip1 = " << flip1);

	    // get as much index information as possible
	    uint pIndex = 0, qIndex = 0, rIndex = 0;
	    HexFunctions::faceOrientation(cntr, i, pIndex, qIndex, rIndex);
	    HexFunctions::faceIndex(i, rIndex, ll);
	    conceptsAssert(pIndex != rIndex, concepts::Assertion());
	    conceptsAssert(qIndex != rIndex, concepts::Assertion());
	    conceptsAssert(pIndex != qIndex, concepts::Assertion());

	    // assume a quad as face
	    unsigned int idx = ctrl2.idx_;
	    for (int p = 0; p < p0_1; ++p) {
	      for (int q = 0; q < p1_1; ++q) {
		if (!trunk_ || ((p + q) <= ((p0_1 > p1_1) ? p0_1 : p1_1) + 1) ) { // trunk space
		  if (addToDim) ++dim_;
		  T1 = TTensor
		    = new concepts::TColumnTensor<Real, 3>(nn, idx++, T1);

		  ll[pIndex] = p + 2;
		  ll[qIndex] = q + 2;

		  // sign of the shape function
		  Real s = 1.0;
		  if ( ( ((int)flip0)*p + ((int)flip1)*q ) % 2 )
		    s = -1.0;

		  (*TTensor)[ll] = s;
		  DEBUGL(hpSpace3DBuildElements_D, cntr.key()
			 << " - active dof: k = "
			 << k+p*p1_1+q << ", ll = (" << ll[0] << ", "
			 << ll[1] << ", " << ll[2] << "), p = (" << ctrl2.p_[0]
			 << ", " << ctrl2.p_[1] << "), idx = " << ctrl2.idx_
			 << ", s = " << s);
		} // if "trunk"
	      } // for q
	    } // for p
	  } // p > 1
	} // active member
	if (! ctrl3.tag_.isMember()) deactivate_(*cntr2);
      }
#endif
      uint pIndex = 0, qIndex = 0, rIndex = 0;
      HexFunctions::faceOrientation(cntr, i, pIndex, qIndex, rIndex);
      k += ( Pmax[pIndex] - 1)*( Pmax[qIndex] - 1); // p max
    } // for i

    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - interior starting, k = "
	   << k);
    if (ctrl3.tag_.isMember()) {
      if ( (ctrl3.p_[0] > 1) && (ctrl3.p_[1] > 1) && (ctrl3.p_[2] > 1) ) { // p int
	ctrl3.idx_ = dim_;
	ctrl3.tag_.validateIdx();
	DEBUGL(hpSpace3Ddof_D, cntr << " has dof " << ctrl3.idx_);

	uint idx = ctrl3.idx_;
	uint PintMax = ctrl3.p_[0] > ctrl3.p_[1] ? ctrl3.p_[0] : ctrl3.p_[1]; // pint
	PintMax = PintMax > ctrl3.p_[2] ? PintMax : ctrl3.p_[2]; // p int
	DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - " << PintMax
	       << " is max of (" << ctrl3.p_[0] << ", " << ctrl3.p_[1] << ", "
	       << ctrl3.p_[2] << ')');
	for (uint& i = ll[0] = 2; i < (uint)ctrl3.p_[0]+1; ++i) { // p int
	  for (uint& j = ll[1] = 2; j < (uint)ctrl3.p_[1]+1; ++j) { // p int
	    for (uint& l = ll[2] = 2; l < (uint)ctrl3.p_[2]+1; ++l, ++k) { // p int
#if buildInnerModes
	      if (!trunk_ || ((i + j + l) <= (PintMax + 1)) ) {
		++dim_;
		T1 = TTensor
		  = new concepts::TColumnTensor<Real, 3>(nn, idx++, T1);
		(*TTensor)[ll] = 1.0;
		DEBUGL(hpSpace3DBuildElements_D, cntr.key()
		       << " - active dof: k = " << k << ", ll = (" << ll[0]
		       << ", " << ll[1] << ", " << ll[2] << ')');
	      } // if "trunk"
#endif
	    } // for l
	  } // for j
	} // for i
      }
      elm->appendT(T1);
      DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - added T columns to "
	     << *elm);
    } else {
      if (T0 || T1)
	recomputeSmatrices(elmH);
      // check subdivision strategy
      const concepts::Hex3dSubdivision* subdiv = elmH->getStrategy();
      const std::auto_ptr<concepts::SMatrixBase<Real> >* Smatrices =
	getSmatrices(subdiv);
      // loop over the children of the cell
      concepts::Hexahedron3d* child = 0;
      for (uint i = 0; (child = cell.child(i)) != 0; ++i) {
#ifdef DEBUG
	concepts::AdaptiveControlP<3>& ctrli(ctrl3_[child->connector().key()]);
#endif
	DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - child " << i
	       << " = " << *child << " with " << ctrli);
	// number of degrees of freedom of the child
	ushort Pmaxi[3] = { 0, 0, 0 };
	if ((T0 != 0) || (T1 != 0))
	  for (uint i = 0; i < 3; ++i) Pmaxi[i] = Pmax[i];
	computePmax_(*child, Pmaxi);
	uint nni[3] = { Pmaxi[0]+1, Pmaxi[1]+1, Pmaxi[2]+1 }; // p max
	DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - child has p max = ("
	       << Pmaxi[0] << ", " << Pmaxi[1] << ", " << Pmaxi[2] << ')');

	// apply S matrices to the already computed columns of the T matrix
	concepts::TColumn<Real>* T2 = 0;
	for (concepts::TColumn<Real>* TT = T0; TT != 0; TT = TT->link()) {
	  DEBUGL(hpSpace3DBuildElements_D, cntr.key()
		 << " - from T0, treating " << *TT);
	  T2 = new concepts::TColumnTensor<Real, 3>(nni, TT->index(), T2);
	  (*Smatrices[i])(*TT, *T2);
	}
	for (concepts::TColumn<Real>* TT = T1; TT != 0; TT = TT->link()) {
	  T2 = new concepts::TColumnTensor<Real, 3>(nni, TT->index(), T2);
	  (*Smatrices[i])(*TT, *T2);
	  DEBUGL(hpSpace3DBuildElements_D, cntr.key()
		 << " - from T1, treating "
		 << *TT << ", stored in " << *T2);
	}

	// buildElements_ is called with the modified columns of the T matrix
	buildElements_(*child, Pmaxi, T2);

	// clean up
	while (T2 != 0) {
	  concepts::TColumn<Real>* foo = T2; T2 = T2->link(); delete foo;
	}
      }
    }

    while (T1 != 0) {
      concepts::TColumn<Real>* foo = T1; T1 = T1->link(); delete foo;
    }

    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - dim = " << dim_
	   << ", done.");
  }

} // namespace hp3d
