/* Functor class for sum factorization
 */

#ifndef sumfact_hh
#define sumfact_hh

#include "toolbox/array.hh"

namespace concepts {

  // forward declarations
  template<typename T>
  class ElementMatrix;

  class InOutParameters;
} // namespace concepts

namespace hp3D {

  using concepts::Real;

  // forward declarations
  template<typename F>
  class ShapeFunction3D;

  // ****************************************************** SumFactorization **

  /** Sum factorization for an element matrix.

      This class is equiped with an interface to get timings of internal
      computations if compiled accordingly (see \c bilinearForm.cc file), see
      \c setTimings() and \c timings().

      @todo Implement changing order of sum factorization, possibility is there
      @bug If the loops are reordered using \c a, \c b and \c c on line 58,
           \c coeff (by use of \c coeffTmp) is not accessed in the correct
	   order in the first loop to fill \c sumfact1_.
      @author Philipp Frauenfelder, 2002
  */
  class SumFactorization {
  public:
    /** Application operator.
	Adds the contribution of \c coeff to \c em.
	@param em Element matrix
	@param jacobian Jacobian
	@param coeff Coefficients for the integration
	@param u Left shape functions
	@param v Right shape functions
    */
    template<typename T>
    void operator()(concepts::ElementMatrix<Real>& em,
		    const concepts::Array<Real>& jacobian, T* coeff,
		    const ShapeFunction3D<Real>& u,
		    const ShapeFunction3D<Real>& v);

    /** @name Timing Interface
     *  These functions are used to get timings from class internal
     *  computations. The values are stored in a user defined
     *  concepts::InOutParameters structure in different arrays (see 
     *  \c setTimings). These arrays can be grouped into a table for
     *  easier postprocessing with
     *  \code
    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::DOUBLE, "jacobian", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "whole_sumfact", output);
    std::ofstream ofs("table.gnuplot");
    ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(ofs);
        \endcode
     */
    //@{
    /** Sets the class to store the timing values in. Additionally,
	the \c timeCntr_ is reset to 0. This counter is used to fill
	in the values into the arrays listed below in subsequent calls.
	The following timings are taken and stored in \c timings:
	<ul>
	<li>computation of the first intermediate matrix in \c first_temp
	<li>computation of the second intermediate matrix in \c second_temp
	<li>computation of the element matrix in \c fill_em
	</ul>
    */
    static void setTimings(concepts::InOutParameters* timings);
    /** Returns true if the class is able to do timings.  The ability
	to do timings depends on a compiler switch in \c sumfact.cc file.
    */
    static bool timings();
    //@}
  private:
    /// Intermediate matrices for sum factorization
    concepts::Array<Real> sumfact1_, sumfact2_;

    /// Place to store timing values
    static concepts::InOutParameters* timings_;
    /// Counter for timing table
    static uint timeCntr_;
  };

} // namespace hp3D


#endif // sumfact_hh
