/* Expansion bases for tensorial shape functions not only for
 * quads and hexahedrons but also for triangles and tetrahedrons.
 * Due to Karniadakis, Sherwin
 */

#include <cmath>
#include "karniadakis.hh"
#include "basics/debug.hh"
#include "basics/exceptions.hh"
#include "toolbox/array.hh"
#include "quadratureBase.hh"

// debugging
#define KarnConstr_D 0

namespace __gnu_cxx {

  // **************************************************************** Orders **

  template<int number>
  size_t hash_value(const concepts::Orders<number>& o) {
    size_t res = o.P();
    for (uint i = 0; i < number; ++i) {
      res = (res << 4) | o.p(i);
    }
    return res;
  }

} // namespace __gnu_cxx

namespace concepts {

  template<int number>
  Orders<number>::Orders() : OrdersBase() {
    for (uint i = 0; i < number; ++i) p_[i] = 0;
  }

  template<int number>
  Orders<number>::Orders(const int p0, const int P,
			 const int p1, const int p2) : OrdersBase(P) {
    if (number > 0) p_[0] = p0;
    if (number > 1) p_[1] = p1;
    if (number > 2) p_[2] = p2;
  }

  template class Orders<3>;
  template class Orders<2>;
  template class Orders<1>;

  // ********************************************************** Karniadakis **

  template<int type, int mode>
  __gnu_cxx::hash_map<Orders<type>, Real*, __gnu_cxx::hash<Orders<type> >,
		      __gnu_cxx::OrdersEqual<type> >
  Karniadakis<type, mode>::principal_H =
  __gnu_cxx::hash_map<Orders<type>, Real*,
		      __gnu_cxx::hash<Orders<type> >,
		      __gnu_cxx::OrdersEqual<type> >();

  template<int type, int mode>
  Karniadakis<type, mode>::Karniadakis(const int P, const Real* xP, 
				       const int NxP, const int Q, 
				       const int R, const bool cache) :
    ShapeFunction1D<Real>((P+1)*(Q+1)*(R+1)+1, NxP), orders_(P, NxP, Q, R),
    cache_(cache) {

    DEBUGL(KarnConstr_D, "orders: " << orders_);
    Assert(1 == 0, Assertion());
  }

  template<> // psi^a, normal
  Karniadakis<1, 0>::Karniadakis(const int P, const Real* xP, 
				 const int NxP, const int Q, const int R,
				 const bool cache) :
    ShapeFunction1D<Real>((P+1)
//  			  +1
			  , NxP), orders_(P, NxP, Q, R), cache_(cache) {

    DEBUGL(KarnConstr_D, "orders: " << orders_ << ", "
	   << (cache_ ? "" : "not ") << "cached");
    __gnu_cxx::hash_map<Orders<1>, Real*, __gnu_cxx::hash<Orders<1> >,
      __gnu_cxx::OrdersEqual<1> >::iterator i;
    i = principal_H.find(orders_);
    if (cache_ && (i != principal_H.end())) {
      // values already computed
      values_ = i->second;
      DEBUGL(KarnConstr_D, "found values of principal function for orders "
	     << orders_);
    } else {
      // need to compute the values
      values_ = new Real[(P+1
//  			  +1
			  )*NxP];
      if (cache_)
	principal_H.insert(std::pair<Orders<1>, Real*>(orders_, values_));

      Real* jacobitmp = 0;
      if (P >= 2) {
	jacobitmp = new Real[(P-1)*NxP];
	JacobiPol(1.0, 1.0, P-2, xP, NxP, jacobitmp);

	for (int i = 1; i < P; ++i) {
	  // loop over all fcts.
	  for (int p = 0; p < NxP; ++p) {
	    // loop over all points
	    values_[(i+1)*NxP + p] = jacobitmp[p*(P-1) + i-1]
	      * (1.0-xP[p])/2.0 * (1.0+xP[p])/2.0;
	  }
	}
      }
      // loop over all points
      for (int p = 0; p < NxP; ++p) {
	// i = 0
	values_[p] = (1.0-xP[p])/2.0;
	// i = 1
	values_[NxP + p] = (1.0+xP[p])/2.0;
	// i = P_+1: f(.) \equiv 1
//  	values_[(P+1)*NxP + p] = 1.0;
      }
      delete[] jacobitmp;
    
#if KarnConstr_D
      DEBUGL(KarnConstr_D, "computed values of principal function for orders "
	     << orders_);
      for (int i = 0; i < P+1
//  	     +1
	     ; ++i) {
	cout << "  ";
	for (int j = 0; j < NxP; ++j) {
	  cout << values_[i*(NxP)+j];
	  if (j == NxP-1)
	    cout << endl;
	  else
	    cout << ", ";
	}
      }
#endif
    }
  }

  template<> // psi^a, derivatives
  Karniadakis<1, 1>::Karniadakis(const int P, const Real* xP,
				 const int NxP, const int Q, const int R,
				 const bool cache) :
    ShapeFunction1D<Real>((P+1)
//  			  +1
			  , NxP), orders_(P, NxP, Q, R), cache_(cache) {

    __gnu_cxx::hash_map<Orders<1>, Real*, __gnu_cxx::hash<Orders<1> >,
      __gnu_cxx::OrdersEqual<1> >::iterator i;
    i = principal_H.find(orders_);
    if (cache_ && (i != principal_H.end())) {
      // values already computed
      values_ = i->second;
      DEBUGL(KarnConstr_D, "found values of principal function for orders "
	     << orders_);
    } else {
      // need to compute the values
      values_ = new Real[(P+1
//  			  +1
			  )*NxP];
      if (cache_)
	principal_H.insert(std::pair<Orders<1>, Real*>(orders_, values_));

      Real* jacobitmp = 0;
      Real* jacobitmpD = 0;
      if (P >= 2) {
	jacobitmp = new Real[(P-1)*NxP];
	JacobiPol(1.0, 1.0, P-2, xP, NxP, jacobitmp);

	if (P >= 3) {
	  jacobitmpD = new Real[(P-2)*NxP];
	  JacobiPol(2.0, 2.0, P-3, xP, NxP, jacobitmpD);
	}

	// loop over all points
	for (int p = 0; p < NxP; ++p) {
	  if (P > 1) // i = 2
	    values_[2*NxP + p] =
	      jacobitmp[p*(P-1)] * (1.0-xP[p])/2.0 * 0.5
	      - jacobitmp[p*(P-1)] * (1.0+xP[p])/2.0 * 0.5;

	  // loop over all fcts.
	  for (int i = 2; i < P; ++i) {
	    values_[(i+1)*NxP + p] =
	      jacobitmp[p*(P-1) + i-1] * (1.0-xP[p])/2.0 * 0.5
	      - jacobitmp[p*(P-1) + i-1] * (1.0+xP[p])/2.0 * 0.5
	      + (jacobitmpD[p*(P-2) + i-2]*0.5*(2.0+i) 
		 * (1.0-xP[p])/2.0 * (1.0+xP[p])/2.0);
	  }
	}
      }
      // loop over all points
      for (int p = 0; p < NxP; ++p) {
	// i = 0
	values_[p] = -0.5;
	// i = 1
	values_[NxP + p] = 0.5;
	// i = P_+1: f(.)' \equiv 0
//  	values_[(P+1)*NxP + p] = 0.0;
      }
      delete[] jacobitmp;
    
#if KarnConstr_D
      DEBUGL(KarnConstr_D, "computed values of principal function for orders "
	     << orders_);
      for (int i = 0; i < P+1
//  	     +1
	     ; ++i) {
	cout << "  ";
	for (int j = 0; j < NxP; ++j) {
	  cout << values_[i*(NxP)+j];
	  if (j == NxP-1)
	    cout << endl;
	  else
	    cout << ", ";
	}
      }
#endif
    }
  }

  template<> // psi^b, normal
  Karniadakis<2, 0>::Karniadakis(const int P, const Real* xP, 
				 const int NxP, const int Q, const int R,
				 const bool cache) :
    ShapeFunction1D<Real>((P+1)*(Q+1)*(R+1), NxP), orders_(P, NxP, Q, R),
    cache_(cache) {

    __gnu_cxx::hash_map<Orders<2>, Real*, __gnu_cxx::hash<Orders<2> >,
      __gnu_cxx::OrdersEqual<2> >::iterator i;
    i = principal_H.find(orders_);
    if (cache_ && (i != principal_H.end())) {
      // values already computed
      values_ = i->second;
      DEBUGL(KarnConstr_D, "found values of principal function for orders " 
	     << orders_);
    } else {
      // need to compute the values
      values_ = new Real[(P+1)*(Q+1)*NxP];
      if (cache_)
	principal_H.insert(std::pair<Orders<2>, Real*>(orders_, values_));

      Array<Real> jacobitmp;
      if (Q >= 2) jacobitmp.resize((Q-1)*NxP);
      for (int i = 0; i < P+1; ++i) {
	if ( (i == 0) || (i == 1) ) {
	  if (Q >= 2)
	    JacobiPol(1.0, 1.0, Q-2, xP, NxP, (Real*)jacobitmp);
	  for (int j = 0; j < Q+1; ++j) {
	    for (int p = 0; p < NxP; ++p) {
	      if (j == 0) {
		values_[i*NxP*(Q+1) + j*NxP + p] = (1.0-xP[p])/2.0;
		continue;
	      }
	      if (j == 1) {
		values_[i*NxP*(Q+1) + j*NxP + p] = (1.0+xP[p])/2.0;
		continue;
	      }
	      values_[i*NxP*(Q+1) + j*NxP + p] = (1.0-xP[p])/2.0 * 
		(1.0+xP[p])/2.0 * jacobitmp[p*(Q-1) + j-2];
	    } // for p
	  } // for j
	} else {
	  if (Q >= 2) JacobiPol(2.0*(i-1)+1.0, 1.0, Q-2, xP, NxP, 
					  (Real*)jacobitmp);
	  // j = 0
	  for (int p = 0; p < NxP; ++p) {
	    values_[i*NxP*(Q+1) + p] = std::pow((1.0-xP[p])/2.0, (Real)i);
	  }
	  // rest of the functions indexed by j
	  for (int j = 1; j < Q; ++j) {
	    for (int p = 0; p < NxP; ++p) {
	      values_[i*NxP*(Q+1) + j*NxP + p] =
		std::pow((1.0-xP[p])/2.0, (Real)i)
		* (1.0+xP[p])/2.0 * jacobitmp[p*(Q-1) + j-1];
	    } // for p
	  } // for j
	} // if ... else ...
      } // for i
#if KarnConstr_D
      DEBUGL(KarnConstr_D, "computed values of principal function for orders "
	     << orders_);
      for (int p = 0; p < P+1; ++p) {
	for (int i = 0; i < Q+1; ++i) {
	  if (i == 0) cout << "p = " << p << "\t";
	  else cout << "\t";
	  for (int j = 0; j < NxP; ++j) {
	    cout << values_[p*NxP*(Q+1) + i*NxP + j];
	    if (j == NxP-1)
	      cout << endl;
	    else
	      cout << ", ";
	  }
	}
      }
#endif
    }
  }

  template<> // psi^b, derivatives
  Karniadakis<2, 1>::Karniadakis(const int P, const Real* xP, 
				 const int NxP, const int Q, const int R,
				 const bool cache) :
    ShapeFunction1D<Real>((P+1)*(Q+1)*(R+1), NxP), orders_(P, NxP, Q, R),
    cache_(cache) {

    __gnu_cxx::hash_map<Orders<2>, Real*, __gnu_cxx::hash<Orders<2> >,
      __gnu_cxx::OrdersEqual<2> >::iterator i;
    i = principal_H.find(orders_);
    if (cache_ && (i != principal_H.end())) {
      // values already computed
      values_ = i->second;
      DEBUGL(KarnConstr_D, "found values of principal function for orders " 
	     << orders_);
    } else {
      // need to compute the values
      values_ = new Real[(P+1)*(Q+1)*NxP];
      if (cache_)
	principal_H.insert(std::pair<Orders<2>, Real*>(orders_, values_));

      Array<Real> jacobitmp, jacobitmpD;
      if (Q >= 2) jacobitmp.resize((Q-1)*NxP);
      if (Q >= 3) jacobitmpD.resize((Q-2)*NxP);
      for (int i = 0; i < P+1; ++i) {
	if ( (i == 0) || (i == 1) ) {
	  if (Q >= 2)
	    JacobiPol(1.0, 1.0, Q-2, xP, NxP, (Real*)jacobitmp);
	  if (Q >= 3)
	    JacobiPol(2.0, 2.0, Q-3, xP, NxP, (Real*)jacobitmpD);
	  for (int j = 0; j < Q+1; ++j) {
	    for (int p = 0; p < NxP; ++p) {
	      if (j == 0) {
		values_[i*NxP*(Q+1) + j*NxP + p] = -0.5;
		continue;
	      }
	      if (j == 1) {
		values_[i*NxP*(Q+1) + j*NxP + p] = 0.5;
		continue;
	      }
	      values_[i*NxP*(Q+1) + j*NxP + p] = 
		jacobitmp[p*(Q-1) + j-2] * (1.0-xP[p])/2.0 * 0.5
		- jacobitmp[p*(Q-1) + j-2] * (1.0+xP[p])/2.0 * 0.5;
	      if (j > 2)
		values_[i*NxP*(Q+1) + j*NxP + p] +=
		  jacobitmpD[p*(Q-2) + j-3] * 0.5 * (1.0+j)
		  * (1.0-xP[p])/2.0 * (1.0+xP[p])/2.0;
	    } // for p
	  } // for j
	} else {
	  if (Q >= 2) JacobiPol(2.0*(i-1)+1.0, 1.0, Q-2, xP, NxP, 
					  (Real*)jacobitmp);
	  if (Q >= 3) JacobiPol(2.0*(i-1)+2.0, 2.0, Q-3, xP, NxP,
					  (Real*)jacobitmpD);
	  // j = 0
	  for (int p = 0; p < NxP; ++p) {
	    values_[i*NxP*(Q+1) + p] =
	      -std::pow((1.0-xP[p])/2.0, (Real)(i-1)) * i * 0.5;
	  }
	  // rest of the functions indexed by j
	  for (int j = 1; j < Q; ++j) {
	    for (int p = 0; p < NxP; ++p) {
	      values_[i*NxP*(Q+1) + j*NxP + p] = 
		(0.5 * std::pow((1.0-xP[p])*0.5, (Real)i) *
		 jacobitmp[p*(Q-1) + j-1])
		-
		(std::pow((1.0-xP[p])*0.5, (Real)(i-1)) * i * 0.5
		 * (xP[p]+1.0)*0.5 * jacobitmp[p*(Q-1) + j-1])
		+
		((j > 1)
		 ?
		 std::pow((1.0-xP[p])*0.5, (Real)i) * (xP[p]+1.0)*0.5 *
		 jacobitmpD[p*(Q-2) + j-2] * (2.0*(i-1)+2.0+j)*0.5
		 :
		 0.0);
	    } // for p
	  } // for j
	} // if ... else ...
      } // for i
#if KarnConstr_D
      DEBUGL(KarnConstr_D, "computed values of principal function for orders "
	     << orders_);
      for (int p = 0; p < P+1; ++p) {
	for (int i = 0; i < Q+1; ++i) {
	  if (i == 0) cout << "p = " << p << "\t";
	  else cout << "\t";
	  for (int j = 0; j < NxP; ++j) {
	    cout << values_[p*NxP*(Q+1) + i*NxP + j];
	    if (j == NxP-1)
	      cout << endl;
	    else
	      cout << ", ";
	  }
	}
      }
#endif
    }
  }

  template<> // psi^b, divided form
  Karniadakis<2, 2>::Karniadakis(const int P, const Real* xP, 
				 const int NxP, const int Q, const int R,
				 const bool cache) :
    ShapeFunction1D<Real>((P+1)*(Q+1)*(R+1), NxP), orders_(P, NxP, Q, R),
    cache_(cache) {

    __gnu_cxx::hash_map<Orders<2>, Real*, __gnu_cxx::hash<Orders<2> >,
      __gnu_cxx::OrdersEqual<2> >::iterator i;
    i = principal_H.find(orders_);
    if (cache_ && (i != principal_H.end())) {
      // values already computed
      values_ = i->second;
      DEBUGL(KarnConstr_D, "found values of principal function for orders " 
	     << orders_);
    } else {
      // need to compute the values
      values_ = new Real[(P+1)*(Q+1)*NxP];
      if (cache_)
	principal_H.insert(std::pair<Orders<2>, Real*>(orders_, values_));

      Array<Real> jacobitmp;
      if (Q >= 2) jacobitmp.resize((Q-1)*NxP);
      for (int i = 0; i < P+1; ++i) {
	if ( (i == 0) || (i == 1) ) {
	  if (Q >= 2)
	    JacobiPol(1.0, 1.0, Q-2, xP, NxP, (Real*)jacobitmp);
	  for (int j = 0; j < Q+1; ++j) {
	    for (int p = 0; p < NxP; ++p) {
	      if (j == 0) {
		values_[i*NxP*(Q+1) + j*NxP + p] = 0.5;
		continue;
	      }
	      if (j == 1) {
		values_[i*NxP*(Q+1) + j*NxP + p] = 0.0;
		continue;
	      }
	      values_[i*NxP*(Q+1) + j*NxP + p] =
		(1.0+xP[p])/2.0 * jacobitmp[p*(Q-1) + j-2]*0.5;
	    } // for p
	  } // for j
	} else {
	  if (Q >= 2) JacobiPol(2.0*(i-1)+1.0, 1.0, Q-2, xP, NxP, 
					  (Real*)jacobitmp);
	  // j = 0
	  for (int p = 0; p < NxP; ++p) {
	    values_[i*NxP*(Q+1) + p] = std::pow((1.0-xP[p])/2.0,
						(Real)(i-1))*0.5;
	  }
	  // rest of the functions indexed by j
	  for (int j = 1; j < Q; ++j) {
	    for (int p = 0; p < NxP; ++p) {
	      values_[i*NxP*(Q+1) + j*NxP + p] =
		std::pow((1.0-xP[p])/2.0, (Real)(i-1))*0.5
		* (1.0+xP[p])/2.0 * jacobitmp[p*(Q-1) + j-1];
	    } // for p
	  } // for j
	} // if ... else ...
      } // for i
#if KarnConstr_D
      DEBUGL(KarnConstr_D, "computed values of principal function for orders "
	     << orders_);
      for (int p = 0; p < P+1; ++p) {
	for (int i = 0; i < Q+1; ++i) {
	  if (i == 0) cout << "p = " << p << "\t";
	  else cout << "\t";
	  for (int j = 0; j < NxP; ++j) {
	    cout << values_[p*NxP*(Q+1) + i*NxP + j];
	    if (j == NxP-1)
	      cout << endl;
	    else
	      cout << ", ";
	  }
	}
      }
#endif
    }
  }

  template<int type, int mode>
  Karniadakis<type, mode>::Karniadakis(const Karniadakis<type, mode>& arg) :
    ShapeFunction1D<Real>(arg.n(), arg.nP()), cache_(arg.cache_) {
    orders_ = arg.orders_;
    values_ = arg.values_;
  }

  template<int type, int mode>
  std::ostream& Karniadakis<type, mode>::info(std::ostream& os) const {

    os << "Karniadakis<";
    switch(type) {
    case 1: os << "psi^a"; break;
    case 2: os << "psi^b"; break;
    case 3: os << "psi^c"; break;
    default: conceptsAssert(1==0, Assertion()); break;
    }
    os << ", ";
    switch(mode) {
    case 0: os << "normal"; break;
    case 1: os << "derivatives"; break;
    case 2: os << "divided"; break;
    default: conceptsAssert(1==0, Assertion()); break;
    }
    os << ">(n = " << n() << ", nP = " << nP()
       << (cache_ ? ", cached" : "") << std::endl;
    switch(type) {
    case 1:
      for (int i = 0; i < orders_.p(0)+1
//  	     +1
	     ; ++i) {
	os << "i = " << i << ":\t";
	for (int j = 0; j < orders_.P(); ++j) {
	  os << values_[i*(orders_.P())+j];
	  if (j == orders_.P()-1)
	    os << std::endl;
	  else
	    os << ", ";
	}
      }
      break;
    case 2:
      for (int p = 0; p < orders_.p(0)+1; ++p) {
	for (int i = 0; i < orders_.p(1)+1; ++i) {
	  if (i == 0) os << "i = " << p << ":\t";
	  else os << '\t';
	  for (int j = 0; j < orders_.P(); ++j) {
	    os << values_[p*orders_.P()*(orders_.p(1)+1) +
			 i*orders_.P() + j];
	    if (j == orders_.P()-1) os << std::endl;
	    else os << ", ";
	  }
	}
      }
      break;
    }
    return os << ')';
  }

  template class Karniadakis<1, 0>; // psi^a, normal
  template class Karniadakis<1, 1>; // psi^a, derivatives
  template class Karniadakis<2, 0>; // psi^b, normal
  template class Karniadakis<2, 1>; // psi^b, derivatives
  template class Karniadakis<2, 2>; // psi^b, divided form

} // namaepsace concepts
