/* Expansion bases for tensorial shape functions not only for
 * quads and hexahedrons but also for triangles and tetrahedrons.
 * Due to Karniadakis, Sherwin
 */

#ifndef karniadakis_hh
#define karniadakis_hh

#include <iostream>

#if __ICC == 800
#  include <hash_map>
#  define __gnu_cxx std
#else
#  if __GNUC__ == 3
#    include <ext/hash_map>
#    if __GNUC_MINOR__ < 1
#      define __gnu_cxx std
#    endif
#  else
#    include <hash_map>
#    define __gnu_cxx std
#  endif
#endif

#include "basics/typedefs.hh"
#include "shapefunction.hh"

namespace concepts {

  // **************************************************************** Orders **

  /** Class containing number of quadrature points. This serves as a
      non-template base class for Orders.
      @author Philipp Frauenfelder, 2001
  */
  class OrdersBase {
  public:
    /// Default constructor
    OrdersBase() : P_(0) {}

    /// Constructor
    OrdersBase(const int P) : P_(P) {}

    /// Returns number of quadrature points
    inline int P() const { return P_; }
  protected:
    /// Number of quadrature points
    int P_;
  };

  template<int number>
  class Orders;

  template<int number>
  std::ostream& operator<< (std::ostream& os, const Orders<number>& o) { 
    for (unsigned int i = 0; i < number; ++i) {
      os << "p(" << i << ") = " << o.p_[i] << ", ";
    }
    os << "P = " << o.P_;
    return os;
  }

  /** Class combining polynomial order and number of quadrature points.

      @see Karniadakis
      @author Philipp Frauenfelder, 2000
  */
  template<int number>
  class Orders : public OrdersBase {
    friend std::ostream& operator<< <>(std::ostream& os,
				       const Orders<number>& o);
  public:
    /// Default constructor
    Orders();

    /// Constructor
    Orders(const int p0, const int P, const int p1 = 0, const int p2 = 0);

    /// Returns ith polynomial degree
    inline int p(const uint i) const { 
      if (i < number) return p_[i];
      else return 0;
    }

    /// Equality operator
    inline bool operator==(const Orders<number>& o) const {
      bool equal = true && (P_ == o.P_);
      for (uint i = 0; i < number; ++i)
	equal &= (p_[i] == o.p_[i]);
      return equal;
    }

    /// < operator
    inline bool operator<(const Orders<number>& o) const {
      for (uint i = 0; i < number; ++i)
	if (p_[i] < o.p_[i]) return true;
      return false;
    }
  private:
    /// Polynomial orders
    int p_[number];
  };

} // namespace concepts

namespace __gnu_cxx {

  template<int number>
  size_t hash_value(const concepts::Orders<number>& o);

  template<int number>
  struct hash<concepts::Orders<number> > {
    size_t operator()(const concepts::Orders<number>& o) const {
      return hash_value(o);
    }
  };

  /// Key equal function class
  template<int number>
  struct OrdersEqual {
    bool operator()(const concepts::Orders<number>& x,
		    const concepts::Orders<number>& y) const {
      return x == y;
    }
  };
} // namespace __gnu_cxx

namespace concepts {

  // *********************************************************** Karniadakis **

  /** Part of the multidimensional expansion bases for the shape
      functions of Karniadakis and Sherwin. Computes one of principal
      functions \f$\psi^a_i(z)\f$, \f$\psi^b_{ij}(z)\f$ or
      \f$\psi^c_{ijk}(z)\f$ for
      given polynomial degrees and a set of points.

      The constructor computes the values if necessary: they are stored
      in static hashes and can be reused if requested a second
      time. There is a method to get a pointer into the array of the
      computed values, use this at initialization time of the element to
      get a pointer to the values. The index of the points in which the shape
      functions are evaluated runs fastest, k is second and i is slowest.

      @param type Template parameter: type of the principal
           function. Can take the values
           <ul>
	   <li> 1 -- The principal function \f$\psi^a_i(z)\f$ is computed in
	        the constructor.
	   <li> 2 -- The principal function \f$\psi^b_{ij}(z)\f$ is computed in
	        the constructor.
	   <li> 3 -- The principal function \f$\psi^c_{ijk}(z)\f$ is computed
	        in the constructor.
           </ul>
      @param mode Template parameter: normal, derivatives, divided
           form. Can take the values
           <ul>
           <li> 0 -- normal
	   <li> 1 -- derivatives. The derivatives of principal
	        function are computed in the constructor,
		ie. \f${\psi^a_i}'(z)\f$, \f${\psi^b_{ij}}'(z)\f$ and
		\f${\psi^c_{ijk}}'(z)\f$ respectively, depending on
		the type of the principal function
	   <li> 2 -- divided form. The principal function is divided
	        by \f$1-z\f$: \f$\frac{\psi^b_{ij}(z)}{1-z}\f$ is
		computed in the constructor. This is only
		possible for the principal function
		\f$\psi^b_{ij}(z)\f$.
	   </ul>

      @test test::KarniadakisTest
      @author Philipp Frauenfelder, 2000	    
  */
  template<int type, int mode>
  class Karniadakis : public ShapeFunction1D<Real> {
  public:
    /** Constructor. Computes the values of the principal function of
	the given order and in the given points.

	@param P Order of the principal function in \f$\eta_1\f$
	@param xP Points
	@param NxP Number of points
	@param Q Order of the principal function in \f$\eta_2\f$
	@param R Order of the principal function in \f$\eta_3\f$
	@param cache Controls if the computed values should be taken
	             from and stored in the cache
    */
    Karniadakis(const int P, const Real* xP, const int NxP,
		const int Q = 0, const int R = 0, const bool cache = true);

    /// Copy constructor
    Karniadakis(const Karniadakis<type, mode>& arg);

    ~Karniadakis() {}
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Combines polynomial degrees and number of quadrature points
    Orders<type> orders_;

    /** Hash of the already computed values of the principal function of
	this type. The hash is static, ie. the data is available to all
	instances of this class.
    */
    static __gnu_cxx::hash_map<Orders<type>, Real*,
			       __gnu_cxx::hash<Orders<type> >,
			       __gnu_cxx::OrdersEqual<type> > principal_H;

    /// Cache the results of the computations in principal_H if set to true
    bool cache_;
  };


} // namespace concepts

#endif // karniadakis_hh
