/* 1D quadrature rules
 */

#ifndef quadrature_hh
#define quadrature_hh

#include <map>

#include "basics/typedefs.hh"
#include "basics/outputOperator.hh"

namespace concepts {

  // ************************************************************ Quadrature **

  /** Basic class for numerical integration. This class provides the
      datastructures for the weights and abscissas and the methods to
      get them.

      The values are computed in the template instantiations.

      @param type Template parameter: type of the quadrature rule. Can
      take the values <ul>
      <li>0: <b>Gauss Lobatto</b> quadrature. This rule
          includes both endpoints.
	  \f[ \int_{-1}^1 f(x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	  is exact for \f$f \in P_{2p-1}\f$ and n = p+1 points.
	  n must be greater or equal to 2.
	  \n
	  The abscissas \f$x_i\f$ are the zeros of 
	  \f$(1-x^2) P_{p-1}^{(1,1)}(x)\f$
	  and the weights are 
	  \f$w_i = 2/(p(p+1) (P_p^{(0,0)}(x_i))^2)\f$.
      <li>1: <b>Gauss Jacobi Lobatto</b> quadrature. This rule 
          includes both endpoints.
	  \f[ \int_{-1}^1 f(x) (1-x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	  is exact for \f$f \in P_{2p-1}\f$ and n = p+1 points.
	  n must be greater or equal to 2.
	  \n
	  The abscissas \f$x_i\f$ are the zeros of 
	  \f$(1-x^2) P_{p-1}^{(2,1)}(x)\f$
	  and the weights are 
	  \f$w_i = 4/(p(p+2) (P_p^{(1,0)}(x_i))^2)\f$ and
	  \f$w_p = 8/(p(p+2) (P_p^{(1,0)}(x_i))^2)\f$.
      <li>2: <b>Gauss Jacobi Lobatto</b> quadrature. This rule 
          includes both endpoints.
	  \f[ \int_{-1}^1 f(x) (1-x)^2 \, dx \approx \sum_{i=0}^p w_i f(x_i)\f]
	  is exact for \f$f \in P_{2p-1}\f$ and n = p+1 points.
	  n must be greater or equal to 2.
	  \n
	  The abscissas \f$x_i\f$ are the zeros of 
	  \f$(1-x^2) P_{p-1}^{(3,1)}(x)\f$
	  and the weights are 
	  \f$w_i = 8/(p(p+3) (P_p^{(2,0)}(x_i))^2)\f$ and
	  \f$w_p = 24/(p(p+3) (P_p^{(2,0)}(x_i))^2)\f$.
      <li>3: <b>Gauss Radau Jacobi</b> quadrature. This rule includes
          only one of the endpoints, ie. -1.
	  \f[ \int_{-1}^1 f(x) (1-x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	  is exact for \f$f \in P_{2p}\f$ and n = p+1 points.
	  n must be greater or equal to 2.
	  \n
	  The abscissas \f$x_i\f$ are the zeros of
	  \f$(1+x) P_p^{(1,1)}(x)\f$
	  and the weights are
	  \f$w_i = 2(1-x_i)/(p(p+1) (P_p^{(1,0)}(x_1))^2)\f$.
      <li>4: <b>Gauss Jacobi</b> quadrature. This rule does not include the
          endpoints -1 and 1.
	  \f[ \int_{-1}^1 f(x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	  is exact for \f$f \in P_{2p+1}\f$ and n = p+1 points.
	  n must be greater or equal to 1.
	  \n
	  The abscissas \f$x_i\f$ are the zeros of
	  \f$P_{p+1}^{(0,0)}(x)\f$
	  and the weights are
	  \f[w_i = \frac{2}{1-x_i^2} \left( \frac{d}{dx} \left. P^{(0,0)}_{p+1}(x) \right|_{x=x_i} \right)^{-2}.\f]
      </ul>

      Requesting the same quadrature rule with the same amount of
      integration points several times does not harm: the values are
      only stored once (internally).

      @test test::QuadratureTest
      @test test::KarniadakisTest
      @author Philipp Frauenfelder, 2000
  */
  template<int type>
  class Quadrature : public OutputOperator {
  public:
    /** Constructor. Computes the quadrature points.

	@param n Number of entries to be computed
    */
    Quadrature(uint n);
    /// Returns a pointer into the array of the abscissas
    const Real* abscissas() const { return abscissas_; }
    /// Returns a pointer into the array of the weights
    const Real* weights() const { return weights_; }
    /// Returns the number of quadrature points
    uint n() const { return n_; }
    /** Lists the integration orders (points) which have been computed so
	far.
	@return a string containing the computed orders.
    */
    static std::string storedPoints();
  protected:
    virtual std::ostream& info(std::ostream& os) const;

    /// Abscissas
    Real* abscissas_;
    /// Weights
    Real* weights_;
  private:
    /** Hash of the already computed values of the abscissas of this
	rule. The hash is static, ie. the data is available to all
	instances of this class.
    */
    static std::map<uint, Real*> abscissas_H;

    /** Hash of the already computed values of the weights of this
	rule. The hash is static, ie. the data is available to all
	instances of this class.
    */
    static std::map<uint, Real*> weights_H;

    /// Number of quadrature points
    uint n_;

    bool find_();
  };

} // namespace concepts

#endif // quadrature_hh
