/* 1D quadrature rules
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* ************************************************************* C routines **/

void JacobiPol(const double alf, const double bet, const int maxn,
	       const double* x, const int m, double* p) {

  double a1 = 0.0, a2 = 0.0, a3 = 0.0, a4 = 0.0;
  const double apb = alf + bet;
  const double a2mb2 = alf*alf - bet*bet;
  double xx = 0.0;
  int i, j;

  p[0] = 1.0;
  /* loop over all points in x */
  for (j = 0; j < m; ++j) {

    xx = x[j];
    /* loop over all needed polynomials */
    p[j*(maxn+1)] = 1.0;
    if (maxn > 0)
      p[j*(maxn+1) + 1] = (alf - bet + xx*(apb + 2.0))/2.0;
    for (i = 1; i < maxn; ++i) {

      a1 = 2.0*(i+1.0)*(i+1.0+apb)*(2.0*i+apb);
      a2 = (2.0*i+apb+1.0)*a2mb2;
      a3 = (2.0*i+apb)*(2.0*i+apb+1.0)*(2.0*i+apb+2.0);
      a4 = 2.0*(i+alf)*(i+bet)*(2.0*i+apb+2.0);

      p[j*(maxn+1) + i + 1] = ( (a2+a3*xx)*p[j*(maxn+1) + i] 
				- a4*p[j*(maxn+1) + i - 1] ) / a1;
    } /* for i */
  } /* for j */
}

void JacobiDerivatives(const double alf, const double bet, const int maxn,
		       const double* x, const int m, const double* p,
		       double* q) {
  double b1 = 0.0, b2 = 0.0, b3 = 0.0;
  double xx = 0.0;
  int i, j;

  /* loop over all point in x */
  for (j = 0; j < m; ++j) {
    xx = x[j];
    /* loop over all needed polynomials */
    q[j*(maxn+1) + 0] = 0.0;
    for (i = 1; i <= maxn; ++i) {
      b1 = (2.0*i+alf+bet)*(1.0 - xx*xx);
      b2 = i*(alf-bet-(2.0*i+alf+bet)*xx);
      b3 = 2*(i+alf)*(i+bet);
      q[j*(maxn+1) + i] = (b2*p[j*(maxn+1) + i] + b3*p[j*(maxn+1) + i-1]) / b1;
    }
  } /* for j */
}

#define EPS 3.0e-14
#define MAXIT 10

void JacobiZeros(double* x, int n, double alf, double bet) {

  int i,its,j;
  double alfbet,an,bn,r1,r2,r3;
  double a,b,c,p1,p2,p3,pp,temp,z = 0.0,z1;

  for (i=1;i<=n;i++) {
    if (i == 1) {
      an=alf/n;
      bn=bet/n;
      r1=(1.0+alf)*(2.78/(4.0+n*n)+0.768*an/n);
      r2=1.0+1.48*an+0.96*bn+0.452*an*an+0.83*an*bn;
      z=1.0-r1/r2;
    } else if (i == 2) {
      r1=(4.1+alf)/((1.0+alf)*(1.0+0.156*alf));
      r2=1.0+0.06*(n-8.0)*(1.0+0.12*alf)/n;
      r3=1.0+0.012*bet*(1.0+0.25*fabs(alf))/n;
      z -= (1.0-z)*r1*r2*r3;
    } else if (i == 3) {
      r1=(1.67+0.28*alf)/(1.0+0.37*alf);
      r2=1.0+0.22*(n-8.0)/n;
      r3=1.0+8.0*bet/((6.28+bet)*n*n);
      z -= (x[1]-z)*r1*r2*r3;
    } else if (i == n-1) {
      r1=(1.0+0.235*bet)/(0.766+0.119*bet);
      r2=1.0/(1.0+0.639*(n-4.0)/(1.0+0.71*(n-4.0)));
      r3=1.0/(1.0+20.0*alf/((7.5+alf)*n*n));
      z += (z-x[n-3])*r1*r2*r3;
    } else if (i == n) {
      r1=(1.0+0.37*bet)/(1.67+0.28*bet);
      r2=1.0/(1.0+0.22*(n-8.0)/n);
      r3=1.0/(1.0+8.0*alf/((6.28+alf)*n*n));
      z += (z-x[n-2])*r1*r2*r3;
    } else {
      z=3.0*x[i-1]-3.0*x[i-2]+x[i-3];
    }
    alfbet=alf+bet;
    for (its=1;its<=MAXIT;its++) {
      temp=2.0+alfbet;
      p1=(alf-bet+temp*z)/2.0;
      p2=1.0;
      for (j=2;j<=n;j++) {
	p3=p2;
	p2=p1;
	temp=2*j+alfbet;
	a=2*j*(j+alfbet)*(temp-2.0);
	b=(temp-1.0)*(alf*alf-bet*bet+temp*(temp-2.0)*z);
	c=2.0*(j-1+alf)*(j-1+bet)*temp;
	p1=(b*p2-c*p3)/a;
      }
      pp=(n*(alf-bet-temp*z)*p1+2.0*(n+alf)*(n+bet)*p2)/(temp*(1.0-z*z));
      z1=z;
      z=z1-p1/pp;
      if (fabs(z-z1) <= EPS) break;
    }
    if (its > MAXIT) {
      printf("too many iterations in gaujac");
      exit(1);
    }
    x[i]=z;
  }
}
#undef EPS
#undef MAXIT
/* (C) Copr. 1986-92 Numerical Recipes Software VsXz&52.!-. */

void GaussLobattoAbscWght(double* x, double* w, const uint p,
			  const uint jacobi) {

  double* xpnt = 0;
  double* xtmppnt = 0;
  double* poly = malloc((p+1)*(p+1)*sizeof(w[0]));
  double* xtmp = malloc((p+1)*sizeof(x[0]));

  int i;

  double alpha1 = 1.0;
  double alpha2 = 0.0;
  double add = 1.0;
  double fact = 2.0;

  if (jacobi) {
    alpha1 += jacobi;
    alpha2 += jacobi;
    add += jacobi;
    for (i = 0; i < jacobi; ++i)
      fact *= 2.0;
  }

  /* computing integration abscissas for Gauss Lobatto integration */

  /* gaujac fills 1...p-1 of x and w, x[0]=-1, x[p]=1 */
  JacobiZeros(xtmp, p-1, alpha1, 1.0);
  xtmp[0] = 1.0;
  xtmp[p] = -1.0;

  /* change sort order */
  xpnt = xtmp + p;
  xtmppnt = x;
  for (i = 0; i < p+1; ++i)
    *xtmppnt++ = *xpnt--;

  /* computing integration weights for Gauss Lobatto integration */
  JacobiPol(alpha2, 0.0, p, x, p+1, poly);

  for (i = 0; i < p+1; ++i)
    w[i] = fact / (poly[i*(p+1)+p]*poly[i*(p+1)+p] * p * (p+add));
  if (jacobi)
    w[p] = (1.0+jacobi) * fact / 
      (poly[p*(p+1)+p]*poly[p*(p+1)+p] * p * (p+add));

  free(poly);
  free(xtmp);
}

void GaussRadauAbscWght(double* x, double* w, const uint p, 
			const uint j) {

  int i;
  double* xpnt = 0;
  double* xtmppnt = 0;
  double* xtmp = malloc((p+1)*sizeof(x[0]));
  double* poly = malloc((p+1)*(p+1)*sizeof(w[0]));

  /* gaujac fills 1...p-1 of x and w, x[0]=-1, x[p]=1 */
  JacobiZeros(xtmp, p, j, 1.0);

  /* change sort order */
  xpnt = xtmp + p;
  xtmppnt = x+1;
  for (i = 0; i < p; ++i)
    *xtmppnt++ = *xpnt--;
  x[0] = -1.0;

  /* computing integration weights for Gauss Lobatto integration */
  JacobiPol(j, 0.0, p, x, p+1, poly);

  for (i = 0; i < p+1; ++i)
    w[i] = pow(2.0, j) * (1-x[i]) / ((p+1.0)*(p+j+1.0)*
				     (poly[i*(p+1)+p]*poly[i*(p+1)+p]));

  free(poly);
  free(xtmp);
}

void GaussJacobiAbscWght(double* x, double* w, const uint p) {
  double* xpnt = 0;
  double* xtmppnt = 0;
  double* xtmp = malloc((p+2)*sizeof(x[0]));
  double* poly = malloc((p+2)*(p+1)*sizeof(w[0]));
  double* dpoly = malloc((p+2)*(p+1)*sizeof(w[0]));
  int i;

  /* get zeros of Jacobi polynomials */
  JacobiZeros(xtmp, p+1, 0.0, 0.0);
  xtmppnt = x;
  xpnt = xtmp + p+1;
  for (i = 0; i < p+1; ++i)
    *xtmppnt++ = *xpnt--;

  /* evaluating Jacobi polynomials */
  JacobiPol(0.0, 0.0, p+1, x, p+1, poly);
  /* evaluating derivatives of Jacobi polynomials */
  JacobiDerivatives(0.0, 0.0, p+1, x, p+1, poly, dpoly);

  for (i = 0; i < p+1; ++i)
    w[i] = 2.0 / (1.0 - x[i]*x[i]) / dpoly[i*(p+2)+p+1] / dpoly[i*(p+2)+p+1];

  free(dpoly);
  free(poly);
  free(xtmp);
}
