/* 1D quadrature rules
 */

#ifndef quadratureBase_hh
#define quadratureBase_hh

#include "basics/typedefs.hh"

namespace concepts {

  // ************************************************************ C routines **

#ifdef __cplusplus
  extern "C" {
#endif

    /** Computes the values of the Jacobi polynomials.

	\f$P^{(\alpha, \beta)}_i (x_j)\f$ for \f$i = 0, \dots, maxn\f$ 
	and \f$x_j \in x\f$, \f$|x| = m\f$.

	@param alf \f$\alpha\f$
	@param bet \f$\beta\f$
	@param maxn Highest polynomial degree to be computed
	@param x Array of size \c m with the points in which the polynomials
	         should be evaluated. \f$x_j \in [-1,1]\f$.
	@param m Size of array \c x
	@param p Array of the size \c m*(maxn+1) (must be allocated), will
	         be filled with the values of the poynomials in the points
		 with i running fastest.

	@pre The space for the array \c p is allocated,
	size: m*(maxn+1)

	@author Philipp Frauenfelder, after the recursion formulae in
	Karniadakis / Sherwin "Spectral/hp Element Methods for CFD"
	p. 350f
    */
    void JacobiPol(const double alf, const double bet, const int maxn,
		   const double* x, const int m, double* p);

    /** Computes the values of the derivatives of the Jacobi polynomials.

        \f$\frac{d}{dx} P^{(\alpha,\beta)}_i (x_j)\f$ for
	\f$i = 0, \dots, maxn\f$ and \f$x_j \in x\f$, \f$|x| = m\f$.

	@param alf \f$\alpha\f$
	@param bet \f$\beta\f$
	@param maxn Highest polynomial degree to be computed
	@param x Array of size \c m with the points in which the polynomials
	         should be evaluated. \f$x_j \in [-1,1]\f$.
	@param m Size of array \c x
	@param p Array of the size \c m*(maxn+1) filled
	         with the Jacobi polynomials.
	@param q Array of the size \c m*(maxn+1) (must be allocated), will
	         be filled with the values of the derivatives of the poynomials
		 in the points with \c i running fastest.

	@pre The space for the array \c q is allocated, size: \c m*(maxn+1)

	@author Philipp Frauenfelder, after the recursion formulae in
	Karniadakis / Sherwin "Spectral/hp Element Methods for CFD"
	p. 350f
    */
    void JacobiDerivatives(const double alf, const double bet, const int maxn,
			   const double* x, const int m, const double* p,
			   double* q);

    /** Computes the zeros of the Jacobi polynomials
	\f$P_{p}^{(\alpha,\beta)}(x)\f$.

	@param x Output: the zeros of the Jacobi polynomials in the elements
	         1, ..., p. The element 0 of the array is untouched.
	@param p Degree of the Jacobi polynomial
	@param alf \f$\alpha\f$
	@param bet \f$\beta\f$

	@pre The space for the arrays x is allocated, size: p+1

	@author (C) Copr. 1986-92 Numerical Recipes Software VsXz&52.!-.
    */
    void JacobiZeros(double* x, int p, double alf, double bet);

    /** Computes and returns the integration weights and abscissas
	for the Gauss (Jacobi) Lobatto integration. The Jacobi version
	can be chosen with a parameter. The abscissas include both endpoints,
	ie. 1 and -1.

	Integration with the quadrature rule
	\f[ \int_{-1}^1 f(x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	is exact for polynomials \f$f \in P_{2p-1}\f$. The Jacobi version
	integrates 
	\f[ \int_{-1}^1 f(x) (1-x)^j \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	exactly for \f$f \in P_{2p-1}\f$.

	The abscissas \f$x_i\f$ are the zeros of
	\f$(1-x^2) P_{p-1}^{(1,1)}(x)\f$
	and the weights are \f$w_i = 2/(p(p+1) (P_p^{(0,0)}(x_i))^2)\f$. For
	the Jacobi version, the abscissas are the zeros of
	\f$(1-x^2) P_{p-1}^{(1+j,1)}(x)\f$ and the weights are
	\f$w_i = 2^{1+j}/(p(p+1+j) (P_p^{(j,0)}(x_i))^2)\f$ and
	\f$w_p = (1+j) \cdot 2^{1+j}/(p(p+1+j) (P_p^{(j,0)}(x_i))^2)\f$.

	@param x Output: the integration abscissas
	@param w Output: the integration weights
	@param p Order of the integration, p+1 points are computed
	@param j The Jacobi version of the integration rule is used if
	this is non-zero.

	@pre The space for the arrays x and w is allocated,
	size: p+1.

	@author Philipp Frauenfelder, 2000
    */
    void GaussLobattoAbscWght(double* x, double* w, const uint p,
			      const uint j = 0);

    /** Computes and returns the integration weights and abscissas
	for the Gauss Radau Jacobi integration. The Jacobi version
	can be chosen with a parameter. The abscissas include only one
	endpoint: -1.

	Integration with the quadrature rule
	\f[ \int_{-1}^1 f(x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	is exact for polynomials \f$f \in P_{2p}\f$. The Jacobi version
	integrates 
	\f[ \int_{-1}^1 f(x) (1-x)^j \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	exactly for \f$f \in P_{2p}\f$.

	The abscissas \f$x_i\f$ are the zeros of \f$(1+x) P_{p}^{(j,1)}(x)\f$
	and the weights are \f$w_i = 2/((p+1)(p+1+j) (P_p^{(j,0)}(x_i))^2)\f$.

	@param x Output: the integration abscissas
	@param w Output: the integration weights
	@param p Order of the integration, p+1 points are computed
	@param j The Jacobi version of the integration rule is used if
	this is non-zero.

	@pre The space for the arrays x and w is allocated,
	size: p+1.

	@author Philipp Frauenfelder, 2001
    */
    void GaussRadauAbscWght(double* x, double* w, const uint p,
			    const uint j = 0);

    /** Computes and returns the integration weights and abscissas for
	the Gauss Jacobi integration. The abscissas no not include the
	endpoints -1 and 1.

	Integration with the quadrature rule
	\f[ \int_{-1}^1 f(x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
	is exact for polynomials \f$f \in P_{2p+1}\f$.

	The abscissas \f$x_i\f$ are the zeros of \f$P_{p+1}^{(0,0)}(x)\f$
	and the weights are
	\f[w_i = \frac{2}{1-x_i^2} \left( \frac{d}{dx} \left. P^{(0,0)}_{p+1}(x) \right|_{x=x_i} \right)^{-2}.\f]

	@pre The space for the arrays x and w is allocated,
	size: p+1.

	@author Philipp Frauenfelder, 2001
    */
    void GaussJacobiAbscWght(double* x, double* w, const uint p);

#ifdef __cplusplus
  }
#endif

} // namespace concepts

#endif // quadratureBase_hh
