// Space for linear DG FEM in 1D

#ifndef linDG1Dspace_hh
#define linDG1Dspace_hh

#include <ostream>

#include "basics/typedefs.hh"
#include "space/space.hh"
#include "space/elementPairs.hh"
#include "linearFEM/element1D.hh"
#include "linearFEM/space1D.hh"
#include "toolbox/scannerConnectors.hh"

namespace concepts {

  // forward declarations
  class Mesh1;

} // namespace concepts

namespace linDG1D {

  using concepts::Real;

  // ************************************************************** Linear1d **

  /** Space for linear DG FEM in 1D.
      @author Thomas Wihler, 2003
  */
  class Linear1d : public concepts::Space<Real> {
  public:
    typedef concepts::Scan<linearFEM::Line> Scan;
    typedef void (*SMap)(const concepts::TColumn<Real>&,
			 concepts::TColumn<Real>&);
    /** Constructor. Recursively calls \c constructor_ and generates the
	list of element pairs \c elmPair_.
	@param msh Mesh
	@param level Number levels for initial refinement
    */
    Linear1d(concepts::Mesh1& msh, uint level);
    virtual ~Linear1d();
    virtual uint dim() const { return dim_; }
    virtual uint nelm() const { return nelm_; }
    virtual Scan* scan() const
    { return new concepts::PListScan<linearFEM::Line>(*elm_); }
    const concepts::ElementPairList<Real>& elmPairs() const
    { return elmPair_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Number of degrees of freedom
    uint dim_;
    /// Number of elements
    uint nelm_;
    /// Mesh which this space is based on
    concepts::Mesh1& msh_;
    /// List of elements
    concepts::Joiner<linearFEM::Line*, 1>* elm_;
    /** List of element pairs. This list contains
	<ul>
	<li>pairs of neighbouring
	elements (with a common vertex whose index is stored in \c
	DGElementPair) and
	<li>pairs of identical elements (with the minimal
	sizes of the neighbouring elements over every vertex stored in
	\c DGElementPair).
	</ul>
    */
    concepts::ElementPairList<Real> elmPair_;
    /** Assigns the degrees of freedom and creates elements. The numbers of
	the degrees of freedom is stored in \c dofMap.

	\c constructor_ is called recursively until \c level is 0. On this
	level, the degrees of freedom are assigned and the elements are
	created and stored in \c elm_.
	@param cell Current cell
	@param level Number of levels left to desired level
    */
    void constructor_(concepts::Cell1& cell, uint level);

    /// Data structure used for temporary data to generate \c DGElementPair
    struct TwoElements {
      TwoElements() { elms[0] = 0; elms[1] = 0; }
      void addElm(const linearFEM::Line* elm, const uint idx) {
	conceptsAssert(elms[1] == 0, concepts::Assertion());
	if (elms[0] == 0) {
	  elms[0] = elm;
	  idxs[0] = idx;
	}
	else {
	  elms[1] = elm;
	  idxs[1] = idx;
	}
      }
      const linearFEM::Line* elms[2];
      uint idxs[2];
    };
  };

} // namespace linDG1D

#endif // linDG1Dspace_hh
