// Bilinear forms for linear DG FEM in 2D

#ifndef linDG2DbilinearForm_hh
#define linDG2DbilinearForm_hh

#include "basics/typedefs.hh"
#include "operator/bilinearForm.hh"
#include "geometry/boundaryConditions.hh"

using concepts::Real;

namespace linDG2D {

  // ************************************************************ IdentityBf **

  /** Volume integral for the identity in DG FEM in 2D. This bilinear form
      integrates \f[ \int_{K} \phi_i \cdot \phi_j \, dx. \f]
      @author Christoph Winkelmann, 2003
  */
  class IdentityBf : public concepts::BilinearForm<Real> {
  public:
    /// Constructor
    IdentityBf() {}
    /// Application operator
    virtual void operator()(const concepts::Element<Real>& elmV,
                            const concepts::Element<Real>& elmU,
                            concepts::ElementMatrix<Real>& em);
  };

  // ********************************************************** LaplaceVolBf **

  /** Volume integral for the Laplacian in DG FEM in 2D. This bilinear form
      integrates
      \f[ \int_{K} \nabla \phi_i \cdot \nabla \phi_j \, dx. \f]
      @author Christoph Winkelmann, 2003
  */
  class LaplaceVolBf : public concepts::BilinearForm<Real> {
  public:
    /// Constructor
    LaplaceVolBf() {}
    /// Application operator
    virtual void operator()(const concepts::Element<Real>& elmV,
                            const concepts::Element<Real>& elmU,
                            concepts::ElementMatrix<Real>& em);
  };

  // ********************************************************* LaplaceFluxBf **

  /** Edge fluxes for the Laplacian in DG FEM in 2D. This bilinear form
      integrates
      \f[ \int_{e} \langle \nabla \phi_i \rangle n_e \cdot [\phi_j] \, ds. \f]
      @author Christoph Winkelmann, 2003
  */
  class LaplaceFluxBf : public concepts::BilinearForm<Real> {
  public:
    /** Constructor
        @param bc Boundary conditions, only the boundary type is relevant
    */
    LaplaceFluxBf(concepts::BoundaryConditions* bc) 
      : bc_(bc) {}
    /// Application operator for volume terms, does nothing
    virtual void operator()(const concepts::Element<Real>& elmV,
                            const concepts::Element<Real>& elmU,
                            concepts::ElementMatrix<Real>& em) {
      conceptsAssert(false, concepts::Assertion());
    }
    /// Application operator for DG term
    virtual void operator()(const concepts::Element<Real>& elmV,
                            const concepts::Element<Real>& elmU,
                            concepts::ElementMatrix<Real>& em,
                            const concepts::ElementPair<Real>& ep);
  private:
    /// Boundary conditions, only the boundary type is relevant
    const concepts::BoundaryConditions* bc_;
  };

  // ********************************************************* LaplaceStabBf **

  /** Stabilizing edge integral for the Laplacian in DG FEM in 2D. This
      bilinear form integrates
      \f[ \frac{\omega}{\vert e\vert} \int_{e} [\phi_i] \cdot [\phi_j] \, ds. \f]
      @author Christoph Winkelmann, 2003
  */
  class LaplaceStabBf : public concepts::BilinearForm<Real> {
  public:
    /** Constructor
        @param omega Stabilization parameter
        @param bc Boundary conditions, only the boundary type is relevant
    */
    LaplaceStabBf(Real omega, concepts::BoundaryConditions* bc)
      : omega_(omega), bc_(bc) {}
    /// Application operator for volume terms, does nothing
    virtual void operator()(const concepts::Element<Real>& elmV,
                            const concepts::Element<Real>& elmU,
                            concepts::ElementMatrix<Real>& em) {
      conceptsAssert(false, concepts::Assertion());
    }
    /// Application operator for DG term
    virtual void operator()(const concepts::Element<Real>& elmV,
                            const concepts::Element<Real>& elmU,
                            concepts::ElementMatrix<Real>& em,
                            const concepts::ElementPair<Real>& ep);
  private:
    /// Stabilization parameter
    const Real omega_;
    /// Boundary conditions, only the boundary type is relevant
    const concepts::BoundaryConditions* bc_;
  };

} // namespace linDG2D

#endif // linDG2DbilinearForm_hh
