// element pair for linear DG FEM in 2D

#ifndef elementPairDG2D_hh
#define elementPairDG2D_hh

#include "space/elementPairs.hh"
#include "edgeInfo.hh"

// debugging
#include "basics/debug.hh"

#define DGElementPair2dConstr_D 0

namespace linDG2D {

  using concepts::Real;

  // *********************************************************** ElementPair **

  /** Element pair for DG FEM in 2D. Combines two elements for computing the
      off-diagonal blocks of the system matrices.
      @author Christoph Winkelmann, 2003
  */
  class ElementPair : public concepts::ElementPair<Real> {
  public:
    /** Constructor.
        @param elm1 First element (with test function)
        @param elm2 Second element (with ansatz function)
        @param edgeInfo edge information
    */
    ElementPair(const Triangle& elm1, const Triangle& elm2, 
                const EdgeInfo& edgeInfo) : 
      elm1_(elm1), elm2_(elm2), edgeInfo_(edgeInfo) {
      DEBUGL(DGElementPair2dConstr_D, elm1 << ", " << elm2);
      idxs_[0] = &(elm1.cell()) == edgeInfo.cell(0) ? 
        edgeInfo.idx(0) : edgeInfo.idx(1);
      idxs_[1] = &(elm2.cell()) == edgeInfo.cell(0) ? 
        edgeInfo.idx(0) : edgeInfo.idx(1);
    }
    /// Returns the first element
    virtual const Triangle& elm1() const { return elm1_; }
    /// Returns the second element
    virtual const Triangle& elm2() const { return elm2_; }
    /// Returns the edge information object
    const EdgeInfo& edgeInfo() const { return edgeInfo_; }
    /** Evaluate trace of shapefunction from element e onto edge
        @param e element to take shapefunction from
        (0 for testfunction, 1 for ansatz function)
        @param i number of the shapefunction (0, 1 or 2)
        @param xi reference coordinate on edge in (0,1)
    */
    Real shapeFct(uint e, uint i, Real xi) const {
      const Triangle& elm = e == 0 ? elm1_ : elm2_;
      Real xim = &(elm.cell()) == edgeInfo_.cell(0) ? xi : 1.0 - xi;
      switch(idxs_[e]) {
      case 0: return elm.shapeFct(i, xim, 0.0);
      case 1: return elm.shapeFct(i, 1.0, xim);
      case 2: return elm.shapeFct(i, 1.0-xim, 1.0-xim);
      default: conceptsAssert(false, concepts::Assertion()); return 0.0;
      }
    }
  private:
    const Triangle& elm1_;
    const Triangle& elm2_;
    const EdgeInfo& edgeInfo_;
    uint idxs_[2];
  };

}

#endif // elementPairDG2D_hh
