// testing linDG2D::EdgeInfo

#include "basics/testsuite.hh"
#include "basics/exceptions.hh"
#include "geometry/meshImport2D.hh"
#include "geometry/uniformlyRefinedMesh2.hh"

#include "edgeInfoTest.hh"
#include "edgeInfo.hh"
#include "elementPair.hh"
#include "meshInfo.hh"

namespace test {

  using concepts::Real;
  
  // ******************************************************** DGEdgeInfoTest **

  void DGEdgeInfoTest::run() {
    concepts::Vertex v0, v1, v2, v3;
    concepts::Edge e0(v0, v1);
    concepts::Edge e1(v1, v2);
    concepts::Edge e2(v2, v0);
    concepts::Edge e3(v1, v3);
    concepts::Edge e4(v3, v2);
    concepts::Triangle tco0(e0, e1, e2);
    concepts::Triangle tco1(e1, e3, e4);
    concepts::Real2d p0(-4.5, -3.0);
    concepts::Real2d p1(3.0, -1.5);
    concepts::Real2d p2(-1.5, 3.0);
    concepts::Real2d p3(3.0, 3.0);
    concepts::VertexTriangle2d tm0(p0, p1, p2);
    concepts::VertexTriangle2d tm1(p2, p1, p3);
    concepts::Triangle2d tce0(tco0, tm0);
    concepts::Triangle2d tce1(tco1, tm1);
    linDG2D::EdgeInfo ei[5];
    ei[0].addCell(&tce0, 0);
    ei[1].addCell(&tce0, 1);
    ei[1].addCell(&tce1, 0);
    ei[2].addCell(&tce0, 2);
    ei[3].addCell(&tce1, 1);
    ei[4].addCell(&tce1, 2);

    testLength(ei);
    testNormal(ei);
    testVertices(ei);
    testMeanHeight(ei);
  }

  void DGEdgeInfoTest::testLength(linDG2D::EdgeInfo* ei) {
    Real dx = 7.5;
    Real dy = 1.5;
    Real len = std::sqrt(dx*dx+dy*dy);
    _numtest(ei[0].length(), len);
    dx = -4.5;
    dy = 4.5;
    len = std::sqrt(dx*dx+dy*dy);
    _numtest(ei[1].length(), len);
    dx = -3.0;
    dy = -6.0;
    len = std::sqrt(dx*dx+dy*dy);
    _numtest(ei[2].length(), len);
  }

  void DGEdgeInfoTest::testNormal(linDG2D::EdgeInfo* ei) {
    Real dx = 7.5;
    Real dy = 1.5;
    Real len = std::sqrt(dx*dx+dy*dy);
    concepts::Real2d normal(dy, -dx);
    normal = normal * (1/len);
    _numtest((ei[0].normal()-normal).l2(), 0);
    dx = -4.5;
    dy = 4.5;
    normal[0] = dy;
    normal[1] = -dx;
    len = std::sqrt(dx*dx+dy*dy);
    normal = normal * (1/len);
    _numtest((ei[1].normal()-normal).l2(), 0);
    dx = -3.0;
    dy = -6.0;
    normal[0] = dy;
    normal[1] = -dx;
    len = std::sqrt(dx*dx+dy*dy);
    normal = normal * (1/len);
    _numtest((ei[2].normal()-normal).l2(), 0);
  }

  void DGEdgeInfoTest::testVertices(linDG2D::EdgeInfo* ei) {
    uint i0 = ei[1].idxs_[0];
    uint i1 = ei[1].idxs_[1];
    const concepts::Triangle2d& c0 = *(ei[1].cell(0));
    const concepts::Triangle2d& c1 = *(ei[1].cell(1));

    concepts::Vertex* ve0 = c0.connector().vertex(i0);
    concepts::Vertex* ve1 = c1.connector().vertex((i1+1)%3);
    _test(ve0 == ve1);

    ve0 = c0.connector().vertex((i0+1)%3);
    ve1 = c1.connector().vertex(i1);
    _test(ve0 == ve1);

    concepts::Real3d pe0 = c0.vertex(i0);
    concepts::Real3d pe1 = c1.vertex((i1+1)%3);
    _numtest((pe0-pe1).l2(), 0);
    
    pe0 = c0.vertex((i0+1)%3);
    pe1 = c1.vertex(i1);
    _numtest((pe0-pe1).l2(), 0);
  }
  
  void DGEdgeInfoTest::testMeanHeight(linDG2D::EdgeInfo* ei) {
    Real a0 = 20.25;
    Real a1 = 10.125;
    Real dx = 7.5;
    Real dy = 1.5;
    Real len = std::sqrt(dx*dx+dy*dy);
    Real mh = a0/len;
    _numtest(ei[0].meanHeight(), mh);
    dx = -4.5;
    dy = 4.5;
    len = std::sqrt(dx*dx+dy*dy);
    mh = 2/(len/a0 + len/a1);
    _numtest(ei[1].meanHeight(), mh);
    dx = -3.0;
    dy = -6.0;
    len = std::sqrt(dx*dx+dy*dy);
    mh = a0/len;
    _numtest(ei[2].meanHeight(), mh);
    dx = 0;
    dy = 4.5;
    len = std::sqrt(dx*dx+dy*dy);
    mh = a1/len;
    _numtest(ei[3].meanHeight(), mh);
    dx = -4.5;
    dy = -0.0;
    len = std::sqrt(dx*dx+dy*dy);
    mh = a1/len;
    _numtest(ei[4].meanHeight(), mh);

    concepts::Import2dMesh coarseMesh
      ("../../../../linDG2D/testsuite/0101.coord",
       "../../../../linDG2D/testsuite/0101.elms",
       "../../../../linDG2D/testsuite/0101.bnd", 1);
    for (uint level=0; level<3; ++level) {
      concepts::UniformlyRefinedMesh2 mesh(coarseMesh, level);
      linDG2D::MeshInfo mi(mesh);
      std::auto_ptr<concepts::Scan<linDG2D::EdgeInfo> > sc(mi.scan());

      const linDG2D::EdgeInfo& fei = (*sc)++;
      Real fmin = fei.meanHeight() / fei.length();
      Real fmax = fmin;

      while (*sc) {
        const linDG2D::EdgeInfo& ei = (*sc)++;
        Real f = ei.meanHeight() / ei.length();
        fmin = f<fmin ? f : fmin;
        fmax = f>fmax ? f : fmax;
      }
    }
  }
}
