#include "bilinearForm.hh"

namespace linDG3D {

  void Assembly::inner(concepts::Matrix<Real>& dest,
		       FvdgBilinearForm& bf,
		       const concepts::ElementPairList<Real>& pairs,
		       bool transpose) {
    concepts::ElementMatrix<Real> A(3, 3);
    concepts::ElementMatrix<Real> B(3, 3);
    
    std::auto_ptr<concepts::ElementPairList<Real>::Scanner> 
      pairSc(pairs.scan());
    conceptsAssert(pairSc.get(), concepts::Assertion());
    while(*pairSc) {
      try {
	const concepts::ElementPair<Real>& epBase = (*pairSc)++;
	const FvdgElementPair& ep =
	  dynamic_cast<const FvdgElementPair&>(epBase);
	DEBUGL(DEBUG_ASSEMBLY, "element pair (" 
	       << ep.elm1().cell().connector().key() << "," 
	       << ep.elm2().cell().connector().key() << ")" );
	
	const FvdgElement* elm[2];
	const concepts::TMatrixBase<Real>* T[2];
	
	elm[0] = &ep.elm1();
	T[0] = &elm[0]->T();
	elm[1] = &ep.elm2();
	T[1] = &elm[1]->T();
	
	for(uint k=0; k<2; ++k) {
	  for(uint l=0; l<2; ++l) {
	    bf(*elm[k], *elm[l], A, ep);
	    if (A.size() > 0) {
	      T[l]->operator()(A, B);
	      B.transpose();
	      T[k]->operator()(B, A);
	      A.transpose();
	      DEBUGL(DEBUG_ASSEMBLY, "assembling " << A << " into " << dest);
	      for(uint i = T[k]->n(); i--;) {
		for(uint j = T[l]->n(); j--;) {
		  if(transpose) {
		    dest(T[l]->index(j), T[k]->index(i)) += A(i, j);
		  } else {
		    dest(T[k]->index(i), T[l]->index(j)) += A(i, j);
		  }
		} // for j
	      } // for i
	    } // for elm l
	  } // for elm k
	} // if size
      } catch (std::bad_cast& e) {
	std::cerr << e.what() << std::endl;
	conceptsAssert(false, concepts::Assertion());
      } // end try/catch
    }
  }
  
  void Assembly::boundary(concepts::Matrix<Real>& dest,
			  FvdgBilinearForm& bf,
			  const concepts::ElementPairList<Real>& pairs,
			  bool transpose) {
    concepts::ElementMatrix<Real> A(3, 3);
    concepts::ElementMatrix<Real> B(3, 3);
    
    std::auto_ptr<concepts::ElementPairList<Real>::Scanner> 
      pairSc(pairs.scan());
    conceptsAssert(pairSc.get(), concepts::Assertion());
    while(*pairSc) {
      try {
	const concepts::ElementPair<Real>& epBase = (*pairSc)++;
	const FvdgElementPair& ep =
	  dynamic_cast<const FvdgElementPair&>(epBase);
	
	const FvdgElement& elm1 = ep.elm1();
	const concepts::TMatrixBase<Real>& T1 = elm1.T();
	const FvdgElement& elm2 = ep.elm2();
	const concepts::TMatrixBase<Real>& T2 = elm2.T();
	
	bf(elm1, elm2, A, ep);
	if (A.size() > 0) {
	  T2(A, B);
	  B.transpose();
	  T1(B, A);
	  A.transpose();
	  for(uint i = T1.n(); i--;) {
	    for(uint j = T2.n(); j--;) {
	      if(transpose) {
		dest(T2.index(j), T1.index(i)) += A(i, j);
	      } else {
		dest(T1.index(i), T2.index(j)) += A(i, j);
	      }
	    } // for j
	  } // for i
	} // if size
      } catch (std::bad_cast& e) {
	std::cerr << e.what() << std::endl;
	conceptsAssert(false, concepts::Assertion());
      } // end try/catch
    }
  }

  //*************************************************************** Identity **
  void Identity::operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em) {
    // check for P0 elements
    const FvdgP0TetElem* eX0 =
      dynamic_cast<const FvdgP0TetElem*>(&elmX);
    const FvdgP0TetElem* eY0 =
      dynamic_cast<const FvdgP0TetElem*>(&elmY);
    if (eX0 && eY0) { this->operator()(*eX0, *eY0, em); return; }
    // check for P1 elements
    const FvdgP1TetElem* eX1 =
      dynamic_cast<const FvdgP1TetElem*>(&elmX);
    const FvdgP1TetElem* eY1 =
      dynamic_cast<const FvdgP1TetElem*>(&elmY);
    if (eX1 && eY1) { this->operator()(*eX1, *eY1, em); return; }
    // if both checks fail, throw exception
    std::stringstream error;
    error << "element combination " << elmX << " with " << elmY 
	  << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }

  void Identity::operator()(const FvdgP0TetElem& eX,
			    const FvdgP0TetElem& eY,
			    concepts::ElementMatrix<Real>& em) {
    const uint m = eX.T().m();
    const uint n = eY.T().m();
    em.resize(m, n);
    em.zeros();
    if (eX.support().key() == eY.support().key()) {
      conceptsAssert(1 == m, concepts::Assertion());
      conceptsAssert(1 == n, concepts::Assertion());
      
      const concepts::Real3d x0 = eX.cell().vertex(0);
      const concepts::Real3d x1 = eX.cell().vertex(1);
      const concepts::Real3d x2 = eX.cell().vertex(2);
	const concepts::Real3d x3 = eX.cell().vertex(3);
	Real volume = (((x1-x0)^(x2-x0))*(x3-x0))/6.0;
	em(0,0) = volume;
	DEBUGL(DEBUG_IDENTITY, em);
    }
  }

  void Identity::operator()(const FvdgP1TetElem& eX,
			    const FvdgP1TetElem& eY,
			    concepts::ElementMatrix<Real>& em) {
    const uint m = eX.T().m();
    const uint n = eY.T().m();
    em.resize(m, n);
    em.zeros();
    if (eX.support().key() == eY.support().key()) {
      conceptsAssert(4 == m, concepts::Assertion());
      conceptsAssert(4 == n, concepts::Assertion());
      
      const concepts::Real3d x0 = eX.cell().vertex(0);
      const concepts::Real3d x1 = eX.cell().vertex(1);
      const concepts::Real3d x2 = eX.cell().vertex(2);
      const concepts::Real3d x3 = eX.cell().vertex(3);
      Real volume = (((x1-x0)^(x2-x0))*(x3-x0))/6.0;
      em(0,0) = volume;
      em(1,1) = volume/5.0;
      em(2,2) = volume/5.0;
      em(3,3) = volume/5.0;
      DEBUGL(DEBUG_IDENTITY, em);
    }
  }

  //*********************************************************** LaplaceVolBF **
  void LaplaceVolBF::operator()(const concepts::Element<Real>& elmX,
				const concepts::Element<Real>& elmY,
				concepts::ElementMatrix<Real>& em) {
    // check for P1 elements
    const FvdgP1TetElem* eX1 =
      dynamic_cast<const FvdgP1TetElem*>(&elmX);
    const FvdgP1TetElem* eY1 =
      dynamic_cast<const FvdgP1TetElem*>(&elmY);
    if (eX1 && eY1) { this->operator()(*eX1, *eY1, em); return; }
    // if check fails, throw exception
    std::stringstream error;
    error << "element combination " << elmX << " with " << elmY 
	  << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
  }

  void LaplaceVolBF::operator()(const FvdgP1TetElem& eX,
				const FvdgP1TetElem& eY,
				concepts::ElementMatrix<Real>& em) {
    const uint m = eX.T().m();
    const uint n = eY.T().m();
    em.resize(m, n);
    em.zeros();
    if (eX.support().key() == eY.support().key()) {
      conceptsAssert(4 == m, concepts::Assertion());
      conceptsAssert(4 == n, concepts::Assertion());
      // calculate volume
      const concepts::Real3d x0 = eX.cell().vertex(0);
      const concepts::Real3d x1 = eX.cell().vertex(1);
      const concepts::Real3d x2 = eX.cell().vertex(2);
      const concepts::Real3d x3 = eX.cell().vertex(3);
      Real volume = (((x1-x0)^(x2-x0))*(x3-x0))/6.0;
      // fetch gradients
      concepts::Real3d gradient[4];
      for(uint i=0; i<4; ++i) {
	gradient[i] = eX.shapeFctGradient(i);
      }
      // fill element matrix
      for(uint p=0; p<4; ++p) {
	em(p,p) = (gradient[p]*gradient[p])*volume;
	for(uint q=p+1; q<4; ++q) {
	  em(p,q) = em(q,p) = (gradient[p]*gradient[q])*volume;
	}
      }
      DEBUGL(DEBUG_LAPLACE, em);
    }
  }

  //********************************************************* LaplaceInnerBF **
  void LaplaceInnerBF::operator()(const FvdgElement& elmX,
				  const FvdgElement& elmY,
				  concepts::ElementMatrix<Real>& em,
				  const FvdgElementPair& ep) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();
    
    const concepts::Key keyX = elmX.support().key();
    const concepts::Key keyY = elmY.support().key();
    const concepts::Key key1 = ep.elm1().support().key();
    const concepts::Key key2 = ep.elm2().support().key();
    conceptsAssert( (keyX == key1) || (keyX == key2), concepts::Assertion());
    conceptsAssert( (keyY == key1) || (keyY == key2), concepts::Assertion());
    conceptsAssert(key1 != key2, concepts::Assertion());
    const uint elmXnbr = keyX == key1 ? 0 : 1;
    const uint elmYnbr = keyY == key1 ? 0 : 1;
    
    Real prefactor = ep.area() * (1./6.);
    concepts::Real3d normal = ep.normal();
    
    Real sign = keyY == key1 ? 1 : -1;
    Real factor = sign * prefactor;
    
    Real* gradientNormal = new Real[m];
    for(uint p=0; p<m; ++p) {
      gradientNormal[p] = (elmX.shapeFctGradient(p) * normal);
    }
    for(uint q=0; q<n; ++q) {
      Real sum = 0;
      for(uint i=0; i<3; ++i) {
	sum += elmY.shapeFctEdgeCenter(q, ep.edgeIndex(elmYnbr, i));
      }
      sum *= factor;
      for(uint p=0; p<m; ++p) {
	em(p,q) -= sum * gradientNormal[p];
      }
    }
    delete[] gradientNormal;
    
    sign = keyX == key1 ? 1 : -1;
    factor = symm_ * sign * prefactor;
    
    gradientNormal = new Real[n];
    for(uint q=0; q<n; ++q) {
      gradientNormal[q] = (elmY.shapeFctGradient(q) * normal);
    }
    for(uint p=0; p<m; ++p) {
      Real sum = 0;
      for(uint i=0; i<3; ++i) {
	sum += elmX.shapeFctEdgeCenter(p, ep.edgeIndex(elmXnbr, i));
      }
      sum *= factor;
      for(uint q=0; q<m; ++q) {
	em(p,q) += sum * gradientNormal[q];
      }
    }
    delete[] gradientNormal;
  }

  //****************************************************** LaplaceBoundaryBF **
  void LaplaceBoundaryBF::operator()(const FvdgElement& elmX,
				   const FvdgElement& elmY,
				   concepts::ElementMatrix<Real>& em,
				   const FvdgElementPair& ep) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();
    
    const concepts::Key keyX = elmX.support().key();
    const concepts::Key keyY = elmY.support().key();
    const concepts::Key key1 = ep.elm1().support().key();
    const concepts::Key key2 = ep.elm2().support().key();
    
    conceptsAssert(keyX == keyY, concepts::Assertion());
    conceptsAssert(keyX == key1, concepts::Assertion());
    conceptsAssert(key1 == key2, concepts::Assertion());
    
    Real prefactor = ep.area() * (1./3.);
    
    for(uint q=0; q<n; ++q) {
      Real sum = 0;
      for(uint i=0; i<3; ++i) {
	sum += elmY.shapeFctEdgeCenter(q, ep.edgeIndex(0, i));
      }
      sum *= prefactor;
      for(uint p=0; p<m; ++p) {
	em(p,q) -= sum * (elmX.shapeFctGradient(p) * ep.normal());
      }
    }
    
    prefactor *= symm_;
    
    for(uint p=0; p<m; ++p) {
      Real sum = 0;
      for(uint i=0; i<3; ++i) {
	sum += elmX.shapeFctEdgeCenter(p, ep.edgeIndex(0, i));
      }
	sum *= prefactor;
	for(uint q=0; q<m; ++q) {
	  em(p,q) += sum * (elmY.shapeFctGradient(q) * ep.normal());
	}
    }
  }
  
  //************************************************************** PenaltyBF **
  void PenaltyBF::operator()(const FvdgElement& elmX, const FvdgElement& elmY,
			     concepts::ElementMatrix<Real>& em,
			     const FvdgElementPair& ep) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();
    
    const concepts::Key keyX = elmX.support().key();
    const concepts::Key keyY = elmY.support().key();
    const concepts::Key key1 = ep.elm1().support().key();
    const concepts::Key key2 = ep.elm2().support().key();
    conceptsAssert( (keyX == key1) || (keyX == key2), concepts::Assertion());
    conceptsAssert( (keyY == key1) || (keyY == key2), concepts::Assertion());
    const uint elmXnbr = keyX == key1 ? 0 : 1;
    const uint elmYnbr = keyY == key1 ? 0 : 1;
    
    Real prefactor = keyX == keyY ? 1 : -1;
    prefactor *= ep.area() * (1./3.) / ep.diam();
    for(uint p=0; p<m; ++p) {
      for(uint q=0; q<n; ++q) {
	Real sum = 0;
	for(uint i=0; i<3; ++i) {
	  sum += elmX.shapeFctEdgeCenter(p, ep.edgeIndex(elmXnbr, i)) *
	    elmY.shapeFctEdgeCenter(q, ep.edgeIndex(elmYnbr, i));
	}
	em(p,q) = sum * prefactor;
      }
    }
  }

} // namespace linDG3D
