/* Classes for DG bilinear forms
 */

#ifndef dgBilinearForms_hh
#define dgBilinearForms_hh

#include "basics/typedefs.hh"
#include "elementPair.hh"
#include "operator/bilinearForm.hh"
#include "operator/matrix.hh"

#define DEBUG_ASSEMBLY 0
#define DEBUG_IDENTITY 0
#define DEBUG_LAPLACE 0

namespace linDG3D {

  using concepts::Real;
  // ***************************************************** FvdgBilinearForm  **
  
  /** Bilinear Form for FV/DG purpose
      @author Christoph Winkelmann, 2003
  */
  class FvdgBilinearForm {
  public:
    /** Evaluates the bilinear form for all shape functions on elmX and elmY
	and stores the result in the matrix em.
	
	@post The returned matrix em has the correct size.
	@param elmX Left element
	@param elmY Right element
	@param em Return element matrix
	@param ep Element pair
    */
    virtual void operator()(const FvdgElement& elmX, const FvdgElement& elmY,
			    concepts::ElementMatrix<Real>& em,
			    const FvdgElementPair& ep) = 0;
  };
  

  // ************************************************************** Assembly **

  /** Provides methods to assemby matrices by lists of FvdgElementPairs.
      @author Christoph Winkelmann, 2003
  */
  class Assembly
  {
  public:
    /** Assembly operator scanning a list of inner element pairs. For each
	element pair (i,j) in \c pairs, The bilinear form \c bf is called
	four times: with i,i, i,j, j,i, and j,j.
	@param dest The destination matrix where the terms are assembled to
	@param bf The bilinear form called on found element pairs
	@param pairs The list of inner ElementPairs scanned
	@param transpose Flag to transpose the global matrix while assembling
    */
    static void inner(concepts::Matrix<Real>& dest,
		      FvdgBilinearForm& bf,
		      const concepts::ElementPairList<Real>& pairs,
		      bool transpose = false);
    /** Assembly operator scanning a list of boundary element pairs. For each
	element pair (i,i) in \c pairs, The bilinear form \c bf is called once
	with elements i,i.
	@param dest The destination matrix where the terms are assembled to
	@param bf The bilinear form called on found element pairs
	@param pairs The list of boundary ElementPairs scanned
	@param transpose Flag to transpose the global matrix while assembling
    */
    static void boundary(concepts::Matrix<Real>& dest,
			 FvdgBilinearForm& bf,
			 const concepts::ElementPairList<Real>& pairs,
			 bool transpose = false);
  private:
    /// Constructor. Is private in order to prevent creating an instance.
    Assembly() {};
  };
  
  // ************************************************************** Identity **
  /** Identity bilinear form. The application operator of the class
      <span class="code">Identity</span> computes the mass matrix for
      FvdgElements.
      @author Christoph Winkelmann, 2002
  */
  class Identity : public concepts::BilinearForm<Real> {
  public:
    Identity() {}
    /// General application operator.
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em);
    /// Application operator for P0 elements
    void operator()(const FvdgP0TetElem& eX,
		    const FvdgP0TetElem& eY,
		    concepts::ElementMatrix<Real>& em);
    /// Application operator for P1 elements
    void operator()(const FvdgP1TetElem& eX,
		    const FvdgP1TetElem& eY,
		    concepts::ElementMatrix<Real>& em);
  };
  
  //*********************************************************** LaplaceVolBF **
  /** Laplace bilinear form. The application operator of the class
      <span class="code">LaplaceVolBF</span> computes the volume integral 
      matrix for FvdgElements.
      @author Christoph Winkelmann, 2003
  */
  class LaplaceVolBF : public concepts::BilinearForm<Real> {
  public:
    LaplaceVolBF() {}
    /// General application operator.
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em);
    /// Application operator for P1 elements
    void operator()(const FvdgP1TetElem& eX,
		    const FvdgP1TetElem& eY,
		    concepts::ElementMatrix<Real>& em);
  };
  
  //********************************************************* LaplaceInnerBF **
  /** Laplace bilinear form. The application operator of the class
      <span class="code">LaplaceInnerBF</span> computes the inner face
      integral matrix for FvdgElements.
      @author Christoph Winkelmann, 2003
  */
  class LaplaceInnerBF : public FvdgBilinearForm {
  public:
    /** Constructor. Creates a symmetric bilinear form if \c symm is true. 
	Otherwise, an antisymmetric one is created.
	@param symm symmetry flag
    */
    LaplaceInnerBF(bool symm = false) { symm_ = symm ? -1 : 1; }
    /// Application operator
    virtual void operator()(const FvdgElement& elmX, const FvdgElement& elmY,
			    concepts::ElementMatrix<Real>& em,
			    const FvdgElementPair& ep);
  private:
    /// symmetry sign
    int symm_;
  };

  //****************************************************** LaplaceBoundaryBF **
  /** Laplace bilinear form. The application operator of the class
      <span class="code">LaplaceBoundaryBF</span> computes the boundary face
      integral matrix for FvdgElements.
      @author Christoph Winkelmann, 2003
  */
  class LaplaceBoundaryBF : public FvdgBilinearForm {
  public:
    /** Constructor. Creates a symmetric bilinear form if \c symm is true. 
	Otherwise, an antisymmetric one is created.
	@param symm symmetry flag
    */
    LaplaceBoundaryBF(bool symm = false) {
      symm_ = symm ? -1 : 1;
    }
    /// Application operator
    virtual void operator()(const FvdgElement& elmX, const FvdgElement& elmY,
			    concepts::ElementMatrix<Real>& em,
			    const FvdgElementPair& ep);
  private:
    /// symmetry sign
    int symm_;
  };
  
  //************************************************************** PenaltyBF **
  /** Penalty bilinear form. The application operator of the class
      <span class="code">PenaltyBF</span> computes the inner and the boundary
      face integral matrix for FvdgElements.
      @author Christoph Winkelmann, 2003
  */
  class PenaltyBF : public FvdgBilinearForm {
  public:
    /// Constructor.
    PenaltyBF() {}
    /// Application operator
    virtual void operator()(const FvdgElement& elmX, const FvdgElement& elmY,
			    concepts::ElementMatrix<Real>& em,
			    const FvdgElementPair& ep);
  };
  
} // namespace linDG3D

#endif // dgBilinearForms_hh
