/* Class for DG linear forms
 */

#ifndef dgLinearForms_hh
#define dgLinearForms_hh

#include "basics/typedefs.hh"
#include "element.hh"
#include "timestepping/vectors.hh"
#include "toolbox/formula.hh"
#include "toolbox/boundary.hh"
#include "geometry/boundaryConditions.hh"

#include <vector>

#define DEBUG_FILLDATA 0

namespace linDG3D {
  using concepts::Real;

  // ************************************************************** ConstRhs **

  /** TimeLinearForm for constant right hand side. Works for all FvdgElements,
      provided the following assumption holds: The first shape function is
      constant = 1, all others have vanishing mean.
      @author Christoph Winkelmann, 2002
  */
  class ConstRhs : public timestepping::TimeLinearForm {
  public:
    ConstRhs(Real c = 1.0) : c_(c) {}
    /** Application operator. Tries to convert elements to FvdgElements and 
        throws exception when conversion fails.
        @param elm Element
        @param em Element matrix
    */
    virtual void operator()(const concepts::Element<Real>& elm,
                            concepts::ElementMatrix<Real>& em);
    /** Application operator. Builds element matrix.
        @param elm Element
        @param em Element matrix	
    */
    void operator()(const FvdgElement& elm,
                    concepts::ElementMatrix<Real>& em);
  private:
    Real c_;
  };

  // *********************************************************** FvdgTimeRhs **
  /// Fvdg right hand side for time dependent problems
  class FvdgTimeRhs : public timestepping::TimeLinearForm {
  public:
    /** Constructor. Takes a formula object that describes the time dependent
        form of the right hand side.
        @param frm Time dependent formula
    */
    FvdgTimeRhs(concepts::Formula& frm) :
      timestepping::TimeLinearForm(),
      frm_(frm) {}
    /** Application operator. Computes the element contribution.
        @param elmBase Element
        @param em Element matrix
    */
    void operator() (const concepts::Element<Real> &elmBase,
                     concepts::ElementMatrix<Real> &em);
  private:
    const concepts::Formula& frm_;
  };

  //******************************************************** FvdgTimeRhsDiff **
    /// Fvdg right hand side for time dependent problems with diffusion
  class FvdgTimeRhsDiff : public timestepping::TimeLinearForm {
  public:
    /** Constructor. Takes a formula object that describes the time dependent
        form of the right hand side.
        @param frm Time dependent formula
        @param eps weighting coefficient of the diffusion terms
        @param bc boundary conditions
        @param symm flag for symmetric discretization of diffusion
    */
    FvdgTimeRhsDiff(concepts::Formula& frm, Real eps,
                    concepts::BoundaryConditions* bc, bool symm = false) :
      timestepping::TimeLinearForm(), frm_(frm), eps_(eps), bc_(bc) {
        symm_ = symm ? -1 : 1;
      }
    /** Application operator. Computes the element contribution.
        @param elmBase Element
        @param em Element matrix
    */
    void operator() (const concepts::Element<Real> &elmBase,
                     concepts::ElementMatrix<Real> &em);
  private:
    const concepts::Formula& frm_;
    const Real eps_;
    const concepts::BoundaryConditions* bc_;
    int symm_;
    };

} // namespace linDG3D

#endif //dgLinearForms_hh
