#include "space.hh"

namespace linDG3D {
  
  //************************************************************** FvdgSpace **
  std::ostream& FvdgSpace::info(std::ostream& os) const {
    return os << "FvdgSpace(dim = " << dim_ << ", nelm = " << nelm_ << ')';
  }
  
  void FvdgSpace::createElementPairList_() {
    // fill in elm1 and elm2 (index is key of a face) with pointers to elements
    std::map<uint, const FvdgElement*> elm1, elm2;
    std::map<uint, uint> idx1, idx2;
    std::auto_ptr<Scanner> sc(scan());
    while (*sc) {
      const FvdgElement& elm = (*sc)++;
      const concepts::Tetrahedron& cntr3 = elm.cell().connector();
      DEBUGL(DEBUG_SPACE, cntr3);
      const concepts::Connector2* face = 0;
      for (uint j = 0; (face = cntr3.face(j)) != 0; ++j) {
	// for face 0, the orientation in the topology is flipped
	const bool rho = (j == 0) ? (! cntr3.rho(j)) : cntr3.rho(j);
	DEBUGL(DEBUG_SPACE, *face << ", rho = " << cntr3.rho(j));
	const uint key = face->key();
	if (rho) {
	  // elm1
	  conceptsAssert( (elm1.find(key) == elm1.end()) ||
			  (elm1.find(key)->second == 0),
			  concepts::Assertion());
	  elm1[key] = &elm; idx1[key] = j;
	  if (elm2.find(key) == elm2.end())
	    elm2[key] = 0;
	} else {
	  // elm2
	  conceptsAssert( (elm2.find(key) == elm2.end()) ||
			  (elm2.find(key)->second == 0),
			  concepts::Assertion());
	  if (elm1.find(key) == elm1.end())
	    elm1[key] = 0;
	  elm2[key] = &elm; idx2[key] = j;
	}
      } // for faces (j)
    } // for elements
    conceptsAssert(elm1.size() == elm2.size(), concepts::Assertion());
    
    // add all faces to boundaryElmPairList_ and innerElmPairList_
    for(std::map<uint, const FvdgElement*>::const_iterator j =
	  elm1.begin(), k = elm2.begin();
	j != elm1.end(); ++j, ++k) {
      if (j->second == 0) {
	// boundary face
	boundaryElmPairList_.add(new FvdgElementPair(*(k->second),
						     idx2[k->first]));
	continue;
      }
      if (k->second == 0) {
	// boundary face
	boundaryElmPairList_.add(new FvdgElementPair(*(j->second),
						     idx1[j->first]));
	continue;
      } // inner face
      innerElmPairList_.add(new FvdgElementPair(*(j->second), *(k->second),
						idx1[j->first],
						idx2[k->first]));
    }
  }
  
  //************************************************************ FvdgSpaceP0 **
  FvdgSpaceP0::FvdgSpaceP0(concepts::Mesh3& msh,
			   concepts::BoundaryConditions* bc) : FvdgSpace(bc) {
    std::auto_ptr<concepts::Scan3> sc(0);
    sc.reset(msh.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      const concepts::Tetrahedron3d* cellT =
	dynamic_cast<const concepts::Tetrahedron3d*>(&cell);
      conceptsAssert(cellT, concepts::Assertion());
      uint idx[1];
      idx[0] = dim_++;
      nelm_++;
      elm_ = new concepts::Joiner<FvdgElement*, 1>
	(new FvdgP0TetElem(*cellT, idx), elm_);
    }
    createElementPairList_();
  }
  
  std::ostream& FvdgSpaceP0::info(std::ostream& os) const {
    return os << "FvdgSpaceP0(dim = " << dim_ << ", nelm = " << nelm_ << ')';
  }

  //************************************************************ FvdgSpaceP1 **
  FvdgSpaceP1::FvdgSpaceP1(concepts::Mesh3& msh,
			   concepts::BoundaryConditions* bc) : FvdgSpace(bc) {
    std::auto_ptr<concepts::Scan3> sc(0);
    sc.reset(msh.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      const concepts::Tetrahedron3d* cellT =
	dynamic_cast<const concepts::Tetrahedron3d*>(&cell);
      conceptsAssert(cellT, concepts::Assertion());
      uint idx[4];
      for(int i=0; i<4; ++i) {
	idx[i] = dim_++;
      }
      nelm_++;
      elm_ = new concepts::Joiner<FvdgElement*, 1>
	(new FvdgP1TetElem(*cellT, idx), elm_);
    }
    createElementPairList_();
  }
  
  std::ostream& FvdgSpaceP1::info(std::ostream& os) const {
    return os << "FvdgSpaceP1(dim = " << dim_ << ", nelm = " << nelm_ << ')';
  }

} // namespace linDG3D
