/* Classes for DG spaces
 */

#ifndef dgSpace_hh
#define dgSpace_hh

#include "basics/typedefs.hh"
#include "element.hh"
#include "elementPair.hh"
#include "space/space.hh"
#include "space/elementPairs.hh"
#include "toolbox/scannerConnectors.hh"
#include "geometry/boundaryConditions.hh"
#include "geometry/mesh.hh"
#include "graphics/spaceTraits.hh"

#include <map>
#include <memory>

#define DEBUG_SPACE 0

namespace concepts {

  // *************************************************************** Scanner **
  /// Scanner over tetrahedral elements in FV/DG-space
  class Scan<linDG3D::FvdgElement> : public Scan<Element<Real> > {
  public:
    /// Returns next element in scanned space
    linDG3D::FvdgElement& operator++(int) = 0;
  };
} // namespace concepts

namespace linDG3D {
  using concepts::Real;

// ****************************************************************** Spaces **

/** Generic space of tetrahedral elements for FV/DGFE method
    @author Christoph Winkelmann, 2002
*/
  class FvdgSpace : public concepts::Space<Real> {
  public:
    virtual ~FvdgSpace() 
    { concepts::Joiner<FvdgElement*, 1>::destructor(elm_); }
    typedef concepts::Scan<FvdgElement> Scanner;
    /// Returns the dimension of the space
    uint dim() const { return dim_; }
    /// Returns the number of elements in the space
    uint nelm() const { return nelm_; }
    /// Returns a new scanner over the elements in the space
    Scanner* scan() const {
      return new concepts::PListScan<FvdgElement>(*elm_);
    }
    /// Returns a reference to the boundary element pair list.
    const concepts::ElementPairList<Real>& boundaryElmPairList() const {
      return boundaryElmPairList_;
    }
    /// Returns a reference to the element pair list of the inner elements.
    const concepts::ElementPairList<Real>& innerElmPairList() const {
      return innerElmPairList_;
    }
    
  protected:
    /** Constructor. It is protected in order not to allow the user to create
	an instance of the base class.
    */
    FvdgSpace::FvdgSpace(concepts::BoundaryConditions* bc) :
      dim_(0), nelm_(0), bc_(bc), elm_(0), innerElmPairList_(0) {};
    /// Returns the dimension and the number of the elements in the space
    virtual std::ostream& info(std::ostream& os) const;
    uint dim_;
    uint nelm_;
    concepts::BoundaryConditions* bc_;
    concepts::Joiner<FvdgElement*, 1>* elm_;
    
    concepts::ElementPairList<Real> innerElmPairList_;
    concepts::ElementPairList<Real> boundaryElmPairList_;
    
    void createElementPairList_();
  };
  
  //************************************************************ FvdgSpaceP0 **
  /** Space of piecewise constant functions on tetrahedrons
      @author Christoph Winkelmann
  */
  
  class FvdgSpaceP0 : public FvdgSpace {
  public:
  /** Constructor.
      @param msh Mesh on which the space is created
      @param bc Boundary conditions for the space
  */
    FvdgSpaceP0(concepts::Mesh3& msh, concepts::BoundaryConditions* bc = 0);
  protected:
    /// Returns the dimension and the number of the elements in the space
    virtual std::ostream& info(std::ostream& os) const;
  };

  //************************************************************ FvdgSpaceP1 **
  /** Space of piecewise linear functions on tetrahedrons
      @author Christoph Winkelmann, 2002
  */
  
  class FvdgSpaceP1 : public FvdgSpace {
  public:
    /** Constructor.
	@param msh Mesh on which the space is created
	@param bc Boundary conditions for the space
    */
    FvdgSpaceP1(concepts::Mesh3& msh, concepts::BoundaryConditions* bc = 0);
  protected:
    /// Returns the dimension and the number of the elements in the space
    virtual std::ostream& info(std::ostream& os) const;
  };

} // namespace linDG3D

// *********************************************** Space Traits for Graphics **

namespace graphics {
/** Traits for FvdgSpace. */
  template<>
  struct spaceTraits<linDG3D::FvdgSpace> {
    static inline uint dim() { return 3; }
    static inline bool positionConnection() { return false; }
    static inline bool boundaryElements() { return false; }
  };
/** Traits for FvdgSpaceP0. */
  template<>
  struct spaceTraits<linDG3D::FvdgSpaceP0> {
    static inline uint dim() { return 3; }
    static inline bool positionConnection() { return false; }
    static inline bool boundaryElements() { return false; }
  };
/** Traits for FvdgSpaceP1. */
  template<>
  struct spaceTraits<linDG3D::FvdgSpaceP1> {
    static inline uint dim() { return 3; }
    static inline bool positionConnection() { return false; }
    static inline bool boundaryElements() { return false; }
  };
} // namespace graphics

#endif // dgSpace_hh
