// bilinear forms for quads and triangles for the Laplacian and the mass matrix

#include <cmath>
#include <sstream>
#include "integration/quadrature.hh"
#include "bilinearForm2D.hh"
#include "element2D.hh"

namespace linearFEM {

  // *************************************************************** Laplace **

  void Laplace2d::operator()(const concepts::Element<Real>& elmX,
			     const concepts::Element<Real>& elmY,
			     concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    if (em.size()) {
      const linearFEM::Quad* eQX = dynamic_cast<const linearFEM::Quad*>(&elmX);
      if (eQX) {
	const linearFEM::Quad* eQY =
	  dynamic_cast<const linearFEM::Quad*>(&elmY);
	if (eQY) { this->operator()(*eQX, *eQY, em); return; }
      }
      const linearFEM::Triangle* eTX =
	dynamic_cast<const linearFEM::Triangle*>(&elmX);
      if (eTX) {
	const linearFEM::Triangle* eTY =
	  dynamic_cast<const linearFEM::Triangle*>(&elmY);
	if (eTY) {
	  this->operator()(*eTX, *eTY, em); return; }
      }
      std::stringstream error;
      error << "element " << elmX << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }
  }

  void Laplace2d::operator()(const linearFEM::Quad& elmX,
			     const linearFEM::Quad& elmY,
			     concepts::ElementMatrix<Real>& em) {
    const uint n = elmX.T().m();
    if (elmX.support().key() == elmY.support().key()) {
      concepts::Real2d y0(elmX.cell().vertex(0));
      concepts::Real2d y1(elmX.cell().vertex(1));
      concepts::Real2d y2(elmX.cell().vertex(2));
      concepts::Real2d y3(elmX.cell().vertex(3));

      concepts::Real2d p(y1 - y0);
      concepts::Real2d q(y3 - y0);
      concepts::Real2d r(y1 - y0 + y3 - y2);

      concepts::Real2d s, t;

      const concepts::Quadrature<4> quadrature(3);
      const Real* abscissas = quadrature.abscissas();
      const Real* weights = quadrature.weights();
      const uint qn = quadrature.n();

      shpfct_.resize(n*qn*qn);
      elmX.evaluateD((concepts::Real2d*)shpfct_, abscissas, qn);

      for (uint qi = 0; qi < qn; ++qi) {
	t.lincomb(q, r, 1.0, -(1.0+abscissas[qi])/2.0);
	for (uint qj = 0; qj < qn; ++qj) {
	  s.lincomb(p, r, 1.0, -(1.0+abscissas[qj])/2.0);
	  const Real dx = std::fabs(4.0 / s.cross(t))
	    * weights[qi] * weights[qj] / 4.0;
	  const Real ss = dx * s.l2_2();
	  const Real st = -dx * (s*t);
	  const Real tt = dx * t.l2_2();
	  for (uint i = 0; i < n; ++i) {
	    concepts::Real2d shpfcti = shpfct_[i*qn*qn+qi*qn+qj];
	    for (uint j = i; j < n; ++j) {
	      concepts::Real2d shpfctj = shpfct_[j*qn*qn+qi*qn+qj];
	      em(i, j) += shpfcti[0] * shpfctj[0] * tt
		+ (shpfcti[0] * shpfctj[1] + shpfcti[1] * shpfctj[0]) * st
		+ shpfcti[1] * shpfctj[1] * ss;
	    }
	  }
	}
      }
      for(uint i = 0; i < n; ++i) 
	for (uint j = i + 1; j < n; ++j) 
	  em(j, i) = em(i, j);
    }
  }

  void Laplace2d::operator()(const linearFEM::Triangle& elmX,
			     const linearFEM::Triangle& elmY,
			     concepts::ElementMatrix<Real>& em) {
    const uint n = elmX.T().m();
    if (elmX.support().key() == elmY.support().key()) {
      concepts::Real2d x0(elmX.cell().vertex(0));
      concepts::Real2d x1(elmX.cell().vertex(1));
      concepts::Real2d x2(elmX.cell().vertex(2));
      // volume of the triangle
      const concepts::MapReal3d volMat(1.0, 1.0, 1.0,
				       x0[0], x1[0], x2[0],
				       x0[1], x1[1], x2[1]);
      const Real volTriangle_4 = 1.0 / (2.0 * volMat.determinant());
      // The stiffness matrix is easily computable from the coordinates of
      // the corners of the triangle
      concepts::Real2d p1_p2(x1 - x2);
      concepts::Real2d p2_p0(x2 - x0);
      concepts::Real2d p0_p1(x0 - x1);

      em(0,0) = p1_p2.l2_2() * volTriangle_4;
      em(0,1) = p1_p2*p2_p0 * volTriangle_4;
      em(0,2) = p1_p2*p0_p1 * volTriangle_4;
      em(1,1) = p2_p0.l2_2() * volTriangle_4;
      em(1,2) = p2_p0*p0_p1 * volTriangle_4;
      em(2,2) = p0_p1.l2_2() * volTriangle_4;

      for (uint i = 0; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
    }
  }

  // ************************************************************** Identity **

  void Identity2d::operator()(const concepts::Element<Real>& elmX,
			      const concepts::Element<Real>& elmY,
			      concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    if (em.size()) {
      const linearFEM::Quad* eQX = dynamic_cast<const linearFEM::Quad*>(&elmX);
      if (eQX) {
	const linearFEM::Quad* eQY =
	  dynamic_cast<const linearFEM::Quad*>(&elmY);
	if (eQY) { this->operator()(*eQX, *eQY, em); return; }
      }
      const linearFEM::Triangle* eTX =
	dynamic_cast<const linearFEM::Triangle*>(&elmX);
      if (eTX) {
	const linearFEM::Triangle* eTY =
	  dynamic_cast<const linearFEM::Triangle*>(&elmY);
	if (eTY) {
	  this->operator()(*eTX, *eTY, em); return; }
      }
      std::stringstream error;
      error << "element " << elmX << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }
  }

  void Identity2d::operator()(const linearFEM::Quad& elmX,
			      const linearFEM::Quad& elmY,
			      concepts::ElementMatrix<Real>& em) {
    const uint n = elmX.T().m();
    if (elmX.support().key() == elmY.support().key()) {
      concepts::Real2d y0(elmX.cell().vertex(0));
      concepts::Real2d y1(elmX.cell().vertex(1));
      concepts::Real2d y2(elmX.cell().vertex(2));
      concepts::Real2d y3(elmX.cell().vertex(3));

      concepts::Real2d p(y1 - y0);
      concepts::Real2d q(y3 - y0);
      concepts::Real2d r(y1 - y0 + y3 - y2);

      concepts::Real2d s, t;

      const concepts::Quadrature<4> quadrature(3);
      const Real* abscissas = quadrature.abscissas();
      const Real* weights = quadrature.weights();
      const uint qn = quadrature.n();

      shpfct_.resize(n*qn*qn);
      elmX.evaluate((Real*)shpfct_, abscissas, qn);

      for (uint qi = 0; qi < qn; ++qi) {
	t.lincomb(q, r, 1.0, -(1.0+abscissas[qi])/2.0);
	for (uint qj = 0; qj < qn; ++qj) {
	  s.lincomb(p, r, 1.0, -(1.0+abscissas[qj])/2.0);
	  const Real dx = std::fabs(s.cross(t)) /
	    4.0 * weights[qi] * weights[qj];
	  for (uint i = 0; i < n; ++i) {
	    concepts::Real shpfcti = shpfct_[i*qn*qn+qi*qn+qj];
	    for (uint j = i; j < n; ++j) {
	      concepts::Real shpfctj = shpfct_[j*qn*qn+qi*qn+qj];
	      em(i, j) += shpfcti * shpfctj * dx;
	    }
	  }
	}
      }
      for(uint i = 0; i < n; ++i) 
	for (uint j = i + 1; j < n; ++j) 
	  em(j, i) = em(i, j);
    }
  }

  void Identity2d::operator()(const linearFEM::Triangle& elmX,
			      const linearFEM::Triangle& elmY,
			      concepts::ElementMatrix<Real>& em) {
    const uint n = elmX.T().m();
    if (elmX.support().key() == elmY.support().key()) {
      concepts::Real2d x0(elmX.cell().vertex(0));
      concepts::Real2d x1(elmX.cell().vertex(1));
      concepts::Real2d x2(elmX.cell().vertex(2));
      // volume of the triangle
      const concepts::MapReal3d volMat(1.0, 1.0, 1.0,
				       x0[0], x1[0], x2[0],
				       x0[1], x1[1], x2[1]);
      const Real volTriangle_12 = volMat.determinant()/24.0;

      em(2,2) = em(1,1) = em(0,0) = volTriangle_12*2.0;
      em(0,1) = em(0,2) = em(1,2) = volTriangle_12;

      for (uint i = 0; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
    }
  }

} // namespace linearFEM
