// bilinear forms for quads and triangles for the Laplacian and the mass matrix

#ifndef bilinearLinFEM_hh
#define bilinearLinFEM_hh

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "operator/bilinearForm.hh"
#include "toolbox/array.hh"

namespace linearFEM {

  using concepts::Real;

  class Quad;
  class Triangle;

  // ************************************************************* Laplace2d **

  /** Discrete equivalent of the Laplacian in 2D for linear FEM. This
      bilinear form computes the stiffness matrix resulting from the
      discretization of the Laplacian (with integration by parts):
      \f[ \int_K \nabla \phi_i \cdot \nabla \phi_j \, dx \f]
      for the element shape functions \f$\phi_i\f$.
      @author Philipp Frauenfelder, 2002
  */
  class Laplace2d : public concepts::BilinearForm<Real> {
  public:
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em);
    /** Computes the element stiffness matrix for a quadrilateral.
	The stiffness matrix has to be integrated numericaly since the
	Jacobian of the element map is not constant.
    */
    void operator()(const linearFEM::Quad& elmX,
		    const linearFEM::Quad& elmY,
		    concepts::ElementMatrix<Real>& em);
    /** Computes the element stiffness matrix for a triangle.
	The stiffness matrix is precomputed since the element map is constant.
    */
    void operator()(const linearFEM::Triangle& elmX,
		    const linearFEM::Triangle& elmY,
		    concepts::ElementMatrix<Real>& em);
  private:
    /// Storage area for values of derivatives of shape functions
    concepts::Array<concepts::Real2d> shpfct_;
  };

  // ************************************************************ Identity2d **

  /** Discrete equivalent for a reaction term in 2D for linear FEM. This
      bilinear form computes the mass matrix resulting from the discretization
      of a reaction term with reaction coefficient 1:
      \f[ \int_K \phi_i \phi_j \, dx \f]
      for the element shape functions \f$\phi_i\f$.
      @author Philipp Frauenfelder, 2002
  */
  class Identity2d : public concepts::BilinearForm<Real> {
  public:
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em);
    /** Computes the element mass matrix for a quadrilateral.
	The mass matrix has to be integrated numericaly since the
	Jacobian of the element map is not constant.
    */
    void operator()(const linearFEM::Quad& elmX,
		    const linearFEM::Quad& elmY,
		    concepts::ElementMatrix<Real>& em);
    /** Computes the element mass matrix for a triangle.
	The mass matrix is precomputed since the element map is constant.
    */
    void operator()(const linearFEM::Triangle& elmX,
		    const linearFEM::Triangle& elmY,
		    concepts::ElementMatrix<Real>& em);
  private:
    /// Storage area for values of shape functions
    concepts::Array<Real> shpfct_;
  };

} // namespace linearFEM

#endif // bilinearLinFEM_hh
