// space for linear FEM in 1D

#ifndef linSpace1D_hh
#define linSpace1D_hh

#include "element1D.hh"
#include "basics/typedefs.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "toolbox/scannerConnectors.hh"
#include "toolbox/dynArray.hh"

namespace concepts {

  // forward declarations
  class Mesh1;
  class BoundaryConditions;

  // ************************************************************** Scanners **

  template<>
  class Scan<linearFEM::Line> : public Scan<linearFEM::Element> {
  public:
    virtual linearFEM::Line& operator++(int) = 0;
  };

} // namespace concepts

namespace linearFEM {

  // ************************************************************** Linear1d **

  /** Space for linear FEM in 1D.
      @author Philipp Frauenfelder, 2002
  */
  class Linear1d : public concepts::Space<Real> {
  public:
    typedef concepts::Scan<Line> Scan;
    typedef void (*SMap)(const concepts::TColumn<Real>&,
			 concepts::TColumn<Real>&);
    /** Constructor. Recursively calls \c enforceBC_ and \c constructor_
	@param msh Mesh
	@param level Number levels for initial refinement
	@param bc Boundary conditions
    */
    Linear1d(concepts::Mesh1& msh, uint level,
	     concepts::BoundaryConditions* bc = 0);
    virtual ~Linear1d();
    virtual uint dim() const { return dim_; }
    virtual uint nelm() const { return nelm_; }
    virtual Scan* scan() const { return new concepts::PListScan<Line>(*elm_); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Number of degrees of freedom
    uint dim_;
    /// Number of elements
    uint nelm_;
    /// Mesh which this space is based on
    concepts::Mesh1& msh_;
    /// Boundary conditions
    concepts::BoundaryConditions* bc_;
    /// List of elements
    concepts::Joiner<Line*, 1>* elm_;
    /** Notes where Dirichlet boundary conditions have to be enforced. This
	information is stored in \c bcMap.

	\c enforceBC_ is called recursively until \c level is 0. On this
	level, the boundary conditions are checked.
	@param cell Current cell
	@param level Number of levels left to desired level
	@param bcMap Stores boundary condition flag for each vertex
    */
    void enforceBC_(concepts::Cell1& cell, uint level,
		    concepts::DynArray<uint>& bcMap);
    /** Assigns the degrees of freedom and creates elements. The numbers of
	the degrees of freedom is stored in \c dofMap.

	\c constructor_ is called recursively until \c level is 0. On this
	level, the degrees of freedom are assigned and the elements are
	created and stored in \c elm_.
	@param cell Current cell
	@param level Number of levels left to desired level
	@param dofMap Stores assigned dof for the vertices
	@param bcMap Stores boundary condition flag for each vertex
    */
    void constructor_(concepts::Cell1& cell, uint level,
		      concepts::DynArray<uint>& dofMap,
		      const concepts::DynArray<uint>& bcMap);
  };

} // namespace linearFEM

#endif // linSpace1D_hh
