// space for linear FEM in 2D

#include <sstream>
#include "space2D.hh"
#include "geometry/mesh.hh"
#include "geometry/boundaryConditions.hh"
#include "toolbox/boundary.hh"

// debugging
#include "basics/debug.hh"

#define spaceRebuild0_D 0
#define spaceRebuild1_D 0
#define spaceDestr_D 0

namespace linearFEM {

  // ************************************************************** Linear2d **

  Linear2d::Linear2d(concepts::Mesh2& msh, uint level,
		     concepts::BoundaryConditions* bc) :
    dim_(0), nelm_(0), rebuild_(true), msh_(msh), bc_(bc), elm_(0) {

    std::auto_ptr<concepts::Scan2> sc(msh_.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;
      concepts::Key key(cell.connector().key());
      adj_[key] = level;
      ctrl2_[key] = 1;
    }
  }

  Linear2d::~Linear2d() {
    concepts::Joiner<Element*, 1>::destructor(elm_);
    DEBUGL(spaceDestr_D, "done.");
  }

  Linear2d::Scan* Linear2d::scan() const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return new concepts::PListScan<Element>(*elm_);
  }

  void Linear2d::adjust(const Quad& elm, const short level) {
    adj_[elm.support().key()] = level;
    rebuild_ = true;
  }

  void Linear2d::rebuild() {
    dof_.clear();
    delete elm_;
    elm_ = 0;
    dim_ = 0;
    nelm_ = 0;
    /* For every cell in the coarsest mesh, rebuild0_is called. It
       fills the control arrays up to very small changes for hanging
       nodes (these changes occur in rebuild1_). */
    std::auto_ptr<concepts::Scan2> sc(msh_.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;
      int L = -1;
      rebuild0_(cell.connector(), 0, L);
    }
    /* Next, rebuild1_ is called for every cell in the coarsest
       mesh. It assigns the degrees of freedom and creates the
       elements. */
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;
      concepts::Quad2d* cellQ = dynamic_cast<concepts::Quad2d*>(&cell);
      if (cellQ) {
	rebuild1_(*cellQ);
	continue;
      }
      concepts::Triangle2d* cellT = dynamic_cast<concepts::Triangle2d*>(&cell);
      if (cellT) {
	rebuild1_(*cellT);
	continue;
      }
      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }
    /* At the end, the rebuild_ flag is turned off and the adjustment
       information are deleted. */
    rebuild_ = false;
    adj_.clear();
    ctrl0_.clear();
    ctrl0L_.clear();
    ctrl1_.clear();
    ctrl1L_.clear();
  }

  const Linear2d::SMap& Linear2d::S(uint i) const {
    conceptsAssert(i < 4, concepts::Assertion());
    return S_[i];
  }

  Linear2d::SMap Linear2d::S_[4] = { Linear2d::S0, Linear2d::S1,
				     Linear2d::S2, Linear2d::S3 };

  void Linear2d::S0(const concepts::TColumn<Real>& src,
		    concepts::TColumn<Real>& dst) {
    dst[0] += 1.0 * src[0];
    dst[1] += 0.5 * src[0] + 0.5 * src[1];
    dst[2] += 0.25 * src[0] + 0.25 * src[1] + 0.25 * src[2] + 0.25 * src[3];
    dst[3] += 0.5 * src[0] + 0.5 * src[3];
  }

  void Linear2d::S1(const concepts::TColumn<Real>& src,
		    concepts::TColumn<Real>& dst) {
    dst[0] += 0.5 * src[0] + 0.5 * src[1];
    dst[1] += 1.0 * src[1];
    dst[2] += 0.5 * src[1] + 0.5 * src[2];
    dst[3] += 0.25 * src[0] + 0.25 * src[1] + 0.25 * src[2] + 0.25 * src[3];
  }

  void Linear2d::S2(const concepts::TColumn<Real>& src,
		    concepts::TColumn<Real>& dst) {
    dst[0] += 0.25 * src[0] + 0.25 * src[1] + 0.25 * src[2] + 0.25 * src[3];
    dst[1] += 0.5 * src[1] + 0.5 * src[2];
    dst[2] += 1.0 * src[2];
    dst[3] += 0.5 * src[2] + 0.5 * src[3];
  }

  void Linear2d::S3(const concepts::TColumn<Real>& src,
		    concepts::TColumn<Real>& dst) {
    dst[0] += 0.5 * src[0] + 0.5 * src[3];
    dst[1] += 0.25 * src[0] + 0.25 * src[1] + 0.25 * src[2] + 0.25 * src[3];
    dst[2] += 0.5 * src[2] + 0.5 * src[3];
    dst[3] += 1.0 * src[3];
  }

  std::ostream& Linear2d::info(std::ostream& os) const {
    return os << "Linear2d(dim = " << dim() << ", nelm = " << nelm() << ')';
  }

  void Linear2d::rebuild0_(concepts::Connector2& cntr, const int l, int& L) {
    DEBUGL(spaceRebuild0_D, l << " - " << cntr);
    short& ctrl(ctrl2_[cntr.key()]);
    if ((L == -1) && (ctrl & 1)) {
      short& adj(adj_[cntr.key()]);
      L = l + adj;
      if (L < 0)
	L = 0;
    }
    if (L == -1 || l < L) {
      int Lmax = -1;
      concepts::Connector2* chld = 0;
      for(uint i = 0; (chld = cntr.child(i)) != 0; ++i) {
	int LL = L;
	rebuild0_(*chld, l + 1, LL);
	if (LL > Lmax)
	  Lmax = LL; 
      }
      L = Lmax;
    }
    if (l < L) {
      /* If the cell is not on the maximal level, it is not member of
	 the space. */
      ctrl = 0;
    } else {
      /* On the maximal level, the cell is member of the space and its
	 edges and vertices too. */
      ctrl = 1;
      concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	ctrl1_[cntr1->key()] |= 1;
	ctrl1L_[cntr1->key()] = l;
      }
      concepts::Connector0* cntr0 = 0;
      for (uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i) {
	ctrl0_[cntr0->key()] |= 1;
	ctrl0L_[cntr0->key()] = l;
      }
    }
    /* At the end, the Dirichlet boundary conditions are enforced by
       setting some topological entities to passive. */
    if (bc_) {
      concepts::Connector1* cntr1 = 0;
      for (uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	if ((*bc_)(cntr1->attrib()).type() == concepts::Boundary::DIRICHLET) {
	  ctrl1_[cntr1->key()] |= 2;
	  DEBUGL(spaceRebuild0_D, l << " - " << *cntr1 << " has Dirichlet bc");
	  concepts::Connector0* cntr0 = 0;
	  for (uint j = 0; (cntr0 = cntr1->vertex(j)) != 0; ++j) {
	    ctrl0_[cntr0->key()] |= 2;
	    DEBUGL(spaceRebuild0_D, l << " - " << *cntr0 << " has Dirichlet bc");
	  }
	} // if DIRICHLET bc
      } // for i
    } // if bc_
  }

  void Linear2d::rebuild1_(concepts::Quad2d& cell,
			   concepts::TColumn<Real>* T0) {
    concepts::Quad& cntr = cell.connector();
    short& ctrl2(ctrl2_[cntr.key()]);
    const int level = cell.level().l_[0];
    DEBUGL(spaceRebuild1_D, cell << ", " << (uint)ctrl2 << ", T at " << T0);
    const uint n = 4;
    concepts::TColumn<Real>* T1 = 0;
    /* Loop over all vertices of the current cell. If the vertex is
       member of the space and active, it gets a degree of freedom
       assigned. */
    concepts::Connector0* cntr0 = 0;
    for(uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i) {
      const concepts::Key& key = cntr0->key();
      short& ctrl0(ctrl0_[key]);
      short& ctrl0L(ctrl0L_[key]);
      if ((ctrl0 & 3) && (ctrl0L == level)) {
	if ((ctrl0 & 3) == 1) {
	  if ((ctrl0 & 4) == 0) {
	    dof_[key] = dim_++;
	    ctrl0 |= 4;
	  }
	  T1 = new concepts::TColumn<Real>(n, dof_[key], T1);
	  (*T1)[i] = 1.0;
	}
      }
    }
    concepts::Connector1* cntr1 = 0;
    for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
      short& ctrl1(ctrl1_[cntr1->key()]);
      short& ctrl1L(ctrl1L_[cntr1->key()]);
      if ((ctrl1 & 3) && (ctrl1L == level)) {
	/* If the edge is passive, the children of the edge (and the
	   vertex in the middle) are deactivated. */
	if ((ctrl2 & 1) == 0) deactivate_(*cntr1);
      }
    }
    // If the current cell is member of the space, a new element is created.
    if (ctrl2 & 1) {
      if ((T0 == 0) && (T1 == 0))
	T1 = new concepts::TColumn<Real>(n, 0, T1);
      Quad* elm = new Quad(cell, T0, T1);
      nelm_++;
      elm_ = new concepts::Joiner<Element*, 1>(elm, elm_);
    } else {
      /* If the current cell not member of the space, the S matrices
	 are applied to the already built columns of the T matrix and
	 rebuild0_ is called on the children of the current cell. If
	 no T columns were created, the S matrices are not applied,
	 ie. the S matrices are only used to resolve hanging nodes. */
      concepts::Cell2* child = 0;
      for (uint i = 0; (child = cell.child(i)) != 0; ++i) {
	concepts::TColumn<Real>* T2 = 0;
	for (concepts::TColumn<Real>* TT = T0; TT != NULL; TT = TT->link()) {
	  T2 = new concepts::TColumn<Real>(n, TT->index(), T2);
	  (S(i))(*TT, *T2);
	}
	for (concepts::TColumn<Real>* TT = T1; TT != NULL; TT = TT->link()) {
	  T2 = new concepts::TColumn<Real>(n, TT->index(), T2);
	  (S(i))(*TT, *T2);
	}
	rebuild1_(*cell.child(i), T2);
	// clean up
	while (T2 != 0) {
	  concepts::TColumn<Real>* foo = T2; T2 = T2->link(); delete foo;
	}
      }
    }
    while (T1 != 0) {
      concepts::TColumn<Real>* foo = T1; T1 = T1->link(); delete foo;
    }
  }

  void Linear2d::rebuild1_(concepts::Triangle2d& cell,
			   concepts::TColumn<Real>* T0) {
    concepts::Triangle& cntr = cell.connector();
    short& ctrl2(ctrl2_[cntr.key()]);
    DEBUGL(spaceRebuild1_D, cell << ", " << (uint)ctrl2 << ", T at " << T0);
    const int level = cell.level();
    const uint n = 3;
    concepts::TColumn<Real>* T1 = 0;
    /* Loop over all vertices of the current cell. If the vertex is
       member of the space and active, it gets a degree of freedom
       assigned. */
    concepts::Connector0* cntr0 = 0;
    for(uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i) {
      const concepts::Key& key = cntr0->key();
      short& ctrl0(ctrl0_[key]);
      short& ctrl0L(ctrl0L_[key]);
      if ((ctrl0 & 3) && (ctrl0L == level)) {
	if ((ctrl0 & 3) == 1) {
	  if ((ctrl0 & 4) == 0) {
	    dof_[key] = dim_++;
	    ctrl0 |= 4;
	    DEBUGL(spaceRebuild1_D, *cntr0 << " has dof " << dof_[key]);
	  }
	  T1 = new concepts::TColumn<Real>(n, dof_[key], T1);
	  (*T1)[i] = 1.0;
	}
      }
    }
    concepts::Connector1* cntr1 = 0;
    for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
      short& ctrl1(ctrl1_[cntr1->key()]);
      short& ctrl1L(ctrl1L_[cntr1->key()]);
      if ((ctrl1 & 3) && (ctrl1L == level)) {
	/* If the edge is passive, the children of the edge (and the
	   vertex in the middle) are deactivated. */
	if ((ctrl2 & 1) == 0) deactivate_(*cntr1);
      }
    }
    // If the current cell is member of the space, a new element is created.
    if (ctrl2 & 1) {
      if ((T0 == 0) && (T1 == 0))
	T1 = new concepts::TColumn<Real>(n, 0, T1);
      Triangle* elm = new Triangle(cell, T0, T1);
      nelm_++;
      elm_ = new concepts::Joiner<Element*, 1>(elm, elm_);
    } else {
      /* If the current cell not member of the space, the S matrices
	 are applied to the already built columns of the T matrix and
	 rebuild0_ is called on the children of the current cell. If
	 no T columns were created, the S matrices are not applied,
	 ie. the S matrices are only used to resolve hanging nodes. */
      concepts::Cell2* child = 0;
      for (uint i = 0; (child = cell.child(i)) != 0; ++i) {
	conceptsAssert(T0 == 0, concepts::Assertion());
	conceptsAssert(T1 == 0, concepts::Assertion());
	rebuild1_(*cell.child(i), 0);
      }
    }
    while (T1 != 0) {
      concepts::TColumn<Real>* foo = T1; T1 = T1->link(); delete foo;
    }
  }

  void Linear2d::deactivate_(concepts::Connector1& edg) {
    ctrl1_[edg.child(0)->key()] |= 2;
    ctrl1_[edg.child(1)->key()] |= 2;
    ctrl0_[edg.child(1)->vertex(0)->key()] |= 2;
  }

} // namespace linearFEM
