// space for linear FEM in 2D

#ifndef linSpace2D_hh
#define linSpace2D_hh

#if __ICC == 800
#  include <hash_map>
#  define __gnu_cxx std
#else
#  if __GNUC__ == 3
#    include <ext/hash_map>
#    if __GNUC_MINOR__ < 1
#      define __gnu_cxx std
#    endif
#  else
#    include <hash_map>
#    define __gnu_cxx std
#  endif
#endif

#include "element2D.hh"
#include "basics/typedefs.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "toolbox/scannerConnectors.hh"

namespace concepts {

  // forward declarations
  class Mesh2;
  class BoundaryConditions;

  // ************************************************************** Scanners **

  template<>
  class Scan<linearFEM::Triangle> : public Scan<linearFEM::Element> {
  public:
    virtual linearFEM::Triangle& operator++(int) = 0;
  };

  template<>
  class Scan<linearFEM::Quad> : public Scan<linearFEM::Element> {
  public:
    virtual linearFEM::Quad& operator++(int) = 0;
  };

} // namespace concepts

namespace linearFEM {

  // ************************************************************** Linear2d **

  /** Space for linear triangular and bilinear quadrilateral FEM in 2D.
      @author Philipp Frauenfelder, 2002
  */
  class Linear2d : public concepts::Space<Real> {
  public:
    typedef concepts::Scan<Element> Scan;
    typedef void (*SMap)(const concepts::TColumn<Real>&,
			 concepts::TColumn<Real>&);
    /** Constructor.
	@param msh Mesh
	@param level Number levels for initial refinement
	@param bc Boundary conditions
    */
    Linear2d(concepts::Mesh2& msh, uint level,
	     concepts::BoundaryConditions* bc = 0);
    virtual ~Linear2d();
    virtual uint dim() const { return dim_; }
    virtual uint nelm() const { return nelm_; }
    virtual Scan* scan() const;
    /// Rebuilds the space
    void rebuild();
    /** Adjusts an element in the space.
	Only quads have S matrices, therefore, only quads can be refined.
	@param elm Element to refine
	@param level Number of refinements
    */
    void adjust(const Quad& elm, const short level);
    /** Returns the ith S matrix.
	The S matrices are used for the irregular refinement of a
	mesh. An S matrix is independent of the geometry but it depends on
	the topological relation of the children with the parent: the
	bottom left child has a different S matrix than the top right etc.
    */
    const SMap& S(uint i) const;
    static void S0(const concepts::TColumn<Real>& src,
		   concepts::TColumn<Real>& dst); // bottom left
    static void S1(const concepts::TColumn<Real>& src,
		   concepts::TColumn<Real>& dst); // bottom right
    static void S2(const concepts::TColumn<Real>& src,
		   concepts::TColumn<Real>& dst); // top right
    static void S3(const concepts::TColumn<Real>& src,
		   concepts::TColumn<Real>& dst); // top left
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Number of degrees of freedom
    uint dim_;
    /// Number of elements
    uint nelm_;
    /// Flag for rebuilding the space
    bool rebuild_;
    /// Mesh which this space is based on
    concepts::Mesh2& msh_;
    /// Boundary conditions
    concepts::BoundaryConditions* bc_;
    /// List of elements
    concepts::Joiner<Element*, 1>* elm_;
    /// Adjustements (ie. level changes) for the elements
    __gnu_cxx::hash_map<uint, short> adj_;
    /// Control information for vertices: active, member of space, valid index
    __gnu_cxx::hash_map<uint, short> ctrl0_;
    /// Level information for vertices
    __gnu_cxx::hash_map<uint, short> ctrl0L_;
    /// Control information for edges: active, member of space
    __gnu_cxx::hash_map<uint, short> ctrl1_;
    /// Level information for edges
    __gnu_cxx::hash_map<uint, short> ctrl1L_;
    /// Control information for elements
    __gnu_cxx::hash_map<uint, short> ctrl2_;
    /// Degree of freedom map
    __gnu_cxx::hash_map<uint, uint> dof_;

    static SMap S_[4];

    /** Determines maximal level and sets control information.
	\c rebuild0_ is called recursively.
	@param cntr Support of the element
	@param l Current level
	@param L Maximal level
    */
    void rebuild0_(concepts::Connector2& cntr, const int l, int& L);
    /** Builds the elements for quads. This routine assigns degress
	of freedom to vertices, creates T matrices, applies S
	matrices. \c rebuild1_ is called recursively.
	@param cell Cell of the element
	@param T0 T columns from element on a higher level
    */
    void rebuild1_(concepts::Quad2d& cell, concepts::TColumn<Real>* T0 = 0);
    /** Builds the elements for triangles. This routines assigns degress
	of freedom to vertices, creates T matrices.
	\c rebuild1_ is called recursively.
	@pre T0 is 0 as there are no S matrices for triangles up to now.
	@todo S matrices for triangles
	@param cell Cell of the element
	@param T0 T columns from element on a higher level
    */
    void rebuild1_(concepts::Triangle2d& cell,
		   concepts::TColumn<Real>* T0 = 0);
    /// Deactives children of this edge and the new vertex
    void deactivate_(concepts::Connector1& edg);
  };

} // namespace linearFEM

#endif // linSpace2D_hh
