// graphics for triangles

#include <sstream>
#include "triangleGraphics.hh"
#include "element2D.hh"
#include "space/postProcess.hh"
#include "graphics/meshEPS.hh"
#include "graphics/dataGnuplot.hh"
#include "graphics/dataDX.hh"

namespace linearFEM {

  // ****************************************************** TriangleGraphics **

  void TriangleGraphics::operator()(const concepts::Element<Real>& elm,
				enum graphicsType type,
				concepts::CellPostprocess<Real>& post) const {
    const Triangle* e = dynamic_cast<const Triangle*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    switch(type) {

    case TOPOLOGY_DOTTY:
    case VERTEX_LIST:
    case MESH_MATLAB:
    case MESH_DX:
    case MESH_EPS:
      post(e->cell()); break;

    case MESH_EPS_INIT: {
      graphics::MeshEPSCell<Real>* eps =
	dynamic_cast<graphics::MeshEPSCell<Real>*>(&post);
      conceptsAssert(eps, concepts::Assertion());
      concepts::Connector0* vtx = 0;
      for (uint i = 0; (vtx = e->support().vertex(i)) != 0; ++i) {
	concepts::Real2d x(e->cell().vertex(i));
	eps->offsetMin(x);
	eps->maxMax(x);
      }
      break;
    }

    case DATA_GNUPLOT: {
      uint vtx[4] = { 0, 1, 2, 2 };
      graphics::DataGnuplotCell<Real>* gnuplot =
	dynamic_cast<graphics::DataGnuplotCell<Real>*>(&post);
      conceptsAssert(gnuplot, concepts::Assertion());

      const concepts::Array<Real>& coeff = gnuplot->coeff(); // coefficients
      std::ofstream* ofs = gnuplot->write();
      for (uint i = 0; i < 2; ++i) {
	const concepts::Real2d v(e->cell().vertex(vtx[i]));
	*ofs << v[0] << " " << v[1] << " "
	     << coeff[vtx[i]] << std::endl;
      }
      *ofs << std::endl;
      for (uint i = 2; i < 4; ++i) {
	const concepts::Real2d v(e->cell().vertex(vtx[i]));
	*ofs << v[0] << " " << v[1] << " "
	     << coeff[vtx[i]] << std::endl;
      }
      *ofs << std::endl;
      *ofs << std::endl;
      break;
    }

    case DATA_DX_COUNTS: {
      graphics::DataDXCounts* counter =
	dynamic_cast<graphics::DataDXCounts*>(&post);
      conceptsAssert(counter, concepts::Assertion());
      counter->add(3);
      break;
    }

    case DATA_DX_COORDINATES: {
      graphics::DataDXCoordinates* coord = 
	dynamic_cast<graphics::DataDXCoordinates*>(&post);
      conceptsAssert(coord, concepts::Assertion());

      for (uint i = 0; i < 3; ++i) {
	concepts::Real2d x(e->cell().vertex(i));
	coord->write(x[0]);
	coord->write(x[1]);
      }
      coord->add(1);

      break;
    }

    case DATA_DX_CONNECTIONS: {
      graphics::DataDXConnections* conn = 
	dynamic_cast<graphics::DataDXConnections*>(&post);
      conceptsAssert(conn, concepts::Assertion());
      uint l = conn->counts();
      uint I[4] = { 0, 1, 2, 2 };
      for (uint i = 0; i < 4; ++i)
	conn->write(l+I[i]);
      conn->add(3);
      break;
    }

    case DATA_DX_DATA: {
      graphics::DataDXData* data = 
	dynamic_cast<graphics::DataDXData*>(&post);
      conceptsAssert(data, concepts::Assertion());

      const concepts::Formula* formula = data->formula();
      Real res = 0;
      const concepts::Array<Real>& coeff = data->coeff();
      for (uint i = 0; i < 3; ++i) {
	if (! formula) {
	  res = coeff[i];
	} else {
	  concepts::Real2d x(e->cell().vertex(i));
	  res = (*formula)(x);
	}
	data->write(res);
      }

      break;
    }

    default: 
      std::stringstream error;
      error << "graphics type " << type << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str())); break;
    }
  }

  std::ostream& TriangleGraphics::info(std::ostream& os) const {
    return os << "TriangleGraphics()";
  }

} // namespace linearFEM
