// include files for the operator subdirectory

#ifndef operator_hh
#define operator_hh

#include "operator/compositions.hh"
#include "operator/permutation.hh"
#include "operator/matrix.hh"
#include "operator/denseMatrix.hh"
#include "operator/hashedSMatrix.hh"
#include "operator/sparseMatrix.hh"
#include "operator/bilinearForm.hh"
#include "operator/solverFabric.hh"
#include "operator/gmres.hh"
#include "operator/gmresFabric.hh"

#include "operator/cg.hh"
#include "operator/cgFabric.hh"

#ifdef HAS_PETSC
#include "operator/PETSc.hh"
#include "operator/PETScFabric.hh"
#endif

#ifdef HAS_SuperLU
#include "operator/superLU.hh"
#include "operator/superLUfabric.hh"
#endif

#include "operator/diagonal.hh"
#include "operator/diagonalFabric.hh"

#include "operator/umfpack.hh"
#include "operator/umfpackFabric.hh"

#include "operator/pardiso.hh"
#include "operator/pardisoFabric.hh"

#include "operator/transpose.hh"

#include "operator/constrained.hh"

#endif // operator_hh

/** @defgroup bilinear Bilinear and Linear Forms -- Matrices and Vectors
    The stiffness matrix and the load vector are assembled from
    elementwise computations of the bilinear and linear forms.
    <h2>Assembling</h2>

    The element matrices (or vectors, concepts::ElementMatrix) need to
    be assembled into the stiffness matrix in order to create the
    global linear system. This assembling needs a mapping from the
    local to the global degrees of freedom. This mapping is stored in
    a so-called T matrix (concepts::TMatrixBase) in each element.

    The element matrices are computed by the application operator of
    the bilinear and linear forms (concepts::BilinearForm and
    concepts::LinearForm). The assembling is then handled by the
    constructor of the global linear system represented by a matrix
    (e.g. concepts::SparseMatrix derived from concepts::Matrix). It is
    also possible to have a matrix free representation of a stiffness
    matrix (derived from concepts::Operator).

    Almost all assembling algorithms for matrices are implemented as
    static methods in concepts::Matrix.
    <h2>Bilinear Forms and Matrices</h2>
    <h3>Bilinear Forms</h3>

    A bilinear form <span class="math">a(u,v)</span> is evaluated for two
    functions of the basis of the space. This is done by evaluating <span
    class="math">a(u,v)</span> for all shape functions on all elements to
    form the element matrices.

    A bilinear form is computed in the application operator of a class
    derived from concepts::BilinearForm.

    The application operator of such an bilinear form takes two elements
    and an element matrix (which is then filled) as arguments. The
    application operator computes the interactions of the shape functions
    of the two elements. In standard FEM, there are only interactions, if
    the two elements are identical.
    <h3>Matrices and Operators</h3>
    
    To form the global matrix, the element matrices have to be
    assembled. This is done in the constructor of a matrix. The base
    class for all operators (and matrices are operators) is
    concepts::Operator.

    It is possible to combine different operators using
    concepts::Compose or concepts::LiCo.
    <h2>Linear Forms and Vectors</h2>
    <h3>Linear Forms</h3>

    A linear form <span class="math">l(v)</span> is evaluated for each
    basis function of the space. This is done similarly to the bilinear
    form: the linear form is evaluated for all shape functions of each
    element. Then, the local contributions to the vector are assembled.

    Linear forms are derived from concepts::LinearForm.
    <h3>Vectors</h3>

    The local contributions to the global vector are assembled in the
    constructor of concepts::Vector, derived from concepts::Function.
    <h2>Solver</h2>

    The application operator of a solver solves the linear
    system. concepts::GMRes and concepts::CG are our own
    implementations, concepts::PETSc includes a lot of (iterative)
    solvers. There are interfaces to the direct solvers
    concepts::SuperLU, concepts::Pardiso, concepts::Umfpack, see
    \ref solver "description of linear solvers".

    @author Philipp Frauenfelder, 2004
*/

/** @defgroup solver Linear Solvers
    Solvers for linear systems.

    All solvers which are just an interface to an external solver
    require that the respective software package is installed and that
    the library is correctly configured to use it.

    <h2>Iterative Solvers</h2>

    concepts::CG is an implementation of the conjugate gradient
    method.

    concepts::GMRes is an implementation of the general minimal
    residual method.

    concepts::PETSc is an interface to PETSc (web links in the class
    description) featuring many different iterative solvers and
    preconditioners.

    <h2>Direct Solvers</h2>

    concepts::SuperLU is an interface to SuperLU (web links in the
    class description), a sparse LU decomposition.

    concepts::Pardiso is an interface to Pardiso (web links in the
    class description), a sparse decomposition (multithreaded if
    requested).

    concepts::Umfpack is an interface to Umfpack (web links in the
    class description) using the unsymmetric multi-frontal method.

    concepts::DiagonalSolver solves a diagonal linear system.
    <h2>Solver Fabrics</h2>

    Most of the solvers have a respective fabric class derived from
    concepts::SolverFabric. The idea is that all parameters (like
    convergence criteria etc.) are given in the construction of the
    fabric and the solver itself is constructed by the fabric given
    the matrix. This construction call is the the same for all solvers
    as all parameters are known before.
    <h2>Constrained Problems</h2>

    concepts::Constrained is a solver for a system with
    (non-homogenous) linear constraints. It modifies the stiffness
    matrix and the right hand side such that the constraints are
    fulfilled, then the modified system is solved with a given solver
    (via a concepts::SolverFabric).

    @author Philipp Frauenfelder, 2004
 */
