/* Basic operator declarations such as compositions of operators
 */

#ifndef compositions_hh
#define compositions_hh

#include "basics/outputOperator.hh"
#include "basics/exceptions.hh"
#include "space/space.hh"
#include "function/basis.hh"
#include "function/vector.hh"
#include "basics/debug.hh"

// debugging
#define ComposeConstr_D 0
#define ComposeAppl_D 0

namespace concepts {

  // forward declaration
  template<class F>
  class SparseMatrix;

  template<class F>
  class DenseMatrix;

  template<class F>
  class Matrix;

  template<class F>
  class HashedSparseMatrix;

  // ************************************************************** Operator **

  /** Abstract class for an operator.
      @ingroup bilinear
  */
  template<class F>
  class Operator : public OutputOperator {
  public:
    virtual ~Operator() {}

    /** Application operator.
	Computes \c fncX = A(\c fncY) where A is this operator.
    */
    virtual void operator()(const Function<F>& fncY, Function<F>& fncX) = 0;

    /// Returns the image space of the operator
    virtual const Space<F>& spaceX() const = 0;

    /// Returns the source space of the operator
    virtual const Space<F>& spaceY() const = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  // *************************************************************** Compose **

  /** Computes the product of two operators.
      In reality, no product is computed but A is applied after B.
      If you really want to compute the matrix product, construct a new
      matrix operator and use the compose object as constructor argument.
      @test test::CompositionsTest
      @test test::DriverTest
      @test test::MoreCompositionsTest
      @test test::DeepCompositionsTest
      @ingroup bilinear
  */
  template<class F>
  class Compose : public Operator<F> {
  public:
    /// Constructor
    inline Compose(Operator<F>& A, Operator<F>& B);

    /** Application operator. Computes $x = A B y$.
     */
    inline void operator()(const Function<F>& fncY, Function<F>& fncX);

    const Space<F>& spaceX() const { return A_.spaceX(); }
    const Space<F>& spaceY() const { return B_.spaceY(); }

    /** Collapses the composition of matrices into one matrix. This
 	only works in some rare combinations, test it out.

 	@return True, if the operation was succesful.
 	@param dest Matrix into which the operator should be collapsed.
 	            Should be zero on startup since the components are added
 		    into the matrix.
    */
    bool collapse(Matrix<F>& dest) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// First operator
    Operator<F>& A_;
    /// Second operator
    Operator<F>& B_;
    /// Intermediate vector to store result after application of B
    Vector<F> f_;

    void collapse_(DenseMatrix<F>& A, DenseMatrix<F>& B,
		   Matrix<F>& dest) const;
  };

  template<class F>
  Compose<F>::Compose(Operator<F>& A, Operator<F>& B)
    : A_(A), B_(B), f_(B.spaceX()) {
    DEBUGL(ComposeConstr_D, "A is (" << A.spaceX().dim() << "x" 
	   << A.spaceY().dim() << ") * B is (" << B.spaceX().dim() << "x" 
	   << B.spaceY().dim() << ")");
    conceptsAssert(A.spaceY().dim() == B.spaceX().dim(), Assertion());
  }

  template<class F>
  void Compose<F>::operator()(const Function<F>& fncY, Function<F>& fncX) {
    DEBUGL(ComposeAppl_D, *this);
    conceptsAssert3(fncY.space().dim() == spaceY().dim(), Assertion(),
		    "fncY.dim = " << fncY.space().dim() << ", spaceY.dim = "
		    << spaceY().dim());
    conceptsAssert3(fncX.space().dim() == spaceX().dim(), Assertion(),
		    "fncX.dim = " << fncX.space().dim() << ", spaceX.dim = "
		    << spaceX().dim());
    DEBUGL(ComposeAppl_D, "y = " << fncY);
    B_(fncY, f_);
    DEBUGL(ComposeAppl_D, "f = " << f_);
    A_(f_, fncX);
    DEBUGL(ComposeAppl_D, "x = " << fncX);
  }

  // ***************************************************************** LiCoI **

  /** Linear combination of an operator with the identity. The application
      operator evaluates a*A + b*I.
      @ingroup bilinear
  */
  template<class F>
  class LiCoI : public Operator<F> {
  public:
    LiCoI(Operator<F>& A, F a = 1.0, F b = 1.0) : A_(A), a_(a), b_(b) {}

    /** Application operator. Calculates 
	\f[ x=a \cdot Ay + b \cdot y = (aA + bI)y \f]
	where \f$A\f$ is the matrix and \f$a\f$ its respective coefficient.
	@param fncY \f$y\f$
	@param fncX \f$x\f$
    */
    inline void operator()(const Function<F>& fncY, Function<F>& fncX);

    const Space<F>& spaceX() const { return A_.spaceX(); }
    const Space<F>& spaceY() const { return A_.spaceY(); }

    /** Collapses the linear combination of matrices into one
 	matrix. This only works, if the linear combination is made out
 	of linear combinations and matrices. This method calls addInto
 	on each matrix which makes part of this linear combination or
 	itself on a linear combination.

 	The method fails if it encounters something else than a LiCo, LiCoI or
	a matrix.

 	@return True, if the operation was succesful
 	@param dest Matrix into which the operator should be collapsed.
 	       Should be zero on startup since the components are added
	       into the matrix.
 	@param fact Factor by which both components should be multiplied.
 	       Defaults to 1, ie. can be omitted.
    */
    bool collapse(Matrix<F>& dest, const F fact = 1.0) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Operator A
    Operator<F>& A_;
    /// Scalar a
    F a_;
    /// Scalar b
    F b_;
  };

  template<class F>
  void LiCoI<F>::operator()(const Function<F>& fncY, Function<F>& fncX) {
    conceptsAssert(fncY.space().dim() == spaceY().dim(), Assertion());
    conceptsAssert(fncX.space().dim() == spaceX().dim(), Assertion());
    A_(fncY, fncX);
    if (a_ != 1.0)
      fncX *= a_;
    if (b_ != 0.0)
      fncX.add(fncY, b_);
  }

  // ****************************************************************** LiCo **

  /** Linear combination of two operators. The application operator evaluates
      a*A + b*B.
      @test test::CompositionsTest
      @test test::DeepCompositionsTest
      @ingroup bilinear
  */
  template<class F>
  class LiCo : public Operator<F> {
  public:
    /** Constructor. The operator consists of $a A + b B$ where $A$ and $B$ are
	two operators and $a$ and $b$ are scalars.

	@param A First operator
	@param B Second operator
	@param a Coefficient of $A$, defaults to 1
	@param b Coefficient of $B$, defaults to 1
    */
    LiCo(Operator<F>& A, Operator<F>& B, F a = 1.0, F b = 1.0) :
      A_(A), B_(B), a_(a), b_(b), f_(A.spaceX()) {
      conceptsAssert(A.spaceX().dim() == B.spaceX().dim(), Assertion());
      conceptsAssert(A.spaceY().dim() == B.spaceY().dim(), Assertion());
    }

    /** Application operator. Calculates 
	\f[x=a \cdot Ay + b \cdot By = (aA + bB)y\f]
	where $A$ and $B$ are the matrices and $a$ and $b$ their 
	respective coefficients.
	@param fncY $y$
	@param fncX $x$
    */
    inline void operator()(const Function<F>& fncY, Function<F>& fncX);

    const Space<F>& spaceX() const { return A_.spaceX(); }
    const Space<F>& spaceY() const { return A_.spaceY(); }

    /** Collapses the linear combination of matrices into one
 	matrix. This only works in some rare combinations, test it
 	out. This method calls addInto on each matrix which makes part
 	of this linear combination or itself on a linear combination.

 	The method fails if it encounters something else than a LiCo, LiCoI or
	a matrix.

 	@return True, if the operation was succesful
 	@param dest Matrix into which the operator should be collapsed.
 	       Should be zero on startup since the components are added
	       into the matrix.
 	@param fact Factor by which both components should be multiplied.
 	       Defaults to 1, ie. can be omitted.
    */
    bool collapse(Matrix<F>& dest, const F fact = 1.0) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Operator A
    Operator<F>& A_;
    /// Operator B
    Operator<F>& B_;
    /// Scalar a
    F a_;
    /// Scalar b
    F b_;
    /// Intermediate vector
    Vector<F> f_;
  };

  template<class F>
  void LiCo<F>::operator()(const Function<F>& fncY, Function<F>& fncX) {
    conceptsAssert(spaceX().dim() == fncX.space().dim(), Assertion());
    conceptsAssert(spaceY().dim() == fncY.space().dim(), Assertion());
    A_(fncY, fncX); 
    fncX *= a_; 
    B_(fncY, f_); 
    fncX.add(f_, b_); 
  }

} // namespace concepts

#endif // compositions_hh
