/* SMatrix
 * sparse matrices
 */

#ifndef hSparseMatrix_hh
#define hSparseMatrix_hh

#include <iostream>
#include <fstream>
#include <limits>
#include "basics/typedefs.hh"
#include "toolbox/pool.hh"

namespace concepts {

  // forward declaration
  template<typename T>
  class Matrix;

  template<typename T>
  class DenseMatrix;

  // **************************************************** HashedSparseMatrix **

  /** A matrix in sparse storage using hashes.
   */
  template <class T>
  class HashedSparseMatrix {
  public:
    /** Values of a matrix in sparse notation.
	@see HashedSparseMatrix
    */
    struct Value {
      Value* lnk;
      uint   idx;
      T      val;

      /** Constructor. Creates a linked list of index-values pairs from
	  tail to head. The already existing list is taken and the link
	  of the item points to the head of this list. The given index and
	  value are stored in to head (the new entry).

	  @param l Linked list of index-value pairs
	  @param i Index of the new entry
	  @param v Value of the new entry
      */
      Value(Value* l, int i, T v = 0.0) { lnk = l; idx = i; val = v; }
      void* operator new(size_t sz) { return new char[sz]; }
      void* operator new
      (size_t, Pool<typename HashedSparseMatrix<T>::Value>& pool) { 
	return pool.alloc(); }
    };

    /** Constructor.
	@param r Number of rows
	@param c Number of columns
	@param h Number of hash bits
    */
    HashedSparseMatrix(uint r, uint c, uint h);
    ~HashedSparseMatrix();

    inline Value** Matrix() const { return matrix; } // new
    inline uint HashBits() const { return hashBits; } // new

    /** Returns the entry with index (r,c). If this entry does not yet exist,
	it is created.

	@param r Row index
	@param c Column index
    */
    T& operator()(const uint r, const uint c);
    T operator()(const uint r, const uint c) const;
    void operator()(const T f[], T g[]) const;
    /** Multiplies the transpose of the matrix with \c f and adds the
	results to \c g.
	@pre \c g is set to 0
    */
    void transpMult(const T f[], T g[]) const;

    /// Returns the number of used entries in the matrix
    uint used() const { return _used; }
    float memory() const { 
      return sizeof(HashedSparseMatrix<T>) + pool.memory() 
	- sizeof(pool); 
    }

    void write(std::ofstream& ofs) const;
    void outputMatlab(std::ofstream& ofs) const;
    void outputSparseQR(std::ofstream& ofs) const;

    /// Multiplies this matrix and \c fact and adds the result to \c dest
    void multiply(const HashedSparseMatrix<T>* const fact,
		  concepts::Matrix<T>& dest) const;
    /// Multiplies this matrix and \c fact and adds the result to \c dest
    void multiply(const concepts::Matrix<T>* const fact,
		  concepts::Matrix<T>& dest) const;

    std::ostream& info(std::ostream& os) const;

    /** Compresses the matrix by dropping small entries. All matrix
	entries which are smaller than a certain \c threshold times
	the largest entry of the matrix are deleted from the
	matrix. To delete a small entry, \c dropEntry_ is used.
     */
    void compress(Real threshold = std::numeric_limits<Real>::epsilon());
  private:
    Pool<Value> pool;
    Value**     matrix;
    uint        _used;

    /// Number of rows
    uint        nofRows;

    /// Number of columns
    uint        nofCols;

    uint        hashBits;
    uint        hashMsk;

    HashedSparseMatrix(const HashedSparseMatrix<T>&);
    void operator =(const HashedSparseMatrix<T>&);

    /** Removes entry (r,c) from the matrix. The storage of the matrix
	entry is not released and will only be released on destruction
	of the whole matrix. Trying to remove a non-existing entry
	simply does nothing.

	@param v Pointer to the entry which should be removed
	@param vPrev Pointer to the entry before the one to be removed
	@param r Row index
	@param c Column index
    */
    void dropEntry_(Value*& v, Value*& vPrev, const uint r, const uint c);
  };

  template<typename T>
  std::ostream& operator<<(std::ostream& os, const HashedSparseMatrix<T>& o) {
#ifdef DEBUG
    os << std::flush;
#endif
    return o.info(os);
  }

} // namespace concepts

#endif // hSparseMatrix_hh
