/* interface for a matrix
 */

#include "matrix.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "toolbox/scannerConnectors.hh"
#include "bilinearForm.hh"

// timings
#define matrixTiming

#ifdef matrixTiming
#include "toolbox/inputOutput.hh"
#include "toolbox/resourceMonitor.hh"
#endif // matrixTiming

// debugging
#include "basics/debug.hh"

#define asseblyop_D 0

namespace concepts {

  // **************************************************************** Matrix **

  template<class F>
  void Matrix<F>::assembly(Matrix<F>& dest,
			   const Space<F>& spc, BilinearForm<F>& bf) {
    conceptsAssert(dest.spaceX().dim() == spc.dim(), Assertion());
    conceptsAssert(dest.spaceY().dim() == spc.dim(), Assertion());

#ifdef matrixTiming
    ResourceMonitor timer;
#endif // matrixTiming

    ElementMatrix<F> A(3, 3);
    ElementMatrix<F> B(3, 3);

    std::auto_ptr<typename Space<F>::Scanner> sc(spc.scan());
    while (*sc) {
      const Element<F>& elm = (*sc)++;
      const TMatrixBase<F>& T = elm.T();

#ifdef matrixTiming
      if (timings_)
	timer.utime();
#endif // matrixTiming
      bf(elm, elm, A);
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("bilinear_form", timeCntr_, timer.utime());
#endif // matrixTiming
      T(A, B); B.transpose(); T(B, A); A.transpose();
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("tmatrix_apply", timeCntr_, timer.utime());
#endif // matrixTiming
      DEBUGL(asseblyop_D, "assembling " << A << " into " << dest);

      for(uint i = T.n(); i--;)
	for(uint j = T.n(); j--;)
	  dest(T.index(i), T.index(j)) += A(i, j);
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("global_assembly", timeCntr_++,
				 timer.utime());
#endif // matrixTiming
    }
    DEBUGL(asseblyop_D, "done.");
  }

  template<class F>
  void Matrix<F>::assembly(Matrix<F>& dest,
			   const Space<F>& spcX, const Space<F>& spcY,
			   BilinearForm<F>& bf) {
    conceptsAssert(dest.spaceX().dim() == spcX.dim(), Assertion());
    conceptsAssert(dest.spaceY().dim() == spcY.dim(), Assertion());

#ifdef matrixTiming
    ResourceMonitor timer;
#endif // matrixTiming

    ElementMatrix<F> A(3, 3);
    ElementMatrix<F> B(3, 3);

    std::auto_ptr<typename Space<F>::Scanner> scX(spcX.scan());
    while ( *(scX.get()) ) {
      const Element<F>& elmX = (*scX)++;
      const TMatrixBase<F>& TX = elmX.T();

      std::auto_ptr<typename Space<F>::Scanner> scY(spcY.scan());
      while ( !scY->eos() ) {
	const Element<F>& elmY = (*scY)++;
	const TMatrixBase<F>& TY = elmY.T();

#ifdef matrixTiming
	if (timings_)
	  timer.utime();
#endif // matrixTiming
	bf(elmX, elmY, A);
	conceptsAssert(A.m() == TX.m(), Assertion());
	conceptsAssert(A.n() == TY.m(), Assertion());
#ifdef matrixTiming
	if (timings_)
	  timings_->addArrayDouble("bilinear_form", timeCntr_, timer.utime());
#endif // matrixTiming
	TY(A, B); B.transpose(); TX(B, A); A.transpose();
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("tmatrix_apply", timeCntr_, timer.utime());
#endif // matrixTiming

	DEBUGL(asseblyop_D, "assembling " << A << " into " << dest);
	for(uint i = TX.n(); i--;) {
	  for(uint j = TY.n(); j--;) {
	    dest(TX.index(i), TY.index(j)) += A(i, j);
	  }
	}
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("global_assembly", timeCntr_++,
				 timer.utime());
#endif // matrixTiming
      } // while scY
    } // while scX
    DEBUGL(asseblyop_D, "done.");
  }

  template<class F>
  void Matrix<F>::assembly(Matrix<F>& dest,
			   BilinearForm<F>& bf,
			   const ElementPairList<F>& pairs) {
    ElementMatrix<F> A(3, 3);
    ElementMatrix<F> B(3, 3);

#ifdef matrixTiming
    ResourceMonitor timer;
#endif // matrixTiming

    std::auto_ptr<typename ElementPairList<F>::Scanner> sc(pairs.scan());
    while (*sc) {
      const ElementPair<F>& elmPair = (*sc)++;
      const Element<F>& elmX = elmPair.elm1();
      const TMatrixBase<F>& TX = elmX.T();
  
      const Element<F>& elmY = elmPair.elm2();
      const TMatrixBase<F>& TY = elmY.T();

#ifdef matrixTiming
      if (timings_)
	timer.utime();
#endif // matrixTiming
      bf(elmX, elmY, A, elmPair);
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("bilinear_form", timeCntr_, timer.utime());
#endif // matrixTiming
      if (A.size() > 0) {
	TY(A, B); B.transpose(); TX(B, A); A.transpose();
#ifdef matrixTiming
	if (timings_)
	  timings_->addArrayDouble("tmatrix_apply", timeCntr_, timer.utime());
#endif // matrixTiming

	for(uint i = TX.n(); i--;) {
	  for(uint j = TY.n(); j--;) {
	    dest(TX.index(i), TY.index(j)) += A(i, j);
	  } // for x
	} // for y
#ifdef matrixTiming
	if (timings_)
	  timings_->addArrayDouble("global_assembly", timeCntr_++,
				   timer.utime());
#endif // matrixTiming
      } // if size
    } // while scX
    DEBUGL(asseblyop_D, "done.");
  }

  template<class F>
  InOutParameters* Matrix<F>::timings_ = 0;

  template<class F>
  uint Matrix<F>::timeCntr_ = 0;

  template<class F>
  void Matrix<F>::setTimings(InOutParameters* timings) {
#ifdef matrixTiming
    timings_ = timings;
    if (timings_) {
      timeCntr_ = 0;
      timings_->addArrayDouble("bilinear_form");
      timings_->addArrayDouble("tmatrix_apply");
      timings_->addArrayDouble("global_assembly");
    }
#endif // matrixTiming
  }

  template<class F>
  bool Matrix<F>::timings() {
#ifdef matrixTiming
    return true;
#else
    return false;
#endif // matrixTiming
  }

} // namespace concepts
