// interface to PARDISO

#include <map>
#include <memory>
#include <cstring>
#include "pardiso.hh"
extern "C" {
#  include "types.h"
#  include "utils_debug.h"
#  include "utils_timing.h"
  int pardiso_(size_t*, int*, int*, int*, int*, uint*, double*, uint*, uint*,
	       int*, int*, int*, int*, double*, double*, int*);

  void  smat_to_fortran_indexing(smat_t * mat) {
    unsigned int  i;
 
    for (i = 0; i < mat->ia[mat->m]; i++)
      mat->ja[i]++;
    for (i = 0; i < (mat->m + 1); i++)
      mat->ia[i]++;
 
    /* sanity check */
    assert(mat->nnz == mat->ia[mat->m]-1);
  }

  void start_timing(timing_t  *time_var) {
    times(time_var);
  }

}
#include "operator/sparseMatrix.hh"
#include "toolbox/array.hh"

// debugging
#include "basics/debug.hh"

#define PardisoConstr_D 0
#define PardisoDestr_D 0
#define PardisoAppl_D 0
#define PardisoPARDISO_D 0

#define fortranNotation 0
#define storeMatrixInFile 0

#if storeMatrixInFile
#  include <fstream>
#endif

namespace concepts {

  smat_t* smatNew(unsigned int m, unsigned int n, int is_symmetric,
		  uint* ia, uint* ja, double* a);
  void smatFree(smat_t*);

  // *************************************************************** Pardiso **

  Pardiso::Pardiso(const SparseMatrix<Real>& matrix, bool symmetric) :
    n_(matrix.spaceX().dim()), spcX_(matrix.spaceY()), spcY_(matrix.spaceX()),
    symmetric_(symmetric), factored_(false),
    nnz_(symmetric ? (matrix.m()->used()-n_)/2 + n_ : matrix.m()->used()),
    A_(0), a_(nnz_), ia_(n_ + 1), ja_(nnz_), handle(64, 0), iparam(64, 0),
    error(0) {
    DEBUGL(PardisoConstr_D, "n = " << n_);
    DEBUGL(PardisoConstr_D, "symmetric = " << symmetric_);
    DEBUGL(PardisoConstr_D, "nnz = " << nnz_ << ", from " << matrix);
    if (symmetric_) {
      if ((matrix.m()->used()-n_)%2 != 0) {
	throw conceptsException(concepts::ExceptionBase());
      }
    }
    conceptsAssert(matrix.spaceY().dim() == matrix.spaceX().dim(),
		   Assertion());

    smat_t* A = smatNew(n_, n_, symmetric_, ia_, ja_, a_);
    A_ = (void*)A;
    A->nnz = nnz_;

    // temporary data structure since the data in a and ja have to ordered
    std::map<uint, Real> atmp;

    // store entries in matrix
    uint nnz = 0;
    const HashedSparseMatrix<Real> *m = matrix.m();
    HashedSparseMatrix<Real>::Value** matrixS = m->Matrix();  
    const int hashBits = m->HashBits();
    const int p = 1<<hashBits;
    for(uint r = 0; r < spcX_.dim(); ++r) {
      ia_[r] = nnz
#if fortranNotation
	+1
#endif
	;
      for (int c_mod = 0; c_mod < p; ++c_mod){
	for (HashedSparseMatrix<Real>::Value* v =
	       matrixS[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk) {
	  if ((! symmetric_) || (v->idx >= r)) {
	    atmp[v->idx] = v->val;
	    ++nnz;
	  } // symmetric
	} // for v
      } // for c_mod
      uint iaTmp = ia_[r]
#if fortranNotation
	-1
#endif
	;
      DEBUGL(PardisoConstr_D, "row " << r << ", " << atmp.size()
	     << " entries, iaTmp = " << iaTmp << ", to store in "
	     << ja_.size());
      for (std::map<uint, Real>::const_iterator iA = atmp.begin();
	   iA != atmp.end(); ++iA, ++iaTmp) {
	ja_[iaTmp] = iA->first
#if fortranNotation
	  +1
#endif
	  ;
	a_[iaTmp] = iA->second;
      } // for iA
      atmp.clear();
    } // for r
    ia_[n_] = nnz
#if fortranNotation
      +1
#endif
      ;
    DEBUGL(PardisoConstr_D, "ia = " << ia_ << ", ja = " << ja_
	   << ", a = " << a_);

#if fortranNotation
    DEBUGL(PardisoConstr_D, A->nnz+1 << " == " << A->ia[A->m]);
    conceptsAssert(A->nnz+1 == A->ia[A->m], Assertion());
    conceptsAssert(A->ia[0] == 1, Assertion());
#else
    DEBUGL(PardisoConstr_D, A->nnz << " == " << A->ia[A->m]);
    conceptsAssert(A->nnz == A->ia[A->m], Assertion());
    conceptsAssert(A->ia[0] == 0, Assertion());
    smat_to_fortran_indexing(A);
#endif

    char* var = getenv("OMP_NUM_THREADS");
    if (var != NULL) sscanf(var, "%d", &max_threads);
    else max_threads = 1;
    if (max_threads < 1) max_threads = 1;

    DEBUGL(PardisoPARDISO_D, "Max_threads for the parallel factorization = "
	   << max_threads);

    // Setup parameters
    iparam[0]  =  1;		// No solver default
    iparam[1]  =  2;		// 2: METIS reordering, 0: minimum degree
    iparam[2]  =  max_threads;	// Value of OMP_NUM_THREADS
    iparam[3]  =  0;		// no CGS iteration
    iparam[4]  =  0;		// no user permutation
    iparam[5]  =  0;		// solution on b
    iparam[6]  =  16;		// Default logical fortran unit number for output
    // max. numbers of iterative refinement steps
    if (symmetric) iparam[7] = 0;
    else iparam[7] = 2;
    iparam[9]  =  13;		// epspivot
    iparam[10] =  1;		// scaling of MC64
    iparam[13] =  0;		// Output: Number of perturbed pivots
    iparam[14] =  0;		// Output: peak memory in KB in phase 1
    iparam[15] =  0;            // Output: perm. memory in KB in phase 1
    iparam[16] =  0;            // Output: total double prec. memory in phase 2
    iparam[17] = -1;		// Output: Number of nonzeros in LU
    iparam[18] = -1;		// Output: Mflops for LU factorization
    iparam[19] =  0;		// Output: Numbers of CG Iterations

    max_factorizations = 1;
    matrix_number = 1;

    // matrix_type
    // 11: real unsymmetric matrix
    // 2: real SPD, -2: real indefinite
    if (symmetric) {
      matrix_type = 2;
    } else {
      matrix_type = 11;
    }

#if PardisoAppl_D
    msglvl = 1;			// statistics in pardiso.stat.$nproc
#else
    msglvl = 0;
#endif
    nb = 1;			// one right hand side
  }

  Pardiso::~Pardiso() {
    smat_t *A = (smat_t*)A_;
    int ido;
    // Clean up
    ido = -1;			/* Free internal memory */
    DEBUGL(PardisoDestr_D, "before clean up call: " << max_factorizations
	   << ", " << matrix_number << ", " << matrix_type << ", "
	   << ido << ", " << nb << ", " << msglvl << ", "
	   << error << ", " << iparam << ", " << handle);
    pardiso_(handle, &max_factorizations, &matrix_number,
	     &matrix_type, &ido, &A->m, A->a, A->ia, A->ja,
	     0, &nb, iparam,
	     &msglvl, 0, 0, &error);
    free((smat_t*)A_);
  }

  void Pardiso::operator()(const Function<Real>& fncY,
			   Function<Real>& fncX) {
    const Vector<Real>* fncYA;
    if ((fncYA = dynamic_cast<const Vector<Real>*>(&fncY))) {
      Vector<Real>* fncXA;
      if ((fncXA = dynamic_cast<Vector<Real>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException(MissingFeature("unknown function"));
  }

  void Pardiso::operator()(const Vector<Real>& fncY, Vector<Real>& fncX) {
#if storeMatrixInFile
    if (! factored_) {
      std::ofstream matrix("pardiso.matrix");

      std::cout << "storing matrix in file 'pardiso.matrix'" << std::endl;
      std::cout << "iparam = " << iparam << std::endl;

      matrix << fncX.space().dim() << std::endl;
      matrix << nnz_ << std::endl;
      for (uint j = 0; j < ia_.size(); ++j)
	matrix << ia_[j] << std::endl;
      conceptsAssert3(ia_.size() == fncX.space().dim()+1,
		      concepts::Assertion(),
		      "ia_.size() = " << ia_.size()
		      << ", fncX.space().dim() = " << fncX.space().dim());
      for (uint j = 0; j < ja_.size(); ++j)
	matrix << ja_[j] << std::endl;
      conceptsAssert3(ja_.size() == nnz_, concepts::Assertion(),
		      "ja_.size() = " << ja_.size() << ", nnz_ = " << nnz_);
      matrix.precision(15);
      for (uint j = 0; j < a_.size(); ++j)
	matrix << a_[j] << std::endl;
      conceptsAssert3(a_.size() == nnz_, concepts::Assertion(),
		      "a_.size() = " << a_.size() << ", nnz_ = " << nnz_);
      for (uint j = 0; j < fncY.space().dim(); ++j)
	matrix << fncY(j) << std::endl;
    }
#endif
    vec_t b, x;
    b.n = fncY.space().dim();
    b.v = fncY;

    Array<Real> xTmp(2*fncX.space().dim());
    x.n = fncX.space().dim();
    x.v = xTmp;

    if (! factored_) {
      int rc = solver_pardiso_factor_();
      conceptsAssert(rc == 1, Assertion());
      factored_ = true;
    }

    DEBUGL(PardisoAppl_D, "b = " << fncY);
    int rc = solver_pardiso_solve_(b.v, x.v);
    conceptsAssert(rc == 1, Assertion());
    DEBUGL(PardisoAppl_D, "x = " << xTmp);

    std::memcpy(fncX, x.v, fncX.space().dim()*sizeof(Real));
  }

  std::ostream& Pardiso::info(std::ostream& os) const {
    return os << "Pardiso(n = " << n_ << ", "
	      << iparam[2] << " thread(s), "
	      << (symmetric_ ? "symmetric" : "non-symmetric") << ", "
	      << (factored_ ? "factored" : "not factored") << ")";
  }

  int Pardiso::solver_pardiso_factor_() {
    DEBUGL(PardisoPARDISO_D, "Solving with PARDISO ...");

    // Symbolic Factorization
    DEBUGL(PardisoPARDISO_D, "PARDISO Analysis");
    smat_t *A = (smat_t*)A_;
    int ido = 11;			/* PARDISO Analysis */
    DEBUGL(PardisoDestr_D, "before analysis call: " << max_factorizations
	   << ", " << matrix_number << ", " << matrix_type << ", "
	   << ido << ", " << nb << ", " << msglvl << ", "
	   << error << ", " << iparam << ", " << handle);
    pardiso_(handle, &max_factorizations, &matrix_number,
	     &matrix_type, &ido, &A->m, A->a, A->ia, A->ja,
	     0, &nb, iparam,
	     &msglvl, NULL, NULL, &error);
    DEBUGL(PardisoPARDISO_D, "pardiso_ error = " << error);
    DEBUGL(PardisoPARDISO_D, "PARDISO Nonzeros in LU " << iparam[17]);
    DEBUGL(PardisoPARDISO_D, "PARDISO Mflops for LU " << iparam[18]);
    conceptsAssert3(error == 0, concepts::Assertion(), "error = " << error);

    // Numerical Factorization
    DEBUGL(PardisoPARDISO_D, "PARDISO Factorization");
    ido = 22;			/* Numerical factorization */
    pardiso_(handle, &max_factorizations, &matrix_number,
	     &matrix_type, &ido, &A->m, A->a, A->ia, A->ja,
	     0, &nb, iparam,
	     &msglvl, NULL, NULL, &error);
    conceptsAssert3(error == 0, concepts::Assertion(), "error = " << error);

    return 1;
  }

  int Pardiso::solver_pardiso_solve_(double *b, double *x) {
    // Forward / backward substitution
    DEBUGL(PardisoPARDISO_D, "PARDISO Backward Substitution");
    /* Allocate memory for solution important 2*neqns */
    smat_t *A = (smat_t*)A_;
    int ido = 33;			/* Substitution */
    pardiso_(handle, &max_factorizations, &matrix_number,
	     &matrix_type, &ido, &A->m, A->a, A->ia, A->ja,
	     0, &nb, iparam,
	     &msglvl, b, x, &error);
    conceptsAssert3(error == 0, concepts::Assertion(), "error = " << error);

    // Status
    return  1;
  }

  smat_t* smatNew(unsigned int m, unsigned int n, int is_symmetric,
		  uint* ia, uint* ja, double* a) {
    smat_t  *new_mat;
    new_mat = (smat_t *) malloc (sizeof(smat_t));
    new_mat->m = m;
    new_mat->n = n;
    new_mat->nnz = 0;
    new_mat->sym = is_symmetric;
    new_mat->ia = ia;
    new_mat->ja = ja;
    new_mat->a  = a;
    return new_mat;
  }

} // namespace concepts
