// interface to PARDISO

#ifndef PARDISO_HH
#define PARDISO_HH

#include <cstdlib>
#include "basics/typedefs.hh"
#include "compositions.hh"

namespace concepts {

  // forwards
  template<typename F>
  class SparseMatrix;

  // *************************************************************** Pardiso **

  /** Sparse direct solver for symmetric and unsymmetric matrices. This class
      interfaces to PARDISO [1]. Use the environment variable OMP_NUM_THREADS
      to change the number of threas used (default is 1).

      @see Olaf Schenk, Klaus Grtner, Wolfgang Fichtner, Andreas Stricker, <em>PARDISO: a high-performance serial and parallel sparse linear solver in semiconductor device simulation</em>, Future Generation Computer System 18 (2001), 69-78.
      @author Philipp Frauenfelder, 2003
      @ingroup solver
  */
  class Pardiso : public Operator<Real> {
  public:
    /** Constructor.
	@param matrix Matrix
	@param symmetric Set this to true if the matrix is symmetric and positive definite
    */
    Pardiso(const SparseMatrix<Real>& matrix, bool symmetric = false);
    virtual ~Pardiso();
    virtual void operator()(const Function<Real>& fncY, Function<Real>& fncX);
    void operator()(const Vector<Real>& fncY, Vector<Real>& fncX);
    virtual const Space<Real>& spaceX() const { return spcX_; }
    virtual const Space<Real>& spaceY() const { return spcY_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// System size
    const uint n_;
    /// Image space
    const Space<Real>& spcX_;
    /// Source space
    const Space<Real>& spcY_;
    /// Symmetric system?
    bool symmetric_;
    /// Stores if the factorization was already computed before
    bool factored_;

    /// PARDISO data: number of non-zeros
    uint nnz_;
    /// PARDISO data: matrix in Compressed Sparse Row format
    void* A_;
    /// PARDISO data: entries of the matrix
    Array<Real> a_;
    /// PARDISO data: row and column indices of the matrix
    Array<uint> ia_, ja_;
    /// PARDISO data: parameter
    Array<size_t> handle;
    /// PARDISO data: parameter
    Array<int> iparam;
    /// PARDISO data: parameter
    int max_factorizations, matrix_number, matrix_type, msglvl, error,
      nb, max_threads;

    /// Calls PARDISO to factor the matrix (symbolically and numerically)
    int solver_pardiso_factor_();
    /// Calls PARDISO to do the forward and backward substiution
    int solver_pardiso_solve_(double*, double*);
  };

} // namespace concepts

#endif // PARDISO_HH
