// permutation and restriction/extension operator

#include "permutation.hh"
#include "matrix.hh"

// debugging
#include "basics/debug.hh"

#define PermCompose_D 0
#define RestrConvert_D 0

namespace concepts {

  // *********************************************************** Permutation **

  template<typename F>
  Permutation<F>::Permutation(const Space<F>& space,
			      const Array<int>& perm, bool transpose) :
    spc_(space), p_(perm) {
    conceptsAssert(space.dim() == perm.size(), Assertion());
    if (transpose) {
      for (uint j = 0; j < perm.size(); ++j) p_[perm[j]] = j;
    }
  }

  template<typename F>
  Permutation<F>::Permutation(const Permutation<F>& perm, bool transpose) :
    spc_(perm.spc_), p_(perm.p_) {
    if (transpose) {
      for (uint j = 0; j < p_.size(); ++j) p_[perm.p_[j]] = j;
    }
  }

  template<typename F>
  void Permutation<F>::composeRestr(const TrivExtendRestrict<F>& restr,
				    Matrix<F>& dest) const {
    DEBUGL(PermCompose_D, "dest = " << dest);
    const uint diff = restr.spaceX().dim() > restr.spaceY().dim() ?
      restr.spaceX().dim()-restr.spaceY().dim() :
      restr.spaceY().dim()-restr.spaceX().dim();
    if (restr.extend()) {
      for (uint i = 0; i < spaceX().dim(); ++i)
	if (p_[i] >= (int)diff) dest(i, p_[i]-diff) = 1.0;
    } else
      for (uint i = diff; i < spaceX().dim(); ++i) dest(i-diff, p_[i]) = 1.0;
  }

  template<typename F>
  void Permutation<F>::convertToMatrix(Matrix<F>& dest) const {
    for (uint i = 0; i < spaceX().dim(); ++i)
      dest(i, p_[i]) = 1;
  }

  template<typename F>
  std::ostream& Permutation<F>::info(std::ostream& os) const {
    return os << "Permutation<" << number<F>::name() << ">(" << spc_ << ", "
	      << p_ << ')';
  }

  // **************************************************** TrivExtendRestrict **

  template<typename F>
  void TrivExtendRestrict<F>::convertToMatrix(Matrix<F>& dest) const {
    DEBUGL(RestrConvert_D, *this << ", extend = " << extend_);
    if (extend_) {
      const uint diff = spcX_.dim() - spcY_.dim();
      for (uint j = 0; j < spcY_.dim(); ++j) {
	DEBUGL(RestrConvert_D, "setting (" << j+diff << ", " << j << " to 1");
	dest(j+diff, j) = 1;
      }
    } else {
      const uint diff = spcY_.dim() - spcX_.dim();
      for (uint i = 0; i < spcX_.dim(); ++i) {
	DEBUGL(RestrConvert_D, "setting (" << i << ", " << i+diff << ") to 1");
	dest(i, i+diff) = 1;
      }
    }
  }

  template<typename F>
  std::ostream& TrivExtendRestrict<F>::info(std::ostream& os) const {
    return os << "TrivExtendRestrict<" << concepts::number<F>::name() << ">("
	      << spcX_ << " <- " << spcY_ << ')';
  }

} // namespace concepts
