/* A sparse matrix
 */

#ifndef sparseMatrix_hh
#define sparseMatrix_hh

#include <map>
#include <limits>
#include <memory>
#include "compositions.hh"
#include "matrix.hh"
#include "hashedSMatrix.hh"
// #include "basics/debug.hh"

namespace concepts {

  // forward declaration
  template<class F>
  class BilinearForm;

  // ********************************************************** SparseMatrix **

  /** Sparse matrix. The matrix has the size m x n where m is the
      dimension of the image space (\c spaceX) and n is the dimension of the
      source space (\c spaceY).

      The matrix is setup and assembled in the constructor. It calls the
      bilinear form on every element of the space and uses the T matrices
      of the elements to assemble the element matrices into the global
      matrix.

      There are quite a few solver which can be used to solve the system.

      @see TMatrixBase
      @see CG
      @see GMRes
      @test test::CompositionsTest
      @test test::MoreCompositionsTest
      @test test::DriverTest
  */
  template<class F>
  class SparseMatrix : public Matrix<F> {
  public:
    /** Constructor. Creates an empty matrix.
     */
    SparseMatrix(const Space<F>& spcX, const Space<F>& spcY) :
      spcX_(spcX), nX_(spcX.dim()), spcY_(spcY), nY_(spcY.dim()), 
      m_(new HashedSparseMatrix<F>(spcX.dim(), spcY.dim(), 2)) {}
    /** Constructor. Computes the global matrix by assembling the
	element matrices.

	This constructor features a double loop over the elements of the
	image and the source space. On each combination, the bilinear form is
	called. You can force this constructor to execute the double
	loop in such a way that only for diagonal combinations of the
	elements in both space the integration and assembling is executed.
	Use \c single and set it to \c true.

	Use this constructor, if \c spcX != \c spcY or if you have
	local matrices which express the interaction of the two
	elements.

	In non-symmetric FEM (eg. DGFEM), one has to solve
	A<sup>T</sup> u = f. This constructor computes A and
	not A<sup>T</sup>.

	@param spcX Image space
	@param spcY Source space
	@param bf Bilinear form
    */
    SparseMatrix(const Space<F>& spcX, const Space<F>& spcY, 
		 BilinearForm<F>& bf);
    /** Constructor. Computes the global matrix by assembling the
	element matrices.
    */
    SparseMatrix(const Space<F>& spc, BilinearForm<F>& bf);

    /** Copy constructor. If \c t is set to true, the matrix is
	transposed during the copy process
    */
    SparseMatrix(const SparseMatrix<F>& m, bool t = false);

    /** Constructor of partial rank 1 matrix. The matrix is set to
	\f$ \vec x \cdot \vec x^\top \f$ where the bottom entries of
	x are given by v and, if i >= 0, x(i) = 1 and x(0:i) = 0.
	@param spc Image and source space
	@param v Partial vector for rank 1 product
	@param i Shift index, \c i==-1 means that v has the full size of x
    */
    SparseMatrix(const Space<F>& spc, const F* const v, const int i = -1);

    /** Constructor of rank 1 matrix. The matrix is set to
	\f$ \vec x \cdot \vec y^\top \f$.
	@param spc Image and source space
	@param x First vector
	@param y Second vector
    */
    SparseMatrix(const Space<F>& spc, const Vector<F>& x, const Vector<F>& y);

    /** Constructor. Converts \c A to a sparse matrix.

        In case, the slow mode is allowed, the matrix is computed by
	multiplying \c A with the standard basis vectors and entering
	the results into the matrix.  This is a O(n^2) operation.

	@param A Operator to store as sparse matrix
	@param slow Also do slow O(n^2) conversion if nothing else is left.
    */
    SparseMatrix(Operator<F>& A, bool slow = false);

    virtual void operator()(const Function<F>& fncY, Function<F>& fncX);
    void operator()(const Vector<F>& fncY, Vector<F>& fncX) {
      (*m_)(fncY, fncX); }

    virtual void transpMult(const Vector<F>& fncY, Vector<F>& fncX);

    virtual F operator()(const uint i, const uint j) const {
      return (const_cast<const HashedSparseMatrix<F>&>(*m_))(i,j); }
    virtual F& operator()(const uint i, const uint j) { return (*m_)(i,j); }

    virtual const Space<F>& spaceX() const { return spcX_; }
    virtual const Space<F>& spaceY() const { return spcY_; }

    /** Writes the matrix to a file.
	@param fname Filename
    */
    void write(char* fname) const;

    void storeMatlab(const char* filename) const;

    /// Returns the sparse matrix itself.
    const HashedSparseMatrix<F>* m() const {
      return const_cast<const HashedSparseMatrix<F>*>(m_.get()); }

    /** This matrix is added into to the given matrix.
	@param dest Matrix into which this matrix should be added.
	@param fact Factor by which this matrix should be multiplied.
    */
    void addInto(Matrix<F>& dest, const F fact) const;

    /// Multiplies this matrix with \c fact and adds the result to \c dest
    void multiply(const SparseMatrix<F>& fact, Matrix<F>& dest) const {
      m_->multiply(fact.m(), dest);
    }
    /// Multiplies this matrix with \c fact and adds the result to \c dest
    void multiply(const Matrix<F>& fact, Matrix<F>& dest) const {
      m_->multiply(&fact, dest);
    }
    /// Returns the number of used entries in the matrix
    uint used() const { return m_->used(); }
    /// Memory usage in byte
    float memory() const { return sizeof(SparseMatrix<F>) + m_->memory(); }
    /// Makes sure a theoretically symmetric matrix is symmetric in memory too
    void symmetrize();

    /** Compresses the matrix by dropping small entries. All matrix
	entries which are smaller than a certain \c threshold times
	the largest entry of the matrix are deleted from the
	matrix.
    */
    void compress(Real threshold = std::numeric_limits<Real>::epsilon()) {
      m_->compress(threshold);
    }

    /// Creates a histogram of the matrix entries
    void histogram(std::map<int, uint>& hist) const;

    /// Copies \c n to this matrix
    void copy(const SparseMatrix<F>& n);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Image space
    const Space<F>& spcX_;
    /// Dimension of image space (spcX_)
    uint                nX_;

    /// Source space
    const Space<F>& spcY_;
    /// Dimension of source space (spcY_)
    uint                nY_;

    /// The matrix
    std::auto_ptr<HashedSparseMatrix<F> > m_;

    SparseMatrix(const SparseMatrix<F>&);
    void operator=(const SparseMatrix<F>&);

    /// Counts number of Matlab outputs (used to uniquely name the matrices)
    static uint storeMatlabCounter_;

    void copyConstructor_(const SparseMatrix<F>& m, bool t);
  };

} // namespace concepts

#endif // sparseMatrix_hh
