// interface to SPOOLES

#include <map>
#include <memory>
#include "spooles.hh"
extern "C" {
#  include "misc.h"
#  include "FrontMtx.h"
#  include "SymbFac.h"
#  include "LinSol/Bridge.h"
}
#include "operator/sparseMatrix.hh"
#include "toolbox/array.hh"

// debugging
#include "basics/debug.hh"

#define SpoolesConstr_D 0
#define SpoolesAppl_D 0

namespace concepts {

  // *************************************************************** Spooles **

  Spooles::Spooles(const SparseMatrix<Real>& A, bool symmetric) :
    n_(A.spaceX().dim()), spcX_(A.spaceY()), spcY_(A.spaceX()),
    symmetric_(symmetric), factored_(false) {
    DEBUGL(SpoolesConstr_D, "n = " << n_);
    mtxA = InpMtx_new();
    InpMtx_init((InpMtx*)mtxA, INPMTX_BY_ROWS, SPOOLES_REAL, A.m()->used(), 1);

    // store entries in matrix
    const HashedSparseMatrix<Real> *m = A.m();
    HashedSparseMatrix<Real>::Value** matrixS = m->Matrix();  
    const int hashBits = m->HashBits();
    const int p = 1<<hashBits;
    for(uint r = 0; r < spcX_.dim(); ++r){
      for (int c_mod = 0; c_mod < p; ++c_mod){
	for (HashedSparseMatrix<Real>::Value* v =
	       matrixS[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk){
	  InpMtx_inputRealEntry((InpMtx*)mtxA, r, v->idx, v->val);
	}  
      }
    }

    InpMtx_changeStorageMode((InpMtx*)mtxA, INPMTX_BY_VECTORS);
#if SpoolesConstr_D
    InpMtx_writeForHumanEye((InpMtx*)mtxA, stdout);
#endif
  }

  Spooles::~Spooles() {
    InpMtx_free((InpMtx*)mtxA);
  }

  void Spooles::operator()(const Function<Real>& fncY,
			   Function<Real>& fncX) {
    const Vector<Real>* fncYA;
    if ((fncYA = dynamic_cast<const Vector<Real>*>(&fncY))) {
      Vector<Real>* fncXA;
      if ((fncXA = dynamic_cast<Vector<Real>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException(MissingFeature("unknown function"));
  }

  void Spooles::operator()(const Vector<Real>& fncY, Vector<Real>& fncX) {
    // right hand side and solution
    DenseMtx* x, * y;
    x = DenseMtx_new();
    y = DenseMtx_new();
#if 0
    Array<int> rowindx(fncX.space().dim()), rowindy(fncY.space().dim());
    int colind = 0;
    int* tmp = rowindx;
    for (uint i = 0; i < rowindx.size(); *tmp++ = i++);
    DEBUGL(SpoolesAppl_D, "rowindx = " << rowindx);
    tmp = rowindy;
    for (uint i = 0; i < rowindy.size(); *tmp++ = i++);
    DenseMtx_initWithPointers(x, SPOOLES_REAL, -1, -1, rowindx.size(), 1,
			      1, 1, rowindx, &colind, fncX);
    DenseMtx_initWithPointers(y, SPOOLES_REAL, -1, -1, rowindy.size(), 1,
			      1, 1, rowindy, &colind, fncY);
#else
    DenseMtx_init(x, SPOOLES_REAL, 0, 0, fncX.space().dim(), 1, 1,
		  fncX.space().dim());
    DenseMtx_init(y, SPOOLES_REAL, 0, 0, fncY.space().dim(), 1, 1,
		  fncY.space().dim());
    DenseMtx_zero(x);
    DenseMtx_zero(y);
    for (uint i = 0; i < fncY.space().dim(); ++i)
      DenseMtx_setRealEntry(y, i, 0, fncY(i));
#endif
#if SpoolesAppl_D
    DEBUGL(SpoolesAppl_D, "x");
    DenseMtx_writeForHumanEye(x, stdout);
    DEBUGL(SpoolesAppl_D, "y");
    DenseMtx_writeForHumanEye(y, stdout);
#endif

#if 0
    // find a low-fill ordering
    graph = Graph_new();
    adjIVL = InpMtx_fullAdjacency((InpMtx*)mtxA);
    int nedges = IVL_tsize((IVL*)adjIVL);
    Graph_init2((Graph*)graph, 0, spcX_.dim(), 0, nedges, spcX_.dim(), nedges,
		(IVL*)adjIVL, 0, 0);
#if SpoolesAppl_D
    DEBUGL(SpoolesAppl_D, "graph");
    Graph_writeForHumanEye((Graph*)graph, stdout);
#endif
    frontETree = orderViaMMD((Graph*)graph, 0, SpoolesAppl_D, stdout);

    // get permutation, apply it
    oldToNewIV = ETree_oldToNewVtxPerm((ETree*)frontETree);
    oldToNew = IV_entries((IV*)oldToNewIV);
    newToOldIV = ETree_newToOldVtxPerm((ETree*)frontETree);
    newToOld = IV_entries((IV*)newToOldIV);
    ETree_permuteVertices((ETree*)frontETree, (IV*)oldToNewIV);
    InpMtx_permute((InpMtx*)mtxA, oldToNew, oldToNew);
    if (symmetric_)
      InpMtx_mapToUpperTriangle((InpMtx*)mtxA);
    InpMtx_changeCoordType((InpMtx*)mtxA, INPMTX_BY_CHEVRONS);
    InpMtx_changeStorageMode((InpMtx*)mtxA, INPMTX_BY_VECTORS);
    DenseMtx_permuteRows(y, (IV*)oldToNewIV);
    symbfacIVL= SymbFac_initFromInpMtx((ETree*)frontETree, (InpMtx*)mtxA);
#if SpoolesAppl_D
    DEBUGL(SpoolesAppl_D, "oldToNewIV");
    IV_writeForHumanEye((IV*)oldToNewIV, stdout);
    DEBUGL(SpoolesAppl_D, "newToOldIV");
    IV_writeForHumanEye((IV*)newToOldIV, stdout);
    DEBUGL(SpoolesAppl_D, "frontETree");
    ETree_writeForHumanEye((ETree*)frontETree, stdout);
    DEBUGL(SpoolesAppl_D, "mtxA");
    InpMtx_writeForHumanEye((InpMtx*)mtxA, stdout);
    DEBUGL(SpoolesAppl_D, "y");
    DenseMtx_writeForHumanEye(y, stdout);
    DEBUGL(SpoolesAppl_D, "symbfacIVL");
    IVL_writeForHumanEye((IVL*)symbfacIVL, stdout);
#endif

    // initialize front matrix object
    frontmtx = FrontMtx_new();
    mtxmanager = SubMtxManager_new();
    SubMtxManager_init((SubMtxManager*)mtxmanager, NO_LOCK, 0);
    FrontMtx_init((FrontMtx*)frontmtx, (ETree*)frontETree, (IVL*)symbfacIVL,
		  SPOOLES_REAL, symmetric_, FRONTMTX_DENSE_FRONTS,
		  SPOOLES_PIVOTING, NO_LOCK, 0, 0, (SubMtxManager*)mtxmanager,
		  SpoolesAppl_D, stdout);

    // compute the numeric factorization
    chvmanager = ChvManager_new();
    ChvManager_init((ChvManager*)chvmanager, NO_LOCK, 1);
    Array<double> cpus(10, 0.0);
    Array<int> stats(20, 0);
    double tau = 100., droptol = 0.0;
    int error = 0;
    rootchv = FrontMtx_factorInpMtx((FrontMtx*)frontmtx, (InpMtx*)mtxA, tau,
				    droptol, (ChvManager*)chvmanager,
				    &error, cpus, stats,
				    SpoolesAppl_D, stdout);
#if SpoolesAppl_D
    DEBUGL(SpoolesAppl_D, "frontmtx");
    FrontMtx_writeForHumanEye((FrontMtx*)frontmtx, stdout);
#endif
    conceptsAssert(rootchv == 0, concepts::Assertion()); // matrix singular
    conceptsAssert(error < 0, concepts::Assertion()); // error at front
#else
    // create and setup a Bridge object
    bridge = Bridge_new();
    Bridge_setMatrixParams((Bridge*)bridge, spcX_.dim(),
			   SPOOLES_REAL, symmetric_);
    Bridge_setMessageInfo((Bridge*)bridge, SpoolesAppl_D, stdout);
    rc = Bridge_setup((Bridge*)bridge, (InpMtx*)mtxA);
    conceptsAssert(rc == 1, concepts::Assertion()); // error from Bridge_setup
    DEBUGL(SpoolesAppl_D, "SETUP: graph constr time: "
	   << ((Bridge*)bridge)->cpus[0]
	   << ", graph compress time: " << ((Bridge*)bridge)->cpus[1]
	   << ", graph order time: " << ((Bridge*)bridge)->cpus[2]
	   << ", symb fact time: " << ((Bridge*)bridge)->cpus[3]
	   << ", total setup time: " << ((Bridge*)bridge)->cpus[4]);
    double nfactorops;
#if SpoolesAppl_D
    int nfront, nfind, nfent, nsolveops;
    rc = Bridge_factorStats((Bridge*)bridge, SPOOLES_REAL, symmetric_, &nfront,
			    &nfind, &nfent, &nsolveops, &nfactorops);
    conceptsAssert(rc == 1, concepts::Assertion());
    DEBUGL(SpoolesAppl_D, "factor matrix statistics: " << nfront
	   << " fronts, " << nfind << " indices, " << nfent << " entries, "
	   << nsolveops << " solve ops, " << nfactorops << " factor ops");
#endif

    // factor the matrix
    int error = 0;
    rc = Bridge_factor((Bridge*)bridge, (InpMtx*)mtxA,
		       SPOOLES_PIVOTING, &error);
    DEBUGL(SpoolesAppl_D, "rc = " << rc << ", error = " << error);
    conceptsAssert(rc == 1, concepts::Assertion());
    DEBUGL(SpoolesAppl_D, "FACTORIZATION: "
	   << "permute orig matrix time:: " << ((Bridge*)bridge)->cpus[5]
	   << ", init factor matrix time: " <<  ((Bridge*)bridge)->cpus[6]
	   << ", compute factorization time: " << ((Bridge*)bridge)->cpus[7]
	   << ", post-proc factorization: " << ((Bridge*)bridge)->cpus[8]
	   << ", total factorization time: " << ((Bridge*)bridge)->cpus[9]);
    DEBUGL(SpoolesAppl_D, "statistics: " << ((Bridge*)bridge)->stats[0]
	   << " pivots, " << ((Bridge*)bridge)->stats[1] << " pivot tests, "
	   << ((Bridge*)bridge)->stats[2] << " delayed vertices, "
	   << ((Bridge*)bridge)->stats[3] << " entries in D, "
	   << ((Bridge*)bridge)->stats[4] << " entries in L, "
	   << ((Bridge*)bridge)->stats[5] << " entries in U");
    DEBUGL(SpoolesAppl_D, "mflops: raw "
	   << 1.e-6*nfactorops/((Bridge*)bridge)->cpus[7] << ", overall "
	   << 1.e-6*nfactorops/((Bridge*)bridge)->cpus[9]);

    // solve the system
    rc = Bridge_solve((Bridge*)bridge, SPOOLES_PIVOTING, x, y);
    conceptsAssert(rc == 1, concepts::Assertion());
    DenseMtx_writeForHumanEye(x, stdout);
//     fprintf(msgFile, "\n\n ----- SOLVE -----\n");
//     fprintf(msgFile, 
// 	    "\n    CPU %8.3f : time to permute rhs into new ordering"
// 	    "\n    CPU %8.3f : time to solve linear system"
// 	    "\n    CPU %8.3f : time to permute solution into old ordering"
// 	    "\n CPU %8.3f : total solve time\n",
// 	    bridge->cpus[10], bridge->cpus[11],
// 	    bridge->cpus[12], bridge->cpus[13]);
//     fprintf(msgFile, "\n\n solve: raw mflops %8.3f, overall mflops %8.3f",
// 	    1.e-6*nsolveops/bridge->cpus[11],
// 	    1.e-6*nsolveops/bridge->cpus[13]);
//     fflush(msgFile);
//     if ( msglvl > 2 ) {
//       fprintf(msgFile, "\n\n solution matrix in original ordering");
//       DenseMtx_writeForHumanEye(mtxX, msgFile);
//       fflush(msgFile);
//     }
#endif
  }

  std::ostream& Spooles::info(std::ostream& os) const {
    return os << "Spooles(n = " << n_ << ", "
	      << (factored_ ? "factored" : "not factored") << ")";
  }


} // namespace concepts
