/* SuperLU: direct sparse solver
 */

#include <sstream>
#include <typeinfo>

#include "superLU.hh"
#include "sparseMatrix.hh"
#include "supermatrix.h"
#include "util.h"
#include "dsp_defs.h"
#include "basics/debug.hh"
#include "basics/fortranException.hh"

#define opSuperLUConstr_D 0
#define SuperLUDestr_D 0
#define SuperLUAppl_D 0

namespace concepts {

  // *************************************************************** SuperLU **

  SuperLU::SuperLU(const Operator<Real>& A, enum order ispec) : 
    n_(A.spaceX().dim()), spcX_(A.spaceY()), spcY_(A.spaceX()),
    factorized_(false), ispec_(ispec) {

    DEBUGL(opSuperLUConstr_D, "A = " << A);

    conceptsAssert(A.spaceX().dim() == A.spaceY().dim(), Assertion());
    if (n_ != A.spaceY().dim())
      throw conceptsException(MissingFeature("matrix is not quadratic"));

    const SparseMatrix<Real>* AL = dynamic_cast<const SparseMatrix<Real>*>(&A);
    SparseMatrix<Real>* AM = 0;
    const SparseMatrix<Real>* AN = 0;
    const HashedSparseMatrix<Real> *m;

    if (!AL) {
      std::stringstream error;
      error << "matrix conversion from " << typeid(A).name()
	    << " to SparseMatrix not supported";
      throw conceptsException(MissingFeature(error.str()));
    } else {
      DEBUGL(opSuperLUConstr_D, "got a SparseMatrix");

      AN = AL;
    }

    m = AN->m();
    int used = m->used();        // >= nnz
    int hashBits = m->HashBits();
    int p = 1<<hashBits;
  
    double *a = doubleMalloc(used);
    int *asub = intMalloc(used);
    int *xa = intMalloc(n_+1);

    HashedSparseMatrix<Real>::Value** matrix = m->Matrix();  
    int nnz = 0; 
    for(uint r=0; r<n_; r++){
      xa[r] = nnz;
      for (int c_mod=0; c_mod<p; c_mod++){
	for (HashedSparseMatrix<Real>::Value* v = matrix[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk){
	  if (v->val != 0.0){
	    a[nnz] = v->val; asub[nnz] = v->idx;
	    nnz++;
	  }  
	}
      }
    }
    xa[n_] = nnz;

    DEBUGL(opSuperLUConstr_D, "building SuperMatrix");
    A_ = new SuperMatrix;
    dCreate_CompCol_Matrix((SuperMatrix*)A_, n_, n_, nnz, a, asub, xa, NR, _D,
			   GE);
    DEBUGL(opSuperLUConstr_D, "built SuperMatrix");

    delete AM;

    if ( !(perm_r_ = intMalloc(n_))) ABORT("Malloc fails for perm_r_[] !");
    if ( !(perm_c_ = intMalloc(n_))) ABORT("Malloc fails for perm_c_[] !");
  }

  SuperLU::~SuperLU() {
    DEBUGL(SuperLUDestr_D, "start.");
    Destroy_CompCol_Matrix((SuperMatrix*)A_);
    delete (SuperMatrix*)A_;
    SUPERLU_FREE (perm_r_);
    SUPERLU_FREE (perm_c_);
    if (factorized_) {
      Destroy_SuperNode_Matrix((SuperMatrix*)L_);
      delete (SuperMatrix*)L_;
      Destroy_CompCol_Matrix((SuperMatrix*)U_);
      delete (SuperMatrix*)U_;
    }
    DEBUGL(SuperLUDestr_D, "done.");
  }

  void SuperLU::operator()(const Function<Real>& fncY,
			   Function<Real>& fncX) {

    const Vector<Real>* fncYA;
    if ((fncYA = dynamic_cast<const Vector<Real>*>(&fncY))) {
      Vector<Real>* fncXA;
      if ((fncXA = dynamic_cast<Vector<Real>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return;
      }
    }

    throw conceptsException(MissingFeature("unknown function"));
    exit(1);
  }

  void SuperLU::operator()(const Vector<Real>& fncY, 
			   Vector<Real>& fncX) {
    // first copy the vector since SuperLU overwrites it
    fncX = fncY;
    conceptsAssert(n_ == fncX.n(), Assertion());

    SuperMatrix B;
    int info;
    double *Vec(fncX);
    char Type('N');

    dCreate_Dense_Matrix(&B, n_, 1, Vec, n_, DN, _D, GE);

    DEBUGL(SuperLUAppl_D, "factorized_ = " << factorized_);
    if (factorized_) {
      dgstrs(&Type, (SuperMatrix*)L_, (SuperMatrix*)U_, perm_r_, perm_c_, &B,
	     &info);
      if (info) {
	Destroy_SuperMatrix_Store(&B);
	throw conceptsException(concepts::FortranException
				(info, "dgstrs"));
      }
    } else {
      L_ = new SuperMatrix;
      U_ = new SuperMatrix;
      get_perm_c(ispec_, (SuperMatrix*)A_, perm_c_);
      dgssv((SuperMatrix*)A_, perm_c_, perm_r_, (SuperMatrix*)L_,
	    (SuperMatrix*)U_, &B, &info);
      factorized_ = true;
      if (info) {
	Destroy_SuperMatrix_Store(&B);
	throw conceptsException(concepts::FortranException
				(info, "dgssv"));
      }
    }

    Destroy_SuperMatrix_Store(&B);
  }

  std::ostream& SuperLU::info(std::ostream& os) const {
    return os << "SuperLU(n = " << n_ << ")";
  }

} // namespace concepts
