/* testsuite for conjugate gradients
 */

#include "basics/exceptions.hh"
#include "operator/matrix.hh"
#include "operator/denseMatrix.hh"
#include "operator/sparseMatrix.hh"
#include "operator/PETSc.hh"
#include "operator/cg.hh"
#include "operator/diagonal.hh"

int main(int, char**) {

  using namespace concepts;

  try {
    DummySpace<Real> spc(5);
    DummySpace<Cmplx> spcC(5);
    std::cout << spc << std::endl;
    std::cout << spcC << std::endl;

    std::cout << "=== DenseMatrix, Real ===" << std::endl;
    DenseMatrix<Real> m(spc, spc);
    m(0,0) = 10.0; m(0,1) = m(1,0) = 0.1; m(0,3) = m(3,0) = 3.6;
    m(1,1) = 2.0;
    m(2,2) = 3.0; m(2,4) = m(4,2) = 0.9;
    m(3,3) = 4.0; m(3,4) = m(4,3) = 1.8;
    m(4,4) = 5.0;
    std::cout << m << std::endl;

    Vector<Real> rhs(spc), sol(spc);
    rhs[0] = 1.0; rhs[1] = 2.0; rhs[2] = 3.0; rhs[3] = 4.0; rhs[4] = 5.0;
    std::cout << "rhs = " << rhs << std::endl;

    CG<Real> cgm(m, 1e-5, 10);
    sol = 0.0;
    cgm(rhs, sol);
    std::cout << "sol = " << sol
	      << " should be [-0.28748, 1.01437, 0.85962, 1.04817, 0.46793]"
	      << std::endl;
    std::cout << cgm << std::endl;

    DiagonalMatrix<Real> diag(spc);
    for (uint i = 0; i < spc.dim(); ++i)
      diag(i, i) = m(i, i);
    DiagonalSolver<Real> diagSolver(diag);
    std::cout << "diagSolver (preconditioner) = " << diagSolver << std::endl;
    CG<Real> Pcgm(m, diagSolver, 1e-5, 10);
    sol = 0.0;
    Pcgm(rhs, sol);
    std::cout << "sol = " << sol
	      << " should be [-0.28748, 1.01437, 0.85962, 1.04817, 0.46793]"
	      << std::endl;
    std::cout << Pcgm << std::endl;


    std::cout << "=== SparseMatrix, Real ===" << std::endl;
    SparseMatrix<Real> mS(spc, spc);
    mS(0,0) = 2.0; mS(0,1) = mS(1,0) = 0.3; mS(0,3) = mS(3,0) = 0.5;
    mS(1,1) = 3.0;
    mS(2,2) = 7.0; mS(2,4) = mS(4,2) = 2.0;
    mS(3,3) = 8.0; mS(3,4) = mS(4,3) = 6.0;
    mS(4,4) = 7.0;
    std::cout << mS << " = " << DenseMatrix<Real>(mS) << std::endl;
    rhs[0] = 5.14; rhs[1] = 9.9; rhs[2] = 10.3; rhs[3] = 3.8; rhs[4] = 0.7;
    std::cout << "rhs = " << rhs << std::endl;

    CG<Real> cgmS(mS, 1e-12, 6);
    sol = 0.0;
    cgmS(rhs, sol);
    std::cout << "sol = " << sol
	      << " should be [1.56618, 3.14338, 2.13892, 2.12927, -2.33621]"
	      << std::endl;
    std::cout << cgmS << std::endl;

    for (uint i = 0; i < spc.dim(); ++i)
      diag(i, i) = mS(i, i);
    DiagonalSolver<Real> diagSolverS(diag);
    std::cout << "diagSolver (preconditioner) = " << diagSolverS << std::endl;
    CG<Real> PcgmS(mS, diagSolverS, 1e-12, 6);
    sol = 0.0;
    PcgmS(rhs, sol);
    std::cout << "sol = " << sol
	      << " should be [1.56618, 3.14338, 2.13892, 2.12927, -2.33621]"
	      << std::endl;
    std::cout << PcgmS << std::endl;
    std::cout << "Solve again with starting value " << sol << std::endl;
    PcgmS(rhs, sol);
    std::cout << "sol = " << sol
	      << " should be [1.56618, 3.14338, 2.13892, 2.12927, -2.33621]"
	      << std::endl;
    std::cout << PcgmS << std::endl;
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
