// testing linear combination and composition

#ifndef testCompositions_hh
#define testCompositions_hh

#include <memory>
#include "basics/testcase.hh"
#include "operator/sparseMatrix.hh"
#include "operator/permutation.hh"
#include "sparseqr/driver.hh"
#include "sparseqr/givensRotations.hh"

namespace test {

  using concepts::Real;

  // *************************************************** CompositionTestBase **

  class CompositionTestBase : public TestCase {
  protected:
    template<typename F>
    void compare(const concepts::Matrix<F>& a, concepts::Operator<F>& b);
  };

  // ****************************************************** CompositionsTest **

  /** Testing linear combination and multiplication of sparse matrices
      and the implementation of the evaluation into a sparse matrix.
      @author Philipp Frauenfelder, 2004
      @see concepts::LiCo
      @see concepts::Compose
      @see concepts::SparseMatrix
  */
  class CompositionsTest : public CompositionTestBase {
  public:
    /// Constructor. Sets up three matrices a, b and c.
    CompositionsTest();
    virtual ~CompositionsTest() {}
    virtual void run();

    /// @name Test routines
    //@{
    /// Tests (a+b)*c
    void test1();
    /// Tests a*(b+c)
    void test2();
    /// Tests a+b*c
    void test3();
    /// Tests a*b+c
    void test4();
    /// Tests a*b-c
    void test5();
    //@}
  private:
    concepts::DummySpace<Real> spc;
    concepts::SparseMatrix<Real> a, b, c;
  };

  // ************************************************** MoreCompositionsTest **

  /** Performs more tests on compositions. Tests are performed with a sparse
      matrix, permutations, extensions and restrictions.
      @see concepts::Permutation
      @see concepts::TrivExtendRestrict
      @see concepts::SparseMatrix
      @author Philipp Frauenfelder, 2004
  */
  class MoreCompositionsTest : public CompositionTestBase {
  public:
    /// Constructor. Sets up a matrix a.
    MoreCompositionsTest();
    virtual ~MoreCompositionsTest() {}
    virtual void run();

    /// @name Test routines
    //@{
    /** Tests if Permutation has consistent application operator and
	conversion to sparse matrix
    */
    void testPermutationConsistency();
    /** Tests if TrivExtendRestrict has consistent application operator and
	conversion to sparse matrix
    */
    void testTrivExtendRestrictConsistency();
    /** Tests composisitions of the matrix a with permutations,
	extensions and restrictions
    */
    void testCompositions();
    //@}
  private:
    concepts::DummySpace<Real> spc;
    concepts::SparseMatrix<Real> a;
  };

  // ************************************************** DeepCompositionsTest **

  /** Tests deeply nested compositions resulting from constrained Eigenvalue
      problems
      @author Philipp Frauenfelder, 2004
      @see eigensolver::Constrained
  */
  class DeepCompositionsTest : public CompositionTestBase {
  public:
    /// Constructor. Sets up a matrix a.
    DeepCompositionsTest();
    virtual ~DeepCompositionsTest() {}
    virtual void run();
    void createSparseQR();

    /// @name Test routines
    //@{
    void testCompositions();
    //@}
  private:
    concepts::DummySpace<Real> spc, cspc;
    concepts::SparseMatrix<Real> a, b, c;
    std::auto_ptr<sparseqr::Driver> qr;
    concepts::Permutation<Real> * Pr;
    std::auto_ptr<concepts::Permutation<Real> > Prt;
    sparseqr::GivensRotations<Real> * Qgt, * Qg;
    concepts::TrivExtendRestrict<Real>* ext, * restr;
    std::auto_ptr<concepts::Compose<Real> > PrQgt, QgPrt, restrPrQgt,
      QgPrtext, restrPrQgtA, restrPrQgtAQgPrtext, restrPrQgtM,
      restrPrQgtMQgPrtext;
  };

}

#endif // testCompositions_hh
