// interface to UMFPACK

#ifndef UMFPACK_HH
#define UMFPACK_HH

#include "basics/typedefs.hh"
#include "compositions.hh"

namespace concepts {

  // forwards
  template<typename F>
  class SparseMatrix;

  // *************************************************************** Umfpack **

  /** Sparse direct solver for unsymmetric matrices.
      @see <a href="http://www.cise.ufl.edu/research/sparse/umfpack/">UMFPACK Homepage</a>
      @author Philipp Frauenfelder, 2003
      @ingroup solver
  */
  class Umfpack : public Operator<Real> {
  public:
    /** Constructor. Has to transpose the matrix \c A while filling the UMFPACK
	data structure since \c SparseMatrix has row compressed form and
	UMFPACK needs column compressed form.
	@param A Matrix
	@param symmetric Set this to true if the matrix is symmetric. In this
	       case, this constructor does not need to transpose the matrix
    */
    Umfpack(const SparseMatrix<Real>& A, bool symmetric = false);
    virtual ~Umfpack();
    virtual void operator()(const Function<Real>& fncY, Function<Real>& fncX);
    void operator()(const Vector<Real>& fncY, Vector<Real>& fncX);
    virtual const Space<Real>& spaceX() const { return spcX_; }
    virtual const Space<Real>& spaceY() const { return spcY_; }
    uint iterations() const { return 0; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// System size
    const uint n_;
    /// Image space
    const Space<Real>& spcX_;
    /// Source space
    const Space<Real>& spcY_;

    /// UMFPACK data
    concepts::Array<long int> Ap_, Ai_;
    /// UMFPACK data
    concepts::Array<Real> Ax_;
    /// UMFPACK data
    void* symbolic_, * numeric_;

    /// Stores if the factorization was already computed before
    bool factored_;
  };

} // namespace concepts

#endif // UMFPACK_HH
