/* Elements
 * Elements to build a space
 */

#ifndef spcElement_hh
#define spcElement_hh

#include <cstring>
#include "basics/typedefs.hh"
#include "basics/outputOperator.hh"
#include "basics/exceptions.hh"
#include "basics/vectorsMatricesForward.hh"
#include "toolbox/array.hh"

#include "basics/debug.hh"

// debugging
#define ElementMatrixAppl_D 0
#define BlockElementMatrixAppl_D 0

namespace concepts {

  // forward declaration
  template<class F>
  class ElementGraphics;

  template<class F>
  class TMatrixBase;

  // *************************************************************** Element **

  /** An abstract class for an element of a space.

      The main property of an element are the T matrix (maps the local
      degrees of freedom of the element onto the global degrees of freedom
      of the space) and the tag of the element (describing some properties
      like: member of the space, active or passive and so on).
      @ingroup space
  */
  template<class F>
  class Element : public OutputOperator {
  public:
    /// Default constructor
    Element() : tag_(0) {}
    virtual ~Element() {}

    /// Returns the T matrix of the element
    virtual const TMatrixBase<F>& T() const = 0;

    virtual const ElementGraphics<F>* graphics() const { return 0; }

    /// Returns the tag
    inline uint& tag() { return tag_; }
  private:
    /// Tag information
    uint tag_;
  };

  // ******************************************************** ElementFormula **

  /** Interface for a formula defined element by element.
      @see Formula for a detailed description
      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class ElementFormula : public OutputOperator {
  public:
    /** Evaluates the formula.
	@param elm Element
	@param p Point in space in local element coordinates
	@param t Point in time
    */
    virtual Real operator() (const Element<F>& elm, const Real p,
			     const Real t = 0.0) const = 0;
    virtual Real operator() (const Element<F>& elm, const Real2d& p,
			     const Real t = 0.0) const = 0;
    virtual Real operator() (const Element<F>& elm, const Real3d& p,
			     const Real t = 0.0) const = 0;
  };

  // ********************************************************* ElementMatrix **

  /** Element matrix. This class stores the element matrix computed by
      a bilinear form. These matrices are assembled in a later step.

      @author Philipp Frauenfelder, 2001
  */
  template<class F>
  class ElementMatrix {
  public:
    ElementMatrix(const uint m = 0, const uint n = 0) :
      data_(n*m), m_(m), n_(n), t_(false), i_(0) {}

    /// Assignement operator
    ElementMatrix<F>& operator=(const ElementMatrix<F>& A);

    /// Returns element (i, j)
    F operator()(const uint i, const uint j) const {
      DEBUGL(ElementMatrixAppl_D, '(' << (t_ ? n_ : m_) << 'x'
	     << (t_ ? m_ : n_) << ") - " << (t_ ? j : i) << ", "
	     << (t_ ? i : j));
      conceptsAssert(t_ ? j < m_ : i < m_, Assertion());
      conceptsAssert(t_ ? i < n_ : j < n_, Assertion());
      return t_ ? data_[j * n_ + i] : data_[i * n_ + j]; }
    /// Returns element (i, j)
    F& operator()(const uint i, const uint j) {
      DEBUGL(ElementMatrixAppl_D, '(' << (t_ ? n_ : m_) << 'x'
	     << (t_ ? m_ : n_) << ") - " << (t_ ? j : i) << ", "
	     << (t_ ? i : j));
      conceptsAssert(t_ ? j < m_ : i < m_, Assertion());
      conceptsAssert(t_ ? i < n_ : j < n_, Assertion());
      return t_ ? data_[j * n_ + i] : data_[i * n_ + j];
    }

    /// Returns the number of rows
    uint m() const { return t_ ? n_ : m_; }
    /// Returns the number of columns
    uint n() const { return t_ ? m_ : n_; }

    /// Sets a new size
    void resize(uint m, uint n);

    /// Returns the number of entries in the matrix
    uint size() const { return data_.size(); }

    /// Transposes the matrix
    bool transpose() { return t_ ^= 1; }

    /// Fills the matrix with zeros
    void zeros() { std::memset((F*)data_, 0, n_*m_*sizeof(F)); }

    /// Returns a pointer to the data array
    operator F*() { return (F*)data_+i_; }
    /// Returns a pointer to the data array
    operator const F*() const { return (const F*)data_; }

    /// sets the internal pointer
    void operator()(const uint i = 0) { i_ = i;}
    /// Returns a pointer to the next entrance of data array
    F* operator++() {
      conceptsAssert(i_<m_*n_, Assertion());
      return (F*)data_ + ++i_;
    }

    /// Returns a pointer to the entrance at same column in next row
    F* nextrow() {
      i_ += n_;
      conceptsAssert(i_<m_*n_, Assertion());
      return (F*)data_ + ++i_;
    }

    /// Returns current row
    uint row() { return t_ ? i_ % n_ : i_ / n_; }
    /// Returns current column
    uint col() { return t_ ? i_ / n_ : i_ % n_; }

    std::ostream& info(std::ostream& os) const;
  private:
    /// Data
    Array<F> data_;
    /// Number of rows
    uint m_;
    /// Number of columns
    uint n_;
    /// Transpose
    bool t_;
    /// Entrance in Array definded by index
    uint i_;
  };


  // **************************************************** BlockElementMatrix **

  /** Container class for Element matrix, which consists of blocks
      This class stores the element matrix computed by
      a bilinear form. Only one block is active, where can be written
      into.

      @author Kersten Schmidt, 2003
  */
  template<class F>
  class BlockElementMatrix {
  public:
    BlockElementMatrix(ElementMatrix<F>& em, const uint offm = 0,
		       const uint offn = 0, const uint bm = 0,
		       const uint bn = 0) :
      em_(em), offn_(offn), offm_(offm), bm_(bm), bn_(bn), i_(0), j_(0) {
      if (bm == 0) { bm_ = em_.m() - offm_;}
      if (bn == 0) { bn_ = em_.n() - offn_;}
    }

    /// Returns a pointer to the data array
    operator F*() { return (F*)&em_(offm_+i_,offn_+j_); }
    /// Returns a pointer to the data array
    operator const F*() const { return (const F*)&em_(offm_+i_,offn_+j_); }

    /// sets the internal pointer
    void operator()(const uint i = 0, uint j = 0) { i_ = i; j_ = j; }
    /// Returns a pointer to the next entrance of the data array
    F* operator++() {
      ++j_;
      if (j_>bn_-1) {
	j_ = 0; ++i_;
      }

      DEBUGL(BlockElementMatrixAppl_D, "i_ = " << i_ << ", j_ = " << j_ <<
	     ", bm_ = " << bm_ << ", bn_ = " << bn_ << ", offm_ = " << offm_ 
	     << ", offn_ = " << offn_ );
      conceptsAssert3(i_ < bm_ | j_== 0, Assertion(), "i_ = " << 
		      i_ << ", bm_ = " << bm_);
      return i_ < bm_ ? (F*)&em_(offm_+i_,offn_+j_) : 0;
    }
    /// Returns a pointer to the entrance at same column in next row
    F* nextrow() {
      ++i_;
      conceptsAssert(i_ < bm_ | j_== 0, Assertion());
      return i_ < bm_ ? (F*)&em_(offm_+i_,offn_+j_) : 0;
    }

    /// Returns current row
    uint row() { return i_; }
    /// Returns current column
    uint col() { return j_; }

    std::ostream& info(std::ostream& os) const;
  private:
    /// Element matrix
    ElementMatrix<F>& em_;
    /// Offset of columns and rows
    uint offn_, offm_;
    /// Number of rows in the block
    uint bm_;
    /// Number of columns in the block
    uint bn_;
    /// Entrance in Array defined by Row and Column
    uint i_, j_;
  };

  template<class F>
  std::ostream& operator<<(std::ostream& os, const ElementMatrix<F>& o) {
#ifdef DEBUG
    os << std::flush;
#endif
    return o.info(os);
  }

  template<class F>
  std::ostream& operator<<(std::ostream& os, const BlockElementMatrix<F>& o) {
#ifdef DEBUG
    os << std::flush;
#endif
    return o.info(os);
  }

} // namespace concepts

#endif // spcElement_hh
