/* S Matrix
 * An S matrix changes a T matrix according to a refinement to handle
 * hanging nodes
 */

#include <cstring>
#include <cmath>
#include <memory>
#include "smatrix.hh"
#include "tmatrix.hh"
#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "integration/shapefunction.hh"
#include "toolbox/array.hh"

#include "basics/debug.hh"

// debugging
#define SMatrix1DConstr_D 0
#define SMatrixTensorConstr_D 0
#define SMatrixTensorAppl_D 0
#define SMatrixGeneralTensorConstr_D 0
#define SMatrixGeneralTensorAppl_D 0
#define SMatrixBlockAppl_D 0

#if defined(__hpux) || defined(_IBMR2)
#  define FORTRAN(X) X
#else
#  define FORTRAN(X) X ## _
#endif

#ifdef __cplusplus
extern "C" {
#endif
  extern void FORTRAN(dgesv)(const int& N, const int& NRHS, 
			     double* A, const int& LDA, int* IPIV, 
			     double* B, const int& LDB, int& INFO);
#ifdef __cplusplus
}
#endif

namespace concepts {

  // *********************************************************** SMatrixBase **

  template<class F>
  std::ostream& SMatrixBase<F>::info(std::ostream& os) const {
    return os << "SMatrixBase()";
  }

  template class SMatrixBase<Real>;

  // ************************************************************* SMatrix1D **

  SMatrix1D::SMatrix1D(const ShapeFunction1D<Real>& small,
		       const ShapeFunction1D<Real>& left,
		       const ShapeFunction1D<Real>& right) :
    n_(small.n()),
    data_(n_*n_*2), left_((Real*)data_), right_((Real*)data_+n_*n_) {
    DEBUGL(SMatrix1DConstr_D, "small = " << small);
    DEBUGL(SMatrix1DConstr_D, "left = " << left);
    DEBUGL(SMatrix1DConstr_D, "right = " << right);

    conceptsAssert(n_ == left.n(), Assertion());
    conceptsAssert(n_ == right.n(), Assertion());
    conceptsAssert(small.nP() == left.nP(), Assertion());
    conceptsAssert(small.nP() == right.nP(), Assertion());
    conceptsAssert(small.nP() >= small.n(), Assertion());

    int n = n_;
    int nP = small.nP();
    int info = 0;

    // copy first values of the shapefunctions on whole intervall into matrix S
    Array<Real> S(small.n()*small.n());
    Real* x = (Real*)S;
    for (uint i = 0; i < n_; ++i) {
      const Real* s = small.values() + nP*i;
      for (uint j = n_; j--;) *x++ = *s++;
    }
    DEBUGL(SMatrix1DConstr_D, "S = " << S);

    // copy first values of shapefunctions on left intervall into matrix left_
    x = left_;
    for (uint i = 0; i < n_; ++i) {
      const Real* l = left.values() + nP*i;
      for (uint j = n_; j--;) *x++ = *l++;
    }
    // copy first values of shapefunctions on right intervall
    // into matrix right_
    x = right_;
    for (uint i = 0; i < n_; ++i) {
      const Real* r = right.values() + nP*i;
      for (uint j = n_; j--;) *x++ = *r++;
    }
    DEBUGL(SMatrix1DConstr_D, "data_ = " << data_);

    Array<int> ipiv(small.n(), 0);

    // Solve System of Equations
    // The data is stored in column major order. Therefore, no problems with
    // Fortran.
    FORTRAN(dgesv)(n, 2*n, (Real*)S, n, (int*)ipiv, (Real*)data_, n, info);
    DEBUGL(SMatrix1DConstr_D, "info = " << info << ", data_ = " << data_);
    conceptsAssert(info == 0, Assertion());

    // The computation above is not very precise. Therefore, adjust
    // the small elements to 0. This works for n < 17 (tested) and
    // perhaps some more.
    x = (Real*)data_;
    for (uint i = 2*n*n; i--; x++)
      *x = (std::fabs(*x) > 1.0/(1<<(n+1)) ? *x : 0.0);
    DEBUGL(SMatrix1DConstr_D, "corrected at " << 1.0/(1<<(n+1))
	   << " data_ = " << data_);
    DEBUGL(SMatrix1DConstr_D, "done.");
  }

  SMatrix1D::SMatrix1D(const SMatrix1D& s, const bool flip) :
    n_(s.n_), data_(flip ? 0 : n_*n_*2),
    left_(flip ? s.right_ : (Real*)data_),
    right_(flip ? 0 : ((Real*)data_)+n_*n_) {
    DEBUGL(SMatrix1DConstr_D, "s = " << s << ", flip = " << flip);
    DEBUGL(SMatrix1DConstr_D, "n_ = " << n_ << ", data_ " << data_
	   << ", left_ = " << left_ << ", right_ = " << right_);
    if (!flip) {
      const Real* sd = (const Real*)s.data_;
      Real* d = (Real*)data_;
      for (uint i = 2*n_*n_; i--;) *d++ = *sd++;
    }
    DEBUGL(SMatrix1DConstr_D, "done.");
  }

  void SMatrix1D::operator()(const TColumn<Real>& src,
			     TColumn<Real>& dest) const {
    // uses the S matrix for the left subelement
    conceptsAssert(dest.n() <= n_, Assertion());
    conceptsAssert(src.n() <= n_, Assertion());

    Real* destD = &(dest[0]);
    for (uint i = 0; i < dest.n(); ++i, ++destD) {
      const Real* s = left_ + i;
      const Real* srcD = src.values();
      *destD = 0.0; 
      for (uint j = 0; j < src.n(); ++j, s += n_)
	*destD += *s * *srcD++;
    }
  }

  void SMatrix1D::scale(const Real factor) {
    
    Real* x = left_;
    for (uint i = n_*n_; i--; x++)
      *x *= factor;
    x = right_;
    if (x)
      for (uint i = n_*n_; i--; x++)
	*x *= factor;
  }

  uint SMatrix1D::size() const {
    return n_;
  }

  std::ostream& SMatrix1D::info(std::ostream& os) const {
    os << "SMatrix1D(n = " << n_ << ", [";
    for (uint i = 0; i < n_; ++i) {
      Real* d = left_ + i;
      for (uint j = n_; j--; d += n_) {
  	os << *d << ((j == 0) ? "" : " ");
      } // for j
      os << ((i == n_-1) ? "" : "\n");
    } // for i
    return os << "])";
  }

  // ********************************************************* SMatrixTensor **

  template<class F, int dim>
  SMatrixTensor<F, dim>::SMatrixTensor(const SMatrixBase<F>& matrix,
				       const uint position) :
    matrix_(matrix), position_(position) {
    conceptsAssert(position_ > 0, Assertion());
    conceptsAssert(position_ <= dim, Assertion());
    DEBUGL(SMatrixTensorConstr_D, "done.");
  }

  template<class F, int dim>
  void SMatrixTensor<F, dim>::operator()(const TColumn<F>& src,
					 TColumn<F>& dest) const {
    DEBUGL(SMatrixTensorAppl_D, "src = " << src << ", dest = " << dest);
    const TColumnTensor<F, dim>* srcT =
      dynamic_cast<const TColumnTensor<F, dim>*>(&src);
    conceptsAssert(srcT, Assertion());
    TColumnTensor<F, dim>* destT = dynamic_cast<TColumnTensor<F, dim>*>(&dest);
    conceptsAssert(destT, Assertion());
    DEBUGL(SMatrixTensorAppl_D, "size of matrix = " << matrix_.size());
    DEBUGL(SMatrixTensorAppl_D, "position = " << position_);
    DEBUGL(SMatrixTensorAppl_D, "rel. size of src = "
	   << srcT->n()[position_-1]);
    conceptsAssert(srcT->n()[position_-1] <= matrix_.size(), Assertion());
    DEBUGL(SMatrixTensorAppl_D, "rel. size of dest = "
	   << destT->n()[position_-1]);
    conceptsAssert(destT->n()[position_-1] <= matrix_.size(), Assertion());
    DEBUGL(SMatrixTensorAppl_D, "src = " << *srcT);

    uint sIdx[dim];
    TColumn<F> tmpS(srcT->n()[position_-1], srcT->index());
    TColumn<F> tmpD(destT->n()[position_-1], destT->index());

    if (dim == 2) {
      DEBUGL(SMatrixTensorAppl_D, "outer loop till "
	     << srcT->n()[position_%dim]);
      DEBUGL(SMatrixTensorAppl_D, "inner loop till "
	     << srcT->n()[position_-1]);
      for (uint i = 0; i < srcT->n()[position_%dim]; ++i) {
	sIdx[position_%dim] = i;
	for (uint j = 0; j < srcT->n()[position_-1]; ++j) {
	  sIdx[position_-1] = j;
	  tmpS[j] = (*srcT)[sIdx];
	}
	DEBUGL(SMatrixTensorAppl_D, "i = " << i << ", tmpS = " << tmpS);
	matrix_(tmpS, tmpD);
	DEBUGL(SMatrixTensorAppl_D, "i = " << i << ", tmpD = " << tmpD);
	for (uint j = 0; j < srcT->n()[position_-1]; ++j) {
	  sIdx[position_-1] = j;
	  (*destT)[sIdx] = tmpD[j];
	}
      }
    }

    if (dim == 3) {
      DEBUGL(SMatrixTensorAppl_D, "outer loop till "
	     << srcT->n()[(position_)%dim]);
      DEBUGL(SMatrixTensorAppl_D, "middle loop till "
	     << srcT->n()[(position_+1)%dim]);
      DEBUGL(SMatrixTensorAppl_D, "inner loop till "
	     << srcT->n()[position_-1]);
      for (uint i = 0; i < srcT->n()[(position_)%dim]; ++i) {
      	sIdx[(position_)%dim] = i;
	for (uint j = 0; j < srcT->n()[(position_+1)%dim]; ++j) {
	  sIdx[(position_+1)%dim] = j;
	  for (uint k = 0; k < srcT->n()[position_-1]; ++k) {
	    sIdx[position_-1] = k;
	    tmpS[k] = (*srcT)[sIdx];
	  }
	  DEBUGL(SMatrixTensorAppl_D, "i = " << i << ", j = " << j
		 << ", tmpS = " << tmpS);
	  matrix_(tmpS, tmpD);
	  DEBUGL(SMatrixTensorAppl_D, "i = " << i << ", tmpD = " << tmpD);
	  for (uint k = 0; k < srcT->n()[position_-1]; ++k) {
	    sIdx[position_-1] = k;
	    (*destT)[sIdx] = tmpD[k];
	  }
	}
      }
    }
    DEBUGL(SMatrixTensorAppl_D, "src = " << src << ", dest = " << dest);
  }

  template<class F, int dim>
  uint SMatrixTensor<F, dim>::size() const {
    return (uint)std::pow(matrix_.size(), (Real)dim);
  }

  template<class F, int dim>
  std::ostream& SMatrixTensor<F, dim>::info(std::ostream& os) const {
    return os << "SMatrixTensor<F, " << dim << ">(size = " << size()
	      << ", on pos. " << position_ << ": " << matrix_ << ')';
  }

  template class SMatrixTensor<Real, 2>;
  template class SMatrixTensor<Real, 3>;

  // ************************************************* SMatrixGeneralTensor **

  template<class F, int dim>
  SMatrixGeneralTensor<F, dim>::SMatrixGeneralTensor(const SMatrixBase<F>& matrix,
						     const uint *n, const uint position) :
    matrix_(matrix), position_(position) {
    std::memcpy(n_, n, dim*sizeof(uint));
    conceptsAssert(position_ > 0, Assertion());
    conceptsAssert(position_ <= dim, Assertion());
    DEBUGL(SMatrixGeneralTensorConstr_D, "done.");
  }

  template<class F, int dim>
  SMatrixGeneralTensor<F, dim>::SMatrixGeneralTensor(const SMatrixBase<F>& matrix,
						     const uint position) :
    matrix_(matrix), position_(position) {
    conceptsAssert(position_ > 0, Assertion());
    conceptsAssert(position_ <= dim, Assertion());
    n_[0] = n_[1] = matrix.size();
    if (dim>2) n_[2] = n_[0];
    DEBUGL(SMatrixGeneralTensorConstr_D, "done.");
  }


  template<class F, int dim>
  void SMatrixGeneralTensor<F, dim>::operator()(const TColumn<F>& src,
						TColumn<F>& dest) const {
    DEBUGL(SMatrixGeneralTensorAppl_D, "src = " << src << ", dest = " << dest);
    const TColumnTensor<F, dim>* srcT =
      dynamic_cast<const TColumnTensor<F, dim>*>(&src);
    conceptsAssert(srcT, Assertion());
    TColumnTensor<F, dim>* destT = dynamic_cast<TColumnTensor<F, dim>*>(&dest);
    conceptsAssert(destT, Assertion());
    DEBUGL(SMatrixGeneralTensorAppl_D, "size of matrix = " << matrix_.size());
    DEBUGL(SMatrixGeneralTensorAppl_D, "position = " << position_);
    DEBUGL(SMatrixGeneralTensorAppl_D, "rel. size of src = "
	   << srcT->n()[position_-1]);
    conceptsAssert3(srcT->n()[position_-1] <= matrix_.size(), Assertion(),
		    "srcT->n()[" << position_-1 << "] = " <<
		    srcT->n()[position_-1] << ", matrix_.size() = " <<
		    matrix_.size());
    DEBUGL(SMatrixGeneralTensorAppl_D, "rel. size of dest = "
	   << destT->n()[position_-1]);
    conceptsAssert(destT->n()[position_-1] <= matrix_.size(), Assertion());
    DEBUGL(SMatrixGeneralTensorAppl_D, "src = " << *srcT);

    uint sIdx[dim];
    TColumn<F> tmpS(srcT->n()[position_-1], srcT->index());
    TColumn<F> tmpD(destT->n()[position_-1], destT->index());

    if (dim == 2) {
      DEBUGL(SMatrixGeneralTensorAppl_D, "outer loop till "
	     << srcT->n()[position_%dim]);
      DEBUGL(SMatrixGeneralTensorAppl_D, "inner loop till "
	     << srcT->n()[position_-1]);
      for (uint i = 0; i < srcT->n()[position_%dim]; ++i) {
	sIdx[position_%dim] = i;
	for (uint j = 0; j < srcT->n()[position_-1]; ++j) {
	  sIdx[position_-1] = j;
	  tmpS[j] = (*srcT)[sIdx];
	}
	DEBUGL(SMatrixGeneralTensorAppl_D, "i = " << i << ", tmpS = " << tmpS);
	matrix_(tmpS, tmpD);
	DEBUGL(SMatrixGeneralTensorAppl_D, "i = " << i << ", tmpD = " << tmpD);
	for (uint j = 0; j < srcT->n()[position_-1]; ++j) {
	  sIdx[position_-1] = j;
	  (*destT)[sIdx] = tmpD[j];
	}
      }
    }

    if (dim == 3) {
      DEBUGL(SMatrixGeneralTensorAppl_D, "outer loop till "
	     << srcT->n()[(position_)%dim]);
      DEBUGL(SMatrixGeneralTensorAppl_D, "middle loop till "
	     << srcT->n()[(position_+1)%dim]);
      DEBUGL(SMatrixGeneralTensorAppl_D, "inner loop till "
	     << srcT->n()[position_-1]);
      for (uint i = 0; i < srcT->n()[(position_)%dim]; ++i) {
      	sIdx[(position_)%dim] = i;
	for (uint j = 0; j < srcT->n()[(position_+1)%dim]; ++j) {
	  sIdx[(position_+1)%dim] = j;
	  for (uint k = 0; k < srcT->n()[position_-1]; ++k) {
	    sIdx[position_-1] = k;
	    tmpS[k] = (*srcT)[sIdx];
	  }
	  DEBUGL(SMatrixGeneralTensorAppl_D, 
		 "i = " << i << ", j = " << j << ", tmpS = " << tmpS);
	  matrix_(tmpS, tmpD);
	  DEBUGL(SMatrixGeneralTensorAppl_D, 
		 "i = " << i << ", tmpD = " << tmpD);
	  for (uint k = 0; k < srcT->n()[position_-1]; ++k) {
	    sIdx[position_-1] = k;
	    (*destT)[sIdx] = tmpD[k];
	  }
	}
      }
    }
    DEBUGL(SMatrixGeneralTensorAppl_D, "src = " << src << ", dest = " << dest);
  }

  template<class F, int dim>
  uint SMatrixGeneralTensor<F, dim>::size() const {
    return n_[0]*n_[1]*(dim>2?n_[2]:1);
  }

  template<class F, int dim>
  std::ostream& SMatrixGeneralTensor<F, dim>::info(std::ostream& os) const {
    os << "SMatrixGeneralTensor<F, " << dim << ">(size = " << size()
       << " = (" << n_[0] << ", " << n_[1];
    if (dim>2) os << ", " << n_[2];
    return os << "), on pos. " << position_ << ": " << matrix_ << ')';
  }

  template class SMatrixGeneralTensor<Real, 2>;
  template class SMatrixGeneralTensor<Real, 3>;

  // ******************************************************** SMatrixCompose **

  template<class F>
  void SMatrixCompose<F>::operator()(const TColumn<F>& src,
				     TColumn<F>& dest) const {
    std::auto_ptr<TColumn<F> > tmp(dest.clone());
    if (C_) {
      std::auto_ptr<TColumn<F> > tmp2(dest.clone());
      (*C_)(src, *tmp2);
      B_(*tmp2, *tmp);
    } else
      B_(src, *tmp);
    A_(*tmp, dest);
  }

  template<class F>
  uint SMatrixCompose<F>::size() const {
    uint sA = A_.size();
    uint sB = B_.size();
    return (sA > sB) ? sA : sB;
  }

  template<class F>
  std::ostream& SMatrixCompose<F>::info(std::ostream& os) const {
    os << "SMatrixCompose(" << A_ << " * " << B_;
    if (C_) os << " * " << *C_;
    return os << ')';
  }

  template class SMatrixCompose<Real>;

  // ******************************************************** SMatrixBlock **

  template<class F, int dim>
  SMatrixBlock<F,dim>::SMatrixBlock(const SMatrixBase<F>& A,
				    const SMatrixBase<F>& B,
				    const SMatrixBase<F>* C) :
    A_(A), B_(B), C_(C) {
    n_ = new uint[dim];
    n_[0] = A.size();
    n_[1] = B.size();
    size_ = n_[0] + n_[1];
    if (dim>2) {
      conceptsAssert(C==0, Assertion());
      n_[2] = C->size();
      size_ += n_[2];
    }
  }

  template<class F, int dim>
  void SMatrixBlock<F,dim>::operator()(const TColumn<F>& src,
				       TColumn<F>& dest) const {
    // number of rows of T matrix has to be equal to size of S matrix
    conceptsAssert3(src.n() <= size_, Assertion(),
		    "Tsrc.n = " << src.n() << ", SMatrix.size = " << size_ <<
		    "(" << n_[0] << " + " << n_[1] << (dim>2?" + ":"") << ")");
    DEBUGL(SMatrixBlockAppl_D, "src = " << src << ", dest = " << dest);

    const TColumnBlock<F, dim>* srcT =
      dynamic_cast<const TColumnBlock<F, dim>*>(&src);
    conceptsAssert(srcT, Assertion());
    // each block of the calculated S matrix should be big enough (source)
    conceptsAssert(srcT->n(0)*srcT->n(1) <= A_.size(), Assertion());
    conceptsAssert(srcT->n(2)*srcT->n(3) <= B_.size(), Assertion());
    conceptsAssert(dim==2 || srcT->n(4)*srcT->n(5) <= C_->size(),
		   Assertion());

    TColumnBlock<F, dim>* destT = dynamic_cast<TColumnBlock<F, dim>*>(&dest);
    conceptsAssert(destT, Assertion());
 
    // each block of the calculated S matrix should be big enough (destiation)
    conceptsAssert(destT->n(0)*destT->n(1) <= A_.size(), Assertion());
    conceptsAssert3(destT->n(2)*destT->n(3) <= B_.size(), Assertion(),
		    "destT->n(2)*destT->n(3) = "
		    << destT->n(2)*destT->n(3) << ", B_.size() = "
		    << B_.size());
    conceptsAssert(dim==2 || destT->n(4)*destT->n(5) <= C_->size(),
		   Assertion());

    DEBUGL(SMatrixBlockAppl_D,"A = " << A_ << ", B = " << B_);

    // loop over blocks
    for(uint l=0; l < dim; ++l){
      // filling temporary source T Matrix while copying k.th block of scrT
      TColumnTensor<F,dim> tmpS(*srcT, l);
      DEBUGL(SMatrixBlockAppl_D,l << ".th block : tmpS filled");

      // initialize temporary destination T Matrix 
      TColumnTensor<F,dim> tmpD(destT->n() + l*dim, destT->index());

      // apply k.th S matrix
      switch(l){
      case 0 : A_(tmpS, tmpD); break;
      case 1 : B_(tmpS, tmpD); break;
      case 2 : (*C_)(tmpS, tmpD); break;
      default:  throw conceptsException(concepts::ExceptionBase());
      }
      DEBUGL(SMatrixBlockAppl_D,
	     l << ".th block : S Matrix applied, tmpD = " << tmpD);
      // filling of k.th block of destination T Matrix
      destT->write(tmpD, l);
      DEBUGL(SMatrixBlockAppl_D, "destT = " << *destT);
      DEBUGL(SMatrixBlockAppl_D,l << ".th block : dest filled");
    } // for l, over blocks
  }

  template<class F, int dim>
  std::ostream& SMatrixBlock<F,dim>::info(std::ostream& os) const {
    os << "SMatrixBlock(" << A_ << " , " << B_;
    if (C_) os << " , " << *C_;
    return os << ')';
  }

  template class SMatrixBlock<Real, 2>;
  template class SMatrixBlock<Real, 3>;

} // namespace concepts
