/* Tmatrix
 * T matrices map local dofs to global dofs and are used for the
 * assembling process
 */

#include <cstring>
#include "basics/exceptions.hh"
#include "function/vector.hh"
#include "element.hh"
#include "tmatrix.hh"

#include "basics/debug.hh"

// debugging TIndex
#define TIndexAppl_D 0

// debugging TMatrix
#define TMatrixConstr_D 0
#define TMatrixAppl_D 0
#define TMatrixAppend_D 0
#define TMatrixInfo_D 0
#define TMatrixExtract_D 0

// debugging TColumnTensor
#define TColumnTensorConstr_D 0
#define TColumnTensorConstrBlock_D 0
// debugging TColumnBlock
#define TColumnBlockConstr_D 0

namespace concepts {

  // *************************************************************** TColumn **

  template<class F>
  TColumn<F>::TColumn(uint n, uint idx, TColumn<F>* lnk) : 
    lnk_(lnk), n_(n), idx_(idx), val_(n) {
    val_.zeros();
  }

  template<class F>
  TColumn<F>::TColumn(const TColumn<F>& t) :
    lnk_(0), n_(t.n_), idx_(t.idx_), val_(t.val_) {
  }

  template<class F>
  TColumn<F>::~TColumn() {}

  template<class F>
  TColumn<F>* TColumn<F>::clone() const {
    return new TColumn<F>(*this);
  }
  template<class F>
  std::ostream& TColumn<F>::info(std::ostream& os) const {

    os << "TColumn(n = " << n_ << ", idx = " << idx_ << ", val = [";
    if (n_ > 0) {
      for (uint i = 0; i < n_ - 1; ++i)
	os << val_[i] << ", ";
      os << val_[n_ - 1];
    }
    return os << "])";
  }

  template class TColumn<Real>;
  template class TColumn<bool>;

  // ********************************************************* TColumnTensor **

  template<class F, int dim>
  TColumnTensor<F, dim>::TColumnTensor(const uint* n, const uint idx,
				       TColumn<F>* lnk) :
    TColumn<F>(dim == 2 ? n[0]*n[1] : n[0]*n[1]*n[2], idx, lnk) {
    std::memcpy(n_, n, dim*sizeof(uint));
    DEBUGL(TColumnTensorConstr_D, "n_ = (" << n_[0] << ", "
	   << n_[1] << "), n = " << TColumn<F>::n());
    DEBUGL(TColumnTensorConstr_D, "done.");
  }

  template<class F, int dim>
  TColumnTensor<F, dim>::TColumnTensor(const TColumnTensor<F, dim>& t) :
    TColumn<F>(t) {
    std::memcpy(n_, t.n_, dim*sizeof(uint));
  }

  template<class F, int dim>
  TColumnTensor<F, dim>::TColumnTensor(const TColumnBlock<F, dim>& t, uint k) :
    TColumn<F>(dim == 2 ? t.n()[2*k]*t.n()[2*k+1] :
	       t.n()[3*k]*t.n()[3*k+1]*t.n()[3*k+2], t.index(), t.link()) {
    // copies range of multiindex
    std::memcpy(n_, t.n() + k*dim, dim*sizeof(uint));
    DEBUGL(TColumnTensorConstrBlock_D, "n_ = (" << n_[0] << ", "
	   << n_[1] << "), n = " << TColumn<F>::n());
    // multiindex
    uint ll[dim+1];
    for(ushort i=0;i < dim; ++i) ll[i]=0;
    ll[dim]=k;
    DEBUGL(TColumnTensorConstrBlock_D, "block k = " << ", first ldof of block= " << t.ldof(ll));
    // copies the values of k.th block in t to own values
    std::memcpy(TColumn<F>::values(), t.values() + t.ldof(ll), TColumn<F>::n()*sizeof(F) );
    DEBUGL(TColumnTensorConstrBlock_D, "TColumnTensor = " << *this);
  }

  template<class F, int dim>
  TColumnTensor<F, dim>* TColumnTensor<F, dim>::clone() const {
    return new TColumnTensor<F, dim>(*this);
  }

  template<class F, int dim>
  std::ostream& TColumnTensor<F, dim>::info(std::ostream& os) const {
    uint n[dim];
    os << "TColumnTensor<F, " << dim << ">(n = (";
    for (ushort i = 0; i < dim; ++i)
      os << n_[i] << ((i==dim-1)?")":"") << ", ";
    os << "idx = " << index() << ", val = ";
    recursiveOut_(dim-1, n, os);
    return os << ')';
  }

  template<class F, int dim>
  void TColumnTensor<F, dim>::recursiveOut_(const int d, uint* n,
					    std::ostream& os) const {
    os << '[';
    for (uint& i = n[d] = 0; i < n_[d]; ++i) {
      if (d == 0) os << operator[](n);
      else recursiveOut_(d-1, n, os);
      if (i < n_[d]-1) os << ", ";
    }
    os << ']';
  }

  template class TColumnTensor<Real, 2>;
  template class TColumnTensor<Real, 3>;
  template class TColumnTensor<bool, 2>;
  template class TColumnTensor<bool, 3>;

  // ********************************************************** TColumnBlock **

  template<class F, int dim>
  TColumnBlock<F, dim>::TColumnBlock(const uint* n, const uint idx,
				     TColumn<F>* lnk) :
    TColumn<F>(dim == 2 ? n[0]*n[1]+n[2]*n[3] :
	       n[0]*n[1]*n[2]+n[3]*n[4]*n[5]+n[6]*n[7]*n[8], idx, lnk) {
    std::memcpy(n_, n, dim*dim*sizeof(uint));
    DEBUGL(TColumnBlockConstr_D, "n_ = (" << n_[0] << ", "
	   << n_[1] << "), n = " << TColumn<F>::n());
    DEBUGL(TColumnBlockConstr_D, "done.");
  }

  template<class F, int dim>
  TColumnBlock<F, dim>::TColumnBlock(const TColumnBlock<F, dim>& t) :
    TColumn<F>(t) {
    std::memcpy(n_, t.n_, dim*dim*sizeof(uint));
  }

  template<class F, int dim>
  TColumnBlock<F, dim>* TColumnBlock<F, dim>::clone() const {
    return new TColumnBlock<F, dim>(*this);
  }

  template<class F, int dim>
  void TColumnBlock<F, dim>::write(TColumnTensor<F,dim>& t, uint k) {
    // determine number of values as product of ranges of multiindex
    uint n = 1;
    for(uint i=0;i<dim;++i){
      conceptsAssert(t.n()[i] == n_[dim*k + i], Assertion());
      n*= t.n()[i];
    }
    DEBUGL(TColumnBlockConstr_D, "block k = " << k << ", n = " << n);
    // multiindex
    uint ll[dim+1];
    for(ushort i=0;i < dim; ++i) ll[i]=0;
    ll[dim]=k;
    // copy values
    std::memcpy(TColumn<F>::values()  + ldof(ll), t.values(), n*sizeof(F) );
  }

  template<class F, int dim>
  uint TColumnBlock<F, dim>::ldof(uint* i) const {
    uint k = dim*i[dim];
#ifdef DEBUG
    //DEBUGL(TColumnBlockConstr_D, "block k = " << i[dim]);
    for (ushort j = 0; j < dim; ++j, ++k)
      conceptsAssert3(i[j] < n_[k], Assertion(), "j = " << j << ", i[j] = "
		      << i[j] << ", n_[" << k << "] = " << n_[k]);
    k = dim*i[dim];
#endif
    uint dof = 0;
    for (ushort j = 0; j < i[dim]; ++j)
      if (dim==2)
	dof += n_[2*j]*n_[2*j+1];
      else
	dof += n_[3*j]*n_[3*j+1]*n_[3*j+2];
    dof += dim == 2 ? i[0]+i[1]*n_[k] : i[0]+i[1]*n_[k]+i[2]*n_[k]*n_[k+1];
    return dof;
  }

  template<class F, int dim>
  std::ostream& TColumnBlock<F, dim>::info(std::ostream& os) const {
    uint n[dim+1];
    os << "TColumnBlock<F, " << dim << ">(n = ";
    for (ushort j = 0; j < dim; ++j){
      os << "(";
      for (ushort i = 0; i < dim; ++i)
	os << n_[i+j*dim] << (i<dim-1?", ":"");
      os << ")" << (j<dim-1?",":"");
    }
    os << ", idx = " << index() << ", val = [";
    for(ushort j = 0; j < dim; ++j){
      n[dim] = j;
      recursiveOut_(dim-1, n, j, os);
      if (j<dim-1) os << ", ";
    }
    return os << "])";
  }
  
  template<class F, int dim>
  void TColumnBlock<F, dim>::recursiveOut_(const int d, uint* n, ushort& position,
					   std::ostream& os) const {
    for (uint& i = n[d] = 0; i < n_[d+position*dim]; ++i) {
      if (d == 0) {
	os << operator[](n);
	if (i < n_[position*dim]-1) os << ", ";
      }
      else recursiveOut_(d-1, n, position, os);
      if ((d>0) && (i < n_[d+position*dim]-1)) os << ", ";
    }
  }

  template class TColumnBlock<Real, 2>;
  template class TColumnBlock<Real, 3>;
  template class TColumnBlock<bool, 2>;
  template class TColumnBlock<bool, 3>;

  // *********************************************************** TMatrixBase **

  template<class F>
  TMatrixBase<F>::TMatrixBase(const uint n) : n_(n), m_(0) {}

  template<class F>
  std::ostream& TMatrixBase<F>::info(std::ostream& os) const {
    return os << "TMatrixBase()";
  }

  template<class F> 
  TMatrixBase<F>::~TMatrixBase() {}

  template class TMatrixBase<Real>;

  // **************************************************************** TIndex **

  template<class F> 
  TIndex<F>::TIndex(const uint m, const uint n, uint idx[]) :
    TMatrixBase<F>(n), idx_(m) {
    std::memcpy((uint*)idx_, idx, m * sizeof(uint));
    m_ = m;
    n_ = n;
  }

  template<class F> 
  TIndex<F>::TIndex(const uint m, const uint n, const Array<uint>& idx) :
    TMatrixBase<F>(n), idx_(idx) {
    conceptsAssert(m == idx_.size(), Assertion());
    m_ = m;
    n_ = n;
  }

  template<class F> 
  TIndex<F>::~TIndex() {}

  template<class F> 
  void TIndex<F>::operator()(const ElementMatrix<F>& A,
			     ElementMatrix<F>& B) const {
    B.resize(A.m(), A.n());
    DEBUGL(TIndexAppl_D, "B = A");
    B = A;
  }

  template<class F> 
  void TIndex<F>::usedIdx(TColumn<bool>& c) const {
    conceptsAssert(m() == c.n(), Assertion());
    for (uint i = 0; i < m(); ++i)
      c[i] = true;
  }

  template<class F>
  void TIndex<F>::extract(const concepts::Vector<F>& solution,
			  concepts::Array<F>& coeff) const {
    coeff.resize(m());
    coeff.zeros();
    for (uint i = 0; i < n(); i++)
      coeff[i] += solution(index(i));
  }

  template<class F>
  std::ostream& TIndex<F>::info(std::ostream& os) const {
    return os << "TIndex(idx = " << idx_ << ')';
  }

  template class TIndex<Real>;
  template class TIndex<Cmplx>;

  // *************************************************************** TMatrix **

  template<class F>
  TMatrix<F>::TMatrix(TColumn<Real>* T) :
    TMatrixBase<F>(0), ctrl_(0), data_(0), sz_(0) {
    if (T)
      append(T);
    DEBUGL(TMatrixConstr_D, "done.");
  }

  template<class F>
  TMatrix<F>::~TMatrix() {}

  template<class F>
  void TMatrix<F>::operator()(const ElementMatrix<F>& A, 
			      ElementMatrix<F>& B) const { 
    DEBUGL(TMatrixAppl_D, "A = " << A);
    DEBUGL(TMatrixAppl_D, "T = " << *this);
    conceptsAssert(A.n() == m(), Assertion());
    uint m = A.m();
    B.resize(m, n_);
    DEBUGL(TMatrixAppl_D, "B = " << B);

    const Data* db = (const Data*)data_;
    for(uint j = 0; j < n_; ++j) {
      const Data* de = db + ctrl_[j].sz;

      for(uint i = 0; i < m; ++i) {
	F s = 0.0;
	for(const Data* d = db; d < de; ++d) s += A(i, d->idx) * d->data;
	B(i, j) = s;
      }

      db = de;
    }
  }

  template<class F>
  void TMatrix<F>::applyOne(const ElementMatrix<F>& A, 
			    ElementMatrix<F>& B) const { 
    DEBUGL(TMatrixAppl_D, "A = " << A);
    DEBUGL(TMatrixAppl_D, "T = " << *this);
    conceptsAssert(A.n() == m(), Assertion());
    uint m = A.m();
    B.resize(m, n_);
    DEBUGL(TMatrixAppl_D, "B = " << B);

    const Data* db = (const Data*)data_;
    for(uint j = 0; j < n_; ++j) {
      const Data* de = db + ctrl_[j].sz;

      for(uint i = 0; i < m; ++i) {
	F s = 0.0;
	for(const Data* d = db; d < de; ++d)
	  if (fabs(d->data) == 1)
	    s += A(i, d->idx) * d->data;
	B(i, j) = s;
      }

      db = de;
    }
  }

  template<class F>
  void TMatrix<F>::append(TColumn<Real>* T) {
#if TMatrixAppend_D
    if (T) {
    } else {
      DEBUGL(TMatrixAppend_D, "T is empty");
    }
#endif
    DEBUGL(TMatrixAppend_D, *this);
    uint n = 0;
    uint sz = 0;
    uint mmax = m_;

    for (TColumn<F>* TT = T; TT != NULL; TT = TT->link()) {
      if (mmax < TT->n())
	mmax = TT->n();
      for (uint i = TT->n(); i--;)
	if ((*TT)[i] != 0.0)
	  ++sz; 
      ++n;
    }
    m_ = mmax;

    n += n_;
    ctrl_.resizePreserve(n);
    Control* c = (Control*)ctrl_ + n_;
    n_ = n;

    sz += sz_;
    data_.resizePreserve(sz);
    Data* d = (Data*)data_ + sz_;
    sz_ = sz;

    for (TColumn<F>* TT = T; TT != NULL; TT = TT->link()) {
      DEBUGL(TMatrixAppend_D, "TT = " << *TT);
      c->idx = TT->index();
      c->sz = 0;

      for (uint i = TT->n(); i--;) 
	if ((*TT)[i] != 0.0) {
	  d->idx = i;
	  d->data = (*TT)[i];
	  ++d; 
	  ++c->sz;
	}

      ++c;
    }
    DEBUGL(TMatrixAppend_D, *this);
  }

  template<class F>
  void TMatrix<F>::usedIdx(TColumn<bool>& c) const {
    // number of local degrees of freedom (ldof) of TMatrix,
    // i.e. included TColumns, has to accord with ldof of c
    conceptsAssert3((m() == 0) || (m() == c.n()), Assertion(),
		    "m() = " << m() << ", c.n() = " << c.n());

    // initialize c
    for (uint i = 0; i < c.n(); ++i) c[i] = false;

    const Data* db = (const Data*)data_;
    for(uint j = 0; j < n_; ++j) {
      const Data* de = db + ctrl_[j].sz;
      for(const Data* d = db; d < de; ++d) {
	if (d->data)
	  c[d->idx] = true;
      }
      db = de;
    }
  }

  template<class F>
  void TMatrix<F>::extract(const concepts::Vector<F>& solution,
			   concepts::Array<F>& coeff) const {
    DEBUGL(TMatrixExtract_D, *this);
    coeff.resize(m());
    coeff.zeros();
    if (coeff.size() && data_.size() && ctrl_.size()) {
      const Data* db = data(0);
      const Control* ct = control(0);

      for (uint i = 0; i < n(); ++i) {
	const Data* de = db + ct[i].sz;
	for (const Data* d = db; d < de; ++d) {
	  coeff[d->idx] += solution(ct[i].idx) * d->data;
	}
	db = de;
      }
    }
  }

  template<class F>
  void TMatrix<F>::extractOne(const concepts::Vector<F>& solution,
			      concepts::Array<F>& coeff) const {
    DEBUGL(TMatrixExtract_D, *this);
    coeff.resize(m());
    coeff.zeros();
    if (coeff.size() && data_.size() && ctrl_.size()) {
      const Data* db = data(0);
      const Control* ct = control(0);

      for (uint i = 0; i < n(); ++i) {
	const Data* de = db + ct[i].sz;
	for (const Data* d = db; d < de; ++d) {
	  if (fabs(d->data) == 1)
	    coeff[d->idx] += solution(ct[i].idx) * d->data;
	}
	db = de;
      }
    }
  }

  template<class F>
  std::ostream& TMatrix<F>::info(std::ostream& os) const {

    os << "TMatrix(" << m_ << 'x' << n_ << ", ctrl = ";
    os << ctrl_ << ", data = " << data_;
#if TMatrixInfo_D
    os << ", matrix = ";
    uint nmax = 0;
    for(uint j = 0; j < n_; ++j) {
      if (ctrl_[j].idx > nmax) nmax = ctrl_[j].idx;
    }
    ElementMatrix<F> Ttmp(m_, nmax + 1);
    Ttmp.zeros();
    const Data* db = (const Data*)data_;
    for(uint j = 0; j < n_; ++j) {
      const Data* de = db + ctrl_[j].sz;
      for(const Data* d = db; d < de; ++d) {
	Ttmp(d->idx, ctrl_[j].idx) = d->data;
      }
      db = de;
    }
    os << Ttmp;
#endif
    return os << ")";
  }

  template<class F>
  std::ostream& TMatrix<F>::Control::info(std::ostream& os) const {
    return os << '(' << idx << ',' << sz << ')';
  }

  template<class F>
  std::ostream& TMatrix<F>::Data::info(std::ostream& os) const {
    return os << '(' << idx << ',' << data << ')';
  }

  template class TMatrix<Real>;

} // namespace concepts
