/* driver class for the sparseqr factorization
 */

#ifndef sparseqrDriver_hh
#define sparseqrDriver_hh

#include <memory>
#include "basics/typedefs.hh"
#include "operator/compositions.hh"
#include "operator/sparseMatrix.hh"
#include "operator/permutation.hh"
#include "sparseqr/sparseqr.hh"
#include "sparseqr/givensRotations.hh"

using concepts::Real;

namespace sparseqr {

  // **************************************************************** Driver **

  /** Driver class for sparse QR decomposition. Performs a QR factorization
      on a sparse mxn matrix A, m>=n, such that
      \f[ P_r \cdot Q^\top \cdot A \cdot P_c^\top =
          \begin{pmatrix} R \\ 0 \end{pmatrix} \f]
      the different matrices can be asked for individually. At the same time,
      this class can act as a linear solver using the QR decomposition of the
      matrix.
      @author Philipp Frauenfelder, 2004
      @test test::DriverTest
      @test test::DeepCompositionsTest
  */
  class Driver : concepts::Operator<Real> {
  public:
    /** Constructor.
	The constructor does not compute or copy anything, just initialises
	local variables.
	@param A Matrix to decompose
    */
    Driver(const concepts::SparseMatrix<Real>& A) :
      rank_(0), A_(A), computed_(false),
      Prow_(0), Pcol_(0), Q_(0), Qt_(0), restr_(0), ext_(0), freeDofs_(0),
      prow_(0), pcol_(0), qr_(0) {}
    virtual ~Driver();
    /// Returns \f$P_r\f$
    concepts::Permutation<Real>* Prow();
    /// Returns \f$P_c\f$
    concepts::Permutation<Real>* Pcol();
    /// Returns \f$Q\f$
    sparseqr::GivensRotations<Real>* Q();
    /// Returns \f$Q^\top\f$
    sparseqr::GivensRotations<Real>* Qt();
    /// Returns the restriction from n to the last n-r degrees of freedom
    concepts::TrivExtendRestrict<Real>* restriction();
    /// Returns the extension from the last n-r to n degrees of freedom
    concepts::TrivExtendRestrict<Real>* extension();
    /// Returns the rank of the matrix
    inline int rank();

    virtual void operator()(const concepts::Function<Real>& fncY,
			    concepts::Function<Real>& fncX);
    void operator()(const concepts::Vector<Real>& fncY,
		    concepts::Vector<Real>& fncX);
    virtual const concepts::Space<Real>& spaceX() const { return A_.spaceY(); }
    virtual const concepts::Space<Real>& spaceY() const { return A_.spaceX(); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    void compute_();
    int rank_;
    const concepts::SparseMatrix<Real>& A_;
    bool computed_;
    std::auto_ptr<concepts::Permutation<Real> > Prow_, Pcol_;
    std::auto_ptr<sparseqr::GivensRotations<Real> > Q_, Qt_;
    std::auto_ptr<concepts::TrivExtendRestrict<Real> > restr_, ext_;
    std::auto_ptr<concepts::DummySpace<Real> > freeDofs_;
    concepts::Array<int> prow_, pcol_;
    std::auto_ptr<QR> qr_;
    sparseqr::Qmatrix q_;
  };

  int Driver::rank() {
    if (!computed_) compute_();
    return rank_;
  }

} // namespace sparseqr

#endif // sparseqrDriver_hh
