// Givens rotations

#ifndef givensRotations_hh
#define givensRotations_hh

#include "sparseqr/sparseqr.hh"
#include "operator/matrix.hh"

namespace sparseqr {

  // ******************************************************* GivensRotations **

  /** Given's Rotations. This operator applies a list of givens rotations
      to a vector.

      @param F Number type (Real or Cmplx)
      @param transpose Applies the permutation (\c false) or its transpose
             (ie. its inverse, \c true)

      @author Philipp Frauenfelder, 2002
      @test test::DeepCompositionsTest
      @test test::GolubExample
      @test test::GolubExampleSum
      @test test::MaxwellTransmissionEVP
  */
  template<typename F>
  class GivensRotations : public concepts::Operator<F> {
  public:
    GivensRotations(const concepts::Space<F>& space,
		    const Qmatrix& q, bool transpose) :
      spc_(space), q_(q), transpose_(transpose) {}
    virtual void operator()(const concepts::Function<F>& fncY,
			    concepts::Function<F>& fncX);
    virtual const concepts::Space<F>& spaceX() const { return spc_; }
    virtual const concepts::Space<F>& spaceY() const { return spc_; }
    /** Computes \c fact times Givens and writes the result to \c dest.
	@pre \c fact == \c dest
    */
    void multiply(const concepts::Matrix<F>& fact,
		  concepts::Matrix<F>& dest) const;
    /** Computes Givens times \c fact and writes the result to \c dest.
	@pre \c fact == \c dest
    */
    void multiplyFirst(const concepts::Matrix<F>& fact,
		       concepts::Matrix<F>& dest) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Source and image space
    const concepts::Space<F>& spc_;
    /// Given's rotations
    const Qmatrix& q_;
    const bool transpose_;
  };

  template<typename F>
  void GivensRotations<F>::operator()
    (const concepts::Function<F>& fncY, concepts::Function<F>& fncX) {
    conceptsAssert(fncY.space().dim() == spc_.dim(), concepts::Assertion());
    conceptsAssert(fncX.space().dim() == spc_.dim(), concepts::Assertion());
    fncX = fncY;
    const J* j = q_.qend;
    if (transpose_) j = q_.qbegin;
    while (j != 0) {
      conceptsAssert((uint)j->row1 < spc_.dim(), concepts::Assertion());
      conceptsAssert((uint)j->row2 < spc_.dim(), concepts::Assertion());
      F t = fncX(j->row1);
      if (transpose_) {
	fncX(j->row1) = j->c*fncX(j->row1) + j->s*fncX(j->row2);
	fncX(j->row2) = j->c*fncX(j->row2) - j->s*t;
      } else {
	fncX(j->row1) = j->c*fncX(j->row1) - j->s*fncX(j->row2);
	fncX(j->row2) = j->c*fncX(j->row2) + j->s*t;
      }
      if (transpose_) j = j->next;
      else  j = j->previous;
    }
  }

  /* The following routine does the same for a matrix as operator() above
     does for a vector.

     It has to compute \c dest = \c fact times Given's rotations whereas
     operator() computes \c fncX = Given's rotations times \c fncY.
     \c dest^T = Given's^T times \c fact^T => each column of \c fact^T is
     treated the same way as \c fncY. A column of \c fact^T is a row
     of \c fact.

     The precondition of the routine is \c fact = \c dest, therefore, we
     can work on \c dest only.
  */
  template<typename F>
  void GivensRotations<F>::multiply
  (const concepts::Matrix<F>& fact, concepts::Matrix<F>& dest) const {
    // check sizes of the matrices
    conceptsAssert(fact.spaceX().dim() == dest.spaceX().dim(),
		   concepts::Assertion());
    conceptsAssert(fact.spaceY().dim() == dest.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert(fact.spaceY().dim() == spc_.dim(), concepts::Assertion());
    // loop over rows of dest
    for (uint i = 0; i < dest.spaceX().dim(); ++i) {
      const J* j = q_.qbegin;
      if (transpose_) j = q_.qend;
      while (j != 0) { // loop over Given's rotations
	conceptsAssert((uint)j->row1 < spc_.dim(), concepts::Assertion());
	conceptsAssert((uint)j->row2 < spc_.dim(), concepts::Assertion());
	F t = dest(i, j->row1);
	if (transpose_) {
	  dest(i, j->row1) = j->c*dest(i, j->row1) - j->s*dest(i, j->row2);
	  dest(i, j->row2) = j->c*dest(i, j->row2) + j->s*t;
	} else {
	  dest(i, j->row1) = j->c*dest(i, j->row1) + j->s*dest(i, j->row2);
	  dest(i, j->row2) = j->c*dest(i, j->row2) - j->s*t;
	}
	if (transpose_) j = j->previous;
	else j = j->next;
      } // while j
    } // for i
  }

  template<typename F>
  void GivensRotations<F>::multiplyFirst
  (const concepts::Matrix<F>& fact, concepts::Matrix<F>& dest) const {
    // check sizes of the matrices
    conceptsAssert(fact.spaceX().dim() == dest.spaceX().dim(),
		   concepts::Assertion());
    conceptsAssert(fact.spaceY().dim() == dest.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert(fact.spaceX().dim() == spc_.dim(), concepts::Assertion());
    // loop over columns of dest
    for (uint i = 0; i < dest.spaceY().dim(); ++i) {
      const J* j = q_.qend;
      if (transpose_) j = q_.qbegin;
      while (j != 0) {
	conceptsAssert((uint)j->row1 < spc_.dim(), concepts::Assertion());
	conceptsAssert((uint)j->row2 < spc_.dim(), concepts::Assertion());
	F t = dest(j->row1, i);
	if (transpose_) {
	  dest(j->row1, i) = j->c*dest(j->row1, i) + j->s*dest(j->row2, i);
	  dest(j->row2, i) = j->c*dest(j->row2, i) - j->s*t;
	} else {
	  dest(j->row1, i) = j->c*dest(j->row1, i) - j->s*dest(j->row2, i);
	  dest(j->row2, i) = j->c*dest(j->row2, i) + j->s*t;
	}
	if (transpose_) j = j->next;
	else j = j->previous;
      }
    }
  }

  template<typename F>
  std::ostream& GivensRotations<F>::info(std::ostream& os) const {
    return os << "GivensRotations(" << spc_ << ", " << q_ << ')';
  }

} // namespace sparseqr

#endif // givensRotations_hh
