/*----------------------------------------------------------------------------
Sparse QR solver for linear systems of equations or least squares problems
Copyright (C) 1996, 1997, 1998 Thomas H. Robey

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Inquiries can be directed to trobey@arc.unm.edu or

Thomas H. Robey
925 Madison NE
Albuquerque, NM  87110
USA
----------------------------------------------------------------------------*/

#include <iostream>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <cstdlib>
#include "sparseqr.hh"
#include "fill.hh"
#include "mindegree.hh"

namespace sparseqr {

  int QR::sparseqr()
  {
    int k,l,p,z,md;
    Smatrix *tptr;
    /* Check for zero entries */
    for (int i=0;i<nr;i++) {
      unsigned m=0;
      for (unsigned j=0;j<pmatrix[i].size;j++) {
	if (pmatrix[i].value[j]!=0.0) {
	  if (j!=m) {
	    pmatrix[i].col[m]=pmatrix[i].col[j];
	    pmatrix[i].col[m]=pmatrix[i].col[j];
	  }
	  m++;
	}
      }
      pmatrix[i].size=m;
    }
    /* Build linked list */
    for (int i=0;i<nr;i++) {
      nrow[i]=pmatrix[i].size;
      Smatrix *sprev=NULL;
      for (unsigned j=0;j<pmatrix[i].size;j++) {
	if (sprev==NULL) {
	  srow[i]=new(i,pmatrix[i].col[j],pmatrix[i].value[j],NULL,NULL) Smatrix;
	  sprev=srow[i];
	}
	else {
	  sprev->right=new(i,pmatrix[i].col[j],pmatrix[i].value[j],NULL,NULL)
	    Smatrix;
	  sprev=sprev->right;
	}
      }
      pmatrix[i].~Packed();
    }
    for (int i=nr-1;i>=0;i--) {
      Smatrix *sptr=srow[i];
      while (sptr!=NULL) {
	sptr->down=scol[sptr->col];
	scol[sptr->col]=sptr;
	sptr=sptr->right;
      }
    }
    /* Calculate degrees of columns */
    MinDegree mindegree(this,nc);
    unsigned nnzc=nc;
    unsigned c=0;
    while (c<nnzc) {
      /* Check for zero column (dependent) and move to the right */
      unsigned i=c;
      while (i<nnzc) {
	if (scol[pcol[i]]==NULL) {
	  nnzc--;
	  if (pcol[i]!=pcol[nnzc]) {
	    int j=pcol[i];
	    pcol[i]=pcol[nnzc];
	    pcol[nnzc]=j;
	  }
	}
	else
	  i++;
      }
      if (c==nnzc)
	break;
      /* Search for column of minimum degree */
      k=mindegree.MinSearch(this,c,nnzc,&md);
      /* Move column to left */
      l=pcol[k];
      pcol[k]=pcol[c];
      pcol[c]=l;
      /* Track columns whose degrees will need to be updated */
      mindegree.Track(this,md,c,nnzc);
      /* Zero out entries using Given's rotations */
      k=0;
      for (Smatrix *cptr=scol[pcol[c]];cptr!=NULL;cptr=cptr->down)
	k++;
      /*  cout << "column = " << pcol[c] << ", number of nonzeros in column = " << 
	  k << "\n"; */
      if (k>2) {
	/* Check for identical patterns */
	Smatrix *cpptr1=NULL;
	Smatrix *cptr1=scol[pcol[c]];
	while (cptr1!=NULL) {
	  Smatrix *cpptr2=cptr1;
	  Smatrix *cptr2=cptr1->down;
	  if (cptr2==NULL)
	    break;
	  while (1) {
	    int j=srow[cptr1->row]->match(srow[cptr2->row]);
	    if (j!=0) {
	      k--;
	      if (cptr1->row<cptr2->row) {
		l=givens(cptr1->row,cptr2->row,pcol[c]);
		assert(l==0);
		cptr2=cptr2->MoveDown(&tptr);
		cpptr2->down=cptr2;
		delete tptr;
	      }
	      else {
		l=givens(cptr2->row,cptr1->row,pcol[c]);
		assert(l==0);
		cptr1=cptr1->MoveDown(&tptr);
		if (cpptr1==NULL)
		  scol[pcol[c]]=cptr1;
		else
		  cpptr1->down=cptr1;
		delete tptr;
		break;
	      }
	    }
	    else
	      cptr2=cptr2->MoveDown(&cpptr2);
	    if (cptr2==NULL) {
	      cptr1=cptr1->MoveDown(&cpptr1);
	      break;
	    }
	  }
	}
      }
      if (k>2) {
	Fill *fill=new Fill(scol[pcol[c]],k);
	fill->Init(this,k);
	fill->IndexSort(this,fill->sortindex1,fill->nlist1);
	p=fill->row1[fill->sortindex1[fill->nlist1-1]];
	z=fill->row2[fill->sortindex1[fill->nlist1-1]];
	l=givens(p,z,pcol[c]);
	assert(l==0);
	k--;
	while (k>2) {
	  fill->nlist2=k-1;
	  fill->Update(p,z);
	  fill->PivotFill(this);
	  fill->IndexSort(this,fill->sortindex2,fill->nlist2);
	  fill->Merge(this);
	  p=fill->row1[fill->sortindex1[fill->nlist1-1]];
	  z=fill->row2[fill->sortindex1[fill->nlist1-1]];
	  l=givens(p,z,pcol[c]);
	  assert(l==0);
	  k--;
	}
	delete fill;
      }
      if (k==2) {
	Smatrix *cptr=scol[pcol[c]];
	while (cptr->row==-1)
	  cptr=cptr->down;
	p=cptr->row;
	cptr=cptr->down;
	while (cptr->row==-1)
	  cptr=cptr->down;
	if (cptr->row<p) {
	  z=p;
	  p=cptr->row;
	}
	else
	  z=cptr->row;
	l=givens(p,z,pcol[c]);
	assert(l==0);
      }
      /* Remove entries */
      p=RemoveMarkedEntries(pcol[c]);
      /* Check for small entry in diagonal */
      if (std::fabs(scol[pcol[c]]->value)<rtol)
	RemoveEntry(p,pcol[c]);
      else {
	prow[c]=p;
	mindegree.Untrack(this,p,md,c,nnzc);
	pmatrix[p].init(nrow[p]);
	l=0;
	Smatrix *rptr=srow[p];
	while (rptr!=NULL) {
	  pmatrix[p].col[l]=rptr->col;
	  pmatrix[p].value[l]=rptr->value;
	  l++;
	  Smatrix *sptr=rptr;
	  rptr=rptr->right;
	  delete sptr;
	}
	c++;
      }
    }
    return(c);
  }

  int QR::givens(int p,int z,int selcol) {
    double t,cos,sin,t1,t2;
    static J *jptr=NULL;
    /*cout << "rows:  " << p << " " << z << "\n"; */
    Smatrix *rptr1=srow[p];
    while (rptr1->col!=selcol)
      rptr1=rptr1->right;
    Smatrix *rptr2=srow[z];
    while (rptr2->col!=selcol)
      rptr2=rptr2->right;
    if (std::fabs(rptr2->value)>=std::fabs(rptr1->value)) {
      t=rptr1->value/rptr2->value;
      sin=1.0/std::sqrt(1.0+t*t);
      cos=sin*t;
    }
    else {
      t=rptr2->value/rptr1->value;
      cos=1.0/std::sqrt(1.0+t*t);
      sin=cos*t;
    }
    /* Check to see if rotations need to be stored */
    if (q!=NULL) {
      jptr=new(cos,sin,p,z,q->qend) J;
      assert(jptr!=NULL);
      if (q->qend==NULL)
	q->qbegin=jptr;
      else
	q->qend->next=jptr;
      q->qend=jptr;
    }
    if (b!=NULL) {
      t=b[p];
      b[p]=cos*t+sin*b[z];
      b[z]=cos*b[z]-sin*t;
    }
#ifdef QRSTAT
    ngivens++;
#endif
    /* Apply Given's rotation to matrix */
    rptr1=srow[p];
    rptr2=srow[z];
    Smatrix *rpptr1=NULL;
    Smatrix *rpptr2=NULL;
    while (rptr1!=NULL || rptr2!=NULL) {
      /* Case 1:  Both rows have nonzero entries */
      if (rptr1!=NULL && rptr2!=NULL && rptr1->col==rptr2->col) {
	if (rptr1->col==selcol) {      /* Selected column */
	  rptr1->value=cos*rptr1->value+sin*rptr2->value;
#ifdef QRSTAT
	  flops+=2;
#endif
	  if (rpptr2==NULL)
	    rptr2=rptr2->MarkEntry(&srow[z]);
	  else
	    rptr2=rptr2->MarkEntry(&rpptr2->right);
	  rptr1=rptr1->MoveRight(&rpptr1);
	  nrow[z]--;
	}
	else {
	  t=rptr1->value;
	  rptr1->value=cos*t+sin*rptr2->value;
	  rptr2->value=cos*rptr2->value-sin*t;
	  t1=std::fabs(rptr1->value);
	  t2=std::fabs(rptr2->value);
#ifdef QRSTAT
	  if (ztol<=0.0)
	    flops+=4;
	  else
	    flops+=5;
#endif
	  if (t1<t2) {
	    if ((ztol<=0.0 && t1==0.0) || (ztol>0.0 && t1<=ztol*t2))
	      rptr1=RemoveEntry(p,rptr2->col,rptr1,rpptr1);
	    else
	      rptr1=rptr1->MoveRight(&rpptr1);
	    rptr2=rptr2->MoveRight(&rpptr2);
	  }
	  else {
	    if ((ztol<=0.0 && t2==0.0) || (ztol>0.0 && t2<=ztol*t1))
	      rptr2=RemoveEntry(z,rptr1->col,rptr2,rpptr2);
	    else
	      rptr2=rptr2->MoveRight(&rpptr2);
	    rptr1=rptr1->MoveRight(&rpptr1);
	  }
	}
      }
      else {
	/* Case 2:  Second row has a zero entry */
	if (rptr2==NULL || (rptr1!=NULL && rptr2->col>rptr1->col)) {
	  scol[rptr1->col]=new(z,rptr1->col,-sin*rptr1->value,rptr2,
			       scol[rptr1->col]) Smatrix;
	  if (rpptr2==NULL)
	    srow[z]=scol[rptr1->col];
	  else
	    rpptr2->right=scol[rptr1->col];
	  nrow[z]++;
	  rptr1->value=cos*rptr1->value;
#ifdef QRSTAT
	  fillin++;
	  flops+=2;
#endif
	  rpptr2=scol[rptr1->col];
	  rptr1=rptr1->MoveRight(&rpptr1);
	}
	else {
	  /* Case 3:  First row has a zero entry */
	  scol[rptr2->col]=new(p,rptr2->col,sin*rptr2->value,rptr1,
			       scol[rptr2->col]) Smatrix;
	  if (rpptr1==NULL)
	    srow[p]=scol[rptr2->col];
	  else
	    rpptr1->right=scol[rptr2->col];
	  nrow[p]++;
	  rptr2->value=cos*rptr2->value;
#ifdef QRSTAT
	  fillin++;
	  flops+=2;
#endif
	  rpptr1=scol[rptr2->col];
	  rptr2=rptr2->MoveRight(&rpptr2);
	}
      }
    }
    /*cout << p << ":  ";
      for (Smatrix *rptr=srow[p];rptr!=NULL;rptr=rptr->right) {
      cout << rptr->col << " ";
      printf("%g ",rptr->value);
      }
      cout << "\n";
      cout << z << ":  ";
      for (Smatrix *rptr=srow[z];rptr!=NULL;rptr=rptr->right) {
      cout << rptr->col << " ";
      printf("%g ",rptr->value);
      }
      cout << "\n"; */
    return(0);
  };
  
  void QR::BackSolve(int rank,double *x) {
    double t=0.0;
    for (int i=rank-1;i>=0;i--) {
      x[pcol[i]]=b[prow[i]];
      for (unsigned j=0;j<pmatrix[prow[i]].size;j++) {
	if (pmatrix[prow[i]].col[j]==pcol[i])
	  t=pmatrix[prow[i]].value[j];
	else
	  x[pcol[i]]-=pmatrix[prow[i]].value[j]*x[pmatrix[prow[i]].col[j]];
      }
      x[pcol[i]]/=t;
    }
    for (int i=rank;i<nc;i++)
      x[pcol[i]]=0.0;
  };

  QR::QR(Packed *qrpmatrix,int qrnr,int qrnc,double *qrb,Qmatrix *qrq,
	 int *qrprow,int *qrpcol,double qrztol,
	 double qrrtol) {
    pmatrix=qrpmatrix;
    nr=qrnr;
    nc=qrnc;
#ifdef QRSTAT
    fillin=0;
    flops=0;
    ngivens=0;
#endif
    nrow=new unsigned[nr];
    assert(nrow!=NULL);
#if defined(__SUNPRO_CC) || defined (__INTEL_COMPILER)
    srow = (Smatrix**)std::malloc(nr*sizeof(Smatrix*));
#else
    srow=new (Smatrix*)[nr];
#endif
    assert(srow!=NULL);
    for (int i=0;i<nr;i++)
      srow[i]=NULL;
#if defined(__SUNPRO_CC) || defined (__INTEL_COMPILER)
    scol = (Smatrix**)std::malloc(nc*sizeof(Smatrix*));
#else
    scol=new (Smatrix *)[nc];
#endif
    assert(scol!=NULL);
    for (int i=0;i<nc;i++)
      scol[i]=NULL;
    ztol=qrztol;
    rtol=qrrtol;
    if (qrprow!=NULL) {
      prow=qrprow;
      del_prow = false;
    } else {
      prow=new int[nr];
      assert(prow!=NULL);
      del_prow = true;
      for (int i=0;i<nr;i++)
	prow[i]=i;
    }
    if (qrpcol!=NULL) {
      pcol=qrpcol;
      del_pcol = false;
    } else {
      pcol=new int[nc];
      assert(pcol!=NULL);
      del_pcol = true;
      for (int i=0;i<nc;i++)
	pcol[i]=i;
    }
    b=qrb;
    q=qrq;
  };
  
  QR::~QR() {
    if ((prow!=NULL) && (del_prow))
      delete[] prow;
    if ((pcol!=NULL) && (del_pcol))
      delete[] pcol;
    delete[] nrow;
#if defined(__SUNPRO_CC) || defined (__INTEL_COMPILER)
    std::free(srow);
    std::free(scol);
#else
    delete[] srow;
    delete[] scol;
#endif
  };

  void QR::RemoveEntry(int row,int col) {
    Smatrix *sprev=NULL;
    Smatrix *sptr=srow[row];
    while (sptr->col!=col) {
      sprev=sptr;
      sptr=sptr->right;
    }
    if (sprev==NULL)
      srow[row]=sptr->right;
    else
      sprev->right=sptr->right;
    delete sptr;
    scol[col]=NULL;
    nrow[row]--;
  };

  Smatrix *QR::RemoveEntry(int row,int col,Smatrix *rptr,Smatrix *rpptr) {
    if (rpptr==NULL)
      srow[row]=rptr->right;
    else
      rpptr->right=rptr->right;
    rptr=rptr->right;
    Smatrix *cptr=scol[col];
    Smatrix *cprev=NULL;
    while (cptr->row!=row) {
      cprev=cptr;
      cptr=cptr->down;
    }
    if (cprev==NULL)
      scol[col]=cptr->down;
    else
      cprev->down=cptr->down;
    delete cptr;
    nrow[row]--;
    return(rptr);
  }

  int QR::RemoveMarkedEntries(int c) {
    int p=-1;
    Smatrix *sptr;
    Smatrix *cptr=scol[c];
    while (cptr!=NULL) {
      if (cptr->row!=-1) {
	scol[c]=cptr;
	p=cptr->row;
	sptr=cptr->down;
	cptr->down=NULL;
      }
      else {
	sptr=cptr->down;
	delete cptr;
      }
      cptr=sptr;
    }
    return(p);
  };

}
