/*----------------------------------------------------------------------------
Sample driver for the Sparse QR solver
Copyright (C) 1996, 1997, 1998 Thomas H. Robey

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Inquiries can be directed to trobey@arc.unm.edu or

Thomas H. Robey
925 Madison NE
Albuquerque, NM  87110
USA
----------------------------------------------------------------------------*/

#include <iostream>
#include <fstream>
#include <cassert>
#include <ctime>
#include <memory>
#include "sparseqr.hh"
#include "lex.hh"

using namespace std;
using namespace sparseqr;

int main(int argc,char *argv[])
{
  unsigned nr,nc,number;
  if (argc<2) {
    cerr << "Input file missing\n";
    exit(0);
  }
  bool rhs = true;
  if (argc == 3) {
    cout << "No RHS\n";
    rhs = false;
  }
  lexFile.open(argv[1],ios::in);
  if (!lexFile) {
    cout << "Input file " << argv[1] << " missing" << "\n";
    exit(-2);
  }
  ofstream outFile("sparseqr.out",ios::out);
  if (!outFile)
    exit(-2);
  outFile << "Sparse QR Decomposition\n";
  outFile << "Input file:  " << argv[1] << "\n\n";
  read_unsigned(&nr);
  outFile << "Number of rows = " << nr << "\n";
  read_unsigned(&nc);
  outFile << "Number of columns = " << nc << "\n\n";
  Packed *pmatrix=new Packed[nr];
  assert(pmatrix!=NULL);
  double *b=new double[nr];
  assert(b!=NULL);
  for (unsigned i=0;i<nr;i++) {
    read_unsigned(&number);
    pmatrix[i].init(number);
    for (unsigned j=0;j<pmatrix[i].size;j++) {
       read_integer(pmatrix[i].col+j);
       pmatrix[i].col[j]--;
    }
    for (unsigned j=0;j<pmatrix[i].size;j++)
      read_double(pmatrix[i].value+j);
    read_double(b+i);
    pmatrix[i].Sort();
  }
  lexFile.close();
  std::auto_ptr<QR> qr(0);
  std::auto_ptr<Qmatrix> q(0);
  if (rhs) {
    qr.reset(new QR(pmatrix,nr,nc,b));
  } else {
    q.reset(new Qmatrix());
    qr.reset(new QR(pmatrix,nr,nc,NULL,q.get()));
  }
  time_t start=time((long *) 0);
  int rank=qr->sparseqr();
  assert(rank>=0);
#ifdef QRSTAT
  long int k=0;
  for (int i=0;i<rank;i++)
    k+=qr->nrow[qr->prow[i]];
  cout << rank << " " << qr->flops+4*qr->ngivens << " " << qr->fillin << " " << 
    qr->ngivens << " " << k;
  outFile << "Rank = " << rank << "\n";
  outFile << "Fill-in = " << qr->fillin << "\n";
  outFile << "Number of Given's rotations = " << qr->ngivens << "\n";
  outFile << "Number of flops = " << qr->flops+4*qr->ngivens << "\n";
  outFile << "Non-zero entries after QR decomposition = " << k << "\n\n";
#endif
  time_t finish=time((long *) 0);
  finish=(long int) difftime(finish,start);
  cout << " time = " << finish << "\n";
  if (! rhs) {
    qr->b=qr->q->ApplyQ(b);
  }
  double *x=new double[nc];
  assert(x!=NULL);
  qr->BackSolve(rank,x);
  outFile << "Solution\n";
  for (unsigned i=0;i<nc;i++)
    outFile << x[i] << "\n";
  outFile.close();
  delete[] pmatrix;
  delete[] b;
  delete[] x;
  return 0;
}
