// alpha strategy

#ifndef alpha_hh
#define alpha_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "timestepping/strategy.hh"

namespace concepts {
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  class TimeVector;

  // ***************************************************************** Alpha **

  /** Timestep strategy for the Alpha algorithm of Hilber, Hughes and Taylor
      to solve second order problems in time with no first order time
      derivative.
      \f[
      [ D_2 \partial_t^2 + D_0 ] y(x,t) = f(x,t)
      \f]
      The scheme has one parameter \c alpha. The scheme is implicit, has
      convergence order 2 and is unconditionally stable. The parameter
      \c alpha controls dissipation. For \c alpha = 0, there is no
      dissipation, and the scheme is equivalent to the Newmark scheme with
      default parameters. For -1/3 <= \c alpha < 0, the scheme is dissipative.
      The damping of higher eigenmodes is stronger.
      @see T.J.R. Hughes <em>The Finite Element Method</em>, Dover, Mineola,
      2000
      @author Christoph Winkelmann, 2004
  */
  class Alpha : public TimeStepStrategy {
  public:
    /** Constructor
        @param fabric Solver fabric for solving the occuring systems
        @param D2 Space opeartor D<sub>2</sub>
        @param D0 Space opeartor D<sub>0</sub>
        @param trhs Timedependent external driver  f(x,t)
        @param Y0 Initial condition y(x,0)
        @param Z0 Initial condition d/dt y(x,0)
        @param dt Time step size
        @param alpha Parameter of the Alpha scheme
    */
    Alpha(concepts::SolverFabric<Real>& fabric,
          concepts::Operator<Real>& D2,
          concepts::Operator<Real>& D0,
          timestepping::TimeVector& trhs,
          const concepts::Vector<Real>& Y0,
          const concepts::Vector<Real>& Z0, 
          Real dt, Real alpha = 0.);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real>& D0_;
    /// External driver function
    timestepping::TimeVector &trhs_;
    /// Store velocity Z and acceleration A
    concepts::Vector<Real> Z_, A_;
    /// Parameters of the scheme
    Real alpha_, beta_, gamma_;
  };
}

#endif // alpha_hh
