// Euler strategy

#ifndef euler_hh
#define euler_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "timestepping/strategy.hh"
#include "timestepping/vectors.hh"

namespace concepts {
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  // ************************************************************** Euler **

  /** Timestep strategy for the explicit Euler algorithm
      to solve first order problems in time.
      \f[
      [ D_1 \partial_t + D_0 ] y(x,t) = f(x,t)
      \f]
      The scheme has convergence order 1.
      Please notice that this algorithm is not absolutely stable and may 
      not converge if the stability condition is not fullfilled.
      @see H. R. Schwarz <em>Numerische Mathematik</em>, Teubner, 1997
      @author Christoph Winkelmann, 2002
  */
  class Euler : public TimeStepStrategy {
  public:
    /** Constructor
       @param D1 Space opeartor D<sub>1</sub>
       @param D0 Space opeartor D<sub>0</sub>
       @param trhs Timedependent external driver  f(x,t)
       @param Y0 Initial condition y(x,0)
       @param dt Time step size
     */
    Euler(concepts::SolverFabric<Real>& fabric,
	  concepts::Operator<Real>& D1,
	  concepts::Operator<Real>& D0,
	  timestepping::TimeVector& trhs,
	  const concepts::Vector<Real>& Y0,
	  Real dt);
    virtual ~Euler();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real> &D1_, &D0_;
    /// External driver function
    TimeVector &trhs_;
    /// Store the latest timestep
    TimeVector Yn1_;    // Yn1 = Y(t_{n-1})
  };


  /** Timestep strategy for the explicit Euler algorithm with limiter
      to solve first order problems in time.
      \f[
      [ D_1 \partial_t + D_0 ] y(x,t) = f(x,t)
      \f]
      The scheme has convergence order 1.
      Please notice that this algorithm is not absolutely stable and may 
      not converge if the stability condition is not fullfilled.
      @see H. R. Schwarz <em>Numerische Mathematik</em>, Teubner, 1997
      @author Christoph Winkelmann, 2002
  */
  class LimitingEuler : public TimeStepStrategy {
  public:
    /** Constructor
       @param D1 Space opeartor D<sub>1</sub>
       @param D0 Space opeartor D<sub>0</sub>
       @param trhs Timedependent external driver  f(x,t)
       @param Y0 Initial condition y(x,0)
       @param dt Time step size
     */
    LimitingEuler(concepts::SolverFabric<Real>& fabric,
		  concepts::Operator<Real>& D1,
		  concepts::Operator<Real>& D0,
		  concepts::Operator<Real>& limiter,
		  timestepping::TimeVector& trhs,
		  const concepts::Vector<Real>& Y0,
		  Real dt);
    virtual ~LimitingEuler();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real> &D1_, &D0_, &limiter_;
    /// External driver function
    TimeVector &trhs_;
    /// Store the latest timestep
    TimeVector Yn1_;    // Yn1 = Y(t_{n-1})
  };

}

#endif // euler_hh
