// nystroem strategy

#ifndef nystroem_hh
#define nystroem_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "timestepping/strategy.hh"

namespace concepts {
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  class TimeVector;

  // ************************************************************** Nystroem **

  /** Timestep strategy for a Nystr&ouml;m algorithm to solve second order
      problems in time with no first order time derivative.
      \f[
      [ D_2 \partial_t^2 + D_0 ] y(x,t) = f(x,t)
      \f]
      The scheme needs no parameters. It is explicit, has convergence
      order 2, is symplectic and conditionally stable. The stability
      criterion for the model problem \f$\partial_t^2 y = -\omega^2
      y\f$ is \f$\Delta t < 2/\omega\f$. For linear operators and
      f(x,t)=0, this scheme is identic to the leap frog scheme.
      @see E. Hairer, S.P. N&oslash;rsett, G. Wanner <em>Solving
      Ordinary Differential Equations I - Nonstiff Problems</em>,
      Springer, Berlin, 1992.
      @author Christoph Winkelmann, 2004
  */
  class Nystroem : public TimeStepStrategy {
  public:
    /** Constructor
        @param fabric Solver fabric for solving the occuring systems
        @param D2 Space opeartor D<sub>2</sub>
        @param D0 Space opeartor D<sub>0</sub>
        @param trhs Timedependent external driver  f(x,t)
        @param Y0 Initial condition y(x,0)
        @param Z0 Initial condition d/dt y(x,0)
        @param dt Time step size
    */
    Nystroem(concepts::SolverFabric<Real>& fabric,
          concepts::Operator<Real>& D2,
          concepts::Operator<Real>& D0,
          timestepping::TimeVector& trhs,
          const concepts::Vector<Real>& Y0,
          const concepts::Vector<Real>& Z0, 
          Real dt);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real>& D0_;
    /// External driver function
    timestepping::TimeVector &trhs_;
    /// Store velocity times timestep in Z
    concepts::Vector<Real> Z_;
  };
}

#endif // nystroem_hh
