
// newmark strategy

#ifndef rungekutta4_hh
#define rungekutta4_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "strategy.hh"

namespace concepts {
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  class TimeVector;

  // ************************************************* Runge Kutta 4th order **

  /** Timestep strategy for the explicit 4th order RungeKutta algorithm 
      to solve first order problems in time.
      \f[
      ( D_1 \partial_t + D_0 ) y(x,t) = f(x,t)
      \f]
      Please notice that this algorithm is not absolutely stable and
      may not converge if the stability condition is not fullfilled.
      @see H. R. Schwarz <em>Numerische Mathematik</em>, Teubner, 1997
      @author Manuel Walser, 2002
  */
  class RungeKutta4 : public TimeStepStrategy {
  public:
    /** Constructor
	@param fabric Solver fabric for solving the occuring systems
	@param D1 Space opeartor D<sub>1</sub>
	@param D0 Space opeartor D<sub>0</sub>
	@param trhs Timedependent external driver  f(x,t)
	@param Y0 Initial condition y(x,0)
	@param dt Time step size
	@param alpha Parameter of the Runge Kutta 4 scheme
    */
    RungeKutta4(concepts::SolverFabric<Real>& fabric,
		concepts::Operator<Real>& D1,
		concepts::Operator<Real>& D0,
		timestepping::TimeVector& trhs,
		const concepts::Vector<Real>& Y0,
		Real dt);
    virtual ~RungeKutta4();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real> &D1_, &D0_;
    /// External driver function
    timestepping::TimeVector &trhs_;
    /// Store the latest timestep
    concepts::Vector<Real> Yn1_;    // Yn1 = Y(t_{n-1})
  };

}

#endif // rungekutta4_hh
