// strategy for timestepping

#ifndef timestpStrg_hh
#define timestpStrg_hh

#include "basics/typedefs.hh"
#include "operator/compositions.hh"
#include "function/vector.hh"
#include "operator/solverFabric.hh"

namespace timestepping {

  using concepts::Real;
  
  class TimeStepping;

  // ****************************************************** TimeStepStrategy **

  /** Abstract strategy class for the class Timestepping. The task of
      a derivated strategy class is to implement the specific
      calculation of the right and left hand side of the equation
      \f[
      ( D + const.(\Delta t) D') Y_n = V(Y_{n-1}, Y_{n-2}, ...),
      \f]
      The initial conditions are given at time 0.
      @see TimeStepping
      @author Manuel Walser, 2002
  */
  class TimeStepStrategy : public concepts::OutputOperator {
  public:
    /** Constructor
	@param lhs Operator on the lhs
	@param sol Solution vector
	@param rhs laod vector
     */
    TimeStepStrategy(concepts::SolverFabric<Real>& fabric,
		     concepts::Operator<Real>* lhs, 
		     const concepts::Vector<Real>& sol, 
		     const concepts::Vector<Real>& rhs,
		     Real dt) : 
      solver_(fabric(*lhs)), 
      liCo_(lhs),
      sol_(sol),
      rhs_(rhs),
      dt_(dt),
      t_(0) {}
  protected:
    friend class TimeStepping;
    /** The overloaded member function next() has to calculate the new
	right hand side and to release the solution vector. Then the
	Timestepping solver can set the new solution.
     */
    virtual void next() = 0;
    /// Solver for the linear system
    std::auto_ptr<concepts::Operator<Real> > solver_;
    /** Operator of the linear equation system which is solved by the 
	friend class TimeStepping. It can be stored as a linear combination 
	of two operators. The exact form depends on the specific scheme.
	@see TimeStepping
     */
    concepts::Operator<Real>* liCo_;
    /** The solution vector of the linear equation system which is
	solved by the friend class TimeStepping.  
	@see TimeStepping
    */
    concepts::Vector<Real> sol_;       // sol_ = Y(t_n)
    /** The right hand side vector of the linear equation system which
	is solved by the friend class TimeStepping.  
	@see TimeStepping
    */
    concepts::Vector<Real> rhs_;       // rhs_ = rhs(t_n)
    /// Time step size
    Real dt_;                          // dt_ = t_n - t_{n-1}, 
    /// Time of the actual solution. 
    Real t_;                           // t_ = t_n
  };

}

#endif // timestpStrg_hh
