/* formula class
 */

#ifndef formula_hh
#define formula_hh

#include <string>
#include "basics/outputOperator.hh"
#include "basics/exceptions.hh"
#include "basics/vectorsMatricesForward.hh"
#include "array.hh"

namespace concepts {

  // **************************************************** FormulaSyntaxError **

  /** Exception indication that a formula contains a syntax error reported
      by the parser.
      @author Philipp Frauenfelder, 2001
  */
  class FormulaSyntaxError : public MissingFeature {
  public:
    /** Constructor.
	@param errMsg Message which will be printed
    */
    FormulaSyntaxError(const std::string& errMsg) throw();
  protected:
    virtual std::ostream& info(std::ostream& os) const throw();
  };

  // *************************************************************** Formula **

  /** Interface for a formula. A general formula would be
      \f$ f : R^n \mapsto R^m \f$. This class covers the possibilities
      \f$ f : R^n \mapsto R \f$ where n is 1, 2 or 3 in space and 0 or 1 in
      time. The image space is just one dimensional since higher dimensions
      can be simulated by taking a \c Formula in every component.

      The application operators of this class evaluate the formula in a given
      point in space \c p and time \c t. If a specialization of this
      interface does not implement a time dependent formula, \c t is simply
      discarded.

      @author Philipp Frauenfelder, 2002
  */
  class Formula : public OutputOperator {
  public:
    /** Creates clone of itself and returns it. This clone function implements
	a virtual constructor (cf. Stroustrup p. 424) which is necessary
	because \c Boundary acts as a container class for a \c Formula and
	its derived classes.

	@see BoundaryCondition
    */
    virtual Formula* clone() const = 0;

    /** Application operator.
	Evaluates the formula.
	@param p Point in space
	@param t Point in time
    */
    virtual Real operator() (const Real p, const Real t = 0.0) const = 0;
    virtual Real operator() (const Real2d& p, const Real t = 0.0) const = 0;
    virtual Real operator() (const Real3d& p, const Real t = 0.0) const = 0;
  };

  // ********************************************************* ParsedFormula **

  /** Parses the given string and evaluates it on request.

      The following operations and functions are allowed in the
      formula: + - / and *, (), ^ for exponentiation (also with broken
      exponents but only with positive bases) and the following formulae:
      <ul>
      <li>sin
      <li>cos
      <li>atan
      <li>sinh
      <li>cosh
      <li>sqr
      <li>sqrt
      <li>exp
      <li>ln
      </ul>
      x, y and z are allowed as variables and pi is defined.

      @see parse
      @see process

      @author Philipp Frauenfelder, 2000
  */
  class ParsedFormula : public Formula {
  public:
    /// Constructor
    ParsedFormula(const char* formula);
    virtual ~ParsedFormula();
    virtual ParsedFormula* clone() const;

    virtual Real operator() (const Real p, const Real t = 0.0) const;
    virtual Real operator() (const Real2d& p, const Real t = 0.0) const;
    virtual Real operator() (const Real3d& p, const Real t = 0.0) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    /// The original formula
    std::string formula_;
    /// Processing the formula
    Real process_(const Real3d& p) const;
    /// Copy constructor
    ParsedFormula(const ParsedFormula& frm);
  private:
    /// The parsed formula
    Array<uchar> pgm_;
    /// Length of the parsed formula
    uint len_;
  };

  // **************************************************** ZylindricalFormula **

  /** Formula in zylindrical coordinates. In such a formula, x has the meaning
      of the radius and y has the meaning of the angle. The input parameters
      of the application operators has to be given in Kartesian coordinates.

      @author Philipp Frauenfelder, 2002
  */
  class ZylindricalFormula : public ParsedFormula {
  public:
    ZylindricalFormula(const char* formula) : ParsedFormula(formula) {}
    virtual ~ZylindricalFormula();
    virtual ZylindricalFormula* clone() const;

    virtual Real operator() (const Real p, const Real t = 0.0) const;
    virtual Real operator() (const Real2d& p, const Real t = 0.0) const;
    virtual Real operator() (const Real3d& p, const Real t = 0.0) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    ZylindricalFormula(const ZylindricalFormula& frm) : ParsedFormula(frm) {}
  };

  // ****************************************************** SphericalFormula **

  /** Formula in spherical polar coordinates. In such a formula, x has
      the meaning of the radius, y has the meaning of the angle phi
      and z has the meaning of the angle theta. The input parameters
      of the application operators has to be given in Kartesian
      coordinates.

      \f[ r^2 = x^2 + y^2 + z^2, \tan \phi = \frac{y}{x},
          \cos \theta = \frac{z}{r}, \f]
      where \f$\phi \in (-\pi, \pi)\f$ and \f$\theta \in (0,\pi)\f$.

      @author Philipp Frauenfelder, 2002
  */
  class SphericalFormula : public ParsedFormula {
  public:
    SphericalFormula(const char* formula) : ParsedFormula(formula) {}
    virtual ~SphericalFormula();
    virtual SphericalFormula* clone() const;

    virtual Real operator() (const Real p, const Real t = 0.0) const;
    virtual Real operator() (const Real2d& p, const Real t = 0.0) const;
    virtual Real operator() (const Real3d& p, const Real t = 0.0) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    SphericalFormula(const SphericalFormula& frm) : ParsedFormula(frm) {}
  };

} // namespace concepts

#endif // formula_hh
