#include <sys/time.h>
#include <sys/utsname.h>
#include <unistd.h>

#include <fstream>
#include <cstring>
#include <cstdio>
#include <sstream>
#include "inputOutput.hh"
#include "basics/exceptions.hh"
#include "basics/debug.hh"

#ifdef DEBUG
#define tbxParametersAdd_D 0
#endif

extern "C" {
  concepts::InputParser::YY_BUFFER_STATE yy_create_buffer(std::FILE *, int);
  void yy_switch_to_buffer(...);
  void yy_delete_buffer(...);
}

int yyparse(void* param);

namespace concepts {

  // ****************************************************** MissingParameter **

  MissingParameter::MissingParameter(const std::string& parameter) throw() : 
    MissingFeature(parameter) {}

  std::ostream& MissingParameter::info(std::ostream& os) const throw() {
    return os << '[' << file_ << ", line " << line_ << "] " << function_ 
	      << " -- parameter " << feature_ << " does not exist";

  }

  // ******************************************************* InOutParameters **

  InOutParameters::~InOutParameters() {

    std::map<const char*, double, ltstr>::iterator id;
    std::map<const char*, std::string, ltstr>::iterator is;
    std::map<const char*, int, ltstr>::iterator ii;
    std::map<const char*, bool, ltstr>::iterator ib;

    while ( (id = double_.begin()) != double_.end()) {
      const char* c = id->first;
      double_.erase(id);
      std::free((char*)c);
    }

    while ( (is = string_.begin()) != string_.end()) {
      const char* c = is->first;
      string_.erase(is);
      std::free((char*)c);
    }

    while ( (ii = int_.begin()) != int_.end()) {
      const char* c = ii->first;
      int_.erase(ii);
      std::free((char*)c);
    }

    while ( (ib = bool_.begin()) != bool_.end()) {
      const char* c = ib->first;
      bool_.erase(ib);
      std::free((char*)c);
    }

    std::map<const char*, std::map<int, double>*, ltstr>::iterator iadouble;
    std::map<const char*, std::map<int, std::string>*,
      ltstr>::iterator iastring;
    std::map<const char*, std::map<int, int>*, ltstr>::iterator iaint;
    std::map<const char*, std::map<int, bool>*, ltstr>::iterator iabool;

    while ( (iadouble = doubleArrays_.begin()) != doubleArrays_.end() ) {
      const char* c = iadouble->first;
      std::map<int, double>* h = iadouble->second;
      doubleArrays_.erase(iadouble);
      delete h;
      std::free((char*)c);
    }

    while ( (iastring = stringArrays_.begin()) != stringArrays_.end() ) {
      const char* c = iastring->first;
      std::map<int, std::string>* h = iastring->second;
      stringArrays_.erase(iastring);
      delete h;
      std::free((char*)c);
    }

    while ( (iaint = intArrays_.begin()) != intArrays_.end() ) {
      const char* c = iaint->first;
      std::map<int, int>* h = iaint->second;
      intArrays_.erase(iaint);
      delete h;
      std::free((char*)c);
    }

    while ( (iabool = boolArrays_.begin()) != boolArrays_.end() ) {
      const char* c = iabool->first;
      std::map<int, bool>* h = iabool->second;
      boolArrays_.erase(iabool);
      delete h;
      std::free((char*)c);
    }
  }

  void InOutParameters::addDouble(const char* name, const double value) {

    std::map<const char*, double, ltstr>::iterator i = 
      double_.find(name);

    if (i == double_.end()) {
      char* n = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
      std::strncpy(n, name, std::strlen(name));
      DEBUGL(tbxParametersAdd_D, "adding " << name << " = " 
	     << value);
      double_.insert(std::pair<const char*, double>(n, value));
    } else {
      i->second = value;
    }
  }

  void InOutParameters::addString(const char* name, const char* value) {

    std::map<const char*, std::string, ltstr>::iterator i = 
      string_.find(name);

    if (i == string_.end()) {
      char* n = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
      std::strncpy(n, name, std::strlen(name));
      DEBUGL(tbxParametersAdd_D, "adding " << name << " = " 
	     << value);
      if (value)
	string_.insert(std::pair<const char*, std::string>(n, value));
      else
	string_.insert(std::pair<const char*, std::string>(n, "(empty)"));
    } else {
      i->second = value;
    }
  }

  void InOutParameters::addInt(const char* name, const int value) {

    std::map<const char*, int, ltstr>::iterator i = 
      int_.find(name);

    if (i == int_.end()) {
      char* n = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
      std::strncpy(n, name, std::strlen(name));
      DEBUGL(tbxParametersAdd_D, "adding " << name << " = " 
	     << value);
      int_.insert(std::pair<const char*, int>(n, value));
    } else {
      i->second = value;
    }
  }

  void InOutParameters::addBool(const char* name, const int value) {

    std::map<const char*, bool, ltstr>::iterator i = bool_.find(name);

    if (i == bool_.end()) {
      char* n = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
      std::strncpy(n, name, std::strlen(name));
      DEBUGL(tbxParametersAdd_D, "adding " << name << " = " 
	     << value);
      bool_.insert(std::pair<const char*, bool>(n, value));
    } else {
      i->second = value;
    }
  }

  double InOutParameters::getDouble(const char* name) const {

    std::map<const char*, double, ltstr>::const_iterator
      i = double_.find(name);

    if (i != double_.end())
      return i->second;
    throw conceptsException(MissingParameter(name));
  }

  std::string InOutParameters::getString(const char* name) const {

    std::map<const char*, std::string, ltstr>::const_iterator
      i = string_.find(name);

    if (i != string_.end())
      return i->second;
    throw conceptsException(MissingParameter(name));
  }

  int InOutParameters::getInt(const char* name) const {

    std::map<const char*, int, ltstr>::const_iterator i = int_.find(name);

    if (i != int_.end())
      return i->second;
    throw conceptsException(MissingParameter(name));
  }

  bool InOutParameters::getBool(const char* name) const {

    std::map<const char*, bool, ltstr>::const_iterator i = bool_.find(name);

    if (i != bool_.end())
      return i->second;
    throw conceptsException(MissingParameter(name));
  }

  void InOutParameters::addArrayDouble(const char* array) { 

    std::map<const char*, std::map<int, double>*, ltstr>::iterator i = 
      doubleArrays_.find(array);
    if (i == doubleArrays_.end()) {
      char* a = (char*)std::calloc(std::strlen(array)+1,sizeof(char));
      std::strncpy(a, array, std::strlen(array));
      DEBUGL(tbxParametersAdd_D, "adding array " << array);
      doubleArrays_.insert
	(std::pair<const char*, std::map<int, double>* >
	 (a, new std::map<int, double>()));
    }
  }

  void InOutParameters::addArrayString(const char* array) { 

    std::map<const char*, std::map<int, std::string>*, ltstr>::iterator i = 
      stringArrays_.find(array);
    if (i == stringArrays_.end()) {
      char* a = (char*)std::calloc(std::strlen(array)+1,sizeof(char));
      std::strncpy(a, array, std::strlen(array));
      DEBUGL(tbxParametersAdd_D, "adding array " << array);
      stringArrays_.insert
	(std::pair<const char*, std::map<int, std::string>* >
	 (a, new std::map<int, std::string>()));
    }
  }

  void InOutParameters::addArrayInt(const char* array) { 

    std::map<const char*, std::map<int, int>*, ltstr>::iterator i = 
      intArrays_.find(array);
    if (i == intArrays_.end()) {
      char* a = (char*)std::calloc(std::strlen(array)+1,sizeof(char));
      std::strncpy(a, array, std::strlen(array));
      DEBUGL(tbxParametersAdd_D, "adding array " << array);
      intArrays_.insert
	(std::pair<const char*, std::map<int, int>* >
	 (a, new std::map<int, int>()));
    }
  }

  void InOutParameters::addArrayBool(const char* array) { 

    std::map<const char*, std::map<int, bool>*, ltstr>::iterator i = 
      boolArrays_.find(array);
    if (i == boolArrays_.end()) {
      char* a = (char*)std::calloc(std::strlen(array)+1,sizeof(char));
      std::strncpy(a, array, std::strlen(array));
      DEBUGL(tbxParametersAdd_D, "adding array " << array);
      boolArrays_.insert
	(std::pair<const char*, std::map<int, bool>* >
	 (a, new std::map<int, bool>()));
    }
  }

  void InOutParameters::addArrayDouble(const char* array, const int number, 
				       const double value) {

    // find the correct array
    std::map<const char*, std::map<int, double>*, ltstr>::iterator i = 
      doubleArrays_.find(array);
    if (i == doubleArrays_.end())
      throw conceptsException(MissingParameter(array));

    // change or add the entry
    std::map<int, double>::iterator j = i->second->find(number);
    if (j == i->second->end()) {
      DEBUGL(tbxParametersAdd_D, "adding " << number << " = " 
	     << value << " to array \"" << array << "\"");
      i->second->insert(std::pair<int, double>(number,value));
    } else
      j->second = value;
  }

  void InOutParameters::addArrayString(const char* array, const int number, 
				       const char* value) {

    // find the correct array
    std::map<const char*, std::map<int, std::string>*, ltstr>::iterator i = 
      stringArrays_.find(array);
    if (i == stringArrays_.end())
      throw conceptsException(MissingParameter(array));

    // change or add the entry
    std::map<int, std::string>::iterator j = i->second->find(number);
    if (j == i->second->end()) {
      DEBUGL(tbxParametersAdd_D, "adding " << number << " = " 
	     << value << " to array \"" << array << "\"");
      i->second->insert(std::pair<int, std::string>(number,value));
    } else
      j->second = value;
  }

  void InOutParameters::addArrayInt(const char* array, const int number, 
				    const int value) {

    // find the correct array
    std::map<const char*, std::map<int, int>*, ltstr>::iterator i = 
      intArrays_.find(array);
    if (i == intArrays_.end())
      throw conceptsException(MissingParameter(array));

    // change or add the entry
    std::map<int, int>::iterator j = i->second->find(number);
    if (j == i->second->end()) {
      DEBUGL(tbxParametersAdd_D, "adding " << number << " = " 
	     << value << " to array \"" << array << "\"");
      i->second->insert(std::pair<int, int>(number,value));
    } else
      j->second = value;
  }

  void InOutParameters::addArrayBool(const char* array, const int number, 
				     const int value) {

    // find the correct array
    std::map<const char*, std::map<int, bool>*, ltstr>::iterator i = 
      boolArrays_.find(array);
    if (i == boolArrays_.end())
      throw conceptsException(MissingParameter(array));

    // change or add the entry
    std::map<int, bool>::iterator j = i->second->find(number);
    if (j == i->second->end()) {
      DEBUGL(tbxParametersAdd_D, "adding " << number << " = " 
	     << value << " to array \"" << array << "\"");
      i->second->insert(std::pair<int, bool>(number,value));
    } else
      j->second = value;
  }

  double InOutParameters::getArrayDouble(const char* array, 
					 const int number) const {

    std::map<const char*, std::map<int, double>*, ltstr>::const_iterator i = 
      doubleArrays_.find(array);
    if (i == doubleArrays_.end()) {
      throw conceptsException(MissingParameter(array));
    } else {
      std::map<int, double>::const_iterator j = i->second->find(number);
      if (j == i->second->end()) {
	std::stringstream error;
	error << number << " in array " << array << std::ends;
	throw conceptsException(MissingParameter(error.str()));
      } else
	return j->second;
    }
  }

  int InOutParameters::getArrayInt(const char* array, 
				   const int number) const {

    std::map<const char*, std::map<int, int>*, ltstr>::const_iterator i = 
      intArrays_.find(array);
    if (i == intArrays_.end()) {
      throw conceptsException(MissingParameter(array));
    } else {
      std::map<int, int>::const_iterator j = i->second->find(number);
      if (j == i->second->end()) {
	std::stringstream error;
	error << number << " in array " << array << std::ends;
	throw conceptsException(MissingParameter(error.str()));
      } else
	return j->second;
    }
  }

  bool InOutParameters::getArrayBool(const char* array, 
				     const int number) const {

    std::map<const char*, std::map<int, bool>*, ltstr>::const_iterator i = 
      boolArrays_.find(array);
    if (i == boolArrays_.end()) {
      throw conceptsException(MissingParameter(array));
    } else {
      std::map<int, bool>::const_iterator j = i->second->find(number);
      if (j == i->second->end()) {
	std::stringstream error;
	error << number << " in array " << array << std::ends;
	throw conceptsException(MissingParameter(error.str()));
      } else
	return j->second;
    }
  }

  std::string InOutParameters::getArrayString(const char* array, 
					 const int number) const {

    std::map<const char*, std::map<int, std::string>*,
      ltstr>::const_iterator i = 
      stringArrays_.find(array);
    if (i == stringArrays_.end()) {
      throw conceptsException(MissingParameter(array));
    } else {
      std::map<int, std::string>::const_iterator j = i->second->find(number);
      if (j == i->second->end()) {
	std::stringstream error;
	error << number << " in array " << array << std::ends;
	throw conceptsException(MissingParameter(error.str()));
      } else
	return j->second;
    }
  }

  const std::map<int, double>& InOutParameters::getMapDouble
  (const char* array) const {
    std::map<const char*, std::map<int, double>*, ltstr>::const_iterator i = 
      doubleArrays_.find(array);
    if (i == doubleArrays_.end()) {
      throw conceptsException(MissingParameter(array));
    } else {
      return *(i->second);
    }
  }

  const std::map<int, std::string>& InOutParameters::getMapString
  (const char* array) const {
    std::map<const char*, std::map<int, std::string>*,
      ltstr>::const_iterator i = 
      stringArrays_.find(array);
    if (i == stringArrays_.end()) {
      throw conceptsException(MissingParameter(array));
    } else {
      return *(i->second);
    }
  }

  const std::map<int, int>& InOutParameters::getMapInt
  (const char* array) const {
    std::map<const char*, std::map<int, int>*, ltstr>::const_iterator i = 
      intArrays_.find(array);
    if (i == intArrays_.end()) {
      throw conceptsException(MissingParameter(array));
    } else {
      return *(i->second);
    }
  }

  const std::map<int, bool>& InOutParameters::getMapBool
  (const char* array) const {
    std::map<const char*, std::map<int, bool>*, ltstr>::const_iterator i = 
      boolArrays_.find(array);
    if (i == boolArrays_.end()) {
      throw conceptsException(MissingParameter(array));
    } else {
      return *(i->second);
    }
  }

  std::ostream& InOutParameters::info(std::ostream& os) const {

    // normal hashes
    for (std::map<const char*, std::string, ltstr>::const_iterator 
	   i = string_.begin(); i != string_.end(); ++i)
      os << "string\t" << i->first << "\t\"" << i->second << "\"" << std::endl;
    for (std::map<const char*, int, ltstr>::const_iterator 
	   i = int_.begin(); i != int_.end(); ++i)
      os << "int\t" << i->first << '\t' << i->second << std::endl;
    for (std::map<const char*, bool, ltstr>::const_iterator 
	   i = bool_.begin(); i != bool_.end(); ++i)
      os << "bool\t" << i->first << '\t' << (i->second ? "true" : "false") 
	 << std::endl;
    for (std::map<const char*, double, ltstr>::const_iterator 
	   i = double_.begin(); i != double_.end(); ++i)
      os << "double\t" << i->first << '\t' << i->second << std::endl;

    // the arrays
    for (std::map<const char*, std::map<int, double>*, ltstr>::const_iterator 
	   i = doubleArrays_.begin();
	 i != doubleArrays_.end(); ++i) {
      os << "array double " << i->first << " {" << std::endl;
      for (std::map<int, double>::const_iterator j = i->second->begin();
	   j != i->second->end(); ++j)
	os << '\t' << j->first << "\t" << j->second << std::endl;
      os << "}" << std::endl;
    }
    for (std::map<const char*, std::map<int, std::string>*, 
	   ltstr>::const_iterator i = stringArrays_.begin();
	 i != stringArrays_.end(); ++i) {
      os << "array string " << i->first << " {" << std::endl;
      for (std::map<int, std::string>::const_iterator j = i->second->begin();
	   j != i->second->end(); ++j)
	os << '\t' << j->first << "\t\"" << j->second << "\"" << std::endl;
      os << "}" << std::endl;
    }
    for (std::map<const char*, std::map<int, int>*, ltstr>::const_iterator i = 
	   intArrays_.begin(); 
	 i != intArrays_.end(); ++i) {
      os << "array int " << i->first << " {" << std::endl;
      for (std::map<int, int>::const_iterator j = i->second->begin();
	   j != i->second->end(); ++j)
	os << '\t' << j->first << "\t" << j->second << std::endl;
      os << "}" << std::endl;
    }
    for (std::map<const char*, std::map<int, bool>*, ltstr>::const_iterator
	   i = boolArrays_.begin(); 
	 i != boolArrays_.end(); ++i) {
      os << "array bool " << i->first << " {" << std::endl;
      for (std::map<int, bool>::const_iterator j = i->second->begin();
	   j != i->second->end(); ++j)
	os << '\t' << j->first << "\t" << (j->second ? "true" : "false") 
	   << std::endl;
      os << "}" << std::endl;
    }

    return os;
  }

  // *********************************************************** InputParser **

  InputParser::InputParser(bool quiet) : current_buffer_(0), quiet_(quiet) {

    // adding standard fields to input
    input_.addString("title", 0);
    input_.addString("author", 0);
    input_.addString("comment", 0);
  }

  InputParser::~InputParser() {

    if (current_buffer_)
      yy_delete_buffer(current_buffer_);
  }

  void InputParser::parse() {

    conceptsAssert(current_buffer_, Assertion());
    yyparse(&input_);
  }

  void InputParser::parse(std::string file) {

    std::FILE* inputFILE = std::fopen(file.c_str(), "r");
    conceptsAssert(inputFILE, concepts::Assertion());
    current_buffer_ = yy_create_buffer(inputFILE, 16384);
    yy_switch_to_buffer(current_buffer_);
    yyparse(&input_);
    yy_delete_buffer(current_buffer_);
    current_buffer_ = 0;
    std::fclose(inputFILE);
  }

  std::ostream& InputParser::info(std::ostream& os) const {
    if (!quiet_) {
      // information on system
      struct utsname *uts = (struct utsname*)std::malloc
	(sizeof(struct utsname));
      uname(uts);

      // get time
      const time_t tme = time(0);
      os << "/* time, date:\t" << ctime(&tme)
	 << " * username:\t" << ( (getlogin()==0) ? "<unknown>" : getlogin() )
	 << std::endl
	 << " * sysname:\t" << uts->sysname << std::endl
	 << " * nodename:\t" << uts->nodename << std::endl
	 << " * release:\t" << uts->release << std::endl
	 << " * version:\t" << uts->release << std::endl
	 << " * machine:\t" << uts->machine << std::endl;
      os << " */" << std::endl;
      std::free(uts);
    }

    os << input_ << "end" << std::endl 
       << "// output starts here" << std::endl
       << output_;
    return os;
  }

} // namespace
