/* write table of results
 */

#include <cstring>
#include "resultsTable.hh"
#include "inputOutput.hh"
#include "basics/debug.hh"

// debugging
#define ResultsTableDestr_D 0
#define ResultsTableAddMap_D 0

namespace concepts {

  // ********************************************************** ResultsTable **

  template<>
  struct ResultsTable::tableDelimTraits<ResultsTable::SCREEN> {
    static inline std::string start() { return ""; }
    static inline std::string starthl() { return ""; }
    static inline std::string starttl() { return ""; }
    static inline std::string hllf() { return "\n"; }
    static inline std::string tllf() { return hllf(); }
    static inline std::string hltd() { return "\t"; }
    static inline std::string tltd() { return hltd(); }
    static inline std::string end() { return ""; }
  };

  template<>
  struct ResultsTable::tableDelimTraits<ResultsTable::GNUPLOT> {
    static inline std::string start() { return ""; }
    static inline std::string starthl() { return "# "; }
    static inline std::string starttl() { return ""; }
    static inline std::string hllf() { return "\n"; }
    static inline std::string tllf() { return hllf(); }
    static inline std::string hltd() { return "\t"; }
    static inline std::string tltd() { return hltd(); }
    static inline std::string end() { return ""; }
  };

  ResultsTable::ResultsTable() : leader_(false), typeL_(DOUBLE), nameL_(0) {}

  ResultsTable::~ResultsTable() {
    std::map<const char*, const std::map<int, double>*,
      ltstr>::iterator iadouble;
    while ( (iadouble = doubleMaps_.begin()) != doubleMaps_.end() ) {
      const char* c = iadouble->first;
      DEBUGL(ResultsTableDestr_D, "removing string " << c);
      doubleMaps_.erase(iadouble);
      std::free((char*)c);
    }

    std::map<const char*, const std::map<int, std::string>*,
      ltstr>::iterator iastring;
    while ( (iastring = stringMaps_.begin()) != stringMaps_.end() ) {
      const char* c = iastring->first;
      DEBUGL(ResultsTableDestr_D, "removing string " << c);
      stringMaps_.erase(iastring);
      std::free((char*)c);
    }

    std::map<const char*, const std::map<int, int>*, ltstr>::iterator iaint;
    while ( (iaint = intMaps_.begin()) != intMaps_.end() ) {
      const char* c = iaint->first;
      DEBUGL(ResultsTableDestr_D, "removing string " << c);
      intMaps_.erase(iaint);
      std::free((char*)c);
    }

    std::map<const char*, const std::map<int, bool>*, ltstr>::iterator iabool;
    while ( (iabool = boolMaps_.begin()) != boolMaps_.end() ) {
      const char* c = iabool->first;
      DEBUGL(ResultsTableDestr_D, "removing string " << c);
      boolMaps_.erase(iabool);
      std::free((char*)c);
    }
    std::free(nameL_);
  }

  void ResultsTable::addMap(enum mapTypes type, const char* name,
			    const InOutParameters& holder) {
    if (! leader_) {
      DEBUGL(ResultsTableAddMap_D, "new leader = " << name);
      leader_ = true;
      typeL_ = type;
      nameL_ = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
      std::strncpy(nameL_, name, std::strlen(name));
    }
    switch(type) {
    case DOUBLE: {
      std::map<const char*, const std::map<int, double>*, ltstr>::iterator i = 
	doubleMaps_.find(name);
      if (i == doubleMaps_.end()) {
	char* a = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
	std::strncpy(a, name, std::strlen(name));
	doubleMaps_.insert
	  (std::pair<const char*, const std::map<int, double>* >
	   (a, &(holder.getMapDouble(name))));
	DEBUGL(ResultsTableAddMap_D, "added new map " << name);
      }}
      break;
    case STRING: {
      std::map<const char*, const std::map<int, std::string>*,
	ltstr>::iterator j = stringMaps_.find(name);
      if (j == stringMaps_.end()) {
	char* a = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
	std::strncpy(a, name, std::strlen(name));
	stringMaps_.insert
	  (std::pair<const char*, const std::map<int, std::string>* >
	   (a, &(holder.getMapString(name))));
	DEBUGL(ResultsTableAddMap_D, "added new map " << name);
      }}
      break;
    case INT: {
      std::map<const char*, const std::map<int, int>*, ltstr>::iterator i = 
	intMaps_.find(name);
      if (i == intMaps_.end()) {
	char* a = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
	std::strncpy(a, name, std::strlen(name));
	intMaps_.insert
	  (std::pair<const char*, const std::map<int, int>* >
	   (a, &(holder.getMapInt(name))));
	DEBUGL(ResultsTableAddMap_D, "added new map " << name);
      }}
      break;
    case BOOL: {
      std::map<const char*, const std::map<int, bool>*, ltstr>::iterator i = 
	boolMaps_.find(name);
      if (i == boolMaps_.end()) {
	char* a = (char*)std::calloc(std::strlen(name)+1,sizeof(char));
	std::strncpy(a, name, std::strlen(name));
	boolMaps_.insert
	  (std::pair<const char*, const std::map<int, bool>* >
	   (a, &(holder.getMapBool(name))));
	DEBUGL(ResultsTableAddMap_D, "added new map " << name);
      }}
      break;
    }
  }

  template<enum ResultsTable::tableTypes t>
  void ResultsTable::print(std::ostream& os) const {
    print<t>(os, typeL_, nameL_);
  }

  template void ResultsTable::print<ResultsTable::GNUPLOT>
  (std::ostream& os) const;

  template<enum ResultsTable::tableTypes t>
  void ResultsTable::print(std::ostream& os,
			   enum mapTypes typeL, const char* nameL) const {
    os << tableDelimTraits<t>::start();
    switch (typeL) {
    case DOUBLE: {
      std::map<const char*, const std::map<int, double>*,
	ltstr>::const_iterator i = doubleMaps_.find(nameL);
      if (i == doubleMaps_.end()) {
	throw conceptsException(MissingParameter(nameL));
      } else {
	// a head line
	os << tableDelimTraits<t>::starthl() << nameL;
	headline_<t>(os);
	os << tableDelimTraits<t>::hllf();

	// start the loop over all indices of the leader
	for (std::map<int, double>::const_iterator j = i->second->begin();
	     j != i->second->end(); ++j) {
	  os << tableDelimTraits<t>::starttl() << j->first;
	  tableline_<t>(os, j->first);
	  os << tableDelimTraits<t>::tllf();
	} // for j
      }}
      break;
    case STRING: {
      std::map<const char*, const std::map<int, std::string>*,
	ltstr>::const_iterator i = stringMaps_.find(nameL);
      if (i == stringMaps_.end()) {
	throw conceptsException(MissingParameter(nameL));
      } else {
	// a head line
	os << tableDelimTraits<t>::starthl() << nameL;
	headline_<t>(os);
	os << tableDelimTraits<t>::hllf();

	// start the loop over all indices of the leader
	for (std::map<int, std::string>::const_iterator j = i->second->begin();
	     j != i->second->end(); ++j) {
	  os << tableDelimTraits<t>::starttl() << j->first;
	  tableline_<t>(os, j->first);
	  os << tableDelimTraits<t>::tllf();
	} // for j
      }}
      break;
    case INT: {
      std::map<const char*, const std::map<int, int>*,
	ltstr>::const_iterator i = intMaps_.find(nameL);
      if (i == intMaps_.end()) {
	throw conceptsException(MissingParameter(nameL));
      } else {
	// a head line
	os << tableDelimTraits<t>::starthl() << nameL;
	headline_<t>(os);
	os << tableDelimTraits<t>::hllf();

	// start the loop over all indices of the leader
	for (std::map<int, int>::const_iterator j = i->second->begin();
	     j != i->second->end(); ++j) {
	  os << tableDelimTraits<t>::starttl() << j->first;
	  tableline_<t>(os, j->first);
	  os << tableDelimTraits<t>::tllf();
	} // for j
      }}
      break;
    case BOOL: {
      std::map<const char*, const std::map<int, bool>*,
	ltstr>::const_iterator i = boolMaps_.find(nameL);
      if (i == boolMaps_.end()) {
	throw conceptsException(MissingParameter(nameL));
      } else {
	// a head line
	os << tableDelimTraits<t>::starthl() << nameL;
	headline_<t>(os);
	os << tableDelimTraits<t>::hllf();

	// start the loop over all indices of the leader
	for (std::map<int, bool>::const_iterator j = i->second->begin();
	     j != i->second->end(); ++j) {
	  os << tableDelimTraits<t>::starttl() << j->first;
	  tableline_<t>(os, j->first);
	  os << tableDelimTraits<t>::tllf();
	} // for j
      }}
      break;
    }
  }

  std::ostream& ResultsTable::info(std::ostream& os) const {
    os << "ResultsTable(" << std::endl;
    print<SCREEN>(os, typeL_, nameL_);
    os << ")";
    return os;
  }

  template<enum ResultsTable::tableTypes t>
  void ResultsTable::headline_(std::ostream& os) const {
    // loop over all double maps
    for (std::map<const char*, const std::map<int, double>*,
	   ltstr>::const_iterator k = doubleMaps_.begin();
	 k != doubleMaps_.end(); ++k) {
      os << tableDelimTraits<t>::hltd() << k->first;
    } // for k
    // loop over all string maps
    for (std::map<const char*, const std::map<int, std::string>*,
	   ltstr>::const_iterator k = stringMaps_.begin();
	 k != stringMaps_.end(); ++k) {
      os << tableDelimTraits<t>::hltd() << k->first;
    } // for k
    // loop over all int maps
    for (std::map<const char*, const std::map<int, int>*,
	   ltstr>::const_iterator k = intMaps_.begin();
	 k != intMaps_.end(); ++k) {
      os << tableDelimTraits<t>::hltd() << k->first;
    } // for k
    // loop over all bool maps
    for (std::map<const char*, const std::map<int, bool>*,
	   ltstr>::const_iterator k = boolMaps_.begin();
	 k != boolMaps_.end(); ++k) {
      os << tableDelimTraits<t>::hltd() << k->first;
    } // for k
  }

  template<enum ResultsTable::tableTypes t>
  void ResultsTable::tableline_(std::ostream& os, const int idx) const {
    // loop over all double maps
    for (std::map<const char*, const std::map<int, double>*,
	   ltstr>::const_iterator k = doubleMaps_.begin();
	 k != doubleMaps_.end(); ++k) {
      // print the entry from the chosen map which matches the
      // index of the leader
      os << tableDelimTraits<t>::tltd() << k->second->find(idx)->second;
    } // for k
    // loop over all string maps
    for (std::map<const char*, const std::map<int, std::string>*,
	   ltstr>::const_iterator k = stringMaps_.begin();
	 k != stringMaps_.end(); ++k) {
      os << tableDelimTraits<t>::tltd() << k->second->find(idx)->second;
    } // for k
    // loop over all int maps
    for (std::map<const char*, const std::map<int, int>*,
	   ltstr>::const_iterator k = intMaps_.begin();
	 k != intMaps_.end(); ++k) {
      os << tableDelimTraits<t>::tltd() << k->second->find(idx)->second;
    } // for k
    // loop over all bool maps
    for (std::map<const char*, const std::map<int, bool>*,
	   ltstr>::const_iterator k = boolMaps_.begin();
	 k != boolMaps_.end(); ++k) {
      os << tableDelimTraits<t>::tltd() << k->second->find(idx)->second;
    } // for k
  }

} // namespace concepts
