#include <sstream>
#include "bramblePasciak.hh"
#include "basics/exceptions.hh"

// debugging

#define BramblePasciakAppl_D 0

namespace vectorial {

  // ******************************************************** BramblePasciak **

  void BramblePasciak::operator()(const concepts::Function<Real>& fncY,
                                  concepts::Function<Real>& fncX) {
    const concepts::Vector<Real>* fncYA;
    if ((fncYA = dynamic_cast<const concepts::Vector<Real>*>(&fncY))) {
      concepts::Vector<Real>* fncXA;
      if ((fncXA = dynamic_cast<concepts::Vector<Real>*>(&fncX))) {
        (*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException(concepts::MissingFeature
                            ("function not supported"));
  }

  void BramblePasciak::operator()(const concepts::Vector<Real>& fncY,
                                  concepts::Vector<Real>& fncX) {
    // subspaces
    const concepts::Space<Real>& spcX = *(spc_.get(0));
    const concepts::Space<Real>& spcY = *(spc_.get(1));

    // subvectors
    concepts::Vector<Real> x(spcX, (Real*)(fncX));
    concepts::Vector<Real> y(spcY, (Real*)(fncX)+spcX.dim());
    const concepts::Vector<Real> f(spcX, (Real*)(fncY));
    const concepts::Vector<Real> g(spcY, (Real*)(fncY)+spcX.dim());

    // temporary vectors
    concepts::Vector<Real> tempX(x);
    concepts::Vector<Real> tempY(y);

    // residuals
    concepts::Vector<Real> rx(x);
    concepts::Vector<Real> ry(y);

    // directions
    concepts::Vector<Real> px(x);
    concepts::Vector<Real> py(y);

    // calculate residuals
    // rx = px = f - A x - B y
    A_(x, tempX); // tempX = A x
    B_(y, rx); // rx = B y
    tempX += rx; // tempX = A x + B y
    rx = f;
    rx -= tempX; // rx = f - A x - B y
    px = rx;

    // ry = px = B^T(W rx + x) - g - C y
    W_(rx, tempX); // tempX = W rx
    tempX += x; // tempX = W rx + x
    Bt_(tempX, ry); // ry = B^T(W rx + x)
    C_(y, py); // py = C y;
    ry -= py; // ry = B^T(W rx + x) - C y
    ry -= g; // ry = B^T(W rx + x) - C y - g
    py = ry;

    Real maxepsY = maxeps_;
    if (relres_) {
      // scale maxepsY with rhs of transformed system
      W_(f, tempX); // tempX = W f;
      Bt_(tempX, tempY); // tempY = B W f
      tempY -= g; // tempY = B W f - g
      maxepsY *= f.l2_2() + tempY.l2_2();
    }

    eps_ = rx.l2_2() + ry.l2_2();
    it_  = 0;

    // helper vectors
    concepts::Vector<Real> v1(x); // v1 = W(A W - 1) px
    concepts::Vector<Real> v2(x); // v2 = W A v1
    concepts::Vector<Real> v3(x); // v3 = W(A W - 1)B py
    concepts::Vector<Real> v4(y); // v4 = (C + B^T W B) py

    for(it_ = 0; !(eps_ < maxepsY || (maxit_ > 0 && it_ > maxit_)); ++it_) {

      // v3 = W(A W - 1)B py
      B_(py, v3); // v3 = B py
      W_(v3, v1); // v1 = W B py
      A_(v1, v2); // v2 = A W B py
      v2 -= v3; // v2 = (A W - 1)B py
      W_(v2, v3); // v3 = W(A W - 1)B py

      // v4 = (C + B^T W B) py
      Bt_(v1, tempY); // tempY = B^T W B py
      C_(py, v4); // v4 = C py
      v4 += tempY; // v3 = (C + B^T W B) py
    
      // v1 = W(A W - 1) px
      W_(px, v1); // v1 = W px
      A_(v1, tempX); // tempX = A W px
      tempX -= px; // tempX = (A W - 1) px
      W_(tempX, v1); // v1 = W(A W - 1) px

      // v2 = W A v1
      A_(v1, tempX); // tempX = A v1
      W_(tempX, v2); // v2 = W A v1

      // update position (x,y)
      Real c1 = v1*rx + ry*py;
      Real c2 = v2*px + 2*(v3*px) + v4*py;
      Real alpha = c1/c2;
      W_(px, tempX); // tempX = W px
      x.add(tempX, alpha); // x += alpha W px
      y.add(py, alpha); // y += alpha py

      // calculate new residual (rx,ry)
      // rx = f - A x - D y
      A_(x, tempX); // tempX = A x
      B_(y, rx); // rx = B y
      tempX += rx; // tempX = A x + B y
      rx = f;
      rx -= tempX; // rx = f - A x - B y

      // ry = B^T(W rx + x) - g - C y
      W_(rx, tempX); // tempX = W rx
      tempX += x; // tempX = W rx + x
      Bt_(tempX, ry); // ry = B^T(W px + x)
      C_(y, tempY); // tempY = C y;
      ry -= tempY; // ry = B^T(W px + x) - C y
      ry -= g; // ry = B^T(W px + x) - C y - g

      // calculate new direction (px,py)
      Bt_(v1, tempY);
      Real c3 = v2*rx + tempY*ry + v3*rx + v4*ry;
      Real beta = c3/c2;
      px *= -beta; px += rx; // px <- rx - beta px
      py *= -beta; py += ry; // py <- ry - beta py

      eps_ = rx.l2_2() + ry.l2_2();
      DEBUGL(BramblePasciakAppl_D, "maxepsY = " << maxepsY << ", it = " << it_
             << ", eps = " << eps_ << ", maxeps_ = " << maxeps_);
    }

    if (!(eps_ < maxepsY)) {
      DEBUGL(BramblePasciakAppl_D, "preparing NoConvergence exception");
      std::stringstream errMsg;
      errMsg << "it: " << it_ << ", eps: " << eps_
             << ", wanted: " << maxepsY << std::ends;
      throw conceptsException(concepts::NoConvergence(errMsg.str()));
    }
    DEBUGL(BramblePasciakAppl_D, "succesfully converged!");
  }

  std::ostream& BramblePasciak::info(std::ostream& os) const {
    return os << "BramblePasciak(A = " << A_ << ", B = " << B_ << ", Bt = "
              << Bt_ << ", C = " << C_ << "W = " << W_ << ", eps = " << eps_
              << ", it = " << it_ << ", relres = " << relres_ << ')';
  }

} // namespace vectorial
