#ifndef cgUzawa_hh
#define cgUzawa_hh

#include <string>

#include "basics/typedefs.hh"
#include "operator/compositions.hh"
#include "vectorial/vectorial.hh"

namespace vectorial {

  using concepts::Real;

  // *************************************************************** CGUzawa **

  /** Uzawa algorithm with conjugate directions for generalized saddle point
      problems.
      \f[ \left(\begin{array}{lr}A&B\\B^{\top}&-C\end{array}\right)
      \left(\begin{array}{c}U\\P\end{array}\right) = 
      \left(\begin{array}{c}L_u\\L_p\end{array}\right) \f]
      A has to be symmetric and positive definite, C has to be symmetric and
      positive semidefinite, and B has to satisfy
      the discrete inf-sup-condition.

      The main idea is to eliminate \f$U\f$ from the above equation, resulting
      in the equation
      \f$(B^{\top} A^{-1} B + C)P = B^{\top} A^{-1}L_u - L_p\f$. This reduced
      system is solved with conjugate gradients. The action of A<sup>-1</sup>
      is a suitable solver that has to be provided at construction time.
      Then, \f$U\f$ is obtained from \f$A \ U = L_u - B \ P\f$. See [1] for
      more details.

      The class is an operator on a
      vectorial::Space consisting of two spaces, say spc1 and spc2. A is
      spc1->spc1, B is spc2->spc1 and C is spc2->spc2. The entire vectorial
      space has to be provided at construction time.

      Constructing an object of this class does not solve the given system.
      Use the application operator to solve the system. If you want to
      specify a starting vector for the cg iterations, set \c fncX
      before calling the application operator to this starting value.
      \c fncX also holds the result after the solve.

      The application operator throws \c NoConvergence if the desired
      residual \c maxeps is not reached within the given number of
      iterations \c maxit.
      @see D. Braess: Finite Elemente: Theorie, schnelle L&ouml;ser und Anwendungen in der Elastitzit&auml;tstheorie. Springer, Berlin, 1997.
  */
  class CGUzawa : public concepts::Operator<Real> {
  public:
    /** Constructor.
        @param A Upper left submatrix
        @param B Upper right submatrix
        @param Bt Lower left submatrix
        @param C Lower right submatrix
        @param Ai Solver for A
        @param spc Entire vectorial space
        @param maxeps Maximal residual
        @param maxit Maximal number of iterations
        @param relres Relative residual
    */
    CGUzawa(concepts::Operator<Real>& A, concepts::Operator<Real>& B,
            concepts::Operator<Real>& Bt, concepts::Operator<Real>& C,
            concepts::Operator<Real>& Ai, Space<Real>& spc,
            Real maxeps, int maxit = 0, uint relres = false) :
      A_(A), B_(B), Bt_(Bt), C_(C), Ai_(Ai), spc_(spc), maxeps_(maxeps*maxeps),
      maxit_(maxit), eps_(1.0), it_(0), relres_(relres) {
      conceptsAssert(spc.vdim() == 2, concepts::Assertion());
      conceptsAssert(&(A.spaceX()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(A.spaceY()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(Ai.spaceX()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(Ai.spaceY()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(B.spaceX()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(B.spaceY()) == spc.get(1), concepts::Assertion());
      conceptsAssert(&(Bt.spaceX()) == spc.get(1), concepts::Assertion());
      conceptsAssert(&(Bt.spaceY()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(C.spaceX()) == spc.get(1), concepts::Assertion());
      conceptsAssert(&(C.spaceY()) == spc.get(1), concepts::Assertion());
    }

    virtual void operator()(const concepts::Function<Real>& fncY,
                            concepts::Function<Real>& fncX);
    void operator()(const concepts::Vector<Real>& fncY,
                    concepts::Vector<Real>& fncX);

    virtual const concepts::Space<Real>& spaceX() const { return spc_; }
    virtual const concepts::Space<Real>& spaceY() const { return spc_; }

    /** Returns the number of (outer) iterations. Calling this method makes
        only sense after a linear system has been solved.
    */
    uint iterations() const { return it_; }

    /** Returns the residual. Calling this method makes
        only sense after a linear system has been solved.
    */
    Real epsilon() const { return eps_; }
  protected:
    std::ostream& info(std::ostream& os) const;
  private:
    /// Upper left submatrix
    concepts::Operator<Real>& A_;
    /// Upper right submatrix
    concepts::Operator<Real>& B_;
    /// Lower left submatrix
    concepts::Operator<Real>& Bt_;
    /// Lower right submatrix
    concepts::Operator<Real>& C_;
    /// Solver for A
    concepts::Operator<Real>& Ai_;
    /// Entire space
    Space<Real>& spc_;
    /// Convergence criterion
    Real maxeps_;
    /// Maximal number of iterations until abortion
    uint maxit_;
    /// Current residual
    Real eps_;
    /// Number of iterations
    uint it_;
    /// false: absolute residual, true: relative residual
    bool relres_;
  };

} // namespace vectorial

#endif // cgUzawa_hh
